# ===========================================================================
# HTSDT-Bias-Rank models: High threshold signal detection model with ranking
#                         and with guessing with position bias parameters
# Date of creation:	July,      2013
# Date of last update:	September, 2013
#
# Author: Siegfried Macho
# ===========================================================================
library(statmod)
library(npmlreg)

HTSDT.Bias.Rk.Tol <- 1e-12
MINWEIGHT         <- 1e-40

#===============================================================================
# 0. COMPUTATION OF QUADRATURE POINTS
#===============================================================================

#-------------------------------------------------------------------------------
# Function for computing the Gauss-Hermite points and weights
# Input:	1. Number of points
#           2. minimal weights
#
# Output: 	Matrix with three columns:    1. Points
#                                         2. Weights of Gauss-Hermite
#                                         3. Weights of adaptive Gauss-Hermite 
#
# Comment:  The points and weights are with respect to the Gaussian Kernel N(0, 1)
#-------------------------------------------------------------------------------
Generate.GH <- function(npt, minweight = MINWEIGHT)
{
      GH.Structure  <- gqz(numnodes = npt, minweight)   # Points and weights for Gauss-Hermite Quadrature
      cbind(GH.Structure, GH.Structure[,2] * exp(GH.Structure[,1]^2/2))	 # Weights for adaptive Gauss-Hermite Quadrature
}

#===============================================================================
# 1. HELP-FUNCTIONS
#===============================================================================

# ------------------------------------------------------------------------------
# Computation positions of parameters
# Input:  	 
#        cfg = configuration list
#     Par.Id = Identifier of the type of parameter whose positions are computed
#              "Prob"  = Positions of probability parameters
#              "Sigma" = Positions of standard deviations
#              "Bias"  = Positions of bias parameters
# Output:
#     Vector of the position of parameters
# ------------------------------------------------------------------------------
HTSDT.Bias.pos.par <- function(cfg, Par.Id = "Prob")
{
      n.rk <- length(cfg$j.rg)                        # Number of ranking models
      n.rt <- cfg$n.sdt                               # Number of rating  models

      if(Par.Id == "Prob")    # Position of probability parameters
      {
            if(n.rk > 1)
                  pos.prb <- c(1, 2, c(1, 2) + rep(cumsum(6 + cfg$k.rg[-n.rk]), each = 2))
            else  
                  pos.prb <- c(1, 2)
            
            if(cfg$rating)
                  return(c(pos.prb, 1:3 + rep(6*n.rk + sum(cfg$k.rg) + 5*(0:(n.rt-1)), each = 3)))
            return(pos.prb)
                              
      }      
      if(Par.Id == "Sigma")    # Position of sigma parameters
      {
            if(n.rk > 1)
                  pos.sigma <- c(4, 6, c(4, 6) + rep(cumsum(6 + cfg$k.rg[-n.rk]), each = 2))      # Position of sigma parameters
            else  
                  pos.sigma <- c(4, 6)
            
            if(cfg$rating)
                  return(c(pos.sigma, 5 + 6*n.rk + sum(cfg$k.rg) + 5*(0:(n.rt-1))))
            return(pos.sigma)
      }   
      if(Par.Id == "Bias")    # Position of bias parameters
      {
            pos.bias <- NULL
            ed       <- 0
            for(i in 1:n.rk)
            {
                  st <- ed + 7
                  ed <- st + cfg$k.rg[i] - 1
                  pos.bias <- c(pos.bias, st:ed)
            }
            return(pos.bias)
      }
      if(Par.Id == "Rank")
            return(1:(6*n.rk + sum(cfg$k.rg)))
        
      stop(paste("\nMIX.Bias.pos.par: Unknown Type Identifier: <", Par.Id, ">\n", sep = ""))
}

#===============================================================================
# 2. COMPUTATION OF PROBABILITIES
#===============================================================================

# ------------------------------------------------------------------------------
# Computation of probabilities from raw vector
# Input:	rawp  = vector of raw values for computing probabilities
#           fixed = binary vector indicates fixed values
#                   F = free, T = fixed
# Output: 	Vector of probabilities
#
# Comment:  The output vector contains one entry more than the input vectors
# ------------------------------------------------------------------------------
HTSDT.Bias.Rk.Cmp.ppars <- function(rawp, fixed)
{
      # Case 1: None fixed
      if(!any(fixed))
      {
            rawp[rawp > 700] <- 700             # Avoid overflow
            ppar <- exp(rawp) / (1.0 + sum(exp(rawp)))
            return(ppar)
      }
      # Case 2: All fixed
      if(all(fixed)) return(rawp)

      # Case 3: Some fixed
      wt <- 1.0 - sum(rawp[fixed])              # Sum raw probabilities
      rawp.free <- rawp[!fixed]
      rawp.free[rawp.free > 700] <- 700         # Avoid overflow
      
      rawp[!fixed] <- wt*(exp(rawp.free) / (1.0 + sum(exp(rawp.free))))
      return(rawp)
}


# ------------------------------------------------------------------------------
# Computation of probabilities in parameter vector
# Input:	parvec  = parameter vector
#           cfg     = configuration
#           fixed   = matrix of fixed constraints
#           ident   = matrix of idenity constraints
#
# Output: 	Parameter vector with raw probabilities replaced by computed probabilities
#
# Comments: - The routine assumes that all identity constraints involved in fixed 
#             constraints are set to fixed.
#           - It is asscumed that identity constraints are oriented from lower to
#             higher parameter numbers
# ------------------------------------------------------------------------------
HTSDT.Bias.Rk.Ppars <- function(parvec, cfg, fixed, ident)
{
      n.rk <- length(cfg$j.rg)                        # Number of ranking models
      n.rt <- cfg$n.sdt                               # Number of rating  models      

      if(is.null(fixed)) fix <- -1
      else               fix <- fixed[2,]
      
      if(is.null(ident)) id1 <- id2 <- -1
      else               
      {
            id1 <- ident[1,]
            id2 <- ident[2,]
      }

      # 1. Compute probabilities of Ranking model 1
      pos <- 1:2        # Position of probabilities of ranking model
      id2.t <- id2[!(id1 %in% pos)]             # Eliminate actual positions from identity constraints
      fx <- pos %in% fix
      fx <- fx | (pos %in% id2.t)
      parvec[pos] <- HTSDT.Bias.Rk.Cmp.ppars(parvec[pos], fx)
      prepos <- pos                       # Compile former positions in a vector
      
      # 2. Compute probabilities of other ranking models
      if(n.rk > 1)
      {
            for(i in 2:n.rk)
            {
                  pos   <- 1:2 + 6*(i-1) + sum(cfg$k.rg[1:(i-1)]) # Position of Recollection and Guessing in Models 2 - s
                  id2.t <- id2[!(id1 %in% pos)]   # Eliminate actual positions from identity constraints
                  fx <- pos %in% fix
                  fx.i  <- (pos %in% id2.t)
                  fx <- fx | fx.i
    
                  # Set identity constraints if necessary      
                  if(any(fx.i))
                  {
                        idp <- ident[, (ident[1,] %in% prepos) & (ident[2,] %in% pos), drop = F]
                        if(!is.null(idp))
                              parvec[idp[2,]] <- parvec[idp[1,]]               
                  }
                  parvec[pos] <- HTSDT.Bias.Rk.Cmp.ppars(parvec[pos], fx)
                  prepos <- c(prepos, pos)
            }
      }
      if(cfg$rating == F) return(parvec)
      
      # 3. Compute probabilities for the rating models
      for(i in 1:n.rt)
      {
            # 3.1 Compute recollection and guessing probabilities
            pos   <- 1:2 + 6*n.rk + sum(cfg$k.rg) + 5*(i-1)     # Position of recollection and guessing parameters
            id2.t <- id2[!(id1 %in% pos)]                       # Eliminate actual positions from identity constraints
            fx <- pos %in% fix
            fx.i  <- (pos %in% id2.t)
            fx <- fx | fx.i
    
            # Set identity constraints if necessary      
            if(any(fx.i))
            {
                  idp <- ident[, (ident[1,] %in% prepos) & (ident[2,] %in% pos), drop = F]
                  if(!is.null(idp))
                        parvec[idp[2,]] <- parvec[idp[1,]]               
            }
            parvec[pos] <- HTSDT.Bias.Rk.Cmp.ppars(parvec[pos], fx)
            prepos <- c(prepos, pos)

            # 3.2 Compute  bias probabilities
            pos   <- 3 + 6*n.rk + sum(cfg$k.rg) + 5*(i-1)  # Position of Bias parameters
            fx <- pos %in% fix
            fx <- fx | (pos %in% id2)
            parvec[pos] <- HTSDT.Bias.Rk.Cmp.ppars(parvec[pos], fx)
      }
      return(parvec)
}


#===============================================================================
# 3. MODEL FUNCTIONS OF THE RANKING PART OF THE RANKING-RATING MODELL
#===============================================================================

# ----------------------------------------------------------------------
# Computation of conditional ranking probability of the target stimulus
# given the strength value x[1] of the target variable multiplied by
# the quadrature weight x[2]
# Input:	x = vector with quadrature point x[1] and quadrature weight x[2]
#       s.pos = start distributions =  distributions of those random variables that 
#               are located in front of the actual value x[1]
#       e.pos = end distributions =  distributions of those random variables that 
#               are located behind the actual value x[1]
#       t.pos = position of the target distribution
#          mu = mean of non-target distributions
#       sigma = standard deviation of non-target distributions
#        bias = vector of position bias parameters
#
# Output: Density
# ----------------------------------------------------------------------
HTSDT.Bias.Rk.cond <- function(x, s.pos, e.pos, t.pos, mu, sigma, bias)
{
      p1 <- p2 <- 1.0
      if(!is.null(s.pos))
            p1 <- prod(1 - pnorm(x[1] + bias[t.pos] - bias[s.pos], mu, sigma))
      if(!is.null(e.pos))
            p2 <- prod(pnorm(x[1] + bias[t.pos] - bias[e.pos], mu, sigma))
      p1 * p2 * x[2]
}

# ----------------------------------------------------------------------
# Computation of the unconditional ranking probability of target stimulus
# Input: s.pos = start distributions =  distributions of those random variables that 
#                are located in front of the actual value x[1]
#        e.pos = end distributions =  distributions of those random variables that 
#                are located behind the actual value x[1]
#        t.pos = position of the target item
#    mu.target = mean of the target distribution
# sigma.target = standard deviation of the target distribution
#         bias = vector of position bias parameters
#           mu = mean of the non-target distribution
#        sigma = standard deviation of the non-target distribution
#
# Output: Probability of the configuration
# ----------------------------------------------------------------------
HTSDT.Bias.Rk  <- function(s.pos, e.pos, t.pos, mu.target, sigma.target, bias, mu = 0, sigma = 1)
{
      GH.Structure[,1] <- GH.Structure[,1] * sigma.target + mu.target
      sum(apply(GH.Structure, 1, HTSDT.Bias.Rk.cond, s.pos = s.pos, e.pos = e.pos, t.pos = t.pos, mu = mu, sigma = sigma, bias = bias))
}

# ------------------------------------------------------------------------------
# Computation of combinations from a set: Selection of m items from set x
# Input:     x = set from which to select
#            m = number of selected items
#        t.pos = position of target alternative
#    mu.target = mean of the target alternative
# sigma.target = stddev of the target alternative
#         bias = vetor of bias parameters
#           mu = mean of the non-target alternative
#        sigma = stddev of the non-target alternative
#
# Output: 	Computed probability
# ------------------------------------------------------------------------------
HTSDT.Bias.combn <- function(x, m, t.pos, mu.target, sigma.target, bias, mu, sigma)
{
      n <- length(x)
      e <- 0
      h <- m
      a <- 1:m
      i <- 2
      nmmp1 <- n - m + 1
      mp1 <- m + 1
      prb <- HTSDT.Bias.Rk(x[a], x[!(x %in% x[a])], t.pos, mu.target, sigma.target, bias, mu, sigma)
      while (a[1] != nmmp1)
      {
            if (e < n - h)
            {
                  h <- 1
                  e <- a[m]
                  j <- 1
            }
            else
            {
                  h <- h + 1
                  e <- a[mp1 - h]
                  j <- 1:h
            }
            a[m - h + j] <- e + j
            prb <- prb + HTSDT.Bias.Rk(x[a], x[!(x %in% x[a])], t.pos, mu.target, sigma.target, bias, mu, sigma)
            i <- i + 1
    }
    return(prb)
}

# ----------------------------------------------------------------------
# Computation of probability of target stimulus
# Input: j.rg  = Rank position of the target
#        k.rg  = Number of items presented
#        t.pos = position of the target item
#           mu = mean of the non-target distribution
#        sigma = standard deviation of the non-target distribution
#    mu.target = mean of the target distribution
# sigma.target = standard deviation of the target distribution
#         bias = vector of position bias parameters
#
# Output: Probability of the configuration
# ----------------------------------------------------------------------
HTSDT.Bias.Rk.1 <- function(j.rg, k.rg, t.pos, mu, sigma, mu.target, sigma.target, bias)
{
      if(j.rg == 1) 
      {
            s.pos <- NULL 
            e.pos <- (1:k.rg)[-t.pos]
            return(HTSDT.Bias.Rk(s.pos, e.pos, t.pos, mu.target, sigma.target, bias, mu, sigma))
      }
      if(j.rg == k.rg)
      {
            s.pos <- (1:k.rg)[-t.pos]
            e.pos <- NULL
            return(HTSDT.Bias.Rk(s.pos, e.pos, t.pos, mu.target, sigma.target, bias, mu, sigma))
      }
      x <- (1:k.rg)[-t.pos]
      
      return(HTSDT.Bias.combn(x, j.rg - 1, t.pos, mu.target, sigma.target, bias, mu, sigma))
}      

#-------------------------------------------------------------------------------
# Computation of ranking probabilities for one model
# Input:
#     parvec = parameter vector with the following entries
#              1. k.rg:       number of alternatives to be ranked
#              2. j.rg:       number of positions to be considered
#              3. t.pos:      position of target
#              4. probability of recollection
#              5. probabilitiy of pure guessing
#              6. mean of the new distribution
#              7. sd   of the new distribution
#              8. mean of the old distribution
#              9. sd   of the old distribution
#
# Output: Ranking probabilities
#-------------------------------------------------------------------------------
HTSDT.Bias.Rank.1 <- function(parvec)
{
      
      sdt.prob <- (1.0 - parvec[4] - parvec[5])
      if(sdt.prob > 0.0)
      {
            prb <- sapply(1:parvec[2], HTSDT.Bias.Rk.1, k.rg = parvec[1], t.pos = parvec[3], mu = parvec[6], sigma = parvec[7],
                          mu.target = parvec[8], sigma.target = parvec[9], bias = parvec[10:(9 + parvec[1])])      
            prb <- sdt.prob * prb
      }
      else
            prb <- rep(0.0, length = parvec[2])
      prb[1] <- prb[1] + parvec[4]
      prb    <- prb + parvec[5]/parvec[1]

      return(c(prb, 1.0 - sum(prb)))
}

#-------------------------------------------------------------------------------
# Computation of ranking probabilties for all ranking models model
#-------------------------------------------------------------------------------
HTSDT.Bias.RankOnly <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
      prb <- NULL
      n.rg <- length(cfg$k.rg)
      for(i in 1:n.rg)              # Loop of all models
      {
            # Extract relevant parameter vector
            st <- ifelse(i == 1, 1, 1 + 6*(i-1) + sum(cfg$k.rg[1:(i-1)]))
            ed <- st + 5 + cfg$k.rg[i]
            for(j in 1:cfg$k.rg[i]) # Loop over all positions
                  prb <- c(prb, HTSDT.Bias.Rank.1(c(cfg$k.rg[i], cfg$j.rg[i] - 1, j, parvec[st:ed])))
      }
      return(prb)
}

#===============================================================================
# 4. MODELL FUNCTIONS OF THE RATING PART OF THE HTSDT-MODEL FOR RANKING AND RATING
#===============================================================================
# ----------------------------------------------------------------------
# Function for computing response distribution of SDT model
# Input:	Vector of parameters:
#           1. Probability of recollection
#           2. Probability of guessing
#           3. Bias for guessing
#           4. mean
#           5. sd
#           6. c[1], c[2], ...., c[n-1])
#
# Output: 	Vector SDT response probabilities for one signal
# ----------------------------------------------------------------------
HTSDT.Bias.SDT.Probs <- function(parvec)
{
      Resp <- pnorm(parvec[6:length(parvec)], mean = parvec[4], sd = max(parvec[5], HTSDT.Bias.Rk.Tol))   # Probability of Response-Category 
	Resp <- c(Resp, 1)							                                        # Response vector
	c(Resp[1], Resp[-1]-Resp[-length(Resp)])				                                       # Compute Differences
}				

# ------------------------------------------------------------------------------
# Function for creating response distribution for on Type of Stimulus
# Input:	Vector of parameters:
#           1. Number indicating the category the recollection probability is assigned to
#           2. Probability of recollection
#           3. Probability of guessing
#           4. Bias for guessing
#           5. mean
#           6. sd
#           7. c[1], c[2], ...., c[n-1])
#
# Output: 	Vector of response probabilities of the HTSDT model for one signal
# ------------------------------------------------------------------------------
HTSDT.Bias.Probs <- function(parvec)
{
	 n    <- length(parvec)
       ncat <- n - 5                                                                         # Number of response categories
       Resp <- pnorm(parvec[7:n], mean = parvec[5], sd = max(parvec[6], HTSDT.Bias.Rk.Tol))  # Probability of Response-Category 
	 Resp <- c(Resp, 1)							                           # Response vector
	 Resp <- c(Resp[1], Resp[-1]-Resp[-ncat])				                           # Compute Differences
	 Resp <- (1-parvec[2]-parvec[3]) * Resp                                                # Non-Recollection and non-guessing probability

       # Treat Recollection: Add recollection porbability to the category given in the first position of the parameter vector
       if(parvec[2] > 0) Resp[parvec[1]] <- Resp[parvec[1]] + parvec[2]

       # Treat Guessing
       if(parvec[3] > 0)
       {
           if(ncat %% 2 == 0) # Even number of Response categories -> Position is left and right of the decision criterion
           {
                  Resp[ncat/2]     <- Resp[ncat/2]     + parvec[3]*parvec[4]
                  Resp[ncat/2 + 1] <- Resp[ncat/2 + 1] + parvec[3]*(1-parvec[4])
           }      
           else              # Uneven number of response categories -> Positon in the middle 
                  Resp[as.integer(ncat/2) + 1] <- Resp[as.integer(ncat/2) + 1] + parvec[3]
       }
     Resp 
}				

# ------------------------------------------------------------------------------------------------------------------
# Computation or probabilities of SDT model with recollection and rating
# Input:	parvec = Vector of parameters, consisting of the following entries
#                      I.  Parameters of the ranking part (ignored)
#                      II. For each signal of the Rating part
#                          1. Probability of recollection
#                          2. Probability of guessing
#                          3. Bias for guessing
#                          4. mean for each model 
#			         5. standard deviation for each model
#                      II.  Identical for all models
#			             thresholds
#
#		n   	= Number of models
#           pos   = vector of numbers indicating for each model where the 
#                   recollection weight has to be placed
#           fixed = 2xp matrix of fixed    constraints
#           ident = 2xq matrix of equality constraints
#
# Output: Vector of probabilities for single models
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.SDT <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
      n           <- cfg$n.sdt                                          # Number of models
      nrating     <- length(cfg$j.rg)                                   # Number of rating models
      start.sdt   <- 1 + 6*nrating + sum(cfg$k.rg)                      # Start of SDT Parameters (first 6 parameters + bias belong to Ranking model
      start.th    <- 5*n + start.sdt                                    # Startindex for thesholdes
	thresholds 	<- parvec[start.th:length(parvec)]                    # Threshold parameters
      ncat        <- length(thresholds) + 1                             # Number of response categories
      
                        
      Par 		<- matrix(parvec[start.sdt:(start.th-1)], nr = n, byrow = T)     # Matrix with probability and Gaussian parameters of HTSDT1 Model
	thresholds 	<- matrix(rep(thresholds, n), nr = n, byrow = T)                 # Matrix with thresholds for each model

      if(is.null(cfg$rec.pos)) 
	     Par <- cbind(c(1, rep(ncat, n-1)), Par, thresholds)          # Matrix with position of recollection category and Gaussian and threshold parameters in each row
      else
	     Par <- cbind(cfg$rec.pos, Par, thresholds)
	# Compute probabilities
	as.vector(apply(Par, 1, HTSDT.Bias.Probs))
}

#===============================================================================
# 5. MODELL FUNCTIONS OF THE (COMBINED) RANKING-RATING HTSDT MODEL
#===============================================================================

#-------------------------------------------------------------------------------
# Computation of ranking and rating probabilties
# Input:
#     parvec = parameter vector with the following entries
#              1: mean of the new distribution
#              2: sd   of the new distribution
#              3: mean of the old distribution
#              4: sd   of the old distribution
#     cfg    = list with configuration information containing the following 
#              entries:
#                 k.rg   = number of alternatives to be ranked
#                 j.rg   = number of positions to be considered
#                 npt    = Number of quadrature points
#                 rating = Flag, indicating whether rating data are present
#      fixed = matrix of fixed constraints
#      ident = matrix of equality constraints
#     with.p = flag indicating whether probability parameters should be 
#              computed from raw parameters (default = TRUE)
#
# Output: Ranking and rating probabilities
#-------------------------------------------------------------------------------
HTSDT.Bias.Rank <- function(parvec, cfg, fixed = NULL, ident = NULL, with.p = T)
{
      if(is.null(cfg$rating)) rating <- T
      else                    rating <- cfg$rating

      # 1. Compute probability parameters if required
      if(with.p)
            parvec <- HTSDT.Bias.Rk.Ppars(parvec, cfg, fixed, ident)

      # 2. Compute model probabilities
      pr.Rg <- HTSDT.Bias.RankOnly(parvec, cfg, fixed, ident)

      if(cfg$rating == F) return(pr.Rg)
      c(pr.Rg, HTSDT.Bias.SDT(parvec, cfg, fixed, ident))
}


# ------------------------------------------------------------------------------------------------------------------
# Model incorporating restrictions
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.Full <- function(parvec, cfg, pos, fixed = NULL, ident = NULL, functional = NULL, with.p = T)
{
      parvec <- Restore.par(parvec, pos, fixed, ident, functional)	# Restore full parameter vector
      HTSDT.Bias.Rank(parvec, cfg, fixed, ident, with.p)
}

#===============================================================================
# 6. JACOBIANS FOR THE RANKING PART OF THE RANKING-RATING MODEL
#===============================================================================

#-------------------------------------------------------------------------------
# Derivative of the non-target functions sith respect to:
#  x, mu, sigma, bias and target bias
#-------------------------------------------------------------------------------
SDT.Bias.Rk.cond.Jacobi  <- function(x, mu, sigma, bias, bias.t)
{
      sigma <- max(sigma, HTSDT.Bias.Rk.Tol)
      z    <- (x - mu - bias + bias.t)/sigma
      fn   <- dnorm(z)
      dm   <- -fn/sigma			# Derivative of Gaussian with respect to m
      db   <- -fn/sigma             # Derivative of Gaussian with respect to bias
      db.t <-  fn/sigma             # Derivative of Gaussian with respect to bias of target
	ds   <- -z*fn/sigma	      # Derivative of Gaussian with respect to sd
      dx   <-  fn/sigma             # Derivative of Gaussian with respect to x
      return(c(dx, dm, ds, db, db.t))
}


#-------------------------------------------------------------------------------
# Derivation vector of the function MIX.Bias.Rk.cond()
# Input:    x     =  actual value 
#           s.pos =  vector of distribution prior to x
#           e.pos =  vector of distribution posterior to x
#              mu =  mean   of non-target distribution
#           sigma =  stddev of non-target distribution
#           bias  = vector  of bias parameters
#
# Output: vector of partial derivatives with respect to the following parameters:
#           [mu, sigma, x, bias.1, bias.2, ..., bias.n]
#-------------------------------------------------------------------------------
HTSDT.Bias.Rk.cond.Jacobi <- function(x, s.pos, e.pos, t.pos, mu, sigma, bias)
{
      lsp <- length(s.pos)
      lep <- length(e.pos)
      dp <- rep(0, 3 + length(bias))
      index <- c(2:3, 1)                  # Reorder parameters: (x, mu, sigma) -> (mu, sigma, x) 
      for(s in s.pos)
      {
          dp.s <- -SDT.Bias.Rk.cond.Jacobi(x[1], mu, sigma, bias[s], bias[t.pos])
          if(lsp > 1)
                  dp.s <- dp.s * prod(1 - pnorm(x[1] + bias[t.pos] - bias[s.pos[s.pos != s]], mu, sigma)) 
          if(lep > 0)
                  dp.s <- dp.s * prod(pnorm(x[1] + bias[t.pos] - bias[e.pos], mu, sigma))
                  
          # reorder parameters to get bias correctly                    
          dp[1:3]     <- dp[1:3] + dp.s[index]
          dp[3+t.pos] <- dp[3+t.pos] + dp.s[5]
          dp[3+s]     <- dp[3+s] + dp.s[4]        
      }
      for(e in e.pos)
      {
          dp.e <- SDT.Bias.Rk.cond.Jacobi(x[1], mu, sigma, bias[e], bias[t.pos])
          if(lep > 1)
                  dp.e <- dp.e * prod(pnorm(x[1] + bias[t.pos] - bias[e.pos[e.pos != e]], mu, sigma)) 
          if(lsp > 0)
                  dp.e <- dp.e * prod(1 - pnorm(x[1] + bias[t.pos] - bias[s.pos], mu, sigma))
                  
          # reorder parameters to get bias correctly                    
          dp[1:3]     <- dp[1:3] + dp.e[index]
          dp[3+t.pos] <- dp[3+t.pos] + dp.e[5]
          dp[3+e]     <- dp[3+e] + dp.e[4]        
      }

      return(dp*x[2])
}      


#-------------------------------------------------------------------------------
# Derivation vector of the function HTSDT.Bias.Rk()
# Input:    s.pos =  vector of distribution prior to x
#           e.pos =  vector of distribution posterior to x
#       mu.target =  mean   of target distribution
#    sigma.target =  stddev of target distribution
#           bias  = vector  of bias parameters
#              mu =  mean   of non-target distribution
#           sigma =  stddev of non-target distribution
#
# Output: vector of partial derivatives with respect to the following parameters:
#           [mu, sigma, mu.target, sigma.target, bias.1, bias.2, ..., bias.n]
#-------------------------------------------------------------------------------
HTSDT.Bias.Rk.Jacobi  <- function(s.pos, e.pos, t.pos, mu.target, sigma.target, bias, mu = 0, sigma = 1)
{
      Save.GH.Struc    <- GH.Structure[,1]
      GH.Structure[,1] <- GH.Structure[,1] * sigma.target + mu.target
      J.Rk <- apply(GH.Structure, 1, HTSDT.Bias.Rk.cond.Jacobi, s.pos = s.pos, e.pos = e.pos, t.pos = t.pos, mu = mu, sigma = sigma, bias = bias)
      J.Rk <- rbind(J.Rk[1:3,], J.Rk[3,]* Save.GH.Struc, J.Rk[-(1:3),])
      rowSums(J.Rk)
}


# ------------------------------------------------------------------------------
# Computation of derviative vector of function HTSDT.Bias.combn()
# Input:     x = set from which to select
#            m = number of selected items
#        t.pos = position of target alternative
#    mu.target = mean of the target alternative
# sigma.target = stddev of the target alternative
#         bias = vetor of bias parameters
#           mu = mean of the non-target alternative
#        sigma = stddev of the non-target alternative
#
# Output: derivation vector [ mu, sigma, mu.target, sigma.target, bias.1, bias.2, ..., bias.n]
# ------------------------------------------------------------------------------
HTSDT.Bias.combn.Jacobi <- function(x, m, t.pos, mu.target, sigma.target, bias, mu, sigma)
{
      n <- length(x)
      e <- 0
      h <- m
      a <- 1:m
      i <- 2
      nmmp1 <- n - m + 1
      mp1 <- m + 1
      deriv.vec <- HTSDT.Bias.Rk.Jacobi(x[a], x[!(x %in% x[a])], t.pos, mu.target, sigma.target, bias, mu, sigma)
      while (a[1] != nmmp1)
      {
            if (e < n - h)
            {
                  h <- 1
                  e <- a[m]
                  j <- 1
            }
            else
            {
                  h <- h + 1
                  e <- a[mp1 - h]
                  j <- 1:h
            }
            a[m - h + j] <- e + j
            deriv.vec <- deriv.vec + HTSDT.Bias.Rk.Jacobi(x[a], x[!(x %in% x[a])], t.pos, mu.target, sigma.target, bias, mu, sigma)
            i <- i + 1
    }
    return(deriv.vec)
}

# ----------------------------------------------------------------------
# Jacobian function corresponding to the function HTSDT.Bias.Rk.1() 
# Input: j.rg  = Rank position of the target
#        k.rg  = Number of items presented
#        t.pos = position of the target item
#           mu = mean of the non-target distribution
#        sigma = standard deviation of the non-target distribution
#    mu.target = mean of the target distribution
# sigma.target = standard deviation of the target distribution
#         bias = vector of position bias parameters
#
# Output: vector of partial derivatives with respect to the following parameters:
#         [mu, sigma, mu.target, sigma.target, bias.1, bias.2, ..., bias.n]
# ----------------------------------------------------------------------
HTSDT.Bias.Rk.1.Jacobi <- function(j.rg, k.rg, t.pos, mu, sigma, mu.target, sigma.target, bias)
{
      if(j.rg == 1) 
      {
            s.pos <- NULL 
            e.pos <- (1:k.rg)[-t.pos]
            return(HTSDT.Bias.Rk.Jacobi(s.pos, e.pos, t.pos, mu.target, sigma.target, bias, mu, sigma))
      }
      if(j.rg == k.rg)
      {
            s.pos <- (1:k.rg)[-t.pos]
            e.pos <- NULL
            return(HTSDT.Bias.Rk.Jacobi(s.pos, e.pos, t.pos, mu.target, sigma.target, bias, mu, sigma))
      }
      x <- (1:k.rg)[-t.pos]
      return(HTSDT.Bias.combn.Jacobi(x, j.rg - 1, t.pos, mu.target, sigma.target, bias, mu, sigma))
}      

#-------------------------------------------------------------------------------
# Computation of Jacobian matrix of the probabilities with respect to parameters
# for a single model and a single position of the target item
# Input:
#     parvec = parameter vector with the following entries
#              1. k.rg:       number of alternatives to be ranked
#              2. j.rg:       number of positions to be considered
#              3. t.pos:      position of target
#              4. probability of recollection
#              5. probabilitiy of pure guessing
#              6. mean of the new distribution
#              7. sd   of the new distribution
#              8. mean of the old distribution
#              9. sd   of the old distribution
#
# Output: Ranking probabilities
#-------------------------------------------------------------------------------
HTSDT.Bias.Rank.1.Jacobi <- function(parvec)
{

      sdt.prob <- (1.0 - parvec[4] - parvec[5])
      if(sdt.prob > 0.0)
      {
            J <- sapply(1:parvec[2], HTSDT.Bias.Rk.1.Jacobi, k.rg = parvec[1], t.pos = parvec[3], mu = parvec[6], sigma = parvec[7],
                        mu.target = parvec[8], sigma.target = parvec[9], bias = parvec[10:(9 + parvec[1])])      
            J <- sdt.prob * J
      }
      else
            J <- matrix(0.0, nr = 4 + parvec[1], nc = parvec[2])

      # Compute derivative of p.recollection
      dp.guess <- dp.rec <- -sapply(1:parvec[2], HTSDT.Bias.Rk.1, k.rg = parvec[1], t.pos = parvec[3], mu = parvec[6], sigma = parvec[7],
                                    mu.target = parvec[8], sigma.target = parvec[9], bias = parvec[10:(9 + parvec[1])])
      dp.rec[1] <- 1 + dp.rec[1]        
    
      # Compute derivative of p.guessing
      dp.guess <- dp.guess + 1/parvec[1]
   
      # Bind results together              
      J <- cbind(dp.rec, dp.guess, t(J))
      return(rbind(J, 0.0 - colSums(J)))
}

#-------------------------------------------------------------------------------
# Jacobian matrix of for all ranking models with bias
#-------------------------------------------------------------------------------
HTSDT.Bias.RankOnly.Jacobi <- function(parvec, cfg)
{
      prb <- NULL
      n.rg <- length(cfg$k.rg)
      J <- NULL
      for(i in 1:n.rg)              # Loop of all models
      {
            # Extract relevant parameter vector
            st <- ifelse(i == 1, 1, 1 + 6*(i-1) + sum(cfg$k.rg[1:(i-1)]))
            ed <- st + 5 + cfg$k.rg[i]
            J.act <- NULL
            for(j in 1:cfg$k.rg[i]) # Loop over all positions
                  J.act <- rbind(J.act, HTSDT.Bias.Rank.1.Jacobi(c(cfg$k.rg[i], cfg$j.rg[i] - 1, j, parvec[st:ed])))

            # Generate Jacobian matrix for all models
            if(i == 1) J <- J.act
            else
            {
                  filler.r <- matrix(0, nr = dim(J.act)[1], nc = dim(J)[2])
                  filler.c <- matrix(0, nr = dim(J)[1],     nc = dim(J.act)[2])
                  J <- rbind(cbind(J, filler.c), cbind(filler.r, J.act))
            }
      }
      return(J)
}            

#===============================================================================
# 7. JACOBIANS FOR THE RATING PART OF THE RANKING MODEL
#===============================================================================

# ------------------------------------------------------------------------------
# Jacobi-Matrix of the Gaussian parameters and thresholds
# Input:	I. Vector of parameters (without thresholds):
#                 1. Number indicating the category the recollection probability
#                    is assigned to
#                 2. Probability of recollection
#                 3. Probability of guessing
#                 4. Bias for guessing
#                 5. mean
#                 6. sd
#           II. Threshold parameters: c[1], c[2], ...., c[n-1])
#
# Output: 	Jacobian of the HTSDT model for one signal
# ------------------------------------------------------------------------------
HTSDT.Bias.Rank.SDT.Probs.Jacobi <- function(gausspar, th)
{
      # 1. Extract parameters
      rec.cat <- gausspar[1]                         # Recollection category
      p.rec   <- gausspar[2]                         # Recolletion
      p.guess <- gausspar[3]                         # Guessing
      p.bias  <- gausspar[4]                         # Bias
	m       <- gausspar[5]                         # Mean parameter
	sd      <- max(HTSDT.Bias.Rk.Tol, gausspar[6]) # Stddev
	std.p   <- 1.0 - p.rec - p.guess
      
      # 2. Compute derivatives for SDT parameters
	n.t <- length(th)					# Number of thresholds
	z  <- (th - m)/sd
	fn <- dnorm(z)
	
	# Derivative with respect to mean
	dm <- -fn/sd						# Derivative of Gaussian with respect to m
	dm <- c(dm[1], dm[-1] - dm[-n.t], -dm[n.t])
      dm <- std.p * dm                 # Adjust for recollection and guessing

	# Derivative with respect to sd
	ds <- -z*fn/sd						# Derivative of Gaussian with respect to sd
	ds <- c(ds[1], ds[-1] - ds[-n.t], -ds[n.t])
      ds <- std.p * ds                 # Adjust for recollection and guessing


	# Derivative with respect to thresholds
	dt <- fn/sd							# Derivative of Gaussian with respect to t[i]
	dt <- std.p * dt                 # Adjust for recollection and guessing
	
	J <- matrix(0, nr = n.t+1, nc = n.t)		# Jacobi matrix for thresholds: columns: Parameter, rows = probabilities
	r <- 1
	for(i in 1:n.t)
	{
		J[r, i] <- dt[i]
		r <- r + 1
		J[r, i] <- -dt[i]
	}
	J.SDT <- cbind(dm, ds, J)				# Full Jacobi matrix of SDT model (adjusted for probabilities)
	
      # 2. Compute derivatives for probability parameters
      prb.SDT <- HTSDT.Bias.SDT.Probs(c(gausspar[-1], th))	            # Compute SDT response probabilities
      dp.rec <- -prb.SDT
      dp.rec[rec.cat] <- dp.rec[rec.cat] + 1
	
	dp.guess <- -prb.SDT
	dp.bias  <- rep(0, n.t+1)
	
      middle <- as.integer((n.t+1)/2)
	if(n.t %% 2 > 0) # Even number of Response categories -> Guessing position is left and right of the decision criterion
      {
            dp.guess[middle]     <- dp.guess[middle] + p.bias
            dp.guess[middle + 1] <- dp.guess[middle + 1] + (1 - p.bias)
            
            dp.bias[middle]     <-  p.guess
            dp.bias[middle + 1] <- -p.guess
      }
      else
            dp.guess[middle + 1] <- dp.guess[middle + 1] + 1
      
      cbind(dp.rec, dp.guess, dp.bias, J.SDT)
}


# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix of the SDT-Model (rating part) of the Ranking-Rating model
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.SDT.Jacobi <- function(parvec, cfg)
{
      start.sdt <- 6*length(cfg$j.rg) + sum(cfg$k.rg)
      start.th <-  start.sdt + 5*cfg$n.sdt + 1
	# Isolate threshold parameters
	th <- parvec[start.th: length(parvec)]	  # Thresholds
	n.t <- length(th)					  # Number of thresholds
      n.cat <- n.t + 1

	# Extract recollection position 
      pos <- cfg$rec.pos                # Extract Recollection position vector 
      if(is.null(pos)) 
            pos <- c(1, rep(n.cat, cfg$n.sdt - 1))   # Position of recollection weight: for first model in cat 1 for other ones in last cat
  
      # Matrix of probabilty and Gaussian parameters for the single models
      Pmatrix <- matrix(parvec[(start.sdt + 1):(start.sdt + 5*cfg$n.sdt)], nr = cfg$n.sdt, byrow = T)
 
      for(i in 1:cfg$n.sdt)
      {
            J.mod <- HTSDT.Bias.Rank.SDT.Probs.Jacobi(c(pos[i], Pmatrix[i,]), th)
            J.G <- J.mod[, 1:5]
            J.t <- J.mod[, 6:(5 + n.t)]

            # New Jacobian matrix for Gaussian model and thresholds
            if(i < 2)
            {
                  J.Gauss <- J.mod[, 1:5]
                  J.th    <- J.mod[, 6:(5 + n.t)]
            }
            else
            {
                  J.Gauss <- rbind(cbind(J.Gauss, matrix(0, nr = n.cat*(i-1), nc = 5)), 
                                   cbind(matrix(0, nr = n.cat, nc = 5*(i-1)), J.mod[, 1:5]))
                  J.th    <- rbind(J.th, J.mod[, 6:(5 + n.t)])
            }      
      }
      cbind(J.Gauss, J.th)
}

#===============================================================================
# 8. JACOBI MATRIX, GRADIENT AND MODEL MATRICES OF THE RATING-RANKING MODEL
#===============================================================================

#-------------------------------------------------------------------------------
# Jacobian of the model function of the Ranking and Rating model:
#-------------------------------------------------------------------------------
HTSDT.Bias.Rank.Jacobi <- function(parvec, cfg, fixed = NULL, ident = NULL, with.p = T)
{
      # 0. Compute probability parameters
      if( with.p)
            parvec <- HTSDT.Bias.Rk.Ppars(parvec, cfg, fixed, ident)

      # 1. Jacobi-Matrix of the ranking part
      rank.J <- HTSDT.Bias.RankOnly.Jacobi(parvec, cfg)
      if(cfg$rating == F)     # No rating 
            return(rank.J %*% HTSDT.Bias.Rank.Ppars.Jacobi(parvec, cfg, fixed, ident))
      
      # 2. Jacobi-Matrix of the rating part
      rating.J <- HTSDT.Bias.Rank.SDT.Jacobi(parvec, cfg)
      
      n.par.Rk   <- 6*length(cfg$j.rg) + sum(cfg$k.rg)      # Number of Ranking parameters
      n.par.SDT  <- length(parvec) - n.par.Rk               # Number of SDT parameters
      filler.Rk  <- matrix(0, nr = sum(cfg$j.rg * cfg$k.rg), nc = n.par.SDT)
      filler.SDT <- matrix(0, nr = dim(rating.J)[1],         nc = n.par.Rk)
      
      J <- rbind(cbind(rank.J, filler.Rk), cbind(filler.SDT, rating.J))

      # 3. Adjust for probability paramters
      return(J %*% HTSDT.Bias.Rank.Ppars.Jacobi(parvec, cfg, fixed, ident))
}

#-------------------------------------------------------------------------------
# Help function that computes those positions of the identity
# that correspond to identity constraints between probabilities 
# of different/identical blocks
# Input:  1. Matrix with identity constraints (2xp) Matrix
#         2. Configuration information  
#         3. Flag indicating whether positions of between or within block 
#            constraints are required (T = between, F = within)
#
# Output: Positions of identity constraints between probability parameters in
#         different or in the same blocks (or NULL if none exist)
#-------------------------------------------------------------------------------
HTSDT.Bias.Rank.Find.Ident.Ppar.Pos <- function(ident, cfg, between)
{
      if(is.null(ident)) return(NULL)
      
      n.id  <- dim(ident)[2]
      
      # 1. Specification of positions of probability parameters without guessing
      ppos <- HTSDT.Bias.pos.par(cfg, Par.Id = "Prob")
      if(cfg$rating)
            ppos.bias <- 3 + 6*length(cfg$j.rg) + sum(cfg$k.rg) + 5*(0:(cfg$n.sdt-1))         # Positions of guessing bias parameters          
      else
            ppos.bias <- NULL

      # 2. Indices of between and within constraints
      #    Constraints involving guessing parameters are treated as within constraints
      if(between)
            i.index  <- (ident[1,] %in% ppos) & (ident[2,] %in% ppos) & (abs(ident[1,] - ident[2,]) > 1)
      else
            i.index  <- ((ident[1,] %in% ppos) & (ident[2,] %in% ppos) & (abs(ident[1,] - ident[2,]) == 1)) |
                        ((ident[1,] %in% c(ppos, ppos.bias)) & (ident[2,] %in% ppos.bias) & (abs(ident[1,] - ident[2,]) > 0)) 

      pos <- (1:n.id)[i.index]
      if(length(pos) < 1) return(NULL)
      return(pos)
}                  
  
#-------------------------------------------------------------------------------
# Wrapper functions
#-------------------------------------------------------------------------------
HTSDT.Bias.Rank.Find.Ident.Ppar.Pos.between <- function(ident, cfg) { HTSDT.Bias.Rank.Find.Ident.Ppar.Pos(ident, cfg, T) }
HTSDT.Bias.Rank.Find.Ident.Ppar.Pos.within  <- function(ident, cfg) { HTSDT.Bias.Rank.Find.Ident.Ppar.Pos(ident, cfg, F) }
      
#-------------------------------------------------------------------------------
# Model matrix of the ranking model 
# implementing functional and identity constraints
# Input:   1. full.par   = Full parameter vector
#          2. cfg        = List with configuration information
#          3. fixed      = Matrix of fixed constraints
#          4. ident      = Matrix of equality constraints
#          5. functional = A function implementing functional constraints
#
# Output: Model matrix of free paraemters
#
# Comment: Not considered are identity constraints between probability 
#          parameters from different blocks
#-------------------------------------------------------------------------------
HTSDT.Bias.Rank.Model.Matrix <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos      <- SavePos.par(full.par, fixed, ident)
      parvec   <- Elim.par(full.par, fixed, ident)
      full.par <- Restore.par(parvec, pos, fixed, ident, functional)	# Restore full parameter vector

      J   <- HTSDT.Bias.Rank.Jacobi(full.par, cfg, fixed, ident)
      if(!is.null(ident))
      {
            J.Id <- diag(length(full.par))

            # Find positions of constraints between probability parameters 
            # from different blocks:
            # These have to be ignored since they are adjusted separately
            pos.ppar.between <- HTSDT.Bias.Rank.Find.Ident.Ppar.Pos.between(ident, cfg)
            if(!is.null(pos.ppar.between))  ident <- ident[, -pos.ppar.between, drop = F]
            if(dim(ident)[2] > 0)
            {
                  Index <- matrix(c(ident[2,], ident[1,]), nc = 2)
                  J.Id[Index] <- 1
                  J <- J %*% J.Id
            }
      }
      
      # Treatment of functional constraints
      if(!is.null(functional))
      {
            J.fct <- attr(functional, "SDT.Jacobian")
            if(is.null(J.fct)) 
                  return(NULL)           # No Jacobian function supplied -> return (NULL)
            J <- J %*% J.fct(full.par)
      }
      J[, pos]
}

# ------------------------------------------------------------------------------------------------------------------
# Numerical computation of the model matrix (SDT-Rank model)
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.Model.Matrix.Num <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)
      parvec <- Elim.par(full.par, fixed, ident)
      J <- jacobian(HTSDT.Bias.Rank.Full, parvec, cfg = cfg, pos = pos, fixed = fixed, ident = ident, functional = functional)
}

# ------------------------------------------------------------------------------------------------------------------
# Gradient of logL for the HTSDT-Rank model
# Input:	Same as for Likelihood function
# Output:	Gradient
#
# Comment:  Identity constraints betreen probability prameters from different blocks are removed from matrix 
#           ident and added to matrix fixed.
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.Grad <- function(parvec, datavec, Model, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
	parvec   <- Restore.par(parvec, pos, fixed, ident, functional)	# Restore full parameter vector
	gradient <- as.vector(SDT.logL.Jacobi(parvec, datavec, Model, n, pos, fixed, ident, functional) %*% HTSDT.Bias.Rank.Jacobi(parvec, n, fixed, ident)) 

	# Treat functional constraints
	if(!is.null(functional))
	{
		gr.fct <- attr(par, "SDT.gradient")		# Isolate gradient function
		if(is.null(gr.fct))	stop("\n\n HTSDT.Bias.Rank.Grad: Error in computation of gradient: No gradient for functional constraints found:\n Set Parameter sym.gr = F in function SDT.Estimate !\n\n")
		gradient <- gr.fct(gradient)
	}	

      # Identity constraints
      if(!is.null(ident))
      {
            # Eliminate identity constraints involving probability parameters 
            # from different blocks from matrix of identity constraints
            # and add them to the matrix of fixed constraints
            pos.ppar.between <- HTSDT.Bias.Rank.Find.Ident.Ppar.Pos.between(ident, n)
            if(!is.null(pos.ppar.between))  
            {
                  fixed <- cbind(fixed, ident[, pos.ppar.between])
                  ident <- ident[, -pos.ppar.between, drop = F]
                  if(dim(ident)[2] < 1) ident <- NULL
            }
 	}  
	Elim.par.grad(gradient, fixed, ident)
}


#===============================================================================
# 7. JACOBIAN FOR ADJUSTING PROBABILITY PARAMETERS
#===============================================================================
# --------------------------------------------------------------------------------
# Computation of Jacobian for adjusting probability paraemters
# Input:	1. parvec = full parameter vector with probability parameters computed
#           2. cfg    = list with configuration information
#           3. fixed  = matrix of fixed constraints
#           4. ident  = matrix of equality constraints
#
# Output:	Jacobian matrix for adjusting probability parameters
#
# Comment:  Fixed constraints are taken into account
# --------------------------------------------------------------------------------
HTSDT.Bias.Rank.Ppars.Jacobi <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
      n.rk <- length(cfg$j.rg)

      # 1. Find probability vectors
      if(n.rk == 1)                                         # Position  of AFC recollection parameter (single model)
            prec <- 1
      else
            prec <- c(1, 1 + cumsum(6 + cfg$k.rg[-n.rk]))   # Positions of AFC recollection parameters (multiple models)
            
      if(cfg$rating == F)
      {
            pgue <- prec + 1                                # Positions of guessing paramters
            ppos <- c(prec, pgue)                           # Positions of all probability parameters
            pbis <- -1                                      # Pseudoposition (because no bias parameters exist)
            parvec <- parvec[1:(6*n.rk + sum(cfg$k.rg))]    # Use reduced parameter vector only
      }
      else                
      {
            prec <- c(prec, 
                      1 + 6*n.rk + sum(cfg$k.rg) + 5*(0:(cfg$n.sdt-1)))   # Positions of recollection parameters
            pgue <- prec + 1                                              # Positions of guessing paramters
            pbis <- 3 + 6*n.rk + sum(cfg$k.rg) + 5*(0:(cfg$n.sdt-1))      # Positions of bias parameters
            ppos <- c(prec, pgue, pbis)                                   # Positions of all probability parameters
      }

      if(!is.null(fixed)) fx <- fixed[2,]
      else                fx <- NULL
      
      if(!is.null(ident)) id <- ident[2, HTSDT.Bias.Rank.Find.Ident.Ppar.Pos.between(ident, cfg)]
      else                id <- NULL

      fx.par <- (ppos %in% fx) | (ppos %in% id)           # Boolean vector indicating fixed probability parameters
      ppos.free <- ppos[!fx.par]
      ppos.fix  <- ppos[fx.par]
      ppar.free <- parvec[ppos.free]
      ppar.fix  <- parvec[ppos.fix]
      prec.free <- prec[prec %in% ppos.free]
      pgue.free <- pgue[pgue %in% ppos.free]
      pbis.free <- pbis[pbis %in% ppos.free]
      
      # Diagonal entries
      dg  <- rep(1, length(parvec))        # Diagonal elements of Jacobian
      dg[ppos.fix]  <- 0
      if(length(pbis.free) > 0) dg[pbis.free] <- parvec[pbis.free] * (1.0 - parvec[pbis.free])  # Bias parameters
      
      J.adjust <- diag(dg)

      # Compute diagonal and cross-diagonal entries for recollcetion and guessing
      for(i in prec)                      # Cycle through recollection parameters (guessing parameters are 1 position to the right)          
      {
            # Case 1: Recollection parameter is free
            if(i %in% prec.free)
            {
                  # Case 11: Guessing parameter is free
                  if((i+1) %in% pgue.free)
                  {
                        J.adjust[i, i]     <- parvec[i]   * (1.0 - parvec[i])             # Diagonal element Recollection
                        J.adjust[i+1, i+1] <- parvec[i+1] * (1.0 - parvec[i+1])           # Diagonal element Guessing
                        J.adjust[i+1, i] <- J.adjust[i, i+1] <- -parvec[i]*parvec[i+1]    # Cross-diagonal element
                  }
                  # Case 12: Guessing parameter is fixed
                  else
                  {
                        if(parvec[i+1] < 1)
                        {
                             p.adj <- parvec[i]/(1.0 - parvec[i+1]) 
                             J.adjust[i, i] <- parvec[i] * (1.0 - p.adj)
                        }
                  }
            }
            # Case 2: Recollection parameter is fixed (and guessing parameter is free)
            else 
            {     
                  if((i+1) %in% pgue.free)   
                  {
                        p.adj <- parvec[i+1]/(1.0 - parvec[i])
                        J.adjust[i+1, i+1] <- parvec[i+1] * (1.0 - p.adj)
                  }
            }
      } 
      # Adjust for equaltity constraints between probability parameters
      J.Id <- HTSDT.Bias.Rank.Ident.Jacobi(parvec, cfg, fixed, ident)
      if(is.null(J.Id))  return(J.adjust)
      return(J.Id %*% J.adjust)
}

# ------------------------------------------------------------------------------------------------------------------
# Jacobian matrix corresponding to equality constraints between probability parameters
# Input:	1. parvec = full parameter vector with probability parameters computed
#           2. cfg    = list with configuration information
#           3. fixed  = matrix of fixed constraints
#           4. ident  = matrix of equality constraints
#
# Output:	Jacobian matrix for adjusting probability parameters
#
# Comments:  - This matrix must be multiplied from left since it corresponds to 
#              equality constraints between probabilities and not between raw parameters
#            - Only equality constraints between probability parameters from different blocks are adjusted
#              The other parameter are adjusted in the usual way.
#              Note that identity constraints within a block of probability parameters can be treated like 
#              common identity constraints since, in this case, raw paramters are set identical 
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.Ident.Jacobi <- function(parvec, cfg, fixed, ident)
{
      if(is.null(ident)) return(NULL)

      # Only identity constraints between probability parameters of different blocks are considered
      pos.ppar.between <- HTSDT.Bias.Rank.Find.Ident.Ppar.Pos.between(ident, cfg)
      
      # If no relevant identity constraints found return
      if(is.null(pos.ppar.between)) return(NULL)

      # Extract relevant positions from identity matrix      
      ident <- ident[, pos.ppar.between, drop = F]
      if(dim(ident)[2] < 1) return(NULL)

      # 1. Set cells for identity constraints      
      d.elem <- rep(1, length(parvec))
      d.elem[ident[2,]] <- 0
      if(!is.null(fixed)) d.elem[fixed[2,]] <- 0
      J.ident <- diag(d.elem)

      Index <- matrix(c(ident[2,], ident[1,]), nc = 2)
      J.ident[Index] <- 1
      
      # 2. Adapt for probabilities
      n.rk <- length(cfg$j.rg)
      if(n.rk == 1)                                                           # Position  of recollection parameter (single model)
           prec <- 1
      else
           prec <- c(1, 1 + cumsum(6 + cfg$k.rg[-n.rk]))                      # Positions of recollection parameters (multiple models)
           
      if(cfg$rating == T) 
           prec <- c(prec, 1 + 6*n.rk + sum(cfg$k.rg) + 5*(0:(cfg$n.sdt-1)))  # Positions of recollection parameters
      pgue <- prec + 1                                                        # Positions of guessing paramter
      ppos <- c(prec, pgue)                                                   # Positions of all probability parameters
     
      # 2.1. Determine fixed positions 
      if(!is.null(fixed)) fx <- fixed[2,]
      else                fx <- NULL
      
      if(!is.null(ident)) id <- ident[2,]
      else                id <- NULL
      fx.par <- ppos[(ppos %in% fx) | (ppos %in% id)]  # fixed probability parameters

      # 2.2 Case 1: Adjustments for recollection and guessing
      ppos.first <- ppos[ppos %in% ident[1,]]  # Consider only probabilty parameters that are the source of equaltiy constraints
      for(source.p in ppos.first)
      {
            target.p <- ident[2, ident[1,] == source.p]         # Find all identity constraints for parameter number source.p
            if(length(target.p) > 0)
            {
                  #-----------------------------------------------------------------------------------------------
                  # if target.p is recollection, other.p is guessing parameter         -> Add 1 to position
                  # if target.p is not recollection, other.p is recollection parameter -> Subtract 1 from position
                  #-----------------------------------------------------------------------------------------------
                  other.p     <- ifelse(target.p %in% prec, target.p + 1, target.p - 1)
                  free.tindex <-  !(other.p %in% fx.par)
                  free.other  <-  other.p[free.tindex]
                  # If guessing parameter is free, recollection has an influence on guessing that has to be taken into account 
                  if(length(free.other) > 0)  J.ident[free.other, source.p] <- -parvec[free.other]/(1-parvec[source.p])       # Guessing paraemter free
            }
      }
      return(J.ident)
}

# ------------------------------------------------------------------------------------------------------------------
# Jacobian for adjusting the covariance matrix due to the delta method
# Input:	1. parvec = full parameter vector with probability parameters computed
#           2. cfg    = list with configuration information
#           3. fixed  = matrix of fixed constraints
#           4. ident  = matrix of equality constraints
#
# Output:	Jacobian matrix for adjusting covariance matrix according to delta method
#           Contains only entries of free probability parameters
#
# Comments:  The matrix is nearly identical to the matrix given by the function HTSDT.Rank.Ppars.Jacobi
#            with one exeception: Probability parameters within one block that are set identical are also
#            considered and derviation of free probability parameters are included.
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.Delta.Matrix <- function(parvec, cfg, fixed, ident)
{
      # 1. Compute Jacobian matrix for the function that transforms raw to 
      #    probability parameters taking fixed and identity constraints between
      #    parameters of different blocks into account
      J <- HTSDT.Bias.Rank.Ppars.Jacobi(parvec, cfg, fixed, ident)
      if(is.null(J)) return(J)

      # 2. Identify free probability parameters
      free.ppos <- HTSDT.Bias.Rank.Free.Ppos(parvec, cfg, fixed, ident)
      if(is.null(free.ppos)) return(NULL)

      # 3. If no identity matrix, done
      if(is.null(ident))
            return(J[free.ppos, free.ppos, drop = F])           
      
      # 2. Now consider identity constraints within blocks
      pos.ppar.within <-  HTSDT.Bias.Rank.Find.Ident.Ppar.Pos.within(ident, cfg)
      ident <- ident[, pos.ppar.within, drop = F]
      if(dim(ident)[2] > 0)
      {
            J.within <- diag(dim(J)[2])
            Index <- matrix(c(ident[2,], ident[1,]), nc = 2)
            J.within[Index] <- 1
            J <- J %*% J.within
            return(J[free.ppos, free.ppos, drop = F])
      }
      return(J[free.ppos, free.ppos, drop = F])           
}

# ------------------------------------------------------------------------------------------------------------------
# Help function that returns the positions of the free probability parameters
# Input:	1. parvec = full parameter vector with probability parameters computed
#           2. cfg    = list with configuration information
#           3. fixed  = matrix of fixed constraints
#           4. ident  = matrix of equality constraints
#           5. full   = flag indicating whether positions are with respect to the full or the restricted parameter vector
#
# Output:	vector of postions of free probability parameters or NULL
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.Free.Ppos <- function(parvec, cfg, fixed, ident, full = T)
{
      # 1. Postions of probability parameters
      ppos <- HTSDT.Bias.pos.par(cfg, Par.Id = "Prob")

      # 2. Identify free probability parameters
      if(is.null(fixed) && is.null(ident)) 
            return(ppos)
      else if(is.null(fixed))
            free.p   <- !(ppos %in% ident[2, ])
      else if(is.null(ident))
            free.p   <- !(ppos %in% fixed[2, ])
      else
            free.p   <- !((ppos %in% fixed[2, ]) | (ppos %in% ident[2, ]))
            
      pos <- SavePos.par(parvec, fixed, ident)  # Positions of free parameters
      if(length(pos) == 0) return(NULL)

      if(full)
            return(pos[pos %in% ppos[free.p]])        # Return position of free probability parameters with respect to full  parameter vector
      return((1:length(pos))[pos %in% ppos[free.p]])  # Return positions of free probability parameters with respect to free parameter vector 
}

# ------------------------------------------------------------------------------------------------------------------
# Computing adjusted standard errors for probability parameters
# Input:	1. cov.matrix = covariance matrix of estimated parameters
#           2. parvec     = full parameter vector with probability parameters computed
#           3. cfg        = list with configuration information
#           4. fixed      = matrix of fixed constraints
#           5. ident      = matrix of equality constraints
#
# Output:	Vector of standand errors (adjusted) or NULL 
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.Delta.Method.Prob <- function(cov.matrix, parvec, cfg, fixed, ident)
{
      # 0. Tests and computation of unadjusted SE vector
      if(is.null(cov.matrix) || is.na(cov.matrix) || is.nan(cov.matrix))
            return(NULL)
      SE.vector <- sqrt(diag(cov.matrix))       # Vector of unadjusted standard errors
      
      # 1. Compute adjustment matrix
      J.adjust <- HTSDT.Bias.Rank.Delta.Matrix(parvec, cfg, fixed, ident)
      if(is.null(J.adjust)) return(SE.vector)
      
      # 2. Positions of free parameters with respect to restricted parameter vector     
      free.ppos <- HTSDT.Bias.Rank.Free.Ppos(parvec, cfg, fixed, ident, full = F)

      # 3. Carve out relevant part of the covariance matrix
      cov.free.ppos <- cov.matrix[free.ppos, free.ppos, drop = F]

      # 4. Construct adjusted SE vector and fill SE vector with adjusted values
      SE.adjust <- sqrt(diag(J.adjust %*% cov.free.ppos %*% t(J.adjust)))
      SE.vector[free.ppos] <- SE.adjust
      return(SE.vector)
}

#===============================================================================
# 8. ADMINISTRATIVE FUNCTIONS
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Check and generate configuration structure
# Input:	cfg = Configuration structure: A list with the following entries
#                 - n.sdt        Number of SDT models
#                 - k.rg         Number of alternatives
#                 - j.rg         Number of relevant ranking positions
#                 - npt          Number of quadrature points
#                 - restriction  string identifying the restriction
#                 - rating       Flag indicating whether ranking and rating are combined
#
# Output: String with parameter values
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.Check.Fill.Cfg <- function(cfg)
{
      n.sdt <- 2
      k.rg  <- 2
      npt   <- 30

      if(is.null(cfg))
      {
            cfg <- list(n.sdt       = n.sdt,
                        k.rg        = k.rg,  
                        j.rg        = k.rg, 
                        npt         = npt, 
                        restriction = "standard",
                        rating      = T)
            return(cfg)
      }
      if(is.numeric(cfg))
      {
            k.rg     <- as.integer(cfg)
            cfg      <- list(n.sdt       = n.sdt,
                             k.rg        = k.rg,  
                             j.rg        = k.rg, 
                             npt         = npt, 
                             restriction = "standard",
                             rating      = T)
            return(cfg)
      }
      cfg.new          <- list(n.sdt = ifelse(is.null(cfg$n.sdt), n.sdt, cfg$n.sdt))
      
      if(is.null(cfg$k.rg))
          if(is.null(cfg$j.rg)) cfg.new$k.rg <- k.rg
          else                  cfg.new$k.rg <- cfg$j.rg
      else 
          cfg.new$k.rg <- cfg$k.rg

      if(is.null(cfg$j.rg))    cfg.new$j.rg <- cfg.new$k.rg
      else                     cfg.new$j.rg <- cfg$j.rg

      cfg.new$npt         <- ifelse(is.null(cfg$npt),  npt, cfg$npt)
      cfg.new$restriction <- HTSDT.Bias.Restriction(cfg$restriction)
      cfg.new$rating      <- ifelse(is.null(cfg$rating),    T, cfg$rating)

      if(!is.null(cfg$rec.pos)) cfg.new$rec.pos <- cfg$rec.pos                      # Position of recollection
      
      # Checks
      if(cfg.new$n.sdt < 2)
            stop(paste("\n FATAL ERROR in HTSDT.Bias.Rank.Check.Fill.Cfg:\n Number of SDT signals <", cfg.new$n.sdt, "> smaller than <2>\n(Minimum number = 2)\n", sep = ""))
     if(any(cfg.new$j.rg < 2))
            stop(paste("\n FATAL ERROR in HTSDT.Bias.Rank.Check.Fill.Cfg: Number of estimated position <", cfg.new$j.rg, "> smaller than <2>\n (Minimum number = 2)\n", sep = ""))
      if(length(cfg.new$k.rg) != length(cfg.new$j.rg))
            stop(paste("\n FATAL ERROR in HTSDT.Bias.Rank.Check.Fill.Cfg:\n Different number of ranking models for k.rg <", length(cfg.new$k.rg), "> and j.rg <",length(cfg.new$j.rg), ">\n", sep = ""))
      if(any(cfg.new$k.rg < cfg.new$j.rg)) 
            stop(c("\n FATAL ERROR in HTSDT.Bias.Rank.Check.Fill.Cfg:\n At least one number of alternatives smaller than one number of relevant ranks\n"))
      return(cfg.new)
}                  
     
# ------------------------------------------------------------------------------------------------------------------
# Check type of restriction
# Input: A string specfiying the typ of restriction
#        Alternatively the flags TRUE (="standard") and FALSE ("No") are allowed
#        Types of restrictions      - "standard"  [ or T]   (Standard restrictions)
#                                   - "SDT"                 (Simulating SDT model)
#                                   - "no"  [or F]          (No restrictions)
#
# Output: String with parameter values
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Restriction <- function(restriction)
{
      if(is.null(restriction) || restriction == T)
            return("standard")
      if(charmatch(restriction, c("standard", "STANDARD", "Standard"), nomatch = 0) > 0)
            return("standard")
      if(charmatch(restriction, c("sdt", "SDT", "Sdt"), nomatch = 0) > 0)
            return("SDT")
      if(charmatch(restriction, c("extended", "EXTENDED", "Extended"), nomatch = 0) > 0)
            return("extended")
      if(charmatch(restriction, c("no", "No", "NO"), nomatch = 0) > 0 || restriction == F)
            return("no")
      warn.msg <- paste("\n\n HTSDT.Bias.Restriction: Unkonwn restriction in configuration list: <", restriction, "> --> setting to: <standard>\n\n", sep = "")      
      warning(warn.msg)
      return("standard")
}            
                        
# ------------------------------------------------------------------------------------------------------------------
# Generate starting parameters due to data and number of signals
# Input:	cfg = Configuration structure: A list with the following entries
#                 - k.rg      Number of alternatives
#                 - j.rg      Number of relevant ranking positions
#                 - npt       Number of quadrature points
#                 -standard   Flag indicating whether standard restrictions have to be specified
#                 -rating     Flag indicating whether ranking and rating are combined
#           data = Data vector
#
# Output: String with parameter values depending on whether only ranking data or
#         ranking as well as reting data are modeled
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.start.par <- function(cfg, datavec)
{
      # If configuration information is not complete, then complete it
      if(is.null(cfg$rating) || is.null(cfg$n.sdt) || is.null(cfg$j.rg) || is.null(cfg$k.rg))
          cfg <- HTSDT.Rank.Check.Fill.Cfg(cfg)  
      
      #-------------------------------------------------------------------------
      # I. Parameters RANKING MODEL
      # 1. p.recollection
      # 2. p.guessing
      # 3. m.new
      # 4. sd.new
      # 5. m.old
      # 6. sd.old
      # 7. Bias parameters: b[1],...b[k]
      #
      # II. Parameters RATING MODEL 1
      # 1. p.recollection
      # 2. p.guessing
      # 3. p. bias (guessing)
      # 4. m.new
      # 5. sd.new
      #
      # III. Parameters RATING MODEL 2
      # 1. p.recollection
      # 2. p.guessing
      # 3. p. bias (guessing)
      # 4. m.old
      # 5. sd.old
      #
      # IV. Threshold parameter
      # c[1], ..., c[k]
      #-------------------------------------------------------------------------
      # Compute ranking starting parameters
      ranking.par <- NULL
      for(i in 1:length(cfg$k.rg))
            ranking.par <- c(ranking.par, c(0.0, 0.0, 0.0, 1.0, 0.0, 1.0), rep(0.0, cfg$k.rg[i]))
      if(!cfg$rating)          # Ranking data only
            return(ranking.par)

      if(is.null(datavec)) stop("\n FATAL ERROR in HTSDT.Bias.Rank.start.par: Cannot compute start parameters due to missing data vector\n")
      n.th <- (length(datavec) - sum(cfg$j.rg * cfg$k.rg))/cfg$n.sdt  - 1    # Number of thresholds
      if(n.th <= 0) 
      {
            msg <- paste("\n\n FATAL ERROR in HTSDT.Bias.Rank.start.par: Not enough data points for estimating specified rating and ranking model:\n",
                         " Length of data vector  : <", length(datavec), ">\n",
                         " Minimum length required: <", sum(cfg$j.rg * cfg$k.rg) + 2*cfg$n.sdt, ">\n\n", sep = "")
            stop(msg)
      }
      c(ranking.par,                              # Parameters of the ranking models
        rep(c(0, 0, 0, 0, 1), cfg$n.sdt),         # Parameters of the sdt     models
        seq(-0.5, 0.5, length = n.th))            # Thresholds
}      

# ------------------------------------------------------------------------------------------------------------------
# Generate upper and lower bounds of parameter
# Input:	parvec = Vector of parameters
#		   cfg = configuration structure
#
# Output: List with lower and upper bounds on parameter vector
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.par.bounds <- function(parvec, cfg)
{
      min.p <- -700
      max.p <-  700

      len <- length(cfg$k.rg)
      
      lower <- upper <- NULL
      for(i in 1:length(cfg$k.rg))
      {
            lower <- c(lower, c(min.p, min.p, -Inf, HTSDT.Bias.Rk.Tol, -Inf, HTSDT.Bias.Rk.Tol), rep(-Inf, cfg$k.rg[i]))
            upper <- c(upper, c(max.p, max.p,  Inf,               Inf,  Inf,               Inf), rep( Inf, cfg$k.rg[i]))
      }
      if(cfg$rating == F)
            return(list(lower = lower, upper = upper))
      
      n.th <- length(parvec) - 6 * len - sum(cfg$k.rg) - 5 * cfg$n.sdt  # Number of thresholds
      lower <- c(lower, rep(c(min.p, min.p, min.p, -Inf, HTSDT.Bias.Rk.Tol), cfg$n.sdt), rep(-Inf, n.th))
      upper <- c(upper, rep(c(max.p, max.p, max.p,  Inf,               Inf), cfg$n.sdt), rep( Inf, n.th))
      
      return(list(lower = lower, upper = upper))
}      

# ------------------------------------------------------------------------------------------------------------------
# Generate strings with names of the parameters for the models
# Input:	parvec         = Vector of parameters
#		   cfg         = configruation structure
#           with.space     = Flag indicating whether names should be with space (default) or not
# Output: String with parameter names
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.par.names <- function(parvec, cfg, with.space = T)
{
      # If configuration information is not complete, then complete it
      if(is.null(cfg$rating) || is.null(cfg$n.sdt) || is.null(cfg$j.rg) || is.null(cfg$k.rg))
          cfg <- HTSDT.Rank.Check.Fill.Cfg(cfg)  
      
      if(with.space)
      {
            Ranking.names <- c("Recollection    (Rank", 
                               "Guessing        (Rank", 
                               "Mean.1          (Rank", 
                               "Stddev.1        (Rank", 
                               "Mean.2          (Rank", 
                               "Stddev.2        (Rank")
                         
            Rating.names  <- c("Recollection    (Gauss", 
                               "Guessing        (Gauss", 
                               "Bias [guessing] (Gauss", 
                               "Mean            (Gauss", 
                               "Stddev          (Gauss")
      }
      else
      {
            Ranking.names <- c("Recollection (Rank", 
                               "Guessing (Rank", 
                               "Mean.1 (Rank", 
                               "Stddev.1 (Rank", 
                               "Mean.2 (Rank", 
                               "Stddev.2 (Rank")
                         
            Rating.names  <- c("Recollection (Gauss", 
                               "Guessing (Gauss", 
                               "Bias [guessing] (Gauss", 
                               "Mean.1 (Gauss", 
                               "Stddev.1 (Gauss")
      }
      Rk.names <- NULL
      for(i in 1:length(cfg$k.rg))
      {
            for(j in 1:length(Ranking.names))
                  Rk.names <- c(Rk.names, paste(Ranking.names[j], " ", i, ")", sep = ""))    # Gaussian parameters
            bias.name <- paste("Bias.", 1:cfg$k.rg[i], ifelse(with.space, "          (Rank ", " (Rank "), i, ")", sep = "")
            Rk.names  <- c(Rk.names, bias.name)
      }      
#      cat("\nRk.names =\n"); print(matrix(Rk.names, nc = 1))
      
      if(cfg$rating)
      {
            models <- rep(1:cfg$n.sdt, each = 5)
            Rating.names <- paste(Rating.names, " ", models, ")", sep = "")
            R.names <- c(Rk.names, Rating.names)
            n.th <- length(parvec) - length(R.names)
            return(c(R.names, paste("c[", 1:n.th, "]", sep = "")))
      }
      return(Rk.names)
}

# ------------------------------------------------------------------------------------------------------------------
# Names of the models 
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.model.name <- function(cfg)
{
      # If configuration information is not complete, then complete it
      if(is.null(cfg$rating) || is.null(cfg$n.sdt) || is.null(cfg$j.rg) || is.null(cfg$k.rg))
          cfg <- HTSDT.Rank.Check.Fill.Cfg(cfg)  

      
      k.rg.s <- paste(cfg$k.rg, collapse = ", ")
      j.rg.s <- paste(cfg$j.rg, collapse = ", ")

      if(cfg$rating)
            mod.name <- paste("SDT ranking and rating model with recollection and guessing, with position bias:", 
                              " Number of Gaussian models: <", cfg$n.sdt, ">", 
                              " Number of ranking-alternatives: <", k.rg.s, ">", 
                              " Number of ranking-positions: <", j.rg.s, ">", 
                              " Number of quadrature points: <", cfg$npt, ">", sep = "")
      else
            mod.name <- paste("SDT ranking model with recollection and guessing, with position bias:", 
                              " Number of ranking-alternatives: <", k.rg.s, ">", 
                              " Number of ranking-positions: <", j.rg.s, ">", 
                              " Number of quadrature points: <", cfg$npt, ">", sep = "")
      
      return(paste(mod.name, " Type of restrictions: <", cfg$restriction, ">", sep = ""))
}       


# ------------------------------------------------------------------------------------------------------------------
# Check number of parameters and data points
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.check <- function(parvec, datavec, cfg)
{
      Msg <- "NOTE: Computation of the expected number of parameters is based on the number of data points and vice versa"
	
      plen     <- length(parvec)
      dlen     <- length(datavec)
      
      # Case 1: Ranking and Rating
      if(cfg$rating)
      {
            n.th     <- (dlen  - sum(cfg$j.rg * cfg$k.rg))/cfg$n.sdt - 1            # Number of thresholds
            plen.dat <- n.th + 6*length(cfg$j.rg) +  sum(cfg$k.rg) + 5*cfg$n.sdt    # Expected parameter length given the data vector
            dlen.par <- sum(cfg$j.rg * cfg$k.rg) + cfg$n.sdt*(n.th + 1)             # Expected data length given the parameter vector
      }

      # Case 2: Ranking only
      else
      {
            plen.dat <- 6 * length(cfg$j.rg) + sum(cfg$k.rg)
            dlen.par <- sum(cfg$j.rg * cfg$k.rg)
      }
      
	Erg <- matrix(c(plen, dlen, plen.dat, dlen.par), nc = 2)
	rownames(Erg) <- c("Parameters", "Data points")
	colnames(Erg) <- c("Found", "Expected")
	Error <- any(Erg[, 1] != Erg[, 2])

	list(Error = Error, Stat = Erg, Msg = Msg)
}

# ------------------------------------------------------------------------------------------------------------------
# MIX Ranking and Rating Model: Set restrictions
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.Restriction <- function(cfg, fixed, ident)
{
      # If configuration information is not complete, then complete it
      cfg <- HTSDT.Bias.Rank.Check.Fill.Cfg(cfg)  

      if(cfg$restriction == "standard")
            return(HTSDT.Bias.Rank.Set.Standard.Restriction(cfg, fixed, ident))
      if(cfg$restriction == "SDT")
            return(HTSDT.Bias.Rank.Set.SDT.Restriction(cfg, fixed, ident))
       if(cfg$restriction == "extended")
            return(HTSDT.Bias.Rank.Set.Extended.Restriction(cfg, fixed, ident))
      return(list(fixed = fixed, ident = ident))
}

# ------------------------------------------------------------------------------------------------------------------
# Ranking and Rating Model with Recollection and Guessing: Set standard restrictions
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.Set.Standard.Restriction <- function(cfg, fixed, ident)
{
      n.rating <- length(cfg$k.rg)
      n.sdt    <- cfg$n.sdt
      
      # ----- Equality constraints on ranking models ------
      # 1. recollection = recollection (Rank 1)     (Parameter 1)
      # 2. mean 2       = mean 2       (Rank 1)     (Parameter 5)
      # ----------------------------------------------------------
      if(n.rating > 1)
      {
            source.id.rank <- rep(c(1, 5), n.rating-1)
            target.id.rank <- c(1, 5) + rep(cumsum(6 + cfg$k.rg[-n.rating]), each = 2)
      }
      else
            source.id.rank <- target.id.rank <- NULL
     # ---- Fixed constraints for each rating model -------------
      # 1. Guessing probability = 0.0           (Parameter 2)
      # 2. Mean 1               = 0.0           (Parameter 3)
      # 3. Stddev 1             = 1.0           (Parameter 4)
      # 4. Stddev 2             = 1.0           (Parameter 6)
      # 5. bias[k]              = 0.0           (parameter 6 + k)
      # ----------------------------------------------------------
      pos.fx.rank <- c(2, 3, 4, 6)
      if(n.rating > 1)
            pos.fx.rank <- c(pos.fx.rank, pos.fx.rank + rep(cumsum(6 + cfg$k.rg[-n.rating]), each = 4))
      val.fx.rank <- rep(c(0.0, 0.0, 1.0, 1.0), n.rating)
      
      # Bias parameters
      pos.fx.bias <- cumsum(6 + cfg$k.rg)             # Constraint on last bias parameters
      val.fx.bias <- rep(0.0, length(cfg$k.rg))
      pos.fx.rank <- c(pos.fx.rank, pos.fx.bias)
      val.fx.rank <- c(val.fx.rank, val.fx.bias)
      
      if(cfg$rating == F)
      {
            # Build fixed matrix and remove redundant restrictions
            fixed.s <- matrix(c(val.fx.rank, pos.fx.rank), nr = 2, byrow = T)
            pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
      	fixed <- cbind(fixed.s[, pos.unique], fixed)   
            # Build identity matrix and remove redundant restrictions
            if(!is.null(source.id.rank))
            {
                  ident.s <- matrix(c(source.id.rank,
                                      target.id.rank), nr = 2, byrow = T)
                  pos.unique <-  !(ident.s[2,] %in% ident[2,])
	            ident <- cbind(ident.s[, pos.unique], ident)
            }
            return(list(fixed = fixed, ident = ident))
      }
     #-------Fixed constraints (Rating) ---------------------------------------
      # FIRST SDT MODEL
      # 1. Recollection          (Rating 1) = 0.0     (Parameter 1)
      # 2. Guessing probability  (Rating 1) = 0.0     (Parameter 2)
      # 3. Bias guessing         (Rating 1) = 0.5     (Parameter 3)
      # 4. Mean                  (Rating 1) = 0.0     (Parameter 4)
      # 5. Stddev1               (Rating 1) = 1.0     (Parameter 5)
      # ------------------------------------------------------------------------
      # FURTHER SDT MODEL
      # 1. Guessing probability  (Rating 1) = 0.0     (Parameter 2)
      # 2. Bias guessing         (Rating 1) = 0.5     (Parameter 3)
      # 3. Stddev                           = 1.0     (Parameter 5)
      # ------------------------------------------------------------------------
      pos.fx.rating.1 <- c(1,  2,   3,  4,  5) + 6*n.rating + sum(cfg$k.rg)
      val.fx.rating.1 <- c(0,  0, 0.5,  0,  1)
      
      pos.fx.rating.2 <- c(2, 3, 5) + 6*n.rating + sum(cfg$k.rg) + rep(5*(1:(n.sdt-1)), each = 3)
      val.fx.rating.2 <- rep(c(0.0, 0.5, 1.0), n.sdt-1)

      fixed.s <- matrix(c(val.fx.rank, val.fx.rating.1, val.fx.rating.2,
                          pos.fx.rank, pos.fx.rating.1, pos.fx.rating.2), nr = 2, byrow = T)
      pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
	fixed <- cbind(fixed.s[, pos.unique], fixed)            # Add to fixed if not alread specified
	
      #------- Identity constraints (Rating) -----------------------------------
      # - Recollection (Ranking) = Recollection (Rating 2): 1 -> 1
      # - Mean2        (Ranking) = Means2       (Rating 2): 5 -> 4
      # ------------------------------------------------------------------------
      source.id.rating <- rep(c(1, 5), n.sdt - 1)
      target.id.rating <- c(1, 4) + 6*n.rating + sum(cfg$k.rg) + rep(5*(1:(n.sdt-1)), each = 2)
      ident.s <- matrix(c(source.id.rank, source.id.rating,
                          target.id.rank, target.id.rating), nr = 2, byrow = T)
      pos.unique <-  !(ident.s[2,] %in% ident[2,])
	ident <- cbind(ident.s[, pos.unique], ident)            # Add to ident if not alread specified
	
      list(fixed = fixed, ident = ident)
}

# ------------------------------------------------------------------------------------------------------------------
# Ranking and Rating Model with Recollection and Guessing: Set SDT restrictions
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.Set.SDT.Restriction <- function(cfg, fixed, ident)
{
      n.rating <- length(cfg$k.rg)
      n.sdt    <- cfg$n.sdt
      
      # ----- Equality constraints on ranking models ------
      # 1. mean 2       = mean 2       (Rank 1)     (Parameter 5)
      # 2. stddev 2     = stddev 2     (Rank 1)     (Parameter 6)
      # ----------------------------------------------------------
      if(n.rating > 1)
      {
            source.id.rank <- rep(c(5, 6), n.rating - 1)
            target.id.rank <- c(5, 6) + rep(cumsum(6 + cfg$k.rg[-n.rating]), each = 2)
      }
      else
            source.id.rank <- target.id.rank <- NULL
      # ---- Fixed constraints for each rating model -------------
      # 1. Recollection probability = 0.0   (Parameter 1)
      # 2. Guessing probability     = 0.0   (Parameter 2)
      # 3. Mean 1                   = 0.0   (Parameter 3)
      # 4. Stddev 1                 = 1.0   (Parameter 4)
      # 5. Bias[k]                  = 0.0   (Parameter 6 + k.rg)
      # ----------------------------------------------------------
      pos.fx.rank <- c(1, 2, 3, 4)
      if(n.rating > 1)
            pos.fx.rank <- c(pos.fx.rank, pos.fx.rank + rep(cumsum(6 + cfg$k.rg[-n.rating]), each = 4))
      val.fx.rank <- rep(c(0.0, 0.0, 0.0, 1.0), n.rating)
      
      # Bias parameters
      pos.fx.bias <- cumsum(6 + cfg$k.rg)             # Constraint on last bias parameters
      val.fx.bias <- rep(0.0, length(cfg$k.rg))
      pos.fx.rank <- c(pos.fx.rank, pos.fx.bias)
      val.fx.rank <- c(val.fx.rank, val.fx.bias)
     
      if(cfg$rating == F)
      {
            # Build fixed matrix and remove redundant restrictions
            fixed.s <- matrix(c(val.fx.rank, pos.fx.rank), nr = 2, byrow = T)
            pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
            fixed <- cbind(fixed.s[, pos.unique], fixed)
            # Build identity matrix and remove redundant restrictions
            if(!is.null(source.id.rank))
            {
                  ident.s <- matrix(c(source.id.rank,
                                      target.id.rank), nr = 2, byrow = T)
                  pos.unique <-  !(ident.s[2,] %in% ident[2,])
	            ident <- cbind(ident.s[, pos.unique], ident)
            }
            return(list(fixed = fixed, ident = ident))
      }
      #-------Fixed constraints (Rating) ---------------------------------------
      # FIRST SDT MODEL
      # 1. Recollection          (Rating 1) = 0       (Parameter 1)
      # 2. Guessing probability  (Rating 1) = 0       (Parameter 2)
      # 3. Bias guessing         (Rating 1) = 0.5     (Parameter 3)
      # 4. Mean                  (Rating 1) = 0       (Parameter 4)
      # 5. Stddev1               (Rating 1) = 1       (Parameter 5)
      # ------------------------------------------------------------------------
      # FURTHER SDT MODEL
      # 1. Recollection          (Rating 1) = 0       (Parameter 1)
      # 2. Guessing probability  (Rating 1) = 0       (Parameter 2)
      # 3. Bias guessing         (Rating 1) = 0.5     (Parameter 3)
      # ------------------------------------------------------------------------
      pos.fx.rating.1 <- c(1,  2,   3,  4,  5) + 6*n.rating + sum(cfg$k.rg)
      val.fx.rating.1 <- c(0,  0, 0.5,  0,  1)
      
      pos.fx.rating.2 <- c(1, 2, 3) + 6*n.rating + sum(cfg$k.rg) + rep(5*(1:(n.sdt-1)), each = 3)
      val.fx.rating.2 <- rep(c(0, 0, 0.5), n.sdt-1)

      fixed.s <- matrix(c(val.fx.rank, val.fx.rating.1, val.fx.rating.2,
                          pos.fx.rank, pos.fx.rating.1, pos.fx.rating.2), nr = 2, byrow = T)
      pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
	fixed <- cbind(fixed.s[, pos.unique], fixed)            # Add to fixed if not alread specified
	
      #------- Identity constraints (Rating) -----------------------------------
      # - Mean2   (Ranking)      = Means2       (Rating 2): 5 = 4
      # - Stddev2 (Ranking)      = Stddev2      (Rating 2): 6 = 5
      # ------------------------------------------------------------------------
      source.id.rating <- rep(c(5, 6), n.sdt - 1)
      target.id.rating <- c(4, 5) + 6*n.rating + sum(cfg$k.rg) + rep(5*(1:(n.sdt-1)), each = 2)
      ident.s <- matrix(c(source.id.rank, source.id.rating,
                          target.id.rank, target.id.rating), nr = 2, byrow = T)
      pos.unique <-  !(ident.s[2,] %in% ident[2,])
	ident <- cbind(ident.s[, pos.unique], ident)            # Add to ident if not alread specified
	
      list(fixed = fixed, ident = ident)
}

# ------------------------------------------------------------------------------------------------------------------
# Ranking and Rating Model with Recollection and Guessing: Set extended restrictions
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.Set.Extended.Restriction <- function(cfg, fixed, ident)
{
      n.rating <- length(cfg$k.rg)
      n.sdt    <- cfg$n.sdt
      
      # ----- Equality constraints on ranking models ------
      # 1. recollection = recollection (Rank 1)     (Parameter 1)
      # 2. mean 2       = mean 2       (Rank 1)     (Parameter 5)
      # 3. stddev 2     = stddev 2     (Rank 1)     (Parameter 6)
      # ----------------------------------------------------------
      if(n.rating > 1)
      {
            source.id.rank <- rep(c(1, 5, 6), n.rating - 1)
            target.id.rank <- c(1, 5, 6) + rep(cumsum(6 + cfg$k.rg[-n.rating]), each = 3)
      }
      else
            source.id.rank <- target.id.rank <- NULL
      # ---- Fixed constraints for each rating model -------------
      # 1. Guessing probability = 0.0     (Parameter 2)
      # 2. Mean 1               = 0.0     (Parameter 3)
      # 3. Stddev 1             = 1.0     (Parameter 4)
      # 5. Bias[k]              = 0.0     (Parameter 6 + k.rg)
      # ----------------------------------------------------------
      pos.fx.rank <- c(2, 3, 4) + rep(6*(0:(n.rating-1)), each = 3) 
      val.fx.rank <- rep(c(0, 0, 1), n.rating)

      pos.fx.rank <- c(2, 3, 4)
      if(n.rating > 1)
            pos.fx.rank <- c(pos.fx.rank, pos.fx.rank + rep(cumsum(6 + cfg$k.rg[-n.rating]), each = 3))
      val.fx.rank <- rep(c(0.0, 0.0, 1.0), n.rating)
      
      # Bias parameters
      pos.fx.bias <- cumsum(6 + cfg$k.rg)             # Constraint on last bias parameters
      val.fx.bias <- rep(0.0, length(cfg$k.rg))
      pos.fx.rank <- c(pos.fx.rank, pos.fx.bias)
      val.fx.rank <- c(val.fx.rank, val.fx.bias)

      if(cfg$rating == F)
      {
            # Build fixed matrix and remove redundant restrictions
            fixed.s <- matrix(c(val.fx.rank, pos.fx.rank), nr = 2, byrow = T)
            pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
      	fixed <- cbind(fixed.s[, pos.unique], fixed)   
            # Build identity matrix and remove redundant restrictions
            if(!is.null(source.id.rank))
            {
                  ident.s <- matrix(c(source.id.rank,
                                      target.id.rank), nr = 2, byrow = T)
                  pos.unique <-  !(ident.s[2,] %in% ident[2,])
	            ident <- cbind(ident.s[, pos.unique], ident)
            }
            return(list(fixed = fixed, ident = ident))
      }
      #-------Fixed constraints (Rating) ---------------------------------------
      # FIRST SDT MODEL
      # 1. Recollection          (Rating 1) = 0.0     (Parameter 1)
      # 2. Guessing probability  (Rating 1) = 0.0     (Parameter 2)
      # 3. Bias guessing         (Rating 1) = 0.5     (Parameter 3)
      # 4. Mean                  (Rating 1) = 0.0     (Parameter 4)
      # 5. Stddev1               (Rating 1) = 1.0     (Parameter 5)
      # ------------------------------------------------------------------------
      # FURTHER SDT MODEL
      # 1. Guessing probability  (Rating 1) = 0.0     (Parameter 2)
      # 2. Bias guessing         (Rating 1) = 0.5     (Parameter 3)
      # ------------------------------------------------------------------------
      pos.fx.rating.1 <- c(1,  2,   3,  4,  5) + 6*n.rating + sum(cfg$k.rg)
      val.fx.rating.1 <- c(0,  0, 0.5,  0,  1)
      
      pos.fx.rating.2 <- c(2, 3) + 6*n.rating + sum(cfg$k.rg) + rep(5*(1:(n.sdt-1)), each = 2)
      val.fx.rating.2 <- rep(c(0.0, 0.5), n.sdt-1)

      fixed.s <- matrix(c(val.fx.rank, val.fx.rating.1, val.fx.rating.2,
                          pos.fx.rank, pos.fx.rating.1, pos.fx.rating.2), nr = 2, byrow = T)
      pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
	fixed <- cbind(fixed.s[, pos.unique], fixed)            # Add to fixed if not alread specified
	
      #------- Identity constraints (Rating) -----------------------------------
      # 1. Recollection (Ranking) = Recollection (Rating 2): 1 -> 1
      # 2. Mean2        (Ranking) = Means2       (Rating 2): 5 -> 4
      # 3. Mean2        (Ranking) = Means2       (Rating 2): 6 -> 5
      # ------------------------------------------------------------------------
      source.id.rating <- rep(c(1, 5, 6), n.sdt - 1)
      target.id.rating <- c(1, 4, 5) + 6*n.rating + sum(cfg$k.rg) + rep(5*(1:(n.sdt-1)), each = 3)
      ident.s <- matrix(c(source.id.rank, source.id.rating,
                          target.id.rank, target.id.rating), nr = 2, byrow = T)
      pos.unique <-  !(ident.s[2,] %in% ident[2,])
	ident <- cbind(ident.s[, pos.unique], ident)            # Add to ident if not alread specified
	
      list(fixed = fixed, ident = ident)
}

# ------------------------------------------------------------------------------------------------------------------
# Format free parameters and compute confidence intervals
# Input:	1. px2 Matrix with parameters and standard errors
#		2. Full parameter vector
#           3. List with configuration information
#		4. Matrix of fixed constraints
#		5. Matrix of equality constraints
#		6. Confidence level
#           7. Number of decimal places
#
# Output: px5 Matrix of parameters, standard errors and confidence limits
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.Format.Par.est <- function(Pars, full.par, cov.matrix, cfg, fixed, ident, alpha, deci)
{
      # 1. Adjustment of sigma: Replace by absolute values
      pos.sigma <- HTSDT.Bias.pos.par(cfg, Par.Id = "Sigma")
      full.par[pos.sigma] <- abs(full.par[pos.sigma])            

      # 2. Compute probabilities
      full.par <- HTSDT.Bias.Rk.Ppars(full.par, cfg, fixed, ident)
      Pars[, 1] <- Elim.par(full.par, fixed, ident)

      # Check for correct cov.matrix
      if(is.null(cov.matrix))
      {
            NAs       <- rep(NA, dim(Pars)[1])
            Pars[,2]  <- NAs
            Pars <- cbind(Pars, NAs, NAs)
            colnames(Pars) <- c("Value", "SE", paste("CFI-", alpha*100, "(Lower)", sep = ""), paste("CFI-", alpha*100, "(Upper)", sep = ""))
            return(round(Pars, deci))
      }
      # Adjust for standard errors for probability parameter
      Pars[,2] <- HTSDT.Bias.Rank.Delta.Method.Prob(cov.matrix, full.par, cfg, fixed, ident)
      qt.cfi <- qnorm((1-alpha)/2)
      cfi.lo <- Pars[,1] + Pars[,2]*qt.cfi
      cfi.up <- Pars[,1] - Pars[,2]*qt.cfi
            
      # Compute indicex with restricted lower bounds & upper bounds            
      index.p <- HTSDT.Bias.pos.par(cfg, Par.Id = "Prob")          # restricted lower and upper bound
      index.s <- HTSDT.Bias.pos.par(cfg, Par.Id = "Sigma")         # restricted lower bound
      index.0 <- c(index.p, index.s)
      
      # Compute lower bounds (= for probabilities and standard errors of the distributions)            
      bd.lo <- rep(-Inf, length(full.par))
      bd.lo[index.0] <- 0
      bd.lo <- Elim.par(bd.lo, fixed, ident)
      cfi.lo <- ifelse(cfi.lo < bd.lo, bd.lo, cfi.lo)

      # Compute higher bounds (= for probability parameters)
      bd.hi <- rep(Inf, length(full.par))
      bd.hi[index.p] <- 1
      bd.hi <- Elim.par(bd.hi, fixed, ident)
      cfi.up <- ifelse(cfi.up > bd.hi, bd.hi, cfi.up)
        
      Pars <- cbind(Pars, cfi.lo, cfi.up)
      colnames(Pars) <- c("Value", "SE", paste("CFI-", alpha*100, "(Lower)", sep = ""), paste("CFI-", alpha*100, "(Upper)", sep = ""))
      round(Pars, deci)
}

# ------------------------------------------------------------------------------------------------------------------
# Format full parameters
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Format.Par.Rank <- function(full.par, cfg, deci, fixed, ident)
{
      Ranking.names <- c("Recollection",
                         "Guessing", 
                         "Mean.1", 
                         "Stddev.1", 
                         "Mean.2", 
                         "Stddev.2")
                         
      Rating.names  <- c("Recollection",
                         "Guessing", 
                         "Bias (guessing)",
                         "Mean", 
                         "Stddev")
      
      # 1. Number of ranking models
      lrank  <-  length(cfg$j.rg)
  
      # 2. Adjustment of sigma: Replace by absolute values
      pos.sigma <- HTSDT.Bias.pos.par(cfg, Par.Id = "Sigma")
      full.par[pos.sigma] <- abs(full.par[pos.sigma])            

      # 3. Compute probabilities
      full.par <- HTSDT.Bias.Rk.Ppars(full.par, cfg, fixed, ident)

      # 4. Extract Ranking parameters:
      R.pars <- HTSDT.Bias.pos.par(cfg, Par.Id = "Rank")
      Ranking.pars <- round(full.par[R.pars], deci)
      
      max.bias <- max(cfg$k.rg)                 # Maximal number of bias parameters
      ed <- 0
      Ranking.df <- NULL
      format.str <- paste("%.", deci, "f", sep = "")  # Format string for sprintf
      for(i in 1:lrank)
      {
            st <- ed + 1
            ed <- ed + 6 + cfg$k.rg[i]
            Rk <- NULL
            for(j in st:ed)
            Rk <- c(Rk, sprintf(format.str, Ranking.pars[j]))
            filler <- rep(" ", max.bias - cfg$k.rg[i])
            Ranking.df <- cbind(Ranking.df, c(Rk, filler))
      }
      Ranking.df <- data.frame(Ranking.df)
            
      cf.matrix <- matrix(c(cfg$k.rg, cfg$j.rg), nr = 2, byrow = T)
      cf.vector <- apply(cf.matrix, 2, fkt <- function(x) { paste(x, collapse = ", j=") })
      colnames(Ranking.df) <- paste("Rank <k=", cf.vector, ">", sep = "")
     	rownames(Ranking.df) <- c(Ranking.names, paste("Bias.", 1:max.bias, sep = ""))
      if(cfg$rating == F)
            return(list(Ranking = Ranking.df))

      # 5. Extract rating parameters
      lrat <-  cfg$n.sdt
	th   <- round(full.par[(6*lrank + sum(cfg$k.rg) + 5*lrat + 1):length(full.par)], deci)	# Isolate thresholds
	n.t  <- length(th)
	th   <- matrix(rep(th, lrat), nc = lrat)     # Copy thresholds and format them
	
	index.rt    <- 1:5 + 6*lrank + sum(cfg$k.rg) + rep(5*(0:(lrat-1)), each = 5)  # Index of parameters for rating
	Rating.pars <- matrix(round(full.par[index.rt], deci), nc = lrat)
	Rating.pars <- rbind(Rating.pars, th)

	colnames(Rating.pars) <- paste("Gauss-", 1:lrat, sep = "")
	rownames(Rating.pars) <- c(Rating.names, paste("t-", 1:n.t, sep = ""))
	list(Ranking = Ranking.df, Rating = Rating.pars)
}

# ------------------------------------------------------------------------------------------------------------------
# Ranking Model with recollection and guessing: Computation of observed p values for ranking model
# Input:    datavec = Vector with data
#           cfg     = configuration information
#
# Output:   Vector with probabilities
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.Compute.Obs.p <- function(datavec, cfg)
{
      p.rank <- NULL
      ed <- 0
      for(i in 1:length(cfg$k.rg))
      {
            st <- ed + 1
            ed <- st + cfg$j.rg[i]*cfg$k.rg[i] - 1
            rank.part <- matrix(datavec[st:ed], nc = cfg$k.rg[i], byrow = T)
            p.rank    <- c(p.rank, as.vector(apply(rank.part, 1, function(x) { x/sum(x) })))
      }
      if(cfg$rating == F) 
            return(p.rank)
      
      sdt.part <- matrix(datavec[-(1:sum(cfg$j.rg * cfg$k.rg))], nr = cfg$n.sdt, byrow = T)     # Extract rating data to matrix with 2 rows
      p.sdt <- apply(sdt.part, 1, function(x) { x/sum(x) })
      c(p.rank, as.vector(p.sdt))
}

# ------------------------------------------------------------------------------------------------------------------
# Computation of observed p values for ranking model
# Input:    datavec = Vector with data
#           cfg     = configuration information
#
# Output:   Vector with observed total counts for each block of data
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.Compute.Obs.N <- function(datavec, cfg)
{
      # N for ranking models      
      N.rank <- NULL
      ed <- 0
      for(i in 1:length(cfg$k.rg))
      {
            st <- ed + 1
            ed <- st + cfg$j.rg[i]*cfg$k.rg[i] - 1
            rank.part  <- matrix(datavec[st:ed], nc = cfg$k.rg[i], byrow = T)
            N.rank.act <- rowSums(rank.part)
            N.rank     <- c(N.rank, rep(N.rank.act, each = cfg$k.rg[i]) )
      }
      if(cfg$rating == F) return(N.rank)

      # N for SDT models      
      sdt.part <- matrix(datavec[-(1:sum(cfg$j.rg * cfg$k.rg))], nr = cfg$n.sdt, byrow = T)     # Extract rating data to matrix with n.sdt rows
      N.rating <- rowSums(sdt.part)
      c(N.rank, rep(N.rating, each = dim(sdt.part)[2]))
}

# ------------------------------------------------------------------------------------------------------------------
# Generation of data labels in output
# Input:    cfg     = configuration information
#           n.dat   = length of the data vector
#
# Output:   Vector with probabilities
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Bias.Rank.Datalabels <- function(cfg, n.dat)
{
      # 1. Labels for ranking data
      label.out <- NULL
      l.rk <- length(cfg$j.rg)
      label.Rk <- paste("Rank", 1:l.rk)
      for(i in 1:l.rk)
            label.out <- c(label.out, paste(label.Rk[i], " [", 1:(cfg$j.rg[i] * cfg$k.rg[i]), "]", sep = ""))
      if(cfg$rating == F) return(label.out)

      # 2. Labels for rating data      
      n.cat <- (n.dat - sum(cfg$j.rg * cfg$k.rg)) / cfg$n.sdt          # Number of response categories
      label.Rt <- paste(" SDT", 1:cfg$n.sdt)
      for(i in 1:cfg$n.sdt)
            label.out <- c(label.out, paste(label.Rt[i], " [", 1:n.cat, "]", sep = ""))
      return(label.out)
}