# ===========================================================================
# HTSDT-Rank models: High threshold signal detection model with ranking
#                    and with guessing
# Date of creation:	    June,     2013
# Date of last update:	Nebember, 2020
#
# Author: Siegfried Macho
# ===========================================================================
options(warn = - 1)
library(statmod)
library(npmlreg)
options(warn = 0)     

HTSDT.Rk.Tol <- 1e-12
MINWEIGHT    <- 1e-40

#===============================================================================
# 0. COMPUTATION OF QUADRATURE POINTS
#===============================================================================

#-------------------------------------------------------------------------------
# Function for computing the Gauss-Hermite points and weights
# Input:	1. Number of points
#           2. minimal weights
#
# Output: 	Matrix with three columns:    1. Points
#                                         2. Weights of Gauss-Hermite
#                                         3. Weights of adaptive Gauss-Hermite 
#
# Comment:  The points and weights are with respect to the Gaussian Kernel N(0, 1)
#-------------------------------------------------------------------------------
Generate.GH <- function(npt, minweight = MINWEIGHT)
{
      GH.Structure  <- gqz(numnodes = npt, minweight)   # Points and weights for Gauss-Hermite Quadrature
      cbind(GH.Structure, GH.Structure[,2] * exp(GH.Structure[,1]^2/2))	 # Weights for adaptive Gauss-Hermite Quadrature
}

#===============================================================================
# 1. PARAMETER POSITIONS
#===============================================================================

# ----------------------------------------------------------------------------------------------------------------------
# HELP FUNCTION
# Input:    cfg     = configuration information
# Output:   A list with three entries:
#           (1) AFC = A 3xn matrix with positions of the recollection parameters in the first,
#                     the gesing parameters in the second, and the d' in the third row (AFC model)
#           (2) SDT = A 3xn matrix with positions of the recollection probabilities means in the first row,
#                     guessing parameters inf the second row, and means in the third row (SDT model)
#           (3) TH  = A list of thresholds for the single SDT models
# ----------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.pos.par_ <- function(cfg)
{
      # 1. Positions of means and sd for AFC
      n.rk <- length(cfg$k.rg)   # Number of ranking models
      pos.afc <- NULL
      for(i in 1:n.rk)
         pos.afc <- c(pos.afc, 1:5 + (i-1)*5)   
      pos.afc <- matrix(c(pos.afc), nr = 5)
      colnames(pos.afc) <- paste("AFC-", 1:n.rk, sep = "")
      rownames(pos.afc) <- c("rec-p", "mu.1", "sigma.1", "mu.2", "sigma.2")
      
      # Postions of means and sd for SDT rating models
      if(cfg$rating == F) return(list(AFC = pos.afc, SDT = NULL, TH = NULL))

      # 2. Positions of means and sd for SDT
      offset <- n.rk*5 + 1       
      pos.sdt <- list(NULL)
      pos.th <- list(NULL)
      for(i in 1:ncol(cfg$n.sdt)) {
          pos.r.sdt <- offset +     (0:(cfg$n.sdt[1, i] - 1) * 3)
          pos.d.sdt <- offset + 1 + (0:(cfg$n.sdt[1, i] - 1) * 3)
          pos.s.sdt <- offset + 2 + (0:(cfg$n.sdt[1, i] - 1) * 3)
          pos.sdt[[i]] <- matrix(c(pos.r.sdt, pos.d.sdt, pos.s.sdt), nr = 3, byrow = T)
          colnames(pos.sdt[[i]]) <- paste("SDT-", 1:ncol(pos.sdt[[i]]), sep = "")
          rownames(pos.sdt[[i]]) <-  c("rec-p", "mean", "sigma")
          
          pos.th[[i]] <- offset + 3*cfg$n.sdt[1, i] + 0:(cfg$n.sdt[2, i] - 2)
          
          if(i < ncol(cfg$n.sdt))
             offset <- offset + 3*cfg$n.sdt[1, i] + cfg$n.sdt[2,i] - 1
      }
      return(list(AFC = pos.afc, SDT = pos.sdt, TH = pos.th))
}

#===============================================================================
# 2. COMPUTATION OF PROBABILITIES
#===============================================================================

# ------------------------------------------------------------------------------
# Computation of recollection probabilities in parameter vector
# Input:	parvec  = parameter vector
#           cfg     = configuration
#           fixed   = matrix of fixed constraints
#           ident   = matrix of idenity constraints
#
# Output: 	Parameter vector with raw probabilities replaced by computed probabilities
# ------------------------------------------------------------------------------
HTSDT.comp.recp <- function(parvec, cfg, fixed, ident)
{
      free.pos <- HTSDT.Rank.Free.Ppos(cfg, fixed)
      if(is.null(free.pos)) return(parvec)
      
      ppar <- parvec[free.pos]
      ppar[ppar > 700] <- 700
      parvec[free.pos] <- exp(ppar)/(1 + exp(ppar))

      return(parvec)
}

# ------------------------------------------------------------------------------------------------------------------
# Help function that returns the positions of the free probability parameters
# Input:	1. cfg    = list with configuration information
#         2. fixed  = matrix of fixed constraints
#
# Output:	vector of positions of free recollection probability parameters or NULL
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.Free.Ppos <- function(cfg, fixed)
{
      # 1. Postions of probability parameters
      n.rk <- length(cfg$k.rg)
      ppos <- 1 + 5*(0:(n.rk - 1))
        
      if(cfg$rating) {
        offset <- 5*n.rk            
        for(i in 1:ncol(cfg$n.sdt)) {
           ppos <- c(ppos, offset + 1 + 3*(0:(cfg$n.sdt[1, i] - 1)))
           offset <- offset + 3*cfg$n.sdt[1, i] + cfg$n.sdt[2,i] - 1
        }   
      }
      # 2. Identify probability parameters that are not fixed
      if(is.null(fixed))
         return(ppos)
         
      free.p   <- !(ppos %in% fixed[2, ])
      if(all(free.p == F)) return(NULL)
      return(ppos[free.p])
}

HTSDT.Rank.Free.Ppos.Delta <- function(parvec, cfg, fixed, ident)
{
     ppos <- HTSDT.Rank.Free.Ppos(cfg, fixed)
     if(length(ppos) == 0) return(NULL)
     
     pos <- SavePos.par(parvec, fixed, ident)  # Positions of free parameters  SDT-Auxiliary
     if(length(pos) == 0) return(NULL)

     free.pos <- (1:length(pos))[pos %in% ppos]
     if(length(free.pos == NULL))
         return(NULL)  # Return positions of free probability parameters with respect to free parameter vector 
     return(free.pos)
}

#===============================================================================
# 3. MODEL FUNCTIONS OF THE RANKING PART OF THE RANKING-RATING MODELL
#===============================================================================

# ----------------------------------------------------------------------
# Computation of conditional ranking probability of the target stimulus
# given the strength value x[1] of the target variable multiplied by
# the quadrature weight x[2]
# Input:	x  = vector with quadrature point x[1] and quadrature weight x[2]
#           j  = position of target variable x within the ranking (j = 1,...,k)
#           k  = Number of alternatives
#           mu =  mean of the non-target variables
#           sigma = standard deviation of the non-target variables
# Output: 	n - Vector, with n being the number of Response categories
# ----------------------------------------------------------------------
HTSDT.Rk.cond  <- function(x, j, k, mu, sigma)
{
      choose(k-1, j-1) * pnorm(x[1], mu, sigma)^(k-j) * (1-pnorm(x[1], mu, sigma))^(j-1) * x[2]
}

# ----------------------------------------------------------------------
# Computation of the unconditional ranking probability of target stimulus
# Input:	j             = position of target variable x within the ranking
#           k             = Number of alternatives
#           mu.target     = mean of the target distribution
#           sigma.targtet = standard deviation of the target distribution
#           mu            = mean of the non-target distribution
#           sigma         = standard deviation of the non-target distribution
#
# Output: 	n - Vector, with n being the number of Response categories
# ----------------------------------------------------------------------
HTSDT.Rk  <- function(j, k, mu.target, sigma.target, mu = 0, sigma =1)
{
      GH.Structure[,1] <- GH.Structure[,1] * sigma.target + mu.target
      sum(apply(GH.Structure, 1, HTSDT.Rk.cond, j = j, k = k, mu = mu, sigma = sigma))
}

#-------------------------------------------------------------------------------
# Computation of ranking probabilties for one model
# Input:
#     rg     = vector with two entries:
#
#     parvec = parameter vector with the following entries
#              1. k.rg:  number of alternatives to be ranked
#              2. j.rg: number of positions to be considered
#              3. probability of recollection
#              4. mean of the new distribution
#              5. sd   of the new distribution
#              6. mean of the old distribution
#              7. sd   of the old distribution
#
# Output: Ranking probabilities
#-------------------------------------------------------------------------------
HTSDT._Rank.1 <- function(parvec)
{
      prb <- sapply(1:parvec[2], HTSDT.Rk, k = parvec[1], parvec[6], parvec[7], parvec[4], parvec[5])
      prb <- (1.0 - parvec[3])*prb
      prb[1] <- prb[1] + parvec[3]
      prb    <- prb

      return(c(prb, 1.0 - sum(prb)))
}

#-------------------------------------------------------------------------------
# Computation of ranking probabilties for all ranking models model
#-------------------------------------------------------------------------------
HTSDT._Rank <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
      # Compile matrix with each row containing: k.rg, j.rg and the parameter vector of the model
      Rg <- matrix(c(cfg$k.rg, cfg$j.rg - 1), nc = 2)
      n  <- dim(Rg)[1]
      Rg <- cbind(Rg, matrix(parvec[1:(5*n)], nr = n, byrow = T))
      
      # Now apply the function HTSDT._Rank.1
      prb <- apply(Rg, 1, HTSDT._Rank.1)

      if(is.list(prb)) return(unlist(prb))
      return(as.vector(prb))
}

#===============================================================================
# 4. MODELL FUNCTIONS OF THE HTSDT-MODEL (RATING PART OF THE RANKING-RATING MODEL)
#===============================================================================
# ----------------------------------------------------------------------
# Function for computing response distribution of SDT model:
# Used for computing the derivative of the rating parameters
# (Function HTSDT.Rank.SDT.Probs.Jacobi)
#
# Input:	Vector of parameters:
#           1. Probability of recollection
#           2. mean
#           3. sd
#           4. c[1], c[2], ...., c[n-1])
#
# Output: 	Vector SDT response probabilities for one signal
# ----------------------------------------------------------------------
HTSDT.SDT.Probs <- function(parvec)
{
      Resp <- pnorm(parvec[4:length(parvec)], mean = parvec[2], sd = max(parvec[3], HTSDT.Rk.Tol)) 	  # Probability of Response-Category 
	    Resp <- c(Resp, 1)							                          # Response vector
	    c(Resp[1], Resp[-1]-Resp[-length(Resp)])				          # Compute Differences
}	

# ------------------------------------------------------------------------------
# Function for creating response distribution for on Type of Stimulus
# Input:	Vector of parameters:
#           1. k = Number indicating the category the recollection probability is assigned to
#           2. par.gauss = vector containing the following parameters
#                           1. Probability of recollection
#                           2. Probability of guessing
#                           3. Bias for guessing
#                           4. mean
#                           4. sd
#           3. par.th = vector of threshold parameters c[1], c[2], ...., c[n-1])
#           4. ncat   = number of response categories
#
# Output: 	Vector of response probabilities of the HTSDT model for one signal
# ------------------------------------------------------------------------------
HTSDT.Probs <- function(k, par.gauss, par.th, ncat)
{
   Resp <- pnorm(par.th, mean = par.gauss[2], sd = max(par.gauss[3], HTSDT.Rk.Tol)) 	   # Probability of Response-Category 
	 Resp <- c(Resp, 1)							                                                       # Response vector
	 Resp <- c(Resp[1], Resp[-1] - Resp[-ncat])				                                     # Compute Differences

   # Treat Recollection: Add recollection probability to the category given in the first position of the parameter vector
   Resp <- (1 - par.gauss[1]) * Resp                                      # Non-Recollection and non-guessing probability
   Resp[k] <- Resp[k] + par.gauss[1]
    return(Resp)
}				

# ------------------------------------------------------------------------------------------------------------------
# Computation or probabilities of SDT model with recollection and rating
# Input:	parvec = Vector of parameters, consisting of the following entries
#                      I.  Parameters of the ranking part (ignored)
#                      II. For each signal of the Rating part
#                          1. Probability of recollection
#                          2. Probability of guessing
#                          3. Bias for guessing
#                          4. mean for each model 
#			         5. standard deviation for each model
#                      II.  Identical for all models
#			             thresholds
#
#		n   	= Number of models
#           pos   = vector of numbers indicating for each model where the 
#                   recollection weight has to be placed
#           fixed = 2xp matrix of fixed    constraints
#           ident = 2xq matrix of equality constraints
#
# Output: Vector of probabilities for single models
# ------------------------------------------------------------------------------------------------------------------
HTSDT.SDT <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
      par.pos <- HTSDT.Rank.pos.par_(cfg)       # List with parameter positions
      sdt.pos <- par.pos$SDT
      th.pos  <- par.pos$TH
      
      probs <- NULL
      for(i in 1:length(sdt.pos)) {
         par.th    <- parvec[th.pos[[i]]]
         ncat <- cfg$n.sdt[2, i]
         rec.cat <- 1
         for(j in 1:ncol(sdt.pos[[i]])) {
            par.gauss <- parvec[sdt.pos[[i]][, j]]
            if(j == 2) rec.cat <- ncat           # for old items recollection to last response category 
            probs <- c(probs, HTSDT.Probs(rec.cat, par.gauss, par.th, ncat))
         }
      }   
      return(probs)
}

#===============================================================================
# 3. MODELL FUNCTIONS OF THE (COMBINED) RANKING-RATING MODEL
#===============================================================================

#-------------------------------------------------------------------------------
# Computation of ranking and rating probabilties
# Input:
#     parvec = parameter vector with the following entries
#              1: mean of the new distribution
#              2: sd   of the new distribution
#              3: mean of the old distribution
#              4: sd   of the old distribution
#     cfg    = list with configuration information containing the following 
#              entries:
#                 k.rg   = number of alternatives to be ranked
#                 j.rg   = number of positions to be considered
#                 npt    = Number of quadrature points
#                 rating = Flag, indicating whether rating data are present
#      fixed = matrix of fixed constraints
#      ident = matrix of equality constraints
#     with.p = flag indicating whether probability parameters should be 
#              computed from raw parameters (default = TRUE)
#
# Output: Ranking and rating probabilities
#-------------------------------------------------------------------------------
HTSDT.Rank <- function(parvec, cfg, fixed = NULL, ident = NULL, with.p = T)
{
      if(is.null(cfg$rating)) rating <- T
      else                    rating <- cfg$rating

      # 1. Compute recollection parameters if required
      if(with.p)
            parvec <- HTSDT.comp.recp(parvec, cfg, fixed, ident)

      # 2. Compute model probabilities
      pr.Rg <- HTSDT._Rank(parvec, cfg, fixed, ident)

      if(cfg$rating == F) return(pr.Rg)
      c(pr.Rg, HTSDT.SDT(parvec, cfg, fixed, ident))
}


# ------------------------------------------------------------------------------------------------------------------
# Model incorporating restrictions
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.Full <- function(parvec, cfg, pos, fixed = NULL, ident = NULL, functional = NULL, with.p = T)
{
      parvec <- Restore.par(parvec, pos, fixed, ident, functional)	# Restore full parameter vector
      HTSDT.Rank(parvec, cfg, fixed, ident, with.p)

}

#===============================================================================
# 4. JACOBIANS FOR THE RANKING PART OF THE RANKING-RATING MODEL
#===============================================================================

#-------------------------------------------------------------------------------
# Derivative of the conditional model function with respect to x, mu and sigma
#-------------------------------------------------------------------------------
SDT.Rk.cond.Jacobi  <- function(x, j, k, mu, sigma)
{
      sigma <- max(sigma, HTSDT.Rk.Tol)
      z  <- (x[1]-mu)/sigma
      fn <- dnorm(z)
      dm <- -fn/sigma			# Derivative of Gaussian with respect to m
	    ds <- -z*fn/sigma		# Derivative of Gaussian with respect to sd
      dx <-  fn/sigma     # Derivative of Gaussian with respect to x

      bin <- choose(k-1, j-1)
      pn  <- pnorm(x[1], mu, sigma)
      p1  <- pn^(k-j)
      p2  <- (1-pn)^(j-1)
      
      if(pn < HTSDT.Rk.Tol)         # 0^0 = 1
      {
            if(k-j-1 == 0)    dp1 <- 1
            else              dp1 <- 0
      }      
      else
            dp1 <- (k-j)*pn^(k-j-1)
            
      if((1-pn) < HTSDT.Rk.Tol)     # 0^0 = 1
      {
            if(j == 2)    dp2 <- 1
            else          dp2 <- 0
      }      
      else
            dp2 <- (j-1)*(1-pn)^(j-2)
      
      bin*(dp1*c(dm, ds, dx)*p2 + p1*c(-dm, -ds, -dx)*dp2)*x[2]
}

#-------------------------------------------------------------------------------
# Derivative of the unconditional SDT model function with respect to:
#     - p.rec           (p.recollection)
#     - mu              (mu.0)
#     - sigma           (sigma.0)
#     - mu.target       (mu.1), 
#     - sigma.target    (sigma.1), 
#
# Input:    j                 actual rank
#           k                 number of ranks
#           p.rec             probability of recollection
#           mu.target         mu.1
#           sigma.targer      sigma.1
#           mu                mu.0
#           sigma             sigma.0
#
# Output: Vector of derviatives in the order shown above
#-------------------------------------------------------------------------------
HTSDT.Rk.Jacobi  <- function(j, k, p.rec, mu.target, sigma.target, mu = 0, sigma = 1)
{
      # 1. Compute of derivatives with respect to Gaussian parameters
      Save.GH.Struc    <- GH.Structure[,1]
      GH.Structure[,1] <- GH.Structure[,1] * sigma.target + mu.target
      J.Rk <- apply(GH.Structure, 1, SDT.Rk.cond.Jacobi, j = j, k = k, mu = mu, sigma = sigma)
      J.Rk <- rbind(J.Rk, J.Rk[3,]* Save.GH.Struc)          # Computation of derivative with respect to sigma.target
      dGauss <- (1 - p.rec) * rowSums(J.Rk)                 # Derivative with respect to Gaussian
      
      # 2. Compute the Gaussian probability
      prb <- HTSDT.Rk(j, k, mu.target, sigma.target, mu, sigma)
      
      # 3. Compute derivative with respect to p.rec
      if(j == 1)  dp.rec <- 1 - prb
      else        dp.rec <- -prb
      
      # 4. Now return full vector
      c(dp.rec, dGauss)
}

#-------------------------------------------------------------------------------
# Jacobian of the model function of the ranking model for a single model:
# Input: 1. parvec = parameter vector of the actual model
#           - p.recollection
#           - mu.0
#           - sigma.0
#           - mu.1
#           -sigma.1
#        2. k.rg   = Number of presented alternatives
#        3. j.rg   = Number of evaluated positions
#        4. fixed  = Matrix of fixed constraints (not used)
#        5. ident  = Matrix of identity constraints (not used)
#
# Jacobian matrix of the probabilities with respect to the six parameters
# of a single ranking model
#-------------------------------------------------------------------------------
HTSDT.__Rank.Jacobi <- function(parvec, k.rg, j.rg, fixed = NULL, ident = NULL)
{
      p.J <- sapply(1:(j.rg-1), HTSDT.Rk.Jacobi, k = k.rg, p.rec = parvec[1], mu.target = parvec[4], 
                                                     sigma.target = parvec[5], mu = parvec[2], sigma = parvec[3])
      rbind(t(p.J), -rowSums(p.J))
}

#-------------------------------------------------------------------------------
# Jacobian of the model function of the ranking model for full ranking model:
# Input: 1. parvec = full parameter vector
#        2. cfg    = Configuration information
#        3. fixed  = Matrix of fixed constraints (not used)
#        4. ident  = Matrix of identity constraints (not used)
#
# Jacobian matrix of the probabilities with respect to the ranking parameters
#-------------------------------------------------------------------------------
HTSDT._Rank.Jacobi <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
      n.rk <- length(cfg$j.rg)
      if(n.rk < 2)                       # Only a single model
            return(HTSDT.__Rank.Jacobi(parvec[1:5], cfg$k.rg, cfg$j.rg, fixed, ident))

      # More than a single model
      Pmatrix <- matrix(parvec[1:(5*n.rk)], nr = n.rk, byrow = T)
      for(i in 1:n.rk)
      {
            J.mod <- HTSDT.__Rank.Jacobi(Pmatrix[i, ], cfg$k.rg[i], cfg$j.rg[i], fixed, ident)
            
            if(i > 1)
                  J <- rbind(cbind(J, matrix(0, nr = sum(cfg$j.rg[1:(i-1)]), nc = 5)), 
                             cbind(matrix(0, nr = cfg$j.rg[i], nc = 5*(i-1)), J.mod))
            else
                  J <- J.mod
      }            
      return(J)
}

#===============================================================================
# 5. JACOBIANS FOR THE RATING PART OF THE RANKING MODEL
#===============================================================================

# ------------------------------------------------------------------------------
# Jacobi-Matrix of the Gaussian parameters and thresholds
# Input:	I. Vector of parameters (without thresholds):
#                 1. Number indicating the category the recollection probability
#                    is assigned to
#                 2. Probability of recollection
#                 3. mean
#                 4. sd
#           II. Threshold parameters: c[1], c[2], ...., c[n-1]
#
# Output: 	Jacobian of the HTSDT model for one signal
# ------------------------------------------------------------------------------
HTSDT.Rank.SDT.Probs.Jacobi <- function(gausspar, th)
{
      # 1. Extract parameters
      rec.cat <- gausspar[1]                    # Recollection category
      p.rec   <- gausspar[2]                    # Recolletion
	    m       <- gausspar[3]                    # Mean parameter
	    sd      <- max(HTSDT.Rk.Tol, gausspar[4]) # Stddev
      
      # 2. Compute derivatives for SDT parameters
	    n.t <- length(th)					# Number of thresholds
	    z  <- (th - m)/sd
	    fn <- dnorm(z)
	
	    # Derivative with respect to mean
	    dm <- -fn/sd						# Derivative of Gaussian with respect to m
	    dm <- c(dm[1], dm[-1] - dm[-n.t], -dm[n.t])
      dm <- (1- p.rec) * dm   # Adjust for recollection

	    # Derivative with respect to sd
	    ds <- -z*fn/sd						# Derivative of Gaussian with respect to sd
	    ds <- c(ds[1], ds[-1] - ds[-n.t], -ds[n.t])
      ds <- (1- p.rec) * ds    # Adjust for recollection


	    # Derivative with respect to thresholds
	    dt <- fn/sd							# Derivative of Gaussian with respect to t[i]
	    dt <- (1- p.rec) * dt   # Adjust for recollection

	
	    J <- matrix(0, nr = n.t+1, nc = n.t)		# Jacobi matrix for thresholds: columns: Parameter, rows = probabilities
	    r <- 1
	    for(i in 1:n.t) {
		      J[r, i] <- dt[i]
		      r <- r + 1
		      J[r, i] <- -dt[i]
	    }
	    J.SDT <- cbind(dm, ds, J)				# Full Jacobi matrix of SDT model (adjusted for probabilities)
	
      # 2. Compute derivatives for probability parameters
      prb.SDT <- HTSDT.SDT.Probs(c(gausspar[-1], th))	            # Compute SDT response probabilities
      dp.rec <- -prb.SDT
      dp.rec[rec.cat] <- dp.rec[rec.cat] + 1
      cbind(dp.rec, J.SDT)
}

# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix of the SDT-Model (rating part) of the Ranking-Rating model
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.SDT.Jacobi <- function(parvec, cfg)
{
       par.pos <- HTSDT.Rank.pos.par_(cfg)              # Positions of parameters in: $SDT and $TH
       SDT.pos <- par.pos$SDT
       TH.pos  <- par.pos$TH 

       J.SDT            <- NULL
       filler.model.col <- NULL
       filler.model.row <- NULL
       for(i in 1:length(SDT.pos)) {
           J.Gauss    <- NULL                             # Jacobian matrix of the Gaussian model
           J.th       <- NULL                             # Jacobian matrix of threshold
           filler.row <- NULL
           filler.col <- NULL
           filler <- matrix(0, nr = cfg$n.sdt[2, i], 3)     # 3 x probs - matrix of zeros as filler
           rec.cat = 1                                      # Recollection category: for fiste Gaussian lowest response categroy
           for(j in 1:ncol(SDT.pos[[i]])) {
               if(j == 2) rec.cat <- cfg$n.sdt[2, i]         # Recollection category = Highest response category
               J       <- HTSDT.Rank.SDT.Probs.Jacobi(c(rec.cat, parvec[SDT.pos[[i]][,j]]), parvec[TH.pos[[i]]])
               J.Gauss <- rbind(cbind(J.Gauss, filler.col), cbind(filler.row, J[, 1:3]))
               J.th    <- rbind(J.th, J[, -(1:3)])
               if(j < ncol(SDT.pos[[i]])) {
                  filler.row <- cbind(filler.row, filler)     # Adapt filler matrices
                  filler.col <- rbind(filler.col, filler)
               }
           }    
           J.actual <- cbind(J.Gauss, J.th)
           J.SDT    <- cbind(rbind(J.SDT, filler.model.row), rbind(filler.model.col, J.actual))
           if(i < length(SDT.pos)) {
              filler.model.row <- matrix(0, nr = cfg$n.sdt[1, i+1] * cfg$n.sdt[2, i+1], nc = sum(cfg$n.sdt[1, 1:i]) * 3 + sum(cfg$n.sdt[2, 1:i] - 1))
              filler.model.col <- matrix(0, nr = sum(cfg$n.sdt[1, 1:i] * cfg$n.sdt[2, 1:i]), nc = cfg$n.sdt[1, i+1] * 3 + cfg$n.sdt[2, i+1] - 1)
          }
       }
       return(J.SDT)
}

#===============================================================================
# 6. JACOBI MATRIX, GRADIENT AND MODEL MATRICES OF THE RATING-RANKING MODEL
#===============================================================================

#-------------------------------------------------------------------------------
# Jacobian of the model function of the Ranking and Rating model:
# 1. Paraemter vector
# 2. Configuration information
# 3. 2xp matrix of fixed constraints (just for compatibility reasons)
# 4. 2xp matrix of identity constrints (just for compatibility reasons)
#-------------------------------------------------------------------------------
HTSDT.Rank.Jacobi <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
      # 1. Compute probability parameters
      parvec <- HTSDT.comp.recp(parvec, cfg, fixed, ident)
            
      # 2. Jacobi-Matrix of the ranking part
      rank.J <- HTSDT._Rank.Jacobi(parvec, cfg, fixed, ident)

      if(cfg$rating == F)     # No rating 
         return(rank.J %*% HTSDT.recp.Jacobi(parvec, cfg, fixed, ident))
      
      # 3. Jacobi-Matrix of the rating part
      rating.J <- HTSDT.Rank.SDT.Jacobi(parvec, cfg)

      filler.Rk  <- matrix(0, nr = dim(rank.J)[1],   nc = dim(rating.J)[2])
      filler.SDT <- matrix(0, nr = dim(rating.J)[1], nc = dim(rank.J)[2])
      
      J <- rbind(cbind(rank.J, filler.Rk), cbind(filler.SDT, rating.J))

      # 3. Adjust for SDT recollection paramters
      return(J %*% HTSDT.recp.Jacobi(parvec, cfg, fixed, ident))
}

#-------------------------------------------------------------------------------
# Model matrix of the ranking model 
# implementing functional and identity constraints
# Input:   1. full.par   = Full parameter vector
#          2. cfg        = List with configuration information
#          3. fixed      = Matrix of fixed constraints
#          4. ident      = Matrix of equality constraints
#          5. functional = A function implementing functional constraints
#
# Output: Model matrix of free paraemters
#
# Comment: Not considered are identity constraints between probability 
#          parameters from different blocks
#-------------------------------------------------------------------------------
HTSDT.Rank.Model.Matrix <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos      <- SavePos.par(full.par, fixed, ident)
      parvec   <- Elim.par(full.par, fixed, ident)
      full.par <- Restore.par(parvec, pos, fixed, ident, functional)	# Restore full parameter vector

      # 1. Compute Jacobian of the model
      J   <- HTSDT.Rank.Jacobi(full.par, cfg, fixed, ident)

      # 2. Treatment of  identity constraints
      if(!is.null(ident))
      {
            J.Id <- diag(length(full.par))
            Index <- matrix(c(ident[2,], ident[1,]), nc = 2)
            J.Id[Index] <- 1
            J <- J %*% J.Id
      }
      
      # 3. Treatment of functional constraints
      if(!is.null(functional))
      {
            J.fct <- attr(functional, "SDT.Jacobian")
            if(is.null(J.fct)) 
                  return(NULL)           # No Jacobian function supplied -> return (NULL)
            J <- J %*% J.fct(full.par)
      }
      
      # 4. Keep only columns with free parameters
      J[, pos]
}

# ------------------------------------------------------------------------------------------------------------------
# Numerical computation of the model matrix (SDT-Rank model)
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.Model.Matrix.Num <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)
      parvec <- Elim.par(full.par, fixed, ident)
      J <- jacobian(HTSDT.Rank.Full, parvec, cfg = cfg, pos = pos, fixed = fixed, ident = ident, functional = functional)
}

# ------------------------------------------------------------------------------------------------------------------
# Gradient of logL for the HTSDT-Rank model
# Input:	Same as for Likelihood function
# Output:	Gradient
#
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.Grad <- function(parvec, datavec, Model, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
	  parvec   <- Restore.par(parvec, pos, fixed, ident, functional)	# Restore full parameter vector
 	  gradient <- as.vector(SDT.logL.Jacobi(parvec, datavec, Model, n, pos, fixed, ident, functional) %*% HTSDT.Rank.Jacobi(parvec, n, fixed, ident)) 

	  # Treat functional constraints
	  if(!is.null(functional)) 	{
		   gr.fct <- attr(par, "SDT.gradient")		# Isolate gradient function
		   if(is.null(gr.fct))	stop("\n\n HTSDT.Rank.Grad: Error in computation of gradient: No gradient for functional constraints found:\n Set Parameter sym.gr = F in function SDT.Estimate !\n\n")
		   gradient <- gr.fct(gradient)
	  }
    Elim.par.grad(gradient, fixed, ident)
}

#===============================================================================
# 7. JACOBIAN FOR ADJUSTING PROBABILITY PARAMETERS
#===============================================================================
# --------------------------------------------------------------------------------
# Computation of Jacobian for adjusting probability parameters
# Input:	1. parvec = full parameter vector with probability parameters computed
#         2. cfg    = list with configuration information
#         3. fixed  = matrix of fixed constraints
#         4. ident  = matrix of equality constraints
#
# Output:	Jacobian matrix for adjusting probability parameters
#
# Comment: Only free parameters are adjusted
# --------------------------------------------------------------------------------
HTSDT.recp.Jacobi <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
      J <- diag(length(parvec))
      rec.pos <- HTSDT.Rank.Free.Ppos(cfg, fixed)              # Helpfunction for finding the columns positions of the recollection parameters within the Jacobian matrix
      if(!is.null(rec.pos)) 
         J[matrix(c(rec.pos, rec.pos), nc = 2)] <- parvec[rec.pos] * (1 - parvec[rec.pos]) # Derivation of probabilities with respect to raw values

      if(cfg$rating == F) {
        plen <- 5*length(cfg$k.rg)
        return(J[1:plen, 1:plen])
      }  
      return(J)
}

#------------------------------------------------------------------------------------------------------------------
# Jacobian for adjusting the covariance matrix due to the delta method
# Input:	  1. parvec = full parameter vector with probability parameters computed
#           2. cfg    = list with configuration information
#           3. fixed  = matrix of fixed constraints
#           4. ident  = matrix of equality constraints
#
# Output:	Jacobian matrix for adjusting covariance matrix according to delta method
#         Contains entries of free probability parameters only
#
#------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.Delta.Matrix <- function(parvec, cfg, fixed, ident)
{
      # 1. Identify free probability parameters
      free.ppos <- HTSDT.Rank.Free.Ppos.Delta(parvec, cfg, fixed, ident)
      if(is.null(free.ppos)) return(NULL)


      # 2. Compute Jacobian matrix for the function that transforms raw to 
      #    probability parameters taking fixed and identity constraints into account
      J <- HTSDT.recp.Jacobi(parvec, cfg, fixed, ident)
      if(is.null(J)) return(J)

      return(J[free.ppos, free.ppos, drop = F])           
}


# ------------------------------------------------------------------------------------------------------------------
# Computing adjusted standard errors for probability parameters
# Input:	1. cov.matrix = covariance matrix of estimated raw parameters
#         2. parvec     = full parameter vector with probability parameters computed
#         3. cfg        = list with configuration information
#         4. fixed      = matrix of fixed constraints
#         5. ident      = matrix of equality constraints
#
# Output:	Vector of standand errors for free parameters (adjusted) or NULL 
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.Delta.Method.Prob <- function(cov.matrix, parvec, cfg, fixed, ident)
{
      # 0. Tests and computation of unadjusted SE vector
      if(is.null(cov.matrix) || is.na(cov.matrix) || is.nan(cov.matrix))
            return(NULL)
      SE.vector <- sqrt(diag(cov.matrix))       # Vector of unadjusted standard errors
      
      # 1. Compute adjustment matrix
      J.adjust <- HTSDT.Rank.Delta.Matrix(parvec, cfg, fixed, ident)
      if(is.null(J.adjust)) return(SE.vector)
      
      # 2. Positions of free parameters with respect to restricted parameter vector     
      free.ppos <- HTSDT.Rank.Free.Ppos.Delta(parvec, cfg, fixed, ident)

      # 3. Carve out relevant part of the covariance matrix
      cov.free.ppos <- cov.matrix[free.ppos, free.ppos, drop = F]

      # 4. Construct adjusted SE vector and fill SE vector with adjusted values
      SE.adjust <- sqrt(diag(J.adjust %*% cov.free.ppos %*% t(J.adjust)))
      SE.vector[free.ppos] <- SE.adjust
      return(SE.vector)
}

#===============================================================================
# 8. ADMINISTRATIVE FUNCTIONS
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Check and generate configuration structure
# Input:	cfg = Configuration structure: A list with the following entries
#                 - n.sdt        2xn matrix with each column representing an SDT model:
#                                first entry of the column  = Number of Gaussians, 
#                                second entry of the column = Number of response categories
#                 - k.rg         Number of alternatives
#                 - j.rg         Number of relevant ranking positions
#                 - npt          Number of quadrature points
#                 - restriction  string identifying the restriction
#                 - rating       Flag indicating whether ranking and rating are combined
#
# Output: Configuration list
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.Check.Fill.Cfg <- function(cfg)
{
      if(!is.null(cfg))
         if(!is.numeric(cfg))
            if(!is.null(cfg$checked))
               if(cfg$checked)   return(cfg)

      k.rg <- 2
      npt <- 30

      if(is.null(cfg))
      {
            cfg <- list(n.sdt       = matrix(c(2, 6), nr = 2, nc = 1),
                        model       = "P",
                        k.rg        = k.rg,
                        j.rg        = k.rg,
                        npt         = npt,
                        restriction = "standard",
                        rating      = T,
                        checked     = T)
            return(cfg)
      }
      if(is.numeric(cfg))
      {
            k.rg     <- as.integer(cfg)
            cfg      <- list(n.sdt       = matrix(c(2, 6), nr = 2, nc = 1),
                             model       = "P",
                             k.rg        = k.rg,
                             j.rg        = k.rg,
                             npt         = npt,
                             restriction = "standard",
                             rating      = T,
                             checked     = T)
            return(cfg)
      }
      cfgneu <- list()
      len <- max(1, length(cfg$k.rg), length(cfg$j.rg))
      if(is.null(cfg$n.sdt))       cfgneu$n.sdt       <- matrix(c(2, 6), nr = 2, nc = 1)           else  cfgneu$n.sdt       <- matrix(cfg$n.sdt, nr = 2)
      if(is.null(cfg$model))       cfgneu$model       <- rep("P", len)                             else  cfgneu$model       <- cfg$model
      if(is.null(cfg$k.rg)) if(is.null(cfg$j.rg)) cfgneu$k.rg <- k.rg else cfgneu$k.rg <- cfg$j.rg else  cfgneu$k.rg        <- cfg$k.rg
      if(is.null(cfg$j.rg))        cfgneu$j.rg        <- cfg$k.rg                                  else  cfgneu$j.rg        <- cfg$j.rg
      if(is.null(cfg$npt))         cfgneu$npt         <- npt                                       else  cfgneu$npt         <- cfg$npt
      if(is.null(cfg$rating))      cfgneu$rating      <- T                                         else  cfgneu$rating      <- cfg$rating
      cfgneu$restriction <- HTSDT.Restriction(cfg$restriction)
      cfg <- cfgneu
      
      # Checks
      if(any(!is.numeric(cfg$n.sdt)))  
         stop(c("\n FATAL ERROR in HTSDT.Rank.Check.Fill.Cfg: Vector with number of Gaussian models within each SDT modesl is not numeric\n"))

      if(any(cfg$k.rg < cfg$j.rg)) {
            k.rg <- paste(cfg$k.rg, collapse = ",")
            j.rg <- paste(cfg$j.rg, collapse = ",")
            stop(c("\n FATAL ERROR in HTSDT.Rank.Check.Fill.Cfg: Number of alternatives <", k.rg, "> smaller than number of relevant ranks <", j.rg, ">\n"))
      }
      if(any(cfg$j.rg < 2)) {
            j.rg <- paste(cfg$j.rg, collapse = ",") 
            stop(c("\n FATAL ERROR in HTSDT.Rank.Check.Fill.Cfg: Number of altenative smaller than 2: <", j.rg, ">\n"))
      }
      if(length(cfg$model) < length(cfg$k.rg)) {
            stop("\n FATAL ERROR in HTSDT.Rank.Check.Fill.Cfg: Length of structure <model> smaller than number of ranking modells\n")
      }  
      if(!is.matrix(cfg$n.sdt))
            stop("\n FATAL ERROR in HTSDT.Rank.Check.Fill.Cfg: <n.sdt> must be a matrix\n")
      if(nrow(cfg$n.sdt) != 2)      
            stop("\n FATAL ERROR in HTSDT.Rank.Check.Fill.Cfg: <n.sdt> must be a 2 x n matrix (n = number of SDT models)\n")
      if(any(cfg$sdt[,1] < 2) && cfg$rating == T)      
            stop("\n FATAL ERROR in HTSDT.Rank.Check.Fill.Cfg: Number of Gaussinas in any SDT model must be greater than 1)\n")

      cfg$checked <- T

      return(cfg)
}   
     
# ------------------------------------------------------------------------------------------------------------------
# Check type of restriction
# Input: A string specfiying the typ of restriction
#        Alternatively the flags TRUE (="standard") and FALSE ("No") are allowed
#        Types of restrictions      - "standard"  [ or T]   (Standard restrictions)
#                                   - "SDT"                 (Simulating SDT model)
#                                   - "EVSDT"               (Simulating SDT model with equal variances)
#                                   - "no"  [or F]          (No restrictions)
#
# Output: String indicating type of restriction
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Restriction <- function(restriction)
{
      if(is.null(restriction) || restriction == T)
            return("standard")
      if(tolower(restriction) == "standard")      # mu = 0 of first Gaussian in each SDT model, and all sd = 1
            return("standard")
      if(tolower(restriction) == "equal")         # standard + equal mu and p.rec
            return("equal")
      if(tolower(restriction) == "uvsdt")         # p.rec = 0, mu = 0 of first Gaussaan in SDT model
            return("UVSDT")
      if(tolower(restriction) == "evsdt")
            return("EVSDT")
      if(tolower(restriction) == "no")
            return("no")
      warn.msg <- paste("\n\n HTSDT.Restriction: Unkonwn restriction in configuration list: <", restriction, "> --> setting to: <standard>\n\n", sep = "")      
      warning(warn.msg)
      return("standard")
}            
                         
# ------------------------------------------------------------------------------------------------------------------
# Generate starting parameters due to data and number of signals
#                 - n.sdt     2xn matrix with number of Gaussians and response categories for the n SDT models
#                 - k.rg      Number of alternatives
#                 - j.rg      Number of relevant ranking positions
#                 - npt       Number of quadrature points
#                 -standard   Flag indicating whether standard restrictions have to be specified
#                 -rating     Flag indicating whether ranking and rating are combined
#
# Output: String with parameter values depending on whether only ranking data or
#         ranking as well as reting data are modeled
#
# I. Parameters RANKING MODEL
#    1. p.recollection
#    2. p.guessing
#    3. m.new
#    4. sd.new
#    5. m.old
#    6. sd.old
#
# II. Parameters for each Gaussing model of the SDT component
#    1. p.recollection
#    2. p.guessing
#    3. p.bias (guessing)
#    4. m.new
#    5. sd.new
#
# IV. Threshold parameter for eaqch SDT model
#    c[1], ..., c[k]
#-------------------------------------------------------------------------
HTSDT.Rank.start.par <- function(cfg)
{
      cfg <- HTSDT.Rank.Check.Fill.Cfg(cfg)

      #-------------------------------------------------------------------------
      # Start parameters of the ranking model:
      # For each model:
      # p.recollection
      # p.guessing 
      # m.new   = mean of new distribution
      # sd.new  = standard deviation of new distribution
      # m.old   = mean of old distribution
      # sd.old  = standard deviation of old distribution
      #-------------------------------------------------------------------------
      par.rank <- rep(c(0, 0, 1, 0.5, 1), length(cfg$k.rg))    # Start parameter

      #-------------------------------------------------------------------------
      # Parameters of the SDT model:
      # For each model:
      # p.recollection
      # bias
      # p.guessing 
      # m    = mean of the distribution
      # sd   = standard deviation of the distribution
      # c[i] = thresholds
      #-------------------------------------------------------------------------
      if(!cfg$rating) return(par.rank)   # No rating

      # Add p-rec, p-guess, bias, mu and sigma, for each model and (at the end) thresholds
      for(i in 1:ncol(cfg$n.sdt))
            par.rank <- c(par.rank, c(0, 0, 1), rep(c(0, 0.5, 1), cfg$n.sdt[1, i]-1), seq(-0.5, 0.5, length = cfg$n.sdt[2, i] - 1))      
      return(par.rank)
}

# ------------------------------------------------------------------------------------------------------------------
# Generate upper and lower bounds of parameter
# Input:	cfg = configruation structure
#
# Output: List with lower and upper bounds on parameter vector
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.par.bounds <- function(cfg)
{
      cfg <- HTSDT.Rank.Check.Fill.Cfg(cfg)

      min.p <- -700
      max.p <-  700

      len <- length(cfg$k.rg)
      lower <- rep(c(min.p, -Inf, HTSDT.Rk.Tol, -Inf, HTSDT.Rk.Tol), len)
      upper <- rep(c(max.p,  Inf,          Inf,  Inf,          Inf), len)
      
      if(cfg$rating == F)
            return(list(lower = lower, upper = upper))
      
           # Add p-rec, p-guess, bias, mu and sigma, for each model and (at the end) thresholds
      for(i in 1:ncol(cfg$n.sdt)) {
            lower <- c(lower, rep(c(min.p, -Inf, HTSDT.Rk.Tol), cfg$n.sdt[1, i]), rep(-Inf, cfg$n.sdt[2, i] - 1))      
            upper <- c(upper, rep(c(max.p,  Inf,          Inf), cfg$n.sdt[1, i]), rep( Inf, cfg$n.sdt[2, i] - 1))      
      }
      return(list(lower = lower, upper = upper))
}      

# ------------------------------------------------------------------------------------------------------------------
# Generate strings with names of the parameters for the models
# Input:	cfg         = configruation structure
#         with.space  = Flag indicating whether names should be with space (default) or not
#
# Output: String with parameter names
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.par.names <- function(cfg, with.space = T)
{
      # If configuration information is not complete, then complete it
      if(is.null(cfg$rating) || is.null(cfg$n.sdt) || is.null(cfg$j.rg) || is.null(cfg$k.rg))
          cfg <- HTSDT.Rank.Check.Fill.Cfg(cfg)  
      
      if(with.space)
            Ranking.names <- c("Recollection (Rank", 
                               "Mean.1       (Rank", 
                               "Stddev.1     (Rank", 
                               "Mean.2       (Rank", 
                               "Stddev.2     (Rank")
      else
            Ranking.names <- c("Recollection (Rank", 
                               "Mean.1 (Rank", 
                               "Stddev.1 (Rank", 
                               "Mean.2 (Rank", 
                               "Stddev.2 (Rank")
                         
      models <- rep(1:length(cfg$k.rg), each = length(Ranking.names))
      Name.Rk  <- paste(Ranking.names, " ", models, ")", sep = "")
      if(!cfg$rating)
         return(Name.Rk) 

      # Now create names for rating models
      for(i in 1:ncol(cfg$n.sdt)) {
          for(j in 1:cfg$n.sdt[1, i]) {
                Name.Rk <- c(Name.Rk, paste("[SDT-", i, "][Gauss-", j,"] Recollection", sep = ""))
                Name.Rk <- c(Name.Rk, paste("[SDT-", i, "][Gauss-", j,"]         Mean", sep = ""))
                Name.Rk <- c(Name.Rk, paste("[SDT-", i, "][Gauss-", j,"]       Stddev", sep = ""))
          }
          Name.Rk <- c(Name.Rk, paste("[SDT-", i, "]      c-", 1:(cfg$n.sdt[2, i] - 1),   sep = ""))
      }    
      return(Name.Rk)
}

# ------------------------------------------------------------------------------------------------------------------
# Names of the models 
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.model.name <- function(cfg)
{
      cfg <- HTSDT.Rank.Check.Fill.Cfg(cfg)

      k.rg  <- paste(cfg$k.rg, collapse = ",")
      j.rg  <- paste(cfg$j.rg, collapse = ",")
      model <- paste(cfg$model, collapse = ",")
      model <- paste(cfg$model, collapse = ",")
      if(ncol(cfg$n.sdt) == 1) 
          n.sdt <- paste(cfg$n.sdt[,1], collapse = ",")  
      else {
          n.sdt = NULL
          for(i in 1:ncol(cfg$n.sdt))
              if(i < ncol(cfg$n.sdt)) 
                 n.sdt <- paste(n.sdt, "<", paste(cfg$n.sdt[,i], collapse = ","), ">, ", sep = "") 
              else
                 n.sdt <- paste(n.sdt, "<", paste(cfg$n.sdt[,i], collapse = ","), ">", sep = "") 
      }
      if(cfg$rating)
            mod.name <- paste("HTSDT ranking and rating model:", 
                              " Number of Gaussians within SDT models: <", n.sdt, ">",
                              " Model types: <", model, ">",
                              " Number of ranking-alternatives: <", k.rg, ">", 
                              " Number of ranking-positions: <", j.rg, ">", 
                              " Number of quadrature points: <", cfg$npt, ">", sep = "",
                              " Restrictions: <", cfg$restriction, ">")
      else
            mod.name <- paste("HTSDT ranking model:", 
                              " Model types: <", model, ">",
                              " Number of ranking-alternatives: <", k.rg, ">",
                              " Number of ranking-positions: <", j.rg, ">", 
                              " Number of quadrature points: <", cfg$npt, ">", sep = "",
                              " Restrictions: <", cfg$restriction, ">")
return(mod.name)
}


# ------------------------------------------------------------------------------------------------------------------
# Check number of parameters and data points
# Error-Codes: 1: Data length wrong
#              2: Parameter lenght wrong
#              3: Both wrong
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.check <- function(cfg,  parvec, datavec)
{
      Msg <- "NOTE: Computation of the expected number of data points and parameters for a given configuration"
      dlen <- length(datavec)  # length of data vector
      plen <- length(parvec)
      cfg  <- HTSDT.Rank.Check.Fill.Cfg(cfg)

      # Case 1: Ranking and Rating
      if(cfg$rating) {
            dlen.cfg <- sum(cfg$j.rg) + sum(cfg$n.sdt[1,]*cfg$n.sdt[2,])                
            plen.cfg <- 5*length(cfg$k.rg) + 3*sum(cfg$n.sdt[1,]) + sum(cfg$n.sdt[2,] - 1)
      }
      # Case 2: Ranking only
      else {
            dlen.cfg <- sum(cfg$j.rg)
            plen.cfg <- 5*length(cfg$k.rg)
      }
	Erg <- matrix(c(dlen, plen, dlen.cfg, plen.cfg), nc = 2)
	colnames(Erg) <- c("Found", "Expected")
  rownames(Erg) <- c("Data-length", "Parameter-length")
	Error <- 1*(Erg[1, 1] != Erg[1, 2])
  Error <- Error + 2*(Erg[2, 1] != Erg[2, 2])
	list(Error = Error, Stat = Erg, Msg = Msg)
}
        
# ------------------------------------------------------------------------------------------------------------------
# HTSDT Ranking and Rating Model: Set restrictions
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.Restriction <- function(cfg, fixed, ident)
{
      # If configuration information is not complete, then complete it
      cfg <- HTSDT.Rank.Check.Fill.Cfg(cfg)  

      if(cfg$restriction == "standard")
            return(HTSDT.Rank.Set.Standard.Restriction(cfg, fixed, ident))
      if(cfg$restriction == "equal")
            return(HTSDT.Rank.Set.Equal.Restriction(cfg, fixed, ident))
      if(cfg$restriction == "UVSDT")
            return(HTSDT.Rank.Set.UVSDT.Restriction(cfg, fixed, ident))
      if(cfg$restriction == "EVSDT")
            return(HTSDT.Rank.Set.EVSDT.Restriction(cfg, fixed, ident))
     return(list(fixed = fixed, ident = ident))
}

# ------------------------------------------------------------------------------------------------------------------
# Ranking and Rating Model with Recollection and Guessing: Set standard restrictions
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.Set.Standard.Restriction <- function(cfg, fixed = NULL, ident = NULL)
{
      cfg  <- HTSDT.Rank.Check.Fill.Cfg(cfg)

      
      # ----- Fixed constraints on ranking models -----------
      # 1. mean of first Gaussian of each model = 0
      # 2. sd   of each  Gaussian of each model = 0
      # ----------------------------------------------------------
      par.pos <- HTSDT.Rank.pos.par_(cfg)
      afc.pos <- par.pos$AFC

      fix.val <- rep(c(0, 1, 1), ncol(afc.pos))
      fix.pos <- NULL
      for(i in 1:ncol(afc.pos))
         fix.pos <- c(fix.pos, afc.pos[c(2:3, 5), i, drop = T])
      fixed.s <- rbind(fix.val, fix.pos)
  
      if(cfg$rating == F)
      {
            # Build fixed matrix and remove redundant restrictions
            pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
      	    fixed <- cbind(fixed.s[, pos.unique], fixed)   
            # Build identity matrix and remove redundant restrictions
            return(list(fixed = fixed, ident = ident))
      }
      #-------Fixed constraints (Rating) ---------------------------------------
      # FIRST FGAUSSIAN IN EACH SDT
      # 1. Recollection          (Rating 1) = 0.0     (Parameter 1)
      # 2. Mean.1                (Rating 1) = 0.0     (Parameter 2)
      # 3. Stddev.1              (Rating 1) = 1.0     (Parameter 3)
      # ------------------------------------------------------------------------
      # FURTHER GAUSSIANS
      # 4. Stddev                           = 1.0     (Parameter 3)
      # ------------------------------------------------------------------------
      sdt.pos <- par.pos$SDT
      len <- length(sdt.pos)
      fix.val.rating <- fix.pos.rating <- NULL
      for(i in 1:len) {
         Gauss.model <- sdt.pos[[i]]              # Exctract 5 x nc matrix of the SDT model
         nc          <- ncol(Gauss.model)         # Number of GAussians
         fix.val.rating <- c(fix.val.rating, c(0, 0, 1))           # First Gausssian
         fix.pos.rating <- c(fix.pos.rating, Gauss.model[, 1])
         if(nc > 1)
            for(j in 2:nc) {                                       # Other Gaussians
               fix.val.rating <- c(fix.val.rating, 1)
               fix.pos.rating <- c(fix.pos.rating,Gauss.model[3, j])
            }
      }
      fixed.s <- cbind(fixed.s, matrix(c(fix.val.rating, fix.pos.rating), nr = 2, byrow = T))
      pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
	    fixed <- cbind(fixed.s[, pos.unique], fixed)            # Add to fixed if not alread specified
	
      list(fixed = fixed, ident = ident)
}

# ------------------------------------------------------------------------------------------------------------------
# Ranking and Rating Model with Recollection and Guessing: Set standard + equalty restricitons on means and probabilities
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.Set.Equal.Restriction <- function(cfg, fixed = NULL, ident = NULL)
{
      cfg  <- HTSDT.Rank.Check.Fill.Cfg(cfg)

      # Only a single ranking model, thus nothing to set
      n.afc <- length(cfg$j.rg)
      if(n.afc < 2 && cfg$rating == F)
         return(list(fixed = fixed, ident = ident))
         
      # Set standard restrictions
      stand.restrictions <- HTSDT.Rank.Set.Standard.Restriction(cfg, fixed, ident)
      fixed <- stand.restrictions$fixed
      ident <- stand.restrictions$ident
      
      par.pos <- HTSDT.Rank.pos.par_(cfg)
      afc.pos <- par.pos$AFC
      if(n.afc > 1) {
        source.pos <- rep(afc.pos[c(1, 4), 1], n.afc - 1)
        target.pos <- as.vector(afc.pos[c(1, 4), 2:n.afc])
        ident.s <- rbind(source.pos, target.pos)
        }
       else ident.s <- NULL 
       if(cfg$rating == F) {
          pos.unique <-  !(ident.s[2,] %in% ident[2,])
	        ident <- cbind(ident.s[, pos.unique], ident)
          return(list(fixed = fixed, ident = ident))
       }
       # Now set equality restritions for SDT part
       sdt.pos <- par.pos$SDT
       
       source.pos <- NULL
       target.pos <- NULL
       for(i in 1:length(sdt.pos)) {
          source.pos <-  c(source.pos, rep(afc.pos[c(1, 4), 1], ncol(sdt.pos[[i]]) - 1))
          target.pos <-  c(target.pos, sdt.pos[[i]][c(1, 2), 2:ncol(sdt.pos[[i]])])
       } 
       ident.s <- cbind(ident.s, rbind(source.pos, target.pos))
       pos.unique <-  !(ident.s[2,] %in% ident[2,])
	     ident <- cbind(ident.s[, pos.unique], ident)
       return(list(fixed = fixed, ident = ident))
}  

# ------------------------------------------------------------------------------------------------------------------
# Ranking and Rating Model with Recollection and Guessing: Set UVSDT restrictions
# 1. Recollection = 0
# 2. Bias = 0
# 3. Guessing = 0.5 (for each Gaussian model)
# 4. Mean and standard deviation of first Gaussian in each model = 0 and 1
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.Set.UVSDT.Restriction <- function(cfg, fixed = NULL, ident = NULL)
{
     cfg  <- HTSDT.Rank.Check.Fill.Cfg(cfg)

      par.pos <- HTSDT.Rank.pos.par_(cfg)
      afc.pos <- par.pos$AFC

      fix.val <- rep(c(0, 0, 1), ncol(afc.pos))
      fix.pos <- NULL
      for(i in 1:ncol(afc.pos))
         fix.pos <- c(fix.pos, afc.pos[1:3, i, drop = T])
      fixed.s <- rbind(fix.val, fix.pos)
      if(cfg$rating == F)
      {
            # Build fixed matrix and remove redundant restrictions
            pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
            fixed <- cbind(fixed.s[, pos.unique], fixed)
            # Build identity matrix and remove redundant restrictions
            return(list(fixed = fixed, ident = ident))
      }
      
      sdt.pos <- par.pos$SDT
      fix.pos <- fix.val <- NULL
      for(i in 1:length(sdt.pos)){
         fix.val <- c(fix.val, c(0, 0, 1))
         fix.pos <- c(fix.pos, sdt.pos[[i]][, 1, drop = T])
         for(j in 2:ncol(sdt.pos[[i]])) {
             fix.val <- c(fix.val, 0)
             fix.pos <- c(fix.pos, sdt.pos[[i]][1, j, drop = T])
         }    
      }
      fixed.s <- cbind(fixed.s, rbind(fix.val, fix.pos))
      pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
	    fixed <- cbind(fixed.s[, pos.unique], fixed)            # Add to fixed if not alread specified
	
      return(list(fixed = fixed, ident = ident))
}

# ------------------------------------------------------------------------------------------------------------------
# Ranking and Rating Model with Recollection and Guessing: Set EVSDT restrictions
# 1. Recollection = 0
# 2. Bias = 0
# 3. Guessing = 0.5 (for each Gaussian model)
# 4. Mean and standard deviation of first Gaussian in each model = 0
# 5. Standard deviation of each GAussian = 1 (in each model)
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.Set.EVSDT.Restriction <- function(cfg, fixed = NULL, ident = NULL)
{
     cfg  <- HTSDT.Rank.Check.Fill.Cfg(cfg)

      par.pos <- HTSDT.Rank.pos.par_(cfg)
      afc.pos <- par.pos$AFC

      fix.val <- rep(c(0, 0, 1, 1), ncol(afc.pos))
      fix.pos <- NULL
      for(i in 1:ncol(afc.pos))
         fix.pos <- c(fix.pos, afc.pos[c(1:3, 5), i, drop = T])
      fixed.s <- rbind(fix.val, fix.pos)
      if(cfg$rating == F)
      {
            # Build fixed matrix and remove redundant restrictions
            pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
            fixed <- cbind(fixed.s[, pos.unique], fixed)
            # Build identity matrix and remove redundant restrictions
            return(list(fixed = fixed, ident = ident))
      }
      
      sdt.pos <- par.pos$SDT
      fix.pos <- fix.val <- NULL
      for(i in 1:length(sdt.pos)){
         fix.val <- c(fix.val, c(0, 0, 1))
         fix.pos <- c(fix.pos, sdt.pos[[i]][, 1, drop = T])
         for(j in 2:ncol(sdt.pos[[i]])) {
             fix.val <- c(fix.val, c(0, 1))
             fix.pos <- c(fix.pos, sdt.pos[[i]][c(1, 3), j, drop = T])
         }    
      }
      fixed.s <- cbind(fixed.s, rbind(fix.val, fix.pos))
      pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
	    fixed <- cbind(fixed.s[, pos.unique], fixed)            # Add to fixed if not alread specified
	
      return(list(fixed = fixed, ident = ident))
}

# ------------------------------------------------------------------------------------------------------------------
# Format free parameters and compute confidence intervals
# Input:	1. px2 Matrix with parameters and standard errors
#		      2. Full parameter vector
#         3. List with configuration information
#		      4. Matrix of fixed constraints
#		      5. Matrix of equality constraints
#		      6. Confidence level
#         7. Number of decimal places
#
# Output: px5 Matrix of parameters, standard errors and confidence limits
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.Format.Par.est <- function(Pars, full.par, cov.matrix, cfg, fixed, ident, alpha, deci)
{
      # 1. Find the position of the parameters
      par.pos <- HTSDT.Rank.pos.par_(cfg)
      afc.pos <- par.pos$AFC
      sdt.pos <- par.pos$SDT
      
      # 2. Find positions of sigmas and recollection parameters
      pos.sigma <- afc.pos[c(3, 5), drop = T]
      pos.recp  <- afc.pos[1, drop = T]
      if(!is.null(sdt.pos)) 
        for(i in 1:length(sdt.pos)) {
           pos.sigma <- c(pos.sigma, sdt.pos[[i]][3, ]) 
           pos.recp  <- c(pos.recp , sdt.pos[[i]][1, ])
        }   
      # 3. Replace negative valus of sigma by positive values      
      full.par[pos.sigma] <- abs(full.par[pos.sigma])
      
      # 4. Compute recollection probabilities from raw parameters
      full.par <- HTSDT.comp.recp(full.par, cfg, fixed, ident)
      Pars[, 1] <- Elim.par(full.par, fixed, ident)

      # 5. Check for correct cov.matrix
      if(is.null(cov.matrix))
      {
            NAs       <- rep(NA, dim(Pars)[1])
            Pars[,2]  <- NAs
            Pars <- cbind(Pars, NAs, NAs)
            colnames(Pars) <- c("Value", "SE", paste("CFI-", alpha*100, "(Lower)", sep = ""), paste("CFI-", alpha*100, "(Upper)", sep = ""))
            return(round(Pars, deci))
      }
      # 6. Adjust for standard errors for probability parameter
      Pars[,2] <- HTSDT.Rank.Delta.Method.Prob(cov.matrix, full.par, cfg, fixed, ident)
      qt.cfi <- qnorm((1-alpha)/2)
      cfi.lo <- Pars[,1] + Pars[,2]*qt.cfi
      cfi.up <- Pars[,1] - Pars[,2]*qt.cfi
            
      # 7. Compute lower bounds of confidence interval (= for probabilities and standard errors of the distributions the lower bound is 0)            
      bd.lo <- rep(-Inf, length(full.par))
      bd.lo[c(pos.sigma, pos.recp)] <- 0
      bd.lo <- Elim.par(bd.lo, fixed, ident)
      cfi.lo <- ifelse(cfi.lo < bd.lo, bd.lo, cfi.lo)

      # 8. Compute upper bounds of confidence interval (= for probability parameters the upper bound i 1)
      bd.hi <- rep(Inf, length(full.par))
      bd.hi[pos.recp] <- 1
      bd.hi <- Elim.par(bd.hi, fixed, ident)
      cfi.up <- ifelse(cfi.up > bd.hi, bd.hi, cfi.up)
        
      Pars <- cbind(Pars, cfi.lo, cfi.up)
      colnames(Pars) <- c("Value", "SE", paste("CFI-", alpha*100, "(Lower)", sep = ""), paste("CFI-", alpha*100, "(Upper)", sep = ""))
      round(Pars, deci)
}

# ------------------------------------------------------------------------------------------------------------------
# Format full parameters
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Format.Par.Rank <- function(full.par, cfg, deci, fixed, ident)
{
      Ranking.names <- c("Recollection", "Mean.1", "Stddev.1", "Mean.2", "Stddev.2") 
      Rating.names  <- c("Recollection", "Mean", "Stddev")
  
      # 1. Find the position of the parameters
      par.pos <- HTSDT.Rank.pos.par_(cfg)
      afc.pos <- par.pos$AFC
      sdt.pos <- par.pos$SDT
       th.pos <- par.pos$TH
      
      # 2. Find positions of sigmas and recollection parameters
      pos.sigma <- afc.pos[c(3, 5), drop = T]
      pos.recp  <- afc.pos[1, drop = T]
      if(!is.null(sdt.pos)) 
        for(i in 1:length(sdt.pos)) {
           pos.sigma <- c(pos.sigma, sdt.pos[[i]][3, ]) 
           pos.recp  <- c(pos.recp, sdt.pos[[i]][1, ])
        }   
      # 3. Rplace negative valus of sigma by positive values      
      full.par[pos.sigma] <- abs(full.par[pos.sigma])

      # 4. Compute probabilities
      full.par <- HTSDT.comp.recp(full.par, cfg, fixed, ident)

      # 5. Extract the afc parameters
      Ranking.pars <- matrix(round(full.par[as.vector(afc.pos)], deci), nc = ncol(afc.pos))
      
      # 6. Row and column names for afc parameters
      cf.matrix <- matrix(c(cfg$k.rg, cfg$j.rg), nr = 2, byrow = T)
      cf.vector <- apply(cf.matrix, 2, fkt <- function(x) { paste(x, collapse = ", j=") })
      colnames(Ranking.pars) <- paste("Rank <k=", cf.vector, ">", sep = "")
     	rownames(Ranking.pars) <- Ranking.names
      if(cfg$rating == F)
            return(list(Ranking = Ranking.pars, Rating = NULL))

     # 6. Compile Gaussian Parameters
      Gauss.pos  <- NULL
      for(i in 1:length(sdt.pos))
           for(j in 1:ncol(sdt.pos[[i]]))
               Gauss.pos <- cbind(Gauss.pos, sdt.pos[[i]][,j])
      Rating.Par             <- round(matrix(full.par[Gauss.pos], nc = sum(cfg$n.sdt[1,])), deci)
      rownames(Rating.Par)   <- Rating.names
      Gauss.names            <- paste("SDT-", 1:ncol(cfg$n.sdt), sep = "")
      SDT.names              <- NULL
      for(i in 1:length(Gauss.names)) SDT.names <- c(SDT.names, paste(Gauss.names[i], "/Gauss-", 1:cfg$n.sdt[1, i], sep = ""))
     	   colnames(Rating.Par) <- SDT.names

       # 7. Compile Thresholds
       th.par <- NULL
       max.lt <- 0
       for(i in 1:length(th.pos)) # Determine maximal number of thresholds
          if(length(th.pos[[i]]) > max.lt) max.lt <- length(th.pos[[i]])
       for(i in 1:length(th.pos)) {
          th     <- c(paste(round(full.par[th.pos[[i]]], deci)), rep("", max.lt - length(th.pos[[i]])))
          th.par <- rbind(th.par, th)   
       }
       th.par <- data.frame(th.par)
       colnames(th.par) <- paste("c-", 1:ncol(th.par), sep = "")
       rownames(th.par) <- paste("SDT-", 1:nrow(th.par), sep = "")
 
      # 8. Return Ranking and rating data
      return(list(Ranking = Ranking.pars, Rating = Rating.Par, Thresholds = th.par))
}

# ------------------------------------------------------------------------------------------------------------------
# Ranking Model with recollection and guessing: Computation of observed p values for ranking model
# Input:    datavec = Vector with data
#           cfg     = configuration information
#
# Output:   Vector with probabilities
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.Compute.Obs.p <- function(datavec, cfg)
{
      cfg <- HTSDT.Rank.Check.Fill.Cfg(cfg)
      
      p.rank <- NULL
      ed <- 0
      for(i in 1:length(cfg$j.rg))
      {
            st     <- ed + 1
            ed     <- ed + cfg$j.rg[i]
            p.rank <- c(p.rank, datavec[st:ed]/sum(datavec[st:ed]))
      }
      if(cfg$rating == F)
            return(p.rank)

      datavec <- datavec[-(1:sum(cfg$j.rg))]
      for(i in 1:ncol(cfg$n.sdt)) {
         sdt.part <- matrix(datavec[1:(cfg$n.sdt[1,i]*cfg$n.sdt[2,i])], nr = cfg$n.sdt[1, i], byrow = T)     # Extract rating data to matrix with n rows
         p.rank   <- c(p.rank, as.vector(apply(sdt.part, 1, function(x) { x/sum(x) })))
         if(i < ncol(cfg$n.sdt))
              datavec <- datavec[-(1:(cfg$n.sdt[1,i]*cfg$n.sdt[2,i]))]
      }
      return(p.rank)
}

# ------------------------------------------------------------------------------------------------------------------
# Ranking Model with recollection and guessing: Computation of observed p values for ranking model
# Input:    datavec = Vector with data
#           cfg     = configuration information
#
# Output:   Vector with probabilities
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.Compute.Obs.N <- function(datavec, cfg)
{
      cfg <- HTSDT.Rank.Check.Fill.Cfg(cfg)

      # N for ranking models
      N.rank <- NULL
      ed <- 0
      for(i in 1:length(cfg$j.rg))
      {
            st     <- ed + 1
            ed     <- ed + cfg$j.rg[i]
            ss     <- sum(datavec[st:ed])
            N.rank <- c(N.rank, rep(ss, cfg$j.rg[i]))
      }
      if(cfg$rating == F) return(N.rank)

      # N for SDT models
      datavec <- datavec[-(1:sum(cfg$j.rg))]
      for(i in 1:ncol(cfg$n.sdt)) {
          sdt.part <- matrix(datavec[1:(cfg$n.sdt[1,i]*cfg$n.sdt[2,i])], nr = cfg$n.sdt[1, i], byrow = T)     # Extract rating data to matrix with n rows
          N.rating <- rowSums(sdt.part)
          N.rank   <- c(N.rank, rep(N.rating, each = dim(sdt.part)[2]))
          if(i < ncol(cfg$n.sdt))
              datavec <- datavec[-(1:(cfg$n.sdt[1,i]*cfg$n.sdt[2,i]))]
      }
     return(N.rank)
}
# ------------------------------------------------------------------------------------------------------------------
# Computation of free data points
# Input:    cfg     = configuration information
#
# Output:   Number of free data points
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.Compute.df.data <- function(cfg)
{
        free.data <- sum(cfg$j.rg - 1)
        if(!cfg$rating) 
          return(free.data)
        
        free.data <- free.data + sum(cfg$n.sdt[1, ]*(cfg$n.sdt[2, ] - 1))
        return(free.data)
}

# ------------------------------------------------------------------------------------------------------------------
# Generation of data labels in output
# Input:    cfg     = configuration information
#           n.dat   = length of the data vector
#
# Output:   Vector with probabilities
# ------------------------------------------------------------------------------------------------------------------
HTSDT.Rank.Datalabels <- function(cfg)
{
      cfg <- HTSDT.Rank.Check.Fill.Cfg(cfg)
      
      # 1. Labels for ranking data
      label.out <- NULL
      l.rk <- length(cfg$j.rg)
      label.Rk <- paste("AFC-", cfg$k.rg, sep = "")
      for(i in 1:l.rk)
            label.out <- c(label.out, paste(label.Rk[i], " [", 1:cfg$j.rg[i], "]", sep = ""))
      if(cfg$rating == F) return(label.out)

      # 2. Labels for rating data
      for(i in 1:ncol(cfg$n.sdt)) {
          for(j in 1:cfg$n.sdt[1, i]) {
             label.out <- c(label.out, paste("SDT-", i, ".", j, " [", 1:cfg$n.sdt[2, i], "]", sep = ""))
          }
      }    
      return(label.out)
}