# ===========================================================================
# MIX-Rank models: Mixture model model for ranking and rating
# Date of creation:	June, 2013
# Date of last update:	July, 2013
#
# Author: Siegfried Macho
# ===========================================================================
library(statmod)
library(npmlreg)

MIX.Rk.Tol <- 1e-12
MINWEIGHT    <- 1e-40

#===============================================================================
# 0. COMPUTATION OF QUADRATURE POINTS
#===============================================================================

#-------------------------------------------------------------------------------
# Function for computing the Gauss-Hermite points and weights
# Input:	1. Number of points
#           2. minimal weights
#
# Output: 	Matrix with three columns:    1. Points
#                                         2. Weights of Gauss-Hermite
#                                         3. Weights of adaptive Gauss-Hermite 
#
# Comment:  The points and weights are with respect to the Gaussian Kernel N(0, 1)
#-------------------------------------------------------------------------------
Generate.GH <- function(npt, minweight = MINWEIGHT)
{
      GH.Structure  <- gqz(numnodes = npt, minweight)   # Points and weights for Gauss-Hermite Quadrature
      cbind(GH.Structure, GH.Structure[,2] * exp(GH.Structure[,1]^2/2))	 # Weights for adaptive Gauss-Hermite Quadrature
}

#===============================================================================
# 1. COMPUTATION OF PROBABILITY PARAMETERS
#===============================================================================

# ------------------------------------------------------------------------------
# Computation of probabilities in parameter vector
# Input:	parvec  = parameter vector
#           cfg     = configuration
#           fixed   = matrix of fixed constraints
#           ident   = matrix of idenity constraints
#
# Output: 	Parameter vector with raw probabilities replaced by 
#           computed probabilities
#
# Comment:  It is assumed that all equality constraints involving fixed 
#           parameters have been converted to fixed constraints
# ------------------------------------------------------------------------------
MIX.Rank.Ppars <- function(parvec, cfg, fixed, ident)
{
      n.rk <- length(cfg$j.rg)                        # Number of ranking models
      n.rt <- cfg$n.sdt                               # Number of rating  models      

      if(is.null(fixed)) fix <- -1
      else               fix <- fixed[2,]
      
      if(cfg$rating)
            pos.ppars <- c(1 + 7*(0:(n.rk-1)), 1 + 7*n.rk + 5*(0:(n.rt-1)))   # Positions of probability parameters
      else
            pos.ppars <- c(1 + 7*(0:(n.rk-1)))
            
      free <- !(pos.ppars %in% fix)
      free.ppars <- pos.ppars[free]                   # Positions of free probability parameters
      parvec[free.ppars] <- ifelse(parvec[free.ppars] > 700, 700, parvec[free.ppars])     # Avoid overflow
      parvec[free.ppars] <- exp(parvec[free.ppars]) / (1.0 + exp(parvec[free.ppars]))   # Compute probability
      return(parvec)
}


#===============================================================================
# 1. MODEL FUNCTIONS OF THE RANKING PART OF THE RANKING-RATING MODELL
#===============================================================================

# ----------------------------------------------------------------------
# Computation of conditional ranking probability of the target stimulus
# given the strength value x[1] of the target variable multiplied by
# the quadrature weight x[2]
# Input:	x  = vector with quadrature point x[1] and quadrature weight x[2]
#           j  = position of target variable x within the ranking (j = 1,...,k)
#           k  = Number of alternatives
#           mu =  mean of the non-target variables
#           sigma = standard deviation of the non-target variables
# Output: 	n - Vector, with n being the number of Response categories
# ----------------------------------------------------------------------
MIX.Rk.cond  <- function(x, j, k, mu, sigma)
{
      choose(k-1, j-1) * pnorm(x[1], mu, sigma)^(k-j) * (1-pnorm(x[1], mu, sigma))^(j-1) * x[2]
}

# ----------------------------------------------------------------------
# Computation of the unconditional ranking probability of target stimulus
# Input:	j             = position of target variable x within the ranking
#           k             = Number of alternatives
#           mu.target     = mean of the target distribution
#           sigma.targtet = standard deviation of the target distribution
#           mu            = mean of the non-target distribution
#           sigma         = standard deviation of the non-target distribution
#
# Output: 	n - Vector, with n being the number of Response categories
# ----------------------------------------------------------------------
MIX.Rk  <- function(j, k, mu.target, sigma.target, mu = 0, sigma = 1)
{
      GH.Structure[,1] <- GH.Structure[,1] * sigma.target + mu.target
      sum(apply(GH.Structure, 1, MIX.Rk.cond, j = j, k = k, mu = mu, sigma = sigma))
}

#-------------------------------------------------------------------------------
# Computation of ranking probabilties for one mixture model
# Input:
#     rg     = vector with two entries:
#
#     parvec = parameter vector with the following entries
#              1. k.rg:  number of alternatives to be ranked
#              2. j.rg: number of positions to be considered
#              3. probability of old.2
#              4. mean of the new distribution
#              5. sd   of the new distribution
#              6. mean of the old.1 distribution (weak targets)
#              7. sd   of the old.1 distribution (weak targets)
#              8. mean of the old.2 distribution (strong targets)
#              9. sd   of the old.2 distribution (strong targets)
#
# Output: Ranking probabilities
#-------------------------------------------------------------------------------
MIX.Rank.1 <- function(parvec, fixed, ident)
{
      # Compute Gaussian probability of first component
      if(parvec[3] > 0.0)            
            prb.1 <- sapply(1:parvec[2], MIX.Rk, k = parvec[1], mu.target = parvec[8], sigma.target = parvec[9], mu = parvec[4], sigma = parvec[5])
      else
            prb.1 <- 0
            
      # Compute Gaussian probability of second component
      if(parvec[3] < 1.0) 
            prb.2 <- sapply(1:parvec[2], MIX.Rk, k = parvec[1], mu.target = parvec[6], sigma.target = parvec[7], mu = parvec[4], sigma = parvec[5])
      else 
            prb.2 <- 0

      # Generate Mixture            
      prb  <- parvec[3] * prb.1 + (1 - parvec[3]) * prb.2
      return(c(prb, 1- sum(prb)))
}

#-------------------------------------------------------------------------------
# Computation of ranking probabilties for all mixtrue ranking models
#-------------------------------------------------------------------------------
MIX.RankOnly <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
      # Compile matrix with each row containing: k.rg, j.rg and the parameter vector of the model
      Rg <- matrix(c(cfg$k.rg, cfg$j.rg - 1), nc = 2)
      n  <- dim(Rg)[1]
      Rg <- cbind(Rg, matrix(parvec[1:(7*n)], nr = n, byrow = T))
      
      # Now apply the function HTSDT._Rank.1
      prb <- apply(Rg, 1, MIX.Rank.1, fixed = fixed, ident = ident)

      if(is.list(prb)) return(unlist(prb))
      return(as.vector(prb))
}

#===============================================================================
# 2. MODELL FUNCTIONS OF THE HTSDT-MODEL (RATING PART OF THE RANKING-RATING MODEL)
#===============================================================================
# --------------------------------------------------------------------------
# Function for computing response distribution of MIX model (three mixtures)
# Input:	Vector of parameters:
#           1. Probability of Target (Gauss 1)
#           2. Mean 0         (non-target)
#           3. Stddev 0       (non-target)
#           4. Mean 1         (target)
#           5. Stddev 1       (target)
#           6. c[1], c[2], ...., c[n-1]
#
# Output: 	Vector SDT response probabilities for a single model 
#           consisting of three Gaussians
# --------------------------------------------------------------------------
MIX.SDT.1 <- function(parvec)
{
      # Extract thrsholds and compute number of response categories
      th    <- parvec[6:length(parvec)]
      n.cat <- length(th) + 1 
      
      # Response probabilities of targets
      if(parvec[1] > 0.0)
      {
            Resp.1 <- c(pnorm(th, mean = parvec[4], sd = max(parvec[5], MIX.Rk.Tol)), 1) # Target weak
            Resp.1 <- c(Resp.1[1], Resp.1[-1] - Resp.1[-n.cat])
      }
      else
            Resp.1 <- 0
            
      # Response probabilities of non-targets
      if(parvec[1] < 1.0)
      {
            Resp.0 <- c(pnorm(th, mean = parvec[2], sd = max(parvec[3], MIX.Rk.Tol)), 1) # Non-Target
            Resp.0 <- c(Resp.0[1], Resp.0[-1] - Resp.0[-n.cat])
      }
      else
            Resp.0 <- 0
      
      # Compute Mixture
	parvec[1] * Resp.1 + (1.0 - parvec[1]) * Resp.0
}				

# ------------------------------------------------------------------------------------------------------------------
# Computation or probabilities of MIX-SDT model
# Input:	parvec = Vector of parameters, consisting of the following entries:
#
#                      I.  Parameters of the ranking part (ignored)
#                      II. For each signal of the Rating part
#                             1. Probability of Strong Gauss (2)
#                             2. Mean 0         (non-target)
#                             3. Stddev 0       (non-target)
#                             4. Mean 1         (target)
#                             5. Stddev 1       (target)
#                      II.  Identical for all models:  threshold parameters
#
#		cfg 	= List with configuration information
#           fixed = 2 x p matrix of fixed    constraints
#           ident = 2 x q matrix of equality constraints
#
# Output: Vector of probabilities for single models
# ------------------------------------------------------------------------------------------------------------------
MIX.SDT <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
      nrating     <- length(cfg$j.rg)                                   # Number of rating models
      start.sdt   <- 7*nrating + 1                                      # Start of SDT Parameters (first 6 parameters belong to Ranking model
      start.th    <- 5*cfg$n.sdt + start.sdt                            # Startindex for thresholds
	thresholds 	<- parvec[start.th:length(parvec)]                    # Threshold parameters
      
                        
      Par 		<- matrix(parvec[start.sdt:(start.th-1)], nr = cfg$n.sdt, byrow = T)  # Matrix with probability and Gaussian parameters of HTSDT1 Model
	thresholds 	<- matrix(rep(thresholds, cfg$n.sdt), nr = cfg$n.sdt, byrow = T)      # Matrix with thresholds for each model
      Par         <- cbind(Par, thresholds)
	     
	# Compute probabilities
	as.vector(apply(Par, 1, MIX.SDT.1))
}

#===============================================================================
# 3. MODELL FUNCTIONS OF THE (COMBINED) RANKING-RATING MODEL
#===============================================================================

#-------------------------------------------------------------------------------
# Computation of ranking and rating probabilties
# Input:
#     parvec = parameter vector with the following entries
#              I: For each Ranking model 7 parameters
#                 1. probability of old.2
#                 2. mean of the new distribution
#                 3. sd   of the new distribution
#                 4. mean of the old.1 distribution (weak targets)
#                 5. sd   of the old.1 distribution (weak targets)
#                 6. mean of the old.2 distribution (strong targets)
#                 7. sd   of the old.2 distribution (strong targets)
#
#             II: For each RAting model 5 parameters
#                 1. probability of target
#                 2. mean of the non-target distribution
#                 3. sd   of the non-target distribution
#                 4. mean of the target distribution
#                 5. sd   of the target distribution
#
#            III: Threshold parameter c[1], c[2], ...., c[k]
#
#     cfg    = list with configuration information containing the following 
#              entries:
#                 n.sdt  = number of MIX Models
#                 k.rg   = number of alternatives to be ranked
#                 j.rg   = number of positions to be considered
#                 npt    = Number of quadrature points
#                 rating = Flag, indicating whether rating data are present
#
#      fixed = matrix of fixed constraints
#      ident = matrix of equality constraints
#     with.p = flag indicating whether probability parameters should be 
#              computed from raw parameters (default = TRUE)
#
# Output: Ranking and rating probabilities
#-------------------------------------------------------------------------------
MIX.Rank <- function(parvec, cfg, fixed = NULL, ident = NULL, with.p = T)
{
      # 1. Compute probability parameters if required
      if(with.p)
            parvec <- MIX.Rank.Ppars(parvec, cfg, fixed, ident)

      # 2. Compute model probabilities
      pr.Rg <- MIX.RankOnly(parvec, cfg, fixed, ident)

      if(cfg$rating == F) return(pr.Rg)
      c(pr.Rg, MIX.SDT(parvec, cfg, fixed, ident))
}

# ------------------------------------------------------------------------------------------------------------------
# Model incorporating restrictions
# ------------------------------------------------------------------------------------------------------------------
MIX.Rank.Full <- function(parvec, cfg, pos, fixed = NULL, ident = NULL, functional = NULL, with.p = T)
{
      parvec <- Restore.par(parvec, pos, fixed, ident, functional)	# Restore full parameter vector
      MIX.Rank(parvec, cfg, fixed, ident, with.p)

}

#===============================================================================
# 4. JACOBIANS FOR THE RANKING PART OF THE RANKING-RATING MODEL
#===============================================================================

#-------------------------------------------------------------------------------
# Derivative of the conditional model function with respect to x, mu and sigma
#-------------------------------------------------------------------------------
SDT.Rk.cond.Jacobi  <- function(x, j, k, mu, sigma)
{
      sigma <- max(sigma, MIX.Rk.Tol)
      z  <- (x[1]-mu)/sigma
      fn <- dnorm(z)
      dm <- -fn/sigma			# Derivative of Gaussian with respect to m
	ds <- -z*fn/sigma			# Derivative of Gaussian with respect to sd
      dx <-  fn/sigma               # Derivative of Gaussian with respect to x

      bin <- choose(k-1, j-1)
      pn  <- pnorm(x[1], mu, sigma)
      p1  <- pn^(k-j)
      p2  <- (1-pn)^(j-1)
      
      if(pn < MIX.Rk.Tol)         # 0^0 = 1
      {
            if(k-j-1 == 0)    dp1 <- 1
            else              dp1 <- 0
      }      
      else
            dp1 <- (k-j)*pn^(k-j-1)
            
      if((1-pn) < MIX.Rk.Tol)     # 0^0 = 1
      {
            if(j == 2)    dp2 <- 1
            else          dp2 <- 0
      }      
      else
            dp2 <- (j-1)*(1-pn)^(j-2)
      
      bin*(dp1*c(dm, ds, dx)*p2 + p1*c(-dm, -ds, -dx)*dp2)*x[2]
}

#-------------------------------------------------------------------------------
# Derivative of the Mixture Rank model for one Rank model
#     - p.mix           (mixing probability of target.strong)
#     - mean.0          (mean, non-target
#     - stddev.0        (stddev, non-target)
#     - mean.1          (mean, target-weak)
#     - stddev.1        (stddev, target-weak) 
#     - mean.2          (mean, target-strong)
#     - stddev.2        (stddev, target-strong) 
#
# Input:    j               actual rank
#           k               number of ranks
#           p.mix           mixing probability
#           mean.0          mean, non-target
#           sigma.0         stddev, non-target
#           mean.1          mean, target-weak
#           sigma.1         stddev, target-weak 
#           mean.2          mean, target-strong
#           sigma.2        stddev, target-strong 
#
# Output: Vector of derviatives in the order shown above
#-------------------------------------------------------------------------------
MIX.Rank.1.Jacobi  <- function(j, k, p.mix, mu.0, sigma.0, mu.1, sigma.1, mu.2, sigma.2)
{
      # 1. Compute of derivatives with respect to Gaussian parameters
      Save.GH.Struc    <- GH.Structure[,1]
      J.Gauss <- rep(0.0, 6)                      # Vector of derivatives with respect to the 6 Gaussian parameters
      
      # mu.2, sigma.2
      if(p.mix > 0.0)
      {
            GH.Structure[,1] <- Save.GH.Struc * sigma.2 + mu.2
            J.Rk <- apply(GH.Structure, 1, SDT.Rk.cond.Jacobi, j = j, k = k, mu = mu.0, sigma = sigma.0)
            J.Rk <- rbind(J.Rk, J.Rk[3,]* Save.GH.Struc)          # Computation of derivative with respect to sigma.2
            J.Gauss[c(1, 2, 5, 6)] <- p.mix * rowSums(J.Rk)
      }
       
      # mu.1, sigma.1
      if(p.mix < 1.0)
      {
            GH.Structure[,1] <- Save.GH.Struc * sigma.1 + mu.1
            J.Rk <- apply(GH.Structure, 1, SDT.Rk.cond.Jacobi, j = j, k = k, mu = mu.0, sigma = sigma.0)
            J.Rk <- rbind(J.Rk, J.Rk[3,]* Save.GH.Struc)          # Computation of derivative with respect to sigma.1
            J.Gauss[1:4] <- J.Gauss[1:4] + (1.0 - p.mix)*rowSums(J.Rk)
      }
      
      # 2. Compute derivative of p.mix
      dp.mix <- MIX.Rk(j, k, mu.2, sigma.2, mu.0, sigma.0) - MIX.Rk(j, k, mu.1, sigma.1, mu.0, sigma.0)

      # 5. Now return full vector
      c(dp.mix, J.Gauss)
      
}

#-------------------------------------------------------------------------------
# Jacobian of the model function of the ranking model for a single model:
# Input: 1. parvec = parameter vector of the actual model
#                       p.mix           mixing probability
#                       mean.0          mean, non-target
#                       sigma.0         stddev, non-target
#                       mean.1          mean, target-weak
#                       sigma.1         stddev, target-weak 
#                       mean.2          mean, target-strong
#                       sigma.2         stddev, target-strong 
#        2. k.rg   = Number of presented alternatives
#        3. j.rg   = Number of evaluated positions
#        4. fixed  = Matrix of fixed constraints (not used)
#        5. ident  = Matrix of identity constraints (not used)
#
# Jacobian matrix of the probabilities with respect to the six parameters
# of a single ranking model
#-------------------------------------------------------------------------------
MIX.__Rank.Jacobi <- function(parvec, k.rg, j.rg, fixed = NULL, ident = NULL)
{
      p.J <- sapply(1:(j.rg-1), MIX.Rank.1.Jacobi, k = k.rg, p.mix = parvec[1], mu.0 = parvec[2], sigma.0 = parvec[3], 
                                                   mu.1  = parvec[4], sigma.1 = parvec[5], mu.2 = parvec[6], sigma.2 = parvec[7])
      rbind(t(p.J), -rowSums(p.J))
}

#-------------------------------------------------------------------------------
# Jacobian of the model function of the ranking part for full ranking model:
# Input: 1. parvec = fulle parameter vector
#        2. cfg    = Configuration information
#        3. fixed  = Matrix of fixed constraints (not used)
#        4. ident  = Matrix of identity constraints (not used)
#
# Jacobian matrix of the probabilities with respect to the parameters of all  
# the ranking models
#-------------------------------------------------------------------------------
MIX._Rank.Jacobi <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
      n.rk <- length(cfg$j.rg)
      if(n.rk < 2)                       # Only a single model
            return(MIX.__Rank.Jacobi(parvec[1:7], cfg$k.rg, cfg$j.rg, fixed, ident))

      # More than a single model
      Pmatrix <- matrix(parvec[1:(7*n.rk)], nr = n.rk, byrow = T)
      for(i in 1:n.rk)
      {
            J.mod <- MIX.__Rank.Jacobi(Pmatrix[i, ], cfg$k.rg[i], cfg$j.rg[i], fixed, ident)
            
            if(i > 1)
                  J <- rbind(cbind(J, matrix(0, nr = sum(cfg$j.rg[1:(i-1)]), nc = 7)), 
                             cbind(matrix(0, nr = cfg$j.rg[i], nc = 7*(i-1)), J.mod))
            else
                  J <- J.mod
      }            
      return(J)
}


#===============================================================================
# 5. JACOBIANS FOR THE RATING PART OF THE RANKING MODEL
#===============================================================================

# ------------------------------------------------------------------------------
# Jacobi-Matrix of parameters and thresholds of a single Gaussian SDT model
# Input:	I. Vector of parameters (without thresholds):
#                 1. Mean
#                 2. Standard deviation
#
#           II. Threshold parameters: c[1], c[2], ...., c[n-1])
#
# Output: 	Jacobian of the SDT model for one signal
# ------------------------------------------------------------------------------
MIX.Rank.Gaussian.Jacobi <- function(gausspar, th)
{
	n.t <- length(th)					# Number of thresholds
	m   <- gausspar[1]
	sd  <- max(MIX.Rk.Tol, gausspar[2])
	z  <- (th - m)/sd
	fn <- dnorm(z)
	
	# Derivative with respect to mean
	dm <- -fn/sd						# Derivative of Gaussian with respect to m
	dm <- c(dm[1], dm[-1] - dm[-n.t], -dm[n.t])

	# Derivative with respect to sd
	ds <- -z*fn/sd						# Derivative of Gaussian with respect to sd
	ds <- c(ds[1], ds[-1] - ds[-n.t], -ds[n.t])


	# Derivative with respect to thresholds
	dt <- fn/sd							# Derivative of Gaussian with respect to t[i]
	
	J <- matrix(0, nr = n.t+1, nc = n.t)		# Jacobi matrix for thresholds: columns: Parameter, rows = probabilities
	r <- 1
	for(i in 1:n.t)
	{
		J[r, i] <- dt[i]
		r <- r + 1
		J[r, i] <- -dt[i]
	}
	cbind(dm, ds, J)						# Full Jacobi matrix (est.p x par)
}

# ------------------------------------------------------------------------------
# Jacobi-Matrix of parameters and thresholds of a single mixture model
# Input:	I. Vector of parameters (without thresholds):
#                 1. Mixing probability
#                 2. Mean (non-target)
#                 3. Standard deviation (non-target)
#                 4. Mean (target)
#                 5. Standard deviation (target)
#
#           II. Threshold parameters: c[1], c[2], ...., c[n-1])
#
# Output: 	Jacobian of the mixture model for a single signal
# ------------------------------------------------------------------------------
MIX.Rank.SDT.1.Jacobi <- function(parvec, th)
{
      n.cat <-length(th) + 1 
      # 1. Compute Jacobian Matrix for Gaussian and threshold parameters
      if(parvec[1] > 0.0)
            J <- parvec[1] * MIX.Rank.Gaussian.Jacobi(parvec[4:5], th)
      if(parvec[1] < 1.0)
      {
            J.2 <- (1.0 - parvec[1]) * MIX.Rank.Gaussian.Jacobi(parvec[2:3], th)
            if(parvec[1] > 0.0)
                  J <- cbind(J.2[ ,1:2], J[ ,1:2], J.2[,-(1:2)] + J[, -(1:2)])
            else
                  J <- cbind(J.2[ ,1:2], matrix(0, nr = n.cat, nc = 2), J.2[, -(1:2)])
      } 
      else
            J <- cbind(matrix(0, nr = n.cat, nc = 2), J)

      # 2. Compute derivation vector for p.mix
      Resp.1 <- c(pnorm(th, mean = parvec[4], sd = max(parvec[5], MIX.Rk.Tol)), 1) # Target weak
      Resp.1 <- c(Resp.1[1], Resp.1[-1] - Resp.1[-n.cat])
     
      Resp.0 <- c(pnorm(th, mean = parvec[2], sd = max(parvec[3], MIX.Rk.Tol)), 1) # Non-Target
      Resp.0 <- c(Resp.0[1], Resp.0[-1] - Resp.0[-n.cat])
     
      cbind(Resp.1 - Resp.0, J)
}
     
# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix of the rating part of the mixture model
# ------------------------------------------------------------------------------------------------------------------
MIX.Rank.SDT.Jacobi <- function(parvec, cfg)
{
      
      # Isolate threshold parameters
      start.th <- 1 + 7*length(cfg$j.rg) + 5*cfg$n.sdt
	th <- parvec[start.th : length(parvec)]	  # Thresholds
	n.t <- length(th)					  # Number of thresholds
      n.cat <- n.t + 1

      # Matrix of paramteters for single models
      Pmatrix <- matrix(parvec[(7*length(cfg$j.rg)+1):(7*length(cfg$j.rg) + 5*cfg$n.sdt)], nr = cfg$n.sdt, byrow = T)
 
      for(i in 1:cfg$n.sdt)
      {
            J.mod <- MIX.Rank.SDT.1.Jacobi(Pmatrix[i,], th)
            J.G <- J.mod[, 1:5]
            J.t <- J.mod[, 6:(5 + n.t)]

            # New Jacobian matrix for Gaussian model and thresholds
            if(i < 2)
            {
                  J.Gauss <- J.mod[, 1:5]
                  J.th    <- J.mod[, 6:(5 + n.t)]
            }
            else
            {
                  J.Gauss <- rbind(cbind(J.Gauss, matrix(0, nr = n.cat*(i-1), nc = 5)), 
                                   cbind(matrix(0, nr = n.cat, nc = 5*(i-1)), J.mod[, 1:5]))
                  J.th    <- rbind(J.th, J.mod[, 6:(5 + n.t)])
            }      
      }
      cbind(J.Gauss, J.th)
}

#===============================================================================
# 6. JACOBI MATRIX, GRADIENT AND MODEL MATRICES OF THE RATING-RANKING MODEL
#===============================================================================

#-------------------------------------------------------------------------------
# Jacobian of the model function of the Ranking and Rating model:
#-------------------------------------------------------------------------------
MIX.Rank.Jacobi <- function(parvec, cfg, fixed = NULL, ident = NULL, with.p = T)
{
      # 0. Compute probability parameters
      if( with.p)
            parvec <- MIX.Rank.Ppars(parvec, cfg, fixed, ident)

      # 1. Jacobi-Matrix of the ranking part
      rank.J <- MIX._Rank.Jacobi(parvec, cfg, fixed, ident)
      if(cfg$rating == F)     # No rating 
            return(rank.J %*% MIX.Rank.Ppars.Jacobi(parvec, cfg, fixed, ident))
      
      # 2. Jacobi-Matrix of the rating part
      rating.J <- MIX.Rank.SDT.Jacobi(parvec, cfg)
      
      n.par.Rk   <- 7*length(cfg$j.rg)                 # Number of Ranking parameters
      n.par.SDT  <- length(parvec) - n.par.Rk          # Number of SDT parameters
      filler.Rk  <- matrix(0, nr = sum(cfg$j.rg), nc = n.par.SDT)
      filler.SDT <- matrix(0, nr = dim(rating.J)[1],  nc = n.par.Rk)
      
      J <- rbind(cbind(rank.J, filler.Rk), cbind(filler.SDT, rating.J))

      # 3. Adjust for probability paramters
      return(J %*% MIX.Rank.Ppars.Jacobi(parvec, cfg, fixed, ident))
}


#-------------------------------------------------------------------------------
# Analytically computed model matrix
# implementing functional and identity constraints
# Input:   1. full.par   = Full parameter vector
#          2. cfg        = List with configuration information
#          3. fixed      = Matrix of fixed constraints
#          4. ident      = Matrix of equality constraints
#          5. functional = A function implementing functional constraints
#
# Output: Model matrix of free parameters
#-------------------------------------------------------------------------------
MIX.Rank.Model.Matrix <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos      <- SavePos.par(full.par, fixed, ident)
      parvec   <- Elim.par(full.par, fixed, ident)
      full.par <- Restore.par(parvec, pos, fixed, ident, functional)	# Restore full parameter vector

      J   <- MIX.Rank.Jacobi(full.par, cfg, fixed, ident)
      if(!is.null(ident))
      {
            J.Id <- diag(length(full.par))
            Index <- matrix(c(ident[2,], ident[1,]), nc = 2)
            J.Id[Index] <- 1
            J <- J %*% J.Id
      }
      
      # Treatment of functional constraints
      if(!is.null(functional))
      {
            J.fct <- attr(functional, "SDT.Jacobian")
            if(is.null(J.fct)) 
                  return(NULL)           # No Jacobian function supplied -> return (NULL)
            J <- J %*% J.fct(full.par)
      }
      J[, pos]
}

# ------------------------------------------------------------------------------------------------------------------
# Numerical computation of the model matrix (MIX-Rank model)
# ------------------------------------------------------------------------------------------------------------------
MIX.Rank.Model.Matrix.Num <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)
      parvec <- Elim.par(full.par, fixed, ident)
      J <- jacobian(MIX.Rank.Full, parvec, cfg = cfg, pos = pos, fixed = fixed, ident = ident, functional = functional)
}

# ------------------------------------------------------------------------------------------------------------------
# Gradient of logL for the HTSDT-Rank model
# Input:	Same as for Likelihood function
# Output:	Gradient
#
# Comment:  Identity constraints betreen probability prameters from different blocks are removed from matrix 
#           ident and added to matrix fixed.
# ------------------------------------------------------------------------------------------------------------------
MIX.Rank.Grad <- function(parvec, datavec, Model, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
	parvec   <- Restore.par(parvec, pos, fixed, ident, functional)	# Restore full parameter vector
	gradient <- as.vector(SDT.logL.Jacobi(parvec, datavec, Model, n, pos, fixed, ident, functional) %*% MIX.Rank.Jacobi(parvec, n, fixed, ident)) 

	# Treat functional constraints
	if(!is.null(functional))
	{
		gr.fct <- attr(par, "SDT.gradient")		# Isolate gradient function
		if(is.null(gr.fct))	stop("\n\n MIX.Rank.Grad: Error in computation of gradient: No gradient for functional constraints found:\n Set Parameter sym.gr = F in function SDT.Estimate !\n\n")
		gradient <- gr.fct(gradient)
	}	
	Elim.par.grad(gradient, fixed, ident)
}

#===============================================================================
# 7. JACOBIAN FOR ADJUSTING PROBABILITY PARAMETERS
#===============================================================================
# --------------------------------------------------------------------------------
# Computation of Jacobian for adjusting probability paraemters
# Input:	1. parvec = full parameter vector with probability parameters computed
#           2. cfg    = list with configuration information
#           3. fixed  = matrix of fixed constraints
#           4. ident  = matrix of equality constraints
#
# Output:	Jacobian matrix for adjusting probability parameters
#
# Comment:  Fixed constraints are taken into account
# --------------------------------------------------------------------------------
MIX.Rank.Ppars.Jacobi <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
      n.rk <- length(cfg$j.rg)
      # 1. Jacobian of probability parameters with respect to raw parameters
      if(cfg$rating == F) 
            ppos <- 1 + 7*(0:(n.rk-1))         # Positions of mixing probabilities in ranking part
      else                
            ppos <- c(1 + 7*(0:(n.rk-1)), 1 + 7*n.rk + 5*(0:(cfg$n.sdt-1))) # Positions of mixing probabilities in ranking and rating part 

      if(!is.null(fixed)) fx <- fixed[2,]
      else                fx <- NULL
      
      # 2. Determine positions of free mixing probabilities
      fx.par <- (ppos %in% fx)           # Boolean vector indicating fixed probability parameters
      ppos.fix  <- ppos[fx.par]
      ppos.free <- ppos[!fx.par]
      ppar.free <- parvec[ppos.free]

      # Diagonal entries
      dg  <- rep(1, length(parvec))        # Diagonal elements of Jacobian
      dg[ppos.fix] <- 0                    # Set fixed parameters to zero
      dg[ppos.free] <- ppar.free * (1.0 - ppar.free)
      
      return(diag(dg))
}

# ------------------------------------------------------------------------------------------------------------------
# Help function that returns the positions of the free probability parameters
# Input:	1. parvec = full parameter vector with probability parameters computed
#           2. cfg    = list with configuration information
#           3. fixed  = matrix of fixed constraints
#           4. ident  = matrix of equality constraints
#           5. full   = flag indicating whether positions are with respect to the full or the restricted parameter vector
#
# Output:	vector of postions of free probability parameters or NULL
# ------------------------------------------------------------------------------------------------------------------
MIX.Rank.Free.Ppos <- function(parvec, cfg, fixed, ident, full = T)
{
      # 1. Postions of probability parameters
      n.rk <- length(cfg$j.rg)
      if(cfg$rating)
            ppos <- c(1 + 7*(0:(n.rk-1)), 1 + 7*n.rk + 5*(0:(cfg$n.sdt-1)))
      else
            ppos <-   1 + 7*(0:(n.rk-1))

      # 2. Identify free probability parameters
      if(is.null(fixed) && is.null(ident)) 
            return(ppos)
      else if(is.null(fixed))
            free.p   <- !(ppos %in% ident[2, ])
      else if(is.null(ident))
            free.p   <- !(ppos %in% fixed[2, ])
      else
            free.p   <- !((ppos %in% fixed[2, ]) | (ppos %in% ident[2, ]))
            
      pos <- SavePos.par(parvec, fixed, ident)  # Positions of free parameters
      if(length(pos) == 0) return(NULL)

      if(full)
            return(pos[pos %in% ppos[free.p]])        # Return position of free probability parameters with respect to full  parameter vector
      return((1:length(pos))[pos %in% ppos[free.p]])  # Return positions of free probability parameters with respect to free parameter vector 
}

# ------------------------------------------------------------------------------------------------------------------
# Computing adjusted standard errors for probability parameters
# Input:	1. cov.matrix = covariance matrix of estimated parameters
#           2. parvec     = full parameter vector with probability parameters computed
#           3. cfg        = list with configuration information
#           4. fixed      = matrix of fixed constraints
#           5. ident      = matrix of equality constraints
#
# Output:	Vector of standand errors (adjusted) or NULL 
# ------------------------------------------------------------------------------------------------------------------
MIX.Rank.Delta.Method.Prob <- function(cov.matrix, parvec, cfg, fixed, ident)
{
      # 1. Extract Standard errors
      if(is.null(cov.matrix) || is.na(cov.matrix) || is.nan(cov.matrix))
            return(NULL)
      SE.vector <- sqrt(diag(cov.matrix))       # Vector of unadjusted standard errors

      # 2. Compute restricted parameter vector
      par.restrict <- Elim.par(parvec, fixed, ident)
            
      # 2. Positions of free parameters with respect to restricted parameter vector     
      free.ppos <- MIX.Rank.Free.Ppos(parvec, cfg, fixed, ident, full = F)

      # 3. Adjuste the standard errors for probabilities
      SE.vector[free.ppos] <- SE.vector[free.ppos] * par.restrict[free.ppos] * (1.0 - par.restrict[free.ppos]) 
      
      # 4. Construct adjusted SE vector and fill SE vector with adjusted values
      return(SE.vector)
}

#===============================================================================
# 8. ADMINISTRATIVE FUNCTIONS
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Check and generate configuration structure
# Input:	cfg = Configuration structure: A list with the following entries
#                 - n.sdt     Number of SDT models
#                 - k.rg      Number of alternatives
#                 - j.rg      Number of relevant ranking positions
#                 - npt       Number of quadrature points
#                 - standard  Flag indicating whether standard restrictions have to be specified
#                 - rating    Flag indicating whether ranking and rating are combined
#
# Output: String with parameter values
# ------------------------------------------------------------------------------------------------------------------
MIX.Rank.Check.Fill.Cfg <- function(cfg)
{
      n.sdt <- 2
      k.rg  <- 2
      npt   <- 30

      if(is.null(cfg))
      {
            cfg <- list(n.sdt       = n.sdt,
                        k.rg        = k.rg,  
                        j.rg        = k.rg, 
                        npt         = npt, 
                        restriction = "standard",
                        rating      = T)
            return(cfg)
      }
      if(is.numeric(cfg))
      {
            k.rg     <- as.integer(cfg)
            cfg      <- list(n.sdt       = n.sdt,
                             k.rg        = k.rg,  
                             j.rg        = k.rg, 
                             npt         = npt, 
                             restriction = "standard",
                             rating      = T)
            return(cfg)
      }
      cfg.new  <- list(n.sdt = ifelse(is.null(cfg$n.sdt), n.sdt, cfg$n.sdt))
      
      if(is.null(cfg$k.rg))
          if(is.null(cfg$j.rg)) cfg.new$k.rg <- k.rg
          else                  cfg.new$k.rg <- cfg$j.rg
      else 
          cfg.new$k.rg <- cfg$k.rg

      if(is.null(cfg$j.rg))    cfg.new$j.rg <- cfg.new$k.rg
      else                     cfg.new$j.rg <- cfg$j.rg

      cfg.new$npt         <- ifelse(is.null(cfg$npt),  npt, cfg$npt)
      cfg.new$restriction <- MIX.Restriction(cfg$restriction)
      cfg.new$rating      <- ifelse(is.null(cfg$rating),    T, cfg$rating)

      # Checks
      if(cfg.new$n.sdt < 2)
            stop(paste("\n FATAL ERROR in MIX.Rank.Check.Fill.Cfg: Number of SDT signals <", cfg.new$n.sdt, "> smaller than <2>\n (Minimum number = 2)\n", sep = ""))
      if(any(cfg.new$j.rg < 2))
            stop(paste("\n FATAL ERROR in MIX.Rank.Check.Fill.Cfg: Number of estimated position <", cfg.new$j.rg, "> smaller than <2>\n (Minimum number = 2)\n", sep = ""))
      if(length(cfg.new$k.rg) != length(cfg.new$j.rg))
            stop(paste("\n FATAL ERROR in MIX.Rank.Check.Fill.Cfg: Different number of ranking models for k.rg <", length(cfg.new$k.rg), "> and j.rg <",length(cfg.new$j.rg), ">\n", sep = ""))
      if(any(cfg.new$k.rg < cfg.new$j.rg)) 
            stop(c("\n FATAL ERROR in MIX.Rank.Check.Fill.Cfg:At least one number of alternatives smaller than one number of relevant ranks\n"))
      return(cfg.new)
}                  
  
# ------------------------------------------------------------------------------------------------------------------
# Check type of restriction
# Input: A string specfiying the typ of restriction
#        Alternatively the flags TRUE (="standard") and FALSE ("No") are allowed
#        Types of restrictions      - "standard"  [ or T]   (Standard restrictions)
#                                   - "SDT"                 (Simulating SDT model)
#                                   - "extended"            (With varying mean in case of missing)
#                                   - "no"  [or F]          (No restrictions)
#
# Output: String with parameter values
# ------------------------------------------------------------------------------------------------------------------
MIX.Restriction <- function(restriction)
{
      if(is.null(restriction) || restriction == T)
            return("standard")
      if(charmatch(restriction, c("standard", "STANDARD", "Standard"), nomatch = 0) > 0)
            return("standard")
      if(charmatch(restriction, c("sdt", "SDT", "Sdt"), nomatch = 0) > 0)
            return("SDT")
      if(charmatch(restriction, c("extended", "EXTENDED", "Extended"), nomatch = 0) > 0)
            return("extended")
      if(charmatch(restriction, c("no", "No", "NO"), nomatch = 0) > 0 || restriction == F)
            return("no")
      warn.msg <- paste("\n\n MIX.Restriction: Unkonwn restriction in configuration list: <", restriction, "> --> setting to: <standard>\n\n", sep = "")      
      warning(warn.msg)
      return("standard")
}            
                        
# ------------------------------------------------------------------------------------------------------------------
# Generate starting parameters due to data and number of signals
# Input:	cfg = Configuration structure: A list with the following entries
#                 - k.rg      Number of alternatives
#                 - j.rg      Number of relevant ranking positions
#                 - npt       Number of quadrature points
#                 -standard   Flag indicating whether standard restrictions have to be specified
#                 -rating     Flag indicating whether ranking and rating are combined
#           data = Data vector
#
# Output: String with parameter values depending on whether only ranking data or
#         ranking as well as reting data are modeled
# ------------------------------------------------------------------------------------------------------------------
MIX.Rank.start.par <- function(cfg, datavec)
{
      
      # If configuration information is not complete, then complete it
      if(is.null(cfg$rating) || is.null(cfg$n.sdt) || is.null(cfg$j.rg) || is.null(cfg$k.rg))
          cfg <- MIX.Rank.Check.Fill.Cfg(cfg)  
      
      #-------------------------------------------------------------------------
      # I. Parameters RANKING MODELS
      # 1. p.mix
      # 2. m.0
      # 3. sd.0
      # 4. m.1
      # 5. sd.1
      # 6. m.2
      # 7. sd.2
      #
      # II. Parameters RATING MODELS
      # 1. p.mix
      # 2. m.0
      # 3. sd.0
      # 4. m.1
      # 5. sd.1
      #
      # IV. Threshold parameter
      # c[1], ..., c[k]
      #-------------------------------------------------------------------------
      if(cfg$rating)          # Ranking and rating data
      {
            if(is.null(datavec)) stop("\n FATAL ERROR in MIX.Rank.start.par: Cannot compute start parameters due to missing data vector\n")
            n.th <- (length(datavec) - sum(cfg$j.rg))/cfg$n.sdt  - 1    # Number of thresholds
            if(n.th <= 0) 
            {
                  msg <- paste("\n\n FATAL ERROR in MIX.Rank.start.par: Not enough data points for estimating specified rating and ranking model:\n",
                               " Length of data vector  : <", length(datavec), ">\n",
                               " Minimum length required: <", sum(cfg$j.rg) + 2*cfg$n.sdt, ">\n\n", sep = "")
                  stop(msg)
            }
            
            return(c(rep(c(0.0, 0.0, 1.0, 0.0, 1.0, 0.5, 1.0), length(cfg$k.rg)), # Parameters of the ranking models
                     rep(c(0.0, 0.0, 1.0, 0.5, 1.0)          , cfg$n.sdt),           # Parameters of the sdt     models
                     seq(-0.5, 0.5, length = n.th)))           # Thresholds
      }
      return(rep(c(0.0, 0.0, 1.0, 0.0, 1.0, 0.5, 1.0), length(cfg$k.rg)))     # Ranking data only
}

# ------------------------------------------------------------------------------------------------------------------
# Generate upper and lower bounds of parameter
# Input:	parvec = Vector of parameters
#		   cfg = configruation structure
#
# Output: List with lower and upper bounds on parameter vector
# ------------------------------------------------------------------------------------------------------------------
MIX.Rank.par.bounds <- function(parvec, cfg)
{
      min.p <- -700
      max.p <-  700

      len <- length(cfg$k.rg)
      len.p <- length(parvec)
      lower <- rep(c(min.p, -Inf, MIX.Rk.Tol, -Inf, MIX.Rk.Tol, -Inf, MIX.Rk.Tol), len)
      upper <- rep(c(max.p,  Inf,        Inf,  Inf,        Inf,  Inf,        Inf), len)
      
      if(cfg$rating == F)
            return(list(lower = lower, upper = upper))
      
      n.th <-  len.p - 7*len - 5*cfg$n.sdt
      lower <- c(lower, rep(c(min.p, -Inf, MIX.Rk.Tol, -Inf, MIX.Rk.Tol), cfg$n.sdt), rep(-Inf, n.th))
      upper <- c(upper, rep(c(max.p,  Inf,        Inf,  Inf,        Inf), cfg$n.sdt), rep( Inf, n.th))

      return(list(lower = lower, upper = upper))
}      

# ------------------------------------------------------------------------------------------------------------------
# Generate strings with names of the parameters for the models
# Input:	parvec         = Vector of parameters
#		   cfg         = configruation structure
#           with.space     = Flag indicating whether names should be with space (default) or not
# Output: String with parameter names
# ------------------------------------------------------------------------------------------------------------------
MIX.Rank.par.names <- function(parvec, cfg, with.space = T)
{
      # If configuration information is not complete, then complete it
      if(is.null(cfg$rating) || is.null(cfg$n.sdt) || is.null(cfg$j.rg) || is.null(cfg$k.rg))
          cfg <- MIX.Rank.Check.Fill.Cfg(cfg)  
      
      if(with.space)
      {
            Ranking.names <- c("p.mix    (Rank", 
                               "Mean.0   (Rank", 
                               "Stddev.0 (Rank", 
                               "Mean.1   (Rank", 
                               "Stddev.1 (Rank", 
                               "Mean.2   (Rank", 
                               "Stddev.2 (Rank")
                         
            Rating.names  <- c("p.mix     (Mix", 
                               "Mean.0    (Mix", 
                               "Stddev.0  (Mix", 
                               "Mean.1    (Mix", 
                               "Stddev.1  (Mix")
     }
      else
      {
            Ranking.names <- c("p.mix (Rank", 
                               "Mean.0 (Rank", 
                               "Stddev.0 (Rank", 
                               "Mean.1 (Rank", 
                               "Stddev.1(Rank", 
                               "Mean.2 (Rank", 
                               "Stddev.2 (Rank")
                         
            Rating.names  <- c("p.mix (Mix", 
                               "Mean.0 (Mix", 
                               "Stddev.0 (Mix", 
                               "Mean.1 (Mix", 
                               "Stddev.1 (Mix")
      }
      models <- rep(1:length(cfg$k.rg), each = 7)
      Ranking.names <- paste(Ranking.names, " ", models, ")", sep = "")

      if(cfg$rating)
      {
            models <- rep(1:cfg$n.sdt, each = 5)
            Rating.names <- paste(Rating.names, " ", models, ")", sep = "")
            R.names <- c(Ranking.names, Rating.names)
            n.th <- length(parvec) - length(R.names)
            return(c(R.names, paste("c[", 1:n.th, "]", sep = "")))
      }
      return(Ranking.names)
}

# ------------------------------------------------------------------------------------------------------------------
# Names of the models 
# ------------------------------------------------------------------------------------------------------------------
MIX.Rank.model.name <- function(cfg)
{
      # If configuration information is not complete, then complete it
      if(is.null(cfg$rating) || is.null(cfg$n.sdt) || is.null(cfg$j.rg) || is.null(cfg$k.rg))
          cfg <- MIX.Rank.Check.Fill.Cfg(cfg)  

      
      k.rg.s <- paste(cfg$k.rg, collapse = ", ")
      j.rg.s <- paste(cfg$j.rg, collapse = ", ")

      if(cfg$rating)
            mod.name <- paste("MIX ranking and rating model:", 
                              " Number of Mixture models: <", cfg$n.sdt, ">", 
                              " Number of ranking-alternatives: <", k.rg.s, ">", 
                              " Number of ranking-positions: <", j.rg.s, ">", 
                              " Number of quadrature points: <", cfg$npt, ">", sep = "")
      else
            mod.name <- paste("MIX ranking model:", 
                              " Number of ranking-alternatives: <", k.rg.s, ">", 
                              " Number of ranking-positions: <", j.rg.s, ">", 
                              " Number of quadrature points: <", cfg$npt, ">", sep = "")
      
     return(paste(mod.name, " Type of restrictions: <", cfg$restriction, ">", sep = ""))
}       

# ------------------------------------------------------------------------------------------------------------------
# Check number of parameters and data points
# ------------------------------------------------------------------------------------------------------------------
MIX.Rank.check <- function(parvec, datavec, cfg)
{
      Msg <- "NOTE: Computation of the expected number of parameters is based on the number of data points and vice versa"
	
      plen     <- length(parvec)
      dlen     <- length(datavec)
      
      # Case 1: Ranking and Rating
      if(cfg$rating)
      {
            n.th     <- (dlen  - sum(cfg$j.rg))/cfg$n.sdt - 1     # Number of thresholds
            plen.dat <- n.th + 7*length(cfg$j.rg) +  5*cfg$n.sdt  # Expected parameter length given the data vector
            dlen.par <- sum(cfg$j.rg) + cfg$n.sdt*(n.th + 1)      # Expected data length given the parameter vector
      }

      # Case 2: Ranking only
      else
      {
            plen.dat <- 7 * length(cfg$j.rg)
            dlen.par <- sum(cfg$j.rg)
      }
      
	Erg <- matrix(c(plen, dlen, plen.dat, dlen.par), nc = 2)
	rownames(Erg) <- c("Parameters", "Data points")
	colnames(Erg) <- c("Found", "Expected")
	Error <- any(Erg[, 1] != Erg[, 2])

	list(Error = Error, Stat = Erg, Msg = Msg)
}

# ------------------------------------------------------------------------------------------------------------------
# MIX Ranking and Rating Model: Set restrictions
# ------------------------------------------------------------------------------------------------------------------
MIX.Rank.Restriction <- function(cfg, fixed, ident)
{
      # If configuration information is not complete, then complete it
      cfg <- MIX.Rank.Check.Fill.Cfg(cfg)  

      if(cfg$restriction == "standard")
            return(MIX.Rank.Set.Standard.Restriction(cfg, fixed, ident))
      if(cfg$restriction == "SDT")
            return(MIX.Rank.Set.SDT.Restriction(cfg, fixed, ident))
      if(cfg$restriction == "extended")
            return(MIX.Rank.Set.Extended.Restriction(cfg, fixed, ident))
      return(list(fixed = fixed, ident = ident))
}

# ------------------------------------------------------------------------------------------------------------------
# MIX Ranking and Rating Model: Set standard restrictions
# ------------------------------------------------------------------------------------------------------------------
MIX.Rank.Set.Standard.Restriction <- function(cfg, fixed, ident)
{
      n.model <- length(cfg$j.rg)
      
      # ----- Equality constraints on ranking and rating models ------
      # 1. pmix   = pmix     (mixture prob)   (Parameter 1)
      # 2. Mean 2 = Mean 2   (strong mean)    (Parameter 6)
      #    Comment: Not for first SDT model: Here: Mean 2 = Mean 1 = 0
      #                                            pmix   = 1.0
      # --------------------------------------------------------------
      if(n.model > 1)
      {
            source.id <- rep(c(1, 6), n.model - 1)
            target.id <- c(1, 6) + rep(7*(1:(n.model - 1)), each = 2)
      }
      else
      {
            source.id <- target.id <- NULL
      }      
      if(cfg$rating)  # Add constraints for SDT models except for the first one
      {
             source.SDT <- rep(c(1, 6), cfg$n.sdt - 1)
             target.SDT <- c(1, 4) + 7*n.model + rep(5*1:(cfg$n.sdt - 1), each = 2) 
             
             source.id  <- c(source.id, source.SDT)
             target.id  <- c(target.id, target.SDT)     
      }
      if(!is.null(source.id) && !is.null(target.id))
      {
            ident.s <- matrix(c(source.id,
                                target.id), nr = 2, byrow = T)
            pos.unique <-  !(ident.s[2,] %in% ident[2,])
            ident <- cbind(ident.s[, pos.unique], ident)    # Add to ident if not alread specified
      }
      # ---- Fixed constraints for each ranking model -----------
      # 1. Mean0   = 0                         (Parameter 2)
      # 2. Stddev0 = 1                         (Parameter 3)
      # 3. Mean1   = 0                         (Parameter 4)
      # 4. Stddev1 = 1                         (Parameter 5)
      # 5. Stddev2 = 1                         (Parameter 7)
      # ----------------------------------------------------------
      pos.fx <- c(2, 3, 4, 5, 7) + rep(7*(0:(n.model - 1)), each = 5) 
      val.fx <- rep(c(0.0, 1.0, 0.0, 1.0, 1.0), n.model)

      # ---------- Fixed constraints for SDT models --------------
      # I. FIRST MODEL
      #     1. pmix    = 1.0                       (Paramerte 1)
      #     2. Mean0   = 0                         (Parameter 2)
      #     3. Stddev0 = 1                         (Parameter 3)
      #     4. Mean1   = 0                         (Parameter 4)
      #     5. Stddev1 = 1                         (Parameter 5)
      #
      # II. FURTHER MODELS
      #     1. Mean0   = 0                         (Parameter 2)
      #     2. Stddev0 = 1                         (Parameter 3)
      #     3. Stddev1 = 1                         (Parameter 5)
      # ----------------------------------------------------------
      if(cfg$rating)
      {
            # First model
            val.SDT.1 <- c(1.0, 0.0, 1.0, 0.0, 1.0)
            pos.SDT.1 <- 7*n.model + 1:5
            
            # Residual models
            val.SDT <- rep(c(0.0, 1.0, 1.0), cfg$n.sdt - 1)
            pos.SDT <- c(2, 3, 5) + 7*n.model + rep(5*(1:(cfg$n.sdt - 1)), each = 3)
 
            val.fx <- c(val.fx, val.SDT.1, val.SDT)
            pos.fx <- c(pos.fx, pos.SDT.1, pos.SDT)
      }
      fixed.s <- matrix(c(val.fx, pos.fx), nr = 2, byrow = T)
      pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
      fixed <- cbind(fixed.s[, pos.unique], fixed)   # Add standard constraints to fixed if not alread specified

      list(fixed = fixed, ident = ident)
}

# ------------------------------------------------------------------------------------------------------------------
# MIX Ranking and Rating Model: Set unequal variance SDT restrictions
# ------------------------------------------------------------------------------------------------------------------
MIX.Rank.Set.SDT.Restriction <- function(cfg, fixed, ident)
{
      n.model <- length(cfg$j.rg)
      
      # ----- Equality constraints on ranking models ------------------------
      # 1. Mean 2   = Mean 2   (strong mean)                 (Parameter 6)
      # 2. Stddev 2 = Stddev 2 (strong standard deviation)   (Parameter 7)
      # ---------------------------------------------------------------------
      if(n.model > 1)
      {
            source.id <- rep(c(6, 7), n.model - 1)
            target.id <- c(6, 7) + rep(7*(1:(n.model - 1)), each = 2)
      }
      else
      {
            source.id <- target.id <- NULL
      }      
      # --- Equality constraints on SDT models except for the first one ---------
      #   Mean 2  (Rank) = Mean 1  (strong mean)               (Parameter 4, SDT)
      #   Sigma 2 (Rank) = Sigma 1 (strong standard deviation) (Parameter 5, SDT)
      # -------------------------------------------------------------------------
      if(cfg$rating)
      {
             source.SDT <- rep(c(6, 7), cfg$n.sdt - 1)
             target.SDT <- c(4, 5) + 7*n.model + rep(5*1:(cfg$n.sdt - 1), each = 2) 
             
             source.id  <- c(source.id, source.SDT)
             target.id  <- c(target.id, target.SDT)     
      }
      if(!is.null(source.id) && !is.null(target.id))
      {
            ident.s <- matrix(c(source.id,
                                target.id), nr = 2, byrow = T)
            pos.unique <-  !(ident.s[2,] %in% ident[2,])
            ident <- cbind(ident.s[, pos.unique], ident)    # Add to ident if not alread specified
      }
      # ---- Fixed constraints for each ranking model -----------
      # 1. pmix    = 1.0                       (Parameter 1)
      # 2. Mean0   = 0.0                       (Parameter 2)
      # 3. Stddev0 = 1.0                       (Parameter 3)
      # 4. Mean1   = 0.0                       (Parameter 4)
      # 5. Stddev1 = 1.0                       (Parameter 5)
      # ----------------------------------------------------------
      pos.fx <- c(1, 2, 3, 4, 5) + rep(7*(0:(n.model - 1)), each = 5) 
      val.fx <- rep(c(1.0, 0.0, 1.0, 0.0, 1.0), n.model)

      # ---------- Fixed constraints for SDT models --------------
      # I. FIRST MODEL
      #     1. pmix    = 1.0                       (Parameter 1)
      #     2. Mean0   = 0.0                       (Parameter 2)
      #     3. Stddev0 = 1.0                       (Parameter 3)
      #     4. Mean1   = 0.0                       (Parameter 4)
      #     5. Stddev1 = 1.0                       (Parameter 5)
      #
      # II. FURTHER MODELS
      #     1. pmix    = 1.0                       (Parameter 1)
      #     2. Mean0   = 0.0                       (Parameter 2)
      #     3. Stddev0 = 1.0                       (Parameter 3)
      # ----------------------------------------------------------
      if(cfg$rating)
      {
            # First model
            val.SDT.1 <- c(1.0, 0.0, 1.0, 0.0, 1.0)
            pos.SDT.1 <- 7*n.model + 1:5
            
            # Residual models
            val.SDT <- rep(c(1.0, 0.0, 1.0), cfg$n.sdt - 1)
            pos.SDT <- c(1, 2, 3) + 7*n.model + rep(5*(1:(cfg$n.sdt - 1)), each = 3)
 
            val.fx <- c(val.fx, val.SDT.1, val.SDT)
            pos.fx <- c(pos.fx, pos.SDT.1, pos.SDT)
      }
      fixed.s <- matrix(c(val.fx, pos.fx), nr = 2, byrow = T)
      pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
      fixed <- cbind(fixed.s[, pos.unique], fixed)   # Add standard constraints to fixed if not alread specified

      list(fixed = fixed, ident = ident)
}

# ------------------------------------------------------------------------------------------------------------------
# MIX Ranking and Rating Model: Set extended restrictions that enable for mean.1 <> 0
# ------------------------------------------------------------------------------------------------------------------
MIX.Rank.Set.Extended.Restriction <- function(cfg, fixed, ident)
{
      n.model <- length(cfg$j.rg)
      
      # ----- Equality constraints on ranking models -----------------
      # 1. pmix   = pmix     (mixture prob)   (Parameter 1)
      # 2. Mean 1 = Mean 1   (weak mean)      (Parameter 4)
      # 3. Mean 2 = Mean 2   (strong mean)    (Parameter 6)
      # --------------------------------------------------------------
      if(n.model > 1)
      {
            source.id <- rep(c(1, 4, 6), n.model - 1)
            target.id <- c(1, 4, 6) + rep(7*(1:(n.model - 1)), each = 3)
      }
      else
      {
            source.id <- target.id <- NULL
      }      
      # - Equality constraints on rating models except for the first one------------
      #  1. pmix    (Rank) = pmix    (mixture probability)(Parameter 1 -> 1)
      #  2. Mean 1  (Rank) = Mean 0  (weak   mean)        (Parameter 4 -> 2)
      #  3. Mean 2  (Rank) = Mean 1  (strong mean)        (Parameter 6 -> 4)
      # ----------------------------------------------------------------------------
      if(cfg$rating)  # Add constraints for SDT models except for the first one
      {
             source.SDT <- rep(c(1, 4, 6), cfg$n.sdt - 1)
             target.SDT <- c(1, 2, 4) + 7*n.model + rep(5*1:(cfg$n.sdt - 1), each = 3) 
             
             source.id  <- c(source.id, source.SDT)
             target.id  <- c(target.id, target.SDT)     
      }
      if(!is.null(source.id) && !is.null(target.id))
      {
            ident.s <- matrix(c(source.id,
                                target.id), nr = 2, byrow = T)
            pos.unique <-  !(ident.s[2,] %in% ident[2,])
            ident <- cbind(ident.s[, pos.unique], ident)    # Add to ident if not alread specified
      }
      # ---- Fixed constraints for each ranking model -----------
      # 1. Mean0   = 0                         (Parameter 2)
      # 2. Stddev0 = 1                         (Parameter 3)
      # 3. Stddev1 = 1                         (Parameter 5)
      # 4. Stddev2 = 1                         (Parameter 7)
      # ----------------------------------------------------------
      pos.fx <- c(2, 3, 5, 7) + rep(7*(0:(n.model - 1)), each = 4) 
      val.fx <- rep(c(0.0, 1.0, 1.0, 1.0), n.model)

      # ---------- Fixed constraints for SDT models --------------
      # I. FIRST MODEL
      #     1. pmix    = 1.0                       (Paramerte 1)
      #     2. Mean0   = 0                         (Parameter 2)
      #     3. Stddev0 = 1                         (Parameter 3)
      #     4. Mean1   = 0                         (Parameter 4)
      #     5. Stddev1 = 1                         (Parameter 5)
      #
      # II. FURTHER MODELS
      #     1. Stddev0 = 1                         (Parameter 3)
      #     2. Stddev1 = 1                         (Parameter 5)
      # ----------------------------------------------------------
      if(cfg$rating)
      {
            # First model
            val.SDT.1 <- c(1.0, 0.0, 1.0, 0.0, 1.0)
            pos.SDT.1 <- 7*n.model + 1:5
            
            # Residual models
            val.SDT <- rep(c(1.0, 1.0), cfg$n.sdt - 1)
            pos.SDT <- c(3, 5) + 7*n.model + rep(5*(1:(cfg$n.sdt - 1)), each = 2)
 
            val.fx <- c(val.fx, val.SDT.1, val.SDT)
            pos.fx <- c(pos.fx, pos.SDT.1, pos.SDT)
      }
      fixed.s <- matrix(c(val.fx, pos.fx), nr = 2, byrow = T)
      pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
      fixed <- cbind(fixed.s[, pos.unique], fixed)   # Add standard constraints to fixed if not alread specified

      list(fixed = fixed, ident = ident)
}


# ------------------------------------------------------------------------------------------------------------------
# Format free parameters and compute confidence intervals
# Input:	1. px2 Matrix with parameters and standard errors
#		2. Full parameter vector
#           3. List with configuration information
#		4. Matrix of fixed constraints
#		5. Matrix of equality constraints
#		6. Confidence level
#           7. Number of decimal places
#
# Output: px5 Matrix of parameters, standard errors and confidence limits
# ------------------------------------------------------------------------------------------------------------------
MIX.Rank.Format.Par.est <- function(Pars, full.par, cov.matrix, cfg, fixed, ident, alpha, deci)
{
  
      # Number of ranking models
      lrank      <-  length(cfg$j.rg)
  
     # 1. Adjustment of sigma: Replace by absolute values
      pos.sigma <- c(3, 5, 7) + rep(7*(0:(lrank-1)), each = 3)
      if(cfg$rating == T)
            pos.sigma <- c(pos.sigma, 7*lrank + rep(c(3, 5) + 5*(0:(cfg$n.sdt-1)), each = 2))
      full.par[pos.sigma] <- abs(full.par[pos.sigma])            

      full.par <- MIX.Rank.Ppars(full.par, cfg, fixed, ident)
      Pars[, 1] <- Elim.par(full.par, fixed, ident)

      # Check for correct cov.matrix
      if(is.null(cov.matrix))
      {
            NAs       <- rep(NA, dim(Pars)[1])
            Pars[,2]  <- NAs
            Pars <- cbind(Pars, NAs, NAs)
            colnames(Pars) <- c("Value", "SE", paste("CFI-", alpha*100, "(Lower)", sep = ""), paste("CFI-", alpha*100, "(Upper)", sep = ""))
            return(round(Pars, deci))
      }
      # Adjust for standard errors for probability parameter
      Pars[,2] <- MIX.Rank.Delta.Method.Prob(cov.matrix, full.par, cfg, fixed, ident)
      qt.cfi <- qnorm((1-alpha)/2)
      cfi.lo <- Pars[,1] + Pars[,2]*qt.cfi
      cfi.up <- Pars[,1] - Pars[,2]*qt.cfi
            
      # Compute lower bounds (= for probabilities and standard errors of the distributions)            
      bd.lo <- rep(-Inf, length(full.par))
      
      # Indices of parameters with lower bound = 0 (ranking model)
      index.rk   <- c(1, 3, 5, 7) + rep(7*(0:(lrank-1)), each = 4)
      index.rk.p <- 1 + 7*(0:(lrank-1))
      
      # Indices of parameters with lower bound = 0 (rating model)
      lrat       <-  cfg$n.sdt
      index.rt   <- c(1, 3, 5) + 7*lrank + rep(5*(0:(lrat-1)), each = 3)
      index.rt.p <- 1 + 7*lrank + 5*(0:(lrat-1))
      
      index.rk.rt   <- c(index.rk,   index.rt)
      index.rk.rt.p <- c(index.rk.p, index.rt.p)
          
      # Compute lower bounds (= for probabilities and standard errors of the distributions)            
      if(cfg$rating == T)  bd.lo[index.rk.rt] <- 0
      else                 bd.lo[index.rk]    <- 0   
      bd.lo <- Elim.par(bd.lo, fixed, ident)
      cfi.lo <- ifelse(cfi.lo < bd.lo, bd.lo, cfi.lo)

      # Compute higher bounds (= for probability parameters)
      bd.hi <- rep(Inf, length(full.par))
      if(cfg$rating == T)  bd.hi[index.rk.rt.p] <- 1
      else                 bd.hi[index.rk.p]    <- 1
      bd.hi <- Elim.par(bd.hi, fixed, ident)
      cfi.up <- ifelse(cfi.up > bd.hi, bd.hi, cfi.up)
        
      Pars <- cbind(Pars, cfi.lo, cfi.up)
      colnames(Pars) <- c("Value", "SE", paste("CFI-", alpha*100, "(Lower)", sep = ""), paste("CFI-", alpha*100, "(Upper)", sep = ""))
      round(Pars, deci)
}

# ------------------------------------------------------------------------------------------------------------------
# Format full parameters
# ------------------------------------------------------------------------------------------------------------------
MIX.Format.Par.Rank <- function(full.par, cfg, deci, fixed, ident)
{
      Ranking.names <- c("p.mix",
                         "Mean.0",
                         "Stddev.0", 
                         "Mean.1", 
                         "Stddev.1", 
                         "Mean.2", 
                         "Stddev.2")
        
      # 1. Number of ranking models
      lrank  <-  length(cfg$j.rg)
  
      # 1. Adjustment of sigma: Replace by absolute values
      pos.sigma <- c(3, 5, 7) + rep(7*(0:(lrank-1)), each = 3)
      if(cfg$rating == T)
            pos.sigma <- c(pos.sigma, 7*lrank + rep(c(3, 5) + 5*(0:(cfg$n.sdt-1)), each = 2))
      full.par[pos.sigma] <- abs(full.par[pos.sigma])            

      # 2. Compute probabilities
      full.par <- MIX.Rank.Ppars(full.par, cfg, fixed, ident)

      # 3. Extract Ranking parameters: Parameters: 1 - 6*j.rg
      R.pars <- 1:7  +  rep(7*(0:(lrank-1)), each = 7)
      Ranking.pars <- matrix(round(full.par[R.pars], deci), nc = lrank)
      
      cf.matrix <- matrix(c(cfg$k.rg, cfg$j.rg), nr = 2, byrow = T)
      cf.vector <- apply(cf.matrix, 2, fkt <- function(x) { paste(x, collapse = ", j=") })
      colnames(Ranking.pars) <- paste("Rank <k=", cf.vector, ">", sep = "")
     	rownames(Ranking.pars) <- Ranking.names
      if(cfg$rating == F)
            return(list(Ranking = Ranking.pars))

      # 4. Extract rating parameters
      lrat <-  cfg$n.sdt
	th   <- round(full.par[(7*lrank + 5*lrat + 1):length(full.par)], deci)	# Isolate thresholds
	n.t  <- length(th)
	th   <- matrix(rep(th, lrat), nc = lrat)     # Copy thresholds and format them
	
	index.rt    <- 1:5 + 7*lrank + rep(5*(0:(lrat-1)), each = 5)  # Index ot Mixture parameters for rating
	Rating.pars <- matrix(round(full.par[index.rt], deci), nc = lrat)
	Rating.pars <- rbind(Rating.pars, th)

	colnames(Rating.pars) <- paste("Gauss-", 1:lrat, sep = "")
	rownames(Rating.pars) <- c(Ranking.names[1:5], paste("t-", 1:n.t, sep = ""))
	list(Ranking = Ranking.pars, Rating = Rating.pars)
}


# ------------------------------------------------------------------------------------------------------------------
# Computation of observed p values for ranking model
# Input:    datavec = Vector with data
#           cfg     = configuration information
#
# Output:   Vector with probabilities
# ------------------------------------------------------------------------------------------------------------------
MIX.Rank.Compute.Obs.p <- function(datavec, cfg)
{
      p.rank <- NULL
      ed <- 0
      for(i in 1:length(cfg$j.rg))
      {
            st     <- ed + 1
            ed     <- ed + cfg$j.rg[i] 
            p.rank <- c(p.rank, datavec[st:ed]/sum(datavec[st:ed]))
      }
      if(cfg$rating == F) 
            return(p.rank)
      
      sdt.part <- matrix(datavec[-(1:sum(cfg$j.rg))], nr = cfg$n.sdt, byrow = T)     # Extract rating data to matrix with 2 rows
      p.sdt <- apply(sdt.part, 1, function(x) { x/sum(x) })
      c(p.rank, as.vector(p.sdt))
}

# ------------------------------------------------------------------------------------------------------------------
# Computation of observed p values for ranking model
# Input:    datavec = Vector with data
#           cfg     = configuration information
#
# Output:   Vector with probabilities
# ------------------------------------------------------------------------------------------------------------------
MIX.Rank.Compute.Obs.N <- function(datavec, cfg)
{
      # N for ranking models      
      N.rank <- NULL
      ed <- 0
      for(i in 1:length(cfg$j.rg))
      {
            st     <- ed + 1
            ed     <- ed + cfg$j.rg[i]
            ss     <- sum(datavec[st:ed]) 
            N.rank <- c(N.rank, rep(ss, cfg$j.rg[i]))
      }
      if(cfg$rating == F) return(N.rank)

      # N for SDT models      
      sdt.part <- matrix(datavec[-(1:sum(cfg$j.rg))], nr = cfg$n.sdt, byrow = T)     # Extract rating data to matrix with n.sdt rows
      N.rating <- rowSums(sdt.part)
      c(N.rank, rep(N.rating, each = dim(sdt.part)[2]))
}

# ------------------------------------------------------------------------------------------------------------------
# Generation of data labels in output
# Input:    cfg     = configuration information
#           n.dat   = length of the data vector
#
# Output:   Vector with probabilities
# ------------------------------------------------------------------------------------------------------------------
MIX.Rank.Datalabels <- function(cfg, n.dat)
{
      # 1. Labels for ranking data
      label.out <- NULL
      l.rk <- length(cfg$j.rg)
      label.Rk <- paste("Rank", 1:l.rk)
      for(i in 1:l.rk)
            label.out <- c(label.out, paste(label.Rk[i], " [", 1:cfg$j.rg[i], "]", sep = ""))
      if(cfg$rating == F) return(label.out)

      # 2. Labels for rating data      
      n.cat <- (n.dat - sum(cfg$j.rg)) / cfg$n.sdt          # Number of response categories
      label.Rt <- paste(" SDT", 1:cfg$n.sdt)
      for(i in 1:cfg$n.sdt)
            label.out <- c(label.out, paste(label.Rt[i], " [", 1:n.cat, "]", sep = ""))
      return(label.out)
}