# ================================================================
# Auxiliary functions SDT model
# Date of creation:	July, 2009
# Date of last update:	April, 2018
# Author: Siegfried Macho
# ================================================================

# ------------------------------------------------------------------------------------------------------------------
# Elimination of parameters with fixed and equality constraints
# Input:	1. Paramter-vector
#		2. 2 x p Matrix with information about to be fixed parameters
#					1. row: Parameter values
#					2. row: Position within Parameter vector that should be fixed
#		3. 2 x q Matrix with identity restrictions
#					1.row: Source index
#					2.row: Target index (paramters with ehis indices are set to the values of the sources)
# Output:	New parameter vector with fixed and identical parameters removed
# ------------------------------------------------------------------------------------------------------------------
Elim.par <- function(par, fixed, ident)
{
	if(is.null(fixed))
  	{
		if(is.null(ident)) return(par)
		return(par[-ident[2,]])
  	}
	if(is.null(ident)) return(par[-fixed[2,]])
	pos <- c(fixed[2, ], ident[2, ])

	return(par[-pos])
}

# ------------------------------------------------------------------------------------------------------------------
# Elimination of the gradients of fixed parameters and adding the gradients of parameters set equal
# Input:	1. Gradient-vector
#		2. 2 x p Matrix with information about to be fixed parameters
#					1. row: Parameter values
#					2. row: Position within Parameter vector that should be fixed
#		3. 2 x q Matrix with identity restrictions
#					1.row: Source index
#					2.row: Target index (paramters with ehis indices are set to the values of the sources)
# Output:	New gradient vector with gradients of fixed parameters removed and gradients of dentical parameters added
# ------------------------------------------------------------------------------------------------------------------
Elim.par.grad <- function(grad, fixed, ident)
{
	if(is.null(fixed))
  	{
		if(is.null(ident)) return(grad)
		for(i in 1:dim(ident)[2])  grad[ident[1, i]] <- grad[ident[1, i]] + grad[ident[2, i]]	# Copy gradients
		return(grad[-ident[2,]])
  	}
	if(is.null(ident)) return(grad[-fixed[2,]])
	for(i in 1:dim(ident)[2])	grad[ident[1, i]] <- grad[ident[1, i]] + grad[ident[2, i]]		# Copy gradients
	pos <- c(fixed[2, ], ident[2, ])
	return(grad[-pos])
}


# ----------------------------------------------------------------------------------------------------------------
# Positions of parameters not removed within the original parameter vector
# Input:	1. Paramter-vector
#		2. 2 x p Matrix with information about to be fixed parameters
#					1. row: Parameter values
#					2. row: Position within Parameter vector that should be fixed
#		3. 2 x q Matrix with identity restrictions
#					1.row: Source index
#					2.row: Target index (paramters with ehis indices are set to the values of the sources)
#
# Output:	Vector with positions of those parameters that are not subjected to constraints
# Comment:	Positions are used for restauration of parameters in function Restore.par()
# ------------------------------------------------------------------------------------------------------------------
SavePos.par <- function(par, fixed, ident)
{
	pos <- 1:length(par)
	if(is.null(fixed))
  	{
		if(is.null(ident)) return(pos)
		return(pos[-ident[2,]])
  	}
	if(is.null(ident)) return(pos[-fixed[2,]])
	return(pos[-c(fixed[2,], ident[2, ])])
}

# ------------------------------------------------------------------------------------------------------------------
# Restauration of the full parameter vector with fixed and quality constarints
# Input:	1. Paramter vector with constrained parameters removed
#		2. Position vector containing positions of the parameters not removed within the original vector
#		3. 2 x p Matrix with information about to be fixed parameters
#					1. row: Parameter values
#					2. row: Position within Parameter vector that should be fixed
#		4. 2 x q Matrix with identity restrictions
#					1.row: Source index
#					2.row: Target index (paramters with these indices are set to the values of the sources)
#		4. A user defined function for specifiying functional constraints: This function takes the full parameter vector 
#		   and returns a new parameter vector with functional constraints set
# Output:	New parameter vector with fixed and 
# ------------------------------------------------------------------------------------------------------------------
Restore.par <- function(par, pos, fixed, ident, functional)
{
	if(is.null(pos))
		return(par)

	if(is.null(fixed))
	{
		if(is.null(ident)) 
		{
			if(!is.null(functional))			# Call function for specifying functional constraints	
				par.rest <- functional(par)
			else
				return(par)
			return(par.rest)
		}
		len <- length(par) + dim(ident)[2]			# Length of restored vector
		par.rest <- numeric(len)
		par.rest[pos] <- par
		par.rest[ident[2,]] <- par.rest[ident[1,]]

		if(!is.null(functional))				# Call function for specifying functional constraints	
			par.rest <- functional(par.rest)

		return(par.rest)
	}
	if(is.null(ident)) 
	{
		len <- length(par) + dim(fixed)[2]			# Length of restored vector
		par.rest <- numeric(len)
		par.rest[pos] <- par
		par.rest[fixed[2,]] <- fixed[1,]

		if(!is.null(functional))				# Call function for specifying functional constraints	
			par.rest <- functional(par.rest)

		return(par.rest)
	}

	len <- length(par) + dim(fixed)[2] + dim(ident)[2]	# Length of restored vector
	par.rest            <- numeric(len)
	par.rest[pos] 	  <- par
	par.rest[fixed[2,]] <- fixed[1,]
	par.rest[ident[2,]] <- par.rest[ident[1,]]

	if(!is.null(functional))					# Call function for specifying functional constraints	
		par.rest <- functional(par.rest)

	return(par.rest)
}

# ------------------------------------------------------------------------------------------------------------------
# Computation of estimated frequencies 
# Input:	1. Vector of estimated probabilities
#		2. Vector of observed frequencies
#
# Output:	Vector of estimated frequencies
#
# Comment:  Takes marginal sums of probabilities into account 
# ------------------------------------------------------------------------------------------------------------------
Compute.Est.N <- function(p, f)
{
	Tol <- 1e-12

	n <- length(p)
	f.est <- numeric(n)
	k <- p[1]
	j <- 1
	for(i in 2:n)
	{
		if(k >= 1.0 - Tol)				# Probabilities reached a sum of 1.0
		{
			N <- sum(f[j:(i-1)])
			f.est[j:(i-1)] <- N*p[j:(i-1)]
			j <- i
			k <- 0
		}
		k <- k + p[i]
	}	
	N <- sum(f[j:n])
	f.est[j:n] <- N*p[j:n]
	f.est
}

# ------------------------------------------------------------------------------------------------------------------
# Computation of observed probabilites
# Input:	1. Vector of observed frequencies
#		2. Number of signals or a vector with number of responses per signal
#
# Output:	Vector of observed probabilties
# ------------------------------------------------------------------------------------------------------------------
Compute.Obs.p <- function(data, n)
{
      if(length(n) > 1)
            return(Cmp.Obs.p.vec(data, n))
            
      data  <- matrix(data, nc = n)
      N     <- colSums(data)
      p <- apply(data, 1, function(x) { x/N } )
      as.vector(t(p))
}

Cmp.Obs.p.vec <- function(data, n)
{
      k <- 1
      p <- NULL
      for(i in 1:length(n))
      {
            p <- c(p, data[k:(k+n[i]-1)]/sum(data[k:(k+n[i]-1)]))
            k <- k + n[i]
      }
      p
}

# ------------------------------------------------------------------------------------------------------------------
# Computation of observed marginal counts
# Input:	1. Vector of observed frequencies
#		2. Number of signals or a vector with number of responses per signal
#
# Output:	Vector of observed marginal counts
# ------------------------------------------------------------------------------------------------------------------
Compute.Obs.N <- function(data, n)
{
	Tol <- 1e-16
	
      if(length(n) > 1)
          return (Cmp.Obs.N.vec(data, n, Tol))

      data  <- matrix(data, nc = n)
      N     <- colSums(data)
      N[N < Tol] <- Tol
      rep(N, each = dim(data)[1])
}

Cmp.Obs.N.vec <- function(data, n, Tol)
{
      k <- 1
      N <- NULL
      for(i in 1:length(n))
      {
            N.i <- max(sum(data[k:(k+n[i]-1)]), Tol)
            N <- c(N, rep(N.i, n[i]))
            k <- k + n[i]
      }
      N
}


# ------------------------------------------------------------------------------------------------------------------
# Computation of degrees of freedom in data
# Input:	Vector of estimated probabilities
#
# Output:	Number of degrees of freedom in the data
#
# Comment:  Takes marginal sums of probabilities into account 
# ------------------------------------------------------------------------------------------------------------------
Compute.df.data <- function(p)
{
	Tol <- 1e-12

	n <- length(p)
	df.data <- 0
	k <- 0
	for(i in 1:n)
	{
		k <- k + p[i]
		if(k < 1.0 - Tol)				# Probabilities reached a sum of 1.0
			df.data <- df.data + 1
		else		
			k <- 0
	}	
	df.data
}

# ------------------------------------------------------------------------------------------------------------------
# Testing Hessian Matrix
# Input:	Hessian matrix
#
# Output:	A list with the following entries
#		1. A message string, indicating whether matrix is positive definite ("OK") or not
#		2. The condition number
#		3. A flag indicating whether the matrix is positive definete and ill-condtioned
#		4. The eigen values of the matrix
#
# ------------------------------------------------------------------------------------------------------------------
Test.Hessian <- function(Hess)
{
	Tol <- 1e-8
	Tol.cond <- 1e8

	Erg <- list(Msg = "OK", kappa = NA, P.d = F, rank = NA, EV = NULL)

	# 1. Test of present
	if(is.null(Hess) || any(is.na(as.vector(Hess)))) 		# Hessian matrix not available
	{
		Erg$Msg = "Hessian not available"
		return(Erg)
	}
	
	# 2. Compute Eigenvalues
	Erg$EV 	  <- svd(Hess)$d				   # Compute Eigen values
	Erg$rank 	<- sum(Erg$EV >= Tol)		 # Compute Rank
  Erg$P.d	  <- all(Erg$EV > Tol)     # Positive definiteness
  
	# 2. Compute condition number
	if(Erg$rank >= 1)
  {
      Erg$kappa <- kappa(Hess, exact = T)
	    Erg$P.d	  <- Erg$P.d  & (Erg$kappa < Tol.cond)   # Ill-conditioned
	}
	if(Erg$P.d == F)	Erg$Msg <- "Hessian ill-conditioned and/or not positive definite"

	return(Erg)
}

# --------------------------------------------------------------------------------------------------------------------------
# Transformation of equality constraints that are also fixed constraints to fixed constraints
# Input:	1. matrix of fixed constraints
#		3. matrix of equality constraints
#
# Output:	List with two elements: new matrices of fixed and equality constraints
# --------------------------------------------------------------------------------------------------------------------------
Ident.to.fixed <- function(fixed, ident)
{
	if(is.null(fixed) || is.null(ident))	
		return(list(fixed = fixed, ident = ident) )		# Nothing to change --> return

      fixed.tmp <- fixed[, !is.na(fixed[1,]), drop = F]                     # Ignore NA's
      
      if(dim(fixed.tmp)[2] < 1)
      	return(list(fixed = fixed, ident = ident) )		# Nothing to change --> return

	pos.fix.1 <- ident[1, ] %in% fixed.tmp[2, ]			# Boolean vector indicating identical positions
	pos.fix.2 <- ident[2, ] %in% fixed.tmp[2, ]			# Boolean vector indicating identical positions

	if(all(pos.fix.1 == F) && all(pos.fix.2 == F))
		return(list(fixed = fixed, ident = ident) )		# Nothing found     --> return
	
	ident.add <- ident[,  (pos.fix.1 | pos.fix.2), drop = F]
	ident     <- ident[, !(pos.fix.1 | pos.fix.2), drop = F]	# Eliminate fixed from identity matrix

	val.1 <- match(ident.add[1,], fixed[2,])				# Find positions of values belonging to fixed positions		
	val.2 <- match(ident.add[2,], fixed[2,])

	val   <- ifelse(is.na(val.1), val.2, val.1)
	fixed.add <- matrix(c(fixed[1, val], fixed[1, val], ident.add[1, ], ident.add[2, ]), nr = 2, byrow = T)
	
	fixed     <- cbind(fixed, fixed.add)
	fixed     <- fixed[, !duplicated(fixed[2,]), drop = F]			# Take only unique positions

	if(dim(ident)[2] == 0) return( list(fixed = fixed, ident = NULL) )
	list(fixed = fixed, ident = ident)
}

# ---  New Funktion (iterates until no more changes are performed ---
Ident.to.fixed <- function(fixed, ident)
{
	if(is.null(fixed) || is.null(ident))	
		return(list(fixed = fixed, ident = ident) )		      # Nothing to change --> return

      fixed.tmp <- fixed[, !is.na(fixed[1,]), drop = F]                 # Ignore NA's

     if(dim(fixed.tmp)[2] < 1)
      	return(list(fixed = fixed, ident = ident) )		      # No fixed constraints --> nothing to change --> return
      
      while(T)
      {
	     pos.fix.1 <- ident[1, ] %in% fixed.tmp[2, ]			# Boolean vector indicating identical positions
	     pos.fix.2 <- ident[2, ] %in% fixed.tmp[2, ]			# Boolean vector indicating identical positions

	     if(all(pos.fix.1 == F) && all(pos.fix.2 == F))
	           return(list(fixed = fixed, ident = ident) )		# Nothing found --> return
	
	     ident.add <- ident[,  (pos.fix.1 | pos.fix.2), drop = F]
	     ident     <- ident[, !(pos.fix.1 | pos.fix.2), drop = F]	# Eliminate fixed from identity matrix

	     val.1 <- match(ident.add[1,], fixed[2,])				# Find positions of values belonging to fixed positions		
	     val.2 <- match(ident.add[2,], fixed[2,])

	     val   <- ifelse(is.na(val.1), val.2, val.1)
	     fixed.add <- matrix(c(fixed[1, val], fixed[1, val], ident.add[1, ], ident.add[2, ]), nr = 2, byrow = T)
	
	     fixed     <- cbind(fixed, fixed.add)
	     fixed     <- fixed[, !duplicated(fixed[2,]), drop = F]			# Take only unique positions

	     if(dim(ident)[2] == 0) return( list(fixed = fixed, ident = NULL) )
	     fixed.tmp <- fixed[, !is.na(fixed[1,]), drop = F]
      }
}

#-------------------------------------------------------------------------------
# Computation of time elapsed
# Input: 1. time structure 1
#        2. time structure 2
#        3. multiplication factor
#
# Output: Vector with hours, minutes and seconds elapsed
#
# Comment: Time structures are output of proc.time() and the third element
#          contains the second elapsed
#-------------------------------------------------------------------------------
time.estimate <- function(ptm1, ptm2, multi = NULL)
{
      secs <- ptm2[3] - ptm1[3]

      # if multiplication factor is given then estimate the time required
      # otherwise take raw seconds
      if(!is.null(multi))
            est.secs <- secs * multi
      else
            est.secs <- secs
      est.hours <- as.integer(est.secs / 3600)
      est.min   <- as.integer((est.secs %% 3600) / 60)
      est.sec   <- as.integer(est.secs %% 60)

      return(c(est.hours, est.min, est.sec))
}

#-------------------------------------------------------------------------------
# Computation of area under the empirical ROC with trapezoid rule
# Input: FA = vector of new item counts  [ordered from "sure new" to  to "sure old"]
#         H = vector of old item counts  [ordered from "sure new" to  to "sure old"]
#    show.p = Flag indicating whether to show points 
#   reverse = Flag indecating that the order of answers is from "sure old" to "sure new"
#
# Output: Formula of Macmillan & Creelman (2005), p.64
#         ag = 0.5 * Sum[(FA[i+1]-FA[i])*(H[i+1]-H[i]))
#-------------------------------------------------------------------------------
ROC.Ag <- function(FA, H, show.p = F, reverse = F)
{
      # 1. Compute probabilities
      FA <-  FA / sum(FA)
       H <-   H / sum(H)
       
      # 2. Compute (x,y)
      if(!reverse) {
            p.FA <- c(0, cumsum(rev(FA)))
            p.H  <- c(0, cumsum(rev(H)))
      }
      else {
            p.FA <- c(0, cumsum(FA))
            p.H  <- c(0, cumsum(H))
      }
  
      if(show.p) {
          cat("\n p.FA =", p.FA)
          cat("\n p.H  =", p.H,  "\n")
      }
  
      # Compute difference and sum of adjacent elements
      d.FA <- p.FA[-1] - p.FA[-length(p.FA)] 
      s.H  <- p.H[-1] + p.H[-length(p.H)]

      if(show.p) {
          cat("\n d.FA =", d.FA)
          cat("\n s.H  =", s.H,  "\n")
      }
      
      # Compute the area
      return(0.5 * sum(d.FA * s.H))
}      

#--------------------------------------------------------------------------------------
# Computation of area under the empirical ROC, 
# with the method of Donaldson & Good (1996)
# Input: FA = vector of new item counts   [ordered from "sure new" to  to "sure old"]
#         H = vector of old item counts   [ordered from "sure new" to  to "sure old"]
#     probs = Flag indicating whether probabiities of hits and FA are used
#   reverse = Flag indecating that the order of answers is from "sure old" to "sure new"
#      coll = Flag indicating to collapse points in case of non-proper ROC
#
# Output: list with two entries:
#         1. A.r = Area A'r according to Donaldson & Good (1996)
#         2. npt = Number of ROC points used for the computation
#-------------------------------------------------------------------------------------
ROC.Ar <- function(FA, H, probs = F, reverse = F, coll = T)
{
    if(!probs) {
        # 1. Compute probabilities
        FA <-  FA / sum(FA)
        H <-   H / sum(H)
       
        # 2. Compute (x,y) of ROC
        if(!reverse) {
          x <- c(0, cumsum(rev(FA)))
          y <- c(0, cumsum(rev(H)))
        }
        else {
          x <- c(0, cumsum(FA))
          y <- c(0, cumsum(H))
        }  
    }
    else {
        x <- FA
        y <- H
    }
    # 3. If ROC is not proper collpse points
    if(coll){
       Erg.coll <- ROC.collapse(x, y)
       x <- Erg.coll$x
       y <- Erg.coll$y
    }
    if(length(x) <= 2)            # Trivial ROC
      return(list(A.r = 0.5, npt = length(x)-2))
    
    # 3. Compute slope b and Intercept a
    m <- c(diff(y) / diff(x), 0)
    b <- y - m*x

    # 3. Compute points
    x.star <- c(0, 0)
    for(k in 1:(length(b)-1)) 
       x.star <- c(x.star, (b[k+1] - b[k-1])/(m[k-1] - m[k+1]))
    x.star <- c(x.star, 1, 1)

    y.star <- 0
    for(k in 1:(length(b)-2)) 
        y.star <- c(y.star, x.star[k+1] * m[k+1] + b[k+1])
    y.star <- c(y.star, 1, 1, 1)
    
    # 4. Compute area
    A.r <- 0.0
    for(k in 2:(length(y.star)-1)) 
       A.r <- A.r + (y.star[k+1] + y.star[k-1]) * (x.star[k+1] - x.star[k-1])
    return(list(A.r = A.r * 0.25, npt = length(x)-2))
}   

#-------------------------------------------------------------------------------
# Computation of area under the empirical ROC with trapezoid rule
# Input: FA = vector of new item counts [ordered from "sure new" to  to "sure old"]
#         H = vector of old item counts [ordered from "sure new" to  to "sure old"]
#    show.p = Flag indicating whether to show points 
#     probs = Flag indicating whether probabiities of hits and FA are used
#   reverse = Flag indecating that the order of answers is from "sure old" to "sure new"
#
# Output: Flag indicating whether curve is concave (=TRUE) upward
#-------------------------------------------------------------------------------
ROC.concave <- function(FA, H, probs = F, reverse = F)
{
    if(!probs) {
        # 1. Compute probabilities
        FA <-  FA / sum(FA)
        H <-   H / sum(H)
       
        # 2. Compute (x,y) of ROC
        if(!reverse) {
          x <- c(0, cumsum(rev(FA)))
          y <- c(0, cumsum(rev(H)))
        }
        else {
          x <- c(0, cumsum(FA))
          y <- c(0, cumsum(H))
        }  
    }
    else {
        x <- FA
        y <- H
    }
    # 3. Compute slope 
    m <- diff(y) / diff(x)

    return((all(diff(m) <= 0)))
}

#-------------------------------------------------------------------------------
# Collapsing ROC points in case of the ROC not being 
# Input:  x = x-coordinates of ROC
#         y = y-coordinates of ROC
#
# Output: List with two entries
#         x = new vector of x-coordinates
#         y = new vector of y-coordinates
#-------------------------------------------------------------------------------
ROC.collapse <- function(x, y)
{
    # Remove zeros by collapsing
    Nonzero.Flag <- (x > 0 & y > 0)
    x <- c(0, x[Nonzero.Flag])
    y <- c(0, y[Nonzero.Flag])

    # Remove multiple 1    
    Noneins.Flag <- (x < 1 & y < 1 )
    x <- c(x[Noneins.Flag], 1)
    y <- c(y[Noneins.Flag], 1)
#cat("\nx(vor) =", x, "\n")
#cat("\ny(vor) =", y)
    
    # Elimnte duplicated(x)
    ddx <- !duplicated(x)
    ddy <- !duplicated(y)
    x <- x[ddx | ddy]
    y <- y[ddx | ddy]

#cat("\nx(nach) =", x)
#cat("\ny(nach) =", y, "\n")

    while(TRUE) {
        # compute slopes
        m <- diff(y) / diff(x)
        if(length(m) < 2) return(list(x = x, y = y))
        if(all(diff(m) <= 0)) return(list(x = x, y = y))  # ROC proper?
        for(i in 2:length(m)) {
            if(m[i-1] < m[i]) {
               x <- x[-i]
               y <- y[-i]
               break
            } 
        }    
    }
    return(list(x = NULL, y = NULL))
}    

#-------------------------------------------------------------------------------
# Computation of slope and intercept of first principal component of z-ROC
# Input: FA = vector of new item counts  [ordered from "sure new" to  to "sure old"]
#         H = vector of old item counts  [ordered from "sure new" to  to "sure old"]
#   reverse = Flag indecating that the order of answers is from "sure old" to "sure new"
#      deci = number of decima places (if NULL no roundig is provided)
#   correct = correct FRequencies that are zero or N
#
# Output: A list with the following elements
#         slope     = slope     of the first principal axis of the z-ROC
#         intercept = intercept of the first principal axis of the z-ROC $
#         d.2       = d2 of Macmillan & Creelman (2005), 
#         d.1       = d1 of Macmillan & Creelman (2005),  
#         d.a      
#         d.e      
#         d.p       = New measure of Vokey (2016)
#         d.YNp     = New measure of Vokey (2016)
#         A.z       = Area under Gaussian ROC
#         A.zp      = Alternative area measure based on d.YNp
#         explained = Variance explained by the first principal component
#
# comment: Source: Vokey (2016)
#-------------------------------------------------------------------------------
ROC.PCA <- function(FA, H, reverse = F, deci = NULL, correct = T)
{
 
      # 1. Compute sums for normalization
      n.FA <- sum(FA)
      n.H  <- sum(H)
 
      if(n.FA <= 0)
      {
          cat("\n No FA frequencies\n")
          return(NULL)
      }    
      if(n.H <= 0)
      {
          cat("\n No H frequencies\n")
          return(NULL)
      }    

      # 2. Compute ROC points
      FA <-  FA / n.FA
       H <-   H / n.H

      if(!reverse) {
            p.FA <- c(cumsum(rev(FA)))
            p.H  <- c(cumsum(rev(H)))
      }
      else {
            p.FA <- c(cumsum(FA))
            p.H  <- c(cumsum(H))
      }

      # 4. Eliminate ending 1.0
      p.FA <-  p.FA[-length(p.FA)]
      p.H  <-  p.H[-length(p.H)]  

      # 5. Correct probabilties of 0 or 1
      if(correct) {
          p.FA <- ROC.PCA.correct(p.FA, n.FA)
          p.H  <- ROC.PCA.correct(p.H, n.H)
      }
  
      # 6. Inverse z-transformation
      z.FA <- qnorm(p.FA)
      z.H  <- qnorm(p.H)
      
      # 7. Compute covariances and means
      Cov.z <- cov(cbind(z.FA, z.H))

      m.FA  <- mean(z.FA, na.rm = T)
      m.H   <- mean(z.H,  na.rm = T)
      
      # 8. Compute eigen values and eigen vectors
      eigenStructure <- eigen(Cov.z)
      
      # 9. Compute slope and intercept of first principal component
      slope     <- abs(eigenStructure$vectors[1,2] / eigenStructure$vectors[1,1])
      intercept <- m.H - slope * m.FA
      
      # 10. Compute relative efficiency of first principal componente
      explained <- eigenStructure$values[1] /  sum(eigenStructure$values)
      
      # 11. Compute measures
      
      # 11.1 dprime measures
      d.2   <- intercept 
      d.1   <- d.2/slope 
      d.a   <- sqrt(2/(1+slope^2))*d.2
      d.e   <- sqrt(2/(1+slope))*d.2
      d.p   <- m.H - m.FA
      d.YNp <- d.p / sqrt(2)
      
      # 11.2 Area measures
      A.z   <- pnorm(d.a/sqrt(2))
      A.zp  <- pnorm(d.YNp)
      
      # 12. Return list of computed values
      if(is.null(deci))
      return(list(slope     = slope, 
                  intercept = intercept, 
                  d.2       = intercept, 
                  d.1       = d.1, 
                  d.a       = d.a,
                  d.e       = d.e,
                  d.p       = d.p,
                  d.YNp     = d.YNp,
                  A.z       = A.z,
                  A.zp      = A.zp,
                  explained = explained))

      return(list(slope     = round(slope, deci), 
                  intercept = round(intercept, deci), 
                  d.2       = round(intercept, deci), 
                  d.1       = round(d.1, deci), 
                  d.a       = round(d.a, deci),
                  d.e       = round(d.e, deci),
                  d.p       = round(d.p, deci),
                  d.YNp     = round(d.YNp, deci),
                  A.z       = round(A.z, deci),
                  A.zp      = round(A.zp, deci),
                  explained = round(explained,  deci)))
                  
}      

#-------------------------------------------------------------------------------
# Correction of frequencies using Lplace-Rule of succession
# Input: cum.p = vector of cumulative probabilities 
#            n = Number of cases (frequency)
#
# Output: cum.p corrected
#
# Comment: Using Laplace's rule of succession
#  Source: Vokey (2016)
#-------------------------------------------------------------------------------
ROC.PCA.correct <- function(cum.p, n)
{
    Tol <- 1e-14
    
    # Laplace Rule of succession for correcting
    cum.p <- ifelse( cum.p < Tol      ,    1    / (n + 2), cum.p)
    cum.p <- ifelse( cum.p > 1.0 - Tol, (n + 1) / (n + 2), cum.p)

    return(cum.p)
}

#-------------------------------------------------------------------------------
# Computation of Area according to Hanley & McNeil
# Input: FA = vector of new item counts  [ordered from "sure new" to  to "sure old"]
#         H = vector of old item counts  [ordered from "sure new" to  to "sure old"]
#      show = Flag inidcating whether intermediate results should be printed
#   neg.exp = Flag that indicates whether the negative exponential model is assummed 
#             to be correct (this simplifies the computation).
#
# Output: A list with the following elements
#         W  = Area under ROC (trapezoid rule) = Wilcox statistic
#         SE = Standard error
#
# comment: Source: Hanley & McNeil (1982)
#-------------------------------------------------------------------------------
Area.HanleyMcNeil <- function(FA, H, show = F, neg.exp = F)
{    
    n.FA <- sum(FA)
    n.H  <- sum(H)
    
    len.FA <- length(FA)
    len.H  <- length(H)
    
    if(len.FA != len.H) { 
        msg <- paste("\nCRITICAL ERROR: Input vectors of different lenght: (FA =", len.FA, "), (H = ", len.H, ")\n", sep = "")
        stop(msg)
    }
    # Compute lines in Table II of Hanley & McNeil (1982), p. 31
    Z1 <- FA
    Z3 <- H
    Z2 <- n.H - cumsum(Z3)
    Z4 <- c(0, cumsum(FA)[-len.FA])
    Z5 <- Z1 * Z2 + 0.5 * Z1 * Z3
    Z6 <- Z3 * (Z4^2 + Z4 * Z1 + Z1^2/3)
    Z7 <- Z1 * (Z2^2 + Z2 * Z3 + Z3^2/3)
    
    
    W <- sum(Z5) / n.H / n.FA
    
    if(neg.exp) {
          Q1 <- W / (2 - W)
          Q2 <-  2*W*W /(1 + W)
    }      
    else {      
          Q1 <- sum(Z7) / n.FA / n.H^2
          Q2 <- sum(Z6) /n.FA^2 / n.H
    }
    SE <- W * (1 - W) + (n.H - 1) * (Q1 - W^2) + (n.FA - 1) * (Q2 - W^2)
    SE <- sqrt(SE / n.H / n.FA)
    
    if(show) {
      Tab.II <-  rbind(Z1, Z2, Z3, Z4, Z5, Z6, Z7)
      colnames(Tab.II) <- paste("x = ", 1:len.H, sep ="")
      rownames(Tab.II) <- paste("Row-", 1:7, sep = "")
      cat("\n Content of Table II of Hanley & McNeil (1982, p. 31):\n")
      print(Tab.II)
      cat("\n Relevant qunatities [cf. Hanley & McNeil (1982), Table II, on p. 31]:")
      cat("\n W = ", W, "\n Q1 =", Q1, "\n Q2 =", Q2, "\n SE =", SE, "\n") 
    }
    
    list(W = W, SE  = SE)   
} 


#-------------------------------------------------------------------------------
# Computation of the standard error of the area according to Hanley & McNeil
# assuming negative exponential distributions
# Input:  W = Area
#     n.new = number of new item  (normal items)
#     n.new = number of old items (abnormal items)
#
# Output: 
#        SE = Standard error of the area
#
# comment: Source: Hanley & McNeil (1982)
#-------------------------------------------------------------------------------
SE.Area.HanleyMcNeil <- function(W, n.new, n.old)
{    
    Q1 <- W / (2 - W)
    Q2 <-  2*W*W /(1 + W)

    Var <- W * (1 - W) + (n.old - 1) * (Q1 - W^2) + (n.new - 1) * (Q2 - W^2)
    sqrt(Var / n.old / n.new)
}      