# ===========================================================================================
# DPSDT model: Dual process model
#
# Date of creation:	November, 2013
# Date of last update:	November, 2013
#
# Author: Siegfried Macho
# ============================================================================================

DPSDT.Tol <- 1e-12

#===============================================================================
# 1. COMPUTATION OF PROBABILITY PARAMETERS
#===============================================================================

# ------------------------------------------------------------------------------
# Computation of probabilities in parameter vector
# Input:	parvec  = parameter vector
#           cfg     = configuration
#           fixed   = matrix of fixed constraints
#
# Output: 	Parameter vector with raw probabilities replaced by 
#           computed probabilities
#
# Comment:  It is assumed that all equality constraints involving fixed 
#           parameters have been converted to fixed constraints
# ------------------------------------------------------------------------------
DPSDT.Ppars <- function(parvec, cfg, fixed)
{
      n    <- cfg$n.sdt        
      npar <- length(parvec)/n              # Number of parameters per models
      if(is.null(fixed)) fix <- -1
      else               fix <- fixed[2,]
      
      pos.ppars <- c(1 + npar * (0:(n-1)))  # Positions of probability parameters
            
      free <- !(pos.ppars %in% fix)
      free.ppars <- pos.ppars[free]                   # Positions of free probability parameters
      parvec[free.ppars] <- ifelse(parvec[free.ppars] > 700, 700, parvec[free.ppars])     # Avoid overflow
      parvec[free.ppars] <- exp(parvec[free.ppars]) / (1.0 + exp(parvec[free.ppars]))   # Compute probability
      return(parvec)
}

#===============================================================================
# 2. COMPUTATION OF MODEL PROBABILITIES
#===============================================================================
#----------------------------------------------------------------------
# HELPFUNCTION: Computation of Gaussian probabilities for one model
# Input:	parvec = Vector of parameters, consisting of the following entries
#			      1. mean parameter 
#			      2. standard deviation
#			      3. thresholds
#
# Output:   vecotr of SDT probabilities.
#
# Comment: The function is required for computing the Jacobian matrix
#----------------------------------------------------------------------
DPSDT.SDT.Probs <- function(parvec)
{
	 Resp <- pnorm(parvec[3:length(parvec)], mean = parvec[1], sd = max(parvec[2], DPSDT.Tol))
	 Resp <- c(Resp, 1.0)					     # Normal probability response vector
	 c(Resp[1], Resp[-1] - Resp[-length(Resp)])	     # Compute Differences
}				


#----------------------------------------------------------------------
# Computation of DPSDT probabilities for one model
# Input:	parvec = Vector of parameters, consisting of the following entries
#                       1. location of recollection probability
#                       2. recollection probability
#			      3. mean parameter 
#			      4. standard deviation
#			      5. thresholds
#----------------------------------------------------------------------
DPSDT_.Probs <- function(parvec)
{
	 Resp <- pnorm(parvec[5:length(parvec)], mean = parvec[3], sd = max(parvec[4], DPSDT.Tol))
	 Resp <- c(Resp, 1)					                             # Normal probability response vector
	 Resp <- (1.0 - parvec[2]) * c(Resp[1], Resp[-1] - Resp[-length(Resp)])	     # Compute Differences and weight them correctly
	 Resp[parvec[1]] <- Resp[parvec[1]] + parvec[2]                              # Add recollection probability to proper category
	 return(Resp)
}				

# ----------------------------------------------------------------------
# Function for creating response distribution for on Type of Stimulus
# Input:	parvec = Vector of parameters, consisting of the following entries
#			1. mean for each model (except noise model)
#			2. standard deviation for each model (except for noise model)
#			3. thresholds (on set for each model)
#		cfg  	 = configuration list
#           fixed  = 2xp matrix of fixed    constraints
#           ident  = 2xq matrix of equality constraints
#           with.p = Flag indicating wether to compute probabilities
#
# Output: Vector of probabilities for single models
# ----------------------------------------------------------------------
DPSDT.Probs <- function(parvec, cfg, fixed = NULL, ident = NULL, with.p = T)
{
      n.sdt <- cfg$n.sdt

      if(with.p)
            parvec <- DPSDT.Ppars(parvec, cfg, fixed)

	Par  <- matrix(parvec, nc = n.sdt)
      ncat <- dim(Par)[1] - 2                         # Number of response categories
	
      if(is.null(cfg$rec.pos)) 
	     Par <- rbind(c(1, rep(ncat, n.sdt-1)), Par)    # Add position of recollection category
	else     
	     Par <- rbind(cfg$rec.pos, Par)                 # Add position of recollection category

	# Compute probabilities
	as.vector(apply(Par, 2, DPSDT_.Probs))
}				

#------------------------------------------------------------------------------------------------------------------
# Model incorporating restrictions
#------------------------------------------------------------------------------------------------------------------
DPSDT.Probs.Full <- function(parvec, cfg, pos, fixed = NULL, ident = NULL, functional = NULL, with.p = T)
{
      parvec <- Restore.par(parvec, pos, fixed, ident, functional)	# Restore full parameter vector
      DPSDT.Probs(parvec, cfg, fixed, ident, with.p)
}

# ------------------------------------------------------------------------------------------------------------------
# Numerical computation of the model matrix
# ------------------------------------------------------------------------------------------------------------------
DPSDT.Model.Matrix.Num <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)
      parvec <- Elim.par(full.par, fixed, ident)
      J <- jacobian(DPSDT.Probs.Full, x = parvec, cfg = cfg, pos = pos, fixed = fixed, ident = ident, functional = functional)
}


# ------------------------------------------------------------------------------------------------------------------
# Numerical computation of Hessian matrix
# ------------------------------------------------------------------------------------------------------------------
DPSDT.Hessian.p <- function(parvec, datavec, cfg, fixed = fixed, ident = ident, functional = functional)
{
      DPSDT.Model <- function(parvec, cfg, fixed = NULL, ident = NULL) { DPSDT.Probs(parvec, cfg, fixed = NULL, ident = NULL, with.p = F) }

	par.p   <- DPSDT.Ppars(parvec, cfg, fixed) 	# Compute probabilities
	pos.par <- SavePos.par(par.p, fixed, ident)	# Positions of free parameters
	par.p	  <- Elim.par(par.p,  fixed, ident)		# Eliminate constrained parameters
	Model   <- DPSDT.Model				      # Function for computing model probabilities
	hess    <- hessian(func = SDT.logL, x = par.p, data = datavec, Model = Model, n = cfg, pos = pos.par, fixed = fixed, ident = ident, functional = functional)
	if(is.null(hess) || is.na(hess) || is.nan(hess)) return(rep(NA, length(par.p))) 	# Hessian absent or not OK -> retuen (NA)

	Erg <- Test.Hessian(hess)	
	if(Erg$P.d == F)		return(rep(NA, length(par.p)))					# Not positive definite -> return(NA) 		
	Cov <- solve(hess)
	return(sqrt(diag(Cov)))
}


#===============================================================================
# 3. ADMINISTRATIVE FUNCTIONS
#===============================================================================
# ------------------------------------------------------------------------------
# Check and generate configuration structure
# Input:	cfg = Configuration structure: A list with the following entries
#                 - n.sdt        Number of SDT models
#                 - rec.pos      Positions of recollections
#                 - restriction  string identifying the restriction
#
# Output: Configuration list
# ------------------------------------------------------------------------------
DPSDT.Check.Fill.Cfg <- function(cfg, datavec = NULL)
{
      n.sdt <- 2
      restriction = "standard"

      if(is.null(cfg))
      {
            cfg <- list(n.sdt       = n.sdt,
                        restriction = restriction)
            return(cfg)
      }
      if(is.numeric(cfg))
      {
            n.sdt    <- as.integer(cfg)
            cfg      <- list(n.sdt       = n.sdt,
                             restriction = restriction)
            return(cfg)
      }
      cfg.new <- list(n.sdt = ifelse(is.null(cfg$n.sdt), n.sdt, cfg$n.sdt))
      cfg.new$restriction <- DPSDT.Restriction(cfg$restriction)
      if(!is.null(cfg$rec.pos)) cfg.new$rec.pos <- cfg$rec.pos                      # Position of recollection
      
      # Checks
      if(cfg.new$n.sdt < 2)
            stop(paste("\n FATAL ERROR in DPSDT.Rank.Fill.Cfg:\n Number of SDT signals <", cfg.new$n.sdt, "> smaller than <2>\n(Minimum number = 2)\n", sep = ""))
      if(cfg.new$restriction == "standard-pairs" && cfg.new$n.sdt %% 2 > 0)
            stop("\n FATAL ERROR in DPSDT.Rank.Fill.Cfg:\n With restriction = \"standard-pairs\" the number of signals must be even\n")
      if(cfg.new$restriction == "standard-pairs-eq" && cfg.new$n.sdt %% 2 > 0)
            stop("\n FATAL ERROR in DPSDT.Rank.Fill.Cfg:\n With restriction = \"standard-pairs-eq\" the number of signals must be even\n")
      if(cfg.new$restriction == "standard-lure-pairs" && cfg.new$n.sdt %% 2 == 0)
            stop("\n FATAL ERROR in DPSDT.Rank.Fill.Cfg:\n With restriction = \"standard-lure-pairs\" the number of signals must be uneven\n")
      if(cfg.new$restriction == "standard-lure-pairs-eq" && cfg.new$n.sdt %% 2 == 0)
            stop("\n FATAL ERROR in DPSDT.Rank.Fill.Cfg:\n With restriction = \"standard-lure-pairs-eq\" the number of signals must be uneven\n")
      if(!is.null(cfg.new$rec.pos))
      {
            if(length(cfg.new$rec.pos) != cfg.new$n.sdt)
                  stop(paste("\n FATAL ERROR in DPSDT.Rank.Fill.Cfg:\n Number of SDT signals <", cfg.new$n.sdt, "> not equal number of recollection positions <", length(cfg$rec.pos), ">\n", sep = ""))
            if(any(cfg.new$rec.pos < 1))
                  stop(paste("\n FATAL ERROR in DPSDT.Rank.Fill.Cfg:\n Positions in rec.pos < 1\n", sep = ""))
            if(!is.null(datavec))
            {
                  n.cat <- length(datavec)/cfg.new$n.sdt
                  if(any(cfg.new$rec.pos > n.cat))
                        stop(paste("\n FATAL ERROR in DPSDT.Rank.Fill.Cfg:\n Positions in rec.pos > number of categories <", n.cat,">\n", sep = ""))
            }      
      }      
      return(cfg.new)
}

# ------------------------------------------------------------------------------------------------------------------
# Check type of restriction
# Input: A string specfiying the typ of restriction
#        Types of restrictions:
# - "standard"                (Standard restrictions: HT-1 model)
# - "standard-2"              (Double HT model) 
# - "SDT"                     (Simulating SDT model)
# - "SDT-EV"                  (Simulating SDT model with equal)
# - "standard-pairs"          (HT-1 model in pairs [works only with even number of distributions])
# - "standard-pairs-eq"       (HT-1 model in pairs, equal variance within pairs but not between [works only with even number of distributions])
# - "standard-lure-pairs"     (First distribution = N(0, 1); for the rest: HT-1 in pairs [works only with uneven number of distributions])
# - "standard-lure-pairs-eq"  (First distribution = N(0, 1); for the rest: HT-1 in pairs, equal variance within pairs [works only with uneven number of distributions])
# - "no"                      (No restrictions)
#
# Output: String indicating type of restriction
# ------------------------------------------------------------------------------------------------------------------
DPSDT.Restriction <- function(restriction)
{
      if(is.null(restriction))
            return("standard")
      if(charmatch(restriction, c("standard", "STANDARD", "Standard"), nomatch = 0) > 0)
            return("standard")
      if(charmatch(restriction, c("no", "No", "NO"), nomatch = 0) > 0 || restriction == F)
            return("no")
      if(charmatch(restriction, c("standard-2", "STANDARD-2", "Standard-2"), nomatch = 0) > 0)
            return("standard-2")
      if(charmatch(restriction, c("standard-2-eq", "STANDARD-2-EQ", "Standard-2-eq"), nomatch = 0) > 0)
            return("standard-2-eq")
      if(charmatch(restriction, c("sdt", "SDT", "Sdt"), nomatch = 0) > 0)
            return("SDT")
      if(charmatch(restriction, c("sdt-ev", "SDT-EV", "Sdt-ev"), nomatch = 0) > 0)
            return("SDT-EV")
      if(charmatch(restriction, c("standard-pairs", "STANDARD-PAIRS", "Standard-pairs", "Standard-Pairs"), nomatch = 0) > 0)
            return("standard-pairs")
      if(charmatch(restriction, c("standard-pairs-eq", "STANDARD-PAIRS-EQ", "Standard-pairs-eq", "Standard-Pairs-Eq"), nomatch = 0) > 0)
            return("standard-pairs-eq")
      if(charmatch(restriction, c("standard-lure-pairs", "STANDARD-LURE-PAIRS", "Standard-lure-pairs", "Standard-Lure-Pairs"), nomatch = 0) > 0)
            return("standard-lure-pairs")
      if(charmatch(restriction, c("standard-lure-pairs-eq", "STANDARD-LURE-PAIRS-EQ", "Standard-lure-pairs-eq", "Standard-Lure-Pairs-Eq"), nomatch = 0) > 0)
            return("standard-lure-pairs-eq")
      warn.msg <- paste("\n\n DPSDT.Restriction: Unkonwn restriction in configuration list: <", restriction, "> --> setting to: <standard>\n\n", sep = "")      
      warning(warn.msg)
      return("standard")
}            

# ------------------------------------------------------------------------------------------------------------------
# Generate starting parameters due to data and number of signals
# Input:	cfg = List with configuration information
#           data = Vector of data ( in case of a single number, 
#                                   this number is interpreted as inidcating the number of response categories)
#
# Output: String with parameter names
# ------------------------------------------------------------------------------------------------------------------
DPSDT.start.par <- function(cfg, datavec)
{
      n.t <- length(datavec) / cfg$n.sdt - 1   # Number of thresholds
	rep(c(0, 0.1, 1.0, seq(-1.0, 1.0, length = n.t)), cfg$n.sdt)
}

# ------------------------------------------------------------------------------------------------------------------
# Compute upper and lower bounds on parameters
# Input:	 parvec = Parameter vector
#		 cfg    = list with configuration information
#       prob.bounds = flag indicating whether to use probability bounds 
#
# Output: A list with lower and upper bounds on parameters
# ------------------------------------------------------------------------------------------------------------------
DPSDT.par.bounds <- function(parvec, cfg, prob.bounds = F)
{
      if(prob.bounds)
      {
            Max.P <- 1.0
            Min.P <- 0.0
      }
      else
      {
            Max.P <- 700
            Min.P <- -10000
      }
      Min.Var   <- DPSDT.Tol
      
      n.t <- length(parvec) / cfg$n.sdt - 3
      lower.mod <- c(Min.P, -Inf, Min.Var, rep(-Inf, n.t))
      upper.mod <- c(Max.P,  Inf,  Inf, rep( Inf, n.t))

      lower <- rep(lower.mod, cfg$n.sdt)
      upper <- rep(upper.mod, cfg$n.sdt)
      
      list(lower = lower, upper = upper)
}

# ------------------------------------------------------------------------------------------------------------------
# Generate strings with names of the parameters for model
# Input:	parvec = Vector of parameters
#		   cfg = list with configuration information
#
# Output: String with parameter names
# ------------------------------------------------------------------------------------------------------------------
DPSDT.par.names <- function(parvec, cfg)
{
      n.t <- length(parvec) / cfg$n.sdt - 3

      m.rec  <- paste("Recollection[", 1:cfg$n.sdt, "]", sep = "")
      m.mean <- paste("Mean[", 1:cfg$n.sdt, "]", sep = "")
      m.std  <- paste("Stddev[", 1:cfg$n.sdt, "]", sep = "")
      m.t    <- paste("t-", 1:n.t, sep = "")
      
      
	p.name   <- NULL
	for(i in 1:cfg$n.sdt) 
            p.name <- c(p.name, m.rec[i], m.mean[i], m.std[i], paste(m.t, "[", i, "]", sep = ""))
      p.name
}

# ------------------------------------------------------------------------------------------------------------------
# Name of the model
# ------------------------------------------------------------------------------------------------------------------
DPSDT.model.name <- function(cfg)
{
      # If configuration information is not complete, then complete it
      cfg <- DPSDT.Check.Fill.Cfg(cfg)  
      
      mod.name <- paste("Dual Process SDT model:", 
                        " Number of Gaussian models: <", cfg$n.sdt, ">", 
                        " Type of restrictions: <", cfg$restriction, ">", 
                        sep = "")
      if(!is.null(cfg$rec.pos))
      {
            rec.pos.string <- paste(cfg$rec.pos, collapse = ", ")
            mod.name <- paste(mod.name, " Recollection positions: <", rec.pos.string, ">", sep = "")
      }
      return(mod.name)
}       


# ---------------------------------------------------------------------------------------------------------------------
# Check number of parameters and data points
# Input:	1. Parameter vector
#		2. Data vector
#		3. Configuration list
#
# Output:	A list with three entries:
#			1. Error flag: T = Error, F = No error
#			2. 2x2 matrix with actual (first row) and predicted (scond row) number of parameters and data points
# ----------------------------------------------------------------------------------------------------------------------
DPSDT.check <- function(parvec, datavec, cfg)
{
      n.par  <-  length(datavec) + 2*cfg$n.sdt     # Predicted number of parameters due to data
      n.data <-  length(parvec) - 2*cfg$n.sdt   # Predicted number of data points due to parameter

      Erg <- matrix(c(length(parvec), length(datavec), n.par, n.data), nc = 2)
	rownames(Erg) <- c("Parameters", "Data points")
	colnames(Erg) <- c("Found", "Expected")

	Error <- any(Erg[, 1] != Erg[, 2])
	list(Error = Error, Stat = Erg, Msg = "NOTE: Computation of the expected number of parameters is based on the number of data points and vice versa")
}


# ------------------------------------------------------------------------------------------------------------------
# DPSDT Ranking and Rating Model: Set restrictions
# - "standard"                (Standard restrictions: HT-1 model)
# - "standard-2"              (Double HT model)
# - "standard-eq"             (Double HT model with all recollection probabilities being equal) 
# - "SDT"                     (Simulating SDT model)
# - "SDT-EV"                  (Simulating SDT model with equal variances)
# - "standard-pairs"          (HT-1 model in pairs [works only with even number of distributions])
# - "standard-pairs-eq"       (HT-1 model in pairs, equal variance within pairs but not between [works only with even number of distributions])
# - "standard-lure-pairs"     (First distribution = N(0, 1); for the rest: HT-1 in pairs [works only with uneven number of distributions])
# - "standard-lure-pairs-eq"  (First distribution = N(0, 1); for the rest: HT-1 in pairs, equal variance within pairs [works only with uneven number of distributions])
# - "no"                      (No restrictions)
# ------------------------------------------------------------------------------------------------------------------
DPSDT.Set.Restriction <- function(parvec, cfg, fixed, ident, functional)
{
      # If configuration information is not complete, then complete it
      cfg <- DPSDT.Check.Fill.Cfg(cfg)  

      if(cfg$restriction == "standard")
            return(DPSDT.Set.Standard.Restriction(parvec, cfg, fixed, ident, functional, fix.flag = 1))
      if(cfg$restriction == "standard-2")
            return(DPSDT.Set.Standard.Restriction(parvec, cfg, fixed, ident, functional, fix.flag = 2))
      if(cfg$restriction == "standard-2-eq")
            return(DPSDT.Set.Standard.Restriction(parvec, cfg, fixed, ident, functional, fix.flag = 5))
      if(cfg$restriction == "standard-pairs")
            return(DPSDT.Set.Standard.Restriction(parvec, cfg, fixed, ident, functional, fix.flag = 6))
      if(cfg$restriction == "standard-pairs-eq")
            return(DPSDT.Set.Standard.Restriction(parvec, cfg, fixed, ident, functional, fix.flag = 7))
      if(cfg$restriction == "standard-lure-pairs")
            return(DPSDT.Set.Standard.Restriction(parvec, cfg, fixed, ident, functional, fix.flag = 8))
      if(cfg$restriction == "standard-lure-pairs-eq")
            return(DPSDT.Set.Standard.Restriction(parvec, cfg, fixed, ident, functional, fix.flag = 9))
      if(cfg$restriction == "SDT")
            return(DPSDT.Set.Standard.Restriction(parvec, cfg, fixed, ident, functional, fix.flag = 3))
      if(cfg$restriction == "SDT-EV")
            return(DPSDT.Set.Standard.Restriction(parvec, cfg, fixed, ident, functional, fix.flag = 4))
      return(list(fixed = fixed, ident = ident, functional))
}

# ------------------------------------------------------------------------------------------------------------------
# Gaussian: Set standard restrictions:
# The following restrictions are added to the existing restictions:
#
# 1. fixed restrictions:
#           mean[1]   = 0
#           stddev[1] = 1
#
# 2. identity constraints:
#           thresholds of all models are set to be identical
# ------------------------------------------------------------------------------------------------------------------
DPSDT.Set.Standard.Restriction <- function(parvec, cfg, fixed, ident, functional, fix.flag = 1)
{
      if(!is.null(fixed))
            fixed.pos <- fixed[2,]
      else
            fixed.pos <- NULL
      
      if(!is.null(ident))
            ident.pos <- ident[2,]
      else
            ident.pos <- NULL
      
      # Set fixed constraints:           
      if(fix.flag == 1)                                                 # 1-HT: fix mean and p.rec of Model 1
            fixed.s <- matrix(c(0, 0,
                                1, 2), nr = 2, byrow = T)
      else if(fix.flag == 2 || fix.flag == 5)                           # 2-HT (without or with equal rec.p): Fix mean of Model 1 
            fixed.s <- matrix(c(0,
                                2), nr = 2, byrow = T)
      else if(fix.flag == 3)                                            # SDT: Fix mean and all p.rec = 0.0 
      {
            fixed.s <- matrix(c(0, 1,
                                2, 3), nr = 2, byrow = T)
            fixed.s <- cbind(fixed.s, DPSDT.Prec.zero(parvec, cfg))     
      }                                                                 
      else if(fix.flag == 4)                                            # SDT-Equal variance SDT
      {                                    
            fixed.s <- matrix(c(0,
                                2), nr = 2, byrow = T)
            fixed.s <- cbind(fixed.s, DPSDT.Prec.zero(parvec, cfg))     
      }
      else if(fix.flag == 6)                                            # 1-HT in pairs: fix 1.mean and p.rec of Model 1, 3, 5, ...
      {
            fixed.s <- matrix(c(0,
                                2), nr = 2, byrow = T)
            fixed.s <- cbind(fixed.s, DPSDT.Prec.zero.unequal(parvec, cfg))     
      }
      else if(fix.flag == 7)                                            # 1-HT in pairs: fix 1.mean and p.rec of Model 1, 3, 5, ...
      {
            fixed.s <- matrix(c(0,
                                2), nr = 2, byrow = T)
            fixed.s <- cbind(fixed.s, DPSDT.Prec.zero.unequal(parvec, cfg))     
      }
      else if(fix.flag == 8 || fix.flag == 9)                           
      {
            fixed.s <- matrix(c(0, 0,                                   # First distribution = Lure distribution = N(0,1)
                                1, 2), nr = 2, byrow = T)
            fixed.s <- cbind(fixed.s, DPSDT.Prec.zero.equal(parvec, cfg))     
      }
     else  warning("\nDPSDT.Set.Standard.Restriction: Wrong value of fixed.flag: No constraints set!\n") 

      if(fix.flag != 3 && fix.flag != 7 && fix.flag != 9)
            fixed.s <- cbind(fixed.s, DPSDT.Sigma.One(parvec, cfg))     # Add fixed constraints: All sigma = 1.0

      # Set identity constraints
      ident.s <- DPSDT.Thresholds.Identical(parvec, cfg)
      if(fix.flag == 5)                                                 # HT-2 with equal recollection parameters
            ident.s <- cbind(ident.s, DPSDT.Prec.id(parvec, cfg))
      if(fix.flag == 7)
      {
            RS.7 <- DPSDT.Sigma.pairs(parvec, cfg)
            fixed.s <- cbind(fixed.s, RS.7$fixed)
            ident.s <- cbind(ident.s, RS.7$ident)      # Sigma of pairs are equal and for the first one equal 1.0
      }
      if(fix.flag == 9)                                # Sigma of pairs are equal except for the first one.
      {
            RS.9 <- DPSDT.Sigma.pairs.from.two(parvec, cfg)
            fixed.s <- cbind(fixed.s, RS.9$fixed)
            ident.s <- cbind(ident.s, RS.9$ident)      # Sigma of pairs are equal and for the first one equal 1.0
      }
     # Eliminate redundant restricitons
      pos.unique <-  !(fixed.s[2,] %in% fixed.pos)
	fixed <- cbind(fixed.s[, pos.unique], fixed)

      pos.unique <-  !(ident.s[2,] %in% ident.pos)
	ident <- cbind(ident.s[, pos.unique], ident)

      list(fixed = fixed, ident = ident, functional = functional)
}

#-------------------------------------------------------------------------------
# Constraints that thresholds of all models are identical
#-------------------------------------------------------------------------------
DPSDT.Thresholds.Identical <- function(parvec, cfg)
{
      n.t      <- length(parvec) / cfg$n.sdt - 3
      pos.s    <- 4:(3 + n.t)
      source.t <- rep(pos.s, cfg$n.sdt - 1)
      
      target.t <- rep((1:(cfg$n.sdt-1))*(n.t + 3), each = n.t) + rep(pos.s, cfg$n.sdt-1)
      matrix(c(source.t, target.t), nr = 2, byrow = T)
}

#-------------------------------------------------------------------------------
# Fixed constraints: sigma = 1.0, for all models
#-------------------------------------------------------------------------------
DPSDT.Sigma.One <- function(parvec, cfg)
{
      plen         <- length(parvec) / cfg$n.sdt
      pos.sigma    <- 3 + 0:(cfg$n.sdt-1)*plen
      
      matrix(c(rep(1.0, cfg$n.sdt), pos.sigma), nr = 2, byrow = T)
}

#------------------------------------------------------------------------------------
# Fixed constraints: sigma = 1.0, for first two models and identical for other pairs
#------------------------------------------------------------------------------------
DPSDT.Sigma.pairs <- function(parvec, cfg)
{
      plen         <- length(parvec) / cfg$n.sdt
      pos.sigma    <- 3 + 0:(cfg$n.sdt-1)*plen
      
      fixed <- matrix(c(1.0, 1.0, 
                        pos.sigma[1:2]), nr = 2, byrow = T)

      if(length(pos.sigma) < 3)
            return(list(fixed = fixed, ident = NULL))
      pos.sigma <- pos.sigma[-(1:2)]
      pos <- 1:length(pos.sigma)
      ident <- matrix(c(pos.sigma[pos %% 2 >  0],
                        pos.sigma[pos %% 2 <= 0]), nr = 2, byrow = T)
      return(list(fixed = fixed, ident = ident))
}

#------------------------------------------------------------------------------------
# Fixed constraints: for pairs of models except for the first one
#------------------------------------------------------------------------------------
DPSDT.Sigma.pairs.from.two <- function(parvec, cfg)
{
      fixed <- matrix(c(1.0, 3), nr = 2, byrow = T)
      if(cfg$n.sdt < 3) return(list(fixed = fixed, ident = NULL))
      
      plen         <- length(parvec) / cfg$n.sdt
      pos.sigma    <- 3 + 1:(cfg$n.sdt-1)*plen
      
      pos <- 1:length(pos.sigma)
      ident <- matrix(c(pos.sigma[pos %% 2 >  0],
                        pos.sigma[pos %% 2 <= 0]), nr = 2, byrow = T)
      return(list(fixed = fixed, ident = ident))
}


#-------------------------------------------------------------------------------
# Fixed constraints: p.recollection = 0, for all models
#-------------------------------------------------------------------------------
DPSDT.Prec.zero <- function(parvec, cfg)
{
      plen         <- length(parvec) / cfg$n.sdt
      pos.rec    <- 1 + 0:(cfg$n.sdt-1)*plen
      
      matrix(c(rep(0.0, cfg$n.sdt), pos.rec), nr = 2, byrow = T)
}

#-------------------------------------------------------------------------------
# Fixed constraints: p.recollection = 0, for unequal model numbers
#-------------------------------------------------------------------------------
DPSDT.Prec.zero.unequal <- function(parvec, cfg)
{
      plen     <- 2 * length(parvec) / cfg$n.sdt
      pos.rec  <- 1 + 0:(as.integer(cfg$n.sdt/2) - 1)*plen
      
      matrix(c(rep(0.0, length(pos.rec)), pos.rec), nr = 2, byrow = T)
}

#-------------------------------------------------------------------------------
# Fixed constraints: p.recollection = 0, for equal model numbers
#-------------------------------------------------------------------------------
DPSDT.Prec.zero.equal <- function(parvec, cfg)
{
      if(cfg$n.sdt < 3) return(NULL)
      plen    <- length(parvec) / cfg$n.sdt
      pos.rec <- 1 - plen + 1:(as.integer(cfg$n.sdt/2))*2*plen
      
      matrix(c(rep(0.0, length(pos.rec)), pos.rec), nr = 2, byrow = T)
}



#-------------------------------------------------------------------------------
# Identity constraints: p.recollection equal for all models
#-------------------------------------------------------------------------------
DPSDT.Prec.id <- function(parvec, cfg)
{
      plen   <- length(parvec) / cfg$n.sdt
      t.p    <- 1 + 1:(cfg$n.sdt - 1) * plen

      matrix(c(rep(1, cfg$n.sdt - 1), t.p), nr = 2, byrow = T)
}


# ------------------------------------------------------------------------------------------------------------------
# Format parameters for HTSDT.1 model
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Par.DPSDT <- function(par, cfg, fixed, dec)
{
	# Compute probabilities for non-fixed probability parameters
	par <- DPSDT.Ppars(par, cfg, fixed) 	

      n.t <- length(par) / cfg$n.sdt - 3
      Fp  <- matrix(round(par, dec), nc = cfg$n.sdt)
      if(!is.null(cfg$rec.pos))
            rec.pos <- cfg$rec.pos
      else
            rec.pos <- c(1, rep(n.t + 1, cfg$n.sdt-1))
      Fp  <- rbind(rec.pos, Fp) 
      
	colnames(Fp) <- c(paste("DPSDT.", 1:cfg$n.sdt, sep = ""))
	rownames(Fp) <- c("Recollection-pos", "Recollection",  "Mean", "Stddev", paste("t-", 1:n.t, sep = ""))
	Fp
}

# ------------------------------------------------------------------------------------------------------------------
# Format free parameters and adapt standard errors
# Input:	1. px2 Matrix with parameters and standard errors
#		2. full parameter vector
#		3. data vector
#		4. number of signals
#		5. matrix of fixed constraints
#		6. matrix of equality constraints
#		7. function implementing functional constraints
#           8. Confidence level
#           9. Number of decimal places
#
# Output: px3 Matrix of parameters and standard errors
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Par.est.DPSDT <- function(Pars, full.par, datavec, cfg, fixed, ident, functional, alpha, deci)
{

      # Compute positions of probability parameters
      n    <- cfg$n.sdt
      npar <- length(full.par)/n        
      pos.ppars <- c(1 + npar * (0:(n-1)))  # Positions of probability parameters
      
    	pos.free <- SavePos.par(full.par, fixed, ident)		# Positions of free parameters
	free.p   <- pos.free %in% pos.ppars			      # Positions of probability parameters in free parameter vector

 	# Compute probabilities and adapt standard errors
 	if(any(free.p))
 	{
 	     Pars[free.p, 1] <- exp(Pars[free.p, 1])/ (1.0 + exp(Pars[free.p, 1]))
	     Pars[free.p, 2] <- Pars[free.p, 2] * (Pars[free.p, 1]*(1.0 - Pars[free.p, 1]))
      }
	# Compute asymptotic standard errors numerically
	Pars <- cbind(Pars, DPSDT.Hessian.p(full.par, datavec, cfg, fixed = fixed, ident = ident, functional = functional))

      # Compute confidence intervals
      bd    <- DPSDT.par.bounds(full.par, cfg, prob.bounds = T)       # Bounds on probabilities
      bd.lo <- Elim.par(bd$lower, fixed, ident)
      bd.up <- Elim.par(bd$upper, fixed, ident)

      qt.cfi <- qnorm((1-alpha)/2)
      cfi.lo <- Pars[,1] + Pars[,2]*qt.cfi
      cfi.up <- Pars[,1] - Pars[,2]*qt.cfi
            
      # Respect lower bounds on sd parameter
      cfi.lo <- ifelse(cfi.lo < bd.lo, bd.lo, cfi.lo)
      cfi.up <- ifelse(cfi.up > bd.up, bd.up, cfi.up)

      # Add columns with confidence intervals and specify new column names
      Pars <- cbind(Pars, cfi.lo, cfi.up)
	colnames(Pars) <- c("Value", "SE(delta)", "SE(numerical)", paste("CFI-", alpha*100, "(Lower)", sep = ""), paste("CFI-", alpha*100, "(Upper)", sep = ""))

	round(Pars, deci)
}

# ------------------------------------------------------------------------------------------------------------------
# Generation of data labels in output
# Input:    cfg     = configuration information
#           n.dat   = length of the data vector
#
# Output:   Vector with Data names
# ------------------------------------------------------------------------------------------------------------------
DPSDT.Datalabels <- function(cfg, n.dat)
{
      n.cat <- n.dat / cfg$n.sdt          # Number of response categories
      label.name <- paste("Signal", 1:cfg$n.sdt)
      label.out <- NULL
      for(i in 1:cfg$n.sdt)
            label.out <- c(label.out, paste(label.name[i], " [", 1:n.cat, "]", sep = ""))
      return(label.out)
}


#===============================================================================
# 4. JACOBIANS AND GRADIENTS
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix for a single DPSDT model inclding thresholds
# Input:    Parameter vector with the following entries
#                 1. position of recollcetion
#                 2. recollection probability
#                 3. mean
#                 4. stddev
#                 5. thresholds
#           
# Output: Jacobi-Matrix: number of probs x number of parameters
# ------------------------------------------------------------------------------------------------------------------
DPSDT_.Probs.Jacobi <- function(parvec)
{
	l   <- length(parvec)			# Number of parameters
	n.t <- l - 4				# Number of thresholds
	rpos<- parvec[1]                    # recollection position
	p   <- parvec[2]                    # Recollection probability
	m   <- parvec[3]                    # Mean
	sd  <- max(DPSDT.Tol, parvec[4])    # Stddv
	t   <- parvec[5:l]
	z  <- (t-m)/sd
	fn <- (1.0 - p) * dnorm(z)
	
	# Derivative with respect to recollection probability
	dp       <- -DPSDT.SDT.Probs(parvec[3:l])       # Compute Gaussian probabilities
	dp[rpos] <- dp[rpos] + 1.0
	
	# Derivative with respect to mean
	dm <- -fn/sd						# Derivative of Gaussian with respect to m
	dm <- c(dm[1], dm[-1] - dm[-n.t], -dm[n.t])

	# Derivative with respect to sd
	ds <- -z*fn/sd						# Derivative of Gaussian with respect to sd
	ds <- c(ds[1], ds[-1] - ds[-n.t], -ds[n.t])


	# Derivative with respect to thresholds
	dt <- fn/sd							# Derivative of Gaussian with respect to t[i]
	
	J <- matrix(0, nr = n.t+1, nc = n.t)		# Jacobi matrix for thresholds: columns: Parameter, rows = probabilities
	r <- 1
	for(i in 1:n.t)
	{
		J[r, i] <- dt[i]
		r <- r + 1
		J[r, i] <- -dt[i]
	}
	cbind(dp, dm, ds, J)					# Full Jacobi matrix
}
# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix for DPSDT model 
# Input:    1. Parameter vector with the following entries, for each model
#                 1. recollection probability
#                 2. mean
#                 3. stddev
#                 4. thresholds
#           2. Configuration structure
#           3. Matrix with fixed constraints
#           3. Flag indicating whether probabilities should be computed (default = T)
#           
# Output: Jacobi-Matrix: number of probs x number of parameters
# ------------------------------------------------------------------------------------------------------------------
DPSDT.Probs.Jacobi <- function(parvec, cfg, fixed = NULL, with.p = T)
{
      if(with.p) parvec <- DPSDT.Ppars(parvec, cfg, fixed)
      
      n <- cfg$n.sdt
      plen <- length(parvec) / n                      # length of parameter vector for a single model
      ncat <- plen - 2

      Par <- matrix(parvec, nc = n)
      if(is.null(cfg$rec.pos)) 
	     Par <- rbind(c(1, rep(ncat, n-1)), Par)    # Add position of recollection category
	else     
	     Par <- rbind(cfg$rec.pos, Par)             # Add position of recollection category

      # Identify free probability positions
      if(!is.null(fixed))
      {
            ppos <- c(1 + (0:(n-1)) * plen)
            free <- !(ppos %in% fixed[2, ])
      }
      else
            free <- rep(T, n)
      
	# Compute Jacobian matrix
      filler <- filler.r <- filler.c <- matrix(0, nr = ncat, nc = plen)
 	J <- DPSDT_.Probs.Jacobi(Par[, 1])              # Compute Jacobian matrix for first signal distribution
 	if(with.p && free[1]) J[, 1] <- J[, 1] * Par[2, 1] * (1.0 - Par[2, 1]) # Adjust for probabilities
	for(i in 2:n)
	{
	      if(i > 2)
	      {
	           filler.r <- cbind(filler.r, filler)
	           filler.c <- rbind(filler.c, filler)
            }
	      J.m <- DPSDT_.Probs.Jacobi(Par[, i])
 	      if(with.p && free[i]) J.m[, 1] <- J.m[, 1] * Par[2, i] * (1.0 - Par[2, i]) # Adjust for probabilities
	      J <- rbind(cbind(J, filler.c), cbind(filler.r, J.m))
	}
	J
}

# ------------------------------------------------------------------------------------------------------------------
# Gradient of logL
# Input:	Same as for Likelihood function
#
# Output:	Gradient vector
# ------------------------------------------------------------------------------------------------------------------
DPSDT.Grad <- function(par, data, Model, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
	par      <- Restore.par(par, pos, fixed, ident, functional)	# Restore full parameter vector
	gradient <- as.vector(SDT.logL.Jacobi(par, data, Model, n, pos, fixed, ident, functional) %*% DPSDT.Probs.Jacobi(par, n, fixed))

	# Treat functional constraints
	if(!is.null(functional))
	{
		gr.fct <- attr(par, "SDT.gradient")		# Isolate gradient function
		if(is.null(gr.fct))	stop("\n\n DPSDT.Grad: Error in computation of gradient: No gradient for functional constraints found:\n Set Parameter sym.gr = F in function SDT.Estimate !\n\n")
		gradient <- gr.fct(gradient)
	}	
	Elim.par.grad(gradient, fixed, ident)
}

# ------------------------------------------------------------------------------
# Analytical computation of the model matrix 
# Input:    1. Full parameter vector
#           2. Number of signals
#           3. 2 x p - matrix of fixed constraints
#           4. 2 x q - matrix of identity constraints
#           5. Jacobian of functional constraints (supplied by the user)
#
# Output:   Modelmatrix
#
# Comment:  The Model matrix of functional constraints must be a full matrix that
#           is premultiplied by the Jacobian
# ------------------------------------------------------------------------------
DPSDT.Model.Matrix <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)

      J   <- DPSDT.Probs.Jacobi(full.par, cfg, fixed)
      if(!is.null(ident))
      {
            J.Id <- diag(length(full.par))
            Index <- matrix(c(ident[2,], ident[1,]), nc = 2)
            J.Id[Index] <- 1
            J <- J %*% J.Id
      }
      
      # Treatment of funtional constraints
      if(!is.null(functional))
      {
            J.fct <- attr(functional, "SDT.Jacobian")
            if(is.null(J.fct)) 
                  return(NULL)           # No Jacobian function supplied -> return (NULL)
            J <- J %*% J.fct(full.par)
      }
      J[, pos]
}
