# ===========================================================================================
# HT-n model: Double high threshold without a signal detection model component for modeling
#		  pairs of signals.
#		  1. 	Only one set of thresholds possible
#		  2. 	Origin fixed is automatically
#		  3. 	An additional foil distribution (additionally to signal pairs) is possible
#
# Date of creation:	September, 2009
# Date of last update:	March,     2010
#
# Author: Siegfried Macho
# ============================================================================================

HT.n.Tol <- 1e-12

# -----------------------------------------------------------------------------------------
# Help function: Computing the positions of the probability parameters within the vector of 
#                free parameters
# Input:	1. Number of Paramaeter
#		2. Vector with positions of probability parameters
#           3. matrix with fixed constraints
#           4. matrix with identity constraints
#
# Output: 	A p x 2 matrix with 
#		1. column contains the positions of probability parameters in the free vector
#           2. column contains the positions of probability parameters in the full vectot
# -----------------------------------------------------------------------------------------
HT.n.Pos.Ppar.in.free <- function(npar, pos.p, fixed, ident)
{
      free.pos <- Elim.par(1:npar, fixed, ident)
      pos.in.free <- (1:length(free.pos))[free.pos %in% pos.p]    # Positions of free probability parameters within vector of free parameters
      pos.in.full <- pos.p[pos.p %in% free.pos]       # Positions of free probability parameters within full parameter vector
      
      matrix(c(pos.in.free, pos.in.full), nc = 2)
}

# -----------------------------------------------------------------------------------------
# Help function: Computing the positions of the probability parameters
# Input:	1. Parameter vector
#		2. Number of models
#
# Output: 	A list with two entries
#		pos.p    = positions of single parameters
#           pos.p.i  = matrix with positions of multiple parmeters per column
#		pos.pp   = positions of single parameters with respect to probabilities only
#           pos.pp.i = matrix with positions of multiple parmeters per column with respect to probabilties only
# -----------------------------------------------------------------------------------------
HT.n.Pos.Ppar <- function(par, n)
{
	n.p              <- (length(par) + 2 - 4*n)/(n+1) + 2         # Number of probability parameters per model
	n.psig           <- n.p + 2                                   # Number of parameters per signal (except first one)
 	pos.par.single   <- seq(1, n*n.psig, n.psig)                  # Position of single probability parameters within raw parameter vector
      pos.par.multiple <- sapply(pos.par.single, function(x, y) {x+y}, y = 1:(n.p-1))
      
      pos.p.single.p   <- seq(1, n*n.p, n.p)                        # Position of single parameters with respect to probabilities
      pos.p.multiple.p <- sapply(pos.p.single.p, function(x, y) {x+y}, y = 1:(n.p-1))
      list(pos.p = pos.par.single, pos.p.i = pos.par.multiple, pos.pp = pos.p.single.p, pos.pp.i = pos.p.multiple.p)
}

# -----------------------------------------------------------------------------------------
# Function that adds the positions of redundant parameters to fixed constraints
# Input:	1. Parameter vector
#		2. Number of models
#		3. Matrix of fixed constraints
#		4. Matrix of equality constraints
#
# Output: 	Matrix of fixed constraints with position of redundant parameters 
#		added. The value of these positions ist NA
#
# Comments:	The last probability parameter not subjected to fixed or identity constraints
#	    	is considered as redundant
#
#           The function HT.n.Pos.Par is used for finding the positions of probability parameters
# -----------------------------------------------------------------------------------------
HT.n.Redundant.to.Fixed <- function(par, n, fixed, ident)
{
      p.par <- HT.n.Pos.Ppar(par, n)$pos.p.i           # Matrix with positions of probability parameters making up the columns
      
	fixed.add <- NULL
	for(i in 1:n)
	{
            pos.par <- p.par[, i]
		free.p <- pos.par[!(pos.par %in% fixed[2,]) & !(pos.par %in% ident[2,])]
		l.p <- length(free.p)
		if(l.p > 0) fixed.add <- cbind(fixed.add, c(NA, free.p[l.p]))	# Add position of last free probability parameter
	}
	cbind(fixed, fixed.add)
}				

#-------------------------------------------------------------------------------------------------------------------------------------
# Function for computing probability parameters
# Input:	1. Parameter vector
#		2. Number of signals
#		3. Matrix of fixed constraints
#		4. Matrix of identity constraints
#           5. Matrix with positions of parameters of single models in the columns
#
# Ouput:	New parameter vector with:
#		1. Redundant  probability added
#		2. Contraints imposed on parameters
#
# Comment: The function requires that the fixed constraints have been set already (so that redundant parameters are marked as redundant)
#-------------------------------------------------------------------------------------------------------------------------------------
SDT.HT.n.Cmp.P.i <- function(par, n, fixed, ident, pos.p)
{
	# 1. Adjust parameter vector by replacing redundant probabilities by 0
      # Find each of the positions of the NA whether due to fixed or due to equality constraints
     
      pos.NA   <- fixed[2, is.na(fixed[1,])]    # NA in fixed constraints
      ident.NA <- ident[1, ] %in% pos.NA        # NA transfered to other parameters due to identity constraints
      pos.NA <- c(pos.NA, ident[2, ident.NA])

	par[pos.NA] <- 0		                 # Replace parameters marked as redundant by 0
	fixed <- fixed[, !is.na(fixed[1,]), drop = F]	     # Eliminate redundancy constraints from fixed

 	# 2. Set constraints and compute probabilities signal after signal
	for(i in 1:n)
	{
            pos.par <- pos.p[,i]                                  # Positions of probability parameters for the actual model
		source.pos <- ident[1, ident[2,] %in% pos.par]		# Vector with positions of source parameters with respect ot the actual signal
		target.pos <- ident[2, ident[2,] %in% pos.par]		# Vector with positions of target parameters with respect ot the actual signal


		prev.signal<- source.pos < min(pos.par)			# Only those positions that are from previous signals are relevant 
		source.pos <- source.pos[prev.signal]			# (i.e. only those probabilities are transferred.
		target.pos <- target.pos[prev.signal]
	
		par[target.pos] <- par[source.pos]				# Set equality constraints: Transfer probabilities
		
		fixed.p   <- (pos.par %in% fixed[2,]) | (pos.par %in% target.pos)	# Determine fixed positions
		pos.fixed <- pos.par[fixed.p]							# Fixed positions
		pos.free  <- pos.par[!fixed.p]						# Free  positions
		
		if(length(pos.free) > 0)
		{
			resid <- 1 - sum(par[pos.fixed])
			if(resid > 0)
                     par[pos.free] <- resid * exp(par[pos.free]) / sum(exp(par[pos.free]))
                  else
			   par[pos.free] <- 0
		}
		# Now check whether probabilities are computed adequately
		sum.p <- sum(par[pos.par])
		
		if(is.na(sum.p) | is.nan(sum.p))
		{
                  msg <- c(paste("\nFATAL ERROR: Sum for signal <", i, "> = NA \nRelevant Parameters: ", sep = ""),
                           paste(" ", par[pos.par]), 
                           "\nRelevant Positions: ", 
                           paste(" ", pos.par, sep = ""),
                           "\nFree     Positions: ",
                           paste(" ", pos.free, sep = "")
                           ) 
 			 cat("\n\n")
                   stop(msg)
            }
	 	if(sum.p < 1.0 - HT.n.Tol || sum.p > 1.0 + HT.n.Tol)
		{
			msg <- paste("\nFATAL ERROR: Computed probabilities for signal <", i, "> out of range: Sum = ", sum.p, sep = "")
                  cat("\n\n")
			stop(msg)
		}
	}
	par
}

#-------------------------------------------------------------------------------------------------------------------------------------
# Function for computing probability of parameters p (with only one probability-parameter being present)
# Input:	1. Parameter vector
#		2. Number of signals
#		3. Matrix of fixed constraints
#           4. Matrix with positions of parameters of single models in the columns
#
# Ouput:	New parameter vector with non-fixed probability parameters transformed to probabilities
#-------------------------------------------------------------------------------------------------------------------------------------
SDT.HT.n.Cmp.P <- function(par, n, fixed, pos.par)
{
	fixed <- pos.par %in% fixed[2,]
	pos.fixed <- pos.par[fixed]
	pos.free  <- pos.par[!fixed]
	par[pos.free] <- exp(par[pos.free]) /(1 + exp(par[pos.free]))		# Compute probabilities for non-constrained parametes

	par
}

#-------------------------------------------------------------------------------------------------------------------------------------
# Function for computing all probabilities from raw probability parameters p
#-------------------------------------------------------------------------------------------------------------------------------------
SDT.HT.n.Cmp.Ppar <- function(par, n, fixed, ident)
{
      pos.par <- HT.n.Pos.Ppar(par, n)
	p.par <- SDT.HT.n.Cmp.P.i(par, n, fixed, ident, pos.par$pos.p.i)
	SDT.HT.n.Cmp.P(p.par, n, fixed, pos.par$pos.p)
}

#-------------------------------------------------------------------------------------------------------------------------------------
# Help function: It computes a Boolean matrix with columns indicating fixed parameters,
# i.e. either fixed or due to an identy constraint with a probability from a previous signal
#
# Input:    1. Matrix with positions of probability paraemters for single models in the columns (computed by the function  HT.n.Pos.Ppar())
#           2. Matrix of fixed constraints
#           3. Matrix of identity constraints
#
# Output:   A Boolean matrix of the same dimension as that in the first argument with TRUE indicating fixed positions
#-------------------------------------------------------------------------------------------------------------------------------------
SDT.HT.n.is.fixed.p <- function(pos.p.i, fixed, ident)
{
      # Helpfunction that computes Boolean vector for a single model
      SDT.HT.n.is.fixed.p.model <- function(pos.p, fixed.pos, ident)
      {
      	source.pos <- ident[1, ident[2,] %in% pos.p]	# Vector with positions of source parameters with respect ot the actual signal
		target.pos <- ident[2, ident[2,] %in% pos.p]	# Vector with positions of target parameters with respect ot the actual signal

            prev.signal<- source.pos < min(pos.p)		# Only those positions that are from previous signals are relevant 
		target.pos <- target.pos[prev.signal]
	      (pos.p %in% target.pos) | (pos.p %in% fixed.pos)
      }
      
      fixed.pos <- fixed[2, !is.na(fixed[1,])]
      apply(pos.p.i, 2, SDT.HT.n.is.fixed.p.model, fixed.pos = fixed.pos, ident = ident)

}

#-------------------------------------------------------------------------------------------------------------------------------------
# Help function finding within signal equality constraints
# Input:    1. Positions of probabilities for the signal with respect to the orginal parameter vector
#           2. Positions of probabilities for the signal with respect to the probability parameter vector (ignoring other constraints)
#           3. Matrix of identity constraints
#           4. Matrix of fixed constraints 
#
# Output:   An identity matrix containing equality constraints with respect to the probability parameters of the actual signal
#
# Comment:  If the matrix of fixed constraints is present (!=NULL), the identity constraint with NA as sources is excluded
#-------------------------------------------------------------------------------------------------------------------------------------
SDT.HT.n.is.ident.within <- function(pos.p, pos.pp, ident, fixed)
{
      if(!is.null(fixed))
      {
            NA.boolean <- ident[1,] %in% fixed[2, is.na(fixed[1,])]
            source.m   <- (ident[1, ] %in% pos.p)  & !NA.boolean          
      }
      else
            source.m <- ident[1, ] %in% pos.p
      if(any(source.m))
      {
            ident <- ident[, source.m, drop = F]
            target.m <- ident[2, ] %in% pos.p
            
            if(any(target.m))
            {
                  diff <- pos.p[1] - pos.pp[1]                    # Compute differences in positions in order to adapt
                  return(ident[, target.m, drop = F] - diff)
            }
      }     
      return(NULL)
}

#-------------------------------------------------------------------------------------------------------------------------------------
# Help function finding between signal equality constraints
# Input:    1. Positions of probabilities for the signal with respect to the orginal parameter vector
#           2. Vector of positions of multiple probabilities with respect to whole parameter vector
#           3. Vector of positions of multiple probabilities with respect to the probability parameter vector
#           4. Matrix of identity constraints
#
# Output:   An identity matrix with equality constraints with respect to the probability parameters of the actual signal
#-------------------------------------------------------------------------------------------------------------------------------------
SDT.HT.n.is.ident.between <- function(pos.p, pos.pp, pos.p.i, pos.pp.i, ident)
{
      target.m <- ident[2, ] %in% pos.p
      if(any(target.m))
      {
            ident <- ident[, target.m, drop = F]
            source.m <- ident[1,] < min(pos.p)                    # Source of identity constraint has to be located in a previous signal
            
            if(any(source.m))
            {
                  dff.s <- NULL
                  for(x in ident[1, source.m])
                  {
                        pos.s <- pos.p.i %in% x
                        dff.s <- c(dff.s, pos.p.i[pos.s] - pos.pp.i[pos.s])   # Compute differences in source positions in order to compute positions with repsect to probability vector
                  }      
                  dff.t <- pos.p[1] - pos.pp[1]                               # Compute differences in target positions in order to compute positions with repsect to probability vector
                  ident <- matrix(c(ident[1, source.m] - dff.s, ident[2, source.m] - dff.t), nr = 2, byrow = T)
                  return(ident)
            }
      }     
      return(NULL)
}

#-------------------------------------------------------------------------------------------------------------------------------------
# Function for computing Jacobian of the transformation of raw to probability parameters
# Input:	1. par   =	Full parameter vector
#		2. n     =	Number of models
#		3. fixed =	matrix of fixed constraints
#		4. ident =	matrix of identity constraints
#           5. Test  =  A test flag that enables the stand alone test of the function
#                       if Test = T then the probability parameters are computed from raw paremters otherwise the
#                       parameter vector with probability parameters already computed is passed to the function 
#
# Ouput:	Jacobian matrix of the transformation
#-------------------------------------------------------------------------------------------------------------------------------------
SDT.HT.n.Cmp.Ppar.Jacobi <- function(par, n, fixed, ident, Test = T)
{
      # 1. Determine positions of probability parameters
      pos.par  <- HT.n.Pos.Ppar(par, n)
      pos.p    <- pos.par$pos.p                      # pos.p = positions of single parameters
      pos.p.i  <- pos.par$pos.p.i                    # pos.p.i = matrix with positions of multiple parameters per column
      pos.pp   <- pos.par$pos.pp                     # pos.pp = positions of single parameters with respect to probabilites
      pos.pp.i <- pos.par$pos.pp.i                   # pos.pp.i = matrix with positions of multiple parameters per column with respect to probabilities

      n.p.model <- dim(pos.pp.i)[1] + 1              # Number of probability parameters per model
      
      # 2. Compute the transformation of raw to probability paremeters, if test flag = T
	if(Test)
	{
            p.par <- SDT.HT.n.Cmp.P(par, n, fixed, pos.p)	               # Compute probabilities for single probability parameters p[n]      
            p.par <- SDT.HT.n.Cmp.P.i(p.par, n, fixed, ident, pos.p.i)     # Compute probabilities for multiple probability parameters p-i[n]
      }
      else
            p.par <- par

      # 3. Generate Jacobian
	n.p.s  <- dim(pos.p.i)[1] + 1             # Number of probability parameters per signal
	n.p    <- n.p.s * n                       # Number of probability parameters
      J      <- matrix(0, n.p, n.p)             # Jacobi-matrix J
      
      # 4. Initialize Information list
      Info.id <- list(NULL)
      n.Info  <- 0
      
      pp.par <- p.par[as.vector(rbind(pos.p, pos.p.i))]    # Cut parameters so that only probabilities remain
      
      # 5. Compute Jacobian for single probabilities
      free.pp <- pos.pp[!(pos.p %in% fixed[2,]) & !(pos.p %in% ident[2,])]            # Boolean vector indicating positions of free single paraemters
      J[matrix(c(free.pp, free.pp), nc = 2)] <- pp.par[free.pp]*(1-pp.par[free.pp])   # Adjust Jacobian

      # 6. Compute fixed positions with respect to probability vector in order to set diagonal elements of J.b and J.w
      pos.p.vec  <- as.vector(rbind(pos.p, pos.p.i))
      free.p.v <- !((pos.p.vec %in% fixed[2,]) | (pos.p.vec %in% ident[2,]))
      J.b <- diag(n.p)
      J.w <- diag(1*free.p.v)                            # Jacobian for between and within signal identity constraints
      
      # 7. Adapt J.b for equality constraints on single parameters
      ident.p.single <- pos.p %in% ident[2,]
      if(any(ident.p.single))
      {

            single.row <- pos.pp[ident.p.single]
            single.col <- NULL
            source.id  <- ident[1, ident[2,] %in% pos.p]
            for(x in source.id)
                  single.col <- c(single.col, pos.pp[pos.p %in% x])
            J.b[matrix(c(single.row, single.col), nc = 2)] <- 1
      }
      
      # 8. Compute Jacobian for multiple probabilities
      fixed.pp   <- SDT.HT.n.is.fixed.p(pos.p.i, fixed, ident)     # Matrix indicating fixed and free positions for each of the signals
      for(i in 1:n)
	{
            fixed.s <- fixed.pp[, i]
            free.s  <- !fixed.s                        # Boolean vectors indicating free and fixed positions of the model
            
            resid <- 0                                 # Reset resid to zero (required for adapting J.b 
            if(any(free.s))
            {
                  free.pos.s  <- pos.pp.i[free.s, i]   # Positions of free parameters for actual model with respect to 
                  pp          <- pp.par[free.pos.s]    # Free parameters
                  
                  # Treatment of free probabilities with repsect to free parameters 
                  # free parameters = not fixed and not subjected to identitiy constraints from previous signal
                  resid <- sum(pp)
                  pp <- pp/resid                       
                  P <- -pp %*% t(pp)                   # -p.i*p.j/resid
                  diag(P) <- diag(P) + pp              # Diagonal: p.i*(1-p.i/res)
                  J[free.pos.s, free.pos.s] <- P*resid

                  # Set within signal equality constraints by coping the relevant columns in J.p
                  ident.s.w <- SDT.HT.n.is.ident.within(pos.p.i[free.s,i], free.pos.s, ident, fixed)
                  if(!is.null(ident.s.w))   # Adapt J.w, by setting relevant entries to 1
                  {
                        one.ident <- matrix(c(ident.s.w[2,], ident.s.w[1,]), nc = 2)
                        J.w[one.ident] <- 1
                  }
            }
            # Now adapt Jacobian due to identity constraints between models
            # Generate the matrix of identity constraints between signals with respect to probability parameters for the actual signal as target
            ident.s.b <- SDT.HT.n.is.ident.between(pos.p.i[, i], pos.pp.i[, i], as.vector(pos.p.i), as.vector(pos.pp.i), ident)
            if(!is.null(ident.s.b))
            {
                  one.ident <- matrix(c(ident.s.b[2,], ident.s.b[1,]), nc = 2)
                  J.b[one.ident] <- J.b[one.ident] + 1
                  if(resid > 0)
                  {
                        for(x in ident.s.b[1,])
                              J.b[free.pos.s, x] <- J.b[free.pos.s, x] - pp
                  }
            }
 
      }
      # 9. Implement between and within signal equality constraints
      J.b %*% J %*% J.w
 }

#-------------------------------------------------------------------------------------------------
# Function for creating response distribution for on Type of Stimulus
# Input:	Vector of parameters with the following entries:
#		1. m1 		= Location of the Gaussian distribution
#		2. sd 		= Scale of distributions
#		3. p			= Probability of invoking the discrete distribution
#		4. p[1]...p[n]	= Probability parameters characterizbning the discrete distribution
#		5. t[1]...t[n-1]  = Decision bounds
#
# Output: 	Vector of length R, with R being the number of Response categories
#		per signal
#-------------------------------------------------------------------------------------------------
HT.n.Probs <- function(par)
{
	n.p 	<- length(par)								# Number of parameters
	n.t	<- (n.p - 4) / 2								# Number of thresholds
	p	<- par[4:(4 + n.t)]							# Extract probability parameters
	t 	<- par[(5 + n.t):n.p]							# Extract thresholds

	# Compute Probabilities for the left signal
	Resp 	<- pnorm(t, mean = par[1], sd = par[2])	                  # Signal 1: Probability of Response-Category 1 to n-1
	Resp 	<- c(Resp, 1)								# Response vector
	R 	<- (1-par[3]) * c(Resp[1], Resp[-1]-Resp[-length(Resp)])	# Compute differences
	
	R + par[3]*p									# Add discrete probabilities

}				

# ------------------------------------------------------------------------------------------------------------------
# Computation of probabilities for high threshold model with rating
# Input:	par = Vector of parameters, consisting of the following entries:
#
#		I. For each signals the following parameters are given, except for the first one, where no location and scale
#		   parameter is specified.
#
#		1. m			= Location of Gaussian distribution (for the first signals no location parameter is specified because m  = 0)
#		2. sd	  		= Scale    of Gaussian distribution (for the first signals no scale    parameter is specified because sd = 1)
#		3. p			= Mixing probability: The probability of using the discrete distribution
#		4. p[1]...p[n]	= Probability parameters characterizing the discrete distribution
#
#		II. One set of threshold parameters (identical for each model) is provided:
#		    t[1], t[2], ...., t[n-1] = Decisions bounds
#
#		fixed = Matrix of fixed constraints
#		ident = Matrix of identity constraints (used for compatability only)
#
# Output: Vector of probabilities for each pair of distributions plus the for the foil (if present)
#
# Comment:
# The function treats raw probability parameters as well as fixed parameters that are constraint between [0, 1]
# Only parameters not fixed are computed from raw parameters
# ------------------------------------------------------------------------------------------------------------------
SDT.HT.n <- function(par, n, fixed = NULL, ident = NULL)
{
      par <- SDT.HT.n.Cmp.Ppar(par, n, fixed, ident)			# Compute probabilities from probability parameters

	par <- c(0, 1, par)			 				# Add location and scale
	n.par <- length(par) 							# Number of parameters
	n.t   <- (n.par - 4*n)/(n+1)						# Number of threshold parameters
	n.sig <-  n.par - n.t							# Number of parameters for the signals, i.e. without thresholds
	t     <-  par[(n.sig+1):n.par]					# Isolate threshold parameters

	par.m   <- matrix(par[1:n.sig], nr = n, byrow = T)		# Transform to matrix
	par.m <- cbind(par.m, matrix(rep(t, n), nr = n, byrow = T))	# Add threshold parameters
	
	as.vector(apply(par.m, 1, HT.n.Probs))
}

# -----------------------------------------------------
# Version without computation of probability parameters
# -----------------------------------------------------
SDT.HT.n.without <- function(par, n, fixed = NULL, ident = NULL)
{
	par <- c(0, 1, par)			 				# Add location and scale
	n.par <- length(par) 							# Number of parameters
	n.t   <- (n.par - 4*n)/(n+1)						# Number of threshold parameters
	n.sig <-  n.par - n.t							# Number of parameters for the signals, i.e. without thresholds
	t     <-  par[(n.sig+1):n.par]					# Isolate threshold parameters

	par.m   <- matrix(par[1:n.sig], nr = n, byrow = T)		# Transform to matrix
	par.m <- cbind(par.m, matrix(rep(t, n), nr = n, byrow = T))	# Add threshold parameters
	
	as.vector(apply(par.m, 1, HT.n.Probs))
}

# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix of the derivatives for a single signal
# Input: Parametervector containing the following entries
#			1. mean
#			2. sd
#			3. p
#			4. p-1, ..., p-n
#			5. thresholds (n-1 threshold parameters)
#
# Output: n x (2*n + 2) Matrix (number of data points x number of parameters)
# ------------------------------------------------------------------------------------------------------------------
HT.n.Probs.Jacobi <- function(par)
{
	p.l   <- length(par)			# Number of parameters
	n.t <- (p.l - 4)/2			# Number of thresholds
	
	m   <- par[1]
	sd  <- par[2]
	p   <- par[3] 
	p.i <- par[4:(4+n.t)]
	t   <- par[(5+n.t):p.l]

	z   <- (t - m)/sd
	fn  <- dnorm(z)
	
	# Derivative with respect to m
	dm  <- -fn/sd							# Derivative of Gaussians with respect to m
	dm  <- (1-p)  * c(dm[1], dm[-1] - dm[-n.t], -dm[n.t]) 	

	# Derivative with respect to sd
	ds  <- -z * fn/sd							# Derivative of Gaussians with respect to sd
	ds  <-  (1-p) * c(ds[1], ds[-1] - ds[-n.t], -ds[n.t]) 	

	# Derivative with respect to probability parameter p
	dp <- -pnorm(t, m, sd)
	dp <- c(dp[1] , dp[-1] - dp[-n.t], -dp[n.t]-1)
	dp <- dp + p.i

	# Derivative with respect to probability parameters p-i
	dp.i <- rep(p, n.t+1)
	Dp.i <- diag(dp.i)

	# Derivative with respect to thresholds
	dt  <- (1-p) * fn/sd						# Derivative of Gaussian with respect to t[i]
	J  <- matrix(0, nr = n.t+1, nc = n.t)			# Jacobi matrix for thresholds: columns: Parameter, rows = probabilities for first signal
	r <- 1
	for(i in 1:n.t)
	{
		J[r, i]  <- dt[i]
		r <- r + 1
		J[r, i]  <- -dt[i]
	}

	cbind(dm, ds, dp, Dp.i, J)			# Full Jacobian matrix
}

# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix of the main HT.n model
# ------------------------------------------------------------------------------------------------------------------
SDT.HT.n.Jacobi <- function(par, n, fixed, ident)
{
      pos.p   <- HT.n.Pos.Ppar(par, n)                            # Determine positions of probability parameters within full parameter vector
      pos.p.v <- as.vector(rbind(pos.p$pos.p, pos.p$pos.p.i)) + 2 # Vector of positions (Add 2 due to mean and sd for model 1)

      # Compute the transformation of raw to probability paremeters
	par.p <- SDT.HT.n.Cmp.P(par, n, fixed, pos.p$pos.p)	           # Compute probabilities for single probability parameters p[n]      
      par.p <- SDT.HT.n.Cmp.P.i(par.p, n, fixed, ident, pos.p$pos.p.i) # Compute probabilities for multiple probability parameters p-i[n]
      
        
      par.full <- c(0, 1, par.p)                            # Add mean and sd to first model
	t <- par.full[(max(pos.p.v)+1):length(par.full)]	# Extract threshold parameters
	n.t <- length(t)					      # Number of thresholds
	n.r <- n.t + 1					      # Length of a data block (=number of rows per signal)
      n.c <- n.t + 4                                  # Number of parameters per signal (=number od columns per signal) without threshold parameters

	fill.c <- fill.r <- fill <- matrix(0, nr = n.r, nc = n.c)	# Filler-Matrix
	for(s in 1:n)								# Generation of the blocks for the single models
	{
		par.act <- c(par.full[((s-1)*n.c+1):(s*n.c)], t)	# Parameter vector for actual signal
		J.act <- HT.n.Probs.Jacobi(par.act)
		if(s == 1)
		{
			J   <- J.act[, 1:n.c]
			J.t <- J.act[, (n.c+1):(n.c+n.t)]
		}
		else
		{
			J   <- rbind(cbind(J, fill.c), cbind(fill.r, J.act[, 1:n.c]))
			J.t <- rbind(J.t, J.act[, (n.c+1):(n.c+n.t)])
			if(s < n)
			{
					fill.c <- rbind(fill.c, fill)
					fill.r <- cbind(fill.r, fill)
			}
		}
	}
	J <- cbind(J, J.t)							    # Bind thresholds
	
	# Now treat probabilities with constraints
 	J.p <- SDT.HT.n.Cmp.Ppar.Jacobi(par.p, n, fixed, ident, Test = F) # Jacobian of probabilitity transformations   
	J[, pos.p.v] <- J[, pos.p.v] %*% J.p
	J[, -(1:2)]
}

# ------------------------------------------------------------------------------------------------------------------
# Gradient of logL for the HT.n model
# Input:	Same as for Likelihood function
#
# Output:	Gradient
#
# Comment:  Due to the fact that identity constraints on probability parameters have been treated previously
#           identity constraints on probability parameters are transformed into fixed constraints. As a result
#           the relevant derivatives are eliminated only
# ------------------------------------------------------------------------------------------------------------------
SDT.HT.n.Grad <- function(par, data, Model, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
	par <- Restore.par(par, pos, fixed, ident, functional)	# Restore full parameter vector

	gradient <- as.vector(SDT.logL.Jacobi(par, data, Model, n, pos, fixed, ident, functional) %*% SDT.HT.n.Jacobi(par, n, fixed, ident)) 

	# Treat functional constraints
	if(!is.null(functional))
	{
		gr.fct <- attr(par, "SDT.gradient")		# Isolate gradient function
		if(is.null(gr.fct))	stop("\n\n SDT.HT.n.Grad: Error in computation of gradient: No gradient for functional constraints found:\n Set Parameter sym.gr = F in function SDT.Estimate !\n\n")
		gradient <- gr.fct(gradient)
	}
      # Transform identity constraints on probabilities to fixed constraints
      pos.p   <- HT.n.Pos.Ppar(par, n)                            # Determine positions of probability parameters within full parameter vector
      pos.p.v <- as.vector(rbind(pos.p$pos.p, pos.p$pos.p.i))     # Vector of positions of multiple probability parameters
      trans <- ident[2,] %in% pos.p.v
      fixed.add <- matrix(c(rep(0, sum(trans)), ident[2, trans]), nr = 2, byrow = T)      
      fixed <- cbind(fixed, fixed.add)                            # Add to fixed constraints
      ident <- ident[,!trans, drop = F]                           # Eliminate from identity constraints

      if(!is.null(ident))
      {
            if(dim(ident)[2] == 0)
                  ident <- NULL
      }      
      Elim.par.grad(gradient, fixed, ident)
}

# ------------------------------------------------------------------------------------------------------------------
# Generate starting parameters due to data and number of signals
# Input:	   n = number of signals
#           data = Either a vector with data or a number indicating the number of categories per signal
#           range.th = a range for the threshold parameters
#
# Output: String with start parameters
# ------------------------------------------------------------------------------------------------------------------
SDT.HT.n.start.par <- function(n, data, range.th = c(-1, 1))
{
      n.data <- length(data)
      if(n.data == 1) 
            n.data <- data*n
            
	n.t <- n.data/n - 1		  # Number of threshold parameters
	n.p <- n.t + 1			  # Number of probability parameters
	c(rep(c(0, 1, 0, rep(0, n.p)), n), seq(range.th[1], range.th[2], length = n.t))[-c(1,2)]	
}
# ------------------------------------------------------------------------------------------------------------------
# Compute upper and lower bounds on parameters
# Input:	         par   = parameter vector
#		           n   = number of signals
#           probs.bounds   = flag indicating whether bounds concern probabilties or raw parameters
#
# Output: A list with lower and upper bounds on parameters
# ------------------------------------------------------------------------------------------------------------------
HT.n.par.bounds  <- function(par, n, probs.bounds = F)
{
      Max.Raw.P <- 300
      Min.Var   <- HT.n.Tol

      lower <- rep(-Inf, length(par))
      upper <- rep(Inf, length(par))
      
      Pos.P <- HT.n.Pos.Ppar(par, n)                              # Positions of probability parameters
      pos.p <- as.vector(rbind(Pos.P$pos.p, Pos.P$pos.p.i))       # Positions as vector

      if(probs.bounds)
      {
            lower[pos.p] <- 0                         # Minimal and maximal possible value of singel probability parameters
            upper[pos.p] <- 1
      }
      else
      {
            lower[pos.p] <- -Max.Raw.P                # Minimal and maximal possible value of singel probability parameters
            upper[pos.p] <-  Max.Raw.P
      } 
       
      pos.p <- Pos.P$pos.p 
      lower[pos.p[-1]-1] <- Min.Var                   # Minimal value of the standard error of the Gaussian distributions
      
      list(lower = lower, upper = upper)
}

# ------------------------------------------------------------------------------------------------------------------
# Generate strings with names of the parameters for model
# Input:	par = Vector of parameters
#		  n = number of signal pairs
#
# Output: String with parameter names
# ------------------------------------------------------------------------------------------------------------------
HT.n.par.names <- function(par, n)
{
	n.t <- (length(par) + 2 - 4*n)/(n+1)		# Number of threshold parameters
	n.p <- n.t + 1						# Number of probability parameters

	m   <- paste("mean["   , 1:n, "]", sep = "")
	sd  <- paste("stddev[" , 1:n, "]", sep = "")
	p   <- paste("p[" , 1:n, "]", sep = "")
	p.i <- paste("p-" , 1:n.p, sep = "")
	
	p.name   <- NULL
	for(i in 1:n) p.name <- c(p.name, m[i], sd[i], p[i], paste(p.i, "[", i, "]", sep = ""))

	c(p.name, paste("t-", 1:n.t, sep = ""))[-c(1,2)]
}

# ------------------------------------------------------------------------------------------------------------------
# Name of the model
# ------------------------------------------------------------------------------------------------------------------
HT.n.model.name <- "HT.n: Double high threshold model for rating data"

# ------------------------------------------------------------------------------------------------------------------
# Format parameters for HT.n model
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Par.HT.n <- function(par, n, fixed, ident, dec)
{
	# Compute probabilities
	par <- SDT.HT.n.Cmp.Ppar(par, n, fixed, ident)			# Compute probabilities from raw parameters
	
	par <- c(0, 1, par)			 				# Add location and scale parameter

	n.t   <- (length(par) - 4*n)/(n+1)					# Number of threshold parameters
	n.p   <- n.t + 1								# Number of probability parameters
	n.sig <- n.p + 3
	
	par <- round(par, dec)
	Fp <- matrix(par[1:(n.sig*n)], nc = n)
	t  <- matrix(rep(par[(n.sig*n+1):length(par)], n), nc = n)
	Fp <- rbind(Fp, t)

	colnames(Fp) <- c(paste("HT.n-", 1:n, sep = ""))
	rownames(Fp) <- c("mean", "stddev",  "p", paste("p-", 1:n.p, sep = ""), paste("t-", 1:n.t, sep = ""))
	
	Fp
}

# ------------------------------------------------------------------------------------------------------------------
# Format free parameters and adapt standard errors
# Input:	1. px2 Matrix with parameters and standard errors
#		2. full parameter vector
#		3. Covariance matrix of estimates 
#		4. Number of signals
#		5. matrix of fixed constraints
#		6. matrix of equality constraints
#		7. function implementing functional constraints
#           8. confidecne level
#           9. number of decimal places
#
# Output: px3 Matrix of parameters and standard errors
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Par.est.HT.n <- function(Pars, full.par, Cov, n, fixed, ident, functional, alpha, deci)
{
	
	# Compute probabilities
	par.p <- SDT.HT.n.Cmp.Ppar(full.par, n, fixed, ident)	# Compute probabilities from raw parameters
      Pars[,1] <- Elim.par(par.p,  fixed, ident)		# Eliminate constrained parameters

	# Compute asymptotic standard errors numerically
	Pars <- cbind(Pars, SDT.HT.n.Delta.num(full.par, Cov, n, fixed = fixed, ident = ident))

      # Compute confidence intervals
      bd    <- HT.n.par.bounds(full.par, n, probs.bounds = T)       # Bounds on probabilities
      bd.lo <- Elim.par(bd$lower, fixed, ident)
      bd.up <- Elim.par(bd$upper, fixed, ident)

      qt.cfi <- qnorm((1-alpha)/2)
      cfi.lo <- Pars[,1] + Pars[,3]*qt.cfi
      cfi.up <- Pars[,1] - Pars[,3]*qt.cfi
            
      # Respect lower bounds on sd parameter
      cfi.lo <- ifelse(cfi.lo < bd.lo, bd.lo, cfi.lo)
      cfi.up <- ifelse(cfi.up > bd.up, bd.up, cfi.up)
      
  	# Add columns with confidence intervals and specify new column names
      Pars <- cbind(Pars, cfi.lo, cfi.up)
	colnames(Pars) <- c("Value", "SE(raw parameters)", "SE(delta numeric)", paste("CFI-", alpha*100, "(Lower)", sep = ""), paste("CFI-", alpha*100, "(Upper)", sep = ""))

	round(Pars, deci)
}

# ---------------------------------------------------------------------------------------------------------------------
# Check number of parameters and data points
# Input:	1. Parameter vector
#		2. Data vector
#		3. Number of signals
#
# Output:	A list with three entries:
#			1. Error flag: T = Error, F = No error
#			2. 2x2 matrix with actual (first row) and predicted (scond row) number of parameters and data points
# ----------------------------------------------------------------------------------------------------------------------
SDT.HT.n.check <- function(par, data, n)
{
	n.t    <- (length(par) + 2 - 4*n)/(n+1)		# Number of threshold parameters
	n.resp <- length(data)/n 				# Number of Response categories
	
	n.par  <- (3 + n.resp)*n + n.resp - 3		# Predicted number of parameters due to data
	n.data <- (n.t+1)*n					# Predicted number of data points 

	Erg <- matrix(c(length(par), length(data), n.par, n.data), nc = 2)
	rownames(Erg) <- c("Parameters", "Data points")
	colnames(Erg) <- c("Found", "Expected")

	Error <- any(Erg[, 1] != Erg[, 2])
	list(Error = Error, Stat = Erg, Msg = "NOTE: Computation of the expected number of parameters is based on the to number of data points and vice versa")
}

# ---------------------------------------------------------------------------------------------------------------------
# Checks that none of the rating parameter enters an identity constraint with another parameter
# Input:	1. Parameter vector
#		2. Number of signals
#		3. matrix with identity constraints
#
# Output:	No (relevant) return value
# ----------------------------------------------------------------------------------------------------------------------
SDT.HT.n.check.identity.constraints <- function(par, n, ident)
{
      if(!is.null(ident))
      {    
            # 1. Find positions of probability parameters
            Pos.P <- HT.n.Pos.Ppar(par, n)                      # List with the positions of probability parameters
            pos.p.i <- as.vector(Pos.P$pos.p.i)                 # Positions of Rating probability Parameters
            pos.p   <- Pos.P$pos.p                              # Positions of recollection probability parameters
            par.names <- HT.n.par.names(par, n)                 # Generate Names of Parameters
             
            # 2. Check rating probability parameters
            p.as.source <- ident[1, ident[2,] %in% pos.p.i]
            p.as.target <- ident[2, ident[1,] %in% pos.p.i]
            
            test.source <- p.as.source %in% pos.p.i
            if(!all(test.source))
            {
                  msg <- c("\nFATAL ERROR (HT.n): Equality constraint between rating parameters and other types of parameters:",
                           "\n Source Parameter:", paste(" ", par.names[p.as.source[!test.source]], sep = ""), "\n")
                  cat("\n")
                  stop(msg)
            }
           
            test.target <- p.as.target %in% pos.p.i
            if(!all(test.target))
            {
                  msg <- c("\nFATAL ERROR (HT.n): Equality constraint between rating parameters and other types of parameters:",
                           "\n Target Parameter:", paste(" ", par.names[p.as.target[!test.target]], sep = ""), "\n")
                  cat("\n")
                  stop(msg)
            }
            
            # 3. Check recollection probability parameters
            p.as.source <- ident[1, ident[2,] %in% pos.p]
            p.as.target <- ident[2, ident[1,] %in% pos.p]
            
            test.source <- p.as.source %in% pos.p
            if(!all(test.source))
            {
                  msg <- c("\nFATAL ERROR (HT.n): Equality constraint between recollection parameters and other types of parameters:",
                           "\n Source Parameter:", paste(" ", par.names[p.as.source[!test.source]], sep = ""), "\n")
                  cat("\n")
                  stop(msg)
            }
           
            test.target <- p.as.target %in% pos.p
            if(!all(test.target))
            {
                  msg <- c("\nFATAL ERROR (HT.n): Equality constraint between recollection parameters and other types of parameters:",
                           "\n Target Parameter:", paste(" ", par.names[p.as.target[!test.target]], sep = ""), "\n")
                  cat("\n")
                  stop(msg)
            }
      }
}

# ---------------------------------------------------------------------------------------------------------------------
# Classification of constraints on multiple probability parameters of the model
# Input:	1. Parameter vector
#		2. Number of signals
#		3. Matrix with fixed    constraints
#		4. Matrix with identity constraints
#
# Output:	Matrix with columns representing the probability parameter of the single signals
#           The entries in the matrix have the following meaning:
#           0: free
#           1: fixed by the user
#           2: fixed due to renundancy
#          -s: equality constraint from signal s (as source)
# ----------------------------------------------------------------------------------------------------------------------
SDT.HT.n.Classify.constraints.p <- function(par, n, fixed, ident)
{
       pos.p.i <- HT.n.Pos.Ppar(par, n)$pos.p.i             # Matrix with the positons of multiple probability parameters in columns
       Class.p <- matrix(0, nr = dim(pos.p.i)[1], nc = n)   # Matrix containing the classification of parameters
       
       for(s in 1:n)                # Work through signals
       {
            pos.s <- pos.p.i[, s]
            if(!is.null(fixed))     # Check fixed constraints
            {
                   fx.1             <- pos.s %in% fixed[2,!is.na(fixed[1,])]   # Constraints set by the user
                   fx.2             <- pos.s %in% fixed[2,is.na(fixed[1,])]    # Constrait due to redundancy
                   Class.p[fx.1, s] <- 1 
                   Class.p[fx.2, s] <- 2
            }
            
            if(!is.null(ident))     # Check fixed constraints
            {
                  id.t  <- pos.s %in% ident[2,]
                  if(any(id.t))     # Equality constraint present (with parameter of the signal as target)
                  {
                        id.s     <- ident[2,] %in% pos.s
                        id.s.pos <- ident[1, id.s]                # Positions of source parameters
                        source.s <- NULL
                        for(x in id.s.pos)
                              for(i in 1:s)
                              {
                                    if(x %in% pos.p.i[,i])
                                          source.s <- c(source.s, -i)
                              }
                        Class.p[id.t, s] <- source.s
                  }
            }
       }
      Class.p
}

# ---------------------------------------------------------------------------------------------------------------------
# Checks for the presence of complex identity constraints on multiple probability parameters
# Input:	1. Parameter vector
#		2. Number of signals
#		3. Matrix with fixed    constraints
#		4. Matrix with identity constraints
#
# Output:	If conditions are violated an error msg is returned, otherwise a NULL is returned
# ----------------------------------------------------------------------------------------------------------------------
SDT.HT.n.check.complex.identity.constraints <- function(par, n, fixed, ident)
{
      # With less than three signals no violation of complex constraints is possible
      if(n < 3) 
            return(NULL)
      Class.p <- SDT.HT.n.Classify.constraints.p(par, n, fixed, ident)
      for(s in 3:n)
      {
            c.p  <- Class.p[, s]
            crit <- c.p < 0 & abs(c.p) < s & abs(c.p) > 1       # Specification of critical cases
            if(any(crit))
            {
                  c.p.rel <- c.p[crit]
                  for(x in abs(c.p.rel))
                  {
                        c.p  <- Class.p[,x]
                        crit <- c.p < 0 & abs(c.p) < x
                        if(any(crit))
                        {
                              msg <- c("\n\nFATAL ERROR: Not allowed equality constraints between signals specified:",
                                       "\n   Number of Signal (target): ", paste(s),
                                       "\n   Number of Signal (source): ", paste(x),
                                       "\nThe source of the constraint is located within a model some of whose parameters",
                                       "\nare itself subjected to equality constraints.",
                                       "\nIn order to fit the model with this type of constraint, set: sym.gr = F",
                                       "\nin the estimation procedure SDT.Estimate().\n\n"
                                       )
                              return(msg)
                        }
                  } 
                  
            }
      }
      return(NULL)
}
# ------------------------------------------------------------------------------------------------------------------
# Functions for computing the Hessian Matrix using the delta rule:
# The Jacobian of the transformation from raw to probability parameters is computed numerically
# Input:	1. Full parameter vector
#		2. Covariance matrix of estimates
#           3. Number of models
#           4. matrix with fixed constraints
#           5. matrix with identity constraints
#
# Output: 	New vector of standard errors of estimates
# ------------------------------------------------------------------------------------------------------------------
SDT.HT.n.Delta.num <- function(full.par, Cov, n, fixed, ident)
{
      # Helpfunction: This function is called by the function jacobian: It sets parameter restrictions
      SDT.HT.n.Cmp.Ppar.jacoby.num <- function(par, n, fixed, ident)
      {
            pos.par <- SavePos.par(par, fixed, ident)					  # Save positions of eliminated parameters	
            par 	  <- Elim.par(par, fixed, ident)
     	      par     <- Restore.par(par, pos.par, fixed, ident, NULL)      

            SDT.HT.n.Cmp.Ppar(par, n, fixed, ident)  
      }

      # No Covariance of estimates present -> return NA
      if(is.null(Cov))
      {
            return(rep(NA, length(Elim.par(full.par, fixed, ident))))
      }                            
            
      full.par[is.na(full.par)] <- 0                                                      # Replace possible NAs by 0
      J <- jacobian(SDT.HT.n.Cmp.Ppar.jacoby.num, x = full.par, n = n, fixed = fixed, ident = ident)   # Now compute Jacobian

      # Compute positions of free probability parameters
      Pos.P <- HT.n.Pos.Ppar(full.par, n)            # Positions of Probability Parameters
      pos.p <- as.vector(rbind(Pos.P$pos.p, Pos.P$pos.p.i))       # Transpose to vector
      
      # Compute Positions of free probability parameters within the original vector and the vector of 
      Pos.p.free <- HT.n.Pos.Ppar.in.free(length(full.par), pos.p, fixed, ident)    
      
      # Cut Jacobian and Covariance matrix
      J  <- J[Pos.p.free[,2], Pos.p.free[,2]]
      Cv <- Cov[Pos.p.free[,1], Pos.p.free[,1]]
      
      # Compute standard errors of probability parameter by means of the delta method.
      sd.p <- sqrt(diag(J %*% Cv %*% t(J)))
      
      # Replace within the vector of standard errors the standard errors of raw parameters with those of probability parameters
      sd <- sqrt(diag(Cov))
      sd[Pos.p.free[,1]] <- sd.p
      sd
}



#===============================================================================
#
# Functions for comupting the model matrix (analytically and numerically)
# for the HT.n model
#
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Model incorporating restrictions
# The parameter passed to the function contains the free parameters only
# ------------------------------------------------------------------------------------------------------------------
SDT.HT.n.Full <- function(rest.par, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
      full.par <- Restore.par(rest.par, pos, fixed, ident, functional)	# Restore full parameter vector
      SDT.HT.n(full.par, n, fixed, ident)
}

# ------------------------------------------------------------------------------------------------------------------
# Numerical computation of the model matrix 
# ------------------------------------------------------------------------------------------------------------------
SDT.HT.n.Model.Matrix.Num <- function(full.par, n, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)
      par <- Elim.par(full.par, fixed, ident)
      J <- jacobian(SDT.HT.n.Full, par, n = n, pos = pos, fixed = fixed, ident = ident, functional = functional)
}

# ------------------------------------------------------------------------------
# Analytical computation of the model matrix 
# Input:    1. Full parameter vector
#           2. Number of signals
#           3. 2 x p - matrix of fixed constraints
#           4. 2 x q - matrix of identity constraints
#           5. Jacobian of functional constraints (supplied by the user)
#
# Output:   Modelmatrix
#
# Comment:  The Modelmatrix of funtional constraints must be a full matrix that
#           is premultiplied by the Jacobian
# ------------------------------------------------------------------------------
SDT.HT.n.Model.Matrix <- function(full.par, n, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)
      J   <- SDT.HT.n.Jacobi(full.par, n = n, fixed, ident)
      
      if(!is.null(ident))
      {
            J.Id <- diag(length(full.par))
            # ignore constraints for probability parameters
            Pos.P <- HT.n.Pos.Ppar(full.par, n)                 # List with the positions of probability parameters
            pos.p <- c(Pos.P$pos.p, as.vector(Pos.P$pos.p.i))   # Convert to vector of positions
             
            # 2. Check rating probability parameters
            pos.p.ident <- unique(c(which(!(ident[2,] %in% pos.p)), which(!(ident[1,] %in% pos.p))))
            ident.res <- ident[, pos.p.ident, drop = F]
            if(!is.null(ident.res))
            {            
                  Index <- matrix(c(ident.res[2,], ident.res[1,]), nc = 2)
                  J.Id[Index] <- 1
                  J <- J %*% J.Id
            }
      }
      
      # Treatment of funtional constraints
      if(!is.null(functional))
      {
            J.fct <- attr(functional, "SDT.Jacobian")
            if(is.null(J.fct)) 
                  return(NULL)           # No Jacobian function supplied -> return (NULL)
            J <- J %*% J.fct(full.par)
      }
      J[, pos]
}
