# ===========================================================================================
# HTSDT-1 model: Double high threshold signal detection model for modeling a pair of signals.
#		  1. 	An additional probability for modeling the frequency of the lowest (first
#			signal of the pair) and	highest response category (second second signal
#			of the pair), respectively
#		  2. 	Only one set of thresholds
#		  3. 	Origin fixed automatically
#		  4. 	Additional foil distribution is possible
#
# Date of creation:	August, 2009
# Date of last update:	March,  2010
#
# Author: Siegfried Macho
# ============================================================================================

HTSDT.1.Tol <- 1e-12
# ----------------------------------------------------------------------
# Function for creating response distribution for on Type of Stimulus
# Input:	Vector of parameters with the following entries:
#		1. m.left 		= Location of distribution for left signal
#		2. sd 		= Scale of distributions
#		3. m.right		= Location of distribution for right signal
#		4. p.left		= Probability of recollection for first distribution
#		5. p.right 		= Probability of recollection for first distribution
#		6. t[1]....t[n-1] = Decision bounds
#
# Output: 	Vector of length 2*R, with R being the number of Response categories
#		per signal
# ----------------------------------------------------------------------
HTSDT.1.Probs <- function(par)
{
	n.p 	<- length(par)								# Number of parameters
	t 	<- par[6:n.p]								# Extract thresholds
	n.R 	<- n.p - 4									# Number of response categories

	# 1. Compute Probabilities for the left signal
	Resp 	<- pnorm(t, mean = par[1], sd = max(par[2], HTSDT.1.Tol))	# Signal 1: Probability of Response-Category 1 to n-1
	Resp 	<- c(Resp, 1)								# Response vector
	R.L 	<- (1-par[4]) * c(Resp[1], Resp[-1]-Resp[-length(Resp)])	# Compute differences
	R.L[1]<- R.L[1] + par[4]							# Add discrete probability

	# 2. Compute Probabilities for the 'right' signal
	Resp 	  <- pnorm(t, mean = par[3], sd = max(par[2], HTSDT.1.Tol))	# Signal 2: Probability of Response-Category 1 to n-1
	Resp 	  <- c(Resp, 1)								# Response vector
	R.R 	  <- (1-par[5]) * c(Resp[1], Resp[-1]-Resp[-length(Resp)])	# Compute differences
	R.R[n.R]<- R.R[n.R] + par[5]							# Add discrete probability

	c(R.L, R.R)
}				

#----------------------------------------
# Probabilities of the foil distribution
#----------------------------------------
HTSDT.1.Foil.Probs <- function(par)
{
	# Compute Probabilities for the foil distribution	
	Resp <- pnorm(par[3:length(par)], mean = par[1], sd = max(par[2], HTSDT.1.Tol))	# Probability of Response-Category 1 to n-1
	Resp <- c(Resp, 1)											# Response vector
	c(Resp[1], Resp[-1]-Resp[-length(Resp)])								# Compute Differences
}				


# ------------------------------------------------------------------------------------------------------------------
# Computation of flag, indicating for which signals the probability parameter is free
# Input:  	1. Number of parameter
#		2. Number of signal pairs
#		3. Matrix of fixed parameters
#		4. Offset = 1. Position of relevant parameter
#
# Output: vector of Flags: 	T = parameter is not fixed for this pair of signals
#					F = parameter is fixed
# ------------------------------------------------------------------------------------------------------------------
HTSDT.1.free.p <- function(n.p, n, fixed, off)
{
	if(is.null(fixed)) return(rep(TRUE, n))

	pos.p   <- seq(0, n.p-1, 5) + off			# Relevant positions
	!(pos.p %in% fixed[2,])					# Boolean vector indicating fixed probability parameters
	
}

# ------------------------------------------------------------------------------------------------------------------
# Computation of probabilities for not fixed probability parameters
# Output: vector of Flags: 	T = parameter is not fixed for this signal
#					F = parameter is fixed
# ------------------------------------------------------------------------------------------------------------------
HTSDT.1.cmp.p <- function(par, n, fixed)
{
	Uneven <- n%%2 != 0				# Boolean variable indicating whether number of models is uneven				

	if(Uneven)	
	{
		m <- (n-1)/2			# Uneven number of signals -> first data points represent foil distribution
		off.1 <- 4				# Offset for identifying positions of left probabilities
		off.2 <- 5				# Offset for identifying positions of left probabilities
	}
	else
	{	
		m <- n/2					# Number of pairs
		par <- c(0, 1, par)			# Insert mean and standard deviation
		off.1 <- 2					# Offset for identifying positions of left probabilities
		off.2 <- 3					# Offset for identifying positions of left probabilities
	}
	n.p <- 5*m						# Number of parameters without thresholds
	t   <- par[(n.p + 1) : length(par)]		# Extract threshold parameters


	free.p1 <- HTSDT.1.free.p(n.p, m, fixed, off = off.1)	# Left  probabilities
	free.p2 <- HTSDT.1.free.p(n.p, m, fixed, off = off.2)	# Right probabilities

	par.m   <- matrix(par[1:n.p], nr = m, byrow = T)					# Transform to matrix
	par.m[free.p1, 4] <- exp(par.m[free.p1, 4]) / (1 + exp(par.m[free.p1, 4])) 	# Compute probabilities
	par.m[free.p2, 5] <- exp(par.m[free.p2, 5]) / (1 + exp(par.m[free.p2, 5])) 	# Compute probabilities

	if(Uneven) return(c(as.vector(t(par.m)), t))
	c(as.vector(t(par.m))[-c(1,2)], t)

}

# ------------------------------------------------------------------------------------------------------------------
# Computation of probabilities for double high threshold signal detection model
# Input:	par = Vector of parameters, consisting of the following entries:
#
#		Two cases:
#
#		A. Without a separate foil distribution (number of models is an even number): 
#		   For each pair of signals the following parameters are presented, except for the first one two signals):
#
#		1. m.right	= Location of right distribution (for the first two signals no location parameter is specified because m = 0)
#			  	  (In Addition, the scale parameter is set to 1.0: Thus nos scale parameter has to be specified) 
#		2. p.left	= first mixing probability
#		3. p.right	= second mixing probability
#
#		B. With a separate foil distribution (numner of models is an uneven number)
#		   For each pair of signals the following parameters are presented:
#
#		1. m.left 	= Location of the left distribution
#		2. sd 	= Scale of distributions
#		3. m.left 	= Location of the right distribution
#		4. p.left	= first mixing probability
#		5. p.right	= second mixing probability
#
#		C. One set of threshold parameters (identical for each model)is provided:
#		   t[1], t[2], ...., t[n-1] decisions bounds
#
#		fixed = matrix of fixed constraints
#		ident = matrix of identity constraints (used for compatability only)
#
# Output: Vector of probabilities for each pair of distributions plus the for the foil (if present)
#
# Comment:
# The function treats raw probability parameters as well as fixed parameters that are constraint between [0, 1]
# Only parameters not fixed are computed from raw parameters
# ------------------------------------------------------------------------------------------------------------------
SDT.HTSDT.1 <- function(par, n, fixed = NULL, ident = NULL)
{
	if(length(n) > 1)	stop("\nFATAL ERROR: Number of signal distributions must not be a vector\n")

	Uneven <- n%%2 != 0				# Boolean variable indicating whether number of models is uneven				

	if(Uneven)	
	{
		m <- (n-1)/2			# Uneven number of signals -> first data points represent foil distribution
		off.1 <- 4				# Offset for identifying positions of left probabilities
		off.2 <- 5				# Offset for identifying positions of left probabilities
	}
	else
	{	
		m <- n/2					# Number of pairs
		par <- c(0, 1, par)			# Insert mean and standard deviation
		off.1 <- 2					# Offset for identifying positions of left probabilities
		off.2 <- 3					# Offset for identifying positions of left probabilities
	}
	n.p <- 5*m						# Number of parameters without thresholds
	t   <- par[(n.p + 1) : length(par)]		# Extract threshold parameters


	free.p1 <- HTSDT.1.free.p(n.p, m, fixed, off = off.1)	# Left  probabilities
	free.p2 <- HTSDT.1.free.p(n.p, m, fixed, off = off.2)	# Right probabilities

	par.m   <- matrix(par[1:n.p], nr = m, byrow = T)					# Transform to matrix
	par.m[free.p1, 4] <- exp(par.m[free.p1, 4]) / (1 + exp(par.m[free.p1, 4])) 	# Compute probabilities
	par.m[free.p2, 5] <- exp(par.m[free.p2, 5]) / (1 + exp(par.m[free.p2, 5])) 	# Compute probabilities
	par.m <- cbind(par.m, matrix(rep(t, m), nr = m, byrow = T))				# Add threshold parameters

	if(Uneven) return(c(HTSDT.1.Foil.Probs(c(0, 1, t)), as.vector(apply(par.m, 1, HTSDT.1.Probs))))
	as.vector(apply(par.m, 1, HTSDT.1.Probs))
}

#-------------------------------------------------------------------------------------------------------------------------------------
# Same function but without computation of probability parameters from raw parameters (assumes that these have been already computed)
#-------------------------------------------------------------------------------------------------------------------------------------
SDT.HTSDT.1.Pr <- function(par, n, fixed = NULL, ident = NULL)
{
	if(length(n) > 1)	stop("\nFATAL ERROR: Number of signal distributions must not be a vector\n")

	Uneven <- n%%2 != 0				# Boolean variable indicating whether number of models is uneven				

	if(Uneven)	
	{
		m <- (n-1)/2			# Uneven number of signals -> first data points represent foil distribution
	}
	else
	{	
		m <- n/2					# Number of pairs
		par <- c(0, 1, par)			# Insert mean and standard deviation
	}
	n.p <- 5*m						# Number of parameters without thresholds
	t   <- par[(n.p + 1) : length(par)]		# Extract threshold parameters

	par.m   <- matrix(par[1:n.p], nr = m, byrow = T)					# Transform to matrix
	par.m <- cbind(par.m, matrix(rep(t, m), nr = m, byrow = T))				# Add threshold parameters

	if(Uneven) return(c(HTSDT.1.Foil.Probs(c(0, 1, t)), as.vector(apply(par.m, 1, HTSDT.1.Probs))))
	as.vector(apply(par.m, 1, HTSDT.1.Probs))
}


# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix of the derivatives for the foil signal
# Input: 	Parametervector for the foil signal = thresholds)
#
# Output:	d x t Matrix (d = number of data points, t = thresholds)
# ------------------------------------------------------------------------------------------------------------------
HTSDT.1.Foils.Jacobi <- function(par)
{
	n.t  <- length(par)			# Number of parameters = number of thrsholds
	fn   <- dnorm(par)
	sd   <- 1.0	

	# Derivative with respect to thresholds
	dt <- fn/sd							# Derivative of Gaussian with respect to t[i]
	
	J <- matrix(0, nr = n.t+1, nc = n.t)		# Jacobi matrix for thresholds: columns: Parameter, rows = probabilities
	r <- 1
	for(i in 1:n.t)
	{
		J[r, i] <- dt[i]
		r <- r + 1
		J[r, i] <- -dt[i]
	}
	J
}

# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix of the derivatives for a single signal pair
# Input: 	1.Parametervector containing the following entries
#			1. mean.left
#			2. sd
#			3. mean.right
#			4. p.left
#			5. p.right
#			6. thresholds (t thresholds)
#		2. Flag inidicating whether p.left  is free, i.e. a raw parameter that has to be transformed (=T)
#		3. Flag inidicating whether p.right is free, i.e. a raw parameter that has to be transformed (=T)
#
# Output:	2*(t+1) x t+6 Matrix (number of data points x parameters)
# ------------------------------------------------------------------------------------------------------------------
HTSDT.1.Pairs.Jacobi <- function(par, free.p.left, free.p.right)
{
	l   <- length(par)			# Number of parameters
	n.t <- l-5					# Number of thresholds

	if(free.p.left)  	p.left  <- exp(par[4]) / (1 + exp(par[4]))
	else		   	p.left  <- par[4]

	if(free.p.right) 	p.right <- exp(par[5]) / (1 + exp(par[5]))
	else 		   	p.right <- par[5]

	m.left  <- par[1]
	sd 	  <- max(HTSDT.1.Tol, par[2])
	m.right <- par[3]
	t   	  <- par[6:l]

	z.left  <- (t - m.left)/sd
	z.right <- (t - m.right)/sd
	fn.left <- dnorm(z.left)
	fn.right<- dnorm(z.right)
	
	# Derivative with respect to probability parameter p.left and p.right
	dp.left  		<- -pnorm(t, m.left, sd)
	dp.left  		<- c(dp.left[1] , dp.left[-1]  - dp.left[-n.t],  -dp.left[n.t])
	dp.left[1]  	<- dp.left[1] + 1
	dp.left[n.t+1]  	<- dp.left[n.t+1] - 1

	dp.right 		<- -pnorm(t, m.right, sd)
	dp.right 		<- c(dp.right[1], dp.right[-1] - dp.right[-n.t], -dp.right[n.t])

	if(free.p.left)	dp.left  <- dp.left  * p.left  * (1-p.left)
	if(free.p.right)	dp.right <- dp.right * p.right * (1-p.right)

	# Derivative with respect to m
	dm.left  <- -fn.left/sd							# Derivative of left   Gaussians with respect to m
	dm.right <- -fn.right/sd						# Derivative of right  Gaussians with respect to m

	dm.first  <- (1-p.left)  * c(dm.left[1]  , dm.left[-1]  - dm.left[-n.t] , -dm.left[n.t]) 	# First signal
	dm.second <- (1-p.right) * c(dm.right[1], dm.right[-1] - dm.right[-n.t], -dm.right[n.t]) 	# Second signal

	# Derivative with respect to sd
	ds.left   <- -z.left  * fn.left/sd					# Derivative of left   Gaussians with respect to sd
	ds.right  <- -z.right * fn.right/sd					# Derivative of right  Gaussians with respect to sd

	ds.first  <-  (1-p.left) * c(ds.left[1]  , ds.left[-1]  - ds.left[-n.t] , -ds.left[n.t]) 	# First signal
	ds.second <-  (1-p.right)* c(ds.right[1], ds.right[-1] - ds.right[-n.t], -ds.right[n.t]) 	# Second signal

	# Derivative with respect to thresholds
	dt.left  <- fn.left/sd							# Derivative of left   Gaussian with respect to t[i]
	dt.right <- fn.right/sd							# Derivative of right  Gaussian with respect to t[i]

	dt.first  <- (1-p.left)  * dt.left					# Derivative with respect to the data for the first  signal
	dt.second <- (1-p.right) * dt.right					# Derivative with respect to the data for the second signal

	J.first  <- matrix(0, nr = n.t+1, nc = n.t)			# Jacobi matrix for thresholds: columns: Parameter, rows = probabilities for first signal
	J.second <- matrix(0, nr = n.t+1, nc = n.t)			# Jacobi matrix for thresholds: columns: Parameter, rows = probabilities
	r <- 1
	for(i in 1:n.t)
	{
		J.first[r, i]  <- dt.first[i]
		J.second[r, i] <- dt.second[i]
		r <- r + 1
		J.first[r, i]  <- -dt.first[i]
		J.second[r, i] <- -dt.second[i]
	}
	fueller <- rep(0, n.t+1)
	cbind(c(dm.first, fueller), 					# Full Jacobi matrix
		c(ds.first, ds.second), 
		c(fueller,  dm.second),
		c(dp.left, fueller), 
		c(fueller, dp.right), 
		rbind(J.first, J.second))				
}

# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix of the Mixture Model with 3 Gaussian distributions per signal pair
# ------------------------------------------------------------------------------------------------------------------
SDT.HTSDT.1.Jacobi <- function(par, n, fixed)
{
	n.c <- 5		# Number of parameters per signal pair (except thresholds)

	Uneven <- n%%2 != 0
	if(Uneven)
	{
		m <- (n-1)/2
		off.l <- 4						# Position of p.left
		off.r <- 5						# Position of p.right
	}
	else
	{
		m <- n/2						# Number of pairs
		par <- c(0, 1, par)				# Add mean and sd for first signal pair
		off.l <- 2						# Position of p.left
		off.r <- 3						# Position of p.right

	}
	n.par <- 5*m
	t <- par[(n.par + 1) : length(par)]			# Extract threshold parameters
	n.t	 <- length(t)					# Number of thresholds
	n.r 	 <- 2*n.t + 2					# Length of a data block (consisting of two signals)

	pos.p.l <- seq(off.l, n.par, 5)			# Positions of left  probability parameters
	pos.p.r <- seq(off.r, n.par, 5)			# Positions of right probability parameters
	free.p.l  <- !(pos.p.l %in% fixed[2,])		# Boolian vector indicating positions of free p.left  parameters
	free.p.r  <- !(pos.p.r %in% fixed[2,])		# Boolian vector indicating positions of free p.right parameters

	fill.c <- fill.r <- fill <- matrix(0, nr = n.r, nc = n.c)	# Filler-Matrix
	for(i in 1:m)								# Generation of the blocks for the single models
	{
		par.act <- c(par[((i-1)*5+1):((i-1)*5+5)], t)		# Parameter vector for actual pair
		J.act <- HTSDT.1.Pairs.Jacobi(par.act, free.p.l[i], free.p.r[i])
		if(i == 1)
		{
			J   <- J.act[, 1:5]
			J.t <- J.act[, 6:(5+n.t)]
		}
		else
		{
			J   <- rbind(cbind(J, fill.c), cbind(fill.r, J.act[, 1:5]))
			J.t <- rbind(J.t, J.act[, 6:(5+n.t)])
			if(i < m)
			{
					fill.c <- rbind(fill.c, fill)
					fill.r <- cbind(fill.r, fill)
			}
		}
	}
	J <- cbind(J, J.t)							# Bind thresholds

	if(Uneven)
	{
		J.first <- cbind(matrix(0, nr = n.t + 1, nc = m * n.c), HTSDT.1.Foils.Jacobi(t))		# Ad first rows representing the foil distribution
		J <- rbind(J.first, J)
	}
	else
	{
		J <- J[, -c(1,2)]
	}
	J
}

# ------------------------------------------------------------------------------------------------------------------
# Gradient of logL for the HTSDT.1 model
# Input:	Same as for Likelihood function
#
# Output:	Gradient
# ------------------------------------------------------------------------------------------------------------------
SDT.HTSDT.1.Grad <- function(par, data, Model, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
	par 	   <- Restore.par(par, pos, fixed, ident, functional)	# Restore full parameter vector
	gradient <- as.vector(SDT.logL.Jacobi(par, data, Model, n, pos, fixed, ident, functional) %*% SDT.HTSDT.1.Jacobi(par, n, fixed)) 

	# Treat functional constraints
	if(!is.null(functional))
	{
		gr.fct <- attr(par, "SDT.gradient")		# Isolate gradient function
		if(is.null(gr.fct))	stop("\n\n SDT.HTSDT.1.Grad: Error in computation of gradient: No gradient for functional constraints found:\n Set Parameter sym.gr = F in function SDT.Estimate !\n\n")
		gradient <- gr.fct(gradient)
	}	
	Elim.par.grad(gradient, fixed, ident)
}

# ------------------------------------------------------------------------------------------------------------------
# Generate starting parameters due to data and number of signals
# Input:	data = Vector of data
#           data = Vector of data ( in case of a single number, 
#                                   this number is interpreted as inidcating the number of response categories)
#
# Output: String with parameter names
# ------------------------------------------------------------------------------------------------------------------
SDT.HTSDT.1.start.par <- function(n, data)
{
      n.resp <- length(data)
      if(n.resp < 2)
            n.resp <- data*n

	n.t <- n.resp/n - 1	   # Number of thresholds
	if(n %% 2 != 0)		   # Uneven number of signals
	{
		par.m <- c(rep(c(0, 1, 0.1, 0, 0), (n-1)/2), seq(-1.5, 1.5, length = n.t))	
	}
	else					# Even number of signals
	{
		par.m <- c(rep(c(0, 1, 0.1, 0, 0), n/2), seq(-1.5, 1.5, length = n.t))
		par.m <- par.m[-c(1,2)]
	}
	par.m
}

# ------------------------------------------------------------------------------------------------------------------
# Compute upper and lower bounds on parameters
# Input:	         par   = Parameter vector
#		           n   = Number of signals
#           probs.bounds   = Flag indicating whether bounds concern probabilties or raw parameters
#
# Output: A list with lower and upper bounds on parameters
# ------------------------------------------------------------------------------------------------------------------
HTSDT.1.par.bounds <- function(par, n, probs.bounds = F)
{
      Max.Raw.P <- 300
      Min.Var   <- HTSDT.1.Tol
      
      uneven <- (n %% 2) != 0

      if(uneven)
      {
            lower <- rep(-Inf, length(par))
            upper <- rep(Inf, length(par))
      }
      else
      {
            lower <- rep(-Inf, length(par)+ 2)
            upper <- rep(Inf, length(par) + 2)
      }

      if(uneven)
            sd.off <- rep(5, (n-1)/2 - 1)
      else
            sd.off <- rep(5, n/2 - 1)
            
      pos.sd <- cumsum(c(2, sd.off))      # Positions of stddev
      pos.p  <- c(pos.sd + 2, pos.sd + 3) # Positions of probability parameters
            
      lower[pos.sd] <- Min.Var            # Lower bounds for stdev
      
      if(probs.bounds)                    # Bounds for probability parameters
      {
            lower[pos.p] <- 0
            upper[pos.p] <- 1
      }
      else
      {
            lower[pos.p] <- -Max.Raw.P
            upper[pos.p] <-  Max.Raw.P
      }      
      if(!uneven)
      {
        lower <- lower[-c(1,2)]
        upper <- upper[-c(1,2)]   
      }

      list(lower = lower, upper = upper)
}

# ------------------------------------------------------------------------------------------------------------------
# Generate strings with names of the parameters for model
# Input:	par = Vector of parameters
#		  n = number of signal pairs
#
# Output: String with parameter names
# ------------------------------------------------------------------------------------------------------------------
HTSDT.1.par.names <- function(par, n)
{
	if(n %% 2 != 0)
	{
		m   <- (n-1)/2
		n.t <- length(par) - m*5

	}
	else		
	{
		m   <- n/2
		n.t <- length(par) - m*5 + 2
	}
	m.L 	<- paste("mean.left["   , 1:m, "]", sep = "")
	sd   <- paste("stddev[" , 1:m, "]", sep = "")
	m.R  <- paste("mean.right[" , 1:m, "]", sep = "")
	p.L  <- paste("p.left[", 1:m, "]", sep = "")
	p.R  <- paste("p.right[", 1:m, "]", sep = "")

	p.name   <- NULL
	for(i in 1:m) p.name <- c(p.name, m.L[i], sd[i], m.R[i], p.L[i], p.R[i])
	if(n %% 2 == 0)	p.name <- p.name[-c(1,2)]

	c(p.name, paste("t-", 1:n.t, sep = ""))
}

# ------------------------------------------------------------------------------------------------------------------
# Name of the model
# ------------------------------------------------------------------------------------------------------------------
HTSDT.1.model.name <- "HTSDT.1: Double high threshold signal detection model"

# ------------------------------------------------------------------------------------------------------------------
# Format parameters for HTSDT.1 model
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Par.HTSDT.1 <- function(par, n, fixed, dec)
{
	# Compute probabilities for non-fixed probability parameters
	par <- HTSDT.1.cmp.p(par, n, fixed) 	

	# Format output
	if(n %% 2 != 0)		# Number of models = uneven
	{
		m   <- (n-1)/2
		t   <- round(par[(m*5 + 1):length(par)], dec)				# Extract threshold parameters
		n.t <- length(t)
		par <- round(c(0, 1, NA, NA, NA, par[1:(m*5)]), dec)
		Fp  <- matrix(par, nc = m+1)
		Fp  <- rbind(Fp, matrix(rep(t, m+1), nc = m+1))
		colnames(Fp) <- c("Foil", paste("HTSDT.1-", 1:m, sep = ""))
		rownames(Fp) <- c("mean.left", "stddev",  "mean.right", "p.left", "p.right", paste("t-", 1:n.t, sep = ""))
	}
	else				# Number of models = even
	{
		m   <- n/2
		t <- round(par[(m*5 - 1):length(par)], dec)				# Extract threshold parameters
		n.t <- length(t)
		par <- c(0, 1, par[1:(m*5 - 2)])
		Fp  <- matrix(round(par, dec), nc = m)
		Fp  <- rbind(Fp, matrix(rep(t, m), nc = m))
		colnames(Fp) <- c(paste("HTSDT.1-", 1:m, sep = ""))
		rownames(Fp) <- c("mean.left", "stddev",  "mean.right", "p.left", "p.right", paste("t-", 1:n.t, sep = ""))
	}
	Fp
}

# ------------------------------------------------------------------------------------------------------------------
# Format free parameters and adapt standard errors
# Input:	1. px2 Matrix with parameters and standard errors
#		2. full parameter vector
#		3. data vector
#		4. number of signals
#		5. matrix of fixed constraints
#		6. matrix of equality constraints
#		7. function implementing functional constraints
#           8. Confidence level
#           9. Number of decimal places
#
# Output: px3 Matrix of parameters and standard errors
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Par.est.HTSDT.1 <- function(Pars, full.par, data, n, fixed, ident, functional, alpha, deci)
{
	pos.free <- SavePos.par(full.par, fixed, ident)		# Positions of free parameters
	
	# Compute Positions of probabilities in full parameter vector
	if(n%%2 != 0)	# Uneven number of signals
	{
		m	<- (n-1)/2
		len.p <- m*5
		pos.p <- c(seq(4, len.p-1, 5), seq(5, len.p, 5))
	
	}
	else			# Even number of signals
	{
		m	<- n/2
		len.p <- m*5 - 2
		pos.p <- c(seq(2, len.p-1, 5), seq(3, len.p, 5))
		
	}
	pos   <- pos.free %in% pos.p					# Positions of probability parameters in free parameter vector

	# Compute probabilities and adapt standard errors
	Pars[pos, 1] <- exp(Pars[pos ,1]) / (1 + exp(Pars[pos ,1]))
	Pars[pos, 2] <- Pars[pos, 2] * (Pars[pos, 1]*(1-Pars[pos, 1]))

	# Compute asymptotic standard errors numerically
	Pars <- cbind(Pars, SDT.HTSDT.1.Hessian.p(full.par, data, n, fixed = fixed, ident = ident, functional = functional))

      # Compute confidence intervals
      bd    <- HTSDT.1.par.bounds(full.par, n, probs.bounds = T)       # Bounds on probabilities
      bd.lo <- Elim.par(bd$lower, fixed, ident)
      bd.up <- Elim.par(bd$upper, fixed, ident)

      qt.cfi <- qnorm((1-alpha)/2)
      cfi.lo <- Pars[,1] + Pars[,2]*qt.cfi
      cfi.up <- Pars[,1] - Pars[,2]*qt.cfi
            
      # Respect lower bounds on sd parameter
      cfi.lo <- ifelse(cfi.lo < bd.lo, bd.lo, cfi.lo)
      cfi.up <- ifelse(cfi.up > bd.up, bd.up, cfi.up)
      
      # Add columns with confidence intervals and specify new column names
      Pars <- cbind(Pars, cfi.lo, cfi.up)
	colnames(Pars) <- c("Value", "SE(delta)", "SE(numerical)", paste("CFI-", alpha*100, "(Lower)", sep = ""), paste("CFI-", alpha*100, "(Upper)", sep = ""))

	round(Pars, deci)
}


# ---------------------------------------------------------------------------------------------------------------------
# Check number of parameters and data points
# Input:	1. Parameter vector
#		2. Data vector
#		3. Number of signals
#
# Output:	A list with three entries:
#			1. Error flag: T = Error, F = No error
#			2. 2x2 matrix with actual (first row) and predicted (scond row) number of parameters and data points
# ----------------------------------------------------------------------------------------------------------------------
SDT.HTSDT.1.check <- function(par, data, n)
{
	if(n%%2 != 0)	# Uneven number of signals
	{
		m	 <- (n-1)/2
		n.par  <- 5*m + length(data)/n - 1		# Predicted number of parameters due to data
		n.data <- (length(par) - 5*m + 1)*n		# Predicted number of data points 
	}
	else			# Even number of signals
	{
		m	 <- n/2
		n.par  <- 5*m - 2 + length(data)/n - 1	# Predicted number of parameters due to data
		n.data <- (length(par) - 5*m + 2 + 1)*n	# Predicted number of data points 
	}

	Erg <- matrix(c(length(par), length(data), n.par, n.data), nc = 2)
	rownames(Erg) <- c("Parameters", "Data points")
	colnames(Erg) <- c("Found", "Expected")

	Error <- any(Erg[, 1] != Erg[, 2])
	list(Error = Error, Stat = Erg, Msg = "NOTE: Computation of the expected number of parameters is based on the to number of data points and vice versa")
}

# ------------------------------------------------------------------------------------------------------------------
# Functions for computing Hessian numerically
# ------------------------------------------------------------------------------------------------------------------
SDT.HTSDT.1.Hessian.p <- function(par, data, n, fixed = fixed, ident = ident, functional = functional)
{
	par.p   <- HTSDT.1.cmp.p(par, n, fixed) 		# Compute probabilities
	pos.par <- SavePos.par(par, fixed, ident)		# Positions of free parameters
	par.p	  <- Elim.par(par.p,  fixed, ident)		# Eliminate constrained parameters
	Model   <- SDT.HTSDT.1.Pr				# Function for computing model probabilities
	hess    <- hessian(func = SDT.logL, x = par.p, data = data, Model = Model, n = n, pos = pos.par, fixed = fixed, ident = ident, functional = functional)

	if(is.null(hess) || is.na(hess) || is.nan(hess)) return(rep(NA, length(par.p))) 	# Hessian absent or not OK -> retuen (NA)

	Erg <- Test.Hessian(hess)	
	if(Erg$P.d == F)		return(rep(NA, length(par.p)))					# Not positive definite -> return(NA) 		
	Cov <- solve(hess)
	return(sqrt(diag(Cov)))
}


#===============================================================================
#
# Functions for comupting the model matrix (analytically and numerically)
# for the HTSDT.1 model
#
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Model incorporating restrictions
# The parameter passed to the function contains the free parameters only
# ------------------------------------------------------------------------------------------------------------------
SDT.HTSDT.1.Full <- function(rest.par, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
      full.par <- Restore.par(rest.par, pos, fixed, ident, functional)	# Restore full parameter vector
      SDT.HTSDT.1(full.par, n, fixed, ident)
}

# ------------------------------------------------------------------------------------------------------------------
# Numerical computation of the model matrix 
# ------------------------------------------------------------------------------------------------------------------
SDT.HTSDT.1.Model.Matrix.Num <- function(full.par, n, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)
      par <- Elim.par(full.par, fixed, ident)
      J <- jacobian(SDT.HTSDT.1.Full, par, n = n, pos = pos, fixed = fixed, ident = ident, functional = functional)
}

# ------------------------------------------------------------------------------
# Analytical computation of the model matrix 
# Input:    1. Full parameter vector
#           2. Number of signals
#           3. 2 x p - matrix of fixed constraints
#           4. 2 x q - matrix of identity constraints
#           5. Jacobian of functional constraints (supplied by the user)
#
# Output:   Modelmatrix
#
# Comment:  The Modelmatrix of funtional constraints must be a full matrix that
#           is premultiplied by the Jacobian
# ------------------------------------------------------------------------------
SDT.HTSDT.1.Model.Matrix <- function(full.par, n, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)

      J   <- SDT.HTSDT.1.Jacobi(full.par, n = n, fixed)
      if(!is.null(ident))
      {
            J.Id <- diag(length(full.par))
            Index <- matrix(c(ident[2,], ident[1,]), nc = 2)
            J.Id[Index] <- 1
            J <- J %*% J.Id
      }
      
      # Treatment of funtional constraints
      if(!is.null(functional))
      {
            J.fct <- attr(functional, "SDT.Jacobian")
            if(is.null(J.fct)) 
                  return(NULL)           # No Jacobian function supplied -> return (NULL)
            J <- J %*% J.fct(full.par)
      }
      J[, pos]
}
