# ===========================================================================
# SDT models with item response function
# Date of creation: December 2017
# Date of last update: aUGUST 2018
#
# Author: Siegfried Macho
# ===========================================================================
SDT.Tol <- 1e-12
MINWEIGHT  <- 1e-40
MAX.TH <- 20

library(statmod)
library(npmlreg)


#===============================================================================
# 0. COMPUTATION OF QUADRATURE POINTS
#===============================================================================

#-------------------------------------------------------------------------------
# Function for computing the Gauss-Hermite points and weights
# Input:	1. Number of points
#           2. minimal weights
#
# Output: 	Matrix with three columns:    1. Points
#                                         2. Weights of Gauss-Hermite
#                                         3. Weights of adaptive Gauss-Hermite
#
# Comment:  The points and weights are with respect to the Gaussian Kernel N(0, 1)
#-------------------------------------------------------------------------------
Generate.GH <- function(npt, minweight = MINWEIGHT)
{
      GH.Structure  <- gqz(numnodes = npt, minweight)   # Points and weights for Gauss-Hermite Quadrature
      cbind(GH.Structure, GH.Structure[,2] * exp(GH.Structure[,1]^2/2))	 # Weights for adaptive Gauss-Hermite Quadrature
}

#===============================================================================
#
# 1. HELP FUNCTIONS
#
#===============================================================================

# ----------------------------------------------------------------------
# Function for computing parametervector with robust thresholds
# Input:	Vector of parameters: (mean.1, sigma.1, mean.2, sigma.2 ...,,
#                                c[1], c[2], ...., c[n-1],
#                                a[1], a[2], ..., a[n-1])  [discrimination parameters]
#
# Output: 	Vector with robust thresholds c[1], c[1]+exp(c[2]),...c[1]+exp(c[2])+...+exp(c[n])
# ----------------------------------------------------------------------
IRF.Robust.Thresholds <- function(parvec, cfg)
{
      n <- cfg$n.sdt
      len.th <- as.integer((length(parvec) - 2*n) / 2)
      if(len.th <= 1)
          return(parvec)

      th <- parvec[(2*n + 1):(2*n + len.th)] # extract thresholds
      th[th > MAX.TH] <- MAX.TH

      c(parvec[1:(2*n)], cumsum(c(th[1], exp(th[2:len.th]))), parvec[(2*n + len.th + 1):(2*n + len.th + len.th)])
}

# ----------------------------------------------------------------------
# Function for computing log incremental thresholds from  parameter with robust theresholds
# Input:	Vector of parameters: (mean.1, sigma.1, mean.2, sigma.2 ...,,
#                                c[1], c[2], ...., c[n-1],
#                                a[1], a[2], ..., a[n-1])  [discrimination parameters]
#
# Output: Vector with log increments: c[1], c[2],...c[n]
# ----------------------------------------------------------------------
IRF.Inverse.Thresholds <- function(parvec, cfg)
{
      n <- cfg$n.sdt
      len.th <- as.integer((length(parvec) - 2*n) / 2)
      if(len.th <= 1)
          return(parvec)
      th <- parvec[(2*n + 1):(2*n + len.th)] # extract thresholds

      c(parvec[1:(2*n)], th[1], log(th[-1] - th[-len.th]), parvec[(2*n + len.th + 1):(2*n + len.th + len.th)])
}

# ----------------------------------------------------------------------
# Function for generating/correcting/filling the configuration file
# Input:	cfg
#
# Output: a configuration list with proper entries.
# ----------------------------------------------------------------------
IRF.Gauss.Check.Fill.Cfg <- function(cfg = NULL)
{
      # Default configuration
      default.n           <- 2
      default.restriction <- "STANDARD"
      default.npt         <- 35
      default.model       <- "ORM"
      default.robust      <- F
      
      # Case 1: cfg does not exist
      if(is.null(cfg))
      {
            cfg <- list(n.sdt       = default.n,               # Number of models
                        restriction = default.restriction,     # Type of restriction
                        npt         = default.npt,             # Number of quadrature points
                        model       = default.model,           # Defauzlt model
                        robust      = default.robust)          # Robust flag 
            return(cfg)
      }
      # Case 2: cfg is a number (inidcating the number of models
      if(is.numeric(cfg))
      {
            cfg      <- list(n.sdt       = as.integer(cfg),     # Number of models
                             restriction = default.restriction, # Type of restriction
                             npt         = default.npt,         # Number of quadrature points
                             model       = default.model,       # Defauzlt model
                             robust      = default.robust)      # Robust flag 
            return(cfg)
      }
      cfg.new <- list(n.sdt       = ifelse(is.null(cfg$n.sdt),       default.n,           cfg$n.sdt),
                      restriction = IRF.Gauss.Check.Restriction(cfg$restriction),
                      npt         = ifelse(is.null(cfg$npt),         default.npt,         cfg$npt),
                      model       = ifelse(is.null(cfg$model),       default.model,       cfg$model),
                      robust      = ifelse(is.null(cfg$robust),      default.robust,      cfg$robust)
                     )
      # Checks
      if(cfg.new$n.sdt < 2)
            stop(paste("\n\n FATAL ERROR in IRF.Check.Fill.Cfg (Check of configuration information):\n Number of SDT signals <", cfg.new$n.sdt, "> smaller than <2>\n(Minimum number = 2)\n\n", sep = ""))
      if(toupper(cfg.new$model) != "GRM" && toupper(cfg.new$model) != "ORM") {
            warn.msg <- paste("\n\n Wrong model: <", cfg.new$model, ">\n", " --> changed to: <", default.model, ">\n\n", sep = "")
            warning(warn.msg)
            cfg.new$model <- default.model
            }
      if(cfg.new$npt < 3) {
            warn.msg <- paste("\n\n Number of quadrature points too small: <", cfg.new$npt, ">\n", " --> changed to: <", default.npt, ">\n\n", sep = "")
            warning(warn.msg)
            cfg.new$npt <- default.npt
            }  
       if(cfg.new$robust != T && cfg.new$robust != F) {      
            warn.msg <- paste("\n\n Value of robust flag not correct: <", cfg.new$robust, ">\n", " --> changed to: <", default.robust, ">\n\n", sep = "")
            warning(warn.msg)
            cfg.new$robust <- default.robust
            }    
      return(cfg.new)
}  

# ----------------------------------------------------------------------
# Function chacking for types of restrictions
# Input:	1. string with restriction information
#
# Output: string with restriction information
# ----------------------------------------------------------------------
IRF.Gauss.Check.Restriction <- function(restriction = "STANDARD")
{                 
      default.restriction <- "STANDARD"

      if(is.null(restriction))
            return(default.restriction)
            
      restriction <- toupper(restriction)                             # convert to upper cases    
      if(charmatch(restriction, c("STANDARD"), nomatch = 0) > 0)
            return("STANDARD")
      if(charmatch(restriction, c("STANDARD-EQUAL-A", "STANDARD.EQUAL.A"), nomatch = 0) > 0)
            return("STANDARD-EQUAL-A")
      if(charmatch(restriction, c("EQUAL-A", "EQUAL.A"), nomatch = 0) > 0)
            return("EQUAL-A")
      if(charmatch(restriction, c("EQUAL-S", "EQUAL.S"), nomatch = 0) > 0)
            return("EQUAL-S")
      if(charmatch(restriction, c("EQUAL-AS", "EQUAL-SA", "EQUAL.AS", "EQUAL.SA"), nomatch = 0) > 0)
            return("EQUAL-AS")
      if(charmatch(restriction, c("EQUAL-S-FIX-A", "EQUAL.S.FIX.A"), nomatch = 0) > 0)
            return("EQUAL-S-FIX-A")
      if(charmatch(restriction, c("NO"), nomatch = 0) > 0 || restriction == F)
            return("NO")
      warn.msg <- paste("\n\n IRF.Restriction: Unkonwn restriction in configuration list: <", restriction, "> --> changed to: <", default.restriction, ">\n\n", sep = "")      
      warning(warn.msg)
      return(default.restriction)
}            

# ------------------------------------------------------------------------------------------------------------------
# Generate starting parameters due to data and number of signals
# Input:	cfg = Configuration structure: A list with the following entries
#                 - n.sdt         Number of SDT models
#                 - model         String "ORM" (Ordinal Rasch Model) or "GRM" (Graded Response Model)
#                 - restriction   String indicating the restrictions
#                 - npt           Number of quadrature points
#                 - robust        Flag indicating whether robust thresholds should be employed
#
#    datavec = Data vector
#
# Output: String with parameter values
# ------------------------------------------------------------------------------------------------------------------
IRF.Gauss.start.par <- function(cfg, datavec)
{
      cfg <- IRF.Gauss.Check.Fill.Cfg(cfg)

      #-------------------------------------------------------------------------
      # Start parameters of the ranking model:
      # For each model: m.new, sd.new, m.old, sd.old
      # m.new   = mean of new distribution
      # sd.new  = standard deviation of new distribution
      # m.old   = mean of old distribution
      # sd.old  = standard deviation of old distribution
      # c[i]    = thresholds: number = categories - 1
      # a[i]    = discriminatiion parameters: number = categories - 1
      #-------------------------------------------------------------------------
      
      # Determine the nomber of thresholds
      n.th <- length(datavec) / cfg$n.sdt - 1 

      par.IRF <- rep(c(0, 1), cfg$n.sdt)    # mean and sd of Gaussian models
      par.IRF <- c(par.IRF, seq(-0.5, 0.5, length = n.th))
      par.IRF <- c(par.IRF, rep(1, n.th))          # Add dscrimination parameters 
      if(!is.null(cfg$robust) && cfg$robust == T)  # Transform to robust threshoolds
         par.IRF <- IRF.Inverse.Thresholds(par.IRF, cfg)
      return(par.IRF)
}

# ------------------------------------------------------------------------------------------------------------------
# Generate parameters names
# Input: parvec = Parameter vector
#
#           cfg = Configuration structure: A list with the following entries
#                 - n.sdt         Number of SDT models
#                 - model         String "ORM" (Ordinal Rasch Model) or "GRM" (Graded Response Model)
#                 - restriction   String indicating the restrictions
#                 - npt           Number of quadrature points
#                 - robust        Flag indicating whether robust thresholds should be employed
#
# Output: String with parameter names
# ------------------------------------------------------------------------------------------------------------------
IRF.Gauss.par.names <- function(parvec, cfg)
{
      cfg <- IRF.Gauss.Check.Fill.Cfg(cfg)

      Name.IRF <- NULL
      # 1. Names of the Models
      for(i in 1:cfg$n.sdt) {
              Name.IRF <- c(Name.IRF, paste("[SDT-", i, "]   Mean", sep = ""))
              Name.IRF <- c(Name.IRF, paste("[SDT-", i, "] Stddev", sep = ""))
           }

      # 2. Thresholds
      n.th <- as.integer((length(parvec) - 2*cfg$n.sdt) / 2)
      Name.IRF <- c(Name.IRF, paste("c[", 1:n.th, "]", sep = ""))
      
      # 3. Discrimination parameters
      Name.IRF <- c(Name.IRF, paste("a[", 1:n.th, "]", sep = ""))
      return(Name.IRF)
}     

# ------------------------------------------------------------------------------------------------------------------
# Generate upper and lower bounds of parameter
# Input:	parvec = Vector of parameters
#		   cfg = configruation structure
#
# Output: List with lower and upper bounds on parameter vector
# ------------------------------------------------------------------------------------------------------------------
IRF.Gauss.par.bounds <- function(parvec, cfg)
{
    Tol.0 <- 1e-7
    Tol.a <- 700

    len.p <- length(parvec)
    n.th  <- as.integer((len.p - 2*cfg$n.sdt) / 2)

    lower <- c(rep(c(-Inf, Tol.0), cfg$n.sdt), rep(-Inf, n.th), rep(Tol.0, n.th))
    upper <- c(rep(c( Inf, Inf),   cfg$n.sdt), rep( Inf, n.th), rep(Tol.a, n.th))
      
    return(list(lower = lower, upper = upper))
}      


# ------------------------------------------------------------------------------------------------------------------
# Generate string with general model information
# Input:    cfg = Configuration structure: A list with the following entries
#                 - n.sdt         Number of SDT models
#                 - model         String "ORM" (Ordinal Rasch Model) or "GRM" (Graded Response Model)
#                 - restriction   String indicating the restrictions
#                 - npt           Number of quadrature points
#                 - robust        Flag indicating whether robust thresholds should be employed
#
# Output: String with mdoel information
# ------------------------------------------------------------------------------------------------------------------
IRF.Gauss.model.name <- function(cfg)
{
     cfg <- IRF.Gauss.Check.Fill.Cfg(cfg)
     return(paste("Gaussian SDT model with probabilistic item response functions <Model = ", cfg$model, ">, <Restriction = ", 
                  cfg$restriction, ">, <Robust thresholds = ", cfg$robust, ">, <Number of quadrature points = ", cfg$npt, ">", sep = ""))
}
     
# ------------------------------------------------------------------------------------------------------------------
# Set restrictions
# ------------------------------------------------------------------------------------------------------------------
IRF.Gauss.Set.Restriction <- function(parvec, cfg, fixed, ident, functional = NULL)
{
        # If configuration information is not complete, then complete it
        cfg <- IRF.Gauss.Check.Fill.Cfg(cfg)

        if(cfg$restriction == "STANDARD")
            return(IRF.Gauss.Set.Standard.Restriction(parvec, cfg, fixed, ident, functional))
        if(cfg$restriction == "STANDARD-EQUAL-A")
            return(IRF.Gauss.Set.Standard.EqualA.Restriction(parvec, cfg, fixed, ident, functional))
        if(cfg$restriction == "EQUAL-A")
            return(IRF.Gauss.Set.EqualA.Restriction(parvec, cfg, fixed, ident, functional))
        if(cfg$restriction == "EQUAL-S")
            return(IRF.Gauss.Set.EqualS.Restriction(parvec, cfg, fixed, ident, functional))
        if(cfg$restriction == "EQUAL-AS")
            return(IRF.Gauss.Set.EqualAS.Restriction(parvec, cfg, fixed, ident, functional))
        if(cfg$restriction == "EQUAL-S-FIX-A")
            return(IRF.Gauss.Set.EqualS.FixA.Restriction(parvec, cfg, fixed, ident, functional))

        return(list(fixed = fixed, ident = ident, functional = functional))
}

# ------------------------------------------------------------------------------------------------------------------
# IRF.Gauss: Set standard restrictions
# Input:      1. Parametervector,           
#             2. Configuration information
#             3. matrix of fixed constraints
#             4. Matrix of identity constraints
#             5. Function implementing functional constrints (not used)
#
# Output:     A list with new fixed, ident and fucntional 
#
# Comment:    Duplicates are removed from fixed and ident
# ------------------------------------------------------------------------------------------------------------------
IRF.Gauss.Set.Standard.Restriction <- function(parvec, cfg, fixed, ident, functional)
{
      # 1. Fix mean and standard devition of Gauss
      fixed.s <- matrix(c(0, 1, 1:2), nr = 2, byrow = T)

      # 2. Fix discrimination parameters
      par.len <- length(parvec)
      n.th    <- (par.len - 2*cfg$n.sdt) / 2 
      fixed.s <- cbind(fixed.s, matrix(c(rep(1, n.th), (par.len - n.th + 1):par.len), nr = 2, byrow = T))
      
      ident.s <- NULL
      return(IRF.Remove.Duplicated(fixed.s, ident.s, fixed, ident, functional))
}

IRF.Gauss.Set.Standard.EqualA.Restriction <- function(parvec, cfg, fixed, ident, functional)
{
      # 1. Fix mean and standard devition of Gauss
      fixed.s <- matrix(c(0, 1, 1:2), nr = 2, byrow = T)

      # 2. Set discrimination parameters to be equal
      par.len <- length(parvec)
      n.th    <- (par.len - 2*cfg$n.sdt) / 2 
      ident.s <- matrix(c(rep(par.len - n.th + 1, n.th - 1), (par.len - n.th + 2):par.len), nr = 2, byrow = T)
      
      return(IRF.Remove.Duplicated(fixed.s, ident.s, fixed, ident, functional))
}

IRF.Gauss.Set.EqualA.Restriction <- function(parvec, cfg, fixed, ident, functional)
{
      # 1. Fix mean and standard devition of Gauss
      fixed.s <- NULL

      # 2. Set discrimination parameters to be equal
      par.len <- length(parvec)
      n.th    <- (par.len - 2*cfg$n.sdt) / 2 
      ident.s <- matrix(c(rep(par.len - n.th + 1, n.th - 1), (par.len - n.th + 2):par.len), nr = 2, byrow = T)
      
      return(IRF.Remove.Duplicated(fixed.s, ident.s, fixed, ident, functional))
}

IRF.Gauss.Set.EqualS.Restriction <- function(parvec, cfg, fixed, ident, functional)
{
      # 1. Fix mean of first and standard deviations of all Gaussian models
      fixed.s <-  matrix(c(0, rep(1, cfg$n.sdt), 1, 2*(1:cfg$n.sdt)), nr = 2, byrow = T)

      ident.s <- NULL
      
      return(IRF.Remove.Duplicated(fixed.s, ident.s, fixed, ident, functional))
}

IRF.Gauss.Set.EqualAS.Restriction <- function(parvec, cfg, fixed, ident, functional)
{
      # 1. Fix mean of first and standard deviations of all Gaussian models
      fixed.s <-  matrix(c(0, rep(1, cfg$n.sdt), 1, 2*(1:cfg$n.sdt)), nr = 2, byrow = T)

      # 2. Set discrimination parameters to be equal
      par.len <- length(parvec)
      n.th    <- (par.len - 2*cfg$n.sdt) / 2 
      ident.s <- matrix(c(rep(par.len - n.th + 1, n.th - 1), (par.len - n.th + 2):par.len), nr = 2, byrow = T)
      
      return(IRF.Remove.Duplicated(fixed.s, ident.s, fixed, ident, functional))
}

IRF.Gauss.Set.EqualS.FixA.Restriction <- function(parvec, cfg, fixed, ident, functional)
{
      # 1. Fix mean of first and standard deviations of all Gaussian models
      fixed.s <-  matrix(c(0, rep(1, cfg$n.sdt), 1, 2*(1:cfg$n.sdt)), nr = 2, byrow = T)

      # 2. Set discrimination parameters to be equal
      par.len <- length(parvec)
      n.th    <- (par.len - 2*cfg$n.sdt) / 2 
      fixed.s <- cbind(fixed.s, matrix(c(rep(1, n.th), (par.len - n.th + 1):par.len), nr = 2, byrow = T))
      
      ident.s <- NULL
      
      return(IRF.Remove.Duplicated(fixed.s, ident.s, fixed, ident, functional))
}


# ------------------------------------------------------------------------------------------------------------------
# Remove duplicated restrictions
# Input: fixed.s    = matrix of fixed    constraints by configuration
#        ident.s    = matrix of identity constraints by configuration
#        fixed      = matrix of fixed    constraints manually specified  
#        ident      = matrix of indentiy constraints manually specified 
#        functional = not used
#
# Output: list with new matrice of constraints with duplicated constraints removed
# ------------------------------------------------------------------------------------------------------------------
IRF.Remove.Duplicated <- function(fixed.s, ident.s, fixed, ident, functional)
{

      # Check for uniqueness add to fixed constraints also respecting identiy constraints
      pos.unique <-  !(fixed.s[2,] %in% fixed[2,]) & !(fixed.s[2,] %in% ident[2,])
	    fixed <- cbind(fixed.s[, pos.unique], fixed)

      # Check for uniqueness add to identity constraints
      pos.unique <-  !(ident.s[2,] %in% ident[2,]) & !(ident.s[2,] %in%  fixed[2,])
	    ident <- cbind(ident.s[, pos.unique], ident)

      # Remove fixed positions from identity constraints
      pos.unique <- !(ident[2,] %in% fixed[2,])
      ident <- ident[, pos.unique, drop = F]

      list(fixed = fixed, ident = ident, functional = functional)
}

# ------------------------------------------------------------------------------------------------------------------
# Check whether parameter and data fit together
# Input: parvec  = parameter vectort
#        datavec = data vector
#        cfg     = list with configuration information
#
# Output: list with new matrice of constraints with duplicated constraints removed
# ------------------------------------------------------------------------------------------------------------------
IRF.Gauss.check <- function(parvec, datavec, cfg)
{
      Msg <- "NOTE: Computation of the expected number of parameters is based on the number of data points and vice versa"
	
      plen     <- length(parvec)
      dlen     <- length(datavec)
      n.th.dat <- as.integer(dlen / cfg$n.sdt - 1) # number of thresholds based on data
      n.th.mod <- as.integer((plen - 2 * cfg$n.sdt) / 2)
      
      plen.exp <- as.integer(2 * (cfg$n.sdt + n.th.dat))  # Expected parameter length due to data and configuration
      dlen.exp <- as.integer(cfg$n.sdt * (n.th.mod + 1))

	    Erg <- matrix(c(plen, dlen, plen.exp, dlen.exp), nc = 2)
	    rownames(Erg) <- c("Parameters", "Data points")
	    colnames(Erg) <- c("Found", "Expected")
	    Error <- any(Erg[, 1] != Erg[, 2])

	    list(Error = Error, Stat = Erg, Msg = Msg)
}

#===============================================================================
#
# 2. ITEM-RESPONSE FUNCTIONS
#
#===============================================================================

# ----------------------------------------------------------------------
# Graded Response Model of Samejima (1969)
# Input: 1. theta = latent trait
#        2. b     = difficutly / threshold paraemters
#        3. a     = vector of discrimination parameters
#
# Ouptut: Vector of response probabilities for different categories
# Comment: IRF.GRM.graph is used for graphical display
# ----------------------------------------------------------------------
IRF.GRM.graph <- function(theta, b, a)
{
    p.star = 1/(1+exp(a*(b - theta)))
    return(c(1, p.star) - c(p.star, 0))
}

# Used for Gauss-Hermite integration
IRF.GRM.cond <- function(x, b, a)
{
    p.star = 1/(1 + exp(a*(b - x[1])))
    return((c(1, p.star) - c(p.star, 0)) * x[2])
}

# ----------------------------------------------------------------------
# Ordinal Rasch Model of Masters (1982)
# Input: 1. theta = latent trait
#        2. b     = difficulty / threshold parameter
#        3. a     = vector of discrimination parameters
#
# Ouptut: Vector of response probabilities for different categories
# Comment: IRF.ORM.graph is used for graphical display
# ----------------------------------------------------------------------
IRF.ORM.cond.graph <- function(theta, b, a)
{
    MAX.VAL <- 700
    
    exp.term <- cumsum(a*(theta - b))
    exp.term[exp.term > MAX.VAL] <- MAX.VAL
    nominator   <- exp(exp.term)
    denominator <- 1 + sum(nominator)
    return(c(1, nominator)/denominator)
}


# Used for Gauss-Hermite integration
IRF.ORM.cond <- function(x, b, a)
{
    MAX.VAL <- 700
    
    exp.term <- cumsum(a*(x[1] - b))
    exp.term[exp.term > MAX.VAL] <- MAX.VAL
    nominator   <- exp(exp.term)
    denominator <- 1 + sum(nominator)
    return(x[2] * c(1, nominator)/denominator)
}    
    
# ----------------------------------------------------------------------
# Graded Response Model of Samejima (1969): Unconditional probabilities
# Input: 1. b     = vector of difficutly / thrreschold paraemters
#        2. a     = vector of discrimination parameters
#        3. mu    = mean of the Gaussian distribution (of theta)
#        4. sigma = standard deviation of the Gaussian distribution (of theta)
#
# Ouptut: Vector of response probabilities for different categories
# ----------------------------------------------------------------------
IRF.GRM.uc  <- function(b, a, mu = 0, sigma = 1)
{
      GH.Structure[,1] <- GH.Structure[,1] * sigma + mu
      rowSums(apply(GH.Structure, 1, IRF.GRM.cond, b = b, a = a))
}

# ----------------------------------------------------------------------
# Ordinal Rasch Model of Masters (1982): Unconditonal probabilities
# Input: 1. b     = vector of difficutly / thrreschold paraemters
#        2. a     = vector of discrimination parameters
#        3. mu    = mean of the Gaussian distribution (of theta)
#        4. sigma = standard deviation of the Gaussian distribution (of theta)
#
# Ouptut: Vector of response probabilities for different categories
# ----------------------------------------------------------------------
IRF.ORM.uc  <- function(b, a, mu = 0, sigma = 1)
{
      GH.Structure[,1] <- GH.Structure[,1] * sigma + mu
      rowSums(apply(GH.Structure, 1, IRF.ORM.cond, b = b, a = a))
}

# ------------------------------------------------------------------------
# Graded Response Model: Full model with arbitrary many Gaussian signals
# Input: 1. parvec = parameter vector comprizing the following parameters
#                    (i)   2 * n.sdt Gaussian parameters (mu, sigma)
#                    (ii)  R-1 threshold parameters  (with R = number of response categories)
#                    (iii) R-1 discrimination parameters
#        2. cfg    = configuration structure
#        3. fixed  = NULL (not used)
#        4. ident  = NULL (not used)
#
# Ouptut: Vector of response probabilities for all categories and signals
# ------------------------------------------------------------------------
IRF.GRM.Gauss  <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
    n.Gauss <- as.integer(2 * cfg$n.sdt)                      # number of parameters of the Gaussian models
    n.th    <- as.integer((length(parvec) - n.Gauss) / 2)     # number of thresholds and discrimination parameters
    b <- parvec[(n.Gauss+1):(n.Gauss + n.th)]                 # extract thresholds
    a <- parvec[(n.Gauss + n.th + 1):(n.Gauss + n.th + n.th)] # extract discrimination parameters
    probs <- NULL
    for(i in 1:cfg$n.sdt) probs <- c(probs, IRF.GRM.uc(b, a, mu = parvec[2 * (i-1) + 1], sigma = parvec[2 * (i-1) + 2]))
    
    return(probs)
}

# ------------------------------------------------------------------------
# Ordinal Rasch Model: Full model with arbitrary many Gaussian signals
# Input: 1. parvec = parameter vector comprizing the following parameters
#                    (i)   2 * n.sdt Gaussian parameters (mu, sigma)
#                    (ii)  R-1 threshold parameters  (with R = number of response categories)
#                    (iii) R-1 discrimination parameters
#        2. cfg    = configuration structure
#        3. fixed  = NULL (not used)
#        4. ident  = NULL (not used)
#
# Ouptut: Vector of response probabilities for all categories and signals
# ------------------------------------------------------------------------
IRF.ORM.Gauss  <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
    n.Gauss <- as.integer(2 * cfg$n.sdt)                      # number of parameters of the Gaussian models
    n.th    <- as.integer((length(parvec) - n.Gauss) / 2)     # number of thresohlds and discrimination parameters
    b <- parvec[(n.Gauss+1):(n.Gauss + n.th)]                 # extract thresholds
    a <- parvec[(n.Gauss + n.th + 1):(n.Gauss + n.th + n.th)] # extract discrimination parameters
    probs <- NULL
    for(i in 1:cfg$n.sdt) probs <- c(probs, IRF.ORM.uc(b, a, mu = parvec[2 * (i-1) + 1], sigma = parvec[2 * (i-1) + 2]))

    return(probs)
}

# ------------------------------------------------------------------------------------------------------------------
# Model function with restoring parameter vector to full vector
# ------------------------------------------------------------------------------------------------------------------
IRF.Gauss.Full <- function(parvec, cfg, pos, fixed, ident, functional)
{
      full.par <- Restore.par(parvec, pos, fixed, ident, functional)	# Restore full parameter vector
      if(cfg$model == "GRM") return(IRF.GRM.Gauss(full.par, cfg))
      return(IRF.ORM.Gauss(full.par, cfg))  
}


#===============================================================================
#
# 3. DERIVATIONS OF MODEL FUNCTIONS
#
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Numerical computation of the model matrix (IRF-Gauss model)
# ------------------------------------------------------------------------------------------------------------------
IRF.Gauss.Matrix.Num <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos    <- SavePos.par(full.par, fixed, ident)
      parvec <- Elim.par(full.par, fixed, ident)
      jacobian(IRF.Gauss.Full, parvec, cfg = cfg, pos = pos, fixed = fixed, ident = ident, functional = functional)
}

# ------------------------------------------------------------------------------------------------------------------
# Numerical derivation of the conditional ORM model function with respect to x, b and a
# ------------------------------------------------------------------------------------------------------------------
IRF.ORM.cond.dxab <- function(x, a, b)
{
   MAX.VAL <- 700
        
    k <- length(a)
    if(k == 1)          # If there are only two categories (a single threshold):  Y/N-response 
        return(IRF.ORM.cond.dxab.Y.N(x, a, b))
    
    exp.term <- cumsum(a*(x[1] - b))
    exp.term[exp.term > MAX.VAL] <- MAX.VAL
      num <-  exp(exp.term)
    denom <-  1 + sum(num)
    cuma <- cumsum(a)
    # Derivation of Numerator   
    da <- diag(num * (x[1] - b))
    db  <- diag(-num * a)
    dx  <- c(0, num * cuma / denom) 

    for(i in 2:k)
        for(j in 1:(i-1)) {
            da[i, j] <- num[i] * (x[1] - b[j])
            db[i, j] <- -num[i] * a[j]
            }
    da <- da / denom 
    db <- db / denom       
    da <- rbind(0, da)       # Insert first row
    db <- rbind(0, db)       # Insert first row
 
    #  Derivation of Denominator
    dn.a <- matrix(0, nr = k+1, nc = k)
    dn.b <- matrix(0, nr = k+1, nc = k)
    dn.x <- numeric(k+1)
    enum <- c(1, num)
    for(i in 1:(k+1)) {
        for(j in 1:k) {
           dn.a[i, j] <- - enum[i] * (sum(num[j:k]) * (x[1] - b[j]) / denom^2)
           dn.b[i, j] <-   enum[i] * (sum(num[j:k]) * a[j] / denom^2)
           dn.x[i]    <-  dn.x[i] - num[j] * cuma[j]
          }
        dn.x[i] <- dn.x[i] * (enum[i] / denom^2)
        }   
    return(cbind(dn.x + dx, dn.b + db, dn.a + da) * x[2])
}    

# ------------------------------------------------------------------------------------------------------------------
# Numerical derivation of the conditional ORM model function with respect to x, b and a in case of two Y/N response
# ------------------------------------------------------------------------------------------------------------------
IRF.ORM.cond.dxab.Y.N <- function(x, a, b)
{
   MAX.VAL <- 700
        
    exp.term <- a*(x[1] - b)
    exp.term[exp.term > MAX.VAL] <- MAX.VAL
      num <-  exp(exp.term)
    denom <-  (1 + num)^2
    
    # Derivation of Numerator   
    dx  <-        a  * num / denom 
    db  <-       -a  * num / denom
    da <- (x[1] - b) * num / denom
    return(cbind(c(-dx, dx) * x[2], c(-db, db) * x[2], c(-da, da) * x[2]))
}

    
# ------------------------------------------------------------------------------------------------------------------
# Numerical derivation of the unconditional GRM model function with respect to a
# ------------------------------------------------------------------------------------------------------------------
IRF.ORM.uc.dxab <- function(a, b, mu, sigma)
{   
    y <- GH.Structure[, 1] 
    GH.Structure[, 1] <- GH.Structure[, 1] * sigma + mu
    for(i in 1:nrow(GH.Structure)) {
        x <- unlist(GH.Structure[i, ])
        if(i == 1) {
            d.xab <- IRF.ORM.cond.dxab(x, a, b)
            d.sigma <- d.xab[,1] * y[i]
            }
        else {
            dd.xab <- IRF.ORM.cond.dxab(x, a, b)
            d.xab <- d.xab +  dd.xab
            d.sigma <- d.sigma + dd.xab[,1] * y[i]
            }
    }   
    k <- length(a)
    if(k == 1) return(cbind(d.xab[, 1], d.sigma, d.xab[, 2:ncol(d.xab)]))
    
    dpar <- cbind(d.xab[, 1], d.sigma, d.xab[, 2:ncol(d.xab)])
    dpar[k + 1, ] <- -colSums(dpar[1:k, ])
    return(dpar)
}     
    
    
# ------------------------------------------------------------------------------------------------------------------
# Numerical derivation of the unconditional GRM model function with respect to a
# ------------------------------------------------------------------------------------------------------------------
IRF.GRM.uc.da <- function(a, b, mu, sigma)
{   
    MAXVAL <- 700

    GH.Structure[,1] <- GH.Structure[,1] * sigma + mu
    for(i in 1:nrow(GH.Structure)) {
        x <- unlist(GH.Structure[i, ])
 
        if(i == 1) { if(all(abs(a*(b-x[1])) < MAXVAL)) da <- - (x[2]*(x[1] - b) * exp(a*(b-x[1])) / (1 + exp(a * (b - x[1])))^2) else da <- 0 }
        else       { if(all(abs(a*(b-x[1])) < MAXVAL)) da <- da - (x[2]*(x[1] - b) * exp(a*(b-x[1])) / (1 + exp(a * (b - x[1])))^2) }
    }    
    if(length(da) == 1) return(matrix(c(da, -da), nc = 1))

    da <- diag(da)
    n <- nrow(da)
    
    for(i in 1:(n-1)) da[i+1, i] <- -da[i, i]
    rbind(da, c(rep(0, n-1), -da[n,n]))
}       
  
# ------------------------------------------------------------------------------------------------------------------
# Numerical derivation of the unconditional GRM model function with respect to a
# ------------------------------------------------------------------------------------------------------------------
IRF.GRM.uc.db <- function(a, b, mu, sigma)
{   
    MAXVAL <- 700
    
    GH.Structure[,1] <- GH.Structure[,1] * sigma + mu
    for(i in 1:nrow(GH.Structure)) {
        x <- unlist(GH.Structure[i, ])
        if(i == 1)  { if(all(abs(a*(b-x[1])) < MAXVAL)) db <- x[2] * a * exp(a*(b-x[1])) / (1 + exp(a * (b - x[1])))^2 else db <- 0 }
        else        { if(all(abs(a*(b-x[1])) < MAXVAL)) db <- db + x[2] * a * exp(a*(b-x[1])) / (1 + exp(a * (b - x[1])))^2   }
    }    

    if(length(db) == 1) return(matrix(c(db, -db), nc = 1))

    db <- diag(db)
    n <- nrow(db)

    for(i in 1:(n-1)) db[i+1, i] <- -db[i, i]
    rbind(db, c(rep(0, n-1), -db[n,n]))
}       
   
# ------------------------------------------------------------------------------------------------------------------
# Numerical derivation of the unconditional GRM model function with respect to mu and sigma
# ------------------------------------------------------------------------------------------------------------------
IRF.GRM.uc.dx <- function(a, b, mu, sigma)
{   
    MAXVAL <- 700

    y <- unlist(GH.Structure[,1])
    GH.Structure[,1] <- GH.Structure[,1] * sigma + mu
    for(i in 1:nrow(GH.Structure)) {
        x <- unlist(GH.Structure[i, ])
        if(i == 1)  { if(all(abs(a*(b-x[1])) < MAXVAL)) {
                        dsigma <- -y[i] * x[2] * a * exp(a*(b-x[1])) / (1 + exp(a * (b - x[1])))^2 
                        dmu    <-        -x[2] * a * exp(a*(b-x[1])) / (1 + exp(a * (b - x[1])))^2 
                        }
                       else dsigma <- dmu <- 0 
                    }
        else        { 
                      if(all(abs(a*(b-x[1])) < MAXVAL)) {
                          dsigma <- dsigma - y[i] * x[2] * a * exp(a*(b-x[1])) / (1 + exp(a * (b - x[1])))^2   
                          dmu    <- dmu    -        x[2] * a * exp(a*(b-x[1])) / (1 + exp(a * (b - x[1])))^2   
                          }
                    }
    }    
    cbind(c(dmu, 0) - c(0, dmu), c(dsigma, 0) - c(0, dsigma))
}       

# ------------------------------------------------------------------------------------------------------------------
# Numerical derivation (Jacobian matrix) of the ORM model function
# ------------------------------------------------------------------------------------------------------------------
IRF.ORM.dGauss  <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
    n.Gauss <- as.integer(2 * cfg$n.sdt)                      # number of parameters of the Gaussian models
    n.th    <- as.integer((length(parvec) - n.Gauss) / 2)     # number of thresholds and discrimination parameters
    b <- parvec[(n.Gauss+1):(n.Gauss + n.th)]                 # extract thresholds
    a <- parvec[(n.Gauss + n.th + 1):(n.Gauss + n.th + n.th)] # extract discrimination parameters
    
    col.blk.vor <- NULL                       # Column-block of zeros
    col.blk.nach <- matrix(0, nr = n.th + 1, nc = 2 * (cfg$n.sdt - 1))
    dGauss  <- NULL
    for(i in 1:cfg$n.sdt) {
       d.abx <- IRF.ORM.uc.dxab(a, b, mu = parvec[2 * (i-1) + 1], sigma = parvec[2 * (i-1) + 2])
       dGauss <- rbind(dGauss, cbind(col.blk.vor,
                                     d.abx[, 1:2],
                                     col.blk.nach,
                                     d.abx[, 3:(2 + n.th + n.th)]
                                    )
                      )
       if(i < cfg$n.sdt) {
          col.blk.vor  <- cbind(col.blk.vor, matrix(0, nr = n.th + 1, nc = 2))
          col.blk.nach <- col.blk.nach[, -(1:2)]
          }
       }
    return(dGauss)
}

# ------------------------------------------------------------------------------------------------------------------
# Numerical derivation (Jacobian matrix) of the GRM model function
# ------------------------------------------------------------------------------------------------------------------
IRF.GRM.dGauss  <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
    n.Gauss <- as.integer(2 * cfg$n.sdt)                      # number of parameters of the Gaussian models
    n.th    <- as.integer((length(parvec) - n.Gauss) / 2)     # number of thresholds and discrimination parameters
    b <- parvec[(n.Gauss+1):(n.Gauss + n.th)]                 # extract thresholds
    a <- parvec[(n.Gauss + n.th + 1):(n.Gauss + n.th + n.th)] # extract discrimination parameters
    
    col.blk.vor <- NULL                       # Column-block of zeros
    col.blk.nach <- matrix(0, nr = n.th + 1, nc = 2 * (cfg$n.sdt - 1))
    dGauss  <- NULL
    for(i in 1:cfg$n.sdt) {
       dGauss <- rbind(dGauss, cbind(col.blk.vor,
                                     IRF.GRM.uc.dx(a, b, mu = parvec[2 * (i-1) + 1], sigma = parvec[2 * (i-1) + 2]),
                                     col.blk.nach,
                                     IRF.GRM.uc.db(a, b, mu = parvec[2 * (i-1) + 1], sigma = parvec[2 * (i-1) + 2]),
                                     IRF.GRM.uc.da(a, b, mu = parvec[2 * (i-1) + 1], sigma = parvec[2 * (i-1) + 2])
                                    )
                      )
       if(i < cfg$n.sdt) {
          col.blk.vor  <- cbind(col.blk.vor, matrix(0, nr = n.th + 1, nc = 2))
          col.blk.nach <- col.blk.nach[, -(1:2)]
          }
       }
    return(dGauss)
}

# ------------------------------------------------------------------------------------------------------------------
# Model (Jacobian matrix) of the ORM and GRM model function including the restrictions
# ------------------------------------------------------------------------------------------------------------------
IRF.Gauss.Model.Matrix <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos      <- SavePos.par(full.par, fixed, ident)
      parvec   <- Elim.par(full.par, fixed, ident)
      full.par <- Restore.par(parvec, pos, fixed, ident, functional)	# Restore full parameter vector

      if(cfg$model == "GRM") J <- IRF.GRM.dGauss(full.par, cfg)
      else                   J <- IRF.ORM.dGauss(full.par, cfg)
      if(!is.null(ident))
      {
            J.Id <- diag(length(full.par))
            Index <- matrix(c(ident[2,], ident[1,]), nc = 2)
            J.Id[Index] <- 1
            J <- J %*% J.Id
      }
      
      # Treatment of functional constraints
      if(!is.null(functional))
      {
            J.fct <- attr(functional, "SDT.Jacobian")
            if(is.null(J.fct)) 
                  return(NULL)           # No Jacobian function supplied -> return (NULL)
            J <- J %*% J.fct(full.par)
      }
      J[, pos]
}


# ------------------------------------------------------------------------------------------------------------------
# Gradient of logL for the SDT-Rank model
# Input:	Same as for Likelihood function
# Output:	Gradient
# ------------------------------------------------------------------------------------------------------------------
IRF.Gauss.Grad <- function(par, data, Model, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
	parvec   <- Restore.par(par, pos, fixed, ident, functional)	# Restore full parameter vector
	if(n$model == "GRM") 
    	gradient <- as.vector(SDT.logL.Jacobi(parvec, data, Model, n, pos, fixed, ident, functional) %*% IRF.GRM.dGauss(parvec, n)) 
  else  	
      gradient <- as.vector(SDT.logL.Jacobi(parvec, data, Model, n, pos, fixed, ident, functional) %*% IRF.ORM.dGauss(parvec, n)) 

	# Treat functional constraints
	if(!is.null(functional))
	{
		gr.fct <- attr(par, "SDT.gradient")		# Isolate gradient function
		if(is.null(gr.fct))	stop("\n\n IRF.Gauss.Grad: Error in computation of gradient: No gradient for functional constraints found:\n Set Parameter sym.gr = F in function SDT.Estimate !\n\n")
		gradient <- gr.fct(gradient)
	}	
	Elim.par.grad(gradient, fixed, ident)
}

#===============================================================================
#
# 4. FUNCTIONS FOR FORMATING THE OUTPUT
#
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Format free parameters and compute confidence intervals
# Input:	1. px2 Matrix with parameters and standard errors
#		      2. Full parameter vector
#         3. Model identification string
#		      4. Number of signals
#		      5. Matrix of fixed constraints
#		      6. Matrix of equality constraints
#		      7. Confidence level
#         8. Number of decimal places
#
# Output: px5 Matrix of parameters, standard errors and confidence limits
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Par.est.IRF.Gauss <- function(Pars, full.par, Model.Id, cfg, fixed, ident, alpha, deci)
{
      qt.cfi <- qnorm((1-alpha)/2)
      cfi.lo <- Pars[,1] + Pars[,2]*qt.cfi
      cfi.up <- Pars[,1] - Pars[,2]*qt.cfi
            
      bd.lo <- Elim.par(IRF.Gauss.par.bounds(full.par, cfg)$lower, fixed, ident)
      cfi.lo <- ifelse(cfi.lo < bd.lo, bd.lo, cfi.lo)
       
        
      Pars <- cbind(Pars, cfi.lo, cfi.up)
      colnames(Pars) <- c("Value", "SE", paste("CFI-", alpha*100, "(Lower)", sep = ""), paste("CFI-", alpha*100, "(Upper)", sep = ""))
      round(Pars, deci)
}

# ------------------------------------------------------------------------------------------------------------------
# Format fuill parameter vector (in fundtion SDT.Statistics
# Input:	1. Full parameter vector
#         2. Configuration information
#		      3. Number of decimal places
#
# Output: Matrix of parameters for each GAussian model
# ------------------------------------------------------------------------------------------------------------------
IRF.Gauss.Format.Par <- function(full.par, cfg, deci)
{
    n       <- cfg$n.sdt              # Number of models     
    len.p   <- length(full.par)       # Length of parameter
    n.t     <- (len.p - 2 * n)/2      # Number of threshold parameters
    start.t <- 2 * n + 1              # Start of threshold parameter
    end.t   <- start.t - 1 + n.t      # End of threshold parameter
   
   
	  th <- round(full.par[start.t:end.t], deci)	   # Isolate thresholds
    a  <- round(full.par[(end.t+1):len.p], deci)	 # Isolate slope parameter
	
    Fp.Gauss <- matrix(round(full.par[1:(2*n)], deci), nc = n)
    Fp.Resp  <- cbind(th, a)

  	colnames(Fp.Gauss) <- paste("Gauss-", 1:n, sep = "")
	  rownames(Fp.Gauss) <- c("Mean", "Stddev")

    colnames(Fp.Resp) <- c("Difficulty", "Discrimination")
	  rownames(Fp.Resp) <- paste(1:n.t)

	  return(list(Gauss.Parameters = Fp.Gauss, Response.Parameters = Fp.Resp)) 
}

# ------------------------------------------------------------------------------------------------------------------
# Generation of data labels in output
# Input:    cfg     = configuration information
#           n.dat   = length of the data vector
#
# Output:   Vector with Data names
# ------------------------------------------------------------------------------------------------------------------
IRF.Gauss.Datalabels <- function(cfg, n.dat)
{
      n.cat <- n.dat / cfg$n.sdt          # Number of response categories
      label.name <- paste("Signal", 1:cfg$n.sdt)
      label.out <- NULL
      for(i in 1:cfg$n.sdt)
            label.out <- c(label.out, paste(label.name[i], " [", 1:n.cat, "]", sep = ""))
      return(label.out)
}


#===============================================================================
#
# 5. COMPUTATION OF SDT MEASURES
#
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Computation of SDT measures
# Input:	1. Full parameter vector
#		      2. Covariance matrix
#		      3. Configuration list
#		      4. Matrix of fixed constraints
#		      5. Matrix of equality constraints
#         6. Function implmenting functional constraints
#         7. Number of decimal places
#		      8. Confidence level
#
# Output: The follwong measures are computed:
#           d.a
#           d.e
#           A.z
#
# Measures are computed only, if first signal is fixed to N(0,1)
# ------------------------------------------------------------------------------------------------------------------
IRF.SDT.measures <- function(full.par, Cov, cfg, fixed, ident, functional, deci, conf)
{
      # Assumes that the parameters of the first signal are fixed
      if(full.par[1] != 0 || full.par[2] != 1 || !all(1:2 %in% fixed[2,]))
        return(NULL)

      n <- cfg$n.sdt
      # In case of functional constraints present and possible parameters are fixed 
      # and therefore in functional then do not compute anything
      if(!is.null(functional) & any((3:(2*n)) %in% fixed[2,]))
        return(NULL)
        
      qt.cfi <- qnorm((1-conf)/2)        

      #-----------------------------------------
      # Functions for computing d.a, d.e and A.z
      #-----------------------------------------
      SDT.d.a <- deriv(~mu*sqrt(2)/sqrt(1 + std^2), c("mu", "std"), function(mu, std) NULL)
      SDT.d.e <- deriv(~mu*2/(1 + std), c("mu", "std"), function(mu, std) NULL)
      SDT.A.z <- deriv(~pnorm(mu/sqrt(1 + std^2)), c("mu", "std"), function(mu, std) NULL)

      d.e.Erg <- d.a.Erg <- A.z.Erg <- NULL
      
      index.red <- Elim.par(1:length(full.par), fixed, ident)     # Positions of existing aprameters
      index     <- 1:length(index.red)

      k <- 3                  # start with the third parameter (since 1. and 2. are fixed
      for(sig in 2:n)         # Loop over signals
      {
            d.a <- SDT.d.a(full.par[k], full.par[k+1])
            d.e <- SDT.d.e(full.par[k], full.par[k+1])
            A.z <- SDT.A.z(full.par[k], full.par[k+1])
            
            d.a.vec <- d.a
            d.e.vec <- d.e
            A.z.vec <- A.z
                         
            
            # Extract relevant part from covariance matrix
            if(!is.null(Cov))
            {
                  CC <- IRF.Compute.CC.measures(sig, Cov, index.red, index, fixed, ident)

                  # Extract gradients
                  d.a.Grad <- matrix(attr(d.a, "gradient"), nc = 1)
                  d.e.Grad <- matrix(attr(d.e, "gradient"), nc = 1)
                  A.z.Grad <- matrix(attr(A.z, "gradient"), nc = 1)
            
                  # Delta method
                  s.d.a  <- sqrt(t(d.a.Grad) %*% CC %*% d.a.Grad)
                  s.d.e  <- sqrt(t(d.e.Grad) %*% CC %*% d.e.Grad)         
                  s.A.z  <- sqrt(t(A.z.Grad) %*% CC %*% A.z.Grad)
                  
                  # CI
                  d.a.lo <- d.a + s.d.a*qt.cfi
                  d.e.lo <- d.e + s.d.e*qt.cfi
                  A.z.lo <- max(A.z + s.A.z*qt.cfi, 0)
                  
                  d.a.up <- d.a - s.d.a * qt.cfi
                  d.e.up <- d.e - s.d.e * qt.cfi
                  A.z.up <- min(A.z - s.A.z * qt.cfi, 1)
                  
                  d.a.vec <- c(d.a.vec, s.d.a, d.a.lo, d.a.up)
                  d.e.vec <- c(d.e.vec, s.d.e, d.e.lo, d.e.up)
                  A.z.vec <- c(A.z.vec, s.A.z, A.z.lo, A.z.up)
            }
            
            d.a.Erg <- rbind(d.a.Erg, d.a.vec)
            d.e.Erg <- rbind(d.e.Erg, d.e.vec)
            A.z.Erg <- rbind(A.z.Erg, A.z.vec)
                  
            k <- k + 2
      }

      # Bind results together and provide row and column names
      Erg <- rbind(round(d.a.Erg, deci), round(d.e.Erg, deci), round(A.z.Erg, deci))
      rownames(Erg) <- c(paste("d.a[", 2:n, "]", sep=""), paste("d.e[", 2:n, "]", sep=""), paste("A.z[", 2:n, "]", sep=""))
      if(!is.null(Cov))
            colnames(Erg) <- c("Value", "SE", paste("CFI-", conf*100, "(Lower)", sep = ""), paste("CFI-", conf*100, "(Upper)", sep = ""))
      else
            colnames(Erg) <- c("Value")
      
      Erg            
}
# ------------------------------------------------------------------------------------------------------------------
# Computation of the relevant Covariance matrix for computing SDT measures
# Input:	1. Actual signal
#	        2. Covariance matrix
#		      3. Vector with numbers of free parameters
#		      4. Index for addressing items in 3
#         5. Matrix iwth fixed constraints
#		      6. Matrix of equality constraints
#
# Output: The relevant (part of the) covariance matrix
# ------------------------------------------------------------------------------------------------------------------
IRF.Compute.CC.measures <- function(sig, Cov, index.red, index, fixed, ident)
{
      k <- 2*(sig-1) + 1
      C.rel <- k:(k+1)

      # Case 1: No fixed or equality constraints
      if(is.null(ident) & is.null(fixed))
            return(Cov[C.rel, C.rel])
      
      # Treat equality constraints
      id <- C.rel %in% ident[2,]
      if(any(id)) C.rel[id] <- ident[1, ident[2,] %in% C.rel]
      
      C.rel.sich <- C.rel
      C.rel <- index[index.red %in% C.rel]
      if(length(C.rel) == 2)                     # Both parameters are free
            return(Cov[C.rel, C.rel])
      if(length(C.rel) == 0)                     # Both parameters are fixed (identity constraints)
            return(matrix(0, nr = 2, nc = 2))
 
      CC <- matrix(0, nr = 2, nc = 2)
      C.rel.t <- (1:2)[C.rel.sich %in% index.red]
      CC[C.rel.t, C.rel.t] <- Cov[C.rel, C.rel]

      CC     
}