# ===========================================================================
# MIX.2 model: Mixture model consisting of two mixtures per signal
# Date of creation:	 July, 2009
# Date of last update:	July, 2010
#
# Author: Siegfried Macho
# ===========================================================================

MIX2.Tol <- 1e-12
# ----------------------------------------------------------------------
# Function for creating response distribution for on Type of Stimulus
# Input:	Vector of parameters with the following entries:
#		1. p 	= mixing probability
#		2. m1 = location parameter of the first normal distribution
#		3. s1 = scale parameter of the first normal distribution
#		4. m2 = location parameter of the second normal distribution
#		5. s2 = scale parameter of the second normal distribution
#		6. t[1], t[2], ...., t[n-1] decision bounds
#
# Output: 	n - Vector, with n being the number of Response categories
# ----------------------------------------------------------------------
MIX2.Probs <- function(par)
{
	# Compute Probabilities for the first distribution	
	Resp <- pnorm(par[6:length(par)], mean = par[2], sd = max(par[3], MIX2.Tol))	# D1: Probability of Response-Category 2 to n-1
	Resp <- c(Resp, 1)										# Response vector
	R1 <- c(Resp[1], Resp[-1]-Resp[-length(Resp)])						# Compute Differences

	# Compute Probabilities for the second distribution	
	Resp <- pnorm(par[6:length(par)], mean = par[4], sd = max(par[5], MIX2.Tol))	# D2: Probability of Response-Category 2 to n-1
	Resp <- c(Resp, 1)										# Response vector
	R2 <- c(Resp[1], Resp[-1]-Resp[-length(Resp)])						# Compute Differences

	# Compute mixture
	par[1]*R1 + (1-par[1])*R2
}				

# ------------------------------------------------------------------------------------------------------------------
# Computation of flag, indicating for which signals the probability parameter is free
# Output: vector of Flags: 	T = parameter is not fixed for this signal
#					F = parameter is fixed
# ------------------------------------------------------------------------------------------------------------------
MIX2.free.p <- function(par, n, fixed)
{
	if(is.null(fixed)) return(rep(TRUE, n))

	l.p 	<- length(par)
	n.p 	<- l.p/n								# Number of parameter per model
	pos.p <- (1:l.p)[((1:l.p) %% n.p == 1)]				# Position of probability parameters in vector
	fixed.p <- intersect(pos.p, fixed[2,])				# Positions of fixed prob. parameters
	if(is.null(fixed.p)) return(rep(TRUE, n))
	
	!((1:n) %in% (as.integer(fixed.p / n.p) + 1)) 			# Boolean vector indicating signals with fixed probs
	
}

# ------------------------------------------------------------------------------------------------------------------
# Computation of probabilities for not fixed probability parameters
# Output: vector of Flags: 	T = parameter is not fixed for this signal
#					F = parameter is fixed
# ------------------------------------------------------------------------------------------------------------------
MIX2.cmp.p <- function(par, n, fixed)
{
	par.m <- matrix(par, nr = n, byrow = T)					# Transform to matrix
	free.p <- MIX2.free.p(par, n, fixed)
	par.m[free.p, 1] <- exp(par.m[free.p, 1]) / (1 + exp(par.m[free.p, 1])) # Compute probabilities
	as.vector(t(par.m))
}

# ------------------------------------------------------------------------------------------------------------------
# Computation of probabilities for Target and Related items for the Conjoint Recognition Model of Stahl & Klauer (2008)
# Input:	par = Vector of parameters, consisting of the following entries:
#		For each signal distribution:
#		1. p 	= mixing probability
#		2. m1 = location parameter of the first normal distribution
#		3. s1 = scale parameter of the first normal distribution
#		4. m2 = location parameter of the second normal distribution
#		5. s2 = scale parameter of the second normal distribution
#		6. t[1], t[2], ...., t[n-1] decision bounds
#
#		fixed = matrix of fixed constraints
#		ident = matrix of identity constraints (used for compatibility only)
#
# Output: Vector of probabilities for each of the distributions
#
# Comment:
# The function treats raw probability parameters as well as fixed parameters that are constraint between [0, 1]
# Only parameters not fixed are computed from raw parameters
# ------------------------------------------------------------------------------------------------------------------
SDT.MIX2 <- function(par, n, fixed = NULL, ident = NULL)
{
	if(length(n) > 1)	stop("\nFATAL ERROR: Number of signal distributions must not be a vector\n")

	par.m <- matrix(par, nr = n, byrow = T)					# Transform to matrix
	free.p <- MIX2.free.p(par, n, fixed)
	par.m[free.p, 1] <- exp(par.m[free.p, 1]) / (1 + exp(par.m[free.p, 1])) # Compute probabilities
	as.vector(apply(par.m, 1, MIX2.Probs))
}


# ------------------------------------------------------------------------------------------------------------------
# Simplified version of SDT.MIX2() that assumes that the pobabilites have been set
# ------------------------------------------------------------------------------------------------------------------
SDT.MIX2.P <- function(par, n, fixed = NULL, ident = NULL)
{
	if(length(n) > 1)	stop("\nFATAL ERROR: Number of signal distributions must not be a vector\n")

	par.m <- matrix(par, nr = n, byrow = T)					# Transform to matrix
	as.vector(apply(par.m, 1, MIX2.Probs))
}


# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix of a single signal
# Input: 	1. Parametervector for the signal
#		2. A flag, indicating, whether the parmeter is free or not
# ------------------------------------------------------------------------------------------------------------------
MIX2.Probs.Jacobi <- function(par, free.p)
{
	l   <- length(par)			# Number of parameters
	n.t <- l-5					# Number of thresholds

	if(free.p)  p <- exp(par[1]) / (1 + exp(par[1]))
	else 		p <- par[1]		
	m1  <- par[2]
	sd1 <- max(MIX2.Tol, par[3])
	m2  <- par[4]
	sd2 <- max(MIX2.Tol, par[5])
	
	t   <- par[6:l]
	z1  <- (t-m1)/sd1
	z2  <- (t-m2)/sd2
	fn1 <- dnorm(z1)
	fn2 <- dnorm(z2)
	
	# Derivative with respect to probability parameter p in [0, 1]
	dp <- pnorm(t, m1, sd1) - pnorm(t, m2, sd2)
	dp <- c(dp[1], dp[-1] - dp[-n.t], -dp[n.t])
	if(free.p) dp <- dp*p*(1-p)

	# Derivative with respect to m1
	dm1 <- -fn1/sd1							# Derivative of Gaussian with respect to m1
	dm1 <- p*c(dm1[1], dm1[-1] - dm1[-n.t], -dm1[n.t])

	# Derivative with respect to sd1
	ds1 <- -z1*fn1/sd1						# Derivative of Gaussian with respect to sd1
	ds1 <- p*c(ds1[1], ds1[-1] - ds1[-n.t], -ds1[n.t])

	# Derivative with respect to m2 
	dm2 <- -fn2/sd2							# Derivative of Gaussian with respect to m2
	dm2 <- (1-p)*c(dm2[1], dm2[-1] - dm2[-n.t], -dm2[n.t])

	# Derivative with respect to sd2
	ds2 <- -z2*fn2/sd2						# Derivative of Gaussian with respect to sd2
	ds2 <- (1-p)*c(ds2[1], ds2[-1] - ds2[-n.t], -ds2[n.t])


	# Derivative with respect to thresholds
	dt1 <- fn1/sd1							# Derivative of Gaussian with respect to t[i]
	dt2 <- fn2/sd2
	dt  <- p*dt1 + (1-p)*dt2

	J <- matrix(0, nr = n.t+1, nc = n.t)			# Jacobi matrix for thresholds: columns: Parameter, rows = probabilities
	r <- 1
	for(i in 1:n.t)
	{
		J[r, i] <- dt[i]
		r <- r + 1
		J[r, i] <- -dt[i]
	}
	cbind(dp, dm1, ds1, dm2, ds2, J)				# Full Jacobi matrix
}

# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix of the Gaussian Mixture Model
# ------------------------------------------------------------------------------------------------------------------
SDT.MIX2.Jacobi <- function(par, n, fixed)
{
	offset <- n.p <- length(par)/n			# Number of parameters
	
	free.p <- MIX2.free.p(par, n, fixed)		# Compute Boolean vector indicating signals with free probability parameters

	filler.r <- filler.c <- filler <- matrix(0, nr = n.p-4, nc = n.p)
	J <- MIX2.Probs.Jacobi(par[1:n.p], free.p[1])
	if(n < 2) return(J)
	for(i in 2:n)
	{
		J.m <- MIX2.Probs.Jacobi(par[(offset+1):(offset+n.p)], free.p[i])
		J <- rbind(cbind(J, filler.c), cbind(filler.r, J.m))
		if(i >= n) return( J )
		
		offset <- offset + n.p
		filler.c <- rbind(filler.c, filler)
		filler.r <- cbind(filler.r, filler)
		
	}
	J
}

# ------------------------------------------------------------------------------------------------------------------
# Gradient of logL for the MIX2 model
# Input:	Same as for Likelihood function
#
# Output:	Gradient
# ------------------------------------------------------------------------------------------------------------------
SDT.MIX2.Grad <- function(par, data, Model, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
	par 	   <- Restore.par(par, pos, fixed, ident, functional)	# Restore full parameter vector
	gradient <- as.vector(SDT.logL.Jacobi(par, data, Model, n, pos, fixed, ident, functional) %*% SDT.MIX2.Jacobi(par, n, fixed)) 

	# Treat functional constraints
	if(!is.null(functional))
	{
		gr.fct <- attr(par, "SDT.gradient")		# Isolate gradient function
		if(is.null(gr.fct))	stop("\n\n SDT.MIX2.Grad: Error in computation of gradient: No gradient for functional constraints found:\n Set Parameter sym.gr = F in function SDT.Estimate !\n\n")
		gradient <- gr.fct(gradient)
	}	
	Elim.par.grad(gradient, fixed, ident)
}

# ------------------------------------------------------------------------------------------------------------------
# Generate starting parameters due to data and number of signals
# Input:	   n = Number of signals
#           data = Vector of data ( in case of a single number, 
#                                   this number is interpreted as inidcating the number of response categories)
#
# Output: String with parameter names
# ------------------------------------------------------------------------------------------------------------------
SDT.MIX2.start.par <- function(n, data)
{
      n.resp <- length(data)
      if(n.resp < 2)
            n.resp <- data*n

	n.t <- as.integer(n.resp/n - 1)				      # Specify Number of thresholds
	par.m <- c(0, 0, 1, 0.5, 1, seq(-1.5, 1.5, length = n.t))	# Parameter vector of one model
	rep(par.m, n)

}

# ------------------------------------------------------------------------------------------------------------------
# Compute upper and lower bounds on parameters
# Input:	         par   = Parameter vector
#		           n   = Number of signals
#           probs.bounds   = Flag indicating whether bounds concern probabilties or raw parameters
#
# Output: A list with lower and upper bounds on parameters
# ------------------------------------------------------------------------------------------------------------------
MIX2.par.bounds <- function(par, n, probs.bounds = F)
{
      Max.Raw.P <- 300
      Min.Var   <- MIX2.Tol
      
      lower <- rep(-Inf, length(par))
      upper <- rep(Inf, length(par))
 
      n.t <- length(par)/n - 5
      p.off <- rep(5 + n.t, n - 1)
            
      pos.p  <- cumsum(c(1, p.off))      # Position of the probability parameter
      pos.sd <- c(pos.p + 2, pos.p + 4)  # Positions of stddevs
      
      lower[pos.sd] <- Min.Var            # Lower bounds for stdev
      
      if(probs.bounds)                    # Bounds for probability parameters
      {
            lower[pos.p] <- 0
            upper[pos.p] <- 1
      }
      else
      {
            lower[pos.p] <- -Max.Raw.P
            upper[pos.p] <-  Max.Raw.P
      }      
 
      list(lower = lower, upper = upper)
}

# ------------------------------------------------------------------------------------------------------------------
# Generate strings with names of the parameters for model MIX2
# Input:	par = Vector of parameters
#		  n = number of models
#
# Output: String with parameter names
# ------------------------------------------------------------------------------------------------------------------
MIX2.par.names <- function(par, n)
{
	p    <- paste("p[", 1:n, "]", sep = "")
	m1   <- paste("Mean-1[" , 1:n, "]", sep = "")
	sd1  <- paste("Stddev-1[", 1:n, "]", sep = "")
	m2   <- paste("Mean-2[" , 1:n, "]", sep = "")
	sd2  <- paste("Stddev-2[", 1:n, "]", sep = "")
	n.t <- length(par)/n - 5
	t   <- paste("t-",  1:n.t, sep = "")
	p.name   <- NULL
	for(i in 1:n) p.name <- c(p.name, p[i], m1[i], sd1[i], m2[i], sd2[i], paste(t, "[", i, "]", sep = ""))
	p.name
}

# ------------------------------------------------------------------------------------------------------------------
# Names of the models 
# ------------------------------------------------------------------------------------------------------------------
MIX2.model.name <- "Mixture Model: Mixtures of two normal distributions per signal"

# ------------------------------------------------------------------------------------------------------------------
# Format parameters for SDT and Gaussian model
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Par.MIX2 <- function(par, n, fixed, dec)
{
	# Compute probabilities for non-fixed probability parameters
	par <- MIX2.cmp.p(par, n, fixed) 	

	# Format output
	n.t <- length(par)/n - 5
	Fp <- matrix(round(par, dec), nc = n)
	
	colnames(Fp) <- paste("Mixture-", 1:n, sep = "")
	rownames(Fp) <- c("p", "Mean-1", "Stddev-1",  "Mean-2", "Stddev-2", paste("t-", 1:n.t, sep = ""))
	Fp
}

# ------------------------------------------------------------------------------------------------------------------
# Format free free parameters for MIX2 model and adapt standard errors
# Input:	1. px2 Matrix with parameters and standard errors
#		2. full parameter vector
#		3. data vector
#		4. number of signals
#		5. matrix of fixed constraints
#		6. matrix of equality constraints
#		7. function implementing functional constraints
#           8. confidecne level
#           9. number of decimal places
#
# Output: p x 5 Matrix of parameters, standard errors, and confidence intervals
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Par.est.MIX2 <- function(Pars, full.par, data, n, fixed, ident, functional, alpha, deci)
{
	pos.free <- SavePos.par(full.par, fixed, ident)		# Positions of free parameters
	
	# Compute Positions of probabities in full parameter vector
	l.p 	<- length(full.par)
	n.p 	<- l.p/n							# Number of parameter per model
	pos.p <- (1:l.p)[((1:l.p) %% n.p == 1)]			# Position of probability parameters in vector
	pos   <- pos.free %in% pos.p					# Positions of probability parameters in free parameter vector

	# Compute probabilities and adapt standard errors
	Pars[pos ,1] <- exp(Pars[pos ,1]) / (1 + exp(Pars[pos ,1]))
	Pars[pos, 2] <- Pars[pos, 2] * (Pars[pos, 1]*(1-Pars[pos, 1]))

	# Compute asymptotic standard errors numerically
	Pars <- cbind(Pars, SDT.MIX2.Hessian.p(full.par, data, n, fixed = fixed, ident = ident, functional = functional))

      # Compute confidence intervals
      bd    <- MIX2.par.bounds(full.par, n, probs.bounds = T)       # Bounds on probabilities
      bd.lo <- Elim.par(bd$lower, fixed, ident)
      bd.up <- Elim.par(bd$upper, fixed, ident)

      qt.cfi <- qnorm((1-alpha)/2)
      cfi.lo <- Pars[,1] + Pars[,2]*qt.cfi
      cfi.up <- Pars[,1] - Pars[,2]*qt.cfi
            
      # Respect lower bounds on sd parameter
      cfi.lo <- ifelse(cfi.lo < bd.lo, bd.lo, cfi.lo)
      cfi.up <- ifelse(cfi.up > bd.up, bd.up, cfi.up)
      
  	# Add columns with confidence intervals and specify new column names
      Pars <- cbind(Pars, cfi.lo, cfi.up)
	colnames(Pars) <- c("Value", "SE(delta method)", "SE(numerical)", paste("CFI-", alpha*100, "(Lower)", sep = ""), paste("CFI-", alpha*100, "(Upper)", sep = ""))

	round(Pars, deci)
}


# ------------------------------------------------------------------------------------------------------------------
# MIX2 Model: Check number of parameters and data points
# ------------------------------------------------------------------------------------------------------------------
SDT.MIX2.check <- function(par, data, n)
{
	n.par  <- as.integer(length(data) + 4*n)
	n.data <- as.integer(length(par)  - 4*n)
  

	Erg <- matrix(c(length(par), length(data), n.par, n.data), nc = 2)
	rownames(Erg) <- c("Parameters", "Data points")
	colnames(Erg) <- c("Found", "Expected")

	Error <- any(Erg[, 1] != Erg[, 2])

	list(Error = Error, Stat = Erg, Msg = "NOTE: Computation of the expected number of parameters is based on the to number of data points and vice versa")
}

# ------------------------------------------------------------------------------------------------------------------
# Functions for computing Hessian numerically
# ------------------------------------------------------------------------------------------------------------------
SDT.MIX2.Hessian.p <- function(par, data, n, fixed = fixed, ident = ident, functional = functional)
{
	par.p   <- MIX2.cmp.p(par, n, fixed) 		# Compute probabilities

	pos.par <- SavePos.par(par, fixed, ident)		# Positions of free parameters
	par.p	  <- Elim.par(par.p,  fixed, ident)		# Set Parameter constraints
	Model   <- SDT.MIX2.P
	hess    <- hessian(func = SDT.logL, x = par.p, data = data, Model = Model, n = n, pos = pos.par, fixed = fixed, ident = ident, functional = functional)

	if(is.null(hess) || is.na(hess) || is.nan(hess)) return(rep(NA, length(par.p))) 	# Hessian absent or not OK -> retuen (NA)

	Erg <- Test.Hessian(hess)	
	if(Erg$P.d == F)		return(rep(NA, length(par.p)))					# Not positive definite -> return(NA) 		
	Cov <- solve(hess)
	return(sqrt(diag(Cov)))
}


#===============================================================================
#
# Functions for comupting the model matrix (analytically and numerically)
# for the MIX2 model
#
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Model incorporating restrictions
# The parameter passed to the function contains the free parameters only
# ------------------------------------------------------------------------------------------------------------------
SDT.MIX2.Full <- function(rest.par, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
      full.par <- Restore.par(rest.par, pos, fixed, ident, functional)	# Restore full parameter vector
      SDT.MIX2(full.par, n, fixed, ident)
}

# ------------------------------------------------------------------------------------------------------------------
# Numerical computation of the model matrix 
# ------------------------------------------------------------------------------------------------------------------
SDT.MIX2.Model.Matrix.Num <- function(full.par, n, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)
      par <- Elim.par(full.par, fixed, ident)
      J <- jacobian(SDT.MIX2.Full, par, n = n, pos = pos, fixed = fixed, ident = ident, functional = functional)
}

# ------------------------------------------------------------------------------
# Analytical computation of the model matrix 
# Input:    1. Full parameter vector
#           2. Number of signals
#           3. 2 x p - matrix of fixed constraints
#           4. 2 x q - matrix of identity constraints
#           5. Jacobian of functional constraints (supplied by the user)
#
# Output:   Modelmatrix
#
# Comment:  The Modelmatrix of funtional constraints must be a full matrix that
#           is premultiplied by the Jacobian
# ------------------------------------------------------------------------------
SDT.MIX2.Model.Matrix <- function(full.par, n, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)

      J   <- SDT.MIX2.Jacobi(full.par, n = n, fixed)
      if(!is.null(ident))
      {
            J.Id <- diag(length(full.par))
            Index <- matrix(c(ident[2,], ident[1,]), nc = 2)
            J.Id[Index] <- 1
            J <- J %*% J.Id
      }
      
      # Treatment of funtional constraints
      if(!is.null(functional))
      {
            J.fct <- attr(functional, "SDT.Jacobian")
            if(is.null(J.fct)) 
                  return(NULL)           # No Jacobian function supplied -> return (NULL)
            J <- J %*% J.fct(full.par)
      }
      J[, pos]
}

#===============================================================================
#
# Plot function:
#
#===============================================================================


# ------------------------------------------------------------------------------------------------------------------
# Plotting for the MiX.2 model
# Input:	1. par.vec = vector of parameters
#		2. n      = number of models
#           3. cols   = number of colors
#           4. ltys   = line types
#		5. labels = labels for axes
#           6. SDT.legend = list with tow entries: 
#                       pos  = c(x, y) Position
#                       text = c("text 1",...,"text n")
#                       cols = colors
#           7. option:  NULL       --> SDT
#                       "ROC"      --> ROC
#                       "zROC"     --> zROC
#                       "ROC+"     --> ROC + Data
#                       "zROC+"    --> zROC + Data
#           8. counts: data     
#
# Comment:  The function extracts the decision bounds as well as the means and sd parameters and calls the funtion
#           SDT.SDT1D.Plot or SDT.ROC.Plot.
# ------------------------------------------------------------------------------------------------------------------
SDT.MIX.2.Plot <- function(par.vec, n, cols, ltys, labels, SDT.legend, option, counts)
{
      Mix.par <- matrix(par.vec, nc = n)
      m  <- as.vector(Mix.par[c(2,4),])
      sd <- as.vector(Mix.par[c(3,5), ])
      t  <- as.vector(Mix.par[-c(1:5),])
      p  <- as.vector(Mix.par[1,])
      p  <- exp(p) / (1 + exp(p))            # Compute p from raw parameters
      
      if(is.null(option))
            SDT.SDT1D.Plot(m, sd, t, n = 2*n, cols, ltys, labels, SDT.legend)
      else if(option == "ROC")
            SDT.ROC.Plot(m, sd, p = p, n, cols, ltys, labels, SDT.legend, zROC = F, counts = NULL)
      else if(option == "ROC+")
            SDT.ROC.Plot(m, sd, p = p, n, cols, ltys, labels, SDT.legend, zROC = F, counts = counts)
      else if(option == "zROC")
            SDT.ROC.Plot(m, sd, p = p, n, cols, ltys, labels, SDT.legend, zROC = T, counts = NULL)
      else if(option == "zROC+")
            SDT.ROC.Plot(m, sd, p = p, n, cols, ltys, labels, SDT.legend, zROC = T, counts = counts)
      else
            cat("\n\n --> SDT.MIX.2.Plot: Unknown option: \"", option, "\"\n\n", sep = "")
}
 