# ========================================================================================
# MIX-PD model: Mixture models consisting of three Gaussians for a pair of signals.
#		  1. Only one set of thresholds
#		  2. Origin fixed automatically
#		  3. Additional foil distribution is possible
#
# Date of creation:	August, 2009
# Date of last update:	July,   2010
#
# Author: Siegfried Macho
# ========================================================================================

MIX.Tol <- 1e-12
# ----------------------------------------------------------------------
# Function for creating response distribution for on Type of Stimulus
# Input:	Vector of parameters with the following entries:
#		1.  m = Location of basic distribution
#		2. sd = Scale of distribtuions
#		3. p1	= first mixing probability
#		4. p2 = second mixing probability
#		5. d1 = displacment of 'left' signal distribution
#		6. d2 = displacment of 'right' signal distribution
#		7. t[1], t[2], ...., t[n-1] decisions bounds
#
# Output: 	n - Vector, with n being the number of Response categories
# ----------------------------------------------------------------------
MIX.Probs <- function(par)
{
	# 1. Compute Probabilities for the basic distribution	
	Resp <- pnorm(par[7:length(par)], mean = par[1], sd = max(par[2], MIX.Tol))		# D.0: Probability of Response-Category 1 to n-1
	Resp <- c(Resp, 1)											# Response vector
	R.0 <- c(Resp[1], Resp[-1]-Resp[-length(Resp)])							# Compute Differences

	# 2. Compute Probabilities for the 'left' distribution	
	Resp <- pnorm(par[7:length(par)], mean = par[1]-par[5], sd = max(par[2], MIX.Tol))	# D.L: Probability of Response-Category 1 to n-1
	Resp <- c(Resp, 1)											# Response vector
	R.L <- c(Resp[1], Resp[-1]-Resp[-length(Resp)])							# Compute Differences

	# 3. Compute Probabilities for the 'right' distribution	
	Resp <- pnorm(par[7:length(par)], mean = par[1]+par[6], sd = max(par[2], MIX.Tol))	# D.R: Probability of Response-Category 1 to n-1
	Resp <- c(Resp, 1)											# Response vector
	R.R <- c(Resp[1], Resp[-1]-Resp[-length(Resp)])							# Compute Differences

	# Compute mixture
	c(par[3]*R.L + (1-par[3])*R.0, par[4]*R.R + (1-par[4])*R.0) 
}				

#----------------------------------------
# Probabilities of the foil distribution
#----------------------------------------
MIX.Foil.Probs <- function(par)
{
	# Compute Probabilities for the foil distribution	
	Resp <- pnorm(par[3:length(par)], mean = par[1], sd = max(par[2], MIX.Tol))	# Probability of Response-Category 1 to n-1
	Resp <- c(Resp, 1)										# Response vector
	c(Resp[1], Resp[-1]-Resp[-length(Resp)])							# Compute Differences
}				


# ------------------------------------------------------------------------------------------------------------------
# Computation of flag, indicating for which signals the probability parameter is free
# Input:  	1. Parameter vector (without thresholds
#		2. Number of signal pairs
#		3. Matrix of fixed parameters
#		4. Offset = 1. Position of relevant parameter
#
# Output: vector of Flags: 	T = parameter is not fixed for this pair of signals
#					F = parameter is fixed
# ------------------------------------------------------------------------------------------------------------------
MIX.free.p <- function(lpar, n, fixed, off)
{
	if(is.null(fixed)) return(rep(TRUE, n))

	pos.p   <- seq(0, lpar, 6) + off			# Relevant positions
	!(pos.p %in% fixed[2,])					# Boolean vector indicating free probability parameters
	
}

# ------------------------------------------------------------------------------------------------------------------
# Computation of probabilities for not fixed probability parameters
# Output: vector of Flags: 	T = parameter is not fixed for this signal
#					F = parameter is fixed
# ------------------------------------------------------------------------------------------------------------------
MIX.PD.cmp.p <- function(par, n, fixed)
{
	if(n%%2 != 0)	# Uneven number of signals -> first data points represent foil distribution
	{
		t <- par[((n-1)*3 + 1) : length(par)]	# Extract threshold parameters

		par.m <- matrix(par[1:((n-1)*3)], nr = (n-1)/2, byrow = T)	# Transform to matrix
		free.p1 <- MIX.free.p((n-1)*3-1, n/2, fixed, off = 3)
		free.p2 <- MIX.free.p((n-1)*3-1, n/2, fixed, off = 4)

		par.m[free.p1, 3] <- exp(par.m[free.p1, 3]) / (1 + exp(par.m[free.p1, 3])) # Compute probabilities
		par.m[free.p2, 4] <- exp(par.m[free.p2, 4]) / (1 + exp(par.m[free.p2, 4])) # Compute probabilities

		c(as.vector(t(par.m)), t)
		
	}
	else			# Even number of models
	{
		t <- par[(n*3 - 1) : length(par)]							# Extract threshold parameters

		free.p1 <- MIX.free.p(n*3-2, n/2, fixed, off = 1)
		free.p2 <- MIX.free.p(n*3-2, n/2, fixed, off = 2)

		par.m <- matrix(c(0, 1, par[1:(n*3 - 2)]), nr = n/2, byrow = T)		   # Inserte location and scale parameter of first pair and transform to matrix
		par.m[free.p1, 3] <- exp(par.m[free.p1, 3]) / (1 + exp(par.m[free.p1, 3])) # Compute probabilities
		par.m[free.p2, 4] <- exp(par.m[free.p2, 4]) / (1 + exp(par.m[free.p2, 4])) # Compute probabilities
	
		c(as.vector(t(par.m))[-c(1,2)], t)
	}
}

# ------------------------------------------------------------------------------------------------------------------
# Computation of Mixture of 3 Gaussian for two signals
# Input:	par = Vector of parameters, consisting of the following entries:
#
#		Two cases:
#
#		A. Without a separate foil distribution (number of models is an even number): 
#		   For each pair of signals the following parameters are presented, except for the first one two signals):
#
#		1.  m = Location of basic distribution (for the first two signals no location parameter is specified because m = 0)
#		2. sd = Scale of distributions (for the first two signals no scale parameter is specified because sd = 1)
#		3. p1	= first mixing probability
#		4. p2 = second mixing probability
#		5. d1 = displacment of 'left' signal distribution
#		6. d2 = displacment of 'right' signal distribution
#
#		B. With a separate foil distribution (numner of models is an uneven number)
#		   For each pair of signals the following parameters are presented:
#
#		1.  m = Location of basic distribution
#		2. sd = Scale of distributions
#		3. p1	= first mixing probability
#		4. p2 = second mixing probability
#		5. d1 = displacment of 'left' signal distribution
#		6. d2 = displacment of 'right' signal distribution
#
#		C. One set of threshold parameters for each model are provided:
#		   t[1], t[2], ...., t[n-1] decisions bounds
#
#		fixed = matrix of fixed constraints
#		ident = matrix of identity constraints (used for compatability only)
#
# Output: Vector of probabilities for each pair of distributions plus 
#
# Comment:
# The function treats raw probability parameters as well as fixed parameters that are constraint between [0, 1]
# Only parameters not fixed are computed from raw parameters
# ------------------------------------------------------------------------------------------------------------------
SDT.MIX.PD <- function(par, n, fixed = NULL, ident = NULL)
{
	if(length(n) > 1)	stop("\nFATAL ERROR: Number of signal distributions must not be a vector\n")

	if(n%%2 != 0)	# Uneven number of signals -> first data points represent foil distribution
	{
		t <- par[((n-1)*3 + 1) : length(par)]	# Extract threshold parameters
		Pr.Foil <- MIX.Foil.Probs(c(0, 1, t))	# Probabilities of foil distribution

		par.m <- matrix(par[1:((n-1)*3)], nr = (n-1)/2, byrow = T)	# Transform to matrix
		free.p1 <- MIX.free.p((n-1)*3-1, n/2, fixed, off = 3)
		free.p2 <- MIX.free.p((n-1)*3-1, n/2, fixed, off = 4)

		par.m[free.p1, 3] <- exp(par.m[free.p1, 3]) / (1 + exp(par.m[free.p1, 3])) 	# Compute probabilities
		par.m[free.p2, 4] <- exp(par.m[free.p2, 4]) / (1 + exp(par.m[free.p2, 4])) 	# Compute probabilities
		par.m <- cbind(par.m, matrix(rep(t, (n-1)/2), nr = (n-1)/2, byrow = T))		# Add threshold parameters

		return(c(Pr.Foil, as.vector(apply(par.m, 1, MIX.Probs))))
	}
	
	# Even number of models
	t <- par[(n*3 - 1) : length(par)]							# Extract threshold parameters

	free.p1 <- MIX.free.p(n*3-2, n/2, fixed, off = 1)
	free.p2 <- MIX.free.p(n*3-2, n/2, fixed, off = 2)

	par.m <- matrix(c(0, 1, par[1:(n*3 - 2)]), nr = n/2, byrow = T)		   # Inserte location and scale parameter of first pair and transform to matrix
	par.m[free.p1, 3] <- exp(par.m[free.p1, 3]) / (1 + exp(par.m[free.p1, 3])) # Compute probabilities
	par.m[free.p2, 4] <- exp(par.m[free.p2, 4]) / (1 + exp(par.m[free.p2, 4])) # Compute probabilities
	
	par.m <- cbind(par.m, matrix(rep(t, n/2), nr = n/2, byrow = T))		# Add threshold parameters
	as.vector(apply(par.m, 1, MIX.Probs))
}
#-------------------------------------------------------------------------------------------------------------------------------------
# Same function but without computation of probability parameters from raw parameters (assumes that these have been already computed)
#-------------------------------------------------------------------------------------------------------------------------------------
SDT.MIX.PD.Pr <- function(par, n, fixed = NULL, ident = NULL)
{
	if(length(n) > 1)	stop("\nFATAL ERROR: Number of signal distributions must not be a vector\n")

	if(n%%2 != 0)	# Uneven number of signals -> first data points represent foil distribution
	{
		t <- par[((n-1)*3 + 1) : length(par)]	# Extract threshold parameters

		Pr.Foil <- MIX.Foil.Probs(c(0, 1, t))	# Probabilities of foil distribution

		par.m <- matrix(par[1:((n-1)*3)], nr = (n-1)/2, byrow = T)	# Transform to matrix
		par.m <- cbind(par.m, matrix(rep(t, (n-1)/2), nr = (n-1)/2, byrow = T))		# Add threshold parameters

		return(c(Pr.Foil, as.vector(apply(par.m, 1, MIX.Probs))))
	}
	
	# Even number of models
	t <- par[(n*3 - 1) : length(par)]							# Extract threshold parameters

	par.m <- matrix(c(0, 1, par[1:(n*3 - 2)]), nr = n/2, byrow = T)		# Inserte location and scale parameter of first pair and transform to matrix
	par.m <- cbind(par.m, matrix(rep(t, n/2), nr = n/2, byrow = T))		# Add threshold parameters
	as.vector(apply(par.m, 1, MIX.Probs))
}


# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix of the derivatives for the foil signal
# Input: 	Parametervector for the foil signal = thresholds)
#
# Output:	d x t Matrix (d = number of data points, t = thresholds)
# ------------------------------------------------------------------------------------------------------------------
MIX.PD.Foils.Jacobi <- function(par)
{
	n.t  <- length(par)			# Number of parameters = number of thrsholds
	fn   <- dnorm(par)
	sd   <- 1.0	

	# Derivative with respect to thresholds
	dt <- fn/sd							# Derivative of Gaussian with respect to t[i]
	
	J <- matrix(0, nr = n.t+1, nc = n.t)		# Jacobi matrix for thresholds: columns: Parameter, rows = probabilities
	r <- 1
	for(i in 1:n.t)
	{
		J[r, i] <- dt[i]
		r <- r + 1
		J[r, i] <- -dt[i]
	}
	J
}

# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix of the derivatives for a single signal pair
# Input: 	1.Parametervector containing the following entries
#			1. mean
#			2. sd
#			3. p.left
#			4. p.right
#			5. d.left
#			6. d.right
#			7. thresholds (t thresholds)
#		2. Flag inidicating whether p.left  is free, i.e. a raw parameter that has to be transformed (=T)
#		3. Flag inidicating whether p.right is free, i.e. a raw parameter that has to be transformed (=T)
#
# Output:	2*(t+1) x t+6 Matrix (number of data points x parameters)
# ------------------------------------------------------------------------------------------------------------------
MIX.PD.Pairs.Jacobi <- function(par, free.p.left, free.p.right)
{
	l   <- length(par)			# Number of parameters
	n.t <- l-6					# Number of thresholds

	if(free.p.left)  	p.left  <- exp(par[3]) / (1 + exp(par[3]))
	else		   	p.left  <- par[3]

	if(free.p.right) 	p.right <- exp(par[4]) / (1 + exp(par[4]))
	else 		   	p.right <- par[4]

	m  	  <- par[1]
	sd 	  <- max(MIX.Tol, par[2])
	d.left  <- par[5]
	d.right <- par[6]
	t   	  <- par[7:l]
	m.left  <- m - d.left
	m.right <- m + d.right

	z.0  	  <- (t - m)/sd
	z.left  <- (t - m.left)/sd
	z.right <- (t - m.right)/sd
	fn.0 	  <- dnorm(z.0)
	fn.left <- dnorm(z.left)
	fn.right<- dnorm(z.right)
	
	# Derivative with respect to probability parameter p.left and p.right
	dp.left  <- pnorm(t, m.left, sd)  - pnorm(t, m, sd)
	dp.right <- pnorm(t, m.right, sd) - pnorm(t, m, sd)
	dp.left  <- c(dp.left[1] , dp.left[-1]  - dp.left[-n.t],  -dp.left[n.t])
	dp.right <- c(dp.right[1], dp.right[-1] - dp.right[-n.t], -dp.right[n.t])

	if(free.p.left)	dp.left  <- dp.left  * p.left  * (1-p.left)
	if(free.p.right)	dp.right <- dp.right * p.right * (1-p.right)

	# Derivative with respect to m
	dm.0 	   <- -fn.0/sd							# Derivative of middle Gaussians with respect to m
	dm.left  <- -fn.left/sd							# Derivative of left   Gaussians with respect to m
	dm.right <- -fn.right/sd						# Derivative of right  Gaussians with respect to m

	dm.first  <-     p.left  * c(dm.left[1]  , dm.left[-1]  - dm.left[-n.t] , -dm.left[n.t]) + 	# First signal
			  (1-p.left) * c(dm.0[1]     , dm.0[-1]     - dm.0[-n.t]    , -dm.0[n.t]) 
	dm.second <-     p.right * c(dm.right[1], dm.right[-1] - dm.right[-n.t], -dm.right[n.t]) + 	# Second signal
			  (1-p.right)* c(dm.0[1]    , dm.0[-1]     - dm.0[-n.t]    , -dm.0[n.t]) 

	# Derivative with respect to sd
	ds.0      <- -z.0     * fn.0/sd					# Derivative of middle Gaussians with respect to sd
	ds.left   <- -z.left  * fn.left/sd					# Derivative of left   Gaussians with respect to sd
	ds.right  <- -z.right * fn.right/sd					# Derivative of right  Gaussians with respect to sd

	ds.first  <-    p.left  * c(ds.left[1]  , ds.left[-1]  - ds.left[-n.t] , -ds.left[n.t]) +	# First signal
			 (1-p.left) * c(ds.0[1]     , ds.0[-1]     - ds.0[-n.t]    , -ds.0[n.t]) 
	ds.second <-    p.right * c(ds.right[1], ds.right[-1] - ds.right[-n.t], -ds.right[n.t]) +	# Second signal
			 (1-p.right)* c(ds.0[1]    , ds.0[-1]     - ds.0[-n.t]    , -ds.0[n.t]) 

	# Derivative with respect to d.left and d.right 
	dd.left  <-  fn.left/sd							# Derivative of (left)  Gaussian with respect to d.left
	dd.right <- -fn.right/sd						# Derivative of (right) Gaussian with respect to d.right
	dd.left  <- p.left  * c(dd.left[1] , dd.left[-1]  - dd.left[-n.t],  -dd.left[n.t])
	dd.right <- p.right * c(dd.right[1], dd.right[-1] - dd.right[-n.t], -dd.right[n.t])


	# Derivative with respect to thresholds
	dt.0     <- fn.0/sd							# Derivative of middle Gaussian with respect to t[i]
	dt.left  <- fn.left/sd							# Derivative of left   Gaussian with respect to t[i]
	dt.right <- fn.right/sd							# Derivative of right  Gaussian with respect to t[i]

	dt.first  <- p.left  * dt.left  + (1-p.left)  * dt.0		# Derivative with respect to the data for the first  signal
	dt.second <- p.right * dt.right + (1-p.right) * dt.0		# Derivative with respect to the data for the second signal

	J.first  <- matrix(0, nr = n.t+1, nc = n.t)			# Jacobi matrix for thresholds: columns: Parameter, rows = probabilities for first signal
	J.second <- matrix(0, nr = n.t+1, nc = n.t)			# Jacobi matrix for thresholds: columns: Parameter, rows = probabilities
	r <- 1
	for(i in 1:n.t)
	{
		J.first[r, i]  <- dt.first[i]
		J.second[r, i] <- dt.second[i]
		r <- r + 1
		J.first[r, i]  <- -dt.first[i]
		J.second[r, i] <- -dt.second[i]
	}
	fueller <- rep(0, n.t+1)
	cbind(c(dm.first, dm.second), 					# Full Jacobi matrix
		c(ds.first, ds.second), 
		c(dp.left, fueller), 
		c(fueller, dp.right), 
		c(dd.left, fueller), 
		c(fueller, dd.right), 
		rbind(J.first, J.second))				
}

# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix of the Mixture Model with 3 Gaussian distributions per signal pair
# ------------------------------------------------------------------------------------------------------------------
SDT.MIX.PD.Jacobi <- function(par, n, fixed)
{
	n.c <- 6		# Number of parameters per signal pair (except thresholds)

	if(n%%2 != 0)	# Case 1: Uneven number of signals -> first data points represent foil distribution
	{
		t 	 <- par[((n-1)*3 + 1) : length(par)]	# Extract threshold parameters
		n.t	 <- length(t)					# Number of thresholds
		n.r 	 <- 2*n.t + 2					# Length of a data block (consistion of two signals
		n.par   <- (n-1)*3
		pos.p.l <- seq(3, n.par, 6)				# Positions of left  probability parameters
		pos.p.r <- seq(4, n.par, 6)				# Positions of right probability parameters
		free.p.l  <- !(pos.p.l %in% fixed[2,])		# Boolian vector indicating positions of free left probability parameters
		free.p.r  <- !(pos.p.r %in% fixed[2,])		# Boolian vector indicating positions of free left probability parameters

		fill   <- matrix(0, nr = n.r, nc = n.c)		# Filler-Matrix
		fill.c <- matrix(0, nr = n.t+1, nc = n.c)		
		fill.r <- NULL

		J.t <- MIX.PD.Foils.Jacobi(t)				# Jacobimatrix of thresholds for foil signal
		J   <- NULL

		m <- (n-1)/2
		for(i in 1:m)
		{
			par.act <- c(par[((i-1)*6+1):((i-1)*6+6)], t)		# Parameter vector for actual pair
			J.act <- MIX.PD.Pairs.Jacobi(par.act, free.p.l[i], free.p.r[i])
			J.t 	<- rbind(J.t, J.act[, 7:(6+n.t)])

			J <- rbind(cbind(J, fill.c), cbind(fill.r, J.act[, 1:6]))
			if(i < m)
			{
				fill.c <- rbind(fill.c, fill)
				fill.r <- cbind(fill.r, fill)
			}
		}
		J <- cbind(J, J.t)					# Add derivatives of thresholds
	
	}
	else			# Case 2: Even number of signals
	{				
		t 	 <- par[(n*3 - 1) : length(par)]		# Extract threshold parameters
		n.t	 <- length(t)					# Number of thresholds
		n.r 	 <- 2*n.t + 2					# Length of a data block (consistion of two signals
		n.par  <- n*3-2
		pos.p.l <- seq(1, n.par, 6)				# Positions of left  probability parameters
		pos.p.r <- seq(2, n.par, 6)				# Positions of right probability parameters
		free.p.l  <- !(pos.p.l %in% fixed[2,])		# Boolian vector indicating positions of free left probability parameters
		free.p.r  <- !(pos.p.r %in% fixed[2,])		# Boolian vector indicating positions of free left probability parameters

		fill.c <- fill.r <- fill <- matrix(0, nr = n.r,   nc = n.c)		# Filler-Matrix

		m   <- n/2
		par <- c(0, 1, par)					# Add mean and sd for first signal pair
		for(i in 1:m)
		{
			par.act <- c(par[((i-1)*6+1):((i-1)*6+6)], t)		# Parameter vector for actual pair
			J.act <- MIX.PD.Pairs.Jacobi(par.act, free.p.l[i], free.p.r[i])
			if(i == 1)
			{
				J   <- J.act[, 1:6]
				J.t <- J.act[, 7:(6+n.t)]
			}
			else
			{
				J   <- rbind(cbind(J, fill.c), cbind(fill.r, J.act[, 1:6]))
				J.t <- rbind(J.t, J.act[, 7:(6+n.t)])
				if(i < m)
				{
					fill.c <- rbind(fill.c, fill)
					fill.r <- cbind(fill.r, fill)
				}
			}
		}
		J <- cbind(J[,-c(1,2)], J.t)		# Add derivatives of thresholds and delete derivatives for m and sd for the first pair of signals
	}
	J
}

# ------------------------------------------------------------------------------------------------------------------
# Gradient of logL for the MIX.PD model
# Input:	Same as for Likelihood function
#
# Output:	Gradient
# ------------------------------------------------------------------------------------------------------------------
SDT.MIX.PD.Grad <- function(par, data, Model, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
	par 	   <- Restore.par(par, pos, fixed, ident, functional)	# Restore full parameter vector
	gradient <- as.vector(SDT.logL.Jacobi(par, data, Model, n, pos, fixed, ident, functional) %*% SDT.MIX.PD.Jacobi(par, n, fixed)) 

	# Treat functional constraints
	if(!is.null(functional))
	{
		gr.fct <- attr(par, "SDT.gradient")		# Isolate gradient function
		if(is.null(gr.fct))	stop("\n\n SDT.MIX.PD.Grad: Error in computation of gradient: No gradient for functional constraints found:\n Set Parameter sym.gr = F in function SDT.Estimate !\n\n")
		gradient <- gr.fct(gradient)
	}	
	Elim.par.grad(gradient, fixed, ident)
}

# ------------------------------------------------------------------------------------------------------------------
# Generate starting parameters due to data and number of signals
# Input:	   n = Number of signals
#           data = Vector of data ( in case of a single number, 
#                                   this number is interpreted as inidcating the number of response categories)
#      range.th = a range for the threshold parameters
#
# Output: Parameter vector
# ------------------------------------------------------------------------------------------------------------------
SDT.MIX.PD.start.par <- function(n, data, range.th = c(-1, 1))
{
      n.resp <- length(data)
      if(n.resp < 2)
            n.resp <- data*n

	n.t <- n.resp/n - 1	   # Number of thresholds
	if(n %% 2 != 0)		   # Uneven number of signals
	{
		par.m <- c(rep(c(0, 1, 0, 0, 1, 1), (n-1)/2), seq(range.th[1], range.th[2], length = n.t))	
	}
	else					# Even number of signals
	{
		par.m <- c(rep(c(0, 1, 0, 0, 1, 1), n/2), seq(range.th[1], range.th[2], length = n.t))
		par.m <- par.m[-c(1,2)]
	}
	par.m
}

# ------------------------------------------------------------------------------------------------------------------
# Compute upper and lower bounds on parameters
# Input:	         par   = Parameter vector
#		           n   = Number of signals
#           probs.bounds   = Flag indicating whether bounds concern probabilties or raw parameters
#
# Output: A list with lower and upper bounds on parameters
# ------------------------------------------------------------------------------------------------------------------
MIX.PD.par.bounds <- function(par, n, probs.bounds = F)
{
      Max.Raw.P <- 300
      Min.Var   <- MIX.Tol
      
      uneven <- (n %% 2) != 0

      if(uneven)
      {
            lower <- rep(-Inf, length(par))
            upper <- rep(Inf, length(par))
      }
      else
      {
            lower <- rep(-Inf, length(par)+ 2)
            upper <- rep(Inf, length(par) + 2)
      }

      if(uneven)
            sd.off <- rep(6, (n-1)/2 - 1)
      else
            sd.off <- rep(6, n/2 - 1)
            
      pos.sd <- cumsum(c(2, sd.off))      # Positions of stddev
      pos.p  <- c(pos.sd + 1, pos.sd + 2) # Positions of probability parameters
      
      lower[pos.sd] <- Min.Var            # Lower bounds for stdev
      
      if(probs.bounds)                    # Bounds for probability parameters
      {
            lower[pos.p] <- 0
            upper[pos.p] <- 1
      }
      else
      {
            lower[pos.p] <- -Max.Raw.P
            upper[pos.p] <-  Max.Raw.P
      }      
      if(!uneven)
      {
        lower <- lower[-c(1,2)]
        upper <- upper[-c(1,2)]   
      }

      list(lower = lower, upper = upper)
}

# ------------------------------------------------------------------------------------------------------------------
# Generate strings with names of the parameters for model MIX2
# Input:	par = Vector of parameters
#		  n = number of signal pairs
#
# Output: String with parameter names
# ------------------------------------------------------------------------------------------------------------------
MIX.PD.par.names <- function(par, n)
{
	if(n %% 2 != 0)
	{
		m   <- (n-1)/2
		n.t <- length(par) - (n-1)*3

	}
	else		
	{
		m   <- n/2
		n.t <- length(par) - n*3 + 2
	}
	mean <- paste("Mean["   , 1:m, "]", sep = "")
	sd   <- paste("Stddev[" , 1:m, "]", sep = "")
	p.L  <- paste("p.left[" , 1:m, "]", sep = "")
	p.R  <- paste("p.right[", 1:m, "]", sep = "")
	d.L  <- paste("d'.left[" , 1:m, "]", sep = "")
	d.R  <- paste("d'.right[", 1:m, "]", sep = "")

	p.name   <- NULL
	for(i in 1:m) p.name <- c(p.name, mean[i], sd[i], p.L[i], p.R[i], d.L[i], d.R[i])
	if(n %% 2 == 0)	p.name <- p.name[-c(1,2)]

	c(p.name, paste("t-", 1:n.t, sep = ""))
}

# ------------------------------------------------------------------------------------------------------------------
# Names of the models 
# ------------------------------------------------------------------------------------------------------------------
MIX.PD.model.name <- "Mixture Model: Mixtures of three normal distributions for each pair of signals"

# ------------------------------------------------------------------------------------------------------------------
# Format parameters for MIX model
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Par.MIX.PD <- function(par, n, fixed, dec)
{
	# Compute probabilities for non-fixed probability parameters
	par <- MIX.PD.cmp.p(par, n, fixed) 	

	# Format output
	if(n %% 2 != 0)		# Number of models = uneven
	{
		m   <- (n-1)/2
		t   <- round(par[(m*6 + 1):length(par)], dec)				# Extract threshold parameters
		n.t <- length(t)
		par <- round(c(0, 1, NA, NA, NA, NA, par[1:((n-1)*3)]), dec)
		Fp  <- matrix(par, nc = m+1)
		Fp  <- rbind(Fp, matrix(rep(t, m+1), nc = m+1))
		colnames(Fp) <- c("Foil", paste("Mixture-", 1:m, sep = ""))
		rownames(Fp) <- c("Mean", "Stddev",  "p.left", "p.right", "d'.left", "d'.right", paste("t-", 1:n.t, sep = ""))
	}
	else				# Number of models = even
	{
		m   <- n/2
		t <- round(par[(m*6 - 1):length(par)], dec)							# Extract threshold parameters
		n.t <- length(t)
		par <- c(0, 1, par[1:(m*6 - 2)])
		Fp <- matrix(round(par, dec), nc = m)
		Fp  <- rbind(Fp, matrix(rep(t, m), nc = m))
		colnames(Fp) <- c(paste("Mixture-", 1:m, sep = ""))
		rownames(Fp) <- c("Mean", "Stddev",  "p.left", "p.right", "d'.left", "d'.right", paste("t-", 1:n.t, sep = ""))
	}
	Fp
}


# ------------------------------------------------------------------------------------------------------------------
# Format free free parameters for MIX-PD model and adapt standard errors
# Input:	1. px2 Matrix with parameters and standard errors
#		2. full parameter vector
#		3. data vector
#		4. number of signals
#		5. matrix of fixed constraints
#		6. matrix of equality constraints
#		7. function implementing functional constraints
#           8. confidecne level
#           9. number of decimal places
#
# Output: p x 5 Matrix of parameters, standard errors, and confidence intervals
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Par.est.MIX.PD <- function(Pars, full.par, data, n, fixed, ident, functional, alpha, deci)
{
	pos.free <- SavePos.par(full.par, fixed, ident)		# Positions of free parameters
	
	# Compute Positions of probabilities in full parameter vector
	if(n%%2 != 0)	# Uneven number of signals
	{
		len.p <- (n-1)*3
		pos.p <- c(seq(3, len.p-1, 6), seq(4, len.p, 6))
	
	}
	else			# Even number of signals
	{
		len.p <- n*3 - 2
		pos.p <- c(seq(1, len.p-1, 6), seq(2, len.p, 6))
		
	}
	pos   <- pos.free %in% pos.p					# Positions of probability parameters in free parameter vector

	# Compute probabilities and adapt standard errors
	Pars[pos, 1] <- exp(Pars[pos ,1]) / (1 + exp(Pars[pos ,1]))
	Pars[pos, 2] <- Pars[pos, 2] * (Pars[pos, 1]*(1-Pars[pos, 1]))

	# Compute asymptotic standard errors numerically
	Pars <- cbind(Pars, SDT.MIX.PD.Hessian.p(full.par, data, n, fixed = fixed, ident = ident, functional = functional))

      # Compute confidence intervals
      bd    <- MIX.PD.par.bounds(full.par, n, probs.bounds = T)       # Bounds on probabilities
      bd.lo <- Elim.par(bd$lower, fixed, ident)
      bd.up <- Elim.par(bd$upper, fixed, ident)

      qt.cfi <- qnorm((1-alpha)/2)
      cfi.lo <- Pars[,1] + Pars[,2]*qt.cfi
      cfi.up <- Pars[,1] - Pars[,2]*qt.cfi
            
      # Respect lower bounds on sd parameter
      cfi.lo <- ifelse(cfi.lo < bd.lo, bd.lo, cfi.lo)
      cfi.up <- ifelse(cfi.up > bd.up, bd.up, cfi.up)
      
  	# Add columns with confidence intervals and specify new column names
      Pars <- cbind(Pars, cfi.lo, cfi.up)
	colnames(Pars) <- c("Value", "SE(delta method)", "SE(numerical)", paste("CFI-", alpha*100, "(Lower)", sep = ""), paste("CFI-", alpha*100, "(Upper)", sep = ""))

	round(Pars, deci)
}


# ---------------------------------------------------------------------------------------------------------------------
# MIX Model: Check number of parameters and data points
# Input:	1. Parameter vector
#		2. Data vector
#		3. Number of signals
#
# Output:	A list with three entries:
#			1. Error flag: T = Error, F = No error
#			2. 2x2 matrix with actual (first row) and predicted (scond row) number of parameters and data points
# ----------------------------------------------------------------------------------------------------------------------
SDT.MIX.PD.check <- function(par, data, n)
{
	if(n%%2 != 0)	# Uneven number of signals
	{
		n.par  <- (n-1)*3 + length(data)/n - 1		# Predicted number of parameters due to data
		n.data <- (length(par) - (n-1)*3 + 1)*n		# Predicted number of data points 
	}
	else			# Even number of signals
	{
		n.par  <- n*3 - 2 + length(data)/n - 1		# Predicted number of parameters due to data
		n.data <- (length(par) - n*3 + 2 + 1)*n		# Predicted number of data points 
	}

	Erg <- matrix(c(length(par), length(data), n.par, n.data), nc = 2)
	rownames(Erg) <- c("Parameters", "Data points")
	colnames(Erg) <- c("Found", "Expected")

	Error <- any(Erg[, 1] != Erg[, 2])
	list(Error = Error, Stat = Erg, Msg = "NOTE: Computation of the expected number of parameters is based on the to number of data points and vice versa")
}

# ------------------------------------------------------------------------------------------------------------------
# Functions for computing Hessian numerically
# ------------------------------------------------------------------------------------------------------------------
SDT.MIX.PD.Hessian.p <- function(par, data, n, fixed = fixed, ident = ident, functional = functional)
{
	par.p   <- MIX.PD.cmp.p(par, n, fixed) 		# Compute probabilities
	pos.par <- SavePos.par(par, fixed, ident)		# Positions of free parameters
	par.p	  <- Elim.par(par.p,  fixed, ident)		# Eliminate constrained parameters
	Model   <- SDT.MIX.PD.Pr				# Function for computing model probabilities
	hess    <- hessian(func = SDT.logL, x = par.p, data = data, Model = Model, n = n, pos = pos.par, fixed = fixed, ident = ident, functional = functional)

	if(is.null(hess) || is.na(hess) || is.nan(hess)) return(rep(NA, length(par.p))) 	# Hessian absent or not OK -> retuen (NA)

	Erg <- Test.Hessian(hess)	
	if(Erg$P.d == F)		return(rep(NA, length(par.p)))					# Not positive definite -> return(NA) 		
	Cov <- solve(hess)
	return(sqrt(diag(Cov)))
}


#===============================================================================
#
# Functions for comupting the model matrix (analytically and numerically)
# for the MIX-PD model
#
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Model incorporating restrictions
# The parameter passed to the function contains the free parameters only
# ------------------------------------------------------------------------------------------------------------------
SDT.MIX.PD.Full <- function(rest.par, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
      full.par <- Restore.par(rest.par, pos, fixed, ident, functional)	# Restore full parameter vector
      SDT.MIX.PD(full.par, n, fixed, ident)
}

# ------------------------------------------------------------------------------------------------------------------
# Numerical computation of the model matrix 
# ------------------------------------------------------------------------------------------------------------------
SDT.MIX.PD.Model.Matrix.Num <- function(full.par, n, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)
      par <- Elim.par(full.par, fixed, ident)
      J <- jacobian(SDT.MIX.PD.Full, par, n = n, pos = pos, fixed = fixed, ident = ident, functional = functional)
}

# ------------------------------------------------------------------------------
# Analytical computation of the model matrix
# Input:    1. Full parameter vector
#           2. Number of signals
#           3. 2 x p - matrix of fixed constraints
#           4. 2 x q - matrix of identity constraints
#           5. Jacobian of functional constraints (supplied by the user)
#
# Output:   Modelmatrix
#
# Comment:  The Modelmatrix of funtional constraints must be a full matrix that
#           is premultiplied by the Jacobian
# ------------------------------------------------------------------------------
SDT.MIX.PD.Model.Matrix <- function(full.par, n, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)

      J   <- SDT.MIX.PD.Jacobi(full.par, n = n, fixed)
      if(!is.null(ident))
      {
            J.Id <- diag(length(full.par))
            Index <- matrix(c(ident[2,], ident[1,]), nc = 2)
            J.Id[Index] <- 1
            J <- J %*% J.Id
      }
      
      # Treatment of funtional constraints
      if(!is.null(functional))
      {
            J.fct <- attr(functional, "SDT.Jacobian")
            if(is.null(J.fct)) 
                  return(NULL)           # No Jacobian function supplied -> return (NULL)
            J <- J %*% J.fct(full.par)
      }
      J[, pos]
}

#===============================================================================
#
# Plot function:
#
#===============================================================================


# ------------------------------------------------------------------------------------------------------------------
# Plotting for the MIX.PD model
# Input:	1. par.vec = vector of parameters
#		2. n      = number of models
#           3. cols   = number of colors
#           4. ltys   = line types
#		5. labels = labels for axes
#           6. SDT.legend = list with tow entries: 
#                       pos = c(x, y) Position
#                       text = c("text 1",...,"text n")
#                       lty  = vector of line types
#           7. option:  NULL       --> SDT
#                       "ROC"       --> ROC
#                       "zROC"      --> zROC
#                       "ROC+"      --> ROC + Data
#                       "zROC+"     --> zROC + Data
#           8. counts: data     
#
# Comment:  - In case of density plots, the function extracts the decision bounds as well as the means and sd parameters 
#             and calls the function SDT.SDT1D.Plot.
#           - In case of ROC plots the mean, and sd vectors are transformed into MIX.2 format and the function
#             SDT.ROC.Plot is called
# ------------------------------------------------------------------------------------------------------------------
SDT.MIX.PD.Plot <- function(par.vec, n, cols, ltys, labels, SDT.legend, option, counts)
{
      m <- sd <- p <- NULL
 	if(n %% 2 == 0)		   # Even number of signals
	{
		k <- 1
		par.vec <- c(0, 1, par.vec)
            for(i in 1:(n/2))
            {
                  if(is.null(option))
                  {
                        m  <- c(m, par.vec[k], par.vec[k] - par.vec[k + 4], par.vec[k] + par.vec[k + 5])
                        sd <- c(sd, rep(par.vec[k+1], 3))
                  }
                  # In case of ROC curves change parameters to MIX.2 format
                  else
                  {
                         m  <- c(m, par.vec[k] - par.vec[k + 4], par.vec[k], par.vec[k] + par.vec[k + 5], par.vec[k])
                        sd <- c(sd, rep(par.vec[k+1], 4))
                  }
                  p  <- c(p, par.vec[k+2], par.vec[k+3])
                  k  <- k + 6
                  
                  
            }
            t <- par.vec[k:length(par.vec)]
 	}
	else					# Uneven number of signals
	{
	      if(is.null(option))
	      {
	           m  <- 0
	           sd <- 1
            }
            else # In case of ROC curves change parameters to MIX.2 format
            {
                  m  <- c(0, 0)
                  sd <- c(1, 1)
                  p  <- 1
            }
		k <- 1
            for(i in 1:((n-1)/2))
            {
                  if(is.null(option))
                  {
                        m <- c(m, par.vec[k], par.vec[k] - par.vec[k + 4], par.vec[k] + par.vec[k + 5])
                        sd <- c(sd, rep(par.vec[k+1], 3))
                  }
                  # In case of ROC curves change parameters to MIX.2 format
                  else
                  {
                         m  <- c(m, par.vec[k] - par.vec[k + 4], par.vec[k], par.vec[k] + par.vec[k + 5], par.vec[k])
                        sd <- c(sd, rep(par.vec[k+1], 4))
                  }
                  p  <- c(p, par.vec[k+2], par.vec[k+3])
                  k <- k + 6
         }
            t <- par.vec[k:length(par.vec)]
	}

      # Compute p from raw parameters
      if(!is.null(p) && !is.null(option))
            p <- exp(p) / (1 + exp(p))            
      
      # Perform plotting
      if(is.null(option))
            SDT.SDT1D.Plot(m, sd, t, n = length(m), cols, ltys, labels, SDT.legend)
      else if(option == "ROC")
            SDT.ROC.Plot(m, sd, p = p, n = length(p), cols, ltys, labels, SDT.legend, zROC = F, counts = NULL)
      else if(option == "ROC+")
            SDT.ROC.Plot(m, sd, p = p, n = length(p), cols, ltys, labels, SDT.legend, zROC = F, counts = counts)
      else if(option == "zROC")
            SDT.ROC.Plot(m, sd, p = p, n = length(p), cols, ltys, labels, SDT.legend, zROC = T, counts = NULL)
      else if(option == "zROC+")
            SDT.ROC.Plot(m, sd, p = p, n = length(p), cols, ltys, labels, SDT.legend, zROC = T, counts = counts)
      else
            cat("\n\n --> SDT.MIX.PD.Plot: Unknown option: \"", option, "\"\n\n", sep = "")
}
 