# ================================================================
# Functions for estimating SDT Model-Parameters
# Date of creation:	July,	      2009
# Date of last update:	September,  2013
#
# Author: Siegfried Macho
# ================================================================
options(warn = - 1)
library(numDeriv)
options(warn = 0)

# ----------------------------------------------------------------------
# Kernel of the negative log-likelihood-function to be minimized
# I. For Gaussian model
# Input:	1. Vector with parameters
#		2. Vector of observed frequencies
#		3. Model that computes probabilities
#		4. Number of Models
#		5. Information about positions of free parameters
#		6. 2xk matrix of fixed constraints
#		7. 2xm matrix of equality constraints
#		8. Function for specifying functional constraints
#
# Output: 	Negative log likelihood
# ----------------------------------------------------------------------
SDT.logL <- function(par, data, Model, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
	Tol <- 1e-16
	
	par <- Restore.par(par, pos, fixed, ident, functional)	# Restore full parameter vector
  est.p <- Model(par, n, fixed, ident)
	est.p[est.p < Tol | is.na(est.p)] <- Tol
	-sum(data*log(est.p))							# Return negative likelihood
}

# ----------------------------------------------------------------------
# Gradient: dlogL/dp 
# ----------------------------------------------------------------------
SDT.logL.Jacobi <- function(par, data, Model, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
	Tol <- 1e-16								# Numeric tolerance level

	est.p <- Model(par, n, fixed, ident)
	est.p[est.p < Tol | is.na(est.p)] <- Tol

	matrix(-data/est.p, nr = 1)		# Gradient of logL
}


# ----------------------------------------------------------------------
# Matrix of second derivatives: d2logL/dp2 
# ----------------------------------------------------------------------
SDT.logL.Jacobi.2 <- function(par, data, Model, n, fixed = NULL, ident = NULL, functional = NULL)
{
	Tol <- 1e-16								# Numeric tolerance level

	est.p <- Model(par, n, fixed, ident)
	est.p[est.p < Tol | is.na(est.p)] <- Tol

	diag(data/est.p^2)		# Hessian of logL
}


# -----------------------------------------------------------------------------------------------------------
# Optimization routine
# Input: 	1. Vector of Parameters
#		2. Data vector
#		3. Model identiication string, for example
#			"SDT" 		= Standard Signal detection model with N(0, 1) as Noise distribution and one set of thresholds
#			"Gaussian"		= Gaussian Models with threhsolds for each model
#		4. Number of Models
#		5. 2xk matrix of fixed constraints
#		6. 2xm matrix of equality constraints
#		7. Function for specifying functional constraints
#	      8. Flag indicating whether a symbolic gradient is computed (T) or a numerical approximation (F)
#	      9. Flag indicating whether the symbolic and numeric gradients should be compared at the end of the optimization process (=T)
#          10. Flag indicating whether to use the optimizer nlminb before optim.
#
# Output: 	Optimization object with new entry: full.par (full parameter vector
#
# Comment: 	If no parameter vector is supplied the starting parameter is generated internally
#	
# -----------------------------------------------------------------------------------------------------------
SDT.Estimate <- function(par = NULL, data, Model.Id = "SDT", n = 2, fixed = NULL, ident = NULL, functional = NULL, sym.gr = T, test = F, use.nlminb = T)
{
      Tol <- 1e-16

	    Model.Name       <- "No Name provided"					# Name of the model
	    Gradient         <- NULL							# Function for computing Gradient
	    Model.Matrix     <- NULL                                          # Function for computing the model matrix
	    Model.Matrix.Num <- NULL                                          # Function for computing the numerical model matrix
	    upper <- lower   <- NULL                                          # Upper and lower bounds for constrained optimization
	    full.data        <- NULL                                          # Full data vector
	
 	    if(Model.Id == "SDT") 								# Model = Standard signal detection model
	    {
            n <- SDT.SDT.Check.Fill.Cfg(n)                              # Check and complete the configuration information 
		        if(is.null(par)) par <- SDT.SDT.start.par(n, data)		# Generate default starting parameter
		        Bounds <- SDT.SDT.par.bounds(par, n)                        # Generate bounds for parameters
		        lower  <- Bounds$lower
		        upper  <- Bounds$upper
 		        Model  <- SDT.SDT          						# Model function
		        if(sym.gr)
            {
                  Gradient <- SDT.SDT.Grad					# Gradient of logL
                  Model.Matrix <- SDT.SDT.Model.Matrix                  # Function for computing the model matrix
            }
            Model.Matrix.Num  <- SDT.SDT.Model.Matrix.Num               # Function for computing the model matrix numerically
            par.names   <- SDT.par.names(par, n)				# Names of parameters
		        Model.Name  <- SDT.model.name(n)					# Name of the model
		        Ch <- SDT.SDT.check(par, data, n)					# Check number of data and parameters

		        if(Ch$Error)
		        {
                  print(Ch$Stat)
			            cat(Ch$Msg, "\n\n")
			            stop("\n SDT: Number of parameters and data points incompatible\n")
            }
            CS <- SDT.SDT.Restriction(n, fixed, ident)                  # Set specified restriction
		        fixed <- CS$fixed
            ident <- CS$ident
 	    }
 	    else if(Model.Id == "SDT.Robust") 						# Model = Standard signal detection model with robust decision bounds
	    {
		        if(is.null(par)) par <- SDT.SDT.start.par.Robust(n, data)	# Generate default starting parameter
		        Model 	<- SDT.SDT.Robust     					# Model function
		        if(sym.gr)
            {
                  Gradient <- SDT.SDT.Grad.Robust				# Gradient of logL
                  Model.Matrix <- SDT.SDT.Model.Matrix.Robust           # Function for computing the model matrix
            }
            Model.Matrix.Num  <- SDT.SDT.Model.Matrix.Num.Robust        # Function for computing the model matrix numerically
		        par.names   <- SDT.par.names(par, n)	                  # Names of parameters
		        Model.Name  <- SDT.model.name.Robust          			# Name of the model
		        Ch <- SDT.SDT.check(par, data, n)		                  # Check number of data and parameters
		        if(Ch$Error)
		        {
			            print(Ch$Stat)
			            cat(Ch$Msg, "\n\n")
			            stop("\n SDT.Robust: Number of parameters and data points incompatible\n")
            }
      }
	    else if(Model.Id == "Gaussian") 						# Model = SDT model with entirely free parameters
	    {
            n <- SDT.Gaussian.Check.Fill.Cfg(n)                         # Check and complete the configuration information 
		        if(is.null(par)) par <- SDT.Gaussian.start.par(n, data)	# Generate default starting parameter
		        Bounds <- SDT.Gaussian.par.bounds(par, n)                   # Generate bounds for parameters
          	lower  <- Bounds$lower
		        upper  <- Bounds$upper
		        Model  <- SDT.Gaussian						      # Model function
		        if(sym.gr)
            {
                  Gradient <- SDT.Gaussian.Grad				      # Gradient of logL
                  Model.Matrix <- SDT.Gaussian.Model.Matrix             # Function for computing the model matrix
            }
            Model.Matrix.Num  <- SDT.Gaussian.Model.Matrix.Num          # Function for computing the model matrix numerically
		        par.names   <- Gaussian.par.names(par, n)				# Names of parameters
		        Model.Name  <- Gaussian.model.name(n)         			# Name of the model
		        Ch <- SDT.Gaussian.check(par, data, n)				# Check number of data and parameters
		        if(Ch$Error)
		        {
			            print(Ch$Stat)
			            cat(Ch$Msg, "\n\n")
			            stop("\n Gaussian: Number of parameters and data points incompatible\n")
            }
	          CS <- SDT.Gaussian.Restriction(par, n, fixed, ident, functional) # Set specified restriction
		        fixed      <- CS$fixed                                           # Fixed constraints
            ident      <- CS$ident                                           # Identity constriants
            functional <- CS$functional                                      # Functional constraints
      }
	    else if(Model.Id == "MIX.PD") 						# Model = MIX model with p and d free parameters
	    {
            if(is.null(par)) par <- SDT.MIX.PD.start.par(n, data)		# Generate default starting parameter
		        Model 	<- SDT.MIX.PD						# Model function
		        if(sym.gr)
            {
                  Gradient <- SDT.MIX.PD.Grad				      # Gradient of logL
                  Model.Matrix <- SDT.MIX.PD.Model.Matrix               # Function for computing the model matrix analytically
            }
            Model.Matrix.Num  <- SDT.MIX.PD.Model.Matrix.Num            # Function for computing the model matrix numerically
		        par.names   <- MIX.PD.par.names(par, n)				# Names of parameters
		        Model.Name  <- MIX.PD.model.name 					# Name of the model
		        Ch <- SDT.MIX.PD.check(par, data, n)				# Check number of data and parameters
		        if(Ch$Error)
		        {
			            print(Ch$Stat)
			            cat(Ch$Msg, "\n\n")
			            stop("\n MIX.PD: Number of parameters and data points incompatible\n")
            }
      }
	    else if(Model.Id == "MIX.2") 							# Model = MIX2 model with entirely free parameters
	    {
		        if(is.null(par)) par <- SDT.MIX2.start.par(n, data)		# Generate default starting parameter
            Model 	<- SDT.MIX2							# Model function
            if(sym.gr)
            {
                  Gradient <- SDT.MIX2.Grad				      # Gradient of logL
                  Model.Matrix <- SDT.MIX2.Model.Matrix                 # Function for computing the model matrix analytically
            }
            Model.Matrix.Num  <- SDT.MIX2.Model.Matrix.Num              # Function for computing the model matrix numerically
		        par.names   <- MIX2.par.names(par, n)				# Names of parameters
		        Model.Name  <- MIX2.model.name 					# Name of the model
		        Ch <- SDT.MIX2.check(par, data, n)					# Check number of data and parameters
		        if(Ch$Error)
		        {
			            print(Ch$Stat)
			            cat(Ch$Msg, "\n\n")
			            stop("\n MIX.2: Number of parameters and data points incompatible\n")
            }
	    }
	    else if(Model.Id == "DPSDT")						      # Model = DPSDT model
	    {
            n <- DPSDT.Check.Fill.Cfg(n, data)                          # Check and complete the configuration information 
		        if(is.null(par)) par <- DPSDT.start.par(n, data)	      # Generate default starting parameter
		        if(use.nlminb)
		        {
                  Bounds <- DPSDT.par.bounds(par, n)                    # Generate bounds for parameters
                  lower  <- Bounds$lower
                  upper  <- Bounds$upper
            }      
		        Model  <- DPSDT.Probs						      # Model function
		        if(sym.gr)
            {
                  Gradient <- DPSDT.Grad				            # Gradient of logL
                  Model.Matrix <- DPSDT.Model.Matrix                    # Function for computing the model matrix analytically
            }
            Model.Matrix.Num  <- DPSDT.Model.Matrix.Num                 # Function for computing the model matrix numerically
		        par.names   <- DPSDT.par.names(par, n)				# Names of parameters
		        Model.Name  <- DPSDT.model.name(n) 					# Name of the model
		        Ch <- DPSDT.check(par, data, n)				      # Check number of data and parameters
		        if(Ch$Error)
		        {
                  print(Ch$Stat)
			            cat(Ch$Msg, "\n\n")
               		stop("\n DPSDT: Number of parameters and data points incompatible\n")
            }
	          CS <- DPSDT.Set.Restriction(par, n, fixed, ident, functional)   # Set specified restriction
		        fixed      <- CS$fixed                                      # Fixed constraints
            ident      <- CS$ident                                      # Identity constriants
            functional <- CS$functional                                 # Functional constraints
      }
	    else if(Model.Id == "HTSDT.1")						# Model = HTSDT.1 model
	    {
		        if(is.null(par)) par <- SDT.HTSDT.1.start.par(n, data)	# Generate default starting parameter
		        Model 	<- SDT.HTSDT.1						# Model function
		        if(sym.gr)
            {
                  Gradient <- SDT.HTSDT.1.Grad				      # Gradient of logL
                  Model.Matrix <- SDT.HTSDT.1.Model.Matrix              # Function for computing the model matrix analytically
            }
            Model.Matrix.Num  <- SDT.HTSDT.1.Model.Matrix.Num           # Function for computing the model matrix numerically
		        par.names   <- HTSDT.1.par.names(par, n)				# Names of parameters
		        Model.Name  <- HTSDT.1.model.name 					# Name of the model
		        Ch <- SDT.HTSDT.1.check(par, data, n)				# Check number of data and parameters
		        if(Ch$Error)
		        {
			            print(Ch$Stat)
			            cat(Ch$Msg, "\n\n")
			            stop("\n HTSDT.1: Number of parameters and data points incompatible\n")
            }
      }
	    else if(Model.Id == "HT.n")							# Model = High threshold model with rating
	    {
            if(is.null(par)) par <- SDT.HT.n.start.par(n, data)		# Generate default starting parameter
		        Bounds <- HT.n.par.bounds(par, n)                           # Generate bounds for parameters
		        lower  <- Bounds$lower
		        upper  <- Bounds$upper
		        Model  <- SDT.HT.n							# Model function
		        if(sym.gr)
            {
                  Gradient <- SDT.HT.n.Grad				      # Gradient of logL
                  Model.Matrix <- SDT.HT.n.Model.Matrix                 # Function for computing the model matrix analytically
            }
            Model.Matrix.Num  <- SDT.HT.n.Model.Matrix.Num              # Function for computing the model matrix numerically
		        par.names   <- HT.n.par.names(par, n)				# Names of parameters
		        Model.Name  <- HT.n.model.name 					# Name of the model
		        Ch <- SDT.HT.n.check(par, data, n)					# Check number of data and parameters
		        if(Ch$Error)
		        {
                  print(Ch$Stat)
			            cat(Ch$Msg, "\n\n")
			            stop("\n HT.n: Number of parameters and data points incompatible\n")
		        }
		        SDT.HT.n.check.identity.constraints(par, n, ident)           # Check, whether rating parameters enter in forbidden identity constraints
            fixed <- HT.n.Redundant.to.Fixed(par, n, fixed, ident)	 # Add positions of redundant parameters to fixed constraints
            if(sym.gr)
		        {
		              msg <- SDT.HT.n.check.complex.identity.constraints(par, n, fixed, ident) # Check for complex identity constraints on multiple
		              if(!is.null(msg))                                                        # probability parameters
                  {
                       cat(msg); stop("")
                  }
           }
     }
	   else if(Model.Id == "SDT.2D")							# Model = Bivariate Gaussain model
	   {
	         n <- SDT.SDT2D.Check.Fill.Cfg(n)                            # Check the configuration information and fill it
		       if(is.null(par)) par <- SDT.SDT2D.start.par(n)		      # Generate default starting parameter
           Bounds <- SDT2D.par.bounds(par, n)                          # Generate bounds for parameters
		       lower  <- Bounds$lower
		       upper  <- Bounds$upper
		       Model  <- SDT.SDT2D							# Model function
		       if(sym.gr)
           {
                  Gradient <- SDT.SDT2D.Grad				      # Gradient of logL
                  Model.Matrix <- SDT.SDT2D.Model.Matrix                # Function for computing the model matrix analytically
           }
           Model.Matrix.Num  <- SDT.SDT2D.Model.Matrix.Num             # Function for computing the model matrix numerically
		       par.names   <- SDT2D.par.names(n)				      # Names of parameters
		       Model.Name  <- SDT2D.model.name(n)				      # Name of the model
		       Ch <- SDT.SDT2D.check(par, data, n)					# Check number of data and parameters
		       if(Ch$Error)
		       {
			            print(Ch$Stat)
			            cat(Ch$Msg, "\n\n")
			            stop("\n SDT2D: Number of parameters and data points incompatible with given configuration information\n")
		       }
           CS <-  SDT.SDT2D.Restriction(n, fixed, ident, functional) # Set restrictions
           fixed      <- CS$fixed
           ident      <- CS$ident
           functional <- CS$functional
            
           if(!is.null(n$struct.zero))
                   data[n$struct.zero] <- 0                             # Eliminate structural zeros
           if(!is.null(n$R))
           {
                  full.data <- data                                     # Save data
                  data <- as.vector(n$R %*% data)                       # Transform data
           }
     }
     else if(Model.Id == "SDT2D.MIX.2") 					      # Model = STD2D model with Mixture of 2 Gaussians
	   {
           SDT.SDT2D.Check.Cfg(n)                                      # Check the configuration information
	         if(is.null(n$ds))  n$ds <- 0                                # If no flag ds specified set to default
	         if(is.null(n$standard)) n$standard <- TRUE                  # Standard restrictions
           if(is.null(par)) par <- SDT2D.MIX.2.start.par(n)	      # Generate default starting parameter
	         Model 	<- SDT2D.MIX.2       					# Model function
		       if(sym.gr)
           {
                  Gradient <- SDT2D.MIX.2.Grad				      # Gradient of logL
                  Model.Matrix <- SDT2D.MIX.2.Model.Matrix              # Function for computing the model matrix analytically
           }
           Model.Matrix.Num  <- SDT2D.MIX.2.Model.Matrix.Num           # Function for computing the model matrix numerically
		       par.names   <- SDT2D.MIX.2.par.names(n)			      # Names of parameters
		       Model.Name  <- SDT2D.MIX.2.model.name(n) 				# Name of the model
		       Ch <- SDT2D.MIX.2.check(par, data, n)				# Check number of data and parameters
		       if(Ch$Error)
		       {
                  print(Ch$Stat)
			            cat(Ch$Msg, "\n\n")
			            stop("\n SDT2D.MIX.2: Number of parameters and data points incompatible\n")
           }
		       if(n$standard)                                              # Set standard restrictions
		       {
                fixed.s <- matrix(c(0, 0, 1, 1, 2:5), nr = 2, byrow = T)
		            pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
		            fixed <- cbind(fixed.s[, pos.unique], fixed)            # Add to fixed if not alread specified
		       }
		       if(!is.null(n$struct.zero))
                data[n$struct.zero] <- 0                             # Eliminate structural zeros
           if(!is.null(n$R))
           {
                full.data <- data                                     # Save data
                data <- as.vector(n$R %*% data)                       # Transform data
           }
     }
   	 else if(Model.Id == "SDT2D.MIX.3") 					      # Model = STD2D model with Mixture of 3 Gaussians
	   {
           SDT.SDT2D.Check.Cfg(n)                                      # Check the configuration information
	         if(is.null(n$ds))  n$ds <- 0                                # If no flag ds specified set to default
	         if(is.null(n$standard)) n$standard <- TRUE                  # Standard restrictions
           if(is.null(par)) par <- SDT2D.MIX.3.start.par(n)	      # Generate default starting parameter
		       Model 	<- SDT2D.MIX.3       					# Model function
		       if(sym.gr)
           {
                Gradient <- SDT2D.MIX.3.Grad				      # Gradient of logL
                Model.Matrix <- SDT2D.MIX.3.Model.Matrix              # Function for computing the model matrix analytically
           }
           Model.Matrix.Num  <- SDT2D.MIX.3.Model.Matrix.Num           # Function for computing the model matrix numerically
		       par.names   <- SDT2D.MIX.3.par.names(n)			      # Names of parameters
		       Model.Name  <- SDT2D.MIX.3.model.name(n) 				# Name of the model
		       Ch <- SDT2D.MIX.3.check(par, data, n)				# Check number of data and parameters
		       if(Ch$Error)
		       {
			          print(Ch$Stat)
			          cat(Ch$Msg, "\n\n")
			          stop("\n SDT2D.MIX.3: Number of parameters and data points incompatible\n")
           }
		       if(n$standard)                                              # Set standard restrictions
		       {
                fixed.s <- matrix(c(0, 0, 1, 1, 2:5), nr = 2, byrow = T)
		            pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
		            fixed <- cbind(fixed.s[, pos.unique], fixed)            # Add to fixed if not alread specified
           }
		       if(!is.null(n$struct.zero))
                   data[n$struct.zero] <- 0                             # Eliminate structural zeros
           if(!is.null(n$R))
           {
                full.data <- data                                     # Save data
                data <- as.vector(n$R %*% data)                       # Transform data
           }
           fixed <- SDT2D.MIX.3.Redundant.to.Fixed(par, n$s, fixed, ident) # Add positions of redundant parameters to fixed constraints
	   }
   	 else if(Model.Id == "mAFC") 					            # Model = mAFC (Alternative forced choice with bias)
	   {
           n <- SDT.mAFC.Check.Fill.Cfg(n)                             # Check the configuration information
           if(is.null(par)) par <- SDT.mAFC.start.par(n)	            # Generate default starting parameter
		       Model 	<- SDT.mAFC       					# Model function
		       if(sym.gr)
           {
                Gradient <- SDT.mAFC.Grad				       # Gradient of logL
                Model.Matrix <- SDT.mAFC.Model.Matrix                 # Function for computing the model matrix analytically
           }
           Model.Matrix.Num  <- SDT.mAFC.Model.Matrix.Num              # Function for computing the model matrix numerically
		       par.names   <- SDT.mAFC.par.names(n)			      # Names of parameters
		       Model.Name  <- SDT.mAFC.model.name(n) 				# Name of the model
		       Ch <- SDT.mAFC.check(par, data, n)				      # Check number of data and parameters
		       if(Ch$Error)
		       {
			          print(Ch$Stat)
			          cat(Ch$Msg, "\n\n")
			          stop("\n SDT.mAFC: Number of parameters and data points incompatible\n")
		       }
		       GH.Structure <<- Generate.GH(npt = n$npt)
     }
   	 else if(Model.Id == "SDT.Rank") 					                   # Model = Combination of ranking and rating
	   {
           n <- SDT.Rank.Check.Fill.Cfg(n)                       # Check the configuration information
           if(is.null(par)) par <- SDT.Rank.start.par(n)	       # Generate default starting parameter
		       Bounds <- SDT.Rank.par.bounds(n)                      # Generate bounds for parameters
		       lower  <- Bounds$lower
		       upper  <- Bounds$upper
		       Model  <- SDT.Rank                				             # Model function
		       if(sym.gr)
           {
                Gradient <- SDT.Rank.Grad				                 # Gradient of logL
                Model.Matrix <- SDT.Rank.Model.Matrix            # Function for computing the model matrix analytically
           }
           Model.Matrix.Num  <- SDT.Rank.Model.Matrix.Num        # Function for computing the model matrix numerically
		       par.names   <- SDT.Rank.par.names(n)			             # Names of parameters
		       Model.Name  <- SDT.Rank.model.name(n) 				         # Name of the model
		       Ch <- SDT.Rank.check(n, par, data)				             # Check number of data and parameters
		       if(Ch$Error > 0)
		       {
			          print(Ch)
			          stop("\n ERROR SDT.Rank: Number of parameters or data points incompatible with configuration information\n")
		       }

           CS <- SDT.Rank.Set.Restriction(n, fixed, ident, functional) # Specify restictions
           fixed <- CS$fixed
           ident <- CS$ident
           GH.Structure <<- Generate.GH(npt = n$npt)

     }
   	 else if(Model.Id == "HTSDT.Rank") 					     # Model = Combination of ranking and rating with recollectioon and guessing
	   {
            n <- HTSDT.Rank.Check.Fill.Cfg(n)                    # Check the configuration information 
            if(is.null(par)) par <- HTSDT.Rank.start.par(n)	     # Generate default starting parameter 
		        Bounds <- HTSDT.Rank.par.bounds(n)                   # Generate bounds for parameters
		        lower  <- Bounds$lower
		        upper  <- Bounds$upper
		        Model 	<- HTSDT.Rank       		     		     # Model function
		        if(sym.gr)
            {
                  Gradient <- HTSDT.Rank.Grad				     # Gradient of logL
                  Model.Matrix <- HTSDT.Rank.Model.Matrix              # Function for computing the model matrix analytically
            }
            Model.Matrix.Num  <- HTSDT.Rank.Model.Matrix.Num   # Function for computing the model matrix numerically
		        par.names   <- HTSDT.Rank.par.names(n)		         # Names of parameters
		        Model.Name  <- HTSDT.Rank.model.name(n) 			     # Name of the model
		        Ch <- HTSDT.Rank.check(n, par, data)			         # Check number of data and parameters
		        if(Ch$Error)
		        {
			            print(Ch$Stat)
			            cat(Ch$Msg, "\n\n")
			            stop("\n HTSDT.Rank: Number of parameters and data points incompatible\n")
            }
		        CS <- HTSDT.Rank.Restriction(n, fixed, ident)               # Specify restictions
            fixed <- CS$fixed
            ident <- CS$ident
		        GH.Structure <<- Generate.GH(npt = n$npt)
     }
   	 else if(Model.Id == "HTSDT.Bias.Rank") 					# Model = Combination of ranking and rating with recollectioon and guessing
	   {
            n <- HTSDT.Bias.Rank.Check.Fill.Cfg(n)                      # Check the configuration information 
            if(is.null(par)) par <- HTSDT.Bias.Rank.start.par(n, data)	# Generate default starting parameter 
		        Bounds <- HTSDT.Bias.Rank.par.bounds(par, n)                # Generate bounds for parameters
		        lower  <- Bounds$lower
		        upper  <- Bounds$upper
		        Model  <- HTSDT.Bias.Rank              				# Model function
		        if(sym.gr)
            {
                  Gradient <- HTSDT.Bias.Rank.Grad				# Gradient of logL
                  Model.Matrix <- HTSDT.Bias.Rank.Model.Matrix          # Function for computing the model matrix analytically
            }
            Model.Matrix.Num  <- HTSDT.Bias.Rank.Model.Matrix.Num       # Function for computing the model matrix numerically
		        par.names   <- HTSDT.Bias.Rank.par.names(par, n)		# Names of parameters
		        Model.Name  <- HTSDT.Bias.Rank.model.name(n) 			# Name of the model
		        Ch <- HTSDT.Bias.Rank.check(par, data, n)				# Check number of data and parameters
		        if(Ch$Error)
		        {
			            print(Ch$Stat)
			            cat(Ch$Msg, "\n\n")
			            stop("\n HTSDT.Bias.Rank: Number of parameters and data points incompatible\n")
		        }
		        CS <- HTSDT.Bias.Rank.Restriction(n, fixed, ident)          # Specify restictions
            fixed <- CS$fixed
            ident <- CS$ident

		        GH.Structure <<- Generate.GH(npt = n$npt)
     }
   	 else if(Model.Id == "MIX.Rank") 					      # Model = mixed model for ranking and rating
	   {
            n <- MIX.Rank.Check.Fill.Cfg(n)                             # Check the configuration information 
            if(is.null(par)) par <- MIX.Rank.start.par(n, data)	      # Generate default starting parameter 
		        Bounds <- MIX.Rank.par.bounds(par, n)                       # Generate bounds for parameters
		        lower  <- Bounds$lower
		        upper  <- Bounds$upper
		        Model  <- MIX.Rank       					      # Model function
		        if(sym.gr)
            {
                  Gradient <- MIX.Rank.Grad				      # Gradient of logL
                  Model.Matrix <- MIX.Rank.Model.Matrix                 # Function for computing the model matrix analytically
            }
            Model.Matrix.Num  <- MIX.Rank.Model.Matrix.Num              # Function for computing the model matrix numerically
		        par.names   <- MIX.Rank.par.names(par, n)			      # Names of parameters
		        Model.Name  <- MIX.Rank.model.name(n) 				# Name of the model
		        Ch <- MIX.Rank.check(par, data, n)				      # Check number of data and parameters
		        if(Ch$Error)
		        {
			            print(Ch$Stat)
			            cat(Ch$Msg, "\n\n")
			            stop("\n MIX.Rank: Number of parameters and data points incompatible\n")
		        }
	          CS <- MIX.Rank.Restriction(n, fixed, ident)                  # Set specified restriction
		        fixed <- CS$fixed
            ident <- CS$ident

		        GH.Structure <<- Generate.GH(npt = n$npt)
	   }
   	 else if(Model.Id == "MIX.Bias.Rank")  	           	              # Model = mixed model for ranking and rating
	   {
            n <- MIX.Bias.Rank.Check.Fill.Cfg(n)                          # Check the configuration information 
            if(is.null(par)) par <- MIX.Bias.Rank.start.par(n, data)      # Generate default starting parameter 
		        Bounds <- MIX.Bias.Rank.par.bounds(par, n)                    # Generate bounds for parameters
		        lower  <- Bounds$lower
		        upper  <- Bounds$upper
		        Model  <- MIX.Bias.Rank       			  	        # Model function
		        if(sym.gr)
            {
                  Gradient <- MIX.Bias.Rank.Grad				  # Gradient of logL
                  Model.Matrix <- MIX.Bias.Rank.Model.Matrix              # Function for computing the model matrix analytically
            }
            Model.Matrix.Num  <- MIX.Bias.Rank.Model.Matrix.Num           # Function for computing the model matrix numerically
		        par.names   <- MIX.Bias.Rank.par.names(par, n)			  # Names of parameters
		        Model.Name  <- MIX.Bias.Rank.model.name(n) 			  # Name of the model
		        Ch <- MIX.Bias.Rank.check(par, data, n)				  # Check number of data and parameters
		        if(Ch$Error)
		        {
			            print(Ch$Stat)
			            cat(Ch$Msg, "\n\n")
			            stop("\n MIX.Bias.Rank: Number of parameters and data points incompatible\n")
		        }
		        CS <- MIX.Bias.Rank.Restriction(n, fixed, ident)              # Specify restictions
            fixed <- CS$fixed
            ident <- CS$ident

		        GH.Structure <<- Generate.GH(npt = n$npt)
	   }
   	 else if(Model.Id == "IRF.Gauss")  	           	                  # Model = SDT model with probabilitistic item response function 
	   {                       
            n <- IRF.Gauss.Check.Fill.Cfg(n)                          # Check the configuration information 
            if(is.null(par)) par <- IRF.Gauss.start.par(n, data)      # Generate default starting parameter 

		        Bounds <- IRF.Gauss.par.bounds(par, n)                    # Generate bounds for parameters
		        lower  <- Bounds$lower
		        upper  <- Bounds$upper
		        if(n$model == "GRM") Model  <- IRF.GRM.Gauss               # Model function
		        else                 Model  <- IRF.ORM.Gauss
		        if(sym.gr)
            {
                  Gradient     <- IRF.Gauss.Grad				              # Gradient of logL
                  Model.Matrix <- IRF.Gauss.Model.Matrix              # Function for computing the model matrix analytically
            }
            Model.Matrix.Num  <- IRF.Gauss.Matrix.Num                 # Function for computing the model matrix numerically
		        par.names         <- IRF.Gauss.par.names(par, n)			    # Names of parameters
		        Model.Name        <- IRF.Gauss.model.name(n) 			        # Name of the model
		        Ch <- IRF.Gauss.check(par, data, n)				                # Check number of data and parameters
		        if(Ch$Error)
		        {
			            print(Ch$Stat)
			            cat(Ch$Msg, "\n\n")
			            stop("\n IRF.Gauss: Number of parameters and data points incompatible\n")
		        }
		        CS <- IRF.Gauss.Set.Restriction(par, n, fixed, ident)              # Set restictions
            fixed <- CS$fixed
            ident <- CS$ident

		        GH.Structure <<- Generate.GH(npt = n$npt)
	   }
   	 else if(Model.Id == "IRF.LC")  	           	                 # Model = SDT model with probabilitistic item response function 
	   {                                                             #         and Latent classes
            n <- IRF.LC.Check.Fill.Cfg(n)                          # Check the configuration information 
            if(is.null(par)) par <- IRF.LC.start.par(n, data)      # Generate default starting parameter 
		        Bounds <- IRF.LC.par.bounds(par, n)                    # Generate bounds for parameters
		        lower  <- Bounds$lower
		        upper  <- Bounds$upper
		        if(n$model == "GRM") Model  <- IRF.LC.GRM              # Model function
		        else                 Model  <- IRF.LC.ORM
		        if(sym.gr)
            {
                  Gradient     <- NULL #IRF.LC.Grad				          # Gradient of logL
                  Model.Matrix <- NULL #IRF.LC.Model.Matrix         # Function for computing the model matrix analytically
            }
            Model.Matrix.Num  <- NULL # IRF.LC.Model.Matrix.Num     # Function for computing the model matrix numerically
		        par.names         <- IRF.LC.par.names(par, n)			      # Names of parameters
		        Model.Name        <- IRF.LC.model.name(n) 			        # Name of the model
		        Ch                <- IRF.LC.check(par, data, n)				  # Check number of data and parameters
		        if(Ch$Error)
		        {
			            print(Ch$Stat)
			            cat(Ch$Msg, "\n\n")
			            stop("\n IRF.LC: Number of parameters and data points incompatible\n")
		        }
		        CS <- IRF.LC.Set.Restriction(par, n, fixed, ident)              # Set restictions
            fixed <- CS$fixed
            ident <- CS$ident
	   }
     else
            stop(paste("\n Unknown Model identification: <", Model.Id, ">\n", sep = ""))


      # Transform equality constraints to fixed ones, if necessary (Function in Auxiliary.R)
      Cst.new <- Ident.to.fixed(fixed, ident)
	    fixed   <- Cst.new$fixed
	    ident   <- Cst.new$ident

     	# Testing of fixed constraints
	    if(!is.null(fixed) & any(duplicated(fixed[2,])))			# Testing for duplicated fixed effects
	    {
		        pos.dup <- (1:dim(fixed)[2])[duplicated(fixed[2,])]
		        pos.dup <- paste(pos.dup, sep = " ")
		        cat("\nFATAL ERROR: Duplicated fixed constraints found on the following positions: ", pos.dup, "\n")
		        stop("")
      }

	    # Testing of equality constraints
	    if(!is.null(ident))
	    {
		        dup  <-  ident[1, ] %in% ident[2,]
		        if(any(dup))								# Testing for duplicated
		        {
			           pos.dup <- (1:dim(ident)[2])[dup]
			           pos.dup <- paste(pos.dup, sep = " ")
			           cat("\nFATAL ERROR: Parameters subjected to identity constraints on the following source positions: ", pos.dup, "\nare also found on target positions\n")
			           stop("")
		        }
	    }
      pos.par <- SavePos.par(par, fixed, ident)				# Save positions of eliminated parameters	
	    par 	  <- Elim.par(par, fixed, ident)					# Set Parameter constraints

      if(is.null(upper) | is.null(lower))                         # Optimization without Box-constraints
	    Optim.Obj <- optim(par, 	 				      # Start parameters
	 		   	               SDT.logL,					# Likelihood function
				                 gr = Gradient,				# Gradient function
                              data = data,				# Data
				                      Model = Model,				# Model for computing probabilies
				                      n = n,					# Number of Models
				                      pos = pos.par,				# Position of redundant parameters
                              fixed = fixed,				# Fixed constraints
				                      ident = ident,				# Equality constraints
				                      functional = functional,	      # Functional constraints
				                      method = "BFGS",			      # Method = Newton without Box-Constraints
                              hessian = TRUE,				# Compute Hessian 
				                      control = list(reltol = Tol, maxit = 1000)
				                  )
      else                                                        # Optimization with Box-constraints
      {
     	        lower <- Elim.par(lower, fixed, ident)			# Set Parameter constraints on lower bounds
     	        upper <- Elim.par(upper, fixed, ident)			# Set Parameter constraints on upper bounds
              if(use.nlminb)                                          # Use nlminb optimization
              {
                      Nlimb.Obj <- nlminb(start      = par,           # Start parameter
                                          objective  = SDT.logL,      # Function to be optimized,
      				                            gradient   = Gradient,	# Gradient function
                                          data       = data,	      # Data
                                          Model      = Model,		# Model for computing probabilies
                                          n          = n,			# Number of Models
                                          pos        = pos.par,		# Position of redundant parameters
				                                  fixed      = fixed,		# Fixed constraints
				                                  ident      = ident,		# Equality constraints
				                                  functional = functional,	# Functional constraints
				                                  lower      = lower,         # Lower bounds on parameters
				                                  upper      = upper,         # Upper bounds on parameters
				                                  control = list(eval.max = 1500, iter.max = 1500)
				                                 )
                      par <- Nlimb.Obj$par
              }
              Optim.Obj <- optim(par, 	 				      # Start parameters
                                 SDT.logL,				      # Likelihood function
                                 gr = Gradient,			  	# Gradient function
                                 data = data,			  	# Data
                                 Model = Model,				# Model for computing probabilies
                                 n = n,					# Number of Models
                                 pos = pos.par,				# Position of redundant parameters
                                 fixed = fixed,				# Fixed constraints
                                 ident = ident,				# Equality constraints
                                 functional = functional,	      # Functional constraints
                			           method = "L-BFGS-B",			# Method = Newton with Box-Constraints
                		 	           hessian = TRUE,			# Compute Hessian
                			           lower = lower,				# Lower limit of parameters
                			           upper = upper,				# Upper limit of parameters
                			           control = list(factr = Tol, maxit = 1000)
				                        )
      }
      Optim.Obj$SDT.full.par 	 <- Restore.par(Optim.Obj$par, pos.par, fixed, ident, functional)	# Full parameter vector
	    Optim.Obj$SDT.par.names	 <- par.names
	    if(!is.null(par.names))	Optim.Obj$SDT.par.names.est	<- Elim.par(par.names, fixed, ident)
	    else				Optim.Obj$SDT.par.names.est	<- NULL
	    if(is.null(full.data))  Optim.Obj$SDT.data <- data                   # Data vector
      else                    Optim.Obj$SDT.data <- full.data              # Data vector (only relevant for SDT-2D model)
	    Optim.Obj$SDT.Model  	        <- Model
	    Optim.Obj$SDT.Model.Matrix.Anal <- Model.Matrix                      # Function for computing the model matrix
	    Optim.Obj$SDT.Model.Matrix.Num  <- Model.Matrix.Num                  # Function for computing the model matrix numerically
	    Optim.Obj$SDT.Model.Id 	        <- Model.Id                          # Model identification string
	    Optim.Obj$SDT.n		        <- n
	    Optim.Obj$SDT.Model.Name        <- Model.Name				   # Name of the model
	    Optim.Obj$SDT.gr.sym.len        <- NA
	    Optim.Obj$SDT.fixed 	        <- fixed
	    Optim.Obj$SDT.ident	        <- ident
	    Optim.Obj$SDT.functional        <- functional
	    Optim.Obj$SDT.sym.gr            <- sym.gr

	    # Computation of gradients at optimum
	    if(!is.null(Gradient))
	    {
		        # Compute Gradient at Optimum
		        gr.sym <- Gradient(Optim.Obj$par, data = data, Model = Model, n = n, pos = pos.par, fixed = fixed, ident = ident, functional = functional)
		        Optim.Obj$SDT.gr.sym.len <- sqrt(sum(gr.sym^2))
		        if(test == T)
		        {	# Compare numerical and symbolical gradient at the optimum
			              gr.num <- as.vector(grad(SDT.logL, Optim.Obj$par, data = data, Model = Model, n = n, pos = pos.par, fixed = fixed, ident = ident, functional = functional))
			              gr.dif <- gr.sym - gr.num
			              gr.erg <- cbind(gr.sym, gr.num, gr.dif)
			              colnames(gr.erg) <- c("Symbolic", "Numeric", "Difference")
			              if(!is.null(Optim.Obj$SDT.par.names.est)) rownames(gr.erg) <- Optim.Obj$SDT.par.names.est

			              cat("\n=======================================================\n")
			              cat(" Symbolic and numeric gradients at the optimum:\n Length of difference vector  :", sqrt(sum(gr.dif^2)))
			              cat("\n Length of gradient at optimum:", Optim.Obj$SDT.gr.sym.len)
			              cat("\n=======================================================\n")
			              print(round(gr.erg, 8))
			              cat("=======================================================\n")
            }
	    }
	    else
	    {
		        # Compute numerical Gradient at Optimum and its length
		        gr.num <- as.vector(grad(SDT.logL, Optim.Obj$par, data = data, Model = Model, n = n, pos = pos.par, fixed = fixed, ident = ident, functional = functional))
		        Optim.Obj$SDT.gr.sym.len <- sqrt(sum(gr.num^2))

      }
	    Optim.Obj
}

# ------------------------------------------------------------------------------------------------------------------
# Computation of test statistics
# Input:	1. Optimization-Object
#		2. Number of decimal places in output (default = 3)
#           3. Size of the confidence region
#		4. A flag whether a warning should be displayed in case of a rank deficient Hessian matrix
#           5. A list with control information
#
# Output: A list with the following entries:
#	I. Statistics: A vector of test statistics with the following entries
#		1. log likelihood at optimum
#		2. Pearson statistic X^2
#		3. Likelihood ratio statistics G^2
#		4. Degrees of freedom
#		5. p of X^2-statistic
#		6. p of G^2-statistic
#		7. AIC
#		8. BIC
#		9. Number of free parameters
#		10.Rank of the Hessian matrix
#		11.Length of symbolic gradient
#		12.Condition number of Hessian matrix
#
#	II. Free.Parameters: A matrix with free parameters in the first and standard deviations in the second column
#
#    III. Full.parametervector: (C+1) x n matrix with all parameters including fixed and constraint parmeters
#
#	IV. Estimated.probabilities: (C x n) Matrix of estimated probabilities 
#		C = number of response categories
#		n = number of models
#
#	 V. Estimated.frequencies: (C x n) Matrix of estimated frequencies 
#		C = number of response categories 
#		n = number of models
#
#	VI. Input Data: (C x n) Matrix
#		C = number of response categories 
#		n = number of models
#
#	 V. Residuals
# ------------------------------------------------------------------------------------------------------------------
SDT.Statistics <- function(Opti, deci = 3, conf = .95, display.warning = T, control = NULL)
{
	Tol <- 1e-12

	if(is.null(Opti$SDT.n))
	{
		cat("\n\n ===> ERROR in SDT.Statistics: Wrong Optimization Object!\n\n")
		return(NULL)
	}
	n 			          <- Opti$SDT.n
	Model			        <- Opti$SDT.Model
	Model.Name		    <- Opti$SDT.Model.Name
	Model.Id		      <- Opti$SDT.Model.Id
	Model.Matrix      <- Opti$SDT.Model.Matrix.Anal
	Model.Matrix.Num  <- Opti$SDT.Model.Matrix.Num
	data <- data.full <- Opti$SDT.data	
	full.par 		      <- Opti$SDT.full.par
	par 			        <- Opti$par
	par.names		      <- Opti$SDT.par.names
	par.names.est	    <- Opti$SDT.par.names.est
	fixed			        <- Opti$SDT.fixed
	ident			        <- Opti$SDT.ident
	functional		    <- Opti$SDT.functional

	q <- length(Opti$par)					# Number of free non redundant parameters
	N <- sum(data)								# Number of observations		

      # Store relevant information in Optimization Object 	
      if(Model.Id == "SDT.2D" || Model.Id == "SDT2D.MIX.2" || Model.Id == "SDT2D.MIX.3")       # Computation of number of data and df specific for SDT2D and SDT2D.MIX3
      {
            n.tmp <- n
            n.tmp$R <- NULL
            est.p  <- Model(full.par, n.tmp, fixed, ident)        # Compute full vector of estimates
            est.h  <- SDT2D.Compute.Est.N(est.p, data, n.tmp)     # Estimated frequencies (File: SDT-SDT-2D.R)
            df.dat <- SDT2D.Compute.df.data(data, n)              # Degress of freedom in the data (File: SDT-SDT-2D.R)
            if(!is.null(n$R)) 
            {
                  data <-  as.vector(n$R %*% data)                      # Transform data and estiamted vectors if necessary
                  est.h <- as.vector(n$R %*% est.h)             
                  est.p <- as.vector(n$R %*% est.p)             
            }
      }
      else if(Model.Id == "SDT.Rank") { 
            est.p <- Model(full.par, n, fixed, ident)	            # Estimated probabilities
            est.h <- est.p * SDT.Rank.Compute.Obs.N(data, n)	    # Estimated frequencies (File: SDT-Rank.R)
            df.dat<- SDT.Rank.Compute.df.data(n)                  # Number of free data points (File: SDT-Rank.R)
      }
      else if(Model.Id == "HTSDT.Rank") { 
            est.p <- Model(full.par, n, fixed, ident)	            # Estimated probabilities
            est.h <- est.p * HTSDT.Rank.Compute.Obs.N(data, n)	  # Estimated frequencies (File: SDT-Rank.R)
            df.dat<- HTSDT.Rank.Compute.df.data(n)                # Number of free data points (File: SDT-Rank.R)
      } 
      else 
      {
            est.p <- Model(full.par, n, fixed, ident)	            # Estimated probabilities
            est.h <- Compute.Est.N(est.p, data)	                  # Estimated frequencies (File: SDT-Auxiliary.R)
            df.dat<- Compute.df.data(est.p) 				              # Number of free data points (File: SDT-Auxiliary.R)
      }
     	nz <- est.h > 0                                             # Criterion for inclusion int pearson statistic (= non zero estimate)
      nzd <- nz & (data > 0)                                      # Criterion for inclusion into LR statistic

 	Res <- numeric(0)
	Res[1] <- -Opti$value							# 1. log likelihood
	
	Res[2] <- sum((est.h[nz] - data[nz])^2/est.h[nz])		# 2. Pearson statistic X^2
	Res[3] <- 2*sum(data[nzd]*log(data[nzd]/est.h[nzd]))	      # 3. Likelihood ratio statistics G^2
	Res[4] <- df.dat - q							# 4. Degrees of freedom
	Res[5] <- ifelse(Res[4] > 0, 1- pchisq(Res[2], Res[4]), NA)	# 5. p of X^2-statistic
	Res[6] <- ifelse(Res[4] > 0, 1- pchisq(Res[3], Res[4]), NA)	# 6. p of G^2-statistic
	Res[7] <- 2*Opti$value + 2*q						# 7. AIC
 	Res[8] <- 2*Opti$value + q*log(N)					# 8. BIC
 	Res[9] <- NA                                                # 9. CAICF
  Res[10]<- NA                                                # 10.ICOMP
  Res[11]<- NA                                                # 11. ICOMP.R
	Res[12]<- q								      # 12. Number of free parameters
	Res[13]<- Opti$SDT.gr.sym.len 					# 13.Length of symbolic gradient at optimum
	Res[14]<- NA								# 14.Rank of the Hessian matrix
	Res[15]<- NA								# 15.Condition number of the Hessian matrix
	Res[16]<- NA                                                # 16.Rank of the t(J) * J
	Res[17]<- NA                                                # 17.Condition number of the t(J) * J


	Res <- matrix(Res, ncol = 1)
	rownames(Res) <- c("log L", "X^2", "G^2", "df", "p(Y > X^2)", "p(Y > G^2)", "AIC", "BIC", "CAICF", "ICOMP", "ICOMP.R", "Free Parameters", 
				 "Length of gradient at optimum", "Rank of Hessian", "Condition number of information matrix",
                         "Rank of model matrix: t(J) * J", "Condition number of model matrix")
	colnames(Res) <- "Statistic"

    # Test of Model matrix
      MM.Analytic <- F
      MM <- NULL

      if(!is.null(Model.Matrix))
      {
         MM <- Model.Matrix(full.par, n, fixed, ident, functional)      # Analytical computation of model matrix
         if(!is.null(MM))     MM.Analytic <- T
      }
      if(is.null(MM) && !is.null(Model.Matrix.Num))
      {
            MM <- Model.Matrix.Num(full.par, n, fixed, ident, functional)  # Numerical computation of model matrix
      }      
 	# Test of Model Matrix (File: "SDT-Auxiliary.R")
	if(!is.null(MM))
	{
            JJ <- t(MM) %*% MM
  }
  else
            JJ <- NULL

  Test.JJ <- Test.Hessian(JJ)
  Res[16] <- Test.JJ$rank
  Res[17] <- Test.JJ$kappa

 	# Test of Hessian matrices (File: "SDT-Auxiliary.R")
	Test.H   <- Test.Hessian(Opti$hessian)
	Res[14]  <- Test.H$rank
	Res[15]  <- Test.H$kappa
	Cov.asym <- NULL                                      # Set asymptotic covariance matrix to zero

	if(Test.H$P.d == T)						# Hessian positive definite
	{
            Cov.asym <- solve(Opti$hessian)		# Asymptotic covariance matrix
            if(Model.Id == "SDT.2D" || Model.Id == "SDT2D.MIX.2" || Model.Id == "SDT2D.MIX.3")
            {
                  Cov.asym <- SDT.2D.Adjust.robust(Cov.asym, full.par, n, fixed, ident)   # SDT-2D/ SDT2D-MIX3: Adjust asymptotic covariance matrix in case of robust thresholds
                  Res[15]  <- kappa(Cov.asym, exact = T)
            }      
            if(Model.Id == "SDT.Robust")
            {
                  Cov.asym <- SDT.Adjust.Robust(Cov.asym, full.par, n, fixed, ident)   # SDT-Robust: Adjust asymptotic covariance matrix in case of robust thresholds
                  Res[15]  <- kappa(Cov.asym, exact = T)
            }
            if(Model.Id == "SDT2D.MIX.2")
            {
                  Delta.M <-  SDT2D.MIX.2.Delta(full.par, n, fixed, ident)      # SDT2D-MIX3: Adjust estimated covariances for probabilities
                  Cov.asym <- Delta.M %*% Cov.asym %*% t(Delta.M)               # Delta rule
                  Res[15]  <- kappa(Cov.asym, exact = T)
            }
            if(Model.Id == "SDT2D.MIX.3")
            {
                  Delta.M <-  SDT2D.MIX.3.Delta(full.par, n, fixed, ident)      # SDT2D-MIX3: Adjust estimated covariances for probabilities
                  Cov.asym <- Delta.M %*% Cov.asym %*% t(Delta.M)               # Delta rule
                  Res[15]  <- kappa(Cov.asym, exact = T)
            }
            var.asym <- diag(Cov.asym)          # Asymptotic variances
            sd.asym  <- sqrt(var.asym)		      # Compute standard errors of probability parameters

		      if(length(sd.asym) == 1) Sd.Asym.Inv <- 1/sd.asym	
        	else                     Sd.Asym.Inv <- diag(1/sd.asym)
        	
            R.asym <- Sd.Asym.Inv %*% Cov.asym %*% Sd.Asym.Inv          # Asymptotic correlation matrix of estimated parameters
			
        		# Compute Measures CAICF, ICOMP, ICOMP.R
          Res[9]  <- 2*Opti$value + q*(log(N)+ 2) + log(det(Opti$hessian))    # 9.  CAICF
        	Res[10] <- 2*Opti$value + q*log(sum(var.asym)/q) - log(det(Cov.asym))	# 10. ICOMP
         	Res[11] <- 2*Opti$value - log(det(R.asym))				      # 11. ICOMP.R
	}
	else
	{
          	sd.asym <- rep(NA, q)
	}
 
  # Display warning
  if(display.warning == T && !Test.H$P.d)
      {
            if(!is.na(Res[14]))
                  cat("\n\n ===> WARNING: Hessian matrix is not positive definite: Rank =", Res[14], "\n\n")
      }
      else if(display.warning == T && !Test.JJ$P.d)
      {
            if(!is.na(Res[16]))
                  cat("\n\n ===> WARNING: Jacobian matrix [t(J) * J] is not positive definite: Rank =", Res[16], "\n\n")
      }
      # Parameters, Standard deviations, Confidence intervals
      Pars <- matrix(c(Opti$par, sd.asym), nc = 2)
      colnames(Pars) <- c("Value", "SE")

      if(!is.null(par.names.est)) 
            rownames(Pars) <- par.names.est
      else 
            rownames(Pars) <- paste("P-", 1:dim(Pars)[1], sep = "")

      if(Model.Id == "SDT" | Model.Id == "Gaussian")
            Pars <- SDT.Format.Par.est.SDT(Pars, full.par, Model.Id, n, fixed, ident, conf, deci)
      else if(Model.Id == "IRF.Gauss") 
            Pars <- SDT.Format.Par.est.IRF.Gauss(Pars, full.par, Model.Id, n, fixed, ident, conf, deci)
      else if(Model.Id == "IRF.LC") 
            Pars <- SDT.Format.Par.est.IRF.LC(Pars, full.par, Model.Id, n, fixed, ident, conf, deci)
      else if(Model.Id == "SDT.Robust")
            Pars <- SDT.Format.Par.est.Robust(Pars, full.par, Model.Id, n, fixed, ident, conf, deci)
      else if(Model.Id == "MIX.2")
            Pars <- SDT.Format.Par.est.MIX2(Pars, full.par, data, n, fixed, ident, functional, conf, deci)
      else if(Model.Id == "MIX.PD")
            Pars <- SDT.Format.Par.est.MIX.PD(Pars, full.par, data, n, fixed, ident, functional, conf, deci)
      else if(Model.Id == "DPSDT")
            Pars <- SDT.Format.Par.est.DPSDT(Pars, full.par, data, n, fixed, ident, functional, conf, deci)
      else if(Model.Id == "HTSDT.1")
            Pars <- SDT.Format.Par.est.HTSDT.1(Pars, full.par, data, n, fixed, ident, functional, conf, deci)
      else if(Model.Id == "HT.n")
            Pars <- SDT.Format.Par.est.HT.n(Pars, full.par, Cov.asym, n, fixed, ident, functional, conf, deci)
      else if(Model.Id == "SDT.2D")
            Pars <- SDT.Format.Par.est.SDT2D(Pars, full.par, n, fixed, ident, conf, deci)
      else if(Model.Id == "SDT2D.MIX.2")
            Pars <- SDT2D.MIX.2.Format.Par.est(Pars, full.par, n, fixed, ident, conf, deci)
      else if(Model.Id == "SDT2D.MIX.3")
            Pars <- SDT2D.MIX.3.Format.Par.est(Pars, full.par, n, fixed, ident, conf, deci)
      else if(Model.Id == "mAFC")
            Pars <- SDT.mAFC.Format.Par.est(Pars, full.par, fixed, ident, conf, deci)
      else if(Model.Id == "SDT.Rank")
            Pars <- SDT.Rank.Format.Par.est(Pars, full.par, n, fixed, ident, conf, deci)
      else if(Model.Id == "HTSDT.Rank")
            Pars <- HTSDT.Rank.Format.Par.est(Pars, full.par, Cov.asym, n, fixed, ident, conf, deci)
      else if(Model.Id == "HTSDT.Bias.Rank")
            Pars <- HTSDT.Bias.Rank.Format.Par.est(Pars, full.par, Cov.asym, n, fixed, ident, conf, deci)
      else if(Model.Id == "MIX.Rank")
            Pars <- MIX.Rank.Format.Par.est(Pars, full.par, Cov.asym, n, fixed, ident, conf, deci)
      else if(Model.Id == "MIX.Bias.Rank")
            Pars <- MIX.Bias.Rank.Format.Par.est(Pars, full.par, Cov.asym, n, fixed, ident, conf, deci)
      else
            Pars <- round(Pars, deci)

	# If required, compute confidence bounds
	if(!is.null(control) & !is.null(control$LR))
	{
            if(control$LR)
            {
                  if(Model.Id == "SDT" | Model.Id == "Gaussian")
                        se.col <- 2
                  else
                        se.col <- 3
                        
                  Erg <- SDT.LR.CFI(Opti.Obj, se = Pars[, se.col], conf, control = control, deci)
                  if(!is.null(Erg))
                  { 
                        cnames <- c(colnames(Pars), paste("LR-", conf*100, "(Lower)", sep =""), paste("LR-", conf*100, "(Upper)", sep =""))
                        Pars <- data.frame(cbind(Pars, Erg))
                        colnames(Pars) <- cnames
                  }
            }	     
	}
	
 	# Full-Parameter vector
	if(Model.Id == "SDT")                    Fp <- SDT.Format.Par.SDT(full.par, n, deci)
	else if(Model.Id == "SDT.Robust") 	     Fp <- SDT.Format.Par.Robust(full.par, n, deci)
	else if(Model.Id == "Gaussian")	         Fp <- SDT.Format.Par.Gaussian(full.par, n, deci)
	else if(Model.Id == "MIX.2")		         Fp <- SDT.Format.Par.MIX2(full.par, n, fixed, deci)
	else if(Model.Id == "MIX.PD")		         Fp <- SDT.Format.Par.MIX.PD(full.par, n, fixed, deci)
	else if(Model.Id == "DPSDT")             Fp <- SDT.Format.Par.DPSDT(full.par, n, fixed, deci)
	else if(Model.Id == "HTSDT.1")	         Fp <- SDT.Format.Par.HTSDT.1(full.par, n, fixed, deci)
	else if(Model.Id == "HT.n")		           Fp <- SDT.Format.Par.HT.n(full.par, n, fixed, ident, deci)
	else if(Model.Id == "SDT.2D")            Fp <- SDT.Format.Par.SDT2D(full.par, n, deci)
	else if(Model.Id == "SDT2D.MIX.2")       Fp <- SDT.Format.Par.SDT2D.MIX.2(full.par, n, deci)
	else if(Model.Id == "SDT2D.MIX.3")       Fp <- SDT.Format.Par.SDT2D.MIX.3(full.par, n, deci)
	else if(Model.Id == "SDT.Rank")          Fp <- SDT.Format.Par.Rank(full.par, n, deci)
	else if(Model.Id == "HTSDT.Rank")        Fp <- HTSDT.Format.Par.Rank(full.par, n, deci, fixed, ident)
	else if(Model.Id == "HTSDT.Bias.Rank")   Fp <- HTSDT.Bias.Format.Par.Rank(full.par, n, deci, fixed, ident)
	else if(Model.Id == "MIX.Rank")          Fp <- MIX.Format.Par.Rank(full.par, n, deci, fixed, ident)
	else if(Model.Id == "MIX.Bias.Rank")     Fp <- MIX.Bias.Format.Par.Rank(full.par, n, deci, fixed, ident)
	else if(Model.Id == "IRF.Gauss")         Fp <- IRF.Gauss.Format.Par(full.par, n, deci)
	else if(Model.Id == "IRF.LC")            Fp <- IRF.LC.Format.Par(full.par, n, deci)
	else
	{	# Default mode
		Fp 	<- matrix(round(full.par, deci), nc = 1)
		colnames(Fp) <- c("Parameters")
		if(!is.null(par.names))     rownames(Fp) <- par.names
		else                        rownames(Fp) <- 1:(dim(Fp)[1])
	}
  # If SDT model or Gaussian IRF model insert section with additional measures
	Add.measures <- NULL
	if(Model.Id == "SDT")              Add.measures <- SDT.measures(full.par, Cov.asym, n, fixed, ident, functional, deci, conf)
	if(Model.Id == "IRF.Gauss")        Add.measures <- IRF.SDT.measures(full.par, Cov.asym, n, fixed, ident, functional, deci, conf)

	# 11. Estimated probabilities and frequencies and input data
	P <- est.p
	H <- est.h
	D <- data

	# 12. Residuals
	Tol.H <- 1e-16
	H[H < Tol.H] <- Tol.H
	Resid <- round((D - H)/sqrt(H), deci)

	# 13. Bind results together
	Freq 	<- cbind(round(D, deci), round(H, deci), round(P, deci), round(Resid, deci))
	colnames(Freq) <- c("Observed", "Expected", "Probabilities", "Residuals")

  # 13.0 Compute Standardized residuals
 	Freq <- SDT.Standard.Residuals(Model.Id, full.par, Freq, MM, data.full, est.p, n, deci, MM.Analytic)

  # 13.1 Construct row names if n is a list or items including s = number of signals
	if(is.list(n))
	{
	      s <- n$s
        if(Model.Id == "SDT.2D")   s <- n$n.sdt
            
	      if(is.null(n$R))
	      {
                  if(Model.Id == "SDT")
                        rownames(Freq) <- SDT.Datalabels(n, dim(Freq)[1])
                  else if(Model.Id == "SDT.Robust")
                        rownames(Freq) <- SDT.Datalabels(n, dim(Freq)[1])
                  else if(Model.Id == "Gaussian")
                        rownames(Freq) <- Gaussian.Datalabels(n, dim(Freq)[1])
                  else if(Model.Id == "DPSDT")
                        rownames(Freq) <- DPSDT.Datalabels(n, dim(Freq)[1])
                  else if(Model.Id == "SDT.Rank")
                        rownames(Freq) <- SDT.Rank.Datalabels(n)
                  else if(Model.Id == "HTSDT.Rank")
                        rownames(Freq) <- HTSDT.Rank.Datalabels(n)
                  else if(Model.Id == "HTSDT.Bias.Rank")
                        rownames(Freq) <- HTSDT.Bias.Rank.Datalabels(n, dim(Freq)[1])
                  else if(Model.Id == "MIX.Rank")
                        rownames(Freq) <- MIX.Rank.Datalabels(n, dim(Freq)[1])
                  else if(Model.Id == "MIX.Bias.Rank")
                        rownames(Freq) <- MIX.Bias.Rank.Datalabels(n, dim(Freq)[1])
                  else if(Model.Id == "IRF.Gauss")
                        rownames(Freq) <- IRF.Gauss.Datalabels(n, nrow(Freq))
                  else if(Model.Id == "IRF.LC")
                        rownames(Freq) <- IRF.LC.Datalabels(n, nrow(Freq))
                  else
                  {
                      if(is.null(s))   stop("\n FATAL ERROR: Number of signals not found\n")
		                  n.dat <- dim(Freq)[1]/s
		                  Datname <- NULL
		                  for(i in 1:s)	Datname <- c(Datname, paste("Data[", i, ", ", 1:n.dat, "]", sep=""))
		                  rownames(Freq) <- Datname
                  }
            }
            else rownames(Freq) <- 1:dim(Freq)[1]
	} 
	
	# 13.2 Construct row names if n is a scalar indicating the number of signals
	else if(length(n) < 2)
	{
		n.dat <- dim(Freq)[1]/n
		Datname <- NULL
		for(i in 1:n)	Datname <- c(Datname, paste("Data[", i, ", ", 1:n.dat, "]", sep=""))
		rownames(Freq) <- Datname
	}
	# 13.3 Construct row names if n is a vector indicating the number of data points per signal
	else
	{
		Datname <- NULL
		l.n <- length(n)
		for(i in 1:l.n)	Datname <- c(Datname, paste("Data[", i, ", ", 1:n[i], "]", sep=""))
		rownames(Freq) <- Datname
	}
  if(Model.Id == "SDT.2D" || Model.Id == "SDT2D.MIX.2" || Model.Id == "SDT2D.MIX.3")
            return(list(Model.description =   Model.Name,
    		            Statistics =          round(Res, deci),
                        Free.parameters =     Pars, 
                        Gaussian.parameters = Fp$GP,
                        Decision.bounds.D1  = Fp$DB.1,
                        Decision.bounds.D2  = Fp$DB.2,
                        Data.and.Estimates =  SDT.Format.Freq.SDT2D(Freq, n))
                  )
   if(Model.Id == "SDT.Rank" || Model.Id == "HTSDT.Rank")
      {
           if(is.null(Fp$Rating))
                 return(list(Model.description   = Model.Name,
                              Statistics         = round(Res, deci),
                              Free.parameters    = Pars,
                              Ranking.parameters = Fp$Ranking,
                              Data.and.Estimates = Freq))

           return(list(Model.description  = Model.Name,
                       Statistics         = round(Res, deci),
                       Free.parameters    = Pars,
                       Ranking.parameters = Fp$Ranking,
                       Rating.parameters  = Fp$Rating,
                       Decision.bounds    = Fp$Thresholds,
                       Data.and.Estimates = Freq))
      }
      if(Model.Id == "HTSDT.Bias.Rank" || Model.Id == "MIX.Rank"   || Model.Id == "MIX.Bias.Rank")
      {
            if(is.null(Fp$Rating))
                  return(list(Model.description  = Model.Name,
                              Statistics         = round(Res, deci),
                              Free.parameters    = Pars, 
                              Ranking.parameters = Fp$Ranking,
                              Data.and.Estimates = Freq))
            return(list(Model.description  = Model.Name,
                        Statistics         = round(Res, deci),
                        Free.parameters    = Pars, 
                        Ranking.parameters = Fp$Ranking,
                        Rating.parameters  = Fp$Rating,
                        Data.and.Estimates = Freq))
      }
      if(Model.Id == "IRF.Gauss")
      {
            if(is.null(Add.measures))
                  return(list(Model.description   = Model.Name,
                              Statistics          = round(Res, deci),
                              Free.parameters     = Pars, 
                              Gaussian.parameters = Fp$Gauss.Parameters,
                              Response.parameters = Fp$Response.Parameters,
                              Data.and.Estimates  = Freq))
            return(list(Model.description   = Model.Name,
                        Statistics          = round(Res, deci),
                        Free.parameters     = Pars, 
                        Gaussian.parameters = Fp$Gauss.Parameters,
                        Response.parameters = Fp$Response.Parameters,
                        SDT.measures        = Add.measures,
                        Data.and.Estimates  = Freq))
      }

      if(Model.Id == "IRF.LC")
      {
            return(list(Model.description   = Model.Name,
                        Statistics          = round(Res, deci),
                        Free.parameters     = Pars, 
                        LC.parameters       = Fp$LC.parameters,
                        Response.parameters = Fp$Response.Parameters,
                        Data.and.Estimates  = Freq))
      }
      # Default
      if(is.null(Add.measures))
            return(list(Model.description = Model.Name,
	                 Statistics = round(Res, deci),
	                 Free.parameters = Pars, 
	                 Full.parametervector = Fp, 
	                 Data.and.Estimates = Freq)
	                 )
      # Default: SDT Model
      list(Model.description = Model.Name,
    		          Statistics = round(Res, deci),
		         Free.parameters = Pars, 
        Full.parametervector = Fp,
                SDT.measures = Add.measures, 
          Data.and.Estimates = Freq)
}


# ------------------------------------------------------------------------------------------------------------------
# Computation of standardized Residuals
# Input:	1. Model identification string
#           2. Full parmeter vector
#           3. Matrix with (rounded)
#                 (i)   observed frequences
#                 (ii)  estimated frequencies, 
#                 (iii) estimated probabilities
#                 (iv)  Pearson residuals
#           4. Model matrix
#		5. Data vector
#           6. Estimated probabilities
#		7. Number of signals or configuration list
#           8. Number of decimal places
#           9. Flag indicating whether model matrix was computed analytically
#
# Output: New Matrix Freq with tow columns containing standardized residuals (numerically and analytically)
# ------------------------------------------------------------------------------------------------------------------
SDT.Standard.Residuals <- function(Model.Id, full.par, Freq, MM, dat, est.p, n, deci, MM.Anal)
{
      Tol <- 1e-7

      if(is.null(MM))
            return(Freq)
      
      # 1. Compute observed probabilities and sqare roots of marginal counts (Auxiliary.R)
      if(Model.Id == "SDT2D.MIX.2" || Model.Id == "SDT2D.MIX.3")
      {
            p      <- Compute.Obs.p(dat, n$s)
            N.sqrt <- sqrt(Compute.Obs.N(dat, n$s))
            if(!is.null(n$R)) 
            {
                  p <- n$R %*% p
                  N.sqrt <- n$R %*% N.sqrt
            }
      }      
      else if(Model.Id == "SDT" || Model.Id == "SDT.Robust" || Model.Id == "Gaussian" || Model.Id == "DPSDT" || Model.Id == "SDT.2D" || Model.Id == "IRF.Gauss" || Model.Id == "IRF.LC")
      {
            p      <- Compute.Obs.p(dat, n$n.sdt)
            N.sqrt <- sqrt(Compute.Obs.N(dat, n$n.sdt))
            if(!is.null(n$R)) 
            {
                  p <- n$R %*% p
                  N.sqrt <- n$R %*% N.sqrt
            }
      }      
      else if(Model.Id == "mAFC")
      {
            nclusters <-  n$s * n$m
            p      <- Compute.Obs.p(dat, nclusters)
            N.sqrt <- sqrt(Compute.Obs.N(dat, nclusters))
      }
      else if(Model.Id == "SDT.Rank")
      {
            p <- SDT.Rank.Compute.Obs.p(dat, n)
            N.sqrt <- sqrt(SDT.Rank.Compute.Obs.N(dat, n))
      }
      else if(Model.Id == "HTSDT.Rank")
      {
            p <- HTSDT.Rank.Compute.Obs.p(dat, n)
            N.sqrt <- sqrt(HTSDT.Rank.Compute.Obs.N(dat, n))
      }
      else if(Model.Id == "HTSDT.Bias.Rank")
      {
            p <- HTSDT.Bias.Rank.Compute.Obs.p(dat, n)
            N.sqrt <- sqrt(HTSDT.Bias.Rank.Compute.Obs.N(dat, n))
      }
      else if(Model.Id == "MIX.Rank")
      {
            p <- MIX.Rank.Compute.Obs.p(dat, n)
            N.sqrt <- sqrt(MIX.Rank.Compute.Obs.N(dat, n))
      }
      else if(Model.Id == "MIX.Bias.Rank")
      {
            p <- MIX.Bias.Rank.Compute.Obs.p(dat, n)
            N.sqrt <- sqrt(MIX.Bias.Rank.Compute.Obs.N(dat, n))
      }
       else
      {
            p      <- Compute.Obs.p(dat, n)
            N.sqrt <- sqrt(Compute.Obs.N(dat, n))
      }
      p.sqrt <- sqrt(est.p)
      p.sqrt[p.sqrt <= 0] <- Tol       # Correct for zeros

      # 2. Compute residuals
      r.raw <- p - est.p
      r <- N.sqrt * r.raw / p.sqrt
      
      # 3. Compute Scale factor for Jacobian matrices
      D.p <- diag(1/p.sqrt)
      
      # 4. Compute scaled model matrix
      A <- D.p %*% MM
 
      # 5. Test scaled model matrix
      Tst <- Test.Hessian(t(A) %*% A)

      # 6. Compute Covariance matrix of residuals
      O <- - p.sqrt %*% t(p.sqrt)
      diag(O) <- diag(O) + 1
      if(Tst$P.d)  O <- O - A %*% solve(t(A) %*% A) %*% t(A)
      
       # 7. Compute possible corrections
      if(Tst$P.d)
      {
            diag.O <- diag(O)
            diag.O[abs(r) < Tol | abs(r.raw) < Tol] <- 1
      }
      # 7. Compute standarized residuals
      if(Tst$P.d)       
      {
            if(any(diag.O < 0))     Res.adj <- rep(NA, length(r))
            else                    Res.adj <- r/sqrt(diag.O)
      }
      else        Res.adj <- rep(NA, length(r))
      
      # 8. Append vectors and adjust colnames
      if(MM.Anal) cn <- c(colnames(Freq), " Std.Resid.(analytical)")
      else        cn <- c(colnames(Freq), " Std.Resid.(numerical)")
      
      Freq <- cbind(Freq, round(Res.adj, deci))
      colnames(Freq) <- cn
      
      Freq
}

# ------------------------------------------------------------------------------------------------------------------
# Comparison of two Statistic objects
# Input:	1. Statistic Object 1
#		2. Statistic Object 2 (of a Model that is nested within the Model of Objekt 1)
#
# Output: A matrix with the following rows
#		1. G2, df, p (of Obj.1)
#		2. G2, df, p (of Obj.2)
#		3. Difference in G2, df, p (of Obj.1 - Obj.2)
#
#		If wrong Objects are transfered to the function a NULL is returned
#
# Comments: if df in Obj.2 < df in Obj.1 the objects are reversed
# ------------------------------------------------------------------------------------------------------------------
SDT.Anova <- function(SO.1, SO.2, deci = 3) 
{
	pos <- c(3, 4, 6)				# Positions of relevant statistics	

	if(is.null(SO.1$Statistics) || is.null(SO.2$Statistics)) return(NULL)
	
	# Extract items
	G1 <- SO.1$Statistics[pos]
	G2 <- SO.2$Statistics[pos]

	if(G2[2] > G1[2])	Erg <- rbind(G2, G1)
	else			Erg <- rbind(G1, G2)

	DG <- Erg[1, 1] - Erg[2, 1]
	df <- Erg[1, 2] - Erg[2, 2]
	G3 <- c(DG, df, 1-pchisq(DG, df))

	Erg <- rbind(Erg, G3)
	colnames(Erg) <- c("G2", "df", "p")
	rownames(Erg) <- c("Model 1", "Model 2", "Difference")
	
	round(Erg, deci)
}


# ------------------------------------------------------------------------------------------------------------------
# Computation of a tableau with parameters their names and information about fixed and idenitity constraints
# Input:	1. Data vector 
#		2. Parameter vector
#		3. Number of models
#		4. Model identification string
#		5. Matrix of fixed constraints
#		6. Matrix of identity constraint
#           7. Number of decimal places
# Output: A list with model identification string and information about parameters
# ------------------------------------------------------------------------------------------------------------------
SDT.Parameter.Info <- function(data = NULL, par = NULL, n = 2, Model.Id = "SDT", fixed = NULL, ident = NULL, deci = 3)
{
      Cst.new <- Ident.to.fixed(fixed, ident)			   # If necessary, transform identity constraints to fixed constraints
	    fixed   <- Cst.new$fixed							         #(Function in Auxiliary.R)
	    ident   <- Cst.new$ident
     	if(Model.Id == "SDT") 					               # Model = Standard signal detection model
	    {
            Cmp.par	   <- SDT.SDT.start.par		       # Function for generating starting parameter
		        Cmp.names  <- SDT.par.names			         # Names of parameters
		        n          <- SDT.SDT.Check.Fill.Cfg(n)  # Generate configuration list
            if(is.null(n) || !is.list(n))
                  stop("\n SDT.Parameter.Info(): Missing configuration information for SDT model: Cannot compute parameter information\n")
            Model.Name  <- SDT.model.name(n)		     # Name of the model
            if(is.null(par) && is.null(data))
                   stop("\n SDT.Parameter.Info(): Neither parameter nor data available: Cannot compute parameter information\n")
      }
	    else if(Model.Id == "Gaussian") 			         # Model = SDT model with entirely free parameters
	    {
		       Cmp.par 	 <- SDT.Gaussian.start.par	     # Function for generating starting parameter
		       Cmp.names <- Gaussian.par.names		       # Function for generating parameter names
 		       n         <- SDT.Gaussian.Check.Fill.Cfg(n)  # Generate configuration list
            if(is.null(n) || !is.list(n))
                  stop("\n SDT.Parameter.Info(): Missing configuration information for Gaussian model: Cannot compute parameter information\n")
            if(is.null(par) && is.null(data))
                  stop("\n Gaussian.Parameter.Info: Neither parameter nor data available: Cannot compute parameter information\n")
           Model.Name  <- Gaussian.model.name(n)	   # Name of the model
      }
	    else if(Model.Id == "DPSDT")			             # Model = Dual process model
	    {
          Cmp.par 	<- DPSDT.start.par	             # Function for generating starting parameter
		      Cmp.names <- DPSDT.par.names		           # Function for generating parameter names
 		      n         <- DPSDT.Check.Fill.Cfg(n, data) # Generate configuration list
            if(is.null(n) || !is.list(n))
                  stop("\n SDT.Parameter.Info(): Missing configuration information for DPSDT model: Cannot compute parameter information\n")
            if(is.null(par) && is.null(data))
                  stop("\n DPSDT.Parameter.Info: Neither parameter nor data available: Cannot compute parameter information\n")
		              Model.Name  <- DPSDT.model.name(n) # Name of the model
      }
	    else if(Model.Id == "MIX.PD") 			           # Model = MIX model with p and d free parameters
	    {
          Cmp.par 	 <- SDT.MIX.PD.start.par		     # Function for generating starting parameter
		      Cmp.names	 <- MIX.PD.par.names		         # Function for generating parameter names
		      Model.Name <- MIX.PD.model.name 		       # Name of the model
      }
	    else if(Model.Id == "MIX.2") 				           # Model = MIX2 model with entirely free parameters
	    {

		      Cmp.par    <- SDT.MIX2.start.par		       # Function for generating starting parameter
		      Cmp.names	 <- MIX2.par.names			         # Function for generating parameter names
		      Model.Name <- MIX2.model.name 		         # Name of the model
      }
	    else if(Model.Id == "HTSDT.1")			           # Model = HTSDT.1 model
	    {
		       Cmp.par 	   <- SDT.HTSDT.1.start.par	     # Function for generating starting parameter
		       Cmp.names	 <- HTSDT.1.par.names		       # Function for generating parameter names
		       Model.Name  <- HTSDT.1.model.name 		     # Name of the model
      }
	    else if(Model.Id == "HT.n")				             # Model = High threshold model with rating
	    {
		       Cmp.par 	  <- SDT.HT.n.start.par		       # Function for generating starting parameter
		       Cmp.names	<- HT.n.par.names			         # Function for generating parameter names
		       Model.Name <- HT.n.model.name 		         # Name of the model
      }
	    else if(Model.Id == "SDT.2D")				           # Model = Bivariate Gaussian signal detections model
	    {
		       n         <- SDT.SDT2D.Check.Fill.Cfg(n)  # Generate configuration list
		       Cmp.par 	 <- SDT.SDT2D.start.par		       # Function for generating starting parameter
		       Cmp.names <- SDT2D.par.names		           # Function for generating parameter names
           if(is.null(n) || !is.list(n))
                  stop("\n SDT.Parameter.Info(): Missing configuration information for SDT-2D model: Cannot compute parameter vector\n")
           Model.Name  <- SDT2D.model.name(n)	       # Name of the model
      }
	    else if(Model.Id == "SDT2D.MIX.2")			       # Model = Bivariate Gaussian signal detections model
	    {
		       Cmp.par 	  <- SDT2D.MIX.2.start.par	     # Function for generating starting parameter
		       Cmp.names	<- SDT2D.MIX.2.par.names	     # Function for generating parameter names
           if(is.null(n) || !is.list(n))
                  stop("\n SDT.Parameter.Info(): Missing configuration information for SDT2D-MIX-2 model: Cannot compute parameter vector\n")
           Model.Name <- SDT2D.MIX.2.model.name(n)  # Name of the model
      }
	    else if(Model.Id == "SDT2D.MIX.3")			      # Model = Bivariate Gaussian mixture signal detections model
	    {
		       Cmp.par 	<- SDT2D.MIX.3.start.par	      # Function for generating starting parameter
		       Cmp.names	<- SDT2D.MIX.3.par.names	    # Function for generating parameter names
           if(is.null(n) || !is.list(n))
                  stop("\n SDT.Parameter.Info(): Missing configuration information for SDT2D-MIX-3 model: Cannot compute parameter vector\n")
           Model.Name  <- SDT2D.MIX.3.model.name(n) # Name of the model
      }
	    else if(Model.Id == "mAFC")			              # Model = Bivariate Gaussian signal detections model
	    {
		       Cmp.par 	   <- SDT.mAFC.start.par	      # Function for generating starting parameter
		       Cmp.names	 <- SDT.mAFC.par.names	      # Function for generating parameter names
           if(is.null(n) || !is.list(n) || is.null(n$m) || is.null(n$s))
                  stop("\n Missing configuration information for SDT.mAFC: Cannot compute parameter vector\n")
           Model.Name <- SDT.mAFC.model.name(n)     # Name of the model
      }
      else if(Model.Id == "SDT.Rank")			          # Model = SDT model for ranking and rating
	    {
           n          <- SDT.Rank.Check.Fill.Cfg(n) # Generate / complete configuration list
           Cmp.par 	  <- SDT.Rank.start.par	        # Function for generating starting parameter
		       Cmp.names  <- SDT.Rank.par.names	        # Function for generating parameter names

           if(is.null(n) || !is.list(n) || is.null(n$j.rg))
                  stop("\n SDT.Parameter.Info(): Missing configuration information for SDT.Rank: Cannot compute parameter vector\n")
           Model.Name <- SDT.Rank.model.name(n)             # Name of the model
      }
	    else if(Model.Id == "HTSDT.Rank")			                # Model = High-threshold-SDT model for ranking and rating
	    {
		       Cmp.par 	  <- HTSDT.Rank.start.par	              # Function for generating starting parameter
		       Cmp.names	<- HTSDT.Rank.par.names	              # Function for generating parameter names
		       n          <- HTSDT.Rank.Check.Fill.Cfg(n)       # Generate configuration list
           if(is.null(n) || !is.list(n) || is.null(n$j.rg))
                  stop("\n HTSDT.Parameter.Info(): Missing configuration information for HTSDT.Rank: Cannot compute parameter vector\n")
		       Model.Name <- HTSDT.Rank.model.name(n)           # Name of the model
      }
	    else if(Model.Id == "HTSDT.Bias.Rank")			          # Model = High-threshold-SDT model for ranking and rating
	    {                                                     #         with position bias
           Cmp.par 	  <- HTSDT.Bias.Rank.start.par	        # Function for generating starting parameter
		       Cmp.names	<- HTSDT.Bias.Rank.par.names	        # Function for generating parameter names
		       n          <- HTSDT.Bias.Rank.Check.Fill.Cfg(n)  # Generate configuration list
		       Model.Name <- HTSDT.Bias.Rank.model.name(n)      # Name of the model
           if(is.null(par) && is.null(data) && n$rating == T)
		                       stop("\n SDT.Parameter.Info(): Neither parameter nor data available: Cannot compute parameter vector\n")
      }
	    else if(Model.Id == "MIX.Rank")			                  # Model = Mixture model for ranking and rating
	    {
		       Cmp.par 	  <- MIX.Rank.start.par	                # Function for generating starting parameter
		       Cmp.names	<- MIX.Rank.par.names	                # Function for generating parameter names
		       n          <- MIX.Rank.Check.Fill.Cfg(n)         # Generate configuration list
		       Model.Name <- MIX.Rank.model.name(n)             # Name of the model
           if(is.null(par) && is.null(data) && n$rating == T)
   		                     stop("\n SDT.Parameter.Info(): Neither parameter nor data available: Cannot compute parameter vector\n")
   	  }
	    else if(Model.Id == "MIX.Bias.Rank")		              # Model = Mixture model for ranking and rating
	    {                                                     #         with position bias
           Cmp.par 	<- MIX.Bias.Rank.start.par	            # Function for generating starting parameter
		       Cmp.names	<- MIX.Bias.Rank.par.names	          # Function for generating parameter names
		       n           <- MIX.Bias.Rank.Check.Fill.Cfg(n)   # Generate configuration list
		       Model.Name  <- MIX.Bias.Rank.model.name(n)       # Name of the model
           if(is.null(par) && is.null(data) && n$rating == T)
   		                     stop("\n SDT.Parameter.Info: Neither parameter nor data available: Cannot compute parameter vector\n")
   	  }
	    else if(Model.Id == "SDT.CAT")		                    # Model = SDT categorization model
	    {
		       Cmp.par 	  <- SDT.CAT.start.par                  # Function for generating starting parameter
		       Cmp.names	<- SDT.CAT.par.names	                # Function for generating parameter names
		       n          <- SDT.CAT.Check.Fill.Cfg(n)          # Generate configuration list
		       Model.Name <- SDT.CAT.model.name(n)              # Name of the model
            if(is.null(par) && is.null(data))
   		    stop("\n SDT.Parameter.Info: Neither parameter nor data available: Cannot compute parameter vector\n")
   	}
	    else if(Model.Id == "IRF.Gauss")		              # Model = Gaussian SDT with probabilisitic IRF
	    {
		       Cmp.par 	  <- IRF.Gauss.start.par            # Function for generating starting parameter
		       Cmp.names	<- IRF.Gauss.par.names	          # Function for generating parameter names
		       n          <- IRF.Gauss.Check.Fill.Cfg(n)    # Generate configuration list
		       Model.Name <- IRF.Gauss.model.name(n)        # Name of the model
            if(is.null(par) && is.null(data))
   		           stop("\n SDT.Parameter.Info: Neither parameter nor data available: Cannot compute parameter vector\n")
   	}
	    else if(Model.Id == "IRF.LC")		                  # Model = Latent class SDT with probabilisitic IRF
	    {
		       Cmp.par 	  <- IRF.LC.start.par               # Function for generating starting parameter
		       Cmp.names	<- IRF.LC.par.names	              # Function for generating parameter names
		       n          <- IRF.LC.Check.Fill.Cfg(n)       # Generate configuration list
		       Model.Name <- IRF.LC.model.name(n)           # Name of the model
            if(is.null(par) && is.null(data))
   		           stop("\n SDT.Parameter.Info: Neither parameter nor data available: Cannot compute parameter vector\n")
   	}
    else
	  {
		    stop(paste("\n SDT.Parameter.Info(): Unknown Model identification: <", Model.Id, ">\n", sep = ""))
    }

	  # If no parameter vector is specified, compute parameter vector from data or from cfg information
	  if(is.null(par))
	  {
		                if(is.null(data) && Model.Id != "SDT.2D"  && Model.Id != "SDT2D.MIX.2"  && Model.Id != "SDT2D.MIX.3"
                             && Model.Id != "mAFC"            && Model.Id != "HTSDT.Bias.Rank" && Model.Id != "MIX.Rank"
                             && Model.Id != "MIX.Bias.Rank"   && Model.Id != "SDT.CAT" && Model.Id != "SDT.Rank" && Model.Id != "HTSDT.Rank")
                             stop("\n SDT.Parameter.Info(): Neither parameter nor data available: Cannot compute parameter vector\n")

                    if(Model.Id == "SDT.2D")
                    {
                             par <- Cmp.par(n)
                             par <- SDT2D.robust.thresholds(par, n)
                    }
                    else if(Model.Id == "SDT2D.MIX.2")
                    {
                             par <- Cmp.par(n)
                             par <- SDT2D.robust.thresholds(par, n, Gauss.par = 11)
                    }
                    else if(Model.Id == "SDT2D.MIX.3")
                    {
                             par <- Cmp.par(n)
                             par <- SDT2D.robust.thresholds(par, n, Gauss.par = 18)
                    }
                    else if(Model.Id == "mAFC" || Model.Id == "SDT.Rank" || Model.Id == "HTSDT.Rank")
                    {
                             par <- Cmp.par(n)
                    }
                    else
                    {
                             par <- Cmp.par(n, data)
                    }
	  }
	  # In case of SDT2D Model with robust estimation compute thresholds
    else  if(Model.Id == "SDT.2D")
	        par <- SDT2D.robust.thresholds(par, n)
    else  if(Model.Id == "SDT2D.MIX.2")
	        par <- SDT2D.robust.thresholds(par, n, Gauss.par = 11)
    else  if(Model.Id == "SDT2D.MIX.3")
	        par <- SDT2D.robust.thresholds(par, n, Gauss.par = 18)
       
    # Compute parameter names
    if(Model.Id == "SDT.2D" || Model.Id == "SDT2D.MIX.2" || Model.Id == "SDT2D.MIX.3" || Model.Id == "mAFC" || Model.Id == "SDT.Rank" || Model.Id == "HTSDT.Rank")
         p.names <- p.names.new <- Cmp.names(n)
    else
	       p.names <- p.names.new <- Cmp.names(par, n)
    # Compute information for fixed and identity constraints
	  fixed.Id <- rep("---", length(par))
	  ident.Id.source <- rep("---", length(par))
	  ident.Id.source.nr <- rep("", length(par))

    # Show standard restrictions
    if(Model.Id == "SDT2D.MIX.3" && (is.null(n$standard) || n$standard == T))
         fixed.Id[2:5] <- paste("<set>", c(0, 0, 1, 1))
    else if(Model.Id == "SDT2D.MIX.2" && (is.null(n$standard) || n$standard == T))
         fixed.Id[2:5] <- paste("<set>", c(0, 0, 1, 1))
    else if(Model.Id == "HTSDT.Rank" || Model.Id == "HTSDT.Bias.Rank" ||
            Model.Id == "MIX.Rank"   || Model.Id == "MIX.Bias.Rank"   ||
            Model.Id == "SDT"        || Model.Id == "Gaussian"        ||
            Model.Id == "SDT.2D"     || Model.Id == "DPSDT"           ||
            Model.Id == "SDT.Rank"   || Model.Id == "SDT.CAT"         ||
            Model.Id == "IRF.Gauss"  || Model.Id == "IRF.LC")
    {
            # New names without more than one space between words
            if(Model.Id == "SDT.2D" || Model.Id == "SDT.Rank" || Model.Id == "HTSDT.Rank")
                  p.names.new <- Cmp.names(n)
            else if(Model.Id == "DPSDT" || Model.Id == "IRF.Gauss" || Model.Id == "IRF.LC")
                  p.names.new <- Cmp.names(par, n)
            else 
                  p.names.new <- Cmp.names(par, n, F)
	      
            if(toupper(n$restriction) != "NO")
            {              
                  # Identify standard restrictions
                  if(Model.Id == "SDT")
                  {
                        CS <- SDT.SDT.Restriction(n, NULL, NULL)
                        CS.full <- SDT.SDT.Restriction(n, fixed, ident)
                  }
                  else if(Model.Id == "Gaussian")
                  {
                        CS <- SDT.Gaussian.Restriction(par, n, NULL, NULL, NULL)
                        CS.full <- SDT.Gaussian.Restriction(par, n, fixed, ident, NULL)
                  }
                  else if(Model.Id == "DPSDT")
                  {
                        CS <- DPSDT.Set.Restriction(par, n, NULL, NULL, NULL)
                        CS.full <- DPSDT.Set.Restriction(par, n, fixed, ident, NULL)
                  }
                  else if(Model.Id == "HTSDT.Rank")
                  {
                        CS <- HTSDT.Rank.Restriction(n, NULL, NULL)
                        CS.full <- HTSDT.Rank.Restriction(n, fixed, ident)
                  }
                  else if(Model.Id == "HTSDT.Bias.Rank")
                  {
                        CS <- HTSDT.Bias.Rank.Restriction(n, NULL, NULL)
                        CS.full <- HTSDT.Bias.Rank.Restriction(n, fixed, ident)
                  }
                  else if(Model.Id == "MIX.Rank")
                  {
                        CS <- MIX.Rank.Restriction(n, NULL, NULL)
                        CS.full <- MIX.Rank.Restriction(n, fixed, ident)
                  }
                  else if(Model.Id == "MIX.Bias.Rank")
                  {
                        CS <- MIX.Bias.Rank.Restriction(n, NULL, NULL)
                        CS.full <- MIX.Bias.Rank.Restriction(n, fixed, ident)
                  }
                  else if(Model.Id == "SDT.2D")
                  {
                        CS <- SDT.SDT2D.Restriction(n, NULL, NULL, NULL)
                        CS.full <- SDT.SDT2D.Restriction(n, fixed, ident, NULL)
                  }
                  else if(Model.Id == "SDT.Rank")
                  {
                        CS <- SDT.Rank.Set.Restriction(n, NULL, NULL, NULL)
                        CS.full <- SDT.Rank.Set.Restriction(n, fixed, ident, NULL)
                  }
                  else if(Model.Id == "SDT.CAT")
                  {
                        CS <- SDT.CAT.Restriction(par, n, NULL, NULL, NULL)
                        CS.full <- SDT.CAT.Restriction(par, n, fixed, ident, NULL)
                  }
                  else if(Model.Id == "IRF.Gauss")
                  {
                        CS <- IRF.Gauss.Set.Restriction(par, n, NULL, NULL, NULL)
                        CS.full <- IRF.Gauss.Set.Restriction(par, n, fixed, ident, NULL)
                  }
                  else if(Model.Id == "IRF.LC")
                  {
                        CS <- IRF.LC.Set.Restriction(par, n, NULL, NULL, NULL)
                        CS.full <- IRF.LC.Set.Restriction(par, n, fixed, ident, NULL)
                  }
                  fixed.std <- CS$fixed
                  fixed <- CS.full$fixed
                  if(!is.null(fixed.std))
                  {
                        # Identify those standard contraints that were set manually to different values by the user
                        len.std <- dim(fixed.std)[2]
                        len.fx  <- dim(fixed)[2]
                        inc     <- rep(T, len.std)  # Indicates those standard restrictions that are set differently by the user
                        for(i in 1:len.std)
                              for(j in 1:len.fx)
                                    if(fixed.std[2, i] == fixed[2, j] && fixed.std[1, i] != fixed[1, j])
                                          inc[i] <- F
                        # Drop standard positions that have values that differ due to the user
                        fixed.std <- fixed.std[, inc, drop = F]
                        fixed.Id[fixed.std[2,]] <- paste(fixed.std[1,], "<set>")          # Only present positions
 
                        # Now eliminate standard constraints from fixed
                        fixed <- fixed[, !(fixed[2,] %in% fixed.std[2,]), drop = F]
                  }
            
	           # Treat identity constraints
                 ident <- CS.full$ident
	           if(!is.null(CS$ident))
	           {
	                 ident.Id.source[CS$ident[2,]] <- paste(p.names.new[CS$ident[1,]], "<set>")
		           ident.Id.source.nr[CS$ident[2,]] <- CS$ident[1,]
	                 ident <- ident[, !(ident[1,] %in% CS$ident[1,]), drop = F]
                 }
            }
      }
	if(!is.null(fixed))
	{
	     fixed.Id[fixed[2,]] <- fixed[1,]
	     Redundant.pos <- fixed[2, is.na(fixed[1,])]
	     fixed.Id[Redundant.pos] <- "Redundant-p"
      }

      # Show equality constraints
	if(!is.null(ident))
		ident.Id.source[ident[2,]] <- p.names.new[ident[1,]]

	if(!is.null(ident))
		ident.Id.source.nr[ident[2,]] <- ident[1,]
		
	x <- data.frame(name = p.names, par = round(par, deci), fixed.value = fixed.Id, ident.source = ident.Id.source, Nr = ident.Id.source.nr)
	list(Model = Model.Name, Parameters.and.Constraints = x)
}

#===============================================================================
#
# FUNCTIONS FOR COMPUTING LR INTERVALS
#
#===============================================================================

# -----------------------------------------------------------------------------------------------------------
# Function to be minimized for computing lower and upper bounds of the LR - Statistics
# Input:
#           1. Value of the target parameter
#		2. Position of the target parameter
#           3. Minimum value of logL
#           4. Full parameter vector
#           5. Data vector
#           6. Model function
#           7. Gradient function
#		8. 2 x k matrix of fixed constraints
#		9. 2 x m matrix of equality constraints
#	     10. Function for specifying functional constraints
#
# Output: 	2*log LR
# -----------------------------------------------------------------------------------------------------------
Profile.LR <- function(val, target.par, logL.min, par, data, Model, Gradient, n, fixed, ident, functional)
{
      Tol <- 1e-16
      
      # Add actual to fixed constraints
      fixed <- cbind(fixed, matrix(c(val, target.par), nr = 2, byrow = T))
      
      pos.par <- SavePos.par(par, fixed, ident)                         # Save positions of eliminated parameters	
	par 	  <- Elim.par(par, fixed, ident)                            # Set Parameter constraints

      # Optimize function
      Optim.Obj  <- optim(par, 	 	          				# Start parameters
	 		  	  SDT.logL,	          				      # To be optimized function
				  gr = Gradient,					      # Gradient function
                          data = data,						# Data
				  Model = Model,						# Model for computing probabilies
				  n = n,							# Number of Models	
				  pos = pos.par,						# Position of redundant parameters
				  fixed = fixed,						# Fixed constraints
				  ident = ident,						# Equality constraints
				  functional = functional,				# Functional constraints
				  method = "BFGS",		                  # Method = Newton without constraints
		 		  hessian = FALSE,					# Compute no Hessian 
                          control = list(factr = Tol, maxit = 1000)
                        )
    # Return 2*log LR
    2*(Optim.Obj$value - logL.min)
               
}
            
# -----------------------------------------------------------------------------------------------------------
# Binary search routine for computating likelihood ratio confidence intervals
# Input:
#
#           1. Positions of the target paramters (with respect to whome intervals have to be evaluated
#           2. Lower bounds
#           3. Upper bounds
#           4. Minimum of the logL for the full model
#           5. Full parameter vector
#           6. Data vector
#		7. Model function
#           8. Gradient function
#	      9. Number of Models
#	     10. 2 x k matrix of fixed constraints
#	     11. 2 x m matrix of equality constraints
#	     12. Function for specifying functional constraints
#          13. Quantil of critical X2 value
#          14. Maximal number of iterations of the search
#          15. Tolerance level for terminating the search 
#
# Output:   Optimization object with new entry: full.par (full parameter vector
#
# Comments: - In case of equality constraints with the target parameter as source parameter, 
#             the position of the target parameter (as the first element) as well as the positions of 
#             those parameters that are set equal to the target paramter are passed in the first argument
#             The same is true for the lower and upper bound(s).
#             In this case the equality constraints have been removed from the matrix of equality constraints
#
#           - In case of more than one target parameter the degrees of freedom have been adjusted
# -----------------------------------------------------------------------------------------------------------
SDT.LR.CFI.Bsearch <- function(target.par, lower, upper, logL.min, par, data, Model, Gradient, n, fixed, ident, functional, q.alpha, maxiter = 50, Tol = 1e-5)
{
      len.par <- length(target.par)

      # Evaluate logL for lower and upper bounds
      LR.L <- Profile.LR(lower, target.par, logL.min, par, data, Model, Gradient, n, fixed, ident, functional)
      LR.U <- Profile.LR(upper, target.par, logL.min, par, data, Model, Gradient, n, fixed, ident, functional)

      # Check, whether lower and upper bounds are too close
      lo.ready <- up.ready <- T
      if(LR.L < q.alpha)
      {
       lower.erg <- c(lower[1], LR.L, pchisq(LR.L, 1))
       lo.ready <- F
      }
      if(LR.U < q.alpha)
      {
       upper.erg <- c(upper[1], LR.U, pchisq(LR.U, 1))
       up.ready <- F
      }
      
      # 1. Estimate lower bound
      if(lo.ready)
      {
            L <- lower
            U <- rep(par[target.par[1]], len.par)
            LR.U <- 0
      
            for(i in 1:maxiter)
            {
                  if(abs(LR.L - LR.U) < Tol)
                  {
                        LR.M <- LR.L
                        break
                  }
                  M <- (U + L) / 2      
                  LR.M <- Profile.LR(M, target.par, logL.min, par, data, Model, Gradient, n, fixed, ident, functional)
                  if(abs(LR.M - q.alpha) < Tol)         # value found
                        break
                  if(LR.M < q.alpha)
                  {
                        U <- M
                        LR.U <- LR.M
                  }
                  else
                  {
                        L <- M
                        LR.L <- LR.M
                  }
                  
            }
            lower.erg <- c(M[1], LR.M, pchisq(LR.M, 1))
      }
      
      # 2. Estimate upper bound
      if(up.ready)
      {
            L <- rep(par[target.par[1]], len.par)
            U <- upper 
            LR.L <- 0
            for(i in 1:maxiter)
            {
                  if(abs(LR.L - LR.U) < Tol)
                  {
                        LR.M <- LR.L
                        break
                  }
                  M <- (U + L) / 2      
                  LR.M <- Profile.LR(M, target.par, logL.min, par, data, Model, Gradient, n, fixed, ident, functional)
                  if(abs(LR.M - q.alpha) < Tol)         # value found
                        break
                  if(LR.M < q.alpha)
                  {
                        L <- M
                        LR.L <- LR.M
                  }
                  else
                  {
                        U <- M
                        LR.U <- LR.M
                  }
            }
            upper.erg <- c(M[1], LR.M, pchisq(LR.M, 1))
      }
#      dimnames(erg) <- list(c("Bound", "logLR", "p", "df"), c("lower", "upper"))
      cbind(lower.erg, upper.erg)
            
}

# -----------------------------------------------------------------------------------------------------------
# Main function that controls the computation of likelihood-ration confidence intervalls
# Input:
#           1. Optimization object
#		2. Vector of standard errors
#		3. Size of confidence region
#           4. Control structure: A list containing the following infromation
#                 maxiter
#			Tol
#                 target.pars
#                 lower
#                 upper
#		5. Number of decimal places
#
# Output: 	
#
# -----------------------------------------------------------------------------------------------------------
SDT.LR.CFI <- function(Opti.Obj, se, cfi, control, deci)
{
      fctr.wald <- 2                    # Factor by which Wald intervals are multiplied to get the lower and upper bounds
      
      # 1. Extract relevant information from Opti.Obj
      logL.min   <- Opti.Obj$value                   # Minimized value of logL 
      par        <- Opti.Obj$SDT.full.par            # Full parameter vector
      fixed      <- Opti.Obj$SDT.fixed               # Matrix of fixed constraints
	ident      <- Opti.Obj$SDT.ident               # Matrix of equality constraints
	functional <- Opti.Obj$SDT.functional          # Matrix of equality constraints
	Model.Id   <- Opti.Obj$SDT.Model.Id            # Model identification string
	n          <- Opti.Obj$SDT.n                   # Number of signals
      dvec       <- Opti.Obj$SDT.data                # Data vector
      
      # 2. Extract model specific information
      if(Model.Id == "SDT")
      {
            Model       <- SDT.SDT                    # Model function
            Gradient    <- SDT.SDT.Grad               # Gradient function
            bd          <- SDT.par.bounds(par, n)
            lower.bd    <- bd$lower
            upper.bd    <- bd$upper
            target.pars <- SDT.SDT.target.pars.LR(par, n)   # Default target parameters for the model
      }
      else if(Model.Id == "Gaussian")
      {
            Model 	<- SDT.Gaussian                # Model function
		Gradient    <- SDT.Gaussian.Grad           # Gradient of logL
            bd          <- Gaussian.par.bounds(par, n)
            lower.bd    <- bd$lower
            upper.bd    <- bd$upper
            target.pars <- SDT.Gaussian.target.pars.LR(par, n)   # Default target parameters for the model
      }
      else
      {
            cat("\n No LR confidendence intervals available for this model <", Model.Id, ">\n")
            return(NULL)
      }

      # 3. Extract information form control structure
      
      # 3.1 Maximal number of search trials
      if(!is.null(control$maxiter)) maxiter <- control$maxiter
      else                          maxiter <- 50

      # 3.2 Tolerance level for terminating stop
      if(!is.null(control$Tol))     Tol <- control$Tol
      else                          Tol <- 1e-5
      
      # 3.3 Factor for computing bounds
      if(!is.null(control$factr))   fctr.wald <- control$factr
 
      # 3.3 Treat and check target parameters
      if(!is.null(control$target.pars))
            target.pars <- control$target.pars

      # Eliminate parameters subject to constraints
      fxd <- target.pars %in% fixed[2,] 
      target.pars <- target.pars[!fxd]
      fxd <- target.pars %in% ident[2,]
      target.pars <- target.pars[!fxd]
      if(length(target.pars) < 1)
            return(NULL)
      target.pars <- sort(target.pars)                         # Sort target parameters according to position
      
      # 3.4 Compute lower and upper bounds
      if(cfi <= Tol | cfi > 1-Tol)
      {
            cat("\n\n Misspecified size of confidence region: <", cfi, ">\n Value in the range (.50,", 1-Tol, ") expected\n\n", sep ="")
            return(NULL)
      }      
 
      pos           <- SavePos.par(par, fixed, ident)
      target.pars.r <- (1:length(pos))[pos %in% target.pars]   # Positions of target parameters with respect to reduced paraemter vector
      z.cfi <- qnorm((1-cfi)/2)

      lower <- par[target.pars] + fctr.wald * se[target.pars.r] * z.cfi
      upper <- par[target.pars] - fctr.wald * se[target.pars.r] * z.cfi
 
      # Check for NAs in bounds
      if(any(is.na(lower)))    return(NULL)
      if(any(is.na(upper)))    return(NULL)
      
      # Check whether lower or upper bounds are too low and correct if necessary
      lower <- ifelse(lower < lower.bd[target.pars], lower.bd[target.pars], lower)
      upper <- ifelse(upper > upper.bd[target.pars], upper.bd[target.pars], upper)
      
      # 4. Compute relevant chi-square quantil value
      df. <- 1
      q.alpha <- qchisq(cfi, df.)
      
      # 5. Now start computation of bounds
      lower.LR <- upper.LR <- rep("", length(se))
      Tol.out  <- 1/(10^deci)
      for(i in 1:length(target.pars))
      { 
            # 5.1 Treatment of equality constraints
            in.id       <- ident[1,] %in% target.pars[i]
            tps         <- c(target.pars[i], ident[2, in.id])
            lo          <- rep(lower[i], length(tps))
            up          <- rep(upper[i], length(tps))
            ident.red   <- ident[, !in.id]

            # 5.2 Now we are ready for the computation of bounds
            Erg <- SDT.LR.CFI.Bsearch(tps, lo, up, logL.min, par, dvec, Model, Gradient, n, fixed, ident.red, functional, q.alpha, maxiter, Tol)
            if(!is.null(Erg))
            {
                  if(abs(Erg[3, 1] - cfi) > Tol.out)  lower.LR[target.pars.r[i]] <- paste(round(Erg[1, 1], deci), "*", sep = "")
                  else                                lower.LR[target.pars.r[i]] <- round(Erg[1, 1], deci)
                  
                  if(abs(Erg[3, 2] - cfi) > Tol.out)  upper.LR[target.pars.r[i]] <- paste(round(Erg[1, 2], deci), "*", sep = "")
                  else                                upper.LR[target.pars.r[i]] <- round(Erg[1, 2], deci)
            }
      }
      data.frame(cbind(lower.LR, upper.LR))
}

#===============================================================================
#
# FUNCTIONS FOR PLOTTING
#
#===============================================================================

# ------------------------------------------------------------------------------
# Function for plotting
# Input:
#           1. Object of the following types
#                 (a) Optimization object
#                 (c) A list with the following entries:
#                       (i)   par  = Parameter vector
#                       (ii)  n    = number of signals or configuration
#                       (iii) data = data points (default = NULL)
#                       (iv)  Model.Id = Model identification string
#
#           2. Vector of colors
#           3. Vector of line types
#           4. Vector of labels for x and y axis
#           5. Legend list
#           6. Options
#
# Output: 	
#
# -----------------------------------------------------------------------------------------------------------
SDT.Plot <- function(Object, cols = NULL, ltys = NULL, labels = NULL, SDT.legend = NULL, option = NULL)
{
      # 1. Estimation Object
      if(!is.null(Object$SDT.Model.Id))   # Estimation object
      {
            Model.Id <- Object$SDT.Model.Id
            if(Model.Id == "SDT.2D")
            {
                  par.vec <- SDT2D.robust.thresholds(Object$SDT.full.par, Object$SDT.n)
                  s       <- Object$SDT.n$n.sdt
                  k1      <- Object$SDT.n$k1
                  k2      <- Object$SDT.n$k2
                  if(is.null(Object$SDT.n$ds)) ds <- 0
                  else ds <- Object$SDT.n$ds
                  SDT.SDT2D.Plot(par.vec, s, k1, k2, ds, cols, ltys, labels, option)
            }
            else if(Model.Id == "SDT2D.MIX.2")
            {
                  par.vec <- SDT2D.MIX.2.cmp.p(Object$SDT.full.par, Object$SDT.n, Object$SDT.fixed)       
                  par.vec <- SDT2D.robust.thresholds(Object$SDT.full.par, Object$SDT.n, Gauss.par = 11)
                  s       <- Object$SDT.n$s
                  k1      <- Object$SDT.n$k1
                  k2      <- Object$SDT.n$k2
                  if(is.null(Object$SDT.n$ds)) ds <- 0
                  else ds <- Object$SDT.n$ds
                  SDT2D.MIX.2.Plot(par.vec, s, k1, k2, ds, cols, ltys, labels, option)
            }
            else if(Model.Id == "SDT")
            {
                  par.vec <- Object$SDT.full.par
                  n       <- Object$SDT.n
                  SDT.SDT.Plot(par.vec, n, cols, ltys, labels, SDT.legend, option, counts = Object$SDT.data)
             }
            else if(Model.Id == "SDT.Robust")
            {
                  par.vec <- SDT.Robust.par(Object$SDT.full.par, Object$SDT.n) 
                  n       <- Object$SDT.n
                  SDT.SDT.Plot(par.vec, n, cols, ltys, labels, SDT.legend, option, counts = Object$SDT.data)
             }
            else if(Model.Id == "Gaussian")
            {
                  par.vec <- Object$SDT.full.par
                  n       <- Object$SDT.n
                  SDT.Gaussian.Plot(par.vec, n, cols, ltys, labels, SDT.legend, option, counts = Object$SDT.data)
             }
             else if(Model.Id == "MIX.PD")
             {
                  par.vec <- Object$SDT.full.par
                  n       <- Object$SDT.n
                  SDT.MIX.PD.Plot(par.vec, n, cols, ltys, labels, SDT.legend, option, counts = Object$SDT.data)
             }
             else if(Model.Id == "MIX.2")
             {
                  par.vec <- Object$SDT.full.par
                  n       <- Object$SDT.n
                  SDT.MIX.2.Plot(par.vec, n, cols, ltys, labels, SDT.legend, option, counts = Object$SDT.data)
             }
             
             else  cat("\n ---> No plotting function for the model <", Model.Id, "> available\n")
      }
      
      # 2. A list with information about the configuration
      else if(!is.null(Object$Model.Id))
      {
            Model.Id <- Object$Model.Id
            if(is.null(Object$par)) 
                  cat("\n ---> Plot SDT: No parameter vector found\n")
                  
            else if(is.null(Object$n)) 
                 cat("\n ---> Plot SDT: No list with configuration information was found\n")
                 
            else if(Model.Id == "SDT.2D")
            {
                  par.vec <- SDT2D.robust.thresholds(Object$par, Object$n)
                  s       <- Object$n$n.sdt
                  k1      <- Object$n$k1
                  k2      <- Object$n$k2
                  if(is.null(Object$n$ds)) ds <- 0
                  else ds <- Object$n$ds
                  SDT.SDT2D.Plot(par.vec, s, k1, k2, ds, cols, ltys, labels, option)
           }
           
           else if(Model.Id == "SDT")
           {
                  par.vec <- Object$par
                  n       <- Object$n
                  SDT.SDT.Plot(par.vec, n, cols, ltys, labels, SDT.legend, option, counts = Object$counts)
           }
           
           else if(Model.Id == "Gaussian")
           {
                  par.vec <- Object$par
                  n       <- Object$n
                  SDT.Gaussian.Plot(par.vec, n, cols, ltys, labels, SDT.legend, option, counts = Object$counts)
           }
           else if(Model.Id == "MIX.PD")
           {
                  par.vec <- Object$par
                  n       <- Object$n
                  SDT.MIX.PD.Plot(par.vec, n, cols, ltys, labels, SDT.legend, option, counts = Object$counts)
           }
           else if(Model.Id == "MIX.2")
           {
                  par.vec <- Object$par
                  n       <- Object$n
                  SDT.MIX.2.Plot(par.vec, n, cols, ltys, labels, SDT.legend, option, counts = Object$counts)
           }
          
           else  cat("\n ---> Plot SDT: No plotting function for the model <", Model.Id, "> available\n")
      }
      else cat("\n ---> Plot SDT Wrong type of Object: No plotting function available\n")
}


# ------------------------------------------------------------------------------------------------------------------
# Performing the plotting of SDT1D models
# Input:	1. m      = vector of means
#		2. sd     = vector of standard errors
#		3. t      = vector with decision bounds
#		4. n      = number of models
#           5. cols   = number of colors
#           6. ltys   = line types
#		7. labels = labels for axes
#           8. SDT.legend = list with position and text of legend
#
# Comment:  The function plots the density functions and the decision bounds for SDT and Gaussian models
#           It computes the relevant x and y range so that the curves fit the figure
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT1D.Plot <- function(m, sd, t, n, cols, ltys, labels, SDT.legend)
{
      # Plotting parameters
     	col.c	<- "grey80"														# colors of decision bounds
	col.d <- c("black", "grey70", "blue", "red", "cyan", "darkgreen", "magenta", "yellow")		# colors of density functions
	lwd.c	<- 2															# line width of decision bounds
	lwd.d <- 2															# line width of density function

      x.range <- 3.5
      y.range <- .005
 
      # Determine plotting area
      min.x <- min(m - x.range * sd)
      max.x <- max(m + x.range * sd)
      
      # Determine Range of y
      max.y <- dnorm(0, mean = 0, sd = min(sd))
      min.y <- 0
      
      # Now generate curves
      x <- seq(min.x, max.x, .01)
      y <- dnorm(x, m[1], sd[1])
      
      # Labels
      if(is.null(labels))
      {
            xl <- ""
            yl <- ""
      }
      else
      {
            xl <- labels[1]
            if(length(labels) > 1) yl <- labels[2]
            else yl <- ""
      }
      
      # Colors
      if(is.null(cols)) c.index <- rep(1, n)
      else              c.index <- cols
      if(length(c.index) < n) c.index <- c(c.index, rep(1, n-length(c.index)))
      c.index <- c.index[1:n]
      
      # Line types
      if(is.null(ltys)) lt.index <- rep(1, n)
      else              lt.index <- ltys
      if(length(lt.index) < n) lt.index <- c(lt.index, rep(1, n-length(lt.index)))
      lt.index <- lt.index[1:n]
      
      # Generate Setup for the legend
      lg <- Legend.setup(min.x, max.x, min.y, max.y, c.index, lt.index, SDT.legend, n = n, option = "SDT")

      # New plot window
      x11()
      
      # Generate the plot
      plot(x, y, type = "n", xlab = xl, ylab = yl, ylim = c(min.y, max.y))
      abline(v = t, col = col.c, lwd = lwd.c)
      lines(x, y, col = col.d[c.index[1]], lwd = lwd.d, lty = lt.index[1])
      for(i in 2:n)
      {
            y <- dnorm(x, m[i], sd[i])
            lines(x, y, col = col.d[c.index[i]], lwd = lwd.d, lty = lt.index[i])
      }
      abline(h = 0, col = col.c, lwd = lwd.c)
      
      # Legend
      if(lg$leg) legend(x = lg$x, y = lg$y, legend = SDT.legend$text, lty = lg$lt.index, col = col.d[lg$c.index], lwd = rep(lwd.d, n),
                     text.col = col.d[lg$c.index],  bg =lg$bg)
                     
}

# ------------------------------------------------------------------------------------------------------------------
# Performing the plotting of ROC curves
# Input:	1. m      = vector of means
#		2. sd     = vector of standard errors
#		3. n      = number of models
#		4. cols   = index on vector of colors
#           5. ltys   = line types
#           6. labels = labels for axes
#           7. SDT.legend = list with position and text of legend
#           8. zROC   = Flag indicating whether to display z-ROCs
#           9. counts = Array with data points
#
# Comment:  The function plots the density functions and the decision bounds for SDT and Gaussian models
#           It computes the relevant x and y range so that the curves fit the figure
# ------------------------------------------------------------------------------------------------------------------
SDT.ROC.Plot <- function(m, sd, n, p, cols, ltys, labels, SDT.legend, zROC, counts)
{

      # Probability bounds for evaluation and stepsize
      Tol.Max <- .999               
      Tol.Min <- .001
      stepsize <- .01
      
      # Plotting parameters
     	col.c	<- "grey80"														# colors of decision bounds
	col.d <- c("black", "grey70", "blue", "red", "cyan", "darkgreen", "magenta", "yellow")		# colors of density functions
	lwd.c	<- 2															# line width of bounds
	lwd.d <- 2															# line width of ROC
      default.pchs <- 23                                                                              # Filled diamond for data
      default.cex <- 1.5                                                                                # Magnification
    
      # 1. Define plot area
      if(zROC)
      {
            max.plotx <- max.ploty <-  3.5
            min.plotx <- min.ploty <- -3.5
      
      }
      else
      {
            max.plotx <- max.ploty <- 1
            min.plotx <- min.ploty <- 0
      }
                  
      # 2. Set labels
      if(is.null(labels))
      {
            xl <- ""
            yl <- ""
      }
      else
      {
            xl <- labels[1]
            if(length(labels) > 1) yl <- labels[2]
            else yl <- ""
      }
      
      # 3. Set colors
      if(is.null(cols)) c.index <- rep(1, n)
      else              c.index <- cols
      if(length(c.index) < n) c.index <- c(c.index, rep(1, n-length(c.index)))
      c.index <- c.index[1:n]
      
      # 4. Set line types
      if(is.null(ltys)) lt.index <- rep(1, n)
      else              lt.index <- ltys
      if(length(lt.index) < n) lt.index <- c(lt.index, rep(1, n-length(lt.index)))
      lt.index <- lt.index[1:n]
      
      # 5. Generate setup for the legend
      lg <- Legend.setup(min.plotx, max.plotx, min.ploty, max.ploty, c.index = c.index, lt.index = lt.index, SDT.legend = SDT.legend, n = n, option = ifelse(zROC, "zROC", "ROC"))
      
      # 6. Open new plot window
      x11()
      
      # 7. Generate an empty plot
      plot(c(min.plotx, max.plotx), c(min.ploty, max.ploty), type = "n", xlab = xl, ylab = yl)
      
      # Add lines 
      if(!zROC) 
      {
            abline(h = c(0, 1), lwd = lwd.c, col = col.c)
            abline(v = c(0, 1), lwd = lwd.c, col = col.c)
      }      
      abline(coef = c(0, 1), lty = 2, lwd = lwd.c, col = col.c)             # 45 degree line
      
      # 8. Compute Range of Evaluation
      eval.min <- qnorm(Tol.Min, min(m), max(sd))
      eval.max <- qnorm(Tol.Max, max(m), max(sd))
      x <- seq(eval.min, eval.max, stepsize)
      
      # 9. Compute data ROC
      if(!is.null(counts))
      {
            D.ROC <- SDT.ROC.Data(counts, n, zROC)
            for(i in 2:n)
                  points(D.ROC[1,], D.ROC[i, ], col = col.d[c.index[i-1]], pch = default.pchs, cex = default.cex, bg = col.d[c.index[i-1]])
      }
      
      # 10.1 Compute ROC or zROC for SDT and Gaussian models
      if(is.null(p))    
      {
            # 10.1.1 Compute False Alarms
            fa <- 1 - pnorm(x, m[1], sd[1])
            if(zROC) fa <- qnorm(fa, 0, 1)
      
            # 10.1.2 Compute Hits
            for(i in 2:n)
            {
                  hit <- 1 - pnorm(x, m[i], sd[i])
                  if(zROC) hit <- qnorm(hit, 0, 1)
                  lines(fa, hit, col = col.d[c.index[i-1]], lwd = lwd.d, lty = lt.index[i-1])
            }
      }
      # 10.2 Compute ROC or zROC for MIX.2 and MIX-PD models
      else         
      {
            # 10.2.1 Compute False Alarms
            fa <- p[1]*(1-pnorm(x, m[1], sd[1])) + (1-p[1])*(1-pnorm(x, m[2], sd[2]))
            if(zROC) fa <- qnorm(fa, 0, 1)
      
            # 10.2.2 Compute Hits
            k <- 3
            for(i in 2:n)
            {
                  hit <- p[i]*(1-pnorm(x, m[k], sd[k])) + (1-p[i])*(1-pnorm(x, m[k+1], sd[k+1]))
                  if(zROC) hit <- qnorm(hit, 0, 1)
                  lines(fa, hit, col = col.d[c.index[i-1]], lwd = lwd.d, lty = lt.index[i-1])
                  k <- k + 2
            }
      }
        
      # 11. Place Legend
      if(lg$leg) legend(x = lg$x, y = lg$y, legend = SDT.legend$text, lty = lg$lt.index, col = col.d[lg$c.index], lwd = rep(lwd.d, n),
                     text.col = col.d[lg$c.index],  bg = lg$bg)
}		


# ------------------------------------------------------------------------------------------------------------------
# Computing Setup of the legend
# Input:	1. min.plotx      = x-Minimum of plot tange
#           2. max.plotx      = x-Maximum of plot tange
#       	3. min.ploty      = y-Minimum of plot tange
#           4. max.ploty      = y-Maximum of plot tange
#           5. c.index        = Index in color table
#           6. lt.index       = Index of line types
#           7. SDT.legend     = list with information about the legend with the following entries:
#                                   pos:        vector (x, y) or (x) containing the position of the legend
#                                   text:       vector of tect strings
#                                   cols:       vector of colors
#                                   ltys:       vector of line types
#
#           8. n              = number of models
#           9. option         = plot option
#                               "SDT"  ---> Density Plot
#                               "ROC"  ---> ROC-Plot
#                               "zROC" ---> zROC-Plot or Density Plot
#
# Output: A list with information (see below)
# ------------------------------------------------------------------------------------------------------------------
Legend.setup <- function(min.plotx, max.plotx, min.ploty, max.ploty, c.index, lt.index, SDT.legend, n, option)
{
      # Initialize
      leg <- F
      leg.bg <- "grey97"                  # Background color legend
      leg.c.index  <- c.index
      leg.lt.index <- lt.index
      
      x.ROC <- max.plotx - 0.4
      y.ROC <- min.ploty + 0.3
     
      if(is.null(SDT.legend))       return(list(leg = F))
      if(is.null(SDT.legend$text))  return(list(leg = F))
      
      leg <- T    # Set Flag, indicating the presenc of the legend to TRUE
      
      # Compute positon of the legend
      if(is.null(SDT.legend$pos))                     
      {
            if(option == "ROC")
            {
                  x.leg <- x.ROC
                  y.leg <- y.ROC
            }
            else
            {
                   x.leg <- min.plotx
                   y.leg <- max.ploty
            }
      }
      else
      {
            x.leg <- SDT.legend$pos[1]
            if(length(SDT.legend$pos) > 1)  
                  y.leg <- SDT.legend$pos[2]
            else
            {
                  if(option == "ROC")     y.leg <- y.ROC
                  else                    y.leg <- max.ploty
            }
      }
      
      # Legend text too long     
      if(length(SDT.legend$text) > n)   
            SDT.legend$text <- SDT.legend$text[1:n]
                        
      # Treat colors of legend
      txt.len <- length(SDT.legend$text)
      if(!is.null(SDT.legend$cols)) 
            leg.c.index <- SDT.legend$cols
      if(length(leg.c.index) > txt.len)
            leg.c.index <- leg.c.index[1:txt.len]
      if(length(leg.c.index) < txt.len)
            leg.c.index <- c(leg.c.index, rep(1, txt.len - length(leg.c.index)))
            
      # Treat index of line types
      if(!is.null(SDT.legend$ltys)) 
            leg.lt.index <- SDT.legend$ltys
      if(length(leg.lt.index) > txt.len)
            leg.lt.index <- leg.lt.index[1:txt.len]
      if(length(leg.lt.index) < txt.len)
            leg.lt.index <- c(leg.lt.index, rep(1, txt.len - length(leg.lt.index)))
      

      #-------------------------------------------------------------------------
      # RETURN LIST WITH THE FOLLOWING ENTRIES
      # 1. leg     = Flag indicating whether leg is present
      # 2. x.leg   = x-positon of legend
      # 3. y.leg   = y-position of legend
      # 4. c.index = Index into color table
      # 5. lt.index = Index of line types
      # 5. Background color
      #-------------------------------------------------------------------------
      list(leg = leg, x = x.leg, y = y.leg, c.index = leg.c.index, lt.index = leg.lt.index, bg = leg.bg)
}

# ------------------------------------------------------------------------------------------------------------------
# Computing Data ROC
# Input:	1. data  = vector with data
#           2. n     = number of signals
#           3. zROC  = a flag indicating whether ZROC should be computed
#
# Output:   A matrix with points: 
#                 The first row  contains the points on x axis
#                 The other rows contain  the points on y axis
# ------------------------------------------------------------------------------------------------------------------
SDT.ROC.Data <- function(counts, n, zROC)
{
      div.last <- function(x, r) { x/r}
      
      
      
      D <- matrix(counts, nc = n)         # Matrix with counts per signal in columns
      D <- apply(D, 2, rev)               # Revert the data vectors
      D <- apply(D, 2, cumsum)            # compute cumulated counts
      
      nr <- dim(D)[1]                     # number of rows of D
      D  <- apply(D, 1, div.last, r = D[nr, ])
      D <- D[,-nr]
      
      if(zROC)
      {
            d <- as.vector(D)
            d <- qnorm(d)
            D <- matrix(d, nc = nr-1)
      }
      D
}      
           