# ===========================================================================
# SDT-Rank models: Signal detection model with ranking
# Date of creation:	      January,  2013
# Date of last update:	  october,  2020
#
# Author: Siegfried Macho
# ===========================================================================
# Load libraries and turn warnigs off
options(warn = - 1)
library(statmod)
library(npmlreg)
options(warn = 0)     

SDT.Rk.Tol <- 1e-12
MINWEIGHT  <- 1e-40

#===============================================================================
# 0. COMPUTATION OF QUADRATURE POINTS
#===============================================================================

#-------------------------------------------------------------------------------
# Function for computing the Gauss-Hermite points and weights
# Input:	1. Number of points
#           2. minimal weights
#
# Output: 	Matrix with three columns:    1. Points
#                                         2. Weights of Gauss-Hermite
#                                         3. Weights of adaptive Gauss-Hermite 
#
# Comment:  The points and weights are with respect to the Gaussian Kernel N(0, 1)
#-------------------------------------------------------------------------------
Generate.GH <- function(npt, minweight = MINWEIGHT)
{
      GH.Structure  <- gqz(numnodes = npt, minweight)   # Points and weights for Gauss-Hermite Quadrature
      cbind(GH.Structure, GH.Structure[,2] * exp(GH.Structure[,1]^2/2))	 # Weights for adaptive Gauss-Hermite Quadrature
}

#===============================================================================
# 1. MODELL FUNCTIONS OF THE RANKING PART OF THE RANKING-RATING MODELL
#===============================================================================

# ----------------------------------------------------------------------
# Computation of conditional ranking probability of the target stimulus
# given the strength value x[1] of the target variable multiplied by
# the quadrature weight x[2]
# Input:	x     = vector with quadrature point x[1] and quadrature weight x[2]
#           j     = position of target variable x within the ranking (j = 1,...,k)
#           k     = Number of alternatives
#           mu    =  mean of the non-target variables
#           sigma = standard deviation of the non-target variables
# Output: 	n - Vector, with n being the number of Response categories
# ----------------------------------------------------------------------
SDT.Rk.cond  <- function(x, j, k, mu, sigma)
{
      choose(k-1, j-1) * pnorm(x[1], mu, sigma)^(k-j) * (1-pnorm(x[1], mu, sigma))^(j-1) * x[2]
}

# ----------------------------------------------------------------------
# Computation of the unconditional ranking probability of target stimulus
# Input:	  j             = position of target variable x within the ranking
#           k             = Number of alternatives
#           mu.target     = mean of the target distribution
#           sigma.targtet = standard deviation of the target distribution
#           mu            = mean of the non-target distribution
#           sigma         = standard deviation of the non-target distribution
#
# Output: 	n - Vector, with n being the number of Response categories
# ----------------------------------------------------------------------
SDT.Rk  <- function(j, k, mu.target, sigma.target, mu = 0, sigma =1)
{
      GH.Structure[,1] <- GH.Structure[,1] * sigma.target + mu.target
      sum(apply(GH.Structure, 1, SDT.Rk.cond, j = j, k = k, mu = mu, sigma = sigma))
}

#-------------------------------------------------------------------------------
# Computation of ranking probabilties
# Input:
#     k      = Number of alternatives
#     j      = Number of positions to be computed (j <= k)
#     parvec = parameter vector with the following entries
#              1: mean of the new distribution
#              2: sd   of the new distribution
#              3: mean of the old distribution
#              4: sd   of the old distribution
# Output: Ranking probabilities
#-------------------------------------------------------------------------------
SDT._Rank <- function(k, j, parvec, fixed = NULL, ident = NULL)
{
      prb <- sapply(1:(j-1), SDT.Rk, k = k, parvec[3], parvec[4], parvec[1], parvec[2])
      return(c(prb, 1.0 - sum(prb)))
}


#===============================================================================
# 1.1 MODELL FUNCTIONS OF THE RANKING PART OF THE RANKING-RATING MODELL:
#     SEQUENTIAL RATING
#===============================================================================
# ----------------------------------------------------------------------
# Computation of probability of selcting the target stimulus
# as the first stimulus given the strength value x[1] of the target variable multiplied by
# the quadrature weight x[2]
# Input:	  x     = vector with quadrature point x[1] and quadrature weight x[2]
#           k     = Number of alternatives
#           mu    =  mean of the non-target variables
#           sigma = standard deviation of the non-target variables
# Output:  Probability of the target being selected as the first stimulus
# ----------------------------------------------------------------------
# Computation of the conditional probability
SDT.Rk.first.cond  <- function(x, k, mu, sigma)
{
      pnorm(x[1], mu, sigma)^(k-1) * x[2]
}
# Computation of the unconditional probability
SDT.Rk.first  <- function(k, mu.target, sigma.target, mu = 0, sigma =1)
{
      GH.Structure[,1] <- GH.Structure[,1] * sigma.target + mu.target
      sum(apply(GH.Structure, 1, SDT.Rk.first.cond, k = k, mu = mu, sigma = sigma))
}

#-------------------------------------------------------------------------------
# Computation of ranking probabilities for the hierarchical model
# Input:
#     k      = Number of alternatives
#     j      = Number of positions to be computed (j <= k)
#     parvec = parameter vector with the following entries
#              1: mean of the new distribution
#              2: sd   of the new distribution
#              3: mean of the old distribution
#              4: sd   of the old distribution
# Output: Ranking probabilities
#-------------------------------------------------------------------------------
SDT._Rank.first <- function(k, j, parvec, fixed = NULL, ident = NULL)
{
      prb <- sapply(k:(k-j+2) , SDT.Rk.first, parvec[3], parvec[4], parvec[1], parvec[2])

      if(j == 2) return(c(prb, 1.0 - prb))

      p <- numeric(j)
      p[1] <- prb[1]
      res <- 1.0 - prb[1]


      for(i in 2:(j-1))
      {
            p[i] <- res * prb[i]
            res  <- res * (1.0 - prb[i])
      }
      p[j] <- res
      return(p)
}

#===============================================================================
# 2. MODELL FUNCTIONS OF THE SDT-MODEL (RATING PART OF THE RANKING-RATING MODEL)
#===============================================================================

# ----------------------------------------------------------------------
# Function for creating response distribution for on Type of Stimulus
# Input:	Vector of parameters: (mean, sigma, c[1], c[2], ...., c[n-1])
# Output: 	n - Vector, with n being the number of Response categories
# ----------------------------------------------------------------------
SDT.Probs <- function(par)
{
	 Resp <- pnorm(par[3:length(par)], mean = par[1], sd = max(par[2], SDT.Rk.Tol)) 	# Probability of Response-Category 2 to n-1
	 Resp <- c(Resp, 1)										      # Response vector
	 c(Resp[1], Resp[-1]-Resp[-length(Resp)])							      # Compute Differences
}				

# ------------------------------------------------------------------------------------------------------------------
# Computation or probabilities of SDT model
# Input:	par 	= Vector of parameters of the SDT part only, consisting of the following entries
#			            1. mean for each model (except noise model)
#			            2. standard deviation for each model (except for noise model)
#			            3. thresholds (one set for each model)
#         n.sdt = Number of SDT models
#		      cfg  	= configuration information
#
# Output: Vector of probabilities for single models
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT.Rank <- function(parvec, cfg)
{
      par.pos <- SDT.Rank.pos.par_(cfg)

      Res <- NULL
      for(i in 1:length(par.pos$SDT))
            for(j in 1:ncol(par.pos$SDT[[i]])) {
                Par <- SDT.Probs(c(parvec[par.pos$SDT[[i]][,j]], parvec[par.pos$TH[[i]]]))
                Res <- c(Res, as.vector(Par))
            }
      return(Res)          
}

#===============================================================================
# 3. MODEL FUNCTIONS OF THE (COMBINED) RANKING-RATING MODEL
#===============================================================================

#-------------------------------------------------------------------------------
# Computation of ranking and rating probabilties
# Input:
#     parvec = parameter vector with the following entries
#              1: mean of the new distribution
#              2: sd   of the new distribution
#              3: mean of the old distribution
#              4: sd   of the old distribution
#     cfg    = list with configuration information containing the following 
#              entries:
#                 model  = "P" (parallel) or "S" (serial)
#                 k.rg   = number of alternatives to be ranked
#                 j.rg   = number of positions to be considered
#                 npt    = Number of quadrature points
#                 rating = Flag, indicating whether rating data are present
#
#     fixed, ident not used (for compatibility reasons only)
# Output: Ranking probabilities
#-------------------------------------------------------------------------------
SDT.Rank <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
      parallel <- (cfg$model == "P")
      p.Rk <- NULL
      n.Rk <- length(cfg$model)
      for(i in 1:n.Rk)
      {
            off.start <- 4*(i-1) + 1
            off.end   <- 4*(i-1) + 4
            pvec <- parvec[off.start:off.end]
            if(parallel[i])
                      p.Rk <- c(p.Rk,  SDT._Rank(cfg$k.rg[i], cfg$j.rg[i], parvec = pvec))
            else
                      p.Rk <- c(p.Rk,  SDT._Rank.first(cfg$k.rg[i], cfg$j.rg[i], parvec = pvec))
      }
      if(cfg$rating == F) return(p.Rk)
#      pvec <- parvec[(n.Rk*4 + 1):length(parvec)]
      c(p.Rk, SDT.SDT.Rank(parvec, cfg))
}

# ------------------------------------------------------------------------------------------------------------------
# Model incorporating restrictions
# ------------------------------------------------------------------------------------------------------------------
SDT.Rank.Full <- function(parvec, cfg, pos, fixed = NULL, ident = NULL, functional = NULL)
{
      parvec <- Restore.par(parvec, pos, fixed, ident, functional)	# Restore full parameter vector
      SDT.Rank(parvec, cfg)

}

#===============================================================================
# 4. JACOBIANS FOR THE RANKING PART OF THE RANKING-RATING MODEL
#===============================================================================

#-------------------------------------------------------------------------------
# Derivative of the conditional model function with respect to x, mu and sigma
#-------------------------------------------------------------------------------
SDT.Rk.cond.Jacobi  <- function(x, j, k, mu, sigma)
{
      z  <- (x[1]-mu)/sigma
      fn <- dnorm(z)
      dm <- -fn/sigma			         # Derivative of Gaussian with respect to m
	    ds <- -z*fn/sigma		         # Derivative of Gaussian with respect to sd
      dx <-  fn/sigma              # Derivative of Gaussian with respect to x

      bin <- choose(k-1, j-1)
      pn  <- pnorm(x[1], mu, sigma)
      p1  <- pn^(k-j)
      p2  <- (1-pn)^(j-1)
      if(pn < SDT.Rk.Tol) 
      {
            if(k-j-1 == 0)    dp1 <- 1
            else              dp1 <- 0
      }      
      else
            dp1 <- (k-j)*pn^(k-j-1)
            
      if((1-pn) < SDT.Rk.Tol) 
      {
            if(j == 2)    dp2 <- 1
            else          dp2 <- 0
      }      
      else
            dp2 <- (j-1)*(1-pn)^(j-2)
      
      bin*(dp1*c(dm, ds, dx)*p2 + p1*c(-dm, -ds, -dx)*dp2)*x[2]
}

# ----------------------------------------------------------------------
# Computation of Jacobian with k alternatives being present for model S
# as the first stimulus given the strength value x[1] of the target variable multiplied by
# the quadrature weight x[2]
# Input:	  x     = vector with quadrature point x[1] and quadrature weight x[2]
#           k     = Number of alternatives
#           mu    =  mean of the non-target variables
#           sigma = standard deviation of the non-target variables
# Output:  Vector of partial derivarives
# ----------------------------------------------------------------------
SDT.Rk.first.cond.Jacobi <- function(x, k, mu, sigma)
{
      if(k < 2) return(1.0)

      z  <- (x[1]-mu)/sigma
      fn <- dnorm(z)
      dm <- -fn/sigma			         # Derivative of Gaussian with respect to m
	    ds <- -z*fn/sigma		         # Derivative of Gaussian with respect to sd
      dx <-  fn/sigma              # Derivative of Gaussian with respect to x

      pn  <- pnorm(x[1], mu, sigma)
      dp1 <- (k-1) * pn^(k-2)

      dp1*c(dm, ds, dx)*x[2]
}

#-------------------------------------------------------------------------------
# Derivative of the unconditional model function with respect to:
# mu.target, sigma.target, mu and sigma
#-------------------------------------------------------------------------------
SDT.Rk.Jacobi  <- function(j, k, mu.target, sigma.target, mu = 0, sigma =1)
{
      Save.GH.Struc    <- GH.Structure[,1]
      GH.Structure[,1] <- GH.Structure[,1] * sigma.target + mu.target
      J.Rk <- apply(GH.Structure, 1, SDT.Rk.cond.Jacobi, j = j, k = k, mu = mu, sigma = sigma)
      J.Rk <- rbind(J.Rk, J.Rk[3,]* Save.GH.Struc)
      rowSums(J.Rk)
}

# Jacobian for model S
SDT.Rk.first.Jacobi  <- function(k, mu.target, sigma.target, mu = 0, sigma = 1)
{
      Save.GH.Struc    <- GH.Structure[,1]
      GH.Structure[,1] <- GH.Structure[,1] * sigma.target + mu.target
      J.Rk <- apply(GH.Structure, 1, SDT.Rk.first.cond.Jacobi, k = k, mu = mu, sigma = sigma)
      J.Rk <- rbind(J.Rk, J.Rk[3,]* Save.GH.Struc)
      rowSums(J.Rk)
}

#-------------------------------------------------------------------------------
# Jacobian of the model function:
# Input:   1. Parameter vector
#          2. Configuration structure
#          3. Index to model
#
# Output: Derivatives of the probabilities with respect the four parameters
#         mu.target, sigma.target, mu and sigma
#-------------------------------------------------------------------------------
SDT._Rank.Jacobi <- function(parvec, cfg, mod)
{
      j <- cfg$j.rg[mod]
      k <- cfg$k.rg[mod]
      p.J <- sapply(1:(j-1), SDT.Rk.Jacobi, k = k, parvec[3], parvec[4], parvec[1], parvec[2])
      rbind(t(p.J), -rowSums(p.J))
}

SDT._Rank.first.Jacobi <- function(parvec, cfg, mod)
{
      j <- cfg$j.rg[mod]
      k <- cfg$k.rg[mod]
      p.J <- t(sapply(k:(k-j+2), SDT.Rk.first.Jacobi, parvec[3], parvec[4], parvec[1], parvec[2]))

      prb <- sapply(k:(k-j+2), SDT.Rk.first, parvec[3], parvec[4], parvec[1], parvec[2])

      p.first.J <-  p.J[1,]                    # Vector of partial derivatives for first probability
      res.J     <- -p.J[1,]
      res       <-  1.0 - prb[1]

      if(j > 2)
      {
           for(i in 2:(j-1))
           {
                 p.first.J <- rbind(p.first.J, res.J * prb[i] + res * p.J[i,])
                 res.J     <- res.J  * (1.0 - prb[i]) - res * p.J[i,]
                 res       <- res * (1.0 - prb[i])
           }
           return(rbind(p.first.J, -colSums(p.first.J)))
      }
      return(rbind(p.first.J, -p.first.J))              # Two probabilities only
}

#-------------------------------------------------------------------------------
# Jacobian of the model function: Ranking part
# Derivatives of the probabilities with respect the ranking parameters
# For each model: mu.target, sigma.target, mu and sigma
#-------------------------------------------------------------------------------
SDT.Rank.Model.Jacobi <- function(parvec, cfg)
{
      parallel <- (cfg$model == "P")
      n.model  <- length(cfg$model)
      for(i in 1:n.model)
      {
            if(parallel[i])
                J.mod <- SDT._Rank.Jacobi(parvec[(1 + 4*(i-1)):(4*i)], cfg, i)
            else
                J.mod <- SDT._Rank.first.Jacobi(parvec[(1 + 4*(i-1)):(4*i)], cfg, i)

            if(i > 1)
            {
                filler.row <- matrix(0, nr = cfg$j.rg[i], nc = 4 * (i-1))
                filler.col <- matrix(0, nr = sum(cfg$j.rg[1:(i-1)]), nc = 4)
                J.mod      <- cbind(filler.row, J.mod)
                Rank.J     <- rbind(cbind(Rank.J, filler.col), J.mod)
            }
            else
                Rank.J <- J.mod
      }
      Rank.J
}

 #===============================================================================
# 5. JACOBIANS FOR THE RATING PART OF THE RANKING MODEL
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix of the Gaussian parameters and thresholds
# ------------------------------------------------------------------------------------------------------------------
SDT.Rank.Probs.Jacobi <- function(gausspar, th)
{
	n.t <- length(th)					# Number of thresholds
	m   <- gausspar[1]
	sd  <- max(SDT.Rk.Tol, gausspar[2])
	z  <- (th - m)/sd
	fn <- dnorm(z)
	
	# Derivative with respect to mean
	dm <- -fn/sd						# Derivative of Gaussian with respect to m
	dm <- c(dm[1], dm[-1] - dm[-n.t], -dm[n.t])

	# Derivative with respect to sd
	ds <- -z*fn/sd						# Derivative of Gaussian with respect to sd
	ds <- c(ds[1], ds[-1] - ds[-n.t], -ds[n.t])


	# Derivative with respect to thresholds
	dt <- fn/sd							# Derivative of Gaussian with respect to t[i]
	
	J <- matrix(0, nr = n.t+1, nc = n.t)		# Jacobi matrix for thresholds: columns: Parameter, rows = probabilities
	r <- 1
	for(i in 1:n.t)
	{
		J[r, i] <- dt[i]
		r <- r + 1
		J[r, i] <- -dt[i]
	}
	cbind(dm, ds, J)						# Full Jacobi matrix
}


# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix of the SDT-Model (rating part) of the Ranking-Rating model
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT.Rank.Jacobi <- function(parvec, cfg)
{
       par.pos <- SDT.Rank.pos.par_(cfg)              # Positions of parameters in: $SDT and $TH
       SDT.pos <- par.pos$SDT
       TH.pos  <- par.pos$TH 

       J.SDT            <- NULL
       filler.model.col <- NULL
       filler.model.row <- NULL
       for(i in 1:length(SDT.pos)) {
           J.Gauss    <- NULL                             # Jacobian matrix of the Gaussian model
           J.th       <- NULL                             # Jacobian matrix of threshold
           filler.row <- NULL
           filler.col <- NULL
           filler <- matrix(0, nr = cfg$n.sdt[2, i], 2)     #  probs x 2 - matrix of zeros as filler
           for(j in 1:ncol(SDT.pos[[i]])) {
               J       <- SDT.Rank.Probs.Jacobi(parvec[SDT.pos[[i]][,j]], parvec[TH.pos[[i]]])
               J.Gauss <- rbind(cbind(J.Gauss, filler.col), cbind(filler.row, J[, 1:2]))
               J.th    <- rbind(J.th, J[, -(1:2)])
               if(j < ncol(SDT.pos[[i]])) {
                  filler.row <- cbind(filler.row, filler)     # Adapt filler matrices
                  filler.col <- rbind(filler.col, filler)
               }
           }    
           J.actual <- cbind(J.Gauss, J.th)
           J.SDT    <- cbind(rbind(J.SDT, filler.model.row), rbind(filler.model.col, J.actual))
           if(i < length(SDT.pos)) {
              filler.model.row <- matrix(0, nr = cfg$n.sdt[1, i+1] * cfg$n.sdt[2, i+1], nc = sum(cfg$n.sdt[1, 1:i]) * 2 + sum(cfg$n.sdt[2, 1:i] - 1))
              filler.model.col <- matrix(0, nr = sum(cfg$n.sdt[1, 1:i] * cfg$n.sdt[2, 1:i]), nc = cfg$n.sdt[1, i+1] * 2 + cfg$n.sdt[2, i+1] - 1)
          }
       }
       return(J.SDT)
}

#===============================================================================
# 6. JACOBI MATRIX, GRADIENT AND MODEL MATRICES OF THE RATING-RANKING MODEL
#===============================================================================

#-------------------------------------------------------------------------------
# Jacobian of the model function of the comnind Ranking & Rating model:
# Input: (a) Parametervector
#        (b) Configuration information
#        (c) fixded and ident not used (just for compatibiltiy)
#-------------------------------------------------------------------------------
SDT.Rank.Jacobi <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
      # 1. Jacobi-Matrix of the ranking part
      rank.J <- SDT.Rank.Model.Jacobi(parvec, cfg)
      if(cfg$rating == F)     # No rating 
            return(rank.J)
      
      # 2. Jacobi-Matrix of the rating part
      SDT.J <- SDT.SDT.Rank.Jacobi(parvec, cfg)
      
      dim.Rk   <- dim(rank.J)
      dim.SDT  <- dim(SDT.J)

      filler.row <- matrix(0, nr = dim.SDT[1], nc = dim.Rk[2])
      filler.col <- matrix(0, nr = dim.Rk[1],  nc = dim.SDT[2])

      rbind(cbind(rank.J, filler.col), cbind(filler.row, SDT.J))
}


#-------------------------------------------------------------------------------
# Model matrix of the ranking model 
# implementing functional and identity constraints
#-------------------------------------------------------------------------------
SDT.Rank.Model.Matrix <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos      <- SavePos.par(full.par, fixed, ident)
      parvec   <- Elim.par(full.par, fixed, ident)
      full.par <- Restore.par(parvec, pos, fixed, ident, functional)	# Restore full parameter vector

      J   <- SDT.Rank.Jacobi(full.par, cfg = cfg)
      if(!is.null(ident))
      {
            J.Id <- diag(length(full.par))
            Index <- matrix(c(ident[2,], ident[1,]), nc = 2)
            J.Id[Index] <- 1
            
            J <- J %*% J.Id
      }
      
      # Treatment of functional constraints
      if(!is.null(functional))
      {
            J.fct <- attr(functional, "SDT.Jacobian")
            if(is.null(J.fct)) 
                  return(NULL)           # No Jacobian function supplied -> return (NULL)
            J <- J %*% J.fct(full.par)
      }
      J[, pos]
}

# ------------------------------------------------------------------------------------------------------------------
# Numerical computation of the model matrix (SDT-Rank model)
# ------------------------------------------------------------------------------------------------------------------
SDT.Rank.Model.Matrix.Num <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)
      parvec <- Elim.par(full.par, fixed, ident)
      J <- jacobian(SDT.Rank.Full, parvec, cfg = cfg, pos = pos, fixed = fixed, ident = ident, functional = functional)
}

# ------------------------------------------------------------------------------------------------------------------
# Gradient of logL for the SDT-Rank model
# Input:	Same as for Likelihood function
# Output:	Gradient
# ------------------------------------------------------------------------------------------------------------------
SDT.Rank.Grad <- function(par, data, Model, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
	parvec   <- Restore.par(par, pos, fixed, ident, functional)	# Restore full parameter vector
	gradient <- as.vector(SDT.logL.Jacobi(parvec, data, Model, n, pos, fixed, ident, functional) %*% SDT.Rank.Jacobi(parvec, n)) 

	# Treat functional constraints
	if(!is.null(functional))
	{
		gr.fct <- attr(par, "SDT.gradient")		# Isolate gradient function
		if(is.null(gr.fct))	stop("\n\n SDT.Rank.Grad: Error in computation of gradient: No gradient for functional constraints found:\n Set Parameter sym.gr = F in function SDT.Estimate !\n\n")
		gradient <- gr.fct(gradient)
	}	
	Elim.par.grad(gradient, fixed, ident)
}

#===============================================================================
# 7. ADMINISTRATIVE FUNCTIONS
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# ChEck and generate configuration structure
# Input:	cfg = Configuration structure: A list with the following entries
#                 - k.rg        A vector indicating the number of ranking alternatives
#                 - j.rg        A vector indicating the number of relevant ranking positions
#                 - n.sdt       2xn matrix (or a vector of this length, inidicating the number Gaussian models within each SDT rating model (for each SDT model)
#                 - npt         Number of quadrature points
#                 - restriction String indicating the restritions
#                 - rating      Flag indicating whether ranking and rating are combined
#
# Output: String with parameter values
# ------------------------------------------------------------------------------------------------------------------
SDT.Rank.Check.Fill.Cfg <- function(cfg)
{
      if(!is.null(cfg))
         if(!is.numeric(cfg))
            if(!is.null(cfg$checked))
               if(cfg$checked)   return(cfg)

      k.rg <- 2
      npt <- 30

      if(is.null(cfg))
      {
            cfg <- list(n.sdt       = matrix(c(2, 6), nr = 2, nc = 1),
                        model       = "P",
                        k.rg        = k.rg,
                        j.rg        = k.rg,
                        npt         = npt,
                        restriction = "standard",
                        rating      = T,
                        checked     = T)
            return(cfg)
      }
      if(is.numeric(cfg))
      {
            k.rg     <- as.integer(cfg)
            cfg      <- list(n.sdt       = matrix(c(2, 6), nr = 2, nc = 1),
                             model       = "P",
                             k.rg        = k.rg,
                             j.rg        = k.rg,
                             npt         = npt,
                             restriction = "standard",
                             rating      = T,
                             checked     = T)
            return(cfg)
      }
      cfgneu <- list()
      len <- max(1, length(cfg$k.rg), length(cfg$j.rg))
      if(is.null(cfg$n.sdt))       cfgneu$n.sdt       <- matrix(c(2, 6), nr = 2, nc = 1)           else  cfgneu$n.sdt       <- matrix(cfg$n.sdt, nr = 2)
      if(is.null(cfg$model))       cfgneu$model       <- rep("P", len)                             else  cfgneu$model       <- cfg$model
      if(is.null(cfg$k.rg)) if(is.null(cfg$j.rg)) cfgneu$k.rg <- k.rg else cfgneu$k.rg <- cfg$j.rg else  cfgneu$k.rg        <- cfg$k.rg
      if(is.null(cfg$j.rg))        cfgneu$j.rg        <- cfg$k.rg                                  else  cfgneu$j.rg        <- cfg$j.rg
      if(is.null(cfg$npt))         cfgneu$npt         <- npt                                       else  cfgneu$npt         <- cfg$npt
      if(is.null(cfg$restriction)) cfgneu$restriction <- "standard"                                else  cfgneu$restriction <- tolower(cfg$restriction)
      if(is.null(cfg$rating))      cfgneu$rating      <- T                                         else  cfgneu$rating      <- cfg$rating
      
      cfg <- cfgneu
      
      # Checks
      if(any(!is.numeric(cfg$n.sdt)))  
         stop(c("\n FATAL ERROR in SDT.Rank.Check.Fill.Cfg: Vector with number of Gaussian models within each SDT modesl is not numeric\n"))

      if(any(cfg$k.rg < cfg$j.rg)) {
            k.rg <- paste(cfg$k.rg, collapse = ",")
            j.rg <- paste(cfg$j.rg, collapse = ",")
            stop(c("\n FATAL ERROR in SDT.Rank.Check.Fill.Cfg: Number of alternatives <", k.rg, "> smaller than number of relevant ranks <", j.rg, ">\n"))
      }
      if(any(cfg$j.rg < 2)) {
            j.rg <- paste(cfg$j.rg, collapse = ",") 
            stop(c("\n FATAL ERROR in SDT.Rank.Check.Fill.Cfg: Number of altenative smaller than 2: <", j.rg, ">\n"))
      }
      if(length(cfg$model) < length(cfg$k.rg)) {
            stop("\n FATAL ERROR in SDT.Rank.Check.Fill.Cfg: Length of structure <model> smaller than number of ranking modells\n")
      }  
      if(!is.matrix(cfg$n.sdt))
            stop("\n FATAL ERROR in SDT.Rank.Check.Fill.Cfg: <n.sdt> must be a matrix\n")
      if(nrow(cfg$n.sdt) != 2)      
            stop("\n FATAL ERROR in SDT.Rank.Check.Fill.Cfg: <n.sdt> must be a 2 x n matrix (n = number of SDT models)\n")
      if(any(cfg$sdt[,1] < 2) && cfg$rating == T)      
            stop("\n FATAL ERROR in SDT.Rank.Check.Fill.Cfg: Number of Gaussinas in any SDT model must be greater than 1)\n")

      cfg$checked <- T

      return(cfg)
}                  
                        
# ------------------------------------------------------------------------------------------------------------------
# Generate starting parameters due to configuration information
# Input:	cfg = Configuration structure: A list with the following entries
#                 - model       Vector of "P" and "S"
#                 - k.rg        Vector of Number of alternatives
#                 - j.rg        Vector of Number of relevant ranking positions
#                 - n.sdt       2xn matrix (or a vector of this length, inidicating the number Gaussian models within each SDT rating model (for each SDT model)
#                 - restriction String indicating the restrictions
#                 - rating:     Flag indicating whether rating data are to be modeled
#                              
# Output: String with parameter values depending on whether only ranking data or
#         ranking as well as reting data are modeled
# ------------------------------------------------------------------------------------------------------------------
SDT.Rank.start.par <- function(cfg)
{
      cfg <- SDT.Rank.Check.Fill.Cfg(cfg)

      #-------------------------------------------------------------------------
      # Start parameters of the ranking model:
      # For each model: m.new, sd.new, m.old, sd.old
      # m.new   = mean of new distribution
      # sd.new  = standard deviation of new distribution
      # m.old   = mean of old distribution
      # sd.old  = standard deviation of old distribution
      #-------------------------------------------------------------------------
      par.rank <- rep(c(0, 1, 0.5, 1), length(cfg$k.rg))    # Start parameter

      #-------------------------------------------------------------------------
      # Parameters of the SDT model:
      # For each model m, sd,
      # thresholds
      # m    = mean of the distribution
      # sd   = standard deviation of the distribution
      # c[i] = thresholds
      #-------------------------------------------------------------------------
      if(!cfg$rating) return(par.rank)   # No rating

      for(i in 1:ncol(cfg$n.sdt))
            par.rank <- c(par.rank, rep(c(0, 1), cfg$n.sdt[1, i]), seq(-0.5, 0.5, length = cfg$n.sdt[2, i] - 1))           # Add mu and sigma, for each model and (at the end) thresholds

      return(par.rank)
}

# ------------------------------------------------------------------------------------------------------------------
# Generate upper and lower bounds of parameter
# Input:	parvec = Vector of parameters
#		   cfg = configruation structure
#
# Output: String with parameter names
# ------------------------------------------------------------------------------------------------------------------
SDT.Rank.par.bounds <- function(cfg)
{
      cfg <- SDT.Rank.Check.Fill.Cfg(cfg)

      lower.Rk  <- rep(c(-Inf, SDT.Rk.Tol, -Inf, SDT.Rk.Tol), length(cfg$k.rg))
      upper.Rk  <- rep(Inf, 4 * length(cfg$k.rg))

      if(cfg$rating == F)
            return(list(lower = lower.Rk, upper = upper.Rk))
      
      for(i in 1:ncol(cfg$n.sdt)) {
            lower.Rk <- c(lower.Rk, rep(c(-Inf, SDT.Rk.Tol), cfg$n.sdt[1, i]), rep(-Inf, cfg$n.sdt[2, i] - 1))
            upper.Rk <- c(upper.Rk, rep(c( Inf, Inf       ), cfg$n.sdt[1, i]), rep( Inf, cfg$n.sdt[2, i] - 1))
      }
      return(list(lower = lower.Rk, upper = upper.Rk))
}      
                  
# ------------------------------------------------------------------------------------------------------------------
# Generate strings with names of the parameters for the models
# Input:	parvec = Vector of parameters
#		   cfg = configruation structure
#
# Output: String with parameter names
# ------------------------------------------------------------------------------------------------------------------
SDT.Rank.par.names <- function(cfg)
{
      cfg <- SDT.Rank.Check.Fill.Cfg(cfg)

      Name.Rk <- NULL
      for(i in 1:length(cfg$model))
           if(cfg$model[i] == "P")
           {
              Name.Rk <- c(Name.Rk, paste("[P-", i, "]   Mean.1", sep = ""))
              Name.Rk <- c(Name.Rk, paste("[P-", i, "] Stddev.1", sep = ""))
              Name.Rk <- c(Name.Rk, paste("[P-", i, "]   Mean.2", sep = ""))
              Name.Rk <- c(Name.Rk, paste("[P-", i, "] Stddev.2", sep = ""))
           }
           else
           {
              Name.Rk <- c(Name.Rk, paste("[S-", i, "]   Mean.1", sep = ""))
              Name.Rk <- c(Name.Rk, paste("[S-", i, "] Stddev.1", sep = ""))
              Name.Rk <- c(Name.Rk, paste("[S-", i, "]   Mean.2", sep = ""))
              Name.Rk <- c(Name.Rk, paste("[S-", i, "] Stddev.2", sep = ""))
           }

      if(cfg$rating == F)
         return(Name.Rk)
      # Now create names for rating models
      for(i in 1:ncol(cfg$n.sdt)) {
          for(j in 1:cfg$n.sdt[1, i]) {
                Name.Rk <- c(Name.Rk, paste("[SDT-", i, "]   Mean.", j, sep = ""))
                Name.Rk <- c(Name.Rk, paste("[SDT-", i, "] Stddev.", j, sep = ""))
          }
          Name.Rk <- c(Name.Rk, paste("[SDT-", i, "]      c-", 1:(cfg$n.sdt[2, i] - 1),   sep = ""))
      }    
      return(Name.Rk)
}

# ------------------------------------------------------------------------------------------------------------------
# Names of the models 
# ------------------------------------------------------------------------------------------------------------------
SDT.Rank.model.name <- function(cfg)
{
      cfg <- SDT.Rank.Check.Fill.Cfg(cfg)

      k.rg  <- paste(cfg$k.rg, collapse = ",")
      j.rg  <- paste(cfg$j.rg, collapse = ",")
      model <- paste(cfg$model, collapse = ",")
      model <- paste(cfg$model, collapse = ",")
      if(ncol(cfg$n.sdt) == 1) 
          n.sdt <- paste(cfg$n.sdt[,1], collapse = ",")  
      else {
          n.sdt = NULL
          for(i in 1:ncol(cfg$n.sdt))
              if(i < ncol(cfg$n.sdt)) 
                 n.sdt <- paste(n.sdt, "<", paste(cfg$n.sdt[,i], collapse = ","), ">, ", sep = "") 
              else
                 n.sdt <- paste(n.sdt, "<", paste(cfg$n.sdt[,i], collapse = ","), ">", sep = "") 
      }
      if(cfg$rating)
            mod.name <- paste("SDT ranking and rating model:", 
                              " Number of Gaussians within SDT models: <", n.sdt, ">",
                              " Model types: <", model, ">",
                              " Number of ranking-alternatives: <", k.rg, ">", 
                              " Number of ranking-positions: <", j.rg, ">", 
                              " Number of quadrature points: <", cfg$npt, ">", sep = "",
                              " Restrictions: <", cfg$restriction, ">")
      else
            mod.name <- paste("SDT ranking model:", 
                              " Model types: <", model, ">",
                              " Number of ranking-alternatives: <", k.rg, ">",
                              " Number of ranking-positions: <", j.rg, ">", 
                              " Number of quadrature points: <", cfg$npt, ">", sep = "",
                              " Restrictions: <", cfg$restriction, ">")
return(mod.name)
}

# ------------------------------------------------------------------------------------------------------------------
# SDT Ranking Model: Check number of parameters and data points on the basis of the configuration information
# Error-Codes: 1: Data length wrong
#              2: Parameter lenght wrong
#              3: Both wrong
# ------------------------------------------------------------------------------------------------------------------
SDT.Rank.check <- function(cfg,  parvec, datavec)
{

      Msg <- "NOTE: Computation of the expected number of data points and parameters for a given configuration"
      dlen <- length(datavec)  # length of data vector
      plen <- length(parvec)
      cfg  <- SDT.Rank.Check.Fill.Cfg(cfg)

      # Case 1: Ranking and Rating
      if(cfg$rating) {
            dlen.cfg <- sum(cfg$j.rg) + sum(cfg$n.sdt[1,]*cfg$n.sdt[2,])                
            plen.cfg <- 4*length(cfg$k.rg) + 2*sum(cfg$n.sdt[1,]) + sum(cfg$n.sdt[2,] - 1)
      }
      # Case 2: Ranking only
      else {
            dlen.cfg <- sum(cfg$j.rg)
            plen.cfg <- 4*length(cfg$k.rg)
      }
	Erg <- matrix(c(dlen, plen, dlen.cfg, plen.cfg), nc = 2)
	colnames(Erg) <- c("Found", "Expected")
  rownames(Erg) <- c("Data-length", "Parameter-length")
	Error <- 1*(Erg[1, 1] != Erg[1, 2])
  Error <- Error + 2*(Erg[2, 1] != Erg[2, 2])
	list(Error = Error, Stat = Erg, Msg = Msg)
}

# ------------------------------------------------------------------------------------------------------------------
# SDT Ranking Model: Set restrictions
# ------------------------------------------------------------------------------------------------------------------
SDT.Rank.Set.Restriction <- function(cfg, fixed = NULL, ident = NULL, functional = NULL)
{
        # If configuration information is not complete, then complete it
        cfg <- SDT.Rank.Check.Fill.Cfg(cfg)

        if(cfg$restriction == "standard")
            return(SDT.Rank.Set.Standard.Restriction(cfg, fixed, ident, functional))
        if(cfg$restriction == "equal")    
            return(SDT.Rank.Set.Equal.Restriction(cfg, fixed, ident, functional))
        return(list(fixed = fixed, ident = ident, functional = functional))
}

# ------------------------------------------------------------------------------------------------------------------
# SDT.Rank: Set standard restrictions
#           Sets restrictions and checks foru double restrictons
# ------------------------------------------------------------------------------------------------------------------
SDT.Rank.Set.Standard.Restriction <- function(cfg, fixed, ident, functional)
{
      SR.Rk <- SDT.Rank.Standard.Restriction(cfg)

      # Check for uniqueness add to fixed constraints
      fixed.s <- SR.Rk$fixed
      pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
	    fixed <- cbind(fixed.s[, pos.unique], fixed)

      list(fixed = fixed, ident = ident, functional = functional)
}
# ------------------------------------------------------------------------------------------------------------------
# SDT.Rank: Here the fixed matrix with stanard restrictions is generated
# ------------------------------------------------------------------------------------------------------------------
SDT.Rank.Standard.Restriction <- function(cfg)
{
      # Fixed constraints:
      # Rank: m = 0, sd = 1 for first Gaussian of each model
      
      par.pos <- SDT.Rank.pos.par_(cfg)
      afc.pos <- par.pos$AFC
      
      pos.mu.rk <- pos.sd.rk <- NULL
      n.rk <- ncol(afc.pos)
      k <- 1
      while(k < n.rk) {
            pos.mu.rk <- c(pos.mu.rk, afc.pos[1,k])
            pos.sd.rk <- c(pos.sd.rk, afc.pos[2,k])
            k <- k + 2
      }
      n.rk <- length(pos.mu.rk)
      fixed <- matrix(c(rep(c(0, 1), each = n.rk), pos.mu.rk, pos.sd.rk), nr = 2, byrow = T)

      if(cfg$rating == F) return(list(fixed = fixed))

      # Set fixed constriants for first Gaussian in each SDT model
      sdt.pos <- par.pos$SDT
      pos.mu.sdt <- pos.sd.sdt <- NULL
      for(i in 1:length(sdt.pos)) {
          pos.mu.sdt <- c(pos.mu.sdt, sdt.pos[[i]][1, 1])
          pos.sd.sdt <- c(pos.sd.sdt, sdt.pos[[i]][2, 1]) 
      }    
      vals <- rep(c(0, 1), c(length( pos.mu.sdt), length(pos.sd.sdt)))
      fixed.sdt <- matrix(c(vals, pos.mu.sdt, pos.sd.sdt), nr = 2, byrow = T)
      fixed <- cbind(fixed, fixed.sdt)

      return(list(fixed = fixed))
}

# ------------------------------------------------------------------------------------------------------------------
# SDT.Rank: Set equality restrictions on all parameters and standard restricitons on the first models
#           Sets restrictions and checks four double restrictons
# ------------------------------------------------------------------------------------------------------------------
SDT.Rank.Set.Equal.Restriction <- function(cfg, fixed, ident, functional)
{
      SR.Rk <- SDT.Rank.Equal.Restriction(cfg)

      # Check for uniqueness add to fixed constraints
      fixed.s <- SR.Rk$fixed
      pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
	    fixed <- cbind(fixed.s[, pos.unique], fixed)

      # Cheque for uniqueness add to identity constraints
      ident.s <- SR.Rk$ident
      pos.unique <-  !(ident.s[2,] %in% ident[2,])
	    ident <- cbind(ident.s[, pos.unique], ident)

      list(fixed = fixed, ident = ident, functional = functional)
}

#------------------------------------------------------------------------------------------------------------------
# SDT.Rank: Here the fixed matrix with stanard restrictions is generated
#------------------------------------------------------------------------------------------------------------------
SDT.Rank.Equal.Restriction <- function(cfg)
{
      # Set fixed standard restrictions
      fixed <- SDT.Rank.Standard.Restriction(cfg)$fixed 

      # Now set equality constraints (on mean and sd of the second Gaussian distribution in each afc model)
      par.pos <- SDT.Rank.pos.par_(cfg)

      afc.pos <- par.pos$AFC
      eq1 <- afc.pos[,2]
      ident <- NULL
      nc <- ncol(afc.pos)
      if(nc >= 4) {
         index <- 4
         while(index <= nc) {
              ident <- cbind(ident, rbind(eq1, afc.pos[, index]))
              index <- index + 2
         }
       }       

      if(cfg$rating == F) return(list(fixed = fixed, ident = ident))

      # Now set equality constraints on mean and sd of last distributions of the SDT models
      sdt.pos <- par.pos$SDT            # List with positions
      for(i in 1:length(sdt.pos))
          for(j in 2:ncol(sdt.pos[[i]]))
             ident <- cbind(ident, rbind(eq1, sdt.pos[[i]][, j]))
      return(list(fixed = fixed, ident = ident))
}

# ------------------------------------------------------------------------------------------------------------------
# Format free parameters and compute confidence intervals
# Input:	1. px2 Matrix with parameters and standard errors
#         2. configuration information
#		      2. Full parameter vector
#		      3. Matrix of fixed constraints
#		      4. Matrix of equality constraints
#		      5. Confidence level
#         6. Number of decimal places
#
# Output: px5 Matrix of parameters, standard errors and confidence limits
# ------------------------------------------------------------------------------------------------------------------
SDT.Rank.Format.Par.est <- function(Pars, full.par, cfg, fixed, ident, alpha, deci)
{
      n.rk <- length(cfg$k.rg)
      pos.sd <- seq(2, 4*n.rk, 2)              # Position of standard deviation parameter
      if(cfg$rating)
      {
           par.pos <- SDT.Rank.pos.par_(cfg)$SDT
           for(i in 1:length(par.pos))
               pos.sd <- c(pos.sd, par.pos[[i]][,2])
      }
      qt.cfi <- qnorm((1-alpha)/2)
      cfi.lo <- Pars[,1] + Pars[,2]*qt.cfi
      cfi.up <- Pars[,1] - Pars[,2]*qt.cfi
            
      # Compute lower bounds (= for standard errors of the distributions)            
      bd.lo <- rep(-Inf, length(full.par))
      bd.lo[pos.sd] <- 0.0
      bd.lo <- Elim.par(bd.lo, fixed, ident)

      cfi.lo <- ifelse(cfi.lo < bd.lo, bd.lo, cfi.lo)
       
      Pars <- cbind(Pars, cfi.lo, cfi.up)
      colnames(Pars) <- c("Value", "SE", paste("CFI-", alpha*100, "(Lower)", sep = ""), paste("CFI-", alpha*100, "(Upper)", sep = ""))
      round(Pars, deci)
}

# ------------------------------------------------------------------------------------------------------------------
# Format full parameters
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Par.Rank <- function(full.par, cfg, deci)
{
      Ranking.names <- c("Mean.1",
                         "Stddev.1",
                         "Mean.2",
                         "Stddev.2")

      Rating.names  <- c("Mean",
                         "Stddev")

      n.rk <- length(cfg$k.rg)

      th.par     <- NULL
      Gauss.pos  <- NULL
      th.pos     <- NULL

      full.par <- round(full.par, deci)
      Ranking.Par <- matrix(full.par[1:(4*n.rk)], nc = n.rk)
      rownames(Ranking.Par) <- Ranking.names
      colnames(Ranking.Par) <- paste("Rank-", 1:n.rk,  "[", cfg$model[1:n.rk], "]", sep = "")

      if(!cfg$rating)
        return(list(Ranking = Ranking.Par))

      # Rating
      par.pos <- SDT.Rank.pos.par_(cfg)
      par.pos.sdt <- par.pos$SDT
      par.pos.th  <- par.pos$TH

      # Compile Gaussian Parameters
      for(i in 1:length(par.pos.sdt))
           for(j in 1:ncol(par.pos.sdt[[i]]))
               Gauss.pos <- cbind(Gauss.pos, par.pos.sdt[[i]][,j])
      Rating.Par             <- matrix(full.par[Gauss.pos], nc = sum(cfg$n.sdt[1,]))
      rownames(Rating.Par)   <- Rating.names
      Gauss.names            <- paste("SDT-", 1:ncol(cfg$n.sdt), sep = "")
      SDT.names              <- NULL
      for(i in 1:length(Gauss.names)) SDT.names <- c(SDT.names, paste(Gauss.names[i], "/Gauss-", 1:cfg$n.sdt[1, i], sep = ""))
     	   colnames(Rating.Par) <- SDT.names

       # Compile Thresholds
       max.lt <- 0
       for(i in 1:length(par.pos.th)) # Detmine maximal number of thresholds
          if(length(par.pos.th[[i]]) > max.lt) max.lt <- length(par.pos.th[[i]])
       for(i in 1:length(par.pos.th)) {
          th     <- c(paste(full.par[par.pos.th[[i]]]), rep("", max.lt - length(par.pos.th[[i]])))
          th.par <- rbind(th.par, th)   
       }
       th.par <- data.frame(th.par)
       colnames(th.par) <- paste("c-", 1:ncol(th.par), sep = "")
       rownames(th.par) <- paste("SDT-", 1:nrow(th.par), sep = "")
 
      # 2. Ranking and rating data
      return(list(Ranking = Ranking.Par, Rating = Rating.Par, Thresholds = th.par))
}

# ------------------------------------------------------------------------------------------------------------------
# Ranking Model with recollection and guessing: Computation of observed p values for ranking model
# Input:    datavec = Vector with data
#           cfg     = configuration information
#
# Output:   Vector with probabilities
# ------------------------------------------------------------------------------------------------------------------
SDT.Rank.Compute.Obs.p <- function(datavec, cfg)
{
      cfg <- SDT.Rank.Check.Fill.Cfg(cfg)
      
      p.rank <- NULL
      ed <- 0
      for(i in 1:length(cfg$j.rg))
      {
            st     <- ed + 1
            ed     <- ed + cfg$j.rg[i]
            p.rank <- c(p.rank, datavec[st:ed]/sum(datavec[st:ed]))
      }
      if(cfg$rating == F)
            return(p.rank)

      datavec <- datavec[-(1:sum(cfg$j.rg))]
      for(i in 1:ncol(cfg$n.sdt)) {
         sdt.part <- matrix(datavec[1:(cfg$n.sdt[1,i]*cfg$n.sdt[2,i])], nr = cfg$n.sdt[1, i], byrow = T)     # Extract rating data to matrix with n rows
         p.rank   <- c(p.rank, as.vector(apply(sdt.part, 1, function(x) { x/sum(x) })))
         if(i < ncol(cfg$n.sdt))
              datavec <- datavec[-(1:(cfg$n.sdt[1,i]*cfg$n.sdt[2,i]))]
      }
      return(p.rank)
}

# ------------------------------------------------------------------------------------------------------------------
# Computation of observed p values
# Input:    datavec = Vector with data
#           cfg     = configuration information
#
# Output:   Vector with marginal Frequencies
# ------------------------------------------------------------------------------------------------------------------
SDT.Rank.Compute.Obs.N <- function(datavec, cfg)
{
      cfg <- SDT.Rank.Check.Fill.Cfg(cfg)

      # N for ranking models
      N.rank <- NULL
      ed <- 0
      for(i in 1:length(cfg$j.rg))
      {
            st     <- ed + 1
            ed     <- ed + cfg$j.rg[i]
            ss     <- sum(datavec[st:ed])
            N.rank <- c(N.rank, rep(ss, cfg$j.rg[i]))
      }
      if(cfg$rating == F) return(N.rank)

      # N for SDT models
      datavec <- datavec[-(1:sum(cfg$j.rg))]
      for(i in 1:ncol(cfg$n.sdt)) {
          sdt.part <- matrix(datavec[1:(cfg$n.sdt[1,i]*cfg$n.sdt[2,i])], nr = cfg$n.sdt[1, i], byrow = T)     # Extract rating data to matrix with n rows
          N.rating <- rowSums(sdt.part)
          N.rank   <- c(N.rank, rep(N.rating, each = dim(sdt.part)[2]))
          if(i < ncol(cfg$n.sdt))
              datavec <- datavec[-(1:(cfg$n.sdt[1,i]*cfg$n.sdt[2,i]))]
      }
     return(N.rank)
}

# ------------------------------------------------------------------------------------------------------------------
# Computation of free data points
# Input:    cfg     = configuration information
#
# Output:   Number of free data points
# ------------------------------------------------------------------------------------------------------------------
SDT.Rank.Compute.df.data <- function(cfg)
{
        free.data <- sum(cfg$j.rg - 1)
        if(!cfg$rating) 
          return(free.data)
        
        free.data <- free.data + sum(cfg$n.sdt[1, ]*(cfg$n.sdt[2, ] - 1))
        return(free.data)
}

# ------------------------------------------------------------------------------------------------------------------
# Generation of data labels in output
# Input:    cfg     = configuration information
#
# Output:   Data labels in output
# ------------------------------------------------------------------------------------------------------------------
SDT.Rank.Datalabels <- function(cfg)
{
      cfg <- SDT.Rank.Check.Fill.Cfg(cfg)
      
      # 1. Labels for ranking data
      label.out <- NULL
      l.rk <- length(cfg$j.rg)
      label.Rk <- paste("AFC-", cfg$k.rg, sep = "")
      for(i in 1:l.rk)
            label.out <- c(label.out, paste(label.Rk[i], " [", 1:cfg$j.rg[i], "]", sep = ""))
      if(cfg$rating == F) return(label.out)

      # 2. Labels for rating data
      for(i in 1:ncol(cfg$n.sdt)) {
          for(j in 1:cfg$n.sdt[1, i]) {
             label.out <- c(label.out, paste("SDT-", i, ".", j, " [", 1:cfg$n.sdt[2, i], "]", sep = ""))
          }
      }    
      return(label.out)
}

# ----------------------------------------------------------------------------------------------------------------------
# HELP FUNCTION
# Input:    cfg     = configuration information
# Output:   A list with three entries:
#           (1) AFC = A 2xn matrix with positions of the means in the first row and the position of the sd in second row (AFC model)
#           (2) SDT = A 2xn matrix with positions of the means in the first row and the position of the sd in second row  (SDT model)
#           (3) TH  = A list of thresholds for the single SDT models
# ----------------------------------------------------------------------------------------------------------------------
SDT.Rank.pos.par_ <- function(cfg)
{
      # 1. Positions of means and sd for AFC
      n.rk <- length(cfg$model)   # Number of ranking models       
      pos.mu.rk <- c(1 + (0:(2*n.rk-1)) * 2)
      pos.sd.rk <- c(2 + (0:(2*n.rk-1)) * 2)
      
      pos.afc <- matrix(c(pos.mu.rk, pos.sd.rk), nr = 2, byrow = T)
      colnames(pos.afc) <- paste("AFC-", 1:length(pos.mu.rk), sep = "")
      rownames(pos.afc) <- c("mean", "sd")
      
      # Postions of means and sd for SDT rating models
      if(cfg$rating == F) return(list(AFC = pos.afc, SDT = NULL, TH = NULL))

      # 2. Positions of means and sd for SDT
      offset <- n.rk*4 + 1       
      pos.sdt <- list()
      for(i in 1:ncol(cfg$n.sdt)) {
          pos.mu.sdt <- offset + (0:(cfg$n.sdt[1, i]-1) * 2)
          pos.sd.sdt <- offset + 1 + (0:(cfg$n.sdt[1, i]-1) * 2)
          pos.sdt[[i]] <- matrix(c(pos.mu.sdt, pos.sd.sdt), nr = 2, byrow = T)
          colnames(pos.sdt[[i]]) <- paste("SDT-", 1:ncol(pos.sdt[[i]]), sep = "")
          rownames(pos.sdt[[i]]) <- c("mean", "sd")
          if(i < ncol(cfg$n.sdt))
             offset <- offset + 2*cfg$n.sdt[1, i] + cfg$n.sdt[2,i] -1
      }
      
      # Position of the thresholds as a list
      offset <- n.rk*4 + 1            # End of ranking
      pos.th <- list(NULL)
      for(i in 1:ncol(cfg$n.sdt)) {
            offset <- offset + 2 * cfg$n.sdt[1,i]
            pos.th[[i]] <- offset:(offset+cfg$n.sdt[2,i] - 2)
            offset <- offset + cfg$n.sdt[2,i]-1
      }
      return(list(AFC = pos.afc, SDT = pos.sdt, TH = pos.th))
}