# ===========================================================================
# SDT models: Standard Signal detection models with robust thresholds 
# Date of creation:	June,   2012
# Date of last update:	June,   2012
#
# Author: Siegfried Macho
# ===========================================================================

SDT.Tol <- 1e-12
MAX.TH  <- 20           # maximal possible threshold: exp(MAX.TH)

#===============================================================================
#
# I. Helpfunctions
#
#===============================================================================

# ----------------------------------------------------------------------
# Function for computing parametervector with robust thresholds
# Input:	Vector of parameters: (mean, sigma, c[1], c[2], ...., c[n-1])
# Output: 	Vector with robust thresholds
# ----------------------------------------------------------------------
SDT.Robust.thresholds <- function(th)
{
      len.t <- length(th)
      if(len.t <= 1) return(th)
      
      th[th > MAX.TH] <- MAX.TH
      cumsum(c(th[1], exp(th[2:len.t])))
}


# ----------------------------------------------------------------------
# Function for computing parametervector with robust thresholds
# Input:	Vector of parameters: (mean, sigma, c[1], c[2], ...., c[n-1])
# Output: 	Vector with robust thresholds
# ----------------------------------------------------------------------
SDT.Robust.par <- function(parvec, n)
{
	m <- n-1								
	Par 		<- matrix(parvec[1:(2*m)], nr = m, byrow = T)
	thresholds 	<- parvec[(2*m + 1):length(parvec)]
      
      c(parvec[1:(2*m)], SDT.Robust.thresholds(thresholds))
}

#===============================================================================
#
# II. Model functions for computing model probabilities
#
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Computation or probabilities for robust SDT model
# Input:	par 	= Vector of parameters, consisting of the following entries
#			1. mean for each model (except noise model)
#			2. standard deviation for each model (except for noise model)
#			3. thresholds (on set for each model)
#		n   	= Number of models
#           fixed = 2xp matrix of fixed    constraints
#           ident = 2xq matrix of equality constraints
#
# Output: Vector of probabilities for single models
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT.Robust <- function(par, n, fixed = NULL, ident = NULL)
{
	# Separate parameters and thresholds for n models
	m <- n-1								
	Par 		<- matrix(par[1:(2*m)], nr = m, byrow = T)
	thresholds 	<- par[(2*m + 1):length(par)]
	
	# Robust thresholds
	thresholds <- SDT.Robust.thresholds(thresholds)
    	thresholds 	<- matrix(rep(thresholds, n), nr = n, byrow = T)	# Matrix with thresholds for each model

	# Add parameters for Noise distribution
	Par <- rbind(c(0, 1), Par)
	Par <- cbind(Par, thresholds)

	# Compute probabilities
	as.vector(apply(Par, 1, SDT.Probs))
}

# ------------------------------------------------------------------------------------------------------------------
# Model incorporating restrictions
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT.Full.Robust <- function(par, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
      par <- Restore.par(par, pos, fixed, ident, functional)	# Restore full parameter vector
      SDT.SDT.Robust(par, n)

}

#===============================================================================
#
# III. Functions computing derivatives
#
#===============================================================================

#-------------------------------------------------------------------------------
# Jacobian matrix for model with robust thresholds
# Input:    1. th = robust thresholds
#           2. k  = Number of thresholds
#
# Output: Jacobian matrix for adaptation of thresholds estimated in a robust way 
#-------------------------------------------------------------------------------
SDT.Jacobian.Robust <- function(th, k)
{
      if(k <= 1) return(1)
      
      th[th > MAX.TH] <- MAX.TH
      
      J.R <- matrix(0, k, k)
      J.R[, 1] <- 1
      for(i in 2:k) 
            J.R[i:k, i] <- rep(exp(th[i]), k-i+1)
      J.R
}                  

# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix of a single Probability model
# Jacobi-Matrix with respect to thresholds
# Jacobi-Matrix with respect to main parameters
# ------------------------------------------------------------------------------------------------------------------
SDT.Probs.Jacobi.Robust <- function(par)
{
	l   <- length(par)			      # Number of parameters
	n.t <- l-2					      # Number of thresholds
	m   <- par[1]
	sd  <- max(SDT.Tol, par[2])
	t.r <- par[3:l]                      # Raw thresholds  
	th  <- SDT.Robust.thresholds(t.r)    # Robust thresholds
	z  <- (th-m)/sd
	fn <- dnorm(z)
	
	# Derivative with respect to mean
	dm <- -fn/sd						# Derivative of Gaussian with respect to m
	dm <- c(dm[1], dm[-1] - dm[-n.t], -dm[n.t])

	# Derivative with respect to sd
	ds <- -z*fn/sd						# Derivative of Gaussian with respect to sd
	ds <- c(ds[1], ds[-1] - ds[-n.t], -ds[n.t])


	# Derivative with respect to thresholds
	dt <- fn/sd							# Derivative of Gaussian with respect to t
	
	J <- matrix(0, nr = n.t+1, nc = n.t)		# Jacobi matrix for thresholds: columns: Parameter, rows = probabilities
	r <- 1
	for(i in 1:n.t)
	{
		J[r, i] <- dt[i]
		r <- r + 1
		J[r, i] <- -dt[i]
	}
	# Adaptation of Jacobian for robust decision bounds
	J <- J %*% SDT.Jacobian.Robust(t.r, n.t)
	cbind(dm, ds, J)						# Full Jacobi matrix
}

# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix with respect to thresholds
# ------------------------------------------------------------------------------------------------------------------
SDT.Threshold.Jacobi.Robust <- function(par)
{
	l   <- length(par)			# Number of parameters
	n.t <- l-2					# Number of thresholds
	m   <- par[1]
	sd  <- max(SDT.Tol, par[2])
	t.r <- par[3:l]
	th   <- SDT.Robust.thresholds(t.r)
	z  <- (th-m)/sd
	fn <- dnorm(z)
	
	# Derivative with respect to thresholds
	dt <- fn/sd							# Derivative of Gaussian with respect to t[i]
	
	J <- matrix(0, nr = n.t+1, nc = n.t)		# Jacobi matrix for thresholds: columns: Parameter, rows = probabilities
	r <- 1
	for(i in 1:n.t)
	{
		J[r, i] <- dt[i]
		r <- r + 1
		J[r, i] <- -dt[i]
	}
      # Adaptation of Jacobian for robust decision bounds
	J %*% SDT.Jacobian.Robust(t.r, n.t)
								# Full Jacobi matrix
}

# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix of the SDT-Model
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT.Jacobi.Robust <- function(par, n)
{
	# Isolate parameters of the single models
	t <- par[(2*n - 1):length(par)]			# Thresholds
	n.t <- length(t)						# Number of thresholds

	J <- filler <- matrix(0, nr = n.t+1, nc = 2)
	filler.r <- filler.c <- NULL
	J.t <- SDT.Threshold.Jacobi.Robust(c(0, 1, t))						# Jacobimatrix of thresholds for Modell 1

	j <- 1
	for(i in 1:(n-1))
	{
		J.m <- SDT.Probs.Jacobi.Robust(c(par[j:(j+1)], t))
		J <- rbind(cbind(J, filler.c), cbind(filler.r, J.m[, 1:2, drop = F]))		# Jacobimatrix of mean and sd for Model i
		J.t <- rbind(J.t, J.m[, -c(1,2), drop = F])						# Add Jacobimatrix of thresholds for Model i

		if(i >= (n-1)) return( cbind(J, J.t))
		
		filler.c <- rbind(filler.c, filler)
		if(i == 1) filler.c <- rbind(filler.c, filler)				# Repeat due to first model
		filler.r <- cbind(filler.r, filler)
		j <- j + 2
		
	}
	cbind(J, J.t)
}

# ------------------------------------------------------------------------------------------------------------------
# Gradient of logL for the SDT and the Gaussian model
# Input:	Same as for Likelihood function
#
# Output:	Gradient
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT.Grad.Robust <- function(par, data, Model, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
	par 	   <- Restore.par(par, pos, fixed, ident, functional)	# Restore full parameter vector
	gradient <- as.vector(SDT.logL.Jacobi(par, data, Model, n, pos, fixed, ident, functional) %*% SDT.SDT.Jacobi.Robust(par, n)) 

	# Treat functional constraints
	if(!is.null(functional))
	{
		gr.fct <- attr(par, "SDTgradient")		# Isolate gradient function
		if(is.null(gr.fct))	stop("\n\n SDT.SDT.Grad.Robust: Error in computation of gradient: No gradient for functional constraints found:\n Set Parameter sym.gr = F in function SDT.Estimate !\n\n")
		gradient <- gr.fct(gradient)
	}	
	Elim.par.grad(gradient, fixed, ident)
}


#===============================================================================
#
# Functions for comupting the model matrix (analytically and numerically)
# for the SDT and Gaussian model
#
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Numerical computation of the model matrix (SDT model)
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT.Model.Matrix.Num.Robust <- function(full.par, n, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)
      par <- Elim.par(full.par, fixed, ident)
      J <- jacobian(SDT.SDT.Full.Robust, par, n = n, pos = pos, fixed = fixed, ident = ident, functional = functional)
}

# ------------------------------------------------------------------------------
# Analytical computation of the model matrix (SDT model)
# Input:    1. Full parameter vector
#           2. Number of signals
#           3. 2 x p - matrix of fixed constraints
#           4. 2 x q - matrix of identity constraints
#           5. Jacobian of functional constraints (supplied by the user)
#
# Output:   Modelmatrix
#
# Comment:  The Modelmatrix of funtional constraints must be a full matrix that
#           is premultiplied by the Jacobian
# ------------------------------------------------------------------------------
SDT.SDT.Model.Matrix.Robust <- function(full.par, n, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)

      J   <- SDT.SDT.Jacobi.Robust(full.par, n = n)
      if(!is.null(ident))
      {
            J.Id <- diag(length(full.par))
            Index <- matrix(c(ident[2,], ident[1,]), nc = 2)
            J.Id[Index] <- 1
            J <- J %*% J.Id
      }
      
      # Treatment of funtional constraints
      if(!is.null(functional))
      {
            J.fct <- attr(functional, "SDT.Jacobian")
            if(is.null(J.fct)) 
                  return(NULL)           # No Jacobian function supplied -> return (NULL)
            J <- J %*% J.fct(full.par)
      }
      J[, pos]
}

# ------------------------------------------------------------------------------------------------------------------
# Adjustment of the asymptotic covariance matrix in case of robust estimation
# Input:	Cov.asym = asymptotic covariance matrix not adjusted
#           full.par = Full parameter vector
#           nmodel   = number of models
#           fixed    = Matrix of fixed constraints
#           ident    = Matrix of identity constraints
#           npar     = Number of parameters per model (5 in case of SDT2D [default]
#                      and 18 in case of SDT2D.MIX3)
#
# Output:	Adjusted asymptotic covariance matrix
# ------------------------------------------------------------------------------------------------------------------
SDT.Adjust.Robust <- function(Cov.asym, full.par, nmodel, fixed, ident)
{
      npar <- length(full.par)
      J.adj <- diag(npar)
      
      begin <- (nmodel-1)*2 + 1
      end   <- npar
      tlen  <- end-begin+1
      th    <- full.par[begin:end]
      J.adj[begin:end, begin:end] <- SDT.Jacobian.Robust(th, tlen)

      # Eliminate irrelevant parameters and perform adjustment
      selec <- Elim.par(1:npar, fixed, ident)
      J.adj <- J.adj[selec, selec]
      J.adj %*% Cov.asym %*% t(J.adj)
}


#===============================================================================
#
# Utility and output format functions
#
#===============================================================================


#------------------------------------------------------------------------------------------------------------------
# Generate starting parameters due to data and number of signals
# Input:	data = Vector of data or number or the number of data points per signal
#		   n = number of signals or (possibly in case of Gaussian model)
#			 a vector of number of response categories per signal
#
# Output: String with parameter names
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT.start.par.Robust <- function(n, data)
{
      n.dat <- length(data)
      if(n.dat < 2) 
            n.dat <- data*n
      
	# Specify Number of thresholds
	n.t <- as.integer(n.dat/n - 1)
	c(rep(c(0, 1), n-1), rep(-1.2, length = n.t))
}

# ------------------------------------------------------------------------------------------------------------------
# Names of the models 
# ------------------------------------------------------------------------------------------------------------------
SDT.model.name.Robust <- "Standard SDT model with noise distribution N(0, 1), ROBUST decision bounds"


# ------------------------------------------------------------------------------------------------------------------
# Format free parameters and compute confidence intervals
# Input:	1. px2 Matrix with parameters and standard errors
#		2. Full parameter vector
#           3. Model identification string
#		4. Number of signals
#		5. Matrix of fixed constraints
#		6. Matrix of equality constraints
#		7. Confidence level
#           8. Number of decimal places
#
# Output: px5 Matrix of parameters, standard errors and confidence limits
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Par.est.Robust <- function(Pars, full.par, Model.Id, n, fixed, ident, alpha, deci)
{
      
      # Compute correct decision bounds
      full.par.adj <- SDT.Robust.par(full.par, n)
      Pars[,1] <- Elim.par(full.par.adj, fixed, ident)
      qt.cfi <- qnorm((1-alpha)/2)
      cfi.lo <- Pars[,1] + Pars[,2]*qt.cfi
      cfi.up <- Pars[,1] - Pars[,2]*qt.cfi
            
      # Respect lower bounds on sd parameter
      bd.lo <- Elim.par(SDT.par.bounds(full.par, n)$lower, fixed, ident)
      cfi.lo <- ifelse(cfi.lo < bd.lo, bd.lo, cfi.lo)
       
        
      Pars <- cbind(Pars, cfi.lo, cfi.up)
      colnames(Pars) <- c("Value", "SE", paste("CFI-", alpha*100, "(Lower)", sep = ""), paste("CFI-", alpha*100, "(Upper)", sep = ""))
      round(Pars, deci)
}

# ------------------------------------------------------------------------------------------------------------------
# Format full parameter vector for output
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Par.Robust <- function(full.par, n, deci)
{
      full.par.adj <- SDT.Robust.par(full.par, n)
	full.par <- c(0, 1, full.par.adj)				       # Add parameters for N(0, 1)
	th  <- round(full.par[(2*n+1): length(full.par)], deci)	 # Isolate thresholds
	n.t <- length(th)
	th <- matrix(rep(th, n), nc = n)			# Copy thresholds and format them
	
	Fp <- matrix(round(full.par[1:(2*n)], deci), nc = n)
	Fp <- rbind(Fp, th)

	colnames(Fp) <- paste("Gauss-", 1:n, sep = "")
	rownames(Fp) <- c("Mean", "Stddev", paste("t-", 1:n.t, sep = ""))

	Fp
}
