# ===========================================================================
# SDT.2D model: Bivariate signal detection model
# Date of creation:	March,      2010
# Date of last update:	September,  2013
#
# Author: Siegfried Macho
# ===========================================================================
library(mvtnorm)	# Load multivariate normal distribution library
library(ellipse)	# Load functions for computing confidence ellipses

SDT2D.Tol <- 1e-12

#===============================================================================
#
# I. Auxiliary functions
#
#===============================================================================
# ------------------------------------------------------------------------------
# Function for computing robust decision bounds
# Input:	1. parvec    = vector of parameters
#           2. cfg       = confiduration information
#           3. Gauss.par = number of Gaussian parameters (default = 5)
#
# Output:   vector of parameters with decision bounds computed
# ------------------------------------------------------------------------------
SDT2D.robust.thresholds <- function(parvec, cfg, Gauss.par = 5)
{
      robust <- cfg$robust
      if(is.null(robust)) return(parvec)
      if(!robust) return(parvec)
      
      ds <- cfg$ds
      if(is.null(ds)) ds <- 0
      s  <- cfg$n.sdt
      k1 <- cfg$k1
      k2 <- cfg$k2
      n1 <- k1 + 1
      n2 <- k2 + 1
      
      if(ds == 0)
      {
            t1 <- parvec[(s*Gauss.par + 1) : (s*Gauss.par + k1)]
            t2 <- parvec[(s*Gauss.par + k1 + 1) : (s*Gauss.par + k1 + k2)]
            
            t1 <- SDT2D.robust.thresholds.simple(t1, k1)
            t2 <- SDT2D.robust.thresholds.simple(t2, k2)
            return(c(parvec[1:(s*Gauss.par)], t1, t2))
      }
      if(ds == 1)
      {
            t1 <- matrix(parvec[(s*Gauss.par + 1): (s*Gauss.par + k1*n2)], nr = k1)
            if(k1 > 1)
            {
                  t.exp  <- exp(t1[2:k1, 1:n2])
                  t1    <- rbind(t1[1, ], t.exp)
                  t1    <- apply(t1, 2, cumsum)
            }
            t2 <- parvec[(s*Gauss.par + k1*n2 + 1) : (s*Gauss.par + k1*n2 + k2)]
            t2 <- SDT2D.robust.thresholds.simple(t2, k2)
            return(c(parvec[1:(s*Gauss.par)], as.vector(t1), t2))
      }
      if(ds == 2)
      {
            t1 <- parvec[(s*Gauss.par + 1) : (s*Gauss.par + k1)]
            t1 <- SDT2D.robust.thresholds.simple(t1, k1)
       
            t2 <- matrix(parvec[(s*Gauss.par + k1 + 1): (s*Gauss.par + k1 + n1*k2)], nr = k2)
            if(k2 > 1)
            {
                  t.exp  <- exp(t2[2:k2, 1:n1])
                  t2    <- rbind(t2[1, ], t.exp)
                  t2    <- apply(t2, 2, cumsum)
            }
            return(c(parvec[1:(s*Gauss.par)], t1, as.vector(t2)))
      }
      return(parvec)
}
# ------------------------------------------------------------------------------
# Function for computing robust decision bounds in case of ds = 0
# Input:	1. t = vector of decision bounds
#           2. k = number of decision bounds
#
# Output:   vector of decision bounds
# ------------------------------------------------------------------------------
SDT2D.robust.thresholds.simple <- function(t, k)
{
      if(k <= 1) return(t)
      t[2:k] <- exp(t[2:k])
      cumsum(t)
}

# ------------------------------------------------------------------------------
# Function for generating a matrix of the decision bounds
# Input:	1. parvec = parameter vector that contains the parameters in the 
#                       following order
#                       i.   For each distribution:     m1, m2, sd1, sd2, r
#                       ii.  Decision bounds on Dimension 1
#                       iii. Decision bounds on Dimension 2
#           2. s      = Number of distributions (signals)
#           3. k1     = Number of decision bounds on Dimension 1
#           4. k2     = Number of decision bounds on Dimension 2
#           5. ds     = Flag indicating violations of decisional separability
#                       ds = 0 --> No violation
#                       ds = 1 --> Violation on Dimension 1
#                       ds = 2 --> Violation on Dimension 2
#           6. robust = Flag inidicting whether decision bounds should be estimated robustly
#           7. Number of Gaussian parameters (default = 5)
#
# Output: 	A [n*(k1+1)*(k2+1) x 4] matrix with lower and upper decision bounds 
#           for each response region, repeated for each signal distribution. 
#           The entries within a row are:
#                       t[1] = lower bound on Dimension 1
#                       t[2] = upper bound on Dimension 1
#                       t[3] = lower bound on Dimension 2
#                       t[4] = upper bound on Dimension 2
#
# Comment:  The function will be used by the model SDT2D-MIX-PD. In this case the number of 
#           Gaussian parameters passed to the function is Gauss.par = 18
# ----------------------------------------------------------------------------------------
SDT.SDT2D.Create.Decision.bounds <- function(parvec, s, k1, k2, ds = 0, robust, Gauss.par = 5)
{
      # Decisional separability not violated
      if(ds == 0)
      {
            
            t1 <- parvec[(s*Gauss.par + 1) : (s*Gauss.par + k1)]
            t2 <- parvec[(s*Gauss.par + k1 + 1) : (s*Gauss.par + k1 + k2)]
 
            n1 <- k1 + 1
            n2 <- k2 + 1

            if(robust)  # Compute thresholds in case of robust thresholds
            {
                  t1 <- SDT2D.robust.thresholds.simple(t1, k1)
                  t2 <- SDT2D.robust.thresholds.simple(t2, k2)
            }
            return(cbind(rep(rep(c(-Inf, t1), each = n2), s),
                         rep(rep(c(t1,  Inf), each = n2), s),
                         rep(rep(c(-Inf, t2), n1), s),
                         rep(rep(c(t2,  Inf), n1), s)
                        )
                  ) 
      }
      if (ds == 1)
      {
            n2 <- k2 + 1
            n1 <- k1 + 1
 
            # Insert -Inf and +Inf
            t1 <- matrix(parvec[(s*Gauss.par + 1): (s*Gauss.par + k1*n2)], nr = k1)
            if(k1 > 1 && robust)
            {
                  t.exp  <- exp(t1[2:k1, 1:n2])
                  t1    <- rbind(t1[1, ], t.exp)
                  t1    <- apply(t1, 2, cumsum)
            }
            
            t1L <- as.vector(t(rbind(rep(-Inf, n2), t1)))
            t1U <- as.vector(t(rbind(t1, rep(Inf, n2))))

            t2  <- parvec[(s*Gauss.par + k1*n2 + 1) : (s*Gauss.par + k1*n2 + k2)]
            if(robust)  t2 <- SDT2D.robust.thresholds.simple(t2, k2)
            
            return(cbind(t1L, t1U,
                         rep(rep(c(-Inf, t2), n1), s),
                         rep(rep(c(t2,  Inf), n1), s)
                        )
                  ) 
             
      }
      # ds = 2

      n2 <- k2 + 1
      n1 <- k1 + 1
 
      t1 <- parvec[(s*Gauss.par + 1) : (s*Gauss.par + k1)]
      if(robust) t1 <- SDT2D.robust.thresholds.simple(t1, k1)
 
    # Insert -Inf and +Inf

      t2 <- matrix(parvec[(s*Gauss.par + k1 + 1): (s*Gauss.par + k1 + n1*k2)], nr = k2)
      if(k2 > 1 && robust)
      {
            t.exp  <- exp(t2[2:k2, 1:n1])
            t2    <- rbind(t2[1, ], t.exp)
            t2    <- apply(t2, 2, cumsum)
      }
 
      t2L <- as.vector(rbind(rep(-Inf, n1), t2))
      t2U <- as.vector(rbind(t2, rep(Inf, n1)))

      return(cbind(rep(rep(c(-Inf, t1), each = n2), s),
                   rep(rep(c(t1,  Inf), each = n2), s),
                   t2L, t2U)
            ) 
}

#===============================================================================
#
# II. Functions for computing the model probabilities
#
#===============================================================================

# ---------------------------------------------------------------------------------
# The bivariate standard Gaussian density function
# Input:	1. x
#		2. y
#		3. r
#
# Output: 	Density of the bivariate Gaussian
#
# Comments:	1. To speed up the process no check concerning correctness is provided
#		2. The function cannot be used as a vector function
# ----------------------------------------------------------------------------------
SDT2D.d2znorm <- function(x, y, r)
{
	TolMin <- -1e12
	TolMax <-  1e12

	if(x >= TolMax || x <= TolMin || y >= TolMax || y <= TolMin) return(0)

	v <- 1 - r*r
	if(v <= SDT2D.Tol) return(0)

      exp(-(x*x - 2*r*x*y + y*y)/2/v) / 2/pi/sqrt(v)
}


# ----------------------------------------------------------------------------------------
# Function for computating the areas under the Gaussian 2D distribution 
# Input:	parvec = parameter vector of length 8:
#                    parvec[1] = mean 1
#                    parvec[2] = mean 2
#                    parvec[3] = sd 1
#                    parvec[4] = sd 2
#                    parvec[5] = r
#                    parvec[6] = lower bound on Dimension 1
#                    parvec[7] = upper bound on Dimension 1
#                    parvec[8] = lower bound on Dimension 2
#                    parvec[9] = upper bound on Dimension 2
#
#           robust: Flag indicating whether decision bounds are estimated in a robust manner
#                   i.e. The order of decision bounds is guarantied
#
# Output:   The probability under the bivariate Gaussian distribution
# ---------------------------------------------------------------------------------------
SDT2D.Prob <- function(parvec, robust)
{
      if(!robust)
      {      # Check whether decision bounds are ordered correctly, if not set them to equal values
            if(parvec[7] < parvec[6])
                  parvec[6] <- parvec[7] <- (parvec[6] + parvec[7])/2

            if(parvec[9] < parvec[8])
                  parvec[8] <- parvec[9] <- (parvec[8] + parvec[9])/2
      }

      # Extract parameters
      mean <- parvec[1:2]
      std  <- parvec[3:4]
      r    <- parvec[5]

      # Check sigmas and r so that thea are within bounds
      std[std < SDT2D.Tol] <- SDT2D.Tol
      r[r > 1 - SDT2D.Tol] <- 1 - SDT2D.Tol
      r[r < SDT2D.Tol - 1] <- SDT2D.Tol - 1

 	# Create covariance matrix		
	Sigma <- diag(std * std)
	Sigma[2,1] <- Sigma[1,2] <- std[1] * std[2] * r

     pmvnorm(lower  = c(parvec[6], parvec[8]),
	     	  upper = c(parvec[7], parvec[9]),
		  mean  = mean,
		  sigma = Sigma)
 }

# ---------------------------------------------------------------------------------
# The Gaussian 2D-SDT Model
# Input:	1. parvec = Parameter vector
#                       i.   For each distribution: m1, m2, sd1, sd2, r
#                       ii.  Decision bounds on Dimension 1
#                       iii. Decision bounds on Dimension 2
#
#           2. A configuration list with the following entries
#                 i.           n  = Number of signal distributions
#                 ii.          k1 = Number of decision bounds on Dimension 1
#                 iii.         k2 = Number of decision bounds on Dimension 2
#                 iv.          ds = Flag indicating the presence of violations of descision separability
#                                   ds = 0 --> No violation
#                                   ds = 1 --> Violation on Dimension 1
#                                   ds = 2 --> Violation on Dimension 2
#                 v.         PA   = Response assignment matrix: This matrix is premaultiplied with response vector
#                 vi. struct.zero = Vector of structural zeros
#
#           3. fixed: fixed constraints (not used, but required for compatibility)
#           4. ident: equality vonstraints  (not used, but required for compatibility)                                       
#
# Output: 	Vector of length n*(k1 + 1)*(k2 + 1) containing the response probabilities
# ----------------------------------------------------------------------------------
SDT.SDT2D <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
      # 1. Extract data from command list
      s  <- cfg$n.sdt
      k1 <- cfg$k1
      k2 <- cfg$k2
      ds <- cfg$ds
      R  <- cfg$R
      struct.zero <- cfg$struct.zero
      ren <- cfg$ren
      conditional <- cfg$conditional
      robust <- cfg$robust
      if(is.null(conditional)) conditional <- F
      if(is.null(robust))      robust <- F
            

      # 2. Second create decision bounds
      DB <- SDT.SDT2D.Create.Decision.bounds(parvec, s, k1, k2, ds, robust)

      # 3. Create parameter matrix
      PM <- NULL
      for(i in 1:s)
      {
            parvec.rep <- rep(parvec[(5*(i-1) + 1): (5*i)], each = (k1+1) * (k2+1))
            PM <- rbind(PM, matrix(parvec.rep, nc = 5))
      }

      p.est <- apply(cbind(PM, DB), 1, SDT2D.Prob, robust = robust)
      
      # 4. Treatment of structural zeros
      if(!is.null(struct.zero)) 
      {
            p.est[struct.zero] <- 0
            if(conditional)               # Renormalization of probabilities
            {
                  p.est <- matrix(p.est, nc = s)
                  csum  <- colSums(p.est)
                  p.est <- as.vector(t(apply(p.est, 1, function(x) { x / csum} )))
            }
      }
      if(!is.null(R)) p.est <- as.vector(R %*% p.est)
      p.est
}

#===============================================================================
#
# III. Functions for computing derivatives: Jacobians and gradients
#
#===============================================================================


#-------------------------------------------------------------------------------------------
# Symbolic gradient of the general bivariate normal distribution with variable upper and
# lower bounds:
#
# Input:  A vector of length 7 with the following entries:
#		1. m1 (Mean on Dimension 1)
#		2. m2 (Mean on Dimension 2)
#		3. s1 (Standard deviation on Dimension 1)
#		4. s2 (Standard deviation on Dimension 2)
#		5. r  (Correlation)
#		6. t1l (Lower decision bound on Dimension 1)
#		7. t1u (Upper decision bound on Dimension 1)
#		8. t2l (Lower decision bound on Dimension 2)
#		9. t2u (Upper decision bound on Dimension 2)
#
# Output: Gradient vector of the bivariate normal distribution with the 3 partial derivatives
#-------------------------------------------------------------------------------------------
SDT2D.Prob.Jacobi <- function(parvec, robust)
{
	m1  <- parvec[1]
	m2  <- parvec[2]
      s1 <- max(parvec[3], SDT2D.Tol)
      s2 <- max(parvec[4], SDT2D.Tol)
      r  <- min(max(parvec[5], SDT2D.Tol-1), 1-SDT2D.Tol)

      if(!robust)
      {
            # Check whether decision bounds are ordered correctly, if not, set them to equal values
            if(parvec[7] < parvec[6])
                  parvec[6] <- parvec[7] <- (parvec[6] + parvec[7])/2
         
            if(parvec[9] < parvec[8])
                  parvec[8] <- parvec[9] <- (parvec[8] + parvec[9])/2
      }

	t1l <- parvec[6]
	t1u <- parvec[7]
	t2l <- parvec[8]
	t2u <- parvec[9]

	# Normalize the thresholds
	tau1l <- (t1l-m1)/s1
	tau1u <- (t1u-m1)/s1

	tau2l <- (t2l-m2)/s2
	tau2u <- (t2u-m2)/s2
	
	# Compute determinant
	v <- 1 - r*r
     	if(v <= SDT2D.Tol) return(rep(0, 9))

	# Compute gradients
	if(tau1l == -Inf)      dtau1l <- 0
      else                   dtau1l <- dnorm(tau1l)*pnorm(tau2l, r*tau1l, sqrt(v)) - dnorm(tau1l)*pnorm(tau2u, r*tau1l, sqrt(v))

      if(tau1u ==  Inf)      dtau1u <- 0
      else                   dtau1u <- dnorm(tau1u)*pnorm(tau2u, r*tau1u, sqrt(v)) - dnorm(tau1u)*pnorm(tau2l, r*tau1u, sqrt(v))
      
      
	if(tau2l == -Inf)      dtau2l <- 0
      else                   dtau2l <- dnorm(tau2l)*pnorm(tau1l, r*tau2l, sqrt(v)) - dnorm(tau2l)*pnorm(tau1u, r*tau2l, sqrt(v))
      
      if(tau2u ==  Inf)      dtau2u <- 0
	else                   dtau2u <- dnorm(tau2u)*pnorm(tau1u, r*tau2u, sqrt(v)) - dnorm(tau2u)*pnorm(tau1l, r*tau2u, sqrt(v))
	
	dm1   <- -(dtau1u + dtau1l)/s1
	dm2	<- -(dtau2u + dtau2l)/s2

	dr    <- SDT2D.d2znorm(tau1u, tau2u, r) -
		   SDT2D.d2znorm(tau1u, tau2l, r) -
		   SDT2D.d2znorm(tau1l, tau2u, r) +
		   SDT2D.d2znorm(tau1l, tau2l, r)

      if(dtau1l == 0) tau1l <- 0
      if(dtau1u == 0) tau1u <- 0
      if(dtau2l == 0) tau2l <- 0
      if(dtau2u == 0) tau2u <- 0
      
	ds1   <- -(tau1u*dtau1u + tau1l*dtau1l)/s1
	ds2   <- -(tau2u*dtau2u + tau2l*dtau2l)/s2

	dt1l  <- 1/s1*dtau1l
	dt1u  <- 1/s1*dtau1u

	dt2l	<- 1/s2*dtau2l
	dt2u	<- 1/s2*dtau2u

	c(dm1, dm2, ds1, ds2, dr, dt1l, dt1u, dt2l, dt2u)
}

    
#-------------------------------------------------------------------------------------------
# Jacobian matrix of SDT.SDT2D model
# Input:    1, The parameter vector
#		2. List with configuration informtion
#		3. Matrix of fixed constraints         (NOT USED, for compaitibility purpose only)
#		4. Matrix of equality constraints      (NOT USED)
#
# Output: Jacobian matrix with parameters making up columns and probabilities making up rows
#-------------------------------------------------------------------------------------------
# Helpfuntion that indexes the items correctly
SDT2D.Index.fct.Jacobi <- function(Working.vector, tlen)
{
            out <- rep(0, tlen)                       # Output vector with 
            target.range <- Working.vector[5:8]       # Part of the vector where indices are located
            index.t <- target.range[target.range > 0] # Use only indices > 0
            index.s <- (1:4)[target.range > 0]
            out[index.t] <- Working.vector[index.s]   # Create Index
            out
}     

#-------------------------------------------------------------------------------------------
# Jacobian matrix for Robust model
# Input:    1. t = thresholds
#           2. k = Number of thresholds
#
# Output: Jacobian matrix with thresholds estimated in a robust way (ds = 0)
#-------------------------------------------------------------------------------------------
SDT2D.Jacobian.Robust <- function(t, k)
{
      J.R <- matrix(0, k, k)
      J.R[1, ] <- 1
      for(i in 2:k) 
            J.R[i, i:k] <- rep(exp(t[i]), k-i+1)
      J.R
}                  


SDT.SDT2D.Jacobi <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
      # 1. Extract data from command list
      s  <- cfg$n.sdt
      k1 <- cfg$k1
      k2 <- cfg$k2
      n1 <- k1 + 1
      n2 <- k2 + 1
      ds <- cfg$ds
      R  <- cfg$R
      struct.zero <- cfg$struct.zero
      ren <- cfg$ren
      conditional <- cfg$conditional
      robust      <- cfg$robust
      if(is.null(conditional)) conditional <- F
      if(is.null(robust))      robust      <- F

      # 2. Second create decision bounds
      DB <- SDT.SDT2D.Create.Decision.bounds(parvec, s, k1, k2, ds, robust)      

      PM <- NULL
      for(i in 1:s)
      {
            parvec.rep <- rep(parvec[(5*(i-1) + 1): (5*i)], each = (k1+1) * (k2+1))
            PM <- rbind(PM, matrix(parvec.rep, nc = 5))
      }

      # Computation of Gradients for all probability points
      Grad.M <- apply(cbind(PM, DB), 1, SDT2D.Prob.Jacobi, robust = robust)

      # Now construct Jacobian of Gaussian parameters from 5 x [k1+1]*[k2+1]-Matrix
      # Or 5 x nrows(R)-Matrix
      npar  <- 5                    # Number of Gaussian Parameters
      npro  <- (k1+1)*(k2+1)        # Number of estimated probabilities per signal
      filler <- matrix(0, nr = npro, nc = npar)
      filler.r <- filler.c <- NULL
      for(i in 1:s)
      {
            # Extract matrix
            J.s <- Grad.M[1:npar, (npro*(i-1)+1):(npro*i)]
            if(i == 1) J.Gauss <- t(J.s)
            else       J.Gauss <- rbind(cbind(J.Gauss, filler.c), cbind(filler.r, t(J.s)))

            if(i < s)
            {
                  filler.c <- rbind(filler.c, filler)
                  filler.r <- cbind(filler.r, filler)
            }
       }
      
      # Now compiling the thresholds
      tlen <- k1 + k2
      if(ds == 0)
      {
            Index.set <- cbind(rep(rep(c(0, 1:k1), each = n2), s),
                               rep(rep(c(1:k1, 0), each = n2), s),
                               rep(rep(c(0, 1:k2+k1), n1), s),
                               rep(rep(c(1:k2+k1, 0), n1), s)
                              )
            Working.array <- rbind(Grad.M[6:9, ], t(Index.set))
      }
      else if(ds == 1)
      {
            tlen <- k1*n2 + k2
 
            t1  <- matrix(1:(k1*n2), nr = k1)
            t1L <- as.vector(t(rbind(rep(0, n2), t1)))
            t1U <- as.vector(t(rbind(t1, rep(0, n2))))
            Index.set <- cbind(t1L, t1U, 
                               rep(rep(c(0, 1:k2 + k1*n2), n1), s),
                               rep(rep(c(1:k2 + k1*n2, 0), n1), s)
                              )
            Working.array <- rbind(Grad.M[6:9, ], t(Index.set))
   }
      
      else  # ds = 2
      {
            tlen <- k1 + k2*n1
            
            t2  <- matrix(1:(k2*n1)+ k1, nr = k2)
            t2L <- as.vector(rbind(rep(0, n1), t2))
            t2U <- as.vector(rbind(t2, rep(0, n1)))
      
            Index.set <- cbind(rep(rep(c(0, 1:k1), each = n2), s),
                               rep(rep(c(1:k1, 0), each = n2), s),
                               t2L, t2U)
            Working.array <- rbind(Grad.M[6:9, ], t(Index.set))
      }
      
      J.t <- apply(Working.array, 2, SDT2D.Index.fct.Jacobi, tlen = tlen)
      
      # Adjustment of thresholds in case of robust estimation
      if(robust && (k1 > 1 || k2 > 1))
      {
            # Extract relevant columns representing thresholds from Jacobian matrix 
            if(ds == 0)
            { 
                  J.t1 <- J.t[1:k1, ]
                  J.t2 <- J.t[(k1+1):(k1+k2), ]
            }
            else if(ds == 1)
            { 
                  J.t1 <- J.t[1:(k1*n2), ]
                  J.t2 <- J.t[(k1*n2 + 1):(k1*n2 + k2), ]
            }
            else
            { 
                  J.t1 <- J.t[1:k1, ]
                  J.t2 <- J.t[(k1 + 1):(k1 + k2*n1), ]
            } 
                       
            # Adjust derivatives for thrsholds
            if(k1 > 1)
            {     # Adjust thresholds on D1
                  if(ds == 0 || ds == 2)
                        t1 <- parvec[(s*5 + 1):(s*5 + k1)]
                  else 
                        t1 <- parvec[(s*5 + 1): (s*5 + k1*n2)]
                  if(ds == 0 || ds == 2)
                  {      
                        J.R <- SDT2D.Jacobian.Robust(t1, k1)    
                        J.t1 <- J.R %*% J.t1
                  }
                  else   # ds = 1
                  {
                        start <- 1
                        end   <- k1
                        for(i in 1:n2)
                        {
                              J.R <- SDT2D.Jacobian.Robust(t1[start:end], k1)    
                              J.t1[start:end,] <- J.R %*% J.t1[start:end, ]
                              start <- start + k1
                              end   <- end   + k1  
                        }     
                  }      
            }
            if(k2 > 1)
            {
                  # Adjust thresholds on D2
                  if(ds == 0)
                        t2 <- parvec[(s*5 + k1 + 1):(s*5 + k1 + k2)]
                  else if(ds == 1)
                        t2 <- parvec[(s*5 + k1*n2 + 1):(s*5 + n2*k1 + k2)]
                  else
                        t2 <- parvec[(s*5 + k1 + 1):(s*5 + k1 + n1*k2)]
                  if(ds == 0 || ds == 1)
                  {      J.R <- SDT2D.Jacobian.Robust(t2, k2)    
                        J.t2 <- J.R %*% J.t2
                  }
                  else  # ds = 2
                  {
                        start <- 1
                        end   <- k2
                        for(i in 1:n1)
                        {
                              J.R <- SDT2D.Jacobian.Robust(t2[start:end], k2)    
                              J.t2[start:end,] <- J.R %*% J.t2[start:end, ]
                              start <- start + k2
                              end   <- end   + k2  
                        }     
                 
                  }
            }
            J.t  <- rbind(J.t1, J.t2)
      }
      
      J <- cbind(J.Gauss, t(J.t))

      # Treat structural zeros
      if(!is.null(cfg$struct.zero))
      {
            J[cfg$struct.zero,] <- 0
            if(conditional)  J <- SDT.SDT2D.Jacobi.conditional(J, parvec, cfg, fixed = NULL, ident = NULL)
       }
      
      # Treat grouping
      if(!is.null(cfg$R))
            J <- cfg$R %*% J
      J
}


# ------------------------------------------------------------------------------------------------------------------
# Adaptation of Modelmatrix (Jacobian matrix) in case of renormalization
# Input:	1. Jacobian = non-adjusted Jacobian
#           2. parvec   = vector of parameters
#           3. cfg      = configuration inforamtion
#           4. fixed (not used)
#           5. ident (not used)
#
# Output:	New jacobian
# ------------------------------------------------------------------------------------------------------------------

#----------------------------------------------------------------
# Help function constructing the Jacobian from the probabilities
#----------------------------------------------------------------
SDT2D.normal.J <- function(p)
{
      s <- sum(p)
      s2 <- s*s
      
      n.p <- length(p)
      J <- NULL
      for(i in 1:n.p)
            J <- rbind(J, rep(-p[i]/s2, n.p))

      diag(J) <- diag(J) + 1/s
      J
      
}

#------------------------------------------------------------------
# Main function for performing the adjustment due to normalization
#------------------------------------------------------------------
SDT.SDT2D.Jacobi.conditional <- function(J, parvec, cfg, fixed = NULL, ident = NULL)
{
      cfg$conditional <- NULL
      cfg$R           <- NULL  
      s <- cfg$n.sdt
      
      p.est <- SDT.SDT2D(parvec, cfg, fixed, ident)   # Compute probabilites igonring condtionalization amd pooling

      J.cond <- NULL
      n.p <- dim(J)[1]/s      # Number of estimated probabilities per model
      for(i in 1:s)
      {
            range.act <- ((n.p)*(i-1) + 1) : (n.p*i)         # Indices of probabilities for actual model
            if(any(cfg$struct.zero %in% range.act))          # Structural zeros for actual model (signal) present
            {
                  p.act <- p.est[range.act]
                  J.cond <- SDT2D.normal.J(p.act)            # Construct Jacobian that represents the normalization         
                  J[range.act, ] <- J.cond %*% J[range.act, ]
            }
      }
      J
}

# ------------------------------------------------------------------------------------------------------------------
# Adjustment of the asymptotic covariance matrix in case of robust estimation
# Input:	Cov.asym = asymptotic covariance matrix not adjusted
#           full.par = Full parameter vector
#           cfg      = List with configuration information
#           fixed    = Matrix of fixed constraints
#           ident    = Matrix of identity constraints
#           npar     = Number of parameters per model (5 in case of SDT2D [default]
#                      and 18 in case of SDT2D.MIX3)
#
# Output:	Adjusted asymptotic covariance matrix
# ------------------------------------------------------------------------------------------------------------------
SDT.2D.Adjust.robust <- function(Cov.asym, full.par, cfg, fixed, ident, npar = 5)
{
      if(is.null(cfg$robust) || cfg$k1 + cfg$k2 <= 2) return(Cov.asym)
      if(cfg$robust == F)                             return(Cov.asym)     
      
      ds <- cfg$ds
      s  <- cfg$n.sdt
      k1 <- cfg$k1
      k2 <- cfg$k2
      
      p <- length(full.par)
      J.adj <- diag(p)
      
      if(ds == 0)
      {
            if(k1 > 1)
            {
                  begin <- s*npar + 1
                  end   <- s*npar + k1
                  t1    <- full.par[begin:end]
                  J.adj[begin:end, begin:end] <- SDT2D.Jacobian.Robust(t1, k1)
            }
            if(k2 > 1)
            {
                  begin <- s*npar + k1 + 1
                  end   <- s*npar + k1 + k2
                  t2    <- full.par[begin:end]
                  J.adj[begin:end, begin:end] <- SDT2D.Jacobian.Robust(t2, k2)
            }
      }      
      else if(ds == 1)
      {
            if(k1 > 1)
            {
                  n2 <- k2 + 1
                  begin <- s*npar + 1
                  end   <- s*npar + k1
                  for(i in 1:n2)
                  {
                        t1    <- full.par[begin:end]
                        J.adj[begin:end, begin:end] <- SDT2D.Jacobian.Robust(t1, k1)
                        begin <- begin + k1
                        end   <- end + k1
                  }
            }
            if(k2 > 1)
            {
                  begin <- s*npar + n2*k1 + 1
                  end   <- s*npar + n2*k1 + k2
                  t2    <- full.par[begin:end]
                  J.adj[begin:end, begin:end] <- SDT2D.Jacobian.Robust(t2, k2)
            }
      }
      else # (ds == 2)
      {
            if(k1 > 1)
            {
                  begin <- s*npar + 1
                  end   <- s*npar + k1
                  t1    <- full.par[begin:end]
                  J.adj[begin:end, begin:end] <- SDT2D.Jacobian.Robust(t1, k1)
            }
            if(k2 > 1)
            {
                  n1 <- k1 + 1
                  begin <- s*npar + k1 + 1
                  end   <- s*npar + k1 + k2
                  for(i in 1:n1)
                  {
                        t2 <- full.par[begin:end]
                        J.adj[begin:end, begin:end] <- SDT2D.Jacobian.Robust(t2, k2)
                        begin <- begin + k2
                        end   <- end + k2
                  }
            }
      }
      
      # Eliminate irrelevant parameters and perform adjustment
      selec <- Elim.par(1:p, fixed, ident)
      J.adj <- J.adj[selec, selec]
      t(J.adj) %*% Cov.asym %*% J.adj
}

# ------------------------------------------------------------------------------------------------------------------
# Gradient of logL for the SDT2D model
# Input:	Same as for Likelihood function
#
# Output:	Gradient
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT2D.Grad <- function(par, data, Model, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
	par 	   <- Restore.par(par, pos, fixed, ident, functional)	# Restore full parameter vector
	gradient <- as.vector(SDT.logL.Jacobi(par, data, Model, n, pos, fixed, ident, functional) %*% SDT.SDT2D.Jacobi(par, n)) 

	# Treat functional constraints
	if(!is.null(functional))
	{
		gr.fct <- attr(par, "SDT.gradient")		# Isolate gradient function
		if(is.null(gr.fct))	stop("\n\n Error in computation of gradient: No gradient for functional constraints found:\n Set Parameter sym.gr = F in function SDT.Estimate !\n\n")
		gradient <- gr.fct(gradient, par)
	}	
	Elim.par.grad(gradient, fixed, ident)
}

#===============================================================================
#
# IV. Functions computing starting parameters, and names of parameters
#
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Generate starting parameters due to data and number of signals
# Input:    cfg = list with configuration information
#			 
# Output:   Vector of starting parameter
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT2D.start.par <- function(cfg)
{
      s      <- cfg$n.sdt
      k1     <- cfg$k1
      k2     <- cfg$k2
      ds     <- cfg$ds
      robust <- cfg$robust
      
      # Start value and increment for robust thresholds
      inc.r <- -1 
      s.r1   <- -0.5*exp(inc.r)*(k1-1)
      s.r2   <- -0.5*exp(inc.r)*(k2-1)
      
      # Start and end value for non-robust thresholds
      s.n <- -1
      s.e <-  1

      if(is.null(ds))         ds <- 0
      if(is.null(robust)) robust <- F
      
	# Specify Number of thresholds
	if(ds == 0)
      { 
            if(robust && k1 > 1)    T1 <- c(rep(c(0, 0, 1, 1, 0), s), c(s.r1, rep(inc.r, k1-1)))
            else                    T1 <- c(rep(c(0, 0, 1, 1, 0), s), seq(s.n, s.e, length = k1))
                  
            if(robust && k2 > 2)    c(T1, c(s.r2, rep(inc.r, k2-1)))
            else                    c(T1, seq(s.n, s.e, length = k2))
	}
      else if(ds == 1)
      {
            if(robust && k1 > 1)    T1 <- c(rep(c(0, 0, 1, 1, 0), s), rep(c(s.r1, rep(inc.r, k1-1)), k2+1))
            else                    T1 <- c(rep(c(0, 0, 1, 1, 0), s), rep(seq(s.n, s.e, length = k1), k2+1)) 
            
            
            if(robust && k2 > 2)    c(T1, c(s.r2, rep(inc.r, k2-1)))
            else                    c(T1, seq(s.n, s.e, length = k2))
      }
	else if(ds == 2)
      {
            if(robust && k1 > 1)    T1 <- c(rep(c(0, 0, 1, 1, 0), s), c(s.r1, rep(inc.r, k1-1)))
            else                    T1 <- c(rep(c(0, 0, 1, 1, 0), s), seq(s.n, s.e, length = k1))
            
            if(robust && k2 > 2)    c(T1, rep(c(s.r2, rep(inc.r, k2-1)), k1+1))
            else                    c(T1, rep(seq(s.n, s.e, length = k2), k1+1))
      }
      else  stop ("\n FATAL ERROR: Wrong value of flag ds \n")
}

# ------------------------------------------------------------------------------------------------------------------
# Generate upper and lower bounds of parameter
# Input:	parvec = Vector of parameters
#		   cfg = configuration structure
#
# Output: List with lower and upper bounds on parameter vector
# ------------------------------------------------------------------------------------------------------------------
SDT2D.par.bounds <- function(parvec, cfg)
{
      n.sdt <- cfg$n.sdt
      min.r <- -1.0 + SDT2D.Tol
      max.r <-  1.0 - SDT2D.Tol
      
      lower <- rep(c(-Inf, -Inf, SDT2D.Tol, SDT2D.Tol, min.r), n.sdt)
      upper <- rep(c( Inf,  Inf,       Inf,       Inf, max.r), n.sdt)

      n.res <- length(parvec) - length(lower)
      lower <- c(lower, rep(-Inf, n.res))
      upper <- c(upper, rep( Inf, n.res))
      
      return(list(lower = lower, upper = upper))
}      


# ------------------------------------------------------------------------------------------------------------------
# Generate strings with names of the parameters for the models
# Input:	cfg = list with configuration information
#
# Output:   String with parameter names
# ------------------------------------------------------------------------------------------------------------------
SDT2D.par.names <- function(cfg)
{
      s  <- cfg$n.sdt
      k1 <- cfg$k1
      k2 <- cfg$k2
      ds <- cfg$ds

     if(is.null(ds)) ds <- 0

	m1   <- paste("Mean.1[" , 1:s, "]", sep = "")
	sd1  <- paste("Stddev.1[", 1:s, "]", sep = "")
	m2   <- paste("Mean.2[" , 1:s, "]", sep = "")
	sd2  <- paste("Stddev.2[", 1:s, "]", sep = "")
	r    <- paste("Corr[", 1:s, "]", sep = "") 
	
	if(ds == 0)
	{
	     t1   <- paste("t.D1-",  1:k1, sep = "")
	     t2   <- paste("t.D2-",  1:k2, sep = "")
      }
      else if(ds == 1)
      {
	     t1   <- paste("t.D1-",  1:k1, sep = "")
	     t1   <- paste(t1, "[D2 = ", rep(1:(k2+1), each = k1), "]", sep = "")
	     t2   <- paste("t.D2-",  1:k2, sep = "")
            
      }
      else if(ds == 2)
      {
	     t1   <- paste("t.D1-",  1:k1, sep = "")
	     t2   <- paste("t.D2-",  1:k2, sep = "")
	     t2   <- paste(t2, "[D1 = ", rep(1:(k1+1), each = k2), "]", sep = "")
      }
      else  stop ("\n FATAL ERROR: Wrong value of flag ds \n")
      
	pname  <- c(m1, m2, sd1, sd2, r)
	index <- matrix(1:(5*s), nr = 5, byrow = T)
	c(pname[as.vector(index)], t1, t2)
}

# ------------------------------------------------------------------------------------------------------------------
# Names of the SDT2D-model
# ------------------------------------------------------------------------------------------------------------------
SDT2D.model.name	<- function(cfg)
{
      if(is.null(cfg$ds))     name <- "Bivariate Gaussian SDT model: No violations of decisional separability"
      else if(cfg$ds == 0)    name <- "Bivariate Gaussian SDT model: No violations of decisional separability"
      else if(cfg$ds == 1)    name <- "Bivariate Gaussian SDT model: Violation of decisional separability on Dimension 1"
      else if(cfg$ds == 2)    name <- "Bivariate Gaussian SDT model: Violation of decisional separability on Dimension 2"
      else name <- "Bivariate Gaussian SDT model: No information about violations of decisional separability"
      
      if(!is.null(cfg$restriction))
            name <- paste(name, " [RESTRICTIONS: ", cfg$restriction, "]", sep = "")
      if(!is.null(cfg$stuct.zero))
            name <- paste(name, "[STRUCTURAL ZEROS]")
      if(!is.null(cfg$R))
            name <- paste(name, "[WITH POOLING MATRIX]")
      if(!is.null(cfg$robust)) 
            if(cfg$robust == T) name <- paste(name, "[ROBUST]")
      if(!is.null(cfg$conditional)) 
            if(cfg$conditional == T) name <- paste(name, "[CONDITIONAL]")
      name
      
}

#===============================================================================
#
# V. Functions for Checking the setup of the SDT2D model
#
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Checking the list with the configuration
# Input:	List of configuration containing the following entries:
#           1.           s = Number of signal distributions
#           2.          k1 = Number of decision bounds on Dimension 1
#           3.          k2 = Number of decision bounds on Dimension 2
#           4.          ds = Flag indicating the presence of violations of descision separability
#                                   ds = 0 --> No violation
#                                   ds = 1 --> Violation on Dimension 1
#                                   ds = 2 --> Violation on Dimension 2
#           5.          PA = Response assignment matrix: This matrix is premaultiplied with response vector
#           6. struct.zero = Vector of structural zeros  
#
# In case of an erroneous entry a message is printed and the program halts.
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT2D.Check.Fill.Cfg <- function(cfg)
{
      # 1. Checks

      if(is.null(cfg))
      {
            msg <- "\n\nNo list with configuration data\n\n"
            stop(msg)
      }
      if(is.null(cfg$n.sdt))
      {
            if(!is.null(cfg$s))
                  cfg$n.sdt <- cfg$s
            else
            {
                  msg <- "\n\nNumber of models (n.sdt) is missing\n\n"
                  stop(msg)
            }
      }
      if (cfg$n.sdt <= 0)
      {
            msg <- "\n\nNumber of models (n.sdt) <= 0\n\n"
            stop(msg)
      }      
      if (is.null(cfg$k1))
      {
            msg <- "\n\nNumber of decision bounds on Dimension 1 (k1) is missing\n\n"
            stop(msg)
      }
      if (cfg$k1 <= 0)
      {
            msg <- "\n\nNumber of decision bounds on Dimension 1 (k1) <= 0\n\n"
            stop(msg)
      }      
      if (is.null(cfg$k2))
      {
            msg <- "\n\nNumber of decision bounds on Dimension 1 (k2) is missing\n\n"
            stop(msg)
      }
      if (cfg$k2 <= 0)
      {
            msg <- "\n\nNumber of decision bounds on Dimension 1 (k2) <= 0\n\n"
            stop(msg)
      }      
      if (!is.null(cfg$R))
      {
            if(dim(cfg$R)[2] != cfg$n.sdt * (cfg$k1 + 1) * (cfg$k2 + 1))
            {
                  msg <- paste("\n\nWrong number of columns of matrix R:", dim(cfg$R)[2], "( Expected:", cfg$n.sdt * (cfg$k1 + 1) * (cfg$k2 + 1), ")\n\n")
                  stop(msg)
            }
      }
      if(!is.null(cfg$ds))
      {
            if(cfg$ds != 0 && cfg$ds != 1 && cfg$ds != 2)
            {
                  msg <- "\n\nFlag indicating violations of decisional separability out of range (Expected: 0, 1, or 2)\n\n"
                  stop(msg)
            }
      }
      # 2. Create new structure
      cfg.new <- list(n.sdt = cfg$n.sdt, k1 = cfg$k1, k2 = cfg$k2)
      if(!is.null(cfg$struct.zero)) cfg.new$struct.zero <- cfg$struct.zero
      if(is.null(cfg$ds))           cfg.new$ds <- 0
      else                          cfg.new$ds <- cfg$ds
      if(!is.null(cfg$robust))      cfg.new$robust <- cfg$robust
      if(!is.null(cfg$conditional)) cfg.new$conditional <- cfg$conditional
      if(!is.null(cfg$R))           cfg.new$R  <- cfg$R
      cfg.new$restriction <- SDT2D.Check.restriction(cfg)
      return(cfg.new)

}

# ------------------------------------------------------------------------------------------------------------------
# SDT2D Model: Check restriction setting in configuration list
# ------------------------------------------------------------------------------------------------------------------
SDT2D.Check.restriction <- function(cfg)
{
      if(is.null(cfg$restriction))  return("standard")
      restriction <- cfg$restriction
      if(charmatch(restriction, c("no", "No", "NO"), nomatch = 0) > 0)
            return("no")
      if(grepl("st", restriction, ignore.case = T))   return(restriction)
      if(grepl("v1", restriction, ignore.case = T))   return(restriction)
      if(grepl("r0", restriction, ignore.case = T))   return(restriction)
      if(grepl("er", restriction, ignore.case = T))   return(restriction)
      if(grepl("qu", restriction, ignore.case = T))   return(restriction)
      warn.msg <- paste("\n\n SDT2D.Check.restriction: Unkonwn restriction in configuration list: <", restriction, "> --> setting to: <standard>\n\n", sep = "")      
      warning(warn.msg)
      return("standard")
}      

# ------------------------------------------------------------------------------------------------------------------
# SDT2D Model: Check number of parameters and data points
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT2D.check <- function(par, data, cfg)
{
      s  <- cfg$n.sdt
      k1 <- cfg$k1
      k2 <- cfg$k2
      ds <- cfg$ds

      exp.d.len         <- s * (k1+1)*(k2+1)
      exp.par.len       <- s * 5
      
      if(ds == 0)       exp.par.len <- exp.par.len + k1 + k2
      else if(ds == 1)  exp.par.len <- exp.par.len + k1*(k2+1) + k2
      else              exp.par.len <- exp.par.len + k1 + k2*(k1+1)
      
      Msg <- "NOTE: Computation of parameter and data length based on given configuration"
	Erg <- matrix(c(length(par), length(data), exp.par.len, exp.d.len), nc = 2)
	rownames(Erg) <- c("Parameters", "Data points")
	colnames(Erg) <- c("Found", "Expected")
	
	Error <- any(Erg[, 1] != Erg[, 2])
	list(Error = Error, Stat = Erg, Msg = Msg)
}


# ------------------------------------------------------------------------------------------------------------------
# MIX Ranking and Rating Model: Set restrictions
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT2D.Restriction <- function(cfg, fixed, ident, functional)
{
      # If configuration information is not complete, then complete it
      cfg <- SDT.SDT2D.Check.Fill.Cfg(cfg)  

      if(cfg$restriction == "standard")
            return(SDT2D.Standard.Restriction(cfg, fixed, ident, functional))
            
      fixed.add <- NULL
      ident.add <- NULL            
      if(grepl("st", cfg$restriction, ignore.case = T))
      {
            RS        <- SDT2D.Standard.Restriction(cfg, fixed, ident, functional)      
            fixed.add <- cbind(fixed.add, RS$fixed)
            ident.add <- cbind(ident.add, RS$ident)
      }
      if(grepl("v1", cfg$restriction, ignore.case = T))     # All variance parameters are equal to 1
      {
            RS        <- SDT2D.v1(cfg)      
            fixed.add <- cbind(fixed.add, RS$fixed)
            ident.add <- cbind(ident.add, RS$ident)
      }
      if(grepl("r0", cfg$restriction, ignore.case = T))     # All correlations equal to zero
      {
            RS        <- SDT2D.r0(cfg)      
            fixed.add <- cbind(fixed.add, RS$fixed)
            ident.add <- cbind(ident.add, RS$ident)
      }
      if(grepl("er", cfg$restriction, ignore.case = T))     # All correlations equal 
      {
            if(!grepl("r0", cfg$restriction, ignore.case = T))    # Perform only if not all correlation equal to zero
            {
                  RS        <- SDT2D.er(cfg)      
                  fixed.add <- cbind(fixed.add, RS$fixed)
                  ident.add <- cbind(ident.add, RS$ident)
            }
      }
      if(grepl("qu", cfg$restriction, ignore.case = T))     # Quadratic restriction
      {
            if(cfg$n.sdt != 4)
                  warning(paste("\n SDT.SDT2D.Restriction: Rectangular restriction only possible with 4 models; number of models is <", cfg$n.sdt, ">\n  Restriction ignored\n", sep = ""))
            else
            {
                  RS        <- SDT2D.qu()      
                  fixed.add <- cbind(fixed.add, RS$fixed)
                  ident.add <- cbind(ident.add, RS$ident)
            }
      }
       if(grepl("quvd", cfg$restriction, ignore.case = T))     # Quadratic restriction
      {
            if(cfg$n.sdt != 4)
                  warning(paste("\n SDT.SDT2D.Restriction: Rectangular restriction only possible with 4 models; number of models is <", cfg$n.sdt, ">\n  Restriction ignored\n", sep = ""))
            else
            {
                  RS        <- SDT2D.quvd()      
                  fixed.add <- cbind(fixed.add, RS$fixed)
                  ident.add <- cbind(ident.add, RS$ident)
            }
      }
      # Only unique constraints
      if(!is.null(fixed.add))
      {
           pos.dup    <- duplicated(fixed.add[2,])
           fixed.add  <- fixed.add[, !pos.dup, drop = F]      

	     pos.unique <-  !(fixed.add[2,] %in% fixed[2,])
	     fixed      <- cbind(fixed.add[, pos.unique, drop = F], fixed)            # Add to fixed if not alread specified
      }
      if(!is.null(ident.add))
      {
           pos.dup    <- duplicated(ident.add[2,])
           ident.add  <- ident.add[, !pos.dup, drop = F]      

	     pos.unique <-  !(ident.add[2,] %in% ident[2,])
	     ident      <- cbind(ident.add[, pos.unique, drop = F], ident)            # Add to identity if not alread specified
      }
      return(list(fixed = fixed, ident = ident, functional = functional))
}

# ------------------------------------------------------------------------------------------------------------------
# Ranking and Rating Model with Recollection and Guessing: Set standard restrictions
# ------------------------------------------------------------------------------------------------------------------
SDT2D.Standard.Restriction <- function(cfg, fixed, ident, functional)
{
      fixed.s <- matrix(c(0, 0, 1, 1, 1:4), nr = 2, byrow = T)
	pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
	fixed <- cbind(fixed.s[, pos.unique, drop = F], fixed)            # Add to fixed if not alread specified

      list(fixed = fixed, ident = ident, functional = functional)
}

# ------------------------------------------------------------------------------------------------------------------
# Variance = 1 restriction:
# Input: configuration information
#        start value: 0: parameters of the first model are included
#                     1: parameters of the first model are excluded  
# ------------------------------------------------------------------------------------------------------------------
SDT2D.v1 <- function(cfg)
{
      pos.var <- c(3 + 5*(0:(cfg$n.sdt-1)), 4 + 5*(0:(cfg$n.sdt-1)))
      fixed   <- matrix(c(rep(1, length(pos.var)), pos.var), nr = 2, byrow = T)

      list(fixed = fixed, ident = NULL)
}

# ------------------------------------------------------------------------------------------------------------------
# All r = 0 restriction:
# Input: configuration information
# ------------------------------------------------------------------------------------------------------------------
SDT2D.r0 <- function(cfg)
{
      pos.r <- c(5*(1:cfg$n.sdt))
      fixed <- matrix(c(rep(0, length(pos.r)), pos.r), nr = 2, byrow = T)

      list(fixed = fixed, ident = NULL)
}

# ------------------------------------------------------------------------------------------------------------------
# Equal r restriction:
# Input: configuration information
# ------------------------------------------------------------------------------------------------------------------
SDT2D.er <- function(cfg)
{
      if(cfg$n.sdt < 2) return(list(fixed = NULL, ident = NULL))
      pos.r <- c(5*(2:cfg$n.sdt))
      ident   <- matrix(c(rep(5, length(pos.r)), pos.r), nr = 2, byrow = T)

      list(fixed = NULL, ident = ident)
}


# ------------------------------------------------------------------------------------------------------------------
# Standard-rectangle-restriction
# ------------------------------------------------------------------------------------------------------------------
SDT2D.qu <- function()
{
      fixed   <- NULL
      ident   <- matrix(c(1,  2,  7, 11, 
                          6, 12, 17, 16), nr = 2, byrow = T)

      list(fixed = fixed, ident = ident)
}

# ------------------------------------------------------------------------------------------------------------------
# Standard-rectangle-restriction and dimension wise restriction of variances
# ------------------------------------------------------------------------------------------------------------------
SDT2D.quvd <- function()
{
      fixed     <- NULL
      ident.m   <- matrix(c(1,  2,  7, 11,                          # Means
                            6, 12, 17, 16), nr = 2, byrow = T)
      ident.v   <- matrix(c(3,  4, 13,  9,                          # Variances per dimension
                            8, 14, 18, 19), nr = 2, byrow = T)
      
      list(fixed = fixed, ident = cbind(ident.m, ident.v))
}




#===============================================================================
#
# V. Formating functions of the SDT2D model
#
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Format full parameter vector for SDT and Gaussian model
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Par.SDT2D <- function(full.par, n, deci)
{
      T.max <- 10										# Max number of bounds for displaying in a row 

      s  <- n$n.sdt
      ds <- n$ds
      k1 <- n$k1
      k2 <- n$k2
      n.par <- 5
      
      # Compute robust thresholds if required
      full.par <- SDT2D.robust.thresholds(full.par, n)
	           
	Gauss.Par <- round(matrix(full.par[1:(s*n.par)], nrow = n.par), deci)
	colnames(Gauss.Par) <- paste("Gauss-", 1:dim(Gauss.Par)[2], sep = "")
	rownames(Gauss.Par) <- c("Mean.1", "Mean.2", "Stddev.1", "Stddev.2", "Corr")
	
	if(ds == 1)
	{
		n1 <- k1*(k2+1)
		bounds.1 		 <- round(matrix(full.par[(s*n.par+1):(s*n.par+n1)], ncol = k1, byrow = T), deci)
		rownames(bounds.1) <- paste("D2: t-", 1:(k2+1), sep = "")
		colnames(bounds.1) <- paste("D1: t-", 1:dim(bounds.1)[2], sep = "")

	}
	else
	{
		bounds.1 	       <- round(matrix(full.par[(s*n.par+1):(s*n.par+k1)], nrow = 1, byrow = T), deci)
		colnames(bounds.1) <- paste("t-", 1:k1, sep = "")
		rownames(bounds.1) <- "Dimension 1"
		if(dim(bounds.1)[2] > T.max) bounds.1 <- t(bounds.1)		# Too many bounds -> form row vector

	}
	if(ds == 2)
	{
		k2 <- (k1+1)*k2
		bounds.2 		 <- round(matrix(full.par[(s*n.par+k1+1):(s*n.par+k1+k2)], ncol = k1+1), deci)
		colnames(bounds.2) <- paste("D1: t-", 1:(k1+1), sep = "")
		rownames(bounds.2) <- paste("D2: t-", 1:dim(bounds.2)[1], sep = "")

	}
	else
	{
	      # Compute beginning
	      if(ds == 1) n1 <- k1*(k2+1)
	      else        n1 <- k1
		bounds.2 	       <- round(matrix(full.par[(s*n.par+n1+1):(s*n.par+n1+k2)], nrow = 1, byrow = T), deci)
		colnames(bounds.2) <- paste("t-", 1:k2, sep = "")
		rownames(bounds.2) <- "Dimension 2"
		if(dim(bounds.2)[2] > T.max) bounds.2 <- t(bounds.2)		# Too many bounds -> form row vector
	}
		
      list(GP = Gauss.Par, DB.1 = bounds.1, DB.2 = bounds.2)

}

# ------------------------------------------------------------------------------------------------------------------
# Format free parameters and compute confidence intervals
# Input:	1. px2 Matrix with parameters and standard errors
#           3. Full parameter vector
#		2. Number of signals
#           3. Fixed constraints
#           4. Equality constraints
#		5. Confidence level
#           6. Number of decimal places
#
# Output: px5 Matrix of parameters, standard errors and confidence limits
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Par.est.SDT2D <- function(Pars, full.par, n, fixed, ident, alpha, deci)
{
	if(!is.null(n$robust))
	{     # Compute thresholds in case of robust thresholds
            if(n$robust) 
            {
                  full.par <- SDT2D.robust.thresholds(full.par, n)
                  Pars[,1] <- Elim.par(full.par, fixed, ident)
            }      
      }	           
      s <- n$n.sdt
      plen <- length(full.par)
      
      if(any(is.na(Pars[, 2]))) return(Pars)     # In case on NA no computation
      
      qt.cfi <- qnorm((1-alpha)/2)
      cfi.lo <- Pars[,1] + Pars[,2]*qt.cfi
      cfi.up <- Pars[,1] - Pars[,2]*qt.cfi
      
      # Respect lower and upper bounds on parameters
      lower <- c(rep(c(-Inf, -Inf,   0,   0, -1), s), rep(-Inf, plen - 5*s))
      upper <- c(rep(c( Inf,  Inf, Inf, Inf,  1), s), rep( Inf, plen - 5*s))
      
      lower <- Elim.par(lower, fixed, ident)
      upper <- Elim.par(upper, fixed, ident)

      cfi.lo[cfi.lo < lower] <- lower[cfi.lo < lower]
      cfi.up[cfi.up > upper] <- upper[cfi.up > upper] 
            
      Pars <- cbind(Pars, cfi.lo, cfi.up)
	colnames(Pars) <- c("Value", "SE", paste("CFI-", alpha*100, "(Lower)", sep = ""), paste("CFI-", alpha*100, "(Upper)", sep = ""))
	
      round(Pars, deci)
}

# ------------------------------------------------------------------------------------------------------------------
# Observed, estimated etc. frequencies to indicate structural zeros
# Input:	1. px5 Observed and estiamted Scores with residuals
#           2. Configuration information
#
# Output: px5 Matrix of parameters, standard errors and confidence limits
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Freq.SDT2D <- function(Freq, cfg)
{
      if(is.null(cfg$R))
      {
            # Compute new row labels
            s  <- 1:cfg$n.sdt
            n1 <- 1:(cfg$k1+1)
            n2 <- 1:(cfg$k2+1)
            Nr <- expand.grid(n2, n1, s)
            rownames(Freq) <- paste("Data[", Nr[, 3], ", ", Nr[, 2], ", ", Nr[, 1], "]", sep = "")
      }

      if(is.null(cfg$struct.zero)) return (Freq)
      
      F1 <- as.character(Freq[,1])
      F1[cfg$struct.zero]  <- paste(F1[cfg$struct.zero],  "***", sep = "")
      F1[-cfg$struct.zero] <- paste(F1[-cfg$struct.zero], "   ", sep = "")
      

      Freq.neu <- data.frame(F1, Freq[,-1])
      colnames(Freq.neu) <- colnames(Freq)
      rownames(Freq.neu) <- rownames(Freq)

      Freq.neu
}

#===============================================================================
#
# VI. Functions for computing teststatistics
#
#===============================================================================
# ------------------------------------------------------------------------------------------------------------------
# Computation of estimated frequencies taking sturctural zeros and regrouping into account
# Input:    1. Parameter vector
#           1. Vector of estimated probabiltiies
#           2. Data vector
#           3. Configuation list
# ------------------------------------------------------------------------------------------------------------------
SDT2D.Compute.Est.N <- function(est.p, data, cfg)
{

      # Neither structural zeros nor regrouping --> Normal computation  (File: SDT-SDT-2D.R)
      if(is.null(cfg$struct.zero)) return (Compute.Est.N(est.p, data))
            
      dat   <- matrix(data,  nc = cfg$n.sdt)
      est   <- matrix(est.p, nc = cfg$n.sdt)
      N     <- colSums(dat)
      erg   <- t(apply(est, 1, function(x) { x*N }))
      as.vector(erg)
}      
  
# ------------------------------------------------------------------------------------------------------------------
# Computation of degrees of freedom of the data
# Input:    1. vector of estimated probabiltiies
#           2. data vector
#           3. configuation list
# ------------------------------------------------------------------------------------------------------------------
SDT2D.Compute.df.data <- function(data, cfg)  
{
      if(is.null(cfg$R))   length(data)  - cfg$n.sdt - length(cfg$struct.zero)
      else                 dim(cfg$R)[1] - cfg$n.sdt - length(cfg$struct.zero)
}

#===============================================================================
#
# VII. Functions for computing the model matrix
#
#===============================================================================

# ------------------------------------------------------------------------------
# Model incorporating restrictions
# The parameter passed to the function contains the free parameters only
# ------------------------------------------------------------------------------
SDT.SDT2D.Full <- function(rest.par, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
      full.par <- Restore.par(rest.par, pos, fixed, ident, functional)	# Restore full parameter vector
      SDT.SDT2D(full.par, n, fixed, ident)
}
 
# ------------------------------------------------------------------------------
# Numerical computation of the model matrix 
# ------------------------------------------------------------------------------
SDT.SDT2D.Model.Matrix.Num <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)
      par <- Elim.par(full.par, fixed, ident)
      J <- jacobian(SDT.SDT2D.Full, par, n = cfg, pos = pos, fixed = fixed, ident = ident, functional = functional)
}

# ------------------------------------------------------------------------------
# Analytical computation of the model matrix
# Input:    1. Full parameter vector
#           2. Number of signals
#           3. 2 x p - matrix of fixed constraints
#           4. 2 x q - matrix of identity constraints
#           5. Jacobian of functional constraints (supplied by the user)
#
# Output:   Modelmatrix
#
# Comment:  The Modelmatrix of funtional constraints must be a full matrix that
#           is premultiplied by the Jacobian
# ------------------------------------------------------------------------------
SDT.SDT2D.Model.Matrix <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)

      J   <- SDT.SDT2D.Jacobi(full.par, cfg = cfg)
      if(!is.null(ident))
      {
            J.Id <- diag(length(full.par))
            Index <- matrix(c(ident[2,], ident[1,]), nc = 2)
            J.Id[Index] <- 1
            J <- J %*% J.Id
      }
      
      # Treatment of funtional constraints
      if(!is.null(functional))
      {
            J.fct <- attr(functional, "SDT.Jacobian")
            if(is.null(J.fct)) 
                  return(NULL)           # No Jacobian function supplied -> return (NULL)
            J <- J %*% J.fct(full.par)
      }
      J[, pos]
}

#===============================================================================
#
# VII. Plot function
#
#===============================================================================


# -------------------------------------------------------------------------------------------------------------
# Drawing the Gaussian Konfiguration of the SDT2D model
# Input:	1. par	 = the whole parameters vector
#		2. k		 = number of Gaussian models that enter a mixture model
#		4. n1  	 = number of decision criteria on Dimension 1
#		5. n2   	 = number of decision criteria on Dimension 2
#		6. ds    	 = Flag indicting violations of decisional spearability
#					0: No violations
#					1: Violations on D1
#					2: Violations on D2
#		7. cols	 = vector with color names or numbers for ellipses
#		8. ltys	 = vector with line types  for ellipses
#		9. labels    = character vector with names of labels for D1 and D2
#           10.option    = list of the number decision bounds to be left out
#
# Output: No sensible output
# -------------------------------------------------------------------------------------------------------------
SDT.SDT2D.Plot <- function(par, k, n1, n2, ds, cols = NULL, ltys = NULL, labels = NULL, option)
{
	n.par <- 5
	min.offset <- max.offset <- 0.0
	y.offset <- x.offset <- 5

	# Printing parameters
	col.c	<- "grey80"														# colors of decision bounds
	col.e <- c("black", "grey70", "blue", "red", "cyan", "darkgreen", "magenta", "yellow")		# colors of confidence ellipses
	lwd.c	<- 2															# line width of decision bounds
	lwd.e <- 2															# line width of confidence ellipse
	cex.e <- 1.5														# size of center of the ellipses
	sym.e <- 16															# symbol used for center of ellipses
		
	# If no color vector has been specified, create default
	if(is.null(cols))	cols <- rep(1, k)							

	# If no vector has line types been specified, create default vector
	if(is.null(ltys)) ltys <- rep(1, k)
	

	# If no labels have been specified, create default
	if(is.null(labels)) labels <- c("Dimension 1", "Dimension 2")

	# Generate a list with coordinates of Ellipses
	E <- NULL
	for(i in 1:k)
	{
		E[[i]] <- ellipse(par[5+(i-1)*n.par],  centre = par[1:2 +(i-1)*n.par],     scale = par[3:4 +(i-1)*n.par]  , level = .90)
		if(i == 1)
		{
			min.x <- min(E[[i]][,1])
			min.y <- min(E[[i]][,2])
			
			max.x <- max(E[[i]][,1])
			max.y <- max(E[[i]][,2])
		}
		else
		{
			min.x <- min(min.x, E[[i]][,1])
			min.y <- min(min.y, E[[i]][,2])

			max.x <- max(max.x, E[[i]][,1])
			max.y <- max(max.y, E[[i]][,2])
		}

	}
	# Generate empty plot
	min.x <- min.x
	min.y <- min.y
	max.x <- max.x
	max.y <- max.y

	# Identical scale on both Axes
	max.co <- max(max.x, max.y)
	min.co <- min(min.x, min.y)

	max.co <- max.co + max.offset
	min.co <- min.co - min.offset
	
	plot(c(min.co, max.co), c(min.co, max.co), type = "n", xlab = labels[1], ylab = labels[2])

      c.x <- c.y <- NULL
	# Plot decision criteria
	if(ds == 0)
	{
		c1 <- par[(n.par*k+1):(n.par*k+n1)]
		c2 <- par[(n.par*k+n1+1):(n.par*k+n1+n2)]
		
		# Compute coordinates for decision bounds on D1
		# The Result is a (n1 * (n2+1)) x 4 matrix (x1, y1, x2, y2) with start and end coordinates
		
		c.x <- cbind(c1, rep(min.y - y.offset, n1), c1, rep(c2[1], n1)) 
		if(n2 > 1)
		{
                  for(i in 2:n2)
		    {
		          actu <- cbind(c1, c2[i-1], c1, c2[i])
		          c.x <- rbind(c.x, actu)    
                  }
            }      
            c.x <- rbind(c.x, cbind(c1, rep(c2[n2], n1), c1, rep(max.y + y.offset, n1)))

		# Now the same for D2
		# The Result is a ((n1+1) * n2) x 4 matrix (x1, y1, x2, y2) with start and end coordinates
 		c.y <- cbind(c2, rep(min.x - x.offset, n2), c2, rep(c1[1], n2)) 
  		if(n1 > 1)
		{
                  for(i in 2:n1)
		    {
		          actu <- cbind(c2, c1[i-1], c2, c1[i])
		          c.y <- rbind(c.y, actu)    
                  }
            }      
            c.y <- rbind(c.y, cbind(c2, rep(c1[n1], n2), c2, rep(max.x + x.offset, n2)))
	}
	else if(ds == 1)
	{
		# Extract thresholds
		c1 <- par[(n.par*k+1):(n.par*k + n1*(n2+1))]
		c2 <- par[(n.par*k + n1*(n2+1) + 1):(n.par*k + n1*(n2+1) + n2)]

		# Compute coordinates for decision bounds on D1
		# The Result is a (n1 * (n2+1)) x 4 matrix (x1, y1, x2, y2) with start and end coordinates
		
		c.x <- cbind(c1[1:n1], rep(min.y - y.offset, n1), c1[1:n1], rep(c2[1], n1))
            dn <- n1 + 1 
		if(n2 > 1)
		{
                  for(i in 2:n2)
		    {
		          actu <- cbind(c1[dn:(dn+n1-1)], c2[i-1], c1[dn:(dn+n1-1)], c2[i])
		          c.x <- rbind(c.x, actu)    
		          dn <- dn + n1
                  }
            }      
            c.x <- rbind(c.x, cbind(c1[dn:(dn+n1-1)], rep(c2[n2], n1), c1[dn:(dn+n1-1)], rep(max.y + y.offset, n1)))

		# Now the same for D2
		# The Result is a ((n1+1) * n2) x 4 matrix (x1, y1, x2, y2) with start and end coordinates
 		c.y <- cbind(c2, rep(min.x - x.offset, n2), c2, rep(c1[1], n2)) 
  		if(n1 > 1)
		{
                  for(i in 2:n1)
		    {
		          actu <- cbind(c2, c1[i-1], c2, c1[i])
		          c.y <- rbind(c.y, actu)    
                  }
            }      
            c.y <- rbind(c.y, cbind(c2, rep(c1[n1], n2), c2, rep(max.x + x.offset, n2)))
	}
	else # ds = 2
	{
		c1 <- par[(n.par*k+1):(n.par*k + n1)]
		c2 <- par[(n.par*k + n1 + 1):(n.par*k + n1 + n2*(n1+1))]

		# Compute coordinates for decision bounds on D1
		# The Result is a (n1 * (n2+1)) x 4 matrix (x1, y1, x2, y2) with start and end coordinates
		
		c.x <- cbind(c1, rep(min.y - y.offset, n1), c1, rep(c2[1], n1)) 
		if(n2 > 1)
		{
                  for(i in 2:n2)
		    {
		          actu <- cbind(c1, c2[i-1], c1, c2[i])
		          c.x <- rbind(c.x, actu)    
                  }
            }      
            c.x <- rbind(c.x, cbind(c1, rep(c2[n2], n1), c1, rep(max.y + y.offset, n1)))


		# Now the same for D2
		# The Result is a ((n1+1) * n2) x 4 matrix (x1, y1, x2, y2) with start and end coordinates
		c.y <- cbind(c2[1:n2], rep(min.x - x.offset, n2), c2[1:n2], rep(c1[1], n2))
            dn <- n2 + 1 
		if(n1 > 1)
		{
                  for(i in 2:n1)
		    {
		          actu <- cbind(c2[dn:(dn+n2-1)], c1[i-1], c2[dn:(dn+n2-1)], c1[i])
		          c.y <- rbind(c.y, actu)    
		          dn <- dn + n2
                  }
            }      
            c.y <- rbind(c.y, cbind(c2[dn:(dn+n2-1)], rep(c1[n1], n2), c2[dn:(dn+n2-1)], rep(max.x + x.offset, n2)))
	}
      # Draw decision bounds
      ac <- 1
      for(i in 1:dim(c.x)[1])
      {
            if((ac %in% option) == F)
                  lines(c.x[i, c(1, 3)], c.x[i, c(2, 4)], col = col.c, lwd = lwd.c)
            ac <- ac + 1
      }      
      for(i in 1:dim(c.y)[1])
      {
            if((ac %in% option) == F)
                  lines(c.y[i, c(2, 4)], c.y[i, c(1, 3)], col = col.c, lwd = lwd.c)
            ac <- ac + 1
      }      

	# Draw centers of ellipes and ellipses
	for(i in 1:k)
	{
		points(par[1+(i-1)*n.par],  par[2+(i-1)*n.par],  pch = sym.e, col = col.e[cols[i]], cex = cex.e)	# Draw center
		lines(E[[i]], col = col.e[cols[i]], lty = ltys[i], lwd = lwd.e)
	}
} 