# ===========================================================================
# SDT models: Standard Signal detection models
# Date of creation:	July,      2009
# Date of last update:	November,  2013
#
# Author: Siegfried Macho
# ===========================================================================

SDT.Tol <- 1e-12

#===============================================================================
#
# 1. Model functions: Computation of model probabilities
#
#===============================================================================

# ----------------------------------------------------------------------
# Function for creating response distribution for on Type of Stimulus
# Input:	Vector of parameters: (mean, sigma, c[1], c[2], ...., c[n-1])
# Output: 	n - Vector, with n being the number of Response categories
# ----------------------------------------------------------------------
SDT.Probs <- function(par)
{
	 Resp <- pnorm(par[3:length(par)], mean = par[1], sd = max(par[2], SDT.Tol)) 	# Probability of Response-Category 2 to n-1
	 Resp <- c(Resp, 1)										      # Response vector
	 c(Resp[1], Resp[-1]-Resp[-length(Resp)])							      # Compute Differences
}				

# ------------------------------------------------------------------------------------------------------------------
# Computation or probabilities of SDT model
# Input:	par 	= Vector of parameters, consisting of the following entries
#			1. mean for each model (except noise model)
#			2. standard deviation for each model (except for noise model)
#			3. thresholds (on set for each model)
#		cfg  	= configuration list
#           fixed = 2xp matrix of fixed    constraints
#           ident = 2xq matrix of equality constraints
#
# Output: Vector of probabilities for single models
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT <- function(par, cfg, fixed = NULL, ident = NULL)
{
      n <- cfg$n.sdt
	# Separate parameters and thresholds for n models
	m <- n-1								
	Par 		<- matrix(par[1:(2*m)], nr = m, byrow = T)
	thresholds 	<- par[(2*m + 1):length(par)]
	thresholds 	<- matrix(rep(thresholds, n), nr = n, byrow = T)	# Matrix with thresholds for each model

	# Add parameters for Noise distribution
	Par <- rbind(c(0, 1), Par)
	Par <- cbind(Par, thresholds)

	# Compute probabilities
	as.vector(apply(Par, 1, SDT.Probs))
}

# ------------------------------------------------------------------------------------------------------------------
# Model incorporating restrictions
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT.Full <- function(par, cfg, pos, fixed = NULL, ident = NULL, functional = NULL)
{
      par <- Restore.par(par, pos, fixed, ident, functional)	# Restore full parameter vector
      SDT.SDT(par, cfg)

}

# ------------------------------------------------------------------------------------------------------------------
# Computation of probabilities for Target and Related items for the Conjoint Recognition Model of Stahl & Klauer (2008)
# Input:	par = Vector of parameters, consisting of the following entries
#			1. mean for each model (except noise model)
#			2. standard deviation for each model (except for noise model)
#			3. thresholds (for each model)
#		cfg   = List with configuration information
#           fixed = 2xp matrix of fixed    constraints
#           ident = 2xq matrix of equality constraints
#
# Output: Vector of probabilities for single models
# ------------------------------------------------------------------------------------------------------------------
SDT.Gaussian <- function(par, cfg, fixed = NULL, ident = NULL)
{
      n <- cfg$n.sdt
	# Compute probabilities: Case n is not a vector
	if(length(n) < 2) return(as.vector(apply(matrix(par, nr = n, byrow = T), 1, SDT.Probs)))

	# Compute probabilitie with data sets of different length
	data.begin <- 1
	probs  <- NULL
	for(i in n)
	{
		data.end 	<- data.begin + i
		probs 	<- c(probs, SDT.Probs(par[data.begin:data.end]))
		data.begin 	<- data.end + 1
	}
	probs
}

# ------------------------------------------------------------------------------------------------------------------
# Model incorporating restrictions
# ------------------------------------------------------------------------------------------------------------------
SDT.Gaussian.Full <- function(par, cfg, pos, fixed = NULL, ident = NULL, functional = NULL)
{
      par <- Restore.par(par, pos, fixed, ident, functional)	# Restore full parameter vector
      SDT.Gaussian(par, cfg)

}


#===============================================================================
#
# 2. Jacobians of model functions
#
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix of a single Probability model
# Jacobi-Matrix with respect to thresholds
# Jacobi-Matrix with respect to main parameters
# ------------------------------------------------------------------------------------------------------------------
SDT.Probs.Jacobi <- function(par)
{
	l   <- length(par)			# Number of parameters
	n.t <- l-2					# Number of thresholds
	m   <- par[1]
	sd  <- max(SDT.Tol, par[2])
	t   <- par[3:l]
	z  <- (t-m)/sd
	fn <- dnorm(z)
	
	# Derivative with respect to mean
	dm <- -fn/sd						# Derivative of Gaussian with respect to m
	dm <- c(dm[1], dm[-1] - dm[-n.t], -dm[n.t])

	# Derivative with respect to sd
	ds <- -z*fn/sd						# Derivative of Gaussian with respect to sd
	ds <- c(ds[1], ds[-1] - ds[-n.t], -ds[n.t])


	# Derivative with respect to thresholds
	dt <- fn/sd							# Derivative of Gaussian with respect to t[i]
	
	J <- matrix(0, nr = n.t+1, nc = n.t)		# Jacobi matrix for thresholds: columns: Parameter, rows = probabilities
	r <- 1
	for(i in 1:n.t)
	{
		J[r, i] <- dt[i]
		r <- r + 1
		J[r, i] <- -dt[i]
	}
	cbind(dm, ds, J)						# Full Jacobi matrix
}

# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix with respect to thresholds
# ------------------------------------------------------------------------------------------------------------------
SDT.Threshold.Jacobi <- function(par)
{
	l   <- length(par)			# Number of parameters
	n.t <- l-2					# Number of thresholds
	m   <- par[1]
	sd  <- max(SDT.Tol, par[2])
	t   <- par[3:l]
	z  <- (t-m)/sd
	fn <- dnorm(z)
	
	# Derivative with respect to thresholds
	dt <- fn/sd							# Derivative of Gaussian with respect to t[i]
	
	J <- matrix(0, nr = n.t+1, nc = n.t)		# Jacobi matrix for thresholds: columns: Parameter, rows = probabilities
	r <- 1
	for(i in 1:n.t)
	{
		J[r, i] <- dt[i]
		r <- r + 1
		J[r, i] <- -dt[i]
	}
	J								# Full Jacobi matrix
}

# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix of the SDT-Model
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT.Jacobi <- function(par, cfg)
{
      n <- cfg$n.sdt
	# Isolate parameters of the single models
	t <- par[(2*n - 1):length(par)]			# Thresholds
	n.t <- length(t)						# Number of thresholds

	J <- filler <- matrix(0, nr = n.t+1, nc = 2)
	filler.r <- filler.c <- NULL
	J.t <- SDT.Threshold.Jacobi(c(0, 1, t))						# Jacobimatrix of thresholds for Modell 1

	j <- 1
	for(i in 1:(n-1))
	{
		J.m <- SDT.Probs.Jacobi(c(par[j:(j+1)], t))
		J <- rbind(cbind(J, filler.c), cbind(filler.r, J.m[, 1:2, drop = F]))		# Jacobimatrix of mean and sd for Model i
		J.t <- rbind(J.t, J.m[, -c(1,2), drop = F])						# Add Jacobimatrix of thresholds for Model i

		if(i >= (n-1)) return( cbind(J, J.t))
		
		filler.c <- rbind(filler.c, filler)
		if(i == 1) filler.c <- rbind(filler.c, filler)				# Repeat due to first model
		filler.r <- cbind(filler.r, filler)
		j <- j + 2
		
	}
	cbind(J, J.t)
}

# ------------------------------------------------------------------------------------------------------------------
# Jacobi-Matrix of the Gaussian-Model:
# (a) General function
# (b) scalar version (equal number of data)
# (c) vector version (with different number of data for different signal distributions)
# ------------------------------------------------------------------------------------------------------------------
SDT.Gaussian.Jacobi <- function(par, cfg)
{
      n <- cfg$n.sdt
	if(length(n) <= 1) SDT.Gaussian.Jacobi.scalar(par, n)
	else			 SDT.Gaussian.Jacobi.vector(par, n)
}

SDT.Gaussian.Jacobi.scalar <- function(par, n)
{
	offset <- n.p <- length(par)/n			# Number of parameters
	
	filler.r <- filler.c <- filler <- matrix(0, nr = n.p-1, nc = n.p)
	J <- SDT.Probs.Jacobi(par[1:n.p])
	for(i in 2:n)
	{
		J.m <- SDT.Probs.Jacobi(par[(offset+1):(offset+n.p)])
		J <- rbind(cbind(J, filler.c), cbind(filler.r, J.m))
		if(i >= n) return( J )
		
		offset <- offset + n.p
		filler.c <- rbind(filler.c, filler)
		filler.r <- cbind(filler.r, filler)
		
	}
	J
}

SDT.Gaussian.Jacobi.vector <- function(par, n)
{
	l.n <- length(n)
	data.begin <- 1					# Start of data

	# Compute first signal distribution
	data.end <- data.begin + n[1]
	J <- SDT.Probs.Jacobi(par[data.begin:data.end])
	for(i in 2:l.n)
	{
		data.begin <- data.end + 1
		data.end <- data.begin + n[i]
		J.m <- SDT.Probs.Jacobi(par[data.begin:data.end])
		filler.c <- filler.r <- NULL
		for(j in 2:i)
		{
			filler.c <- rbind(filler.c, matrix(0, nr = n[j-1], nc = n[i]+1)) 
			filler.r <- cbind(filler.r, matrix(0, nr = n[i]  , nc = n[j-1]+1)) 
		}
		J <- rbind(cbind(J, filler.c), cbind(filler.r, J.m))
	}
	J
}

# ------------------------------------------------------------------------------------------------------------------
#  Computation of the derivation of the Jacobian matrix: d^2p/theta^2
#  Input:   1. parvec = Full parameter vector (but without (0,1))
#           2. n.s    = number of Gaussian models (signals)
#
#  Output:  (n.par x n.par x n.s*R) - array of second partial derivatives 
#
# Comment:  n.par = Number of paramters (except (0,1) of the first model)
#           R     = Number of response categories
# ------------------------------------------------------------------------------------------------------------------
SDT.Deriv.Jacobian <- function(parvec, n.s)
{
      parvec <- c(0, 1, parvec)           # Insert mean and st parameters of first model
	len   <- length(parvec)			# Number of parameters
	n.t <- len - 2*n.s			# Number of thresholds
	index.m <- 1                        # Index to the mean and standard deviation (-1 indeicates position of (fixed parameters 0,1)
      index.t <- 2*(n.s-1)                # Index of thresholds (1 must be added)
      t <- parvec[(2*n.s + 1):len]                        # Extract thresholds
      Deriv.J <- array(0, dim = c(len-2, len-2, n.s*(n.t+1)))    # Initialize derviative of the Jacobian matrix

      p.i  <- 1:(n.t+1)                   # Index of probabilities
      for(i in 1:n.s)                     # Iterate over models
      {
            m <-  parvec[index.m]
            sd <- parvec[index.m+1]
            z  <- (t - m)/sd
            fn <- dnorm(z)

            dm2  <- dt2 <-    -z/sd/sd * fn        # Vector of second derivatives with respect to mean and thresholds
            ds2  <- (2*z - z*z*z)/sd/sd * fn       # Vector of second derivatives with respect to standard deviation
            dmdt <- -dt2                           # Vector of derivatives with respect to mean and thresholds
            dmds <- (1-z*z)/sd/sd * fn             # Vector of derivatives with respect to mean and standard deviation
            dsdt <- -dmds                          # Vector of derivatives with respect to threshold and standard deviation   

            # Computation of indices in the array and fill the array
            # 1. Derivatives with respect to mean and sd
            if(index.m > 2)                           
            {
                  mean.i <- index.m - 2
                  sd.i   <- index.m - 1
                  Deriv.J[mean.i, mean.i, p.i] <- c(dm2[1], dm2[-1] - dm2[-n.t], -dm2[n.t])
                  Deriv.J[sd.i,   sd.i,   p.i] <- c(ds2[1], ds2[-1] - ds2[-n.t], -ds2[n.t])
                  Deriv.J[mean.i, sd.i,   p.i] <- Deriv.J[sd.i,   mean.i, p.i] <- c(dmds[1],dmds[-1] - dmds[-n.t], -dmds[n.t])
                   
            }  
            # 2. Derivatives with respect to thresholds
            for(k in 1:n.t)
            {
                  t.i <- index.t + k
                  Deriv.J[t.i, t.i, p.i[k]]   <-  dt2[k]
                  Deriv.J[t.i, t.i, p.i[k+1]] <- -dt2[k]
                  if(index.m > 2)                           
                  {
                        Deriv.J[mean.i, t.i, p.i[k]] <- Deriv.J[t.i, mean.i, p.i[k]]     <-  dmdt[k]
                        Deriv.J[sd.i,   t.i, p.i[k]] <- Deriv.J[t.i, sd.i,   p.i[k]]     <-  dsdt[k]
                        Deriv.J[mean.i, t.i, p.i[k+1]] <- Deriv.J[t.i, mean.i, p.i[k+1]] <- -dmdt[k]
                        Deriv.J[sd.i,   t.i, p.i[k+1]] <- Deriv.J[t.i, sd.i,   p.i[k+1]] <- -dsdt[k]
                  }  
            }            

            index.m <- index.m + 2
            p.i <- p.i + n.t + 1
      }
      Deriv.J
}


#===============================================================================
#
# 3. Functions for comupting the model matrix (analytically and numerically)
#
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Numerical computation of the model matrix (SDT model)
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT.Model.Matrix.Num <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)
      par <- Elim.par(full.par, fixed, ident)
      J <- jacobian(SDT.SDT.Full, par, cfg = cfg, pos = pos, fixed = fixed, ident = ident, functional = functional)
}

# ------------------------------------------------------------------------------------------------------------------
# Numerical computation of the model matrix  (Gaussian Model)
# ------------------------------------------------------------------------------------------------------------------
SDT.Gaussian.Model.Matrix.Num <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)
      par <- Elim.par(full.par, fixed, ident)
      J <- jacobian(SDT.Gaussian.Full, par, cfg = cfg, pos = pos, fixed = fixed, ident = ident, functional = functional)
}

# ------------------------------------------------------------------------------
# Analytical computation of the model matrix (SDT model)
# Input:    1. Full parameter vector
#           2. List with configuration information
#           3. 2 x p - matrix of fixed constraints
#           4. 2 x q - matrix of identity constraints
#           5. Jacobian of functional constraints (supplied by the user)
#
# Output:   Modelmatrix
#
# Comment:  The Modelmatrix of funtional constraints must be a full matrix that
#           is premultiplied by the Jacobian
# ------------------------------------------------------------------------------
SDT.SDT.Model.Matrix <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)

      J   <- SDT.SDT.Jacobi(full.par, cfg = cfg)
      if(!is.null(ident))
      {
            J.Id <- diag(length(full.par))
            Index <- matrix(c(ident[2,], ident[1,]), nc = 2)
            J.Id[Index] <- 1
            J <- J %*% J.Id
      }
      
      # Treatment of funtional constraints
      if(!is.null(functional))
      {
            J.fct <- attr(functional, "SDT.Jacobian")
            if(is.null(J.fct)) 
                  return(NULL)           # No Jacobian function supplied -> return (NULL)
            J <- J %*% J.fct(full.par)
      }
      J[, pos]
}

# ------------------------------------------------------------------------------
# Analytical computation of the model matrix (Gaussian Model)
# Input:    1. Full parameter vector
#           2. List with configuration information
#           3. 2 x p - matrix of fixed constraints
#           4. 2 x q - matrix of identity constraints
#           5. Jacobian of functional constraints (supplied by the user)
#
# Output:   Model matrix
#
# Comment:  The Modelmatrix of funtional constraints must be a full matrix that 
#           is premultiplied by the Jacobian
# ------------------------------------------------------------------------------
SDT.Gaussian.Model.Matrix <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)

      J   <- SDT.Gaussian.Jacobi(full.par, cfg = cfg)
      if(!is.null(ident))
      {
            J.Id <- diag(length(full.par))
            Index <- matrix(c(ident[2,], ident[1,]), nc = 2)
            J.Id[Index] <- 1
            J <- J %*% J.Id
      }

      # Treatment of functional constraints
      if(!is.null(functional))
      {
            J.fct <- attr(functional, "SDT.Jacobian")
            if(is.null(J.fct)) 
                  return(NULL)           # No Jacobian function supplied -> return (NULL)
            J <- J %*% J.fct(full.par)
      }
      J[, pos]
}

#===============================================================================
#
# 4. Computation of Gradients
#
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Gradient of logL for the SDT and the Gaussian model
# Input:	Same as for Likelihood function
#
# Output:	Gradient
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT.Grad <- function(par, data, Model, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
	par 	   <- Restore.par(par, pos, fixed, ident, functional)	# Restore full parameter vector
	gradient <- as.vector(SDT.logL.Jacobi(par, data, Model, n, pos, fixed, ident, functional) %*% SDT.SDT.Jacobi(par, n)) 

	# Treat functional constraints
	if(!is.null(functional))
	{
		gr.fct <- attr(par, "SDT.gradient")		# Isolate gradient function
		if(is.null(gr.fct))	stop("\n\n SDT.SDT.Grad: Error in computation of gradient: No gradient for functional constraints found:\n Set Parameter sym.gr = F in function SDT.Estimate !\n\n")
		gradient <- gr.fct(gradient)
	}	
	Elim.par.grad(gradient, fixed, ident)
}

SDT.Gaussian.Grad <- function(par, data, Model, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
	par 	    <- Restore.par(par, pos, fixed, ident, functional)	# Restore full parameter vector
	gradient  <- as.vector(SDT.logL.Jacobi(par, data, Model, n, pos, fixed, ident, functional) %*% SDT.Gaussian.Jacobi(par, n))

	# Treat functional constraints
	if(!is.null(functional))
	{
		gr.fct <- attr(par, "SDT.gradient")		# Isolate gradient function
		if(is.null(gr.fct))                       # Internally constructed gradient function
		{   # Look if gradient may be found with functional function
		    gr.fct <- attr(functional, "SDT.gradient")
		    if(is.null(gr.fct))	stop("\n\n SDT.Gaussian.Grad: Error in computation of gradient: No gradient for functional constraints found:\n Set Parameter sym.gr = F in function SDT.Estimate !\n\n")
		}
            gradient <- gr.fct(gradient)
	}	
 	Elim.par.grad(gradient, fixed, ident)
}


#===============================================================================
#
# 5. Analytical computation of Hessian matrix
#
#===============================================================================

# ------------------------------------------------------------------------------
#  Computation of the Hessian matrix: 
#  Input:   1. parvec         = Full parameter vector (but without (0,1))
#           2. datavec        = Data vector
#           3. cfg            = Configuration list
#           4. fixed          = 2 x f matrix of fixed constraints
#           5. ident          = 2 x i matrix of identity constraints
#           6. functional     = function implementing functional constraints
#
#  Output:  Hessian matrix
#
# Comment:  
# ------------------------------------------------------------------------------
SDT.Hessian <- function(parvec, datavec, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      # 1. Compute part 1: dlogL * D.Jacobi
      Hess.1 <- NULL
      
      # 1.1 Compute derivative of Jacobian matrix
      J.Deriv <- SDT.Deriv.Jacobian(parvec, cfg$n.sdt)

      # 1.2 Compute gradient
      pos       <- SavePos.par(parvec, fixed, ident)
      Grad.logL <- SDT.logL.Jacobi(parvec, datavec, SDT.SDT, n = cfg, pos = pos, fixed = fixed, ident = ident, functional = functional)
      
      # 1.3 Compute part 1 of the Hessian
      for(i in 1:dim(J.Deriv)[3])
      {
            if(i == 1)
                  Hess.1 <- Grad.logL[i] * J.Deriv[,,i]
            else
                  Hess.1 <- Hess.1 + Grad.logL[i] * J.Deriv[,,i]      
      }
      
      # 2. Compute part 2: t(J) %*% d2logL %*% J 
      # 2.1 Second derivative matrix of logL
      Grad2.logL <- SDT.logL.Jacobi.2(parvec, datavec, SDT.SDT, n = cfg, fixed = fixed, ident = ident, functional = functional)
      # 2.2 Model matrix
      J <- SDT.SDT.Jacobi(parvec, cfg)
      Hess <- Hess.1 + t(J) %*% Grad2.logL %*% J

      if(is.null(fixed) && is.null(ident))
            return(Hess)
            
      # 3. Now adjust Hessian for constraints
      J.adj <- NULL      
      
      # 3.1 Treatment of Equality constreints
      if(!is.null(ident))
      {
            J.adj <- diag(length(parvec))
            Index <- matrix(c(ident[2,], ident[1,]), nc = 2)
            J.adj[Index] <- 1
      }
      
      # 3.2 Treatment of functional constraints
      if(!is.null(functional))
      {
            J.fct <- attr(functional, "SDT.Jacobian")
            if(is.null(J.fct)) 
                  cat("\n\n SDT.Hessian: No Jacobian of nonlinear constraints specified. Hessian incorrect\n\n")
            else 
                  J.adj <- J.adj %*% J.fct(parvec)
      }
      if(is.null(J.adj))
            return(Hess[pos, pos])                          # Fixed constraints only
            
      J.adj <- J.adj[, pos]
      return(t(J.adj) %*% Hess %*% J.adj)
}      

#===============================================================================
#
# 6. Help functions
#
#===============================================================================


# ------------------------------------------------------------------------------------------------------------------
# Check and generate configuration structure
# Input:	cfg = Configuration structure: A list with the following entries
#                 - n.sdt        Number of SDT models
#                 - restriction  string identifying the restriction
#
# Output: Configuration list
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT.Check.Fill.Cfg <- function(cfg)
{
      n.sdt <- 2
      if(is.null(cfg))
      {
            cfg <- list(n.sdt       = n.sdt,
                        restriction = "no")
            return(cfg)
      }
      if(is.numeric(cfg))
      {
            cfg      <- list(n.sdt       = as.integer(cfg),
                             restriction = "no")
            return(cfg)
      }
      cfg.new             <- list(n.sdt = ifelse(is.null(cfg$n.sdt), n.sdt, cfg$n.sdt))
      cfg.new$restriction <- SDT.SDT.Check.Restriction(cfg$restriction)

      # Checks
      if(cfg.new$n.sdt < 2)
            stop(paste("\n FATAL ERROR in SDT.SDT.Check.Fill.Cfg:\n Number of SDT signals <", cfg.new$n.sdt, "> smaller than <2>\n(Minimum number = 2)\n", sep = ""))
      return(cfg.new)
}  
#-------------------------------------------------------------------------------
# Gaussian version
#-------------------------------------------------------------------------------
SDT.Gaussian.Check.Fill.Cfg <- function(cfg)
{
      n.sdt <- 2
      if(is.null(cfg))
      {
            cfg <- list(n.sdt = n.sdt)
            if(length(cfg$n.sdt) > 1)
                        cfg$restriction = "no"
            else
                        cfg$restriction = "standard"
            return(cfg)
      }
      if(is.numeric(cfg))
      {
            cfg      <- list(n.sdt = as.integer(cfg))
            if(length(cfg$n.sdt) > 1)
                        cfg$restriction = "no"
            else
                        cfg$restriction = "standard"
            return(cfg)
      }
      if(is.null(cfg$n.sdt))
            cfg.new <- list(n.sdt = n.sdt)
      else
            cfg.new <- list(n.sdt = cfg$n.sdt)
            
      if(length(cfg.new$n.sdt) > 1)      # Restrictions only in case of equal number of data per signal
            cfg.new$restriction <- "no"
      else
            cfg.new$restriction <- SDT.Gaussian.Check.Restriction(cfg$restriction)
      
      # Checks
      if(length(cfg.new$n.sdt) < 2 && sum(cfg.new$n.sdt) < 2)
            stop(paste("\n FATAL ERROR in SDT.Gaussian.Check.Fill.Cfg:\n Total number of Gaussians <", sum(cfg.new$n.sdt), "> smaller than <2>\n(Minimum number = 2)\n", sep = ""))
      if(cfg.new$restriction == "pairs" && length(cfg.new$n.sdt) > 1)
            stop(paste("\n FATAL ERROR in SDT.Gaussian.Check.Fill.Cfg:\n Restriciton =\"pairs\" does not work with vectors of (different) observations\n", sep = ""))
      if(cfg.new$restriction == "standard-pairs" && length(cfg.new$n.sdt) > 1)
            stop(paste("\n FATAL ERROR in SDT.Gaussian.Check.Fill.Cfg:\n Restriciton =\"standard-pairs\" does not work with vectors of (different) observations\n", sep = ""))
      if(cfg.new$restriction == "pairs-equalvar" && length(cfg.new$n.sdt) > 1)
            stop(paste("\n FATAL ERROR in SDT.Gaussian.Check.Fill.Cfg:\n Restriciton =\"pairs-equalvar\" does not work with vectors of (different) observations\n", sep = ""))
      if(cfg.new$restriction == "standard-pairs-equalvar" && length(cfg.new$n.sdt) > 1)
            stop(paste("\n FATAL ERROR in SDT.Gaussian.Check.Fill.Cfg:\n Restriciton =\"standard-pairs-equalvar\" does not work with vectors of (different) observations\n", sep = ""))

      if(cfg.new$restriction == "pairs" && cfg.new$n.sdt %% 2 > 0)
            stop(paste("\n FATAL ERROR in SDT.Gaussian.Check.Fill.Cfg:\n Restriciton =\"pairs\" requires an equal number of Gaussian models. Actual number = <", sum(cfg.new$n.sdt), ">\n", sep = ""))
      if(cfg.new$restriction == "standard-pairs" && cfg.new$n.sdt %% 2 > 0)
            stop(paste("\n FATAL ERROR in SDT.Gaussian.Check.Fill.Cfg:\n Restriciton =\"standard-pairs\" requires an equal number of Gaussian models. Actual number = <", sum(cfg.new$n.sdt), ">\n", sep = ""))
      if(cfg.new$restriction == "pairs-equalvar" && cfg.new$n.sdt %% 2 > 0)
            stop(paste("\n FATAL ERROR in SDT.Gaussian.Check.Fill.Cfg:\n Restriciton =\"pairs-var\" requires an equal number of Gaussian models. Actual number = <", sum(cfg.new$n.sdt), ">\n", sep = ""))
      if(cfg.new$restriction == "standard-pairs-equalvar" && cfg.new$n.sdt %% 2 > 0)
            stop(paste("\n FATAL ERROR in SDT.Gaussian.Check.Fill.Cfg:\n Restriciton =\"standard-pairs-var\" requires an equal number of Gaussian models. Actual number = <", sum(cfg.new$n.sdt), ">\n", sep = ""))
      return(cfg.new)
}                  
                
# ------------------------------------------------------------------------------------------------------------------
# Check type of restriction
# Input: A string specfiying the typ of restriction
#        Types of restrictions      - "equalvar"      (Equal variance)
#                                   - "no"  [or F]    (No restrictions)
#
# Output: String with parameter values
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT.Check.Restriction <- function(restriction)
{
      if(is.null(restriction))
            return("no")
      if(charmatch(restriction, c("equalvar", "EQUALVAR", "Equalvar"), nomatch = 0) > 0)
            return("equalvar")
      if(charmatch(restriction, c("no", "No", "NO"), nomatch = 0) > 0 || restriction == F)
            return("no")
      warn.msg <- paste("\n\n SDT.Restriction: Unkonwn restriction in configuration list: <", restriction, "> --> setting to: <no>\n\n", sep = "")      
      warning(warn.msg)
      return("no")
}            
#-------------------------------------------------------------------------------
# Gaussian version
#-------------------------------------------------------------------------------
SDT.Gaussian.Check.Restriction <- function(restriction)
{
      if(is.null(restriction))
            return("standard")
      if(charmatch(restriction, c("equalvar", "EQUALVAR", "Equalvar"), nomatch = 0) > 0)
            return("equalvar")
      if(charmatch(restriction, c("no", "No", "NO"), nomatch = 0) > 0 || restriction == F)
            return("no")
      if(charmatch(restriction, c("standard", "Standard", "STANDARD"), nomatch = 0) > 0 || restriction == T)
            return("standard")
      if(charmatch(restriction, c("symmetric", "Symmetric", "SYMMETRIC"), nomatch = 0) > 0 || restriction == T)
            return("symmetric")
      if(charmatch(restriction, c("equalvar-symmetric", "equal-symmetric", "EQUALVAR-SYMMETRIC", "EQUAL-SYMMETRIC"), nomatch = 0) > 0 || restriction == T)
            return("equalvar-symmetric")
      if(charmatch(restriction, c("pairs", "PAIRS", "Pairs"), nomatch = 0) > 0 || restriction == T)
            return("pairs")
      if(charmatch(restriction, c("pairs-equalvar", "PAIRS-EQUALVAR", "Pairs-Equalvar", "Pairs-equalvar"), nomatch = 0) > 0 || restriction == T)
            return("pairs-equalvar")
     if(charmatch(restriction, c("standard-pairs", "STANDARD-PAIRS", "Standard-Pairs"), nomatch = 0) > 0 || restriction == T)
            return("standard-pairs")
      if(charmatch(restriction, c("standard-pairs-equalvar", "STANDARD-PAIRS-EQUALVAR", "Standard-Pairs-Equalvar", "Pairs-equalvar"), nomatch = 0) > 0 || restriction == T)
            return("standard-pairs-equalvar")
      warn.msg <- paste("\n\n Gaussian.Restriction: Unkonwn restriction in configuration list: <", restriction, "> --> setting to: <standard>\n\n", sep = "")      
      warning(warn.msg)
      return("standard")
}            


# ------------------------------------------------------------------------------------------------------------------
# Generate starting parameters due to data and number of signals
# Input:	data = Vector of data or number or the number of data points per signal
#		 cfg = configuration list)
#
# Output: String with parameter names
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT.start.par <- function(cfg, data)
{
      n <- cfg$n.sdt
      n.dat <- length(data)
      if(n.dat < 2) 
            n.dat <- data*n
      
	# Secify Number of thresholds
	n.t <- as.integer(n.dat/n - 1)
	c(rep(c(0, 1), n-1), seq(-0.9, 0.9, length = n.t))
}

SDT.Gaussian.start.par <- function(cfg, data)
{
      n <- cfg$n.sdt
	if(length(n) < 2)		# n indicates number of signals
	{
            n.dat <- length(data)
            if(n.dat < 2) 
                  n.dat <- data*n

            n.t <- as.integer(n.dat/n - 1)			# Specify Number of thresholds
      	par.m <- c(0, 1, seq(-1.5, 1.5, length = n.t))	# Parameter vector of one model

	      return(rep(par.m, n))
	}
	# n indicates vector of number of reponses per signal
	unlist(sapply(n, function(n) c(0,1, seq(-1.5, 1.5, length = n-1))))
}
# ------------------------------------------------------------------------------------------------------------------
# Generate upper and lower bounds of parameter
# Input:	parvec = Vector of parameters
#		   cfg = configruation structure
#
# Output: List with lower and upper bounds on parameter vector
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT.par.bounds <- function(parvec, cfg)
{
      len.p <- length(parvec)
      n.th  <- len.p - 2*(cfg$n.sdt - 1)

      lower <- c(rep(c(-Inf, SDT.Tol), cfg$n.sdt-1), rep(-Inf, n.th))
      upper <- rep(Inf, len.p)
      
      return(list(lower = lower, upper = upper))
}      

SDT.Gaussian.par.bounds <- function(parvec, cfg)
{
      n <- cfg$n.sdt
      len.p <- length(parvec)
      upper <- rep(Inf, len.p)
      if(length(n) < 2)
 	{
            n.th  <- as.integer(len.p/n) - 2    # Specify Number of thresholds            
            lower <- rep(c(-Inf, SDT.Tol, rep(-Inf, n.th)), n)
	      return(list(lower = lower, upper = upper))
	}
     
	lower <- unlist(sapply(n, function(n) c(-Inf, SDT.Tol, rep(-Inf, n-1))))
      return(list(lower = lower, upper = upper))
}      


# ------------------------------------------------------------------------------------------------------------------
# Generate default target parameters for which LR intervals are computed
# Input:	par   = Vector of parameters
#		cfg   = configuration list
#           fixed = 2 x k matrix of fixed constraints
#           ident = 2 x m matrix of equality constraints
#
# Output: vector with positions of parameters
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT.target.pars.LR <- function(par, cfg)
{
      n <- cfg$n.sdt
      c(seq(1, 2*(n-1)- 1, by = 2), seq(2, 2*(n-1), by = 2))
}

SDT.Gaussian.target.pars.LR <- function(par, cfg)
{
      n <- cfg$n.sdt
	if(length(n) < 2) SDT.Gaussian.target.pars.LR.scalar(par, n)
	else			SDT.Gaussian.target.pars.LR.vector(par, n)
}

SDT.Gaussian.target.pars.LR.scalar <- function(par, n)
{
      n.t <- length(par)/n - 2
      pos.sd  <- seq(2, (2+n.t)*n, by = 2+n.t) 
      c(pos.sd - 1, pos.sd)     
}

SDT.Gaussian.target.pars.LR.vector <- function(par, n)
{
      m <- cumsum(c(1, 2 + (n-1)))
      m <- m[-length(m)]
	c(m, m + 1)
}

# ------------------------------------------------------------------------------------------------------------------
# Generate strings with names of the parameters for the models
# Input:	par = Vector of parameters
#		  n = number of models
#           flag = F (not used)
#
# Output: String with parameter names
# ------------------------------------------------------------------------------------------------------------------
SDT.par.names <- function(par, cfg, flag = F)
{
      n <- cfg$n.sdt
	m  <- paste("Mean["  , 2:n, "]", sep = "")
	sd <- paste("Stddev[", 2:n, "]", sep = "")
	p  <- c(m, sd)
	t  <- paste("t-" , 1:(length(par)- 2*(n-1)), sep = "")

	index <- matrix(1:(n+n-2), nr = 2, byrow = T)
	c(p[as.vector(index)], t)
}

Gaussian.par.names <- function(par, cfg, flag = F)
{
      n <- cfg$n.sdt
	if(length(n) < 2) Gaussian.par.names.scalar(par, n)
	else			Gaussian.par.names.vector(par, n)
}

# ------------------------------------------------------------------------------------------------------------------
# Names with n being a number (indicating the number of signals)
# ------------------------------------------------------------------------------------------------------------------
Gaussian.par.names.scalar <- function(par, n)
{
      
	m   <- paste("Mean[" , 1:n, "]", sep = "")
	sd  <- paste("Stddev[", 1:n, "]", sep = "")
	n.t <- length(par)/n - 2
	t   <- paste("t-",  1:n.t, sep = "")
	p   <- NULL
	for(i in 1:n) p <- c(p, m[i], sd[i], paste(t, "[", i, "]", sep = ""))
	p
}
# ------------------------------------------------------------------------------------------------------------------
# Names with n being a vector with different number of data per signal
# ------------------------------------------------------------------------------------------------------------------
Gaussian.par.names.vector <- function(par, n)
{
	l.n <- length(n)

	names <- NULL
	for(i in 1:l.n)
	{
		names <- c(names, paste("Mean[" , i, "]", sep = ""), paste("Stddev[", i, "]", sep = ""))
		t     <- paste("t-",  1:(n[i]-1), sep = "")
		t     <- paste(t, "[", i, "]", sep = "")
		names <- c(names, t)
	}
	names
}

# ------------------------------------------------------------------------------------------------------------------
# Computation of bounds on parameters:
# ------------------------------------------------------------------------------------------------------------------
SDT.par.bounds  <- function(par, cfg)
{
      Min.Var   <- SDT.Tol


      lower <- rep(-Inf, length(par))
      upper <- rep(Inf, length(par))
      
      pos.sd <- seq(2, 2*(cfg$n.sdt - 1), by = 2)
      lower[pos.sd] <- Min.Var           # Minimal value of the scale parameter of the Gaussian distributions
      
      list(lower = lower, upper = upper)
}

Gaussian.par.bounds  <- function(par, cfg)
{
      n <- cfg$n.sdt
	if(length(n) < 2) Gaussian.par.bounds.scalar(par, n)
	else			Gaussian.par.bounds.vector(par, n)
}      

Gaussian.par.bounds.scalar  <- function(par, n)
{

      Min.Var   <- SDT.Tol

      lower <- rep(-Inf, length(par))
      upper <- rep(Inf, length(par))
      
     	n.t <- length(par)/n - 2
     	
      pos.sd <- seq(2, (n.t+2)*n, by = n.t+2)
      lower[pos.sd] <- Min.Var           # Minimal value of the scale parameter of the Gaussian distributions
      
      list(lower = lower, upper = upper)
}

Gaussian.par.bounds.vector  <- function(par, n)
{

      Min.Var   <- SDT.Tol

      lower <- rep(-Inf, length(par))
      upper <- rep(Inf, length(par))
      
      pos.sd <- cumsum(c(2, n[-length(n)]+1))
      lower[pos.sd] <- Min.Var           # Minimal value of the scale parameter of the Gaussian distributions
      
      list(lower = lower, upper = upper)
}

# ------------------------------------------------------------------------------------------------------------------
# Names of the models 
# ------------------------------------------------------------------------------------------------------------------
SDT.model.name <- function(cfg)
{              
      mod.name <- "Standard SDT model with noise distribution N(0, 1), Number of models: <"
      mod.name <- paste(mod.name, cfg$n.sdt, ">", sep = "")
      mod.name <- paste(mod.name, ", Type of restrictions: <", cfg$restriction, ">", sep = "")
      return(mod.name)
}

Gaussian.model.name <- function(cfg)
{
      if(length(cfg$n.sdt) > 1)  G.mods <- paste(cfg$n.sdt, collapse = ", ")
      else                       G.mods <- cfg$n.sdt

      if(length(cfg$n.sdt) > 1)  
            mod.name <- "Gaussian SDT model with freely estimable parameters for each model, data points per signal: <"
      else
            mod.name <- "Gaussian SDT model with freely estimable parameters for each model, Number of models: <"
      mod.name <- paste(mod.name, G.mods, ">", sep = "")
      mod.name <- paste(mod.name, ", Type of restrictions: <", cfg$restriction, ">", sep = "")
      return(mod.name)
}


# ------------------------------------------------------------------------------------------------------------------
# SDT: Setting predefined restrictions
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT.Restriction <- function(cfg, fixed, ident)
{
      # If configuration information is not complete, then complete it
      cfg <- SDT.SDT.Check.Fill.Cfg(cfg)  

      if(cfg$restriction == "equalvar")
            return(SDT.SDT.Set.Equalvar.Restriction(cfg, fixed, ident))
     return(list(fixed = fixed, ident = ident))
}
# ------------------------------------------------------------------------------------------------------------------
# Gaussian: Setting predefined restrictions
# ------------------------------------------------------------------------------------------------------------------

SDT.Gaussian.Restriction <- function(parvec, cfg, fixed, ident, functional)
{

     # If configuration information is not complete, then complete it
     cfg <- SDT.Gaussian.Check.Fill.Cfg(cfg)  
     
     if(cfg$restriction == "standard")
            return(SDT.Gaussian.Set.Standard.Restriction(parvec, cfg, fixed, ident, functional))
     if(cfg$restriction == "equalvar")
            return(SDT.Gaussian.Set.Equalvar.Restriction(parvec, cfg, fixed, ident, functional))
     if(cfg$restriction == "symmetric")
            return(SDT.Gaussian.Set.Symmetric.Restriction(parvec, cfg, fixed, ident, functional))
     if(cfg$restriction == "equalvar-symmetric")
            return(SDT.Gaussian.Set.Equalvar.Symmetric.Restriction(parvec, cfg, fixed, ident, functional))
     if(cfg$restriction == "pairs")
            return(SDT.Gaussian.Set.Pairs.Restriction(parvec, cfg, fixed, ident, functional))
     if(cfg$restriction == "pairs-equalvar")
            return(SDT.Gaussian.Set.Pairs.Equalvar.Restriction(parvec, cfg, fixed, ident, functional))
     if(cfg$restriction == "standard-pairs")
            return(SDT.Gaussian.Set.Standard.Pairs.Restriction(parvec, cfg, fixed, ident, functional))
     if(cfg$restriction == "standard-pairs-equalvar")
            return(SDT.Gaussian.Set.Standard.Pairs.Equalvar.Restriction(parvec, cfg, fixed, ident, functional))
     return(list(fixed = fixed, ident = ident, functional = functional))
}

# ------------------------------------------------------------------------------------------------------------------
# SDT: Set equal variance restrictions
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT.Set.Equalvar.Restriction <- function(cfg, fixed, ident)
{
      val.fx <- rep(1.0, cfg$n.sdt-1)
      if(cfg$n.sdt > 2)
            pos.fx <- 2*(1:(cfg$n.sdt - 1))
      else
            pos.fx <- 2

      fixed.s <- matrix(c(val.fx,
                          pos.fx), nr = 2, byrow = T)
                          
      # Cheque for uniqueness add to fixed constraints                         
      pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
	fixed <- cbind(fixed.s[, pos.unique], fixed)
      list(fixed = fixed, ident = ident)
}

# ------------------------------------------------------------------------------------------------------------------
# Gaussian: Set standard restrictions
# ------------------------------------------------------------------------------------------------------------------
SDT.Gaussian.Set.Standard.Restriction <- function(parvec, cfg, fixed, ident, functional)
{
      SR.G <- SDT.Gaussian.Set.Standard.Restriction.Scalar(parvec, cfg)

      # Cheque for uniqueness add to fixed constraints                         
      fixed.s <- SR.G$fixed         
      pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
	fixed <- cbind(fixed.s[, pos.unique], fixed)

      # Cheque for uniqueness add to identity constraints                         
      ident.s <- SR.G$ident       
      pos.unique <-  !(ident.s[2,] %in% ident[2,])
	ident <- cbind(ident.s[, pos.unique], ident)

      list(parvec = parvec, fixed = fixed, ident = ident, functional = functional)
}

SDT.Gaussian.Set.Standard.Restriction.Scalar <- function(parvec, cfg)
{
      # 1. First Gaussian = N(0,1)
      fixed <- matrix(c(0.0, 1.0, 
                          1,   2), nr = 2, byrow = T)
 
      # 2. Identity constraints on thrsholds
      n.t <- length(parvec) / cfg$n.sdt - 2
      source.ident <- rep(2 + 1:n.t, cfg$n.sdt - 1)
      
      incr.ident <- (2 + n.t) * 1:(cfg$n.sdt - 1)
      target.ident <- rep(incr.ident, each = n.t)
      ident <- matrix(c(source.ident, source.ident + target.ident), nr = 2, byrow = T)
            
      list(fixed = fixed, ident = ident)
}

# ------------------------------------------------------------------------------------------------------------------
# Gaussian: Set equal variance restrictions + standard restriction
# ------------------------------------------------------------------------------------------------------------------
SDT.Gaussian.Set.Equalvar.Restriction <- function(parvec, cfg, fixed, ident, functional)
{
      # Set Standard restrictions
      SR.G <- SDT.Gaussian.Set.Standard.Restriction.Scalar(parvec, cfg)

      # Add fixed variance constraints
      val.fx <- rep(1.0, cfg$n.sdt-1)
      n.t <- length(parvec) / cfg$n.sdt - 2
      pos.fx <- 2 + (2 + n.t)*(1:(cfg$n.sdt - 1))

      fixed.s <- matrix(c(val.fx,
                          pos.fx), nr = 2, byrow = T)
      fixed.s <- cbind(fixed.s, SR.G$fixed)
     
     # Cheque for uniqueness add to fixed constraints                         
      pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
	fixed <- cbind(fixed.s[, pos.unique], fixed)

      # Cheque for uniqueness add to identity constraints                         
      ident.s <- SR.G$ident       
      pos.unique <-  !(ident.s[2,] %in% ident[2,])
	ident <- cbind(ident.s[, pos.unique], ident)

      list(fixed = fixed, ident = ident, functional = functional)
}

# ------------------------------------------------------------------------------------------------------------------
# Gaussian: Set symmetric restrictions + equal thresholds
# ------------------------------------------------------------------------------------------------------------------
SDT.Gaussian.Set.Symmetric.Restriction <- function(parvec, cfg, fixed, ident, functional)
{
      # Set Standard restrictions
      SR.G <- SDT.Gaussian.Set.Standard.Restriction.Scalar(parvec, cfg)
     
     # Cheque for uniqueness add to fixed constraints                         
      fixed.s <- SR.G$fixed
      pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
	fixed <- cbind(fixed.s[, pos.unique], fixed)

      # Cheque for uniqueness add to identity constraints                         
      ident.s <- SR.G$ident       
      pos.unique <-  !(ident.s[2,] %in% ident[2,])
	ident <- cbind(ident.s[, pos.unique], ident)
	
	# Function that computes functional constraints
	fct.SDT <- function(parvec)
      {
           tindex <- 1                                            # Target index: m1
           sindex <- sdt.gaussian.symmetric.restriction.index__   # Source index: m2

	     parvec[tindex] <- -parvec[sindex]
	     return(parvec)
      }

      # Function that computes Gradient function
	gr.fct <- function(grad)
	{
	     tindex <- 1
           sindex <- sdt.gaussian.symmetric.restriction.index__   # Source index: m2
	     grad[sindex] <- grad[sindex] - grad[tindex]
	     return(grad)
	}
	     
      # Function computing the Jacobian for functional constraints
      J.fct <- function(parvec)
      {
            sindex <- sdt.gaussian.symmetric.restriction.index__  # Source index: m2
            J <- diag(length(parvec))
            J[1, sindex] <- -1
            J
      }
      # Positions of the mean 2
      npar.m <- length(parvec) / cfg$n.sdt          # Number of parameters per model
      sdt.gaussian.symmetric.restriction.index__ <<- 1 + npar.m

      if(is.null(attr(fct.SDT, "SDT.gradient")))
             attr(fct.SDT, "SDT.gradient") <- gr.fct	    # Jacobian function as attribute of the funcitonal constraint
      else
            stop("\n FATAL ERROR in SDT.Gaussian.Set.Symmetric.Restriction: Gradient function already exists\n")
      if(is.null(attr(fct.SDT, "SDT.M.Jacobian")))
             attr(fct.SDT, "SDT.M.Jacobian") <- J.fct	 # Jacobian function as attribute of the funcitonal constraint
      else
            stop("\n FATAL ERROR in SDT.Gaussian.Set.Symmetric.Restriction: Jacobian function already exists\n")
  
      list(fixed = fixed, ident = ident, functional = fct.SDT)
}

# ------------------------------------------------------------------------------------------------------------------
# Gaussian: Set equal variance restrictions + standard restriction
# ------------------------------------------------------------------------------------------------------------------
SDT.Gaussian.Set.Equalvar.Symmetric.Restriction <- function(parvec, cfg, fixed, ident, functional)
{
      # Set symmetric restrictions
      SR.S <- SDT.Gaussian.Set.Symmetric.Restriction(parvec, cfg, fixed, ident, functional)

      # Add fixed variance constraints
      val.fx <- rep(1.0, cfg$n.sdt-1)
      n.t <- length(parvec) / cfg$n.sdt - 2
      pos.fx <- 2 + (2 + n.t)*(1:(cfg$n.sdt - 1))

      fixed.s <- matrix(c(val.fx,
                          pos.fx), nr = 2, byrow = T)
      fixed.s <- cbind(fixed.s, SR.S$fixed)
     
     # Cheque for uniqueness add to fixed constraints                         
      pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
	fixed <- cbind(fixed.s[, pos.unique], fixed)

      # Cheque for uniqueness add to identity constraints                         
      ident.s <- SR.S$ident       
      pos.unique <-  !(ident.s[2,] %in% ident[2,])
	ident <- cbind(ident.s[, pos.unique], ident)

      list(fixed = fixed, ident = ident, functional = SR.S$functional)
}

# ------------------------------------------------------------------------------------------------------------------
# Gaussian: Set standard restrictions for pairs:
# - N(0, 1) for models 1, 3, 5, ...
# - identical thresholds for pairs of models
# ------------------------------------------------------------------------------------------------------------------
SDT.Gaussian.Set.Pairs.Restriction <- function(parvec, cfg, fixed, ident, functional)
{
      # 0. Determine length of parameters and generate matrix with parameters of models in columns
      plen <- length(parvec) / cfg$n.sdt
      Pmat <- matrix(1:length(parvec), nc = cfg$n.sdt)

      # 1. Fix m = 0 and sd = 1, for 1. 3. 5. etc. model
      col.source <- seq(1, cfg$n.sdt, 2)
      fixed.s <- matrix(c(rep(c(0, 1), as.integer(cfg$n.sdt/2)), as.vector(Pmat[1:2, col.source])), nr = 2, byrow = T)
      
      # 2. Identical thresholds for pairs
      col.target <- seq(2, cfg$n.sdt, 2)
      
      ident.s <- matrix(c(as.vector(Pmat[c(3:plen), col.source]), 
                          as.vector(Pmat[c(3:plen), col.target])), nr = 2, byrow = T)

     # 3. Cheque for uniqueness of fixed and identity constraints
      pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
	fixed <- cbind(fixed.s[, pos.unique], fixed)

      pos.unique <-  !(ident.s[2,] %in% ident[2,])
	ident <- cbind(ident.s[, pos.unique], ident)

      list(fixed = fixed, ident = ident, functional = functional)
}

# ------------------------------------------------------------------------------------------------------------------
# Gaussian: Set standard restrictions for pairs:
# - mean = 0 for models 1, 3, 5, ... 
# - all variance parameters are 1.0
# - identical thresholds for pairs of models
# ------------------------------------------------------------------------------------------------------------------
SDT.Gaussian.Set.Pairs.Equalvar.Restriction <- function(parvec, cfg, fixed, ident, functional)
{
      # 0. Determine length of parameters and generate matrix with parameters of models in columns
      plen <- length(parvec) / cfg$n.sdt
      Pmat <- matrix(1:length(parvec), nc = cfg$n.sdt)

      # 1. Fix m = 0 and sd = 1, for 1. 3. 5. etc. model
      col.source <- seq(1, cfg$n.sdt, 2)
      fixed.s <- matrix(c(rep(0, as.integer(cfg$n.sdt/2)), as.vector(Pmat[1, col.source])), nr = 2, byrow = T)
      fixed.s <- cbind(fixed.s, matrix(c(rep(1, cfg$n.sdt), as.vector(Pmat[2,])), nr = 2, byrow = T)) 
      
      # 2. Identical thresholds for pairs
      col.target <- seq(2, cfg$n.sdt, 2)
      
      ident.s <- matrix(c(as.vector(Pmat[c(3:plen), col.source]), 
                          as.vector(Pmat[c(3:plen), col.target])), nr = 2, byrow = T)

     # 3. Cheque for uniqueness of fixed and identity constraints
      pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
	fixed <- cbind(fixed.s[, pos.unique], fixed)

      pos.unique <-  !(ident.s[2,] %in% ident[2,])
	ident <- cbind(ident.s[, pos.unique], ident)

      list(fixed = fixed, ident = ident, functional = functional)
}

# ------------------------------------------------------------------------------------------------------------------
# Gaussian: Set standard restrictions for pairs:
# - N(0, 1) for models 1, 3, 5, ...
# - identical thresholds for all models
# ------------------------------------------------------------------------------------------------------------------
SDT.Gaussian.Set.Standard.Pairs.Restriction <- function(parvec, cfg, fixed, ident, functional)
{
      # 0. Determine length of parameters and generate matrix with parameters of models in columns
      plen <- length(parvec) / cfg$n.sdt
      Pmat <- matrix(1:length(parvec), nc = cfg$n.sdt)

      # 1. Fix m = 0 and sd = 1, for 1. 3. 5. etc. model
      col.source <- seq(1, cfg$n.sdt, 2)
      fixed.s <- matrix(c(rep(c(0, 1), as.integer(cfg$n.sdt/2)), as.vector(Pmat[1:2, col.source])), nr = 2, byrow = T)
      
      # 2. Identical thresholds for all models
      col.target <- seq(2, cfg$n.sdt, 2)
      
      ident.s <- matrix(c(rep(as.vector(Pmat[3:plen, 1]), cfg$n.sdt-1),
                          as.vector(Pmat[3:plen, 2:cfg$n.sdt])), nr = 2, byrow = T)

     # 3. Cheque for uniqueness of fixed and identity constraints
      pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
	fixed <- cbind(fixed.s[, pos.unique], fixed)

      pos.unique <-  !(ident.s[2,] %in% ident[2,])
	ident <- cbind(ident.s[, pos.unique], ident)

      list(fixed = fixed, ident = ident, functional = functional)
}

# ------------------------------------------------------------------------------------------------------------------
# Gaussian: Set standard restrictions for pairs:
# - mean = 0 for models 1, 3, 5, ... 
# - all variance parameters are 1.0
# - identical thresholds for all models
# ------------------------------------------------------------------------------------------------------------------
SDT.Gaussian.Set.Standard.Pairs.Equalvar.Restriction <- function(parvec, cfg, fixed, ident, functional)
{
      # 0. Determine length of parameters and generate matrix with parameters of models in columns
      plen <- length(parvec) / cfg$n.sdt
      Pmat <- matrix(1:length(parvec), nc = cfg$n.sdt)

      # 1. Fix m = 0 and sd = 1, for 1. 3. 5. ... model
      col.source <- seq(1, cfg$n.sdt, 2)
      fixed.s <- matrix(c(rep(0, as.integer(cfg$n.sdt/2)), as.vector(Pmat[1, col.source])), nr = 2, byrow = T)
      fixed.s <- cbind(fixed.s, matrix(c(rep(1, cfg$n.sdt), as.vector(Pmat[2,])), nr = 2, byrow = T)) 
      
      # 2. Identical thresholds for pairs
      col.target <- seq(2, cfg$n.sdt, 2)
      
      ident.s <- matrix(c(rep(as.vector(Pmat[3:plen, 1]), cfg$n.sdt-1), 
                          as.vector(Pmat[3:plen, 2:cfg$n.sdt])), nr = 2, byrow = T)

     # 3. Cheque for uniqueness of fixed and identity constraints
      pos.unique <-  !(fixed.s[2,] %in% fixed[2,])
	fixed <- cbind(fixed.s[, pos.unique], fixed)

      pos.unique <-  !(ident.s[2,] %in% ident[2,])
	ident <- cbind(ident.s[, pos.unique], ident)

      list(fixed = fixed, ident = ident, functional = functional)
}


# ------------------------------------------------------------------------------------------------------------------
# Format parameters for SDT and Gaussian model
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Par.SDT <- function(par, cfg, dec)
{
      n <- cfg$n.sdt
	par <- c(0, 1, par)					# Add parameters for N(0, 1)
	t   <- round(par[(2*n+1): length(par)], dec)	# Isolate thresholds
	n.t <- length(t)
	t <- matrix(rep(t, n), nc = n)			# Copy thresholds and format them
	
	Fp <- matrix(round(par[1:(2*n)], dec), nc = n)
	Fp <- rbind(Fp, t)

	colnames(Fp) <- paste("Gauss-", 1:n, sep = "")
	rownames(Fp) <- c("Mean", "Stddev", paste("t-", 1:n.t, sep = ""))

	Fp
}

SDT.Format.Par.Gaussian <- function(par, cfg, dec)
{
      n <- cfg$n.sdt
	if(length(n) < 2) SDT.Format.Par.Gaussian.scalar(par, n, dec)
	else			SDT.Format.Par.Gaussian.vector(par, n, dec)
}

# ------------------------------------------------------------------------------------------------------------------
# Format if n is number (indicating the number of signals)
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Par.Gaussian.scalar <- function(par, n, dec)
{
      
	n.t <- length(par)/n - 2
	Fp <- matrix(round(par, dec), nc = n)
	
	colnames(Fp) <- paste("Gauss-", 1:n, sep = "")
	rownames(Fp) <- c("Mean", "Stddev", paste("t-", 1:n.t, sep = ""))
	
	Fp
}
# ------------------------------------------------------------------------------------------------------------------
# Names with n being a vector with different number of data per signal
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Par.Gaussian.vector <- function(par, n, dec)
{
	l.n <- length(n)
	max.n <- max(n)
	
	Fp <- matrix(NA, nr = max.n+1, nc = l.n)
	colnames(Fp) <- paste("Gauss-", 1:l.n, sep = "")
	rownames(Fp) <- c("Mean", "Stddev", paste("t-", 1:(max.n-1), sep = ""))

	# Fill the matrix
	par.start <- 1
	for(i in 1:l.n)
	{
		par.end       <- par.start + n[i]
	 	Fp[1:(n[i]+1), i] <- round(par[par.start:par.end], dec)
		par.start     <- par.end + 1
	}
	Fp
}

# ------------------------------------------------------------------------------------------------------------------
# SDT Model: Check number of parameters and data points
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT.check <- function(par, data, cfg)
{
      n <- cfg$n.sdt
	l.n <- length(n)
	if(l.n > 1)
	{
		Error <- T
		n.par <- n.data <- NA
		Msg 	<- "The Number of signal distribution must be a scalar for the SDT model"
	}
	else
	{
		n.par  <- 2*(n-1) + length(data)/n - 1
		n.data <- (length(par) - 2*(n-1) + 1)*n
		Msg    <- "NOTE: Computation of the expected number of parameters is based on the number of data points and vice versa"
	}  

	Erg <- matrix(c(length(par), length(data), n.par, n.data), nc = 2)
	rownames(Erg) <- c("Parameters", "Data points")
	colnames(Erg) <- c("Found", "Expected")
	Error <- any(Erg[, 1] != Erg[, 2])

	list(Error = Error, Stat = Erg, Msg = Msg)
}

# ------------------------------------------------------------------------------------------------------------------
# Gaussian Model: Check number of parameters and data points
# ------------------------------------------------------------------------------------------------------------------
SDT.Gaussian.check <- function(par, data, cfg)
{
     n <- cfg$n.sdt
	l.n <- length(n)
	if(l.n > 1)
	{
		n.par  <- sum(n) + l.n
		n.data <- sum(n)
	}
	else
	{
		n.par  <- length(data) + n
		n.data <- length(par)  - n
	}  

	Erg <- matrix(c(length(par), length(data), n.par, n.data), nc = 2)
	rownames(Erg) <- c("Parameters", "Data points")
	colnames(Erg) <- c("Found", "Expected")

	Error <- any(Erg[, 1] != Erg[, 2])

	list(Error = Error, Stat = Erg, Msg = "NOTE: Computation of the expected number of parameters is based on the to number of data points and vice versa")
}

# ------------------------------------------------------------------------------------------------------------------
# Format free parameters and compute confidence intervals
# Input:	1. px2 Matrix with parameters and standard errors
#		2. Full parameter vector
#           3. Model identification string
#		4. Number of signals
#		5. Matrix of fixed constraints
#		6. Matrix of equality constraints
#		7. Confidence level
#           8. Number of decimal places
#
# Output: px5 Matrix of parameters, standard errors and confidence limits
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Par.est.SDT <- function(Pars, full.par, Model.Id, cfg, fixed, ident, alpha, deci)
{
      qt.cfi <- qnorm((1-alpha)/2)
      cfi.lo <- Pars[,1] + Pars[,2]*qt.cfi
      cfi.up <- Pars[,1] - Pars[,2]*qt.cfi
            
      # Respect lower bounds on sd parameter
      if(Model.Id == "SDT" || Model.Id == "SDT.Robust")
           bd.lo <- Elim.par(SDT.par.bounds(full.par, cfg)$lower, fixed, ident)
      else
            bd.lo <- Elim.par(Gaussian.par.bounds(full.par, cfg)$lower, fixed, ident)
      cfi.lo <- ifelse(cfi.lo < bd.lo, bd.lo, cfi.lo)
       
        
      Pars <- cbind(Pars, cfi.lo, cfi.up)
      colnames(Pars) <- c("Value", "SE", paste("CFI-", alpha*100, "(Lower)", sep = ""), paste("CFI-", alpha*100, "(Upper)", sep = ""))
      round(Pars, deci)
}

# ------------------------------------------------------------------------------------------------------------------
# Generation of data labels in output
# Input:    cfg     = configuration information
#           n.dat   = length of the data vector
#
# Output:   Vector with Data names
# ------------------------------------------------------------------------------------------------------------------
SDT.Datalabels <- function(cfg, n.dat)
{
      n.cat <- n.dat / cfg$n.sdt          # Number of response categories
      label.name <- paste("Signal", 1:cfg$n.sdt)
      label.out <- NULL
      for(i in 1:cfg$n.sdt)
            label.out <- c(label.out, paste(label.name[i], " [", 1:n.cat, "]", sep = ""))
      return(label.out)
}

Gaussian.Datalabels <- function(cfg, n.dat)
{
      if(length(cfg$n.sdt) < 2)
            return(SDT.Datalabels(cfg, n.dat))
       
      label.name <- paste("Signal", 1:length(cfg$n.sdt))
      label.out <- NULL
      for(i in 1:length(cfg$n.sdt))
            label.out <- c(label.out, paste(label.name[i], " [", 1:cfg$n.sdt[i], "]", sep = ""))
      return(label.out)
}


#===============================================================================
#
# 6. Functions for computing additional measures of discriminability and bias 
#    for the SDT model only
#
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Computation of SDT measures
# Input:	1. Full parameter vector
#		2. Covariance matrix
#		3. Configuration list
#		4. Matrix of fixed constraints
#		5. Matrix of equality constraints
#           6. Function implmenting functional constraints
#           7. Number of decimal places
#		8. Confidence level
#
# Output: The follwong measures are computed:
#           d.a
#           d.e
#           A.z
# ------------------------------------------------------------------------------------------------------------------
SDT.measures <- function(full.par, Cov, cfg, fixed, ident, functional, deci, conf)
{
      n <- cfg$n.sdt
      # In case of functional constraints present and possible parameters in fixed 
      # and therefore in functional then doe not compute anything
      if(!is.null(functional) & any((1:(2*(n-1))) %in% fixed[2,]))
            return(NULL)

      qt.cfi <- qnorm((1-conf)/2)        

      #-----------------------------------------
      # Functions for computing d.a, d.e and A.z
      #-----------------------------------------
      SDT.d.a <- deriv(~mu*sqrt(2)/sqrt(1 + std^2), c("mu", "std"), function(mu, std) NULL)
      SDT.d.e <- deriv(~mu*2/(1 + std), c("mu", "std"), function(mu, std) NULL)
      SDT.A.z <- deriv(~pnorm(mu/sqrt(1 + std^2)), c("mu", "std"), function(mu, std) NULL)

      d.e.Erg <- d.a.Erg <- A.z.Erg <- NULL
      
      index.red <- Elim.par(1:length(full.par), fixed, ident)     # Positions of existing aprameters
      index     <- 1:length(index.red)

       CC.0 <- matrix(0, nr = 2, nc = 2)
      
      k <- 1
      for(sig in 1:(n-1))
      {
            d.a <- SDT.d.a(full.par[k], full.par[k+1])
            d.e <- SDT.d.e(full.par[k], full.par[k+1])
            A.z <- SDT.A.z(full.par[k], full.par[k+1])
            
            d.a.vec <- d.a
            d.e.vec <- d.e
            A.z.vec <- A.z
                         
            
            # Extract relevant part from covariance matrix
            if(!is.null(Cov))
            {
                  CC <- SDT.Compute.CC(sig, Cov, index.red, index, fixed, ident)

                  # Extract gradients
                  d.a.Grad <- matrix(attr(d.a, "gradient"), nc = 1)
                  d.e.Grad <- matrix(attr(d.e, "gradient"), nc = 1)
                  A.z.Grad <- matrix(attr(A.z, "gradient"), nc = 1)
            
                  # Delta method
                  s.d.a  <- sqrt(t(d.a.Grad) %*% CC %*% d.a.Grad)
                  s.d.e  <- sqrt(t(d.e.Grad) %*% CC %*% d.e.Grad)         
                  s.A.z  <- sqrt(t(A.z.Grad) %*% CC %*% A.z.Grad)
                  
                  # CI
                  d.a.lo <- d.a + s.d.a*qt.cfi
                  d.e.lo <- d.e + s.d.e*qt.cfi
                  A.z.lo <- max(A.z + s.A.z*qt.cfi, 0)
                  
                  d.a.up <- d.a - s.d.a*qt.cfi
                  d.e.up <- d.e - s.d.e*qt.cfi
                  A.z.up <- min(A.z - s.A.z*qt.cfi, 1)
                  
                  d.a.vec <- c(d.a.vec, s.d.a, d.a.lo, d.a.up)
                  d.e.vec <- c(d.e.vec, s.d.e, d.e.lo, d.e.up)
                  A.z.vec <- c(A.z.vec, s.A.z, A.z.lo, A.z.up)
            }
            
            d.a.Erg <- rbind(d.a.Erg, d.a.vec)
            d.e.Erg <- rbind(d.e.Erg, d.e.vec)
            A.z.Erg <- rbind(A.z.Erg, A.z.vec)
                  
            k <- k + 2
      }

      # Bind results together and provide row and column names
      Erg <- rbind(round(d.a.Erg, deci), round(d.e.Erg, deci), round(A.z.Erg, deci))
      rownames(Erg) <- c(paste("d.a[", 2:n, "]", sep=""), paste("d.e[", 2:n, "]", sep=""), paste("A.z[", 2:n, "]", sep=""))
      if(!is.null(Cov))
            colnames(Erg) <- c("Value", "SE", paste("CFI-", conf*100, "(Lower)", sep = ""), paste("CFI-", conf*100, "(Upper)", sep = ""))
      else
            colnames(Erg) <- c("Value")
      
      Erg            
}
# ------------------------------------------------------------------------------------------------------------------
# Computation of the relevant Covariance matrix for computing SDT measures
# Input:	1. Actual signal
#		2. Covariance matrix
#		3. Vector with numbers of free parameters
#		4. Index for addressing items in 3
#           5. Matrix iwth fixed constraints
#		6. Matrix of equality constraints
#
# Output: The relevant (part of the) covariance matrix
# ------------------------------------------------------------------------------------------------------------------
SDT.Compute.CC <- function(sig, Cov, index.red, index, fixed, ident)
{
      k <- 2*(sig-1) + 1
      C.rel <- k:(k+1)

      # Case 1: No fixed or equality constraints
      if(is.null(ident) & is.null(fixed))
            return(Cov[C.rel, C.rel])
      
      # Treat equality constraints
      id <- C.rel %in% ident[2,]
      if(any(id)) C.rel[id] <- ident[1, ident[2,] %in% C.rel]
      
      C.rel.sich <- C.rel
      C.rel <- index[index.red %in% C.rel]
      if(length(C.rel) == 2)                     # Both parameters are free
            return(Cov[C.rel, C.rel])
      if(length(C.rel) == 0)                     # Both parameters are fixed
            return(matrix(0, nr = 2, nc = 2))
 
      CC <- matrix(0, nr = 2, nc = 2)
      C.rel.t <- (1:2)[C.rel.sich %in% index.red]
      CC[C.rel.t, C.rel.t] <- Cov[C.rel, C.rel]

      CC     
}

#===============================================================================
#
# 7. Plot functions
#
#===============================================================================


# ------------------------------------------------------------------------------------------------------------------
# Plotting for the SDT and Gaussian models
# Input:	1. par.vec = vector of parameters
#		2. cfg     = configuration list
#           3. cols   = number of colors
#           4. ltys   = line types
#		5. labels = labels for axes
#           6. SDT.legend = list with tow entries: 
#                       pos = c(x, y) Position
#                       text = c("text 1",...,"text n")
#                       cols = colors
#           7. option:  NULL       --> SDT
#                       "ROC"       --> ROC
#                       "zROC"      --> zROC
#                       "ROC+"      --> ROC + Data
#                       "zROC+"     --> zROC + Data
#           8. counts:  data
#
# Comment:  The function extracts the decision bounds as well as the means and sd parameters and calls the funtion
#           SDT.SDT1D.Plot or the function SDT.ROC.Plot (in module SDT-Main.R)
# ------------------------------------------------------------------------------------------------------------------
SDT.SDT.Plot <- function(par.vec, cfg, cols, ltys, labels, SDT.legend, option, counts)
{
      n <- cfg$n.sdt
      par.vec <- c(0, 1, par.vec)
      t <- par.vec[(2*n + 1) : length(par.vec)]
      
      # Determine Range the Range of x
      m  <- par.vec[seq(1, 2*n-1, 2)]
      sd <- par.vec[seq(2, 2*n, 2)]
      
      if(is.null(option))
            SDT.SDT1D.Plot(m, sd, t, n, cols, ltys, labels, SDT.legend)
      else if(option == "ROC")
            SDT.ROC.Plot(m, sd, p = NULL, n, cols, ltys, labels, SDT.legend, zROC = F, counts = NULL)
      else if(option == "ROC+")
            SDT.ROC.Plot(m, sd, p = NULL, n, cols, ltys, labels, SDT.legend, zROC = F, counts = counts)
      else if(option == "zROC")
            SDT.ROC.Plot(m, sd, p = NULL, n, cols, ltys, labels, SDT.legend, zROC = T, counts = NULL)
      else if(option == "zROC+")
            SDT.ROC.Plot(m, sd, p = NULL, n, cols, ltys, labels, SDT.legend, zROC = T, counts = counts)
      else
            cat("\n\n --> SDT.SDT.Plot: Unknown option: \"", option, "\"\n\n", sep = "")
}
      
SDT.Gaussian.Plot <- function(par.vec, cfg, cols, ltys, labels, SDT.legend, option, counts)
{
      n <- cfg$n.sdt
      if(length(n) < 2)
      {
            par.matrix <- matrix(par.vec, nc = n)
            
            t <- as.vector(par.matrix[-c(1,2),])
            m  <- as.vector(par.matrix[1, ])
            sd <- as.vector(par.matrix[2, ])
      }
      else  # Unequal number of responses per signal
      {
            t <- m <- sd <- NULL
            k <- 1
            for(i in 1:length(n))
            {
            
                  m  <- c(m,  par.vec[k])
                  sd <- c(sd, par.vec[k+1])
                  t  <- c(t,  par.vec[(k+2):(k+n[i])])
                  k  <- k + 1 + n[i] 
            }
      }
      if(is.null(option))
            SDT.SDT1D.Plot(m, sd, t, n, cols, ltys, labels, SDT.legend)
      else if(option == "ROC")
            SDT.ROC.Plot(m, sd, p = NULL, n, cols, ltys, labels, SDT.legend, zROC = F, counts = NULL)
      else if(option == "ROC+")
            SDT.ROC.Plot(m, sd, p = NULL, n, cols, ltys, labels, SDT.legend, zROC = F, counts = counts)
      else if(option == "zROC")
            SDT.ROC.Plot(m, sd, p = NULL, n, cols, ltys, labels, SDT.legend, zROC = T, counts = NULL)
      else if(option == "zROC+")
            SDT.ROC.Plot(m, sd, p = NULL, n, cols, ltys, labels, SDT.legend, zROC = T, counts = counts)
      else
            cat("\n\n --> SDT.Gaussian.Plot: Unknown option: \"", option, "\"\n\n", sep = "")

}
 