# ===========================================================================
# SDT models: mAFC modell with bias
# Source: DeCarlo (2012)
#
# Date of creation:	October, 2012
# Date of last update:	October, 2012
#
# Author: Siegfried Macho
# ===========================================================================
library(statmod)
library(npmlreg)

mAFC.Tol  <- 1e-12
MINWEIGHT <- 1e-40

#===============================================================================
# 0. Computation of Quadrature points
#===============================================================================

#-------------------------------------------------------------------------------
# Function for computing the Gauss-Hermite points and weights
# Input:	1. Number of points
#           2. minimal weights
#
# Output: 	Matrix with three columns:    1. Points
#                                         2. Weights of Gauss-Hermite
#                                         3. Weights of adaptive Gauss-Hermite 
#
# Comment:  The points and weights are with respect to the Gaussian Kernel N(0, 1)
#-------------------------------------------------------------------------------
Generate.GH <- function(npt, minweight = MINWEIGHT)
{
      GH.Structure  <- gqz(numnodes = npt, minweight)   # Points and weights for Gauss-Hermite Quadrature
      cbind(GH.Structure, GH.Structure[,2] * exp(GH.Structure[,1]^2/2))	 # Weights for adaptive Gauss-Hermite Quadrature
}

#===============================================================================
# 1. Model function
#===============================================================================

# ---------------------------------------------------------------------------
# Function for computing the conditional probability of selecting Response i, 
# given a fixed residual error
# Input:	Vector of parameters: 
#           1. eps:            Error
#           2. act.pos:        Position whose probability is computed 
#           3. targtet.pos:    Position of the target
#           4. parvec:         Vector of parameters: [d', sigma, b[1], b[2], ...., b[m-1], 0]
#
# Output: 	Probability of selecting Item i
# ----------------------------------------------------------------------
SDT.mAFC.i.cond <- function(eps, act.pos, target.pos, parvec)
{
      bias             <- parvec[-(1:2)]                   # Extract bias parameters
      bias[target.pos] <- bias[target.pos] + parvec[1]     # Add d' to target position
      sigma            <- parvec[2]                        # Standard deviation 
      
      # Case 1: actual position corresponds to target position
      if(act.pos == target.pos)
      {
            act   <- bias[act.pos] + eps * sigma           # Extract actual evidence and bias
            other <- bias[-act.pos]                        # Values of other items
            return(prod(pnorm(act - other)) )
      }
      # Case2: actual position does not correspond to target position
      act   <- bias[act.pos] + eps                   # Extract actual evidence and bias
      other.target <- bias[target.pos]               # Bias of target item
      if(length(bias) < 3)                           # Two alternatives
            return( pnorm((act - other.target)/sigma))
      other <- bias[-c(act.pos, target.pos)]
      if(length(bias) < 4)                           # Three alternatives
            return( pnorm(act - other) * pnorm((act - other.target)/sigma))
      return(prod(pnorm(act - other)) * pnorm((act - other.target)/sigma) )
}	

# ------------------------------------------------------------------------------
# Functions required for computing Gaussian and adaptive Gaussian quadrature
# Input:	Vector of parameters: 
#           1. act.pos:     Position whose probability is computed 
#           2. targtet.pos: Position of the target
#           3. parvec:      Vector of parameters: [d', b[1], b[2], ...., b[m-1], 0]
#
# Output: 	Probability of item on act.pos i
# ------------------------------------------------------------------------------
SDT.mAFC.i.Cond.GH <- function(x, act.pos, target.pos, parvec)
{
      SDT.mAFC.i.cond(x[1], act.pos, target.pos, parvec) * x[2]
}


SDT.mAFC.i.Cond.aGH <- function(x, act.pos, target.pos, parvec, u.m, u.s)
{
      z <- u.m + u.s * x[1]
      SDT.mAFC.i.cond(z, act.pos, target.pos, parvec)*dnorm(z, 0, 1.0) * x[3]
}

# ------------------------------------------------------------------------------
# Function for computing the unconditional probability of selecting Response i
# given a fixed residual error
# Input:	Vector of parameters: 
#           1. act.pos:     Position whose probability is computed 
#           2. targtet.pos: Position of the target
#           3. parvec:      Vector of parameters: [d', b[1], b[2], ...., b[m-1], 0]
#
# Output: 	Probability of item on act.pos i
# ------------------------------------------------------------------------------
SDT.mAFC.i <- function(act.pos, target.pos, parvec)
{
      sum(apply(GH.Structure, 1, SDT.mAFC.i.Cond.GH, act.pos = act.pos, target.pos = target.pos, parvec = parvec))
}

SDT.mAFC.i.aGH <- function(act.pos, target.pos, parvec)
{
	g <- function(x, ...) { erg <- SDT.mAFC.i.cond(x, ...) * dnorm(x, 0.0, 1.0); if(erg > 0) return(-log(erg)); return(1e16) } 

      par = 0
	opti.obj <- nlm(g, par, hessian = T, act.pos = act.pos, target.pos = target.pos, parvec = parvec)
	u.m <- opti.obj$estimate
	u.s <- sqrt(1/opti.obj$hessian)

      sqrt(2*pi) * u.s * sum(apply(GH.Structure, 1, SDT.mAFC.i.Cond.aGH, act.pos = act.pos, target.pos = target.pos, parvec = parvec, u.m = u.m, u.s = u.s))
}


# ------------------------------------------------------------------------------
# Function for application to structure 
# Input:	1. x:      Vector with three components: 
#                       x[1] = actual.position that is evaluated
#                       x[2] = target.position where target is located
#                       x[3] = model   
#           2. parvec: Parameter vector (without the zero for the bias parameter of the last option)
#           3. plen:   Number of parameters per model
#
# Output: 	Probability of item on act.pos i
# ------------------------------------------------------------------------------
SDT.mAFC.i.All <- function(x, parvec, plen, quad)
{
      if(quad == "aGH")
            SDT.mAFC.i.aGH(act.pos = x[1], target.pos = x[2], c(parvec[(x[3]-1)*plen + 1:plen], 0))
      else
            SDT.mAFC.i(act.pos = x[1], target.pos = x[2], c(parvec[(x[3]-1)*plen + 1:plen], 0))
}      
# ------------------------------------------------------------------------------
# Function for computing the unconditional probability of response selection
# Input:	1. Vector of parameters
#           2. Number of models
#           3. fixed and ident (not used)
#
# Output: 	Probabilities of AFC items for the different target postions of the 
#           target item
# ------------------------------------------------------------------------------
SDT.mAFC <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
      plen <- length(parvec)/cfg$s
      m    <- cfg$m                 # Number of response options
                    
      x <- expand.grid(a = 1:(m-1), t = 1:m, s = 1:cfg$s) # target and actual position as well as model index
      Res <- apply(x, 1, SDT.mAFC.i.All, parvec = parvec, plen = plen, cfg$quad)     # Compute Results

      # Transform to matrix and add probabilities of last response option
      Res <- matrix(Res, nr = m-1)
      Res <- rbind(Res, 1.0 - colSums(Res))
      as.vector(Res)
}

#===============================================================================
# 2. Derivatives and Model Matrices
#===============================================================================

# ---------------------------------------------------------------------------
# Function for computing the derivatives with respect to d', s, b[1], b[m-1]
# given a fixed residual error
# Input:	Vector of parameters: 
#           1. eps:            Error
#           2. act.pos:        Position whose probability is computed 
#           3. targtet.pos:    Position of the target
#           4. parvec:         Vector of parameters: [d', sigma, b[1], b[2], ...., b[m-1], 0]
#
# Output: 	Probability of selecting Item i
# ----------------------------------------------------------------------
pd.product <- function(x, p, d, eps)
{
      prod(p[-x])*d[x]*eps
}            


SDT.mAFC.i.cond.J <- function(eps, act.pos, target.pos, parvec)
{
      bias             <- parvec[-(1:2)]                   # Extract bias parameters
      bias[target.pos] <- bias[target.pos] + parvec[1]     # Add d' to target position
      sigma            <- parvec[2]                        # Standard deviation 
      m                <- length(bias)
      k                <- m-1                              # Exclude last 
      
      dbias <- numeric(k)
      sq <- 1:k
      non.act.pos <- sq[sq != act.pos]
      if(act.pos <= k) take <- 1:(k-1)
      else             take <- sq
      
      # Case 1: actual position corresponds to target position
      if(act.pos == target.pos)
      {
            act   <- bias[act.pos] + eps * sigma           # Extract actual evidence and bias
            other <- bias[-act.pos]                        # Values of other items
            p <- pnorm(act - other)
            d <- dnorm(act - other)
            
            pdp <- sapply(sq, pd.product, p = p, d = d, eps = 1.0)
            dd.prime <- sum(pdp)
            ds       <- sum(sapply(sq, pd.product, p = p, d = d, eps = eps))

            if(act.pos <= k) dbias[act.pos] <- dd.prime
            dbias[non.act.pos] <- -pdp[take]
            return(c(dd.prime, ds, dbias))
      }
      # Case2: actual position does not correspond to target position

      act   <- bias[act.pos] + eps                   # Extract actual evidence and bias
      other.target <- bias[target.pos]               # Bias of target item
      if(length(bias) < 3)                           # Two alternatives
      {
            dd.prime <- -dnorm((act - other.target)/sigma)/sigma
            ds       <- -dnorm((act - other.target)/sigma)*(act - other.target)/sigma/sigma
            if(act.pos == 1) db <- -dd.prime
            else             db <-  dd.prime
            return(c(dd.prime, ds, db))
      }
 
      # Compute dervative for d' and sd
      other    <- bias[-c(act.pos, target.pos)]
      p        <- prod(pnorm(act - other))
      d.other  <- -dnorm((act - other.target)/sigma)
      dd.prime <-  p * d.other/sigma
      ds       <-  p * d.other * (act - other.target)/sigma/sigma
 
      # Compute derivatives for b[1], ..., b[m-1]
      v.sigma <- rep(1, m)
      v.sigma[target.pos] <- sigma
      v.sigma <- v.sigma[-act.pos]
      other               <- bias[-act.pos]
      p <- pnorm((act - other)/v.sigma)
      d <- dnorm((act - other)/v.sigma)/v.sigma
      
      pdp <- sapply(sq, pd.product, p = p, d = d, eps = 1.0)
      
      if(act.pos <= k)  dbias[act.pos] <- sum(pdp)
      dbias[non.act.pos] <- -pdp[take]
      return(c(dd.prime, ds, dbias))
}	

# ------------------------------------------------------------------------------
# Functions required for computing Gaussian and adaptive Gaussian quadrature
# for Jacobians
# Input:	Vector of parameters: 
#           1. act.pos:     Position whose probability is computed 
#           2. targtet.pos: Position of the target
#           3. parvec:      Vector of parameters: [d', b[1], b[2], ...., b[m-1], 0]
#
# Output: 	Probability of item on act.pos i
# ------------------------------------------------------------------------------
SDT.mAFC.i.Cond.GH.J <- function(x, act.pos, target.pos, parvec)
{
      SDT.mAFC.i.cond.J(x[1], act.pos, target.pos, parvec) * x[2]
}


SDT.mAFC.i.Cond.aGH.J <- function(x, act.pos, target.pos, parvec, u.m, u.s)
{
      z <- u.m + u.s * x[1]
      SDT.mAFC.i.cond.J(z, act.pos, target.pos, parvec)*dnorm(z, 0, 1.0) * x[3]
}


# ------------------------------------------------------------------------------
# Function for computing the unconditional probability of selecting Response i
# given a fixed residual error
# Input:	Vector of parameters: 
#           1. act.pos:     Position whose probability is computed 
#           2. targtet.pos: Position of the target
#           3. parvec:      Vector of parameters: [d', b[1], b[2], ...., b[m-1], 0]
#
# Output: 	Probability of item on act.pos i
# ------------------------------------------------------------------------------
SDT.mAFC.i.J <- function(act.pos, target.pos, parvec)
{
     rowSums(apply(GH.Structure, 1, SDT.mAFC.i.Cond.GH.J, act.pos = act.pos, target.pos = target.pos, parvec = parvec))
}

SDT.mAFC.i.aGH.J <- function(act.pos, target.pos, parvec)
{
	g <- function(x, ...) { erg <- SDT.mAFC.i.cond(x, ...) * dnorm(x, 0.0, 1.0); if(erg > 0) return(-log(erg)); return(1e16) } 

      par = 0
	opti.obj <- nlm(g, par, hessian = T, act.pos = act.pos, target.pos = target.pos, parvec = parvec)
	u.m <- opti.obj$estimate
	u.s <- sqrt(1/opti.obj$hessian)

      sqrt(2*pi) * u.s * rowSums(apply(GH.Structure, 1, SDT.mAFC.i.Cond.aGH.J, act.pos = act.pos, target.pos = target.pos, parvec = parvec, u.m = u.m, u.s = u.s))
}


# ------------------------------------------------------------------------------
# Function for application to structure 
# Input:	1. x:      Vector with three components: 
#                       x[1] = actual.position that is evaluated
#                       x[2] = target.position where target is located
#                       x[3] = model   
#           2. parvec: Parameter vector (without the zero for the bias parameter of the last option)
#           3. plen:   Number of parameters per model
#
# Output: 	Probability of item on act.pos i
# ------------------------------------------------------------------------------
SDT.mAFC.i.All.J <- function(x, parvec, quad)
{
      if(quad == "aGH")
            SDT.mAFC.i.aGH.J(act.pos = x[1], target.pos = x[2], c(parvec, 0))
      else
            SDT.mAFC.i.J(act.pos = x[1], target.pos = x[2], c(parvec, 0))
}      
# ------------------------------------------------------------------------------
# Function for computing the Jacobian of the unconditional probability of 
# response selection
# Input:	1. Vector of parameters
#           2. Number of models
#           3. fixed and ident (not used)
#
# Output: 	Probabilities of AFC items for the different target postions of the 
#           target item
# ------------------------------------------------------------------------------
SDT.mAFC.J <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
      plen <- length(parvec)/cfg$s
      m    <- cfg$m                 # Number of response options
      
      x <- cbind(1:(m-1), 1) # target and actual position as well as model index
      
      mAFC.J.act <-  mAFC.J <- J <- row.Filler <- col.Filler <- NULL
      Filler <- matrix(0, nr <- m*m, nc <- m+1)
      # Loop through the models
      for(i in 1:cfg$s)
      {
            par.model <- parvec[((i-1)*plen + 1:plen)]
            for(target in 1:m)
            {
                  mAFC.J.act <- apply(x, 1, SDT.mAFC.i.All.J, parvec = par.model, cfg$quad)  # Compute Results
                  if(target == 1)
                        mAFC.J <- cbind(mAFC.J.act, 0.0 - rowSums(mAFC.J.act))               # Rows must sum to zero
                  else
                        mAFC.J <- cbind(mAFC.J, mAFC.J.act, 0.0 - rowSums(mAFC.J.act))       # Rows must sum to zero
                  if(target < m)
                        x[,2] <- target+1
            }      
            J  <- rbind(cbind(J, col.Filler), cbind(row.Filler, t(mAFC.J)))
            if(i < cfg$s)
            {
                  x[,2]      <- 1
                  row.Filler <- cbind(row.Filler, Filler)
                  col.Filler <- rbind(col.Filler, Filler)
            }      
      }
      J      
}

# ------------------------------------------------------------------------------------------------------------------
# Gradient of logL for the mAFC model
# Input:	Same as for Likelihood function
#
# Output:	Gradient
# ------------------------------------------------------------------------------------------------------------------
SDT.mAFC.Grad <- function(par, data, Model, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
	parvec   <- Restore.par(par, pos, fixed, ident, functional)	# Restore full parameter vector
	gradient <- as.vector(SDT.logL.Jacobi(parvec, data, Model, n, pos, fixed, ident, functional) %*% SDT.mAFC.J(parvec, n)) 

	# Treat functional constraints
	if(!is.null(functional))
	{
		gr.fct <- attr(par, "SDT.gradient")		# Isolate gradient function
		if(is.null(gr.fct))	stop("\n\n SDT.mAFC.Grad: Error in computation of gradient: No gradient for functional constraints found:\n Set Parameter sym.gr = F in function SDT.Estimate !\n\n")
		gradient <- gr.fct(gradient)
	}	
	Elim.par.grad(gradient, fixed, ident)
}

# ------------------------------------------------------------------------------
# Analytical computation of the model matrix
# Input:    1. Full parameter vector
#           2. Number of signals
#           3. 2 x p - matrix of fixed constraints
#           4. 2 x q - matrix of identity constraints
#           5. Jacobian matrix of functional constraints (supplied by the user)
#
# Output:   Modelmatrix
#
# Comment:  The Modelmatrix of funtional constraints must be a full matrix that
#           is premultiplied by the Jacobian
# ------------------------------------------------------------------------------
SDT.mAFC.Model.Matrix <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)

      J   <- SDT.mAFC.J(full.par, cfg = cfg, fixed, ident)
      if(!is.null(ident))
      {
            J.Id <- diag(length(full.par))
            Index <- matrix(c(ident[2,], ident[1,]), nc = 2)
            J.Id[Index] <- 1
            J <- J %*% J.Id
      }
      
      # Treatment of functional constraints
      if(!is.null(functional))
      {
            J.fct <- attr(functional, "SDT.Jacobian")
            if(is.null(J.fct)) 
                  return(NULL)           # No Jacobian function supplied -> return (NULL)
            J <- J %*% J.fct(full.par)
      }
      J[, pos]
}

# ------------------------------------------------------------------------------
# Model incorporating restrictions
# The parameter passed to the function contains the free parameters only
# ------------------------------------------------------------------------------
SDT.mAFC.Full <- function(rest.par, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
      full.par <- Restore.par(rest.par, pos, fixed, ident, functional)	# Restore full parameter vector
      SDT.mAFC(full.par, n, fixed, ident)
}

# ------------------------------------------------------------------------------
# Numerical computation of the model matrix 
# ------------------------------------------------------------------------------
SDT.mAFC.Model.Matrix.Num <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)
      parvec <- Elim.par(full.par, fixed, ident)
      J <- jacobian(SDT.mAFC.Full, parvec, n = cfg, pos = pos, fixed = fixed, ident = ident, functional = functional)
}

#===============================================================================
# 3. Help functions
#===============================================================================

# ------------------------------------------------------------------------------
# Name of the model
# ------------------------------------------------------------------------------
SDT.mAFC.model.name <- function(cfg)
{
      if(is.null(cfg))    return("m-AFC Model with Bias, Assuming Gaussian Distributions")
      if(is.numeric(cfg)) return(paste("-AFC Model with Bias, Assuming Gaussian Distributions", cfg, sep = ""))
      
      mess <- paste(cfg$m, "-AFC Model with Bias, Assuming Gaussian Distributions", sep = "")
      if(!is.null(cfg$s))
            mess <- paste(mess, ", Number of Models: ", cfg$s, sep = "")
      if(!is.null(cfg$quad))
      {
            if(cfg$quad == "GH")
                  mess <- paste(mess, ", Method of quadrature: ", "Gauss-Hermite", sep = "")
            else if(cfg$quad == "aGH")
                  mess <- paste(mess, ", Method of quadrature: ", "Adaptive Gauss-Hermite", sep = "")
      }
      if(!is.null(cfg$npt))
            mess <- paste(mess, ", Number of quadrature points: ", cfg$npt, sep = "")
            
      return(mess)
}       

# ------------------------------------------------------------------------------
# Function for checking and completing the configuration
# Input:	cfg = configuration information
# Output: 	cfg = configuration information filled.
# ------------------------------------------------------------------------------
SDT.mAFC.Check.Fill.Cfg <- function(cfg) 
{
      if(is.null(cfg))       stop("\n\nFATAL ERROR: No configuration information for SDT.mAFC\n\n")
      if(is.numeric(cfg))    return(list(s = 1, m = cfg, npt = 25, quad = "aGH"))
      if(is.null(cfg$m))     stop("\n\nFATAL ERROR: Number of alternatives (m) not specified in configuration information for SDT.mAFC\n\n")
      if(is.null(cfg$npt))   cfg$npt <- 25
      if(is.null(cfg$quad))  cfg$quad <- "aGH"
      if(is.null(cfg$s))     cfg$s <- 1
      
      return(cfg)
}      

# ------------------------------------------------------------------------------
# Function for computing the startparameter
# Input:	cfg = configuration information
# Output: 	parameter vector (d1, sigma1, b1.1, ... b1.m-1, ... ds, sigmas, bs.1, ...bs.m-1)
# ------------------------------------------------------------------------------
SDT.mAFC.start.par <- function(cfg)
{
      d.prime <- 0.5
      sigma   <- 1
      b       <- rep(0, cfg$m-1)
      return(rep(c(d.prime, sigma, b), cfg$s))
}

# ------------------------------------------------------------------------------------------------------------------
# Generate strings with names of the parameters for the models
# Input:	cfg = Configuration information
# Output:   String with parameter names
# ------------------------------------------------------------------------------------------------------------------
SDT.mAFC.par.names <- function(cfg)
{
      if(cfg$s == 1) return(c("d'", "sd", paste("b", 1:(cfg$m-1), sep = "")))
      
      b <- paste("b", 1:(cfg$m-1), sep = "")
      paste(c("d'", "sd", b), "[", rep(1:cfg$s, each = cfg$m + 1), "]", sep = "")
}


# ------------------------------------------------------------------------------------------------------------------
# Checking data and parameter vector
# Input:	1. parvec  = Parameter vector
#           2. datavec = Data vector
#           3. cfg     = Configuration information
#
# Output:   Structure with Error and message
# ------------------------------------------------------------------------------------------------------------------
SDT.mAFC.check <- function(parvec, datavec, cfg)
{
      # Expected length
      exp.plen   <- cfg$s * (cfg$m + 1)
      exp.dlen   <- cfg$s * (cfg$m * cfg$m)  
      
      Msg <- "NOTE: Computation of parameter and data length based on given configuration"
	Erg <- matrix(c(length(parvec), length(datavec), exp.plen, exp.dlen), nc = 2)
	rownames(Erg) <- c("Parameters", "Data points")
	colnames(Erg) <- c("Found", "Expected")
	
	Error <- any(Erg[, 1] != Erg[, 2])
	list(Error = Error, Stat = Erg, Msg = Msg)

}

#===============================================================================
# 4. Function for output formating
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Format free parameters and compute confidence intervals
# Input:	1. px2 Matrix with parameters and standard errors
#		2. Full parameter vector
#		3. Matrix of fixed constraints
#		4. Matrix of equality constraints
#		5. Confidence level
#           6. Number of decimal places
#
# Output: px5 Matrix of parameters, standard errors and confidence limits
# ------------------------------------------------------------------------------------------------------------------
SDT.mAFC.Format.Par.est <- function(Pars, full.par, fixed, ident, alpha, deci)
{
      qt.cfi <- qnorm((1-alpha)/2)
      cfi.lo <- Pars[,1] + Pars[,2]*qt.cfi
      cfi.up <- Pars[,1] - Pars[,2]*qt.cfi
            
      Pars <- cbind(Pars, cfi.lo, cfi.up)
      colnames(Pars) <- c("Value", "SE", paste("CFI-", alpha*100, "(Lower)", sep = ""), paste("CFI-", alpha*100, "(Upper)", sep = ""))
      round(Pars, deci)
}
