# ===========================================================================
# SDT2D-MIX-3 model:  Bivariate signal detection model mixture model 
#                     Mixture of three distributions
# Date of creation:	May, 2012
# Date of last update:	May, 2012
#
# Author: Siegfried Macho
# ===========================================================================
library(mvtnorm)	# Load multivariate normal distribution library
library(ellipse)	# Load functions for computing confidence ellipses

SDT2D.MIX.3.Tol <- 1e-12

#===============================================================================
#
# I. Auxiliary functions
#
#===============================================================================
# ------------------------------------------------------------------------------
# Function for generating a matrix of the decision bounds
# Input:	1. parvec = parameter vector that contains the parameters in the 
#                       following order
#                       i.   For each distribution:     m1, m2, sd1, sd2, r
#                       ii.  Decision bounds on Dimension 1
#                       iii. Decision bounds on Dimension 2
#           2. s      = Number of distributions (signals)
#           3. k1     = Number of decision bounds on Dimension 1
#           4. k2     = Number of decision bounds on Dimension 2
#           5. ds     = Flag indicating violations of decisional separability
#                       ds = 0 --> No violation
#                       ds = 1 --> Violation on Dimension 1
#                       ds = 2 --> Violation on Dimension 2
#           6. robust = Flag inidicting whether decision bounds should be estimated robustly
#
# Output: 	A [n*(k1+1)*(k2+1) x 4] matrix with lower and upper decision bounds 
#           for each response region, repeated for each signal distribution. 
#           The entries within a row are:
#                       t[1] = lower bound on Dimension 1
#                       t[2] = upper bound on Dimension 1
#                       t[3] = lower bound on Dimension 2
#                       t[4] = upper bound on Dimension 2
#
# Comment:  nblock gibt die Anzahl der zu erzeugenden Blcke der Grsse fr ein Modell.
#           Im Normalfall ist nblock = s, aber fr MIXTUR ist nblock abhngig von Anzahl der 
#           Mixturen.
# ----------------------------------------------------------------------------------------
SDT2D.MIX.3.Create.Decision.bounds <- function(parvec, s, k1, k2, ds, robust)
{
      ppar <- parvec[1+6*(0:(3*s-1))]
      nblock <- sum(ppar > SDT2D.MIX.3.Tol)

      DB   <- SDT.SDT2D.Create.Decision.bounds(parvec, s, k1, k2, ds, robust, Gauss.par = 18)
      
      miss <- nblock - s
      if(miss <= 0) return(DB)
      
      # There are missing blocks --> Add them
      Add.DB <- DB[1:((k1+1)*(k2+1)), ]
      
      for(i in 1:miss)
            DB <- rbind(DB, Add.DB)
      return(DB)
}

# ------------------------------------------------------------------------------------------------------------------
# Computation of probabilities for probability parameters that are not fixed
# Input:  	1. Parameter vector
#           2. Configuration information (only the number of models is required)
#           3. Matrix of fixed constraints
#
# Output:   Parameter vector with raw probability parameters replaced by computed parameters
# ------------------------------------------------------------------------------------------------------------------
SDT2D.MIX.3.cmp.p <- function(parvec, cfg, fixed)
{
      MAX.P <- 700

      # 1. Extract information about free and fixed probability parameters
      pos.p <- 1+6*(0:(3*cfg$s-1))                          # Vector with positions of probability parameters

      if(is.null(fixed)) fixed.tmp <- NULL
      else               fixed.tmp <- fixed[, !is.na(fixed[1,]), drop = F]     # Eliminate fixed positions with value NA from fixed vector
      if(is.null(fixed.tmp))  free.p <- rep(T, length(pos.p))
      else                    free.p <- !(pos.p %in% fixed.tmp[2,])   # Vector with positions of free probability parameters

      ppar <- parvec[pos.p]                         # Extract probability parameters
      ppar[is.na(ppar)] <- 0                        # Replace NA by zeros
  
      # 2. Compute probabilites
      free.pm <- matrix(free.p, nc = cfg$s)                         # Matrix of free parameter indicator
      ppar.m  <- matrix(ppar, nc = cfg$s)                           # Matrix of probability parameters
		
	comp.p <- NULL
	for(i in 1:dim(ppar.m)[2])
	{
	    free.pv <- ppar.m[free.pm[,i], i]
	    if(length(free.pv) > 1) 
          {
                sum.notfixed <- sum(ppar.m[!free.pm[,i], i])
                pp <- ifelse(free.pv > MAX.P, 1.0, exp(free.pv))
                comp.p <- c(comp.p, (1 - sum.notfixed) * pp / sum(pp))
          }  
	    else if(length(free.pv) > 0) 
          {
                sum.notfixed <- sum(ppar.m[!free.pm[,i], i])
                comp.p <- c(comp.p, 1 - sum.notfixed)
          }  
	}  
      parvec[pos.p[free.p]] <- comp.p
      return(parvec)
}

# -----------------------------------------------------------------------------------------
# Function that adds the positions of redundant parameters to fixed constraints
# Input:	1. Parameter vector
#		2. Number of models
#		3. Matrix of fixed constraints
#		4. Matrix of equality constraints
#
# Output: 	Matrix of fixed constraints with position of redundant parameters 
#		added. The value of these positions ist NA
#
# Comments:	The last probability parameter per model not subjected to a fixed or identity
#	    	constraint is considered as redundant.
# -----------------------------------------------------------------------------------------
SDT2D.MIX.3.Redundant.to.Fixed <- function(parvec, n.s, fixed, ident)
{
      
      # Case 1: If there is no matrix of constraints then generate one
      if(is.null(fixed) & is.null(ident))
      {
            redundant.pos <- 13+18*0:(n.s-1)
            return(matrix(c(rep(NA, length(redundant.pos)), redundant.pos), nr = 2, byrow = T))
      }
      # Case 2: Add matrix of redundant constraints to fixed matrix
      pos.p  <- 1+6*(0:(3*n.s-1))                           # Vector with positions of probability parameters
      free.p <- rep(T, length(pos.p))
      if(!is.null(fixed))
            free.p <- free.p & !(pos.p %in% fixed[2,])      # Vector with flags indicating free probability parameters
      if(!is.null(ident))
            free.p <- free.p & !(pos.p %in% ident[2,])      

      pos.pm <- matrix(pos.p, nr = 3)
      free.pm <- matrix(free.p, nr = 3)      
      add.pos <- NULL
      for(i in 1:n.s)                                      # Find last free parameter (it is redundant)
      {
            x <- pos.pm[free.pm[,i], i]
            len <- length(x)
            if(len > 0) add.pos <- c(add.pos, x[len])
      }      
      if(is.null(add.pos)) 
            return(fixed)
      
      return(cbind(fixed, matrix(c(rep(NA, length(add.pos)), add.pos), nr = 2, byrow =T)))
}				

#===============================================================================
#
# II. Functions for computing the model probabilities
#
#===============================================================================

# ---------------------------------------------------------------------------------
# The Mixture Gaussian 2D-SDT Model with 3 possible Gaussians per model
# Input:	1. parvec = Parameter vector
#                       i.   For each model: p1, m11, m12, sd11, sd12, r1
#                                            p2, m21, m22, sd21, sd22, r2
#                                            p3, m31, m32, sd31, sd32, r3
#                       ii.  Decision bounds on Dimension 1
#                       iii. Decision bounds on Dimension 2
#
#           2. A configuration list with the following entries
#                 i.           n  = Number of signal distributions
#                 ii.          k1 = Number of decision bounds on Dimension 1
#                 iii.         k2 = Number of decision bounds on Dimension 2
#                 iv.          ds = Flag indicating the presence of violations of descision separability
#                                   ds = 0 --> No violation
#                                   ds = 1 --> Violation on Dimension 1
#                                   ds = 2 --> Violation on Dimension 2
#                 v.         PA   = Response assignment matrix: This matrix is premaultiplied with response vector
#                 vi. struct.zero = Vector of structural zeros
#
#           3. fixed: Fixed constraints (not used, but required for compatibility)
#           4. ident: Equality vonstraints  (not used, but required for compatibility)
#
# Output: 	Vector of length n*(k1 + 1)*(k2 + 1) containing the response probabilities
# ----------------------------------------------------------------------------------
SDT2D.MIX.3 <- function(parvec, cfg, fixed = NULL, ident = NULL)
{
      # 1. Extract data from command list
      s  <- cfg$s
      k1 <- cfg$k1
      k2 <- cfg$k2
      ds <- cfg$ds
      R  <- cfg$R
      struct.zero <- cfg$struct.zero
      ren <- cfg$ren
      conditional <- cfg$conditional
      robust <- cfg$robust
      if(is.null(conditional)) conditional <- F
      if(is.null(robust))      robust <- F
       
      # 2. Compute probabilites (must be executed before the creation of decision bounds)
      parvec <- SDT2D.MIX.3.cmp.p(parvec, cfg, fixed)

      # 3. Create decision bounds
      DB <- SDT2D.MIX.3.Create.Decision.bounds(parvec, s, k1, k2, ds, robust)
      
      # 4. Create parameter matrix
      PM  <- NULL
      p.off <- 1
      for(i in 1:s)
      {
            parvec.rep <- NULL
            for(j in 1:3)
            {
                  if(parvec[p.off] > SDT2D.MIX.3.Tol)                                                   # Ignore Gaussians with mixture probability = 0
                        parvec.rep <- c(parvec.rep, rep(parvec[(p.off+1):(p.off+5)], (k1+1) * (k2+1)))  # Extract parameters of Gaussian models without probabilities
                  p.off <- p.off + 6
            }            
            PM <- rbind(PM, matrix(parvec.rep, nc = 5, byrow = T))
      }
      # 5. Compute Gaussian estimates      
      p.est <- apply(cbind(PM, DB), 1, SDT2D.Prob, robust = robust)

      # 6. Integrate estimates from different Gaussians of the mixture
      p.est.mix <- NULL
      p.off <- est.off <- 1
      for(i in 1:s)
      {
            p.est.sum <- 0
            for(j in 1:3)
            {
                  if(parvec[p.off] > SDT2D.MIX.3.Tol)                                         # Ignore Gaussians with mixture probability = 0
                  {
                        est.end <- est.off+(k1+1)*(k2+1)-1
                        p.est.sum <- p.est.sum + parvec[p.off] * p.est[est.off:est.end]       # Add to mixture
                        est.off <- est.end + 1
                  }
                  p.off <- p.off + 6      
            }          
            p.est.mix <- c(p.est.mix, p.est.sum)  
      }
      p.est <- p.est.mix

      # 7. Treatment of structural zeros
      if(!is.null(struct.zero)) 
      {
            p.est[struct.zero] <- 0
            if(conditional)               # Renormalization of probabilities
            {
                  p.est <- matrix(p.est, nc = s)
                  csum  <- colSums(p.est)
                  p.est <- as.vector(t(apply(p.est, 1, function(x) { x / csum} )))
            }
      }
      
      
      # 8. Pool results
      if(!is.null(R)) p.est <- as.vector(R %*% p.est)
      p.est
}

#===============================================================================
#
# III. Functions for computing derivatives: Jacobians and gradients
#
#===============================================================================

#-------------------------------------------------------------------------------------------
# Jacobian matrix of the function that computes mixture probabilities
#
# Input:  1. Vector of probability parameters   (of length 3)
#         2. Vector of integer indicating, fixed probability parameters (of length 3)
#            0: free 
#            1: fixed
#            2: redundant
#
# Output: Jacobian matrix Gradient vector of the mixture model
#-------------------------------------------------------------------------------------------
SDT2D.MIX.3.Mixprob.Jacobi <- function(prob.par, is.fixed)
{
      # Case 1: Only free and redundant parameters
      if(is.null(is.fixed) || all(is.fixed == 0 | is.fixed == 2))
      {
            J.p <- -prob.par %*% t(prob.par)
            diag(J.p) <- diag(J.p) + prob.par
            J.p[, is.fixed == 2] <- 0                # Redundant parameters (is.fixed = 2) are pure functions of the other parameters
            return(J.p)
      }
      J.p <- matrix(0, 3, 3)

      # Case 2: All parameters fixed
      if(all(is.fixed == 1)) return(J.p)
      
      # Case 3: Some parameters are fixed
      res <- sum(prob.par[is.fixed != 1])

      pp <- prob.par[is.fixed != 1]/res
      J.teil.p <- -pp %*% t(pp)
      diag(J.teil.p) <- diag(J.teil.p) + pp
      J.teil.p <- J.teil.p * res
      
      J.p[is.fixed != 1, is.fixed != 1] <- J.teil.p
      J.p[, is.fixed == 2] <- 0                      # Redundant parameters (is.fixed = 2) are pure functions of the other parameters
      return(J.p)
}


#-------------------------------------------------------------------------------------------
# Symbolic gradient of the general mixture model of three bivariate normal distribution
#  with variable upper and lower bounds:
#
# Input:  1. A vector of length 22 with the following entries:
#            For each of the three Gaussian of the mixture
#            1. p  (Mixture probability of Gaussian)  NOT RAW PARAMETERS! 
#		 2. m1 (Mean on Dimension 1)
#		 3. m2 (Mean on Dimension 2)
#		 4. s1 (Standard deviation on Dimension 1)
#		 5. s2 (Standard deviation on Dimension 2)
#		 6. r  (Correlation)
#
#	      19. t1l (Lower decision bound on Dimension 1)
#	      20. t1u (Upper decision bound on Dimension 1)
#	      21. t2l (Lower decision bound on Dimension 2)
#	      22. t2u (Upper decision bound on Dimension 2)
#
#        2. A flag indicating robust estimation of thresholds
#        3. A vector inidcating which of the probabilities are fixed
#                 0 = free
#                 1 = fixed
#                 2 = redundant
#
# Output: Gradient vector of the mixture model
#
# Comment:  The function uses the follwing functions from the module SDT2D:
#           1. SDT2D.Prob() for computing Gaussian probabilities
#           2. SDT2D.Prob.Jacobi() to compute derivatives of the parameters 
#              of the bivariate Gaussian distribution as well as of the 
#              thresholds as well as the function
#-------------------------------------------------------------------------------------------
SDT2D.MIX.3.Prob.Grad <- function(parvec, robust, is.fixed)
{
      if(is.null(is.fixed)) is.fixed <- c(0, 0, 0)      
      dtheta <- NULL 
      dtaus <- rep(0, 4)
      for(i in 1:3)
      {
            parvec.actual <- parvec[c(2:6 + (i-1)*6, 19:22)]
            # 1. Derviative of probability
            if(is.fixed[i] != 1)    dtheta <- c(dtheta, SDT2D.Prob(parvec.actual, robust))
            else                    dtheta <- c(dtheta, 0)
            mix.p <- parvec[1+(i-1)*6]
            
            # 2. Derivative of Gaussian parameters and threshold parameters
            if(mix.p > 0) 
            {
                  
                  dt.Gauss <- mix.p * SDT2D.Prob.Jacobi(parvec.actual, robust)
                  dtaus <-  dtaus + dt.Gauss[6:9]
                  dtheta <- c(dtheta, dt.Gauss[1:5])
            }
            else
                  dtheta <- c(dtheta, rep(0, 5))
      }
      # Adapt for probability parameters
      ppos <- c(1, 7, 13)
      dtheta[ppos] <-  dtheta[ppos] %*% SDT2D.MIX.3.Mixprob.Jacobi(parvec[ppos], is.fixed)
      
      return (c(dtheta, dtaus))
}

    
#-------------------------------------------------------------------------------------------
# Jacobian matrix of SDT2D.MIX.3 model
# Input:    1, The parameter vector
#		2. List with configuration informtion
#		3. Matrix of fixed constraints         (NOT USED, for compaitibility purpose only)
#		4. Matrix of equality constraints      (NOT USED)
#
# Output: Jacobian matrix with parameters making up columns and probabilities making up rows
#
# Comment: The following function form the SDT2D Module are used
#           - SDT.SDT2D.Create.Decision.bounds()
#           - SDT2D.Index.fct.Jacobi()
#           - SDT2D.Jacobian.Robust()
#-------------------------------------------------------------------------------------------
SDT2D.MIX.3.Jacobi <- function(parvec, cfg, fixed = NULL)
{
      npar.model <- 18  # Number of parameters per model

      # 1. Extract data from command list
      s  <- cfg$s
      k1 <- cfg$k1
      k2 <- cfg$k2
      n1 <- k1 + 1
      n2 <- k2 + 1
      ds <- cfg$ds
      R  <- cfg$R
      struct.zero <- cfg$struct.zero
      ren <- cfg$ren
      conditional <- cfg$conditional
      robust      <- cfg$robust
      if(is.null(conditional)) conditional <- F
      if(is.null(robust))      robust      <- F

      # 2. Compute probabilites (must be executed before the creation of decision bounds)
      parvec <- SDT2D.MIX.3.cmp.p(parvec, cfg, fixed)

      # 3. Compute Decision bounds for one model
      parvec.DS <- c(1:npar.model, parvec[(npar.model*s + 1):length(parvec)])          # Create a Pseudo parametervector
      DB <- SDT.SDT2D.Create.Decision.bounds(parvec.DS, s = 1, k1, k2, ds, robust, Gauss.par = npar.model)
      nprob <- dim(DB)[1]           # Number of probabilities to be modeled
      
      # 4. Generate fixed vectors of positions fo fixed and redundant parameters
      fixed.without.NA <- NULL
      fixed.NA <- NULL
      if(!is.null(fixed))
      {
            fixed.without.NA <- fixed[2, !is.na(fixed[1,])]
            fixed.NA <- fixed[2, is.na(fixed[1,])]
      }      
      
      J.Model     <- NULL                       # The Jacobian of the model prameters
      J.t         <- NULL                       # The Jacobian of the threshold parameters
      J.filler.h  <- J.filler.v <- NULL         # Horizontal and vertical filler matrices
      J.filler.0  <- matrix(0, nr = nprob, nc = npar.model)       
          
      # 5. Now generate Jacobian matrix
      for(i in 1:s)
      {
            # 5.1 Extract parameters: p, m1, m2, s1, s2, r
            par.actual <- parvec[1:npar.model + npar.model*(i-1)]
            PM <- matrix(rep(par.actual, nprob), nc = npar.model, byrow = T)      
            
            # 5.2 Determine fixed.values of the probability parameters: 0: not fixed, 1: fixed, 2: redundant
            pos.ppar <- c(1, 7, 13)+18*(i-1)
            is.fixed <- as.integer(pos.ppar %in% fixed.without.NA)  
            is.redundant <- as.integer((pos.ppar %in% fixed.NA)*2)
            is.fixed <- is.fixed + is.redundant
            
            # 5.3 Compute Jacobian matrix of the model
            J.m <- apply(cbind(PM, DB), 1, SDT2D.MIX.3.Prob.Grad, robust = robust, is.fixed = is.fixed)

            # 5.4 Create the Model matrix without threshold these are treted separately
            J.Model <- rbind(cbind(J.Model, J.filler.v), cbind(J.filler.h, t(J.m[1:npar.model, ])))
            if(i < s) 
            {
                  J.filler.h <- cbind(J.filler.h, J.filler.0)
                  J.filler.v <- rbind(J.filler.v, J.filler.0)
            }            
            
            # 5.5 Compile the Jacobian of thresholds
            J.t <- cbind(J.t, J.m[19:22, ])
      }

      # 6. Generate the threshold matrix
      tlen <- k1 + k2
      if(ds == 0)
      {
            Index.set <- cbind(rep(rep(c(0, 1:k1), each = n2), s),
                               rep(rep(c(1:k1, 0), each = n2), s),
                               rep(rep(c(0, 1:k2+k1), n1), s),
                               rep(rep(c(1:k2+k1, 0), n1), s)
                              )
            Working.array <- rbind(J.t, t(Index.set))
      }
      else if(ds == 1)
      {
            tlen <- k1*n2 + k2
 
            t1  <- matrix(1:(k1*n2), nr = k1)
            t1L <- as.vector(t(rbind(rep(0, n2), t1)))
            t1U <- as.vector(t(rbind(t1, rep(0, n2))))
            Index.set <- cbind(t1L, t1U, 
                               rep(rep(c(0, 1:k2 + k1*n2), n1), s),
                               rep(rep(c(1:k2 + k1*n2, 0), n1), s)
                              )
            Working.array <- rbind(J.t, t(Index.set))
   }
      
      else  # ds = 2
      {
            tlen <- k1 + k2*n1
            
            t2  <- matrix(1:(k2*n1)+ k1, nr = k2)
            t2L <- as.vector(rbind(rep(0, n1), t2))
            t2U <- as.vector(rbind(t2, rep(0, n1)))
      
            Index.set <- cbind(rep(rep(c(0, 1:k1), each = n2), s),
                               rep(rep(c(1:k1, 0), each = n2), s),
                               t2L, t2U)
            Working.array <- rbind(J.t, t(Index.set))
      }
      J.t <- apply(Working.array, 2, SDT2D.Index.fct.Jacobi, tlen = tlen)
      
      # 7. Adjustment of thresholds in case of robust estimation
      if(robust && (k1 > 1 || k2 > 1))
      {
            # Extract relevant columns representing thresholds from Jacobian matrix 
            if(ds == 0)
            { 
                  J.t1 <- J.t[1:k1, ]
                  J.t2 <- J.t[(k1+1):(k1+k2), ]
            }
            else if(ds == 1)
            { 
                  J.t1 <- J.t[1:(k1*n2), ]
                  J.t2 <- J.t[(k1*n2 + 1):(k1*n2 + k2), ]
            }
            else
            { 
                  J.t1 <- J.t[1:k1, ]
                  J.t2 <- J.t[(k1 + 1):(k1 + k2*n1), ]
            } 
                       
            # Adjust derivatives of thresholds
            if(k1 > 1)
            {     # Adjust thresholds on D1
                  if(ds == 0 || ds == 2)
                        t1 <- parvec[(s*npar.model + 1):(s*npar.model + k1)]
                  else 
                        t1 <- parvec[(s*npar.model + 1): (s*npar.model + k1*n2)]
                  if(ds == 0 || ds == 2)
                  {      
                        J.R <- SDT2D.Jacobian.Robust(t1, k1)    
                        J.t1 <- J.R %*% J.t1
                  }
                  else   # ds = 1
                  {
                        start <- 1
                        end   <- k1
                        for(i in 1:n2)
                        {
                              J.R <- SDT2D.Jacobian.Robust(t1[start:end], k1)    
                              J.t1[start:end,] <- J.R %*% J.t1[start:end, ]
                              start <- start + k1
                              end   <- end   + k1  
                        }     
                  }      
            }
            if(k2 > 1)
            {
                  # Adjust thresholds on D2
                  if(ds == 0)
                        t2 <- parvec[(s*npar.model + k1 + 1):(s*npar.model + k1 + k2)]
                  else if(ds == 1)
                        t2 <- parvec[(s*npar.model + k1*n2 + 1):(s*npar.model + n2*k1 + k2)]
                  else
                        t2 <- parvec[(s*npar.model + k1 + 1):(s*npar.model + k1 + n1*k2)]
                  if(ds == 0 || ds == 1)
                  {     
                        J.R <- SDT2D.Jacobian.Robust(t2, k2)    
                        J.t2 <- J.R %*% J.t2
                  }
                  else  # ds = 2
                  {
                        start <- 1
                        end   <- k2
                        for(i in 1:n1)
                        {
                              J.R <- SDT2D.Jacobian.Robust(t2[start:end], k2)    
                              J.t2[start:end,] <- J.R %*% J.t2[start:end, ]
                              start <- start + k2
                              end   <- end   + k2  
                        }     
                 
                  }
            }
            J.t  <- rbind(J.t1, J.t2)
      }
      
      J <- cbind(J.Model, t(J.t))

      # Treat structural zeros
      if(!is.null(cfg$struct.zero))
      {
            J[cfg$struct.zero,] <- 0
            if(conditional)  J <- SDT2D.MIX.3.Jacobi.conditional(J, parvec, cfg, fixed = NULL, ident = NULL)
       }
      
      # Treat grouping
      if(!is.null(cfg$R))
            J <- cfg$R %*% J
      J
}

# ------------------------------------------------------------------------------
# Analytical computation of the model matrix
# Input:    1. Full parameter vector
#           2. Number of signals
#           3. 2 x p - matrix of fixed constraints
#           4. 2 x q - matrix of identity constraints
#           5. Jacobian of functional constraints (supplied by the user)
#
# Output:   Modelmatrix
#
# Comment:  The Modelmatrix of funtional constraints must be a full matrix that
#           is premultiplied by the Jacobian
# ------------------------------------------------------------------------------
SDT2D.MIX.3.Model.Matrix <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)

      J   <- SDT2D.MIX.3.Jacobi(full.par, cfg = cfg, fixed = fixed)
      if(!is.null(ident))
      {
            J.Id <- diag(length(full.par))
            Index <- matrix(c(ident[2,], ident[1,]), nc = 2)
            J.Id[Index] <- 1
            J <- J %*% J.Id
      }
      
      # Treatment of funtional constraints
      if(!is.null(functional))
      {
            J.fct <- attr(functional, "SDT.Jacobian")
            if(is.null(J.fct)) 
                  return(NULL)           # No Jacobian function supplied -> return (NULL)
            J <- J %*% J.fct(full.par)
      }
      J[, pos]
}
# ------------------------------------------------------------------------------------------------------------------
# Adaptation of Modelmatrix (Jacobian matrix) in case of renormalization
# Input:	1. Jacobian = non-adjusted Jacobian
#           2. parvec   = vector of parameters
#           3. cfg      = configuration inforamtion
#           4. fixed (not used)
#           5. ident (not used)
#
# Output:	New jacobian
#
# Comment: The following function from the SDT2D module is used
#           SDT2D.normal.J()
# ------------------------------------------------------------------------------------------------------------------
SDT2D.MIX.3.Jacobi.conditional <- function(J, parvec, cfg, fixed = NULL, ident = NULL)
{
      cfg$conditional <- NULL
      cfg$R           <- NULL  
      s <- cfg$s
      
      p.est <- SDT2D.MIX.3(parvec, cfg, fixed, ident)   # Compute probabilities igonring conditionalization amd pooling

      J.cond <- NULL
      n.p <- dim(J)[1]/s      # Number of estimated probabilities per model
      for(i in 1:s)
      {
            range.act <- ((n.p)*(i-1) + 1) : (n.p*i)         # Indices of probabilities for actual model
            if(any(cfg$struct.zero %in% range.act))          # Structural zeros for actual model (signal) present
            {
                  p.act <- p.est[range.act]
                  J.cond <- SDT2D.normal.J(p.act)            # Construct Jacobian that represents the normalization        
                  J[range.act, ] <- J.cond %*% J[range.act, ]
            }
      }
      J
}

# ------------------------------------------------------------------------------------------------------------------
# Gradient of logL for the SDT2D.MIX.3 model
# Input:	Same as for Likelihood function
#
# Output:	Gradient
# ------------------------------------------------------------------------------------------------------------------
SDT2D.MIX.3.Grad <- function(par, data, Model, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
	par 	   <- Restore.par(par, pos, fixed, ident, functional)	# Restore full parameter vector
	gradient <- as.vector(SDT.logL.Jacobi(par, data, Model, n, pos, fixed, ident, functional) %*% SDT2D.MIX.3.Jacobi(par, n, fixed)) 
	
	# Treat functional constraints
	if(!is.null(functional))
	{
		gr.fct <- attr(par, "SDT.gradient")		# Isolate gradient function
		if(is.null(gr.fct))	stop("\n\n SDT2D.MIX.3.Grad: Error in computation of gradient: No gradient for functional constraints found:\n Set Parameter sym.gr = F in function SDT.Estimate !\n\n")
		gradient <- gr.fct(gradient, par)
	}	
	Elim.par.grad(gradient, fixed, ident)
}


# ------------------------------------------------------------------------------------------------------------------
# Computation of Delta rule for probability parameters
# Input:	1. Full parameter vector (with raw parameters)
#           2. Information about model configuration
#           3. Matrix of fixed constraints
#           4. Matrix of identity constraints
#
# Output:	Jacobian matrix for adaptation of covariance matrix for probability parameters
# ------------------------------------------------------------------------------------------------------------------
SDT2D.MIX.3.Delta <- function(full.parvec, cfg, fixed, ident)
{
      # 1. Compute probability parameters from raw parameters
      parvec <- SDT2D.MIX.3.cmp.p(full.parvec, cfg, fixed)
      
      # 2. Now adapt for each model the probabilities
      pnum  <- 1:length(full.parvec)    
      pnum  <- Elim.par(pnum, fixed, ident)                       # Isolate the positions of free parameters
      pos.p <- 1+6*(0:(3*cfg$s-1))                                # Vector with positions of probability parameters
      k <- 1
      
      # 3. Generate fixed vectors of positions fo fixed and redundant parameters
      fixed.without.NA <- NULL
      fixed.NA <- NULL
      if(!is.null(fixed))
      {
            fixed.without.NA <- fixed[2, !is.na(fixed[1,])]
            fixed.NA <- fixed[2, is.na(fixed[1,])]
      }      
      if(is.null(ident))
            ident.t <- NULL
      else 
            ident.t <- ident[2,]

      # 4. Adapt probability parameters for each model
      J.p   <- diag(length(full.parvec))                          # Generate Jacobian of adaptation
      for(i in 1:cfg$s)
      {
            # Determine fixed.values of the probability parameters: 0: not fixed, 1: fixed, 2: redundant
            pos.ppar <- pos.p[k:(k+2)]                               # Extract positions of probability parameters for actual model
            is.fixed <- as.integer(pos.ppar %in% fixed.without.NA)  
            is.redundant <- as.integer((pos.ppar %in% fixed.NA)*2)
            is.fixed <- is.fixed + is.redundant

            if(any(is.fixed != 1))                                         # Adaptation only if there is any free paraameter
                  J.p[pos.ppar, pos.ppar] <- SDT2D.MIX.3.Mixprob.Jacobi(parvec[pos.ppar], is.fixed)     # Compute Jacobian
            k <- k + 3
      }
      # 5. Adapt for identity constraints
      if(any(pos.p %in% ident.t))
      {
            J.Id <- diag(length(full.parvec))
            Id <- ident.t %in% pos.p
            J.Id[ident[1, Id], ident[2, Id]] <- 1 
            J.p <- J.Id %*% J.p
      }
      J.p <- J.p[pnum, pnum]
      return(J.p)                  
}

#===============================================================================
#
# IV. Functions computing starting parameters, and names of parameters
#
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Generate starting parameters due to data and number of signals
# Input:    cfg = list with configuration information
#			 
# Output:   Vector of starting parameter
# ------------------------------------------------------------------------------------------------------------------
SDT2D.MIX.3.start.par <- function(cfg)
{
      s      <- cfg$s
      k1     <- cfg$k1
      k2     <- cfg$k2
      ds     <- cfg$ds
      robust <- cfg$robust
      
      # Start value and increment for robust thresholds
      inc.r <- -1 
      s.r1   <- -0.5*exp(inc.r)*(k1-1)
      s.r2   <- -0.5*exp(inc.r)*(k2-1)
      
      # Start and end value for non-robust thresholds
      s.n <- -1
      s.e <-  1

      if(is.null(ds))         ds <- 0
      if(is.null(robust)) robust <- F
      
      # Generate parameter vector for Gaussian part
      Mix.part <- c(0,    0,    0, 1, 1, 0,        # p, m1, m2, sd1, sd2, r (Gauss 1)
                    0, -0.1, -0.1, 1, 1, 0,        # p, m1, m2, sd1, sd2, r (Gauss 2)
                    0,  0.1,  0.1, 1, 1, 0)        # p, m1, m2, sd1, sd2, r (Gauss 3)
                    
      
	# Specify Number of thresholds
	if(ds == 0)
      { 
            if(robust && k1 > 1)    T1 <- c(rep(Mix.part, s), c(s.r1, rep(inc.r, k1-1)))
            else                    T1 <- c(rep(Mix.part, s), seq(s.n, s.e, length = k1))
                  
            if(robust && k2 > 2)    c(T1, c(s.r2, rep(inc.r, k2-1)))
            else                    c(T1, seq(s.n, s.e, length = k2))
	}
      else if(ds == 1)
      {
            if(robust && k1 > 1)    T1 <- c(rep(Mix.part, s), rep(c(s.r1, rep(inc.r, k1-1)), k2+1))
            else                    T1 <- c(rep(Mix.part, s), rep(seq(s.n, s.e, length = k1), k2+1)) 
            
            
            if(robust && k2 > 2)    c(T1, c(s.r2, rep(inc.r, k2-1)))
            else                    c(T1, seq(s.n, s.e, length = k2))
      }
	else if(ds == 2)
      {
            if(robust && k1 > 1)    T1 <- c(rep(Mix.part, s), c(s.r1, rep(inc.r, k1-1)))
            else                    T1 <- c(rep(Mix.part, s), seq(s.n, s.e, length = k1))
            
            if(robust && k2 > 2)    c(T1, rep(c(s.r2, rep(inc.r, k2-1)), k1+1))
            else                    c(T1, rep(seq(s.n, s.e, length = k2), k1+1))
      }
      else  stop ("\n FATAL ERROR: Wrong value of flag ds \n")
}


# ------------------------------------------------------------------------------------------------------------------
# Generate strings with names of the parameters for the models
# Input:	cfg = list with configuration information
#
# Output:   String with parameter names
# ------------------------------------------------------------------------------------------------------------------
SDT2D.MIX.3.par.names <- function(cfg)
{
      s  <- cfg$s
      k1 <- cfg$k1
      k2 <- cfg$k2
      ds <- cfg$ds

     if(is.null(ds)) ds <- 0

     n.mix <- 3

	p    <- paste("p[" , 1:s, ".", sep = "")
      p    <- paste(rep(p, each = n.mix), 1:n.mix, "]", sep = "")
      m1   <- paste("Mean.x[" , 1:s, ".", sep = "")
      m1   <- paste(rep(m1, each = n.mix), 1:n.mix, "]", sep = "")
	m2   <- paste("Mean.y[" , 1:s, ".", sep = "")
	m2   <- paste(rep(m2, each = n.mix), 1:n.mix, "]", sep = "")
	sd1  <- paste("Stddev.x[", 1:s, ".", sep = "")
      sd1  <- paste(rep(sd1, each = n.mix), 1:n.mix, "]", sep = "")
      sd2  <- paste("Stddev.y[", 1:s, ".", sep = "")
      sd2  <- paste(rep(sd2, each = n.mix), 1:n.mix, "]", sep = "")
	r    <- paste("Corr[", 1:s, ".", sep = "") 
      r    <- paste(rep(r, each = n.mix), 1:n.mix, "]", sep = "")

	pname  <- c(p, m1, m2, sd1, sd2, r)

	# Transpose the matrix to get the right order
	index <- matrix(1:(6*s*n.mix), nc = s*n.mix, byrow = T)
	pname <- pname[as.vector(index)]
	
	if(ds == 0)
	{
	     t1   <- paste("t.D1-",  1:k1, sep = "")
	     t2   <- paste("t.D2-",  1:k2, sep = "")
      }
      else if(ds == 1)
      {
	     t1   <- paste("t.D1-",  1:k1, sep = "")
	     t1   <- paste(t1, "[D2 = ", rep(1:(k2+1), each = k1), "]", sep = "")
	     t2   <- paste("t.D2-",  1:k2, sep = "")
            
      }
      else if(ds == 2)
      {
	     t1   <- paste("t.D1-",  1:k1, sep = "")
	     t2   <- paste("t.D2-",  1:k2, sep = "")
	     t2   <- paste(t2, "[D1 = ", rep(1:(k1+1), each = k2), "]", sep = "")
      }
      else  stop ("\n FATAL ERROR: Wrong value of flag ds \n")
      
	c(pname, t1, t2)
}

# ------------------------------------------------------------------------------------------------------------------
# Names of the SDT2D.MIX.3-model
# ------------------------------------------------------------------------------------------------------------------
SDT2D.MIX.3.model.name	<- function(cfg)
{
      if(is.null(cfg$ds))     name <- "Bivariate Gaussian 3 Mixture model: No violations of decisional separability"
      else if(cfg$ds == 0)    name <- "Bivariate Gaussian 3 Mixture model: No violations of decisional separability"
      else if(cfg$ds == 1)    name <- "Bivariate Gaussian 3 Mixture model: Violation of decisional separability on Dimension 1"
      else if(cfg$ds == 2)    name <- "Bivariate Gaussian 3 Mixture model: Violation of decisional separability on Dimension 2"
      else name <- "Bivariate Gaussian 3 Mixture model: No information about violations of decisional separability"
      
      if(!is.null(cfg$standard))
            if(cfg$standard) name <- paste(name, "[STANDARD RESTRICTIONS]")
      if(is.null(cfg$standard))
            name <- paste(name, "[STANDARD RESTRICTIONS]")
      if(!is.null(cfg$stuct.zero))
            name <- paste(name, "[STRUCTURAL ZEROS]")
      if(!is.null(cfg$R))
            name <- paste(name, "[WITH POOLING MATRIX]")
      if(!is.null(cfg$robust)) 
            if(cfg$robust == T) name <- paste(name, "[ROBUST]")
      name
      
}

#===============================================================================
#
# V. Functions for Checking the setup of the SDT2D.MIX.3 model
#
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# SDT Model: Check number of parameters and data points
# ------------------------------------------------------------------------------------------------------------------
SDT2D.MIX.3.check <- function(parvec, data, cfg)
{
      s  <- cfg$s
      k1 <- cfg$k1
      k2 <- cfg$k2
      ds <- cfg$ds

      exp.d.len         <- s * (k1+1)*(k2+1)     # Expected number of data points
      exp.par.len       <- s * 18                # Expected number of parameters without thesholds
      
      if(ds == 0)       exp.par.len <- exp.par.len + k1 + k2
      else if(ds == 1)  exp.par.len <- exp.par.len + k1*(k2+1) + k2
      else              exp.par.len <- exp.par.len + k1 + k2*(k1+1)
      
      Msg <- "NOTE: Computation of parameter and data length based on given configuration"
	Erg <- matrix(c(length(parvec), length(data), exp.par.len, exp.d.len), nc = 2)
	rownames(Erg) <- c("Parameters", "Data points")
	colnames(Erg) <- c("Found", "Expected")
	
	Error <- any(Erg[, 1] != Erg[, 2])
	list(Error = Error, Stat = Erg, Msg = Msg)
}

#===============================================================================
#
# V. Formating functions of the SDT2D model
#
#===============================================================================

# ------------------------------------------------------------------------------------------------------------------
# Format full parameter vector for SDT and Gaussian model
# ------------------------------------------------------------------------------------------------------------------
SDT.Format.Par.SDT2D.MIX.3 <- function(full.par, n, deci)
{
      T.max <- 10										# Max number of bounds for displaying in a row 

      s  <- n$s
      ds <- n$ds
      k1 <- n$k1
      k2 <- n$k2
      n.par <- 18
      
      # Compute robust thresholds if required
      full.par <- SDT2D.robust.thresholds(full.par, n, Gauss.par = n.par)
	
      # Compute probabilities           
	full.par <- SDT2D.MIX.3.cmp.p(full.par, n, fixed)           
	           
	Gauss.Par <- round(matrix(full.par[1:(s*n.par)], nrow = n.par), deci)
	colnames(Gauss.Par) <- paste("Mixture-", 1:dim(Gauss.Par)[2], sep = "")
	rownames(Gauss.Par) <- c("p[1]", "Mean.x[1]", "Mean.y[1]", "Stddev.x[1]", "Stddev.y[1]", "Corr[1]",
                               "p[2]", "Mean.x[2]", "Mean.y[2]", "Stddev.x[2]", "Stddev.y[2]", "Corr[2]",
                               "p[3]", "Mean.x[3]", "Mean.y[3]", "Stddev.x[3]", "Stddev.y[3]", "Corr[3]")
	if(ds == 1)
	{
		n1 <- k1*(k2+1)
		bounds.1 		 <- round(matrix(full.par[(s*n.par+1):(s*n.par+n1)], ncol = k1, byrow = T), deci)
		rownames(bounds.1) <- paste("D2: t-", 1:(k2+1), sep = "")
		colnames(bounds.1) <- paste("D1: t-", 1:dim(bounds.1)[2], sep = "")

	}
	else
	{
		bounds.1 	       <- round(matrix(full.par[(s*n.par+1):(s*n.par+k1)], nrow = 1, byrow = T), deci)
		colnames(bounds.1) <- paste("t-", 1:k1, sep = "")
		rownames(bounds.1) <- "Dimension 1"
		if(dim(bounds.1)[2] > T.max) bounds.1 <- t(bounds.1)		# Too many bounds -> form row vector

	}
	if(ds == 2)
	{
		k2 <- (k1+1)*k2
		bounds.2 		 <- round(matrix(full.par[(s*n.par+k1+1):(s*n.par+k1+k2)], ncol = k1+1), deci)
		colnames(bounds.2) <- paste("D1: t-", 1:(k1+1), sep = "")
		rownames(bounds.2) <- paste("D2: t-", 1:dim(bounds.2)[1], sep = "")

	}
	else
	{
		bounds.2 	       <- round(matrix(full.par[(s*n.par+k1+1):(s*n.par+k1+k2)], nrow = 1, byrow = T), deci)
		colnames(bounds.2) <- paste("t-", 1:k2, sep = "")
		rownames(bounds.2) <- "Dimension 2"
		if(dim(bounds.2)[2] > T.max) bounds.2 <- t(bounds.2)		# Too many bounds -> form row vector
	}
		
      list(GP = Gauss.Par, DB.1 = bounds.1, DB.2 = bounds.2)

}

# ------------------------------------------------------------------------------------------------------------------
# Format free parameters and compute confidence intervals
# Input:	1. px2 Matrix with parameters and standard errors
#           3. Full parameter vector
#		2. Number of signals
#           3. Fixed constraints
#           4. Equality constraints
#		5. Confidence level
#           6. Number of decimal places
#
# Output: px5 Matrix of parameters, standard errors and confidence limits
# ------------------------------------------------------------------------------------------------------------------
SDT2D.MIX.3.Format.Par.est <- function(Pars, full.par, n, fixed, ident, alpha, deci)
{
      npar <- 18
      
      # Compute probabilities           
      full.par <- SDT2D.MIX.3.cmp.p(full.par, n, fixed)       
      Pars[,1] <- Elim.par(full.par, fixed, ident)    

	if(!is.null(n$robust))
	{     # Compute thresholds in case of robust thresholds
            if(n$robust) 
            {
                  full.par <- SDT2D.robust.thresholds(full.par, n, Gauss.par = npar)
                  Pars[,1] <- Elim.par(full.par, fixed, ident)
            }      
      }	           
      s <- n$s
      plen <- length(full.par)
      
      if(any(is.na(Pars[, 2]))) return(Pars)     # In case on NA no computation
      
      qt.cfi <- qnorm((1-alpha)/2)
      cfi.lo <- Pars[,1] + Pars[,2]*qt.cfi
      cfi.up <- Pars[,1] - Pars[,2]*qt.cfi
      
      # Respect lower and upper bounds on parameters
      lower <- c(rep(c(0, -Inf, -Inf,   0,   0, -1), 3*s), rep(-Inf, plen - npar*s))
      upper <- c(rep(c(1,  Inf,  Inf, Inf, Inf,  1), 3*s), rep( Inf, plen - npar*s))
      
      lower <- Elim.par(lower, fixed, ident)
      upper <- Elim.par(upper, fixed, ident)

      cfi.lo[cfi.lo < lower] <- lower[cfi.lo < lower]
      cfi.up[cfi.up > upper] <- upper[cfi.up > upper] 
            
      Pars <- cbind(Pars, cfi.lo, cfi.up)
	colnames(Pars) <- c("Value", "SE", paste("CFI-", alpha*100, "(Lower)", sep = ""), paste("CFI-", alpha*100, "(Upper)", sep = ""))
	
      round(Pars, deci)
}

#===============================================================================
#
# VI. Functions for computing teststatistics
#
#===============================================================================

#===============================================================================
#
# VII. Functions for computing the model matrix numerically
#
#===============================================================================

# ------------------------------------------------------------------------------
# Model incorporating restrictions
# The parameter passed to the function contains the free parameters only
# ------------------------------------------------------------------------------
SDT2D.MIX.3.Full <- function(rest.par, n, pos, fixed = NULL, ident = NULL, functional = NULL)
{
      full.par <- Restore.par(rest.par, pos, fixed, ident, functional)	# Restore full parameter vector
      SDT2D.MIX.3(full.par, n, fixed, ident)
}
 
# ------------------------------------------------------------------------------
# Numerical computation of the model matrix 
# ------------------------------------------------------------------------------
SDT2D.MIX.3.Model.Matrix.Num <- function(full.par, cfg, fixed = NULL, ident = NULL, functional = NULL)
{
      pos <- SavePos.par(full.par, fixed, ident)
      parvec <- Elim.par(full.par, fixed, ident)
      J <- jacobian(SDT2D.MIX.3.Full, parvec, n = cfg, pos = pos, fixed = fixed, ident = ident, functional = functional)
}



#===============================================================================
#
# VII. Plot function
#
#===============================================================================


# -------------------------------------------------------------------------------------------------------------
# Drawing the Gaussian Konfiguration of the SDT2D model
# Input:	1. par	 = the whole parameters vector
#		2. k		 = number of Gaussian models that enter a mixture model
#		4. n1  	 = number of decision criteria on Dimension 1
#		5. n2   	 = number of decision criteria on Dimension 2
#		6. ds    	 = Flag indicting violations of decisional spearability
#					0: No violations
#					1: Violations on D1
#					2: Violations on D2
#		7. cols	 = vector with color names or numbers for ellipses
#		8. ltys	 = vector with line types  for ellipses
#		9. labels    = character vector with names of labels for D1 and D2
#           10.option    = list of the number decision bounds to be left out
#
# Output: No sensible output
# -------------------------------------------------------------------------------------------------------------
SDT.SDT2D.Plot <- function(par, k, n1, n2, ds, cols = NULL, ltys = NULL, labels = NULL, option)
{
	n.par <- 5
	min.offset <- max.offset <- 0.0
	y.offset <- x.offset <- 5

	# Printing parameters
	col.c	<- "grey80"														# colors of decision bounds
	col.e <- c("black", "grey70", "blue", "red", "cyan", "darkgreen", "magenta", "yellow")		# colors of confidence ellipses
	lwd.c	<- 2															# line width of decision bounds
	lwd.e <- 2															# line width of confidence ellipse
	cex.e <- 1.5														# size of center of the ellipses
	sym.e <- 16															# symbol used for center of ellipses
		
	# If no color vector has been specified, create default
	if(is.null(cols))	cols <- rep(1, k)							

	# If no vector has line types been specified, create default vector
	if(is.null(ltys)) ltys <- rep(1, k)
	

	# If no labels have been specified, create default
	if(is.null(labels)) labels <- c("Dimension 1", "Dimension 2")

	# Generate a list with coordinates of Ellipses
	E <- NULL
	for(i in 1:k)
	{
		E[[i]] <- ellipse(par[5+(i-1)*n.par],  centre = par[1:2 +(i-1)*n.par],     scale = par[3:4 +(i-1)*n.par]  , level = .90)
		if(i == 1)
		{
			min.x <- min(E[[i]][,1])
			min.y <- min(E[[i]][,2])
			
			max.x <- max(E[[i]][,1])
			max.y <- max(E[[i]][,2])
		}
		else
		{
			min.x <- min(min.x, E[[i]][,1])
			min.y <- min(min.y, E[[i]][,2])

			max.x <- max(max.x, E[[i]][,1])
			max.y <- max(max.y, E[[i]][,2])
		}

	}
	# Generate empty plot
	min.x <- min.x
	min.y <- min.y
	max.x <- max.x
	max.y <- max.y

	# Identical scale on both Axes
	max.co <- max(max.x, max.y)
	min.co <- min(min.x, min.y)

	max.co <- max.co + max.offset
	min.co <- min.co - min.offset
	
	plot(c(min.co, max.co), c(min.co, max.co), type = "n", xlab = labels[1], ylab = labels[2])

      c.x <- c.y <- NULL
	# Plot decision criteria
	if(ds == 0)
	{
		c1 <- par[(n.par*k+1):(n.par*k+n1)]
		c2 <- par[(n.par*k+n1+1):(n.par*k+n1+n2)]
		
		# Compute coordinates for decision bounds on D1
		# The Result is a (n1 * (n2+1)) x 4 matrix (x1, y1, x2, y2) with start and end coordinates
		
		c.x <- cbind(c1, rep(min.y - y.offset, n1), c1, rep(c2[1], n1)) 
		if(n2 > 1)
		{
                  for(i in 2:n2)
		    {
		          actu <- cbind(c1, c2[i-1], c1, c2[i])
		          c.x <- rbind(c.x, actu)    
                  }
            }      
            c.x <- rbind(c.x, cbind(c1, rep(c2[n2], n1), c1, rep(max.y + y.offset, n1)))

		# Now the same for D2
		# The Result is a ((n1+1) * n2) x 4 matrix (x1, y1, x2, y2) with start and end coordinates
 		c.y <- cbind(c2, rep(min.x - x.offset, n2), c2, rep(c1[1], n2)) 
  		if(n1 > 1)
		{
                  for(i in 2:n1)
		    {
		          actu <- cbind(c2, c1[i-1], c2, c1[i])
		          c.y <- rbind(c.y, actu)    
                  }
            }      
            c.y <- rbind(c.y, cbind(c2, rep(c1[n1], n2), c2, rep(max.x + x.offset, n2)))
	}
	else if(ds == 1)
	{
		# Extract thresholds
		c1 <- par[(n.par*k+1):(n.par*k + n1*(n2+1))]
		c2 <- par[(n.par*k + n1*(n2+1) + 1):(n.par*k + n1*(n2+1) + n2)]

		# Compute coordinates for decision bounds on D1
		# The Result is a (n1 * (n2+1)) x 4 matrix (x1, y1, x2, y2) with start and end coordinates
		
		c.x <- cbind(c1[1:n1], rep(min.y - y.offset, n1), c1[1:n1], rep(c2[1], n1))
            dn <- n1 + 1 
		if(n2 > 1)
		{
                  for(i in 2:n2)
		    {
		          actu <- cbind(c1[dn:(dn+n1-1)], c2[i-1], c1[dn:(dn+n1-1)], c2[i])
		          c.x <- rbind(c.x, actu)    
		          dn <- dn + n1
                  }
            }      
            c.x <- rbind(c.x, cbind(c1[dn:(dn+n1-1)], rep(c2[n2], n1), c1[dn:(dn+n1-1)], rep(max.y + y.offset, n1)))

		# Now the same for D2
		# The Result is a ((n1+1) * n2) x 4 matrix (x1, y1, x2, y2) with start and end coordinates
 		c.y <- cbind(c2, rep(min.x - x.offset, n2), c2, rep(c1[1], n2)) 
  		if(n1 > 1)
		{
                  for(i in 2:n1)
		    {
		          actu <- cbind(c2, c1[i-1], c2, c1[i])
		          c.y <- rbind(c.y, actu)    
                  }
            }      
            c.y <- rbind(c.y, cbind(c2, rep(c1[n1], n2), c2, rep(max.x + x.offset, n2)))
	}
	else # ds = 2
	{
		c1 <- par[(n.par*k+1):(n.par*k + n1)]
		c2 <- par[(n.par*k + n1 + 1):(n.par*k + n1 + n2*(n1+1))]

		# Compute coordinates for decision bounds on D1
		# The Result is a (n1 * (n2+1)) x 4 matrix (x1, y1, x2, y2) with start and end coordinates
		
		c.x <- cbind(c1, rep(min.y - y.offset, n1), c1, rep(c2[1], n1)) 
		if(n2 > 1)
		{
                  for(i in 2:n2)
		    {
		          actu <- cbind(c1, c2[i-1], c1, c2[i])
		          c.x <- rbind(c.x, actu)    
                  }
            }      
            c.x <- rbind(c.x, cbind(c1, rep(c2[n2], n1), c1, rep(max.y + y.offset, n1)))


		# Now the same for D2
		# The Result is a ((n1+1) * n2) x 4 matrix (x1, y1, x2, y2) with start and end coordinates
		c.y <- cbind(c2[1:n2], rep(min.x - x.offset, n2), c2[1:n2], rep(c1[1], n2))
            dn <- n2 + 1 
		if(n1 > 1)
		{
                  for(i in 2:n1)
		    {
		          actu <- cbind(c2[dn:(dn+n2-1)], c1[i-1], c2[dn:(dn+n2-1)], c1[i])
		          c.y <- rbind(c.y, actu)    
		          dn <- dn + n2
                  }
            }      
            c.y <- rbind(c.y, cbind(c2[dn:(dn+n2-1)], rep(c1[n1], n2), c2[dn:(dn+n2-1)], rep(max.x + x.offset, n2)))
	}
      # Draw decision bounds
      ac <- 1
      for(i in 1:dim(c.x)[1])
      {
            if((ac %in% option) == F)
                  lines(c.x[i, c(1, 3)], c.x[i, c(2, 4)], col = col.c, lwd = lwd.c)
            ac <- ac + 1
      }      
      for(i in 1:dim(c.y)[1])
      {
            if((ac %in% option) == F)
                  lines(c.y[i, c(2, 4)], c.y[i, c(1, 3)], col = col.c, lwd = lwd.c)
            ac <- ac + 1
      }      

	# Draw centers of ellipes and ellipses
	for(i in 1:k)
	{
		points(par[1+(i-1)*n.par],  par[2+(i-1)*n.par],  pch = sym.e, col = col.e[cols[i]], cex = cex.e)	# Draw center
		lines(E[[i]], col = col.e[cols[i]], lty = ltys[i], lwd = lwd.e)
	}
} 